#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Script for building optiops: buildout, model libraries and SIMO db

import sys
sys.path.insert(0, 'dev')
import os
import platform
import shutil
import glob
import stat
from optparse import OptionParser
import conf
import re
import logging
logging.basicConfig(level=logging.INFO)

# set permissions to rwxr-xr-x (755)
PERM = stat.S_IRUSR|stat.S_IWUSR|stat.S_IXUSR\
       |stat.S_IRGRP|stat.S_IXGRP|stat.S_IROTH|stat.S_IXOTH

def build_model_libraries_and_simo_db(sys_id, dll_ini):
    """
    Builds model libraries specified in the dll ini file

    sys_id -- which platform are we on
    """
    py_exe = 'python'
    if sys_id != 'win':
        py_exe = 'bin/python'
    copied = _copy_temp_files()
    _compile_ext_et_models(py_exe, dll_ini)
    if not hasattr(conf, 'RUN_BUILDER') or conf.RUN_BUILDER:
        _run_builder(py_exe)
    else:
        print 'conf.RUN_BUILDER set to False, skipping builder'
    if copied:
        _remove_temp_files(copied)

    if sys_id != 'win':
        # Create redis default configuration
        with open('redis_default.conf', 'w') as f:
            f.write('bind 127.0.0.1\n')

        # Create a link to redis-server
        file = open('bin/redis-server', 'w')
        try:
            lines = []
            lines.append('#!/bin/sh\n')
            lines.append('parts/redis/src/redis-server redis_default.conf\n')
            file.writelines(lines)
        finally:
            file.close()
        # set permissions to rwxr-xr-x (755)
        os.chmod('bin/redis-server', PERM)

def copy_sources(what, source_dir, target_dir, exclude_dirs=[]):
    msg = '(Re)placing %s at %s' % (what, target_dir)
    logging.info(msg)
    if not os.path.exists(source_dir):
        msg = 'No such path %s' % os.path.abspath(source_dir)
        logging.error(msg)
        return
    if os.path.exists(target_dir):
        for entry in os.listdir(target_dir):
            full_entry = os.path.join(target_dir, entry)
            if entry in exclude_dirs + ['.svn']:
                continue
            elif os.path.isdir(full_entry):
                shutil.rmtree(full_entry)
            else:
                os.remove(full_entry)
    else:
        os.mkdir(target_dir)
    for item in os.walk(source_dir):
        skip_path = False
        path, folders, files = item
        for excl in exclude_dirs:
            if excl in path:
                skip_path = True
        if skip_path:
            continue
        targetpath = path.replace(source_dir, target_dir, 1)
        for folder in folders:
            if folder in exclude_dirs:
                continue
            if '.svn' in folder or '.svn' in path or \
               'egg-info' in folder or 'egg-info' in path:
                continue
            targetfolder = os.path.join(targetpath, folder)
            if not os.path.exists(targetfolder):
                os.mkdir(targetfolder)
        for file in files:
            if '.svn' in file or '.svn' in path or \
               'egg-info' in path or file.endswith('.pyc'):
                continue
            shutil.copyfile(os.path.join(path, file),
                            os.path.join(targetpath, file))

def merge_sources(what, source_dir, target_dir, exclude_dirs=[]):
    msg = '(Re)placing %s at %s' % (what, target_dir)
    logging.info(msg)
    if not os.path.exists(source_dir):
        msg = 'No such path %s' % os.path.abspath(source_dir)
        logging.error(msg)
        return
    if not os.path.exists(target_dir):
        os.mkdir(target_dir)
    for item in os.walk(source_dir):
        skip_path = False
        path, folders, files = item
        for excl in exclude_dirs:
            if excl in path:
                skip_path = True
        if skip_path:
            continue
        targetpath = path.replace(source_dir, target_dir, 1)
        for folder in folders:
            if folder in exclude_dirs:
                continue
            if '.svn' in folder or '.svn' in path or \
               'egg-info' in folder or 'egg-info' in path:
                continue
            targetfolder = os.path.join(targetpath, folder)
            if not os.path.exists(targetfolder):
                os.mkdir(targetfolder)
        for file in files:
            if '.svn' in file or '.svn' in path or \
               'egg-info' in path or file.endswith('.pyc'):
                continue
            shutil.copyfile(os.path.join(path, file),
                            os.path.join(targetpath, file))

def copy_postgis(sys_id):
    if sys_id == 'win':
        return
    if not os.path.exists('parts/postgresql'):
        msg = 'No postgresql installation in parts, moving on...'
        logging.info(msg)
        return
    libs = glob.glob('parts/geos/lib/*.so*')+glob.glob('parts/proj4/lib/*.so*')
    target = 'parts/postgresql/lib'
    for lib in libs:
        base = os.path.basename(lib)
        tgt = os.path.join(target, base)
        if os.path.exists(tgt):
            logging.info('Removing old %s' % tgt)
            os.remove(tgt)
        logging.info('Copying %s to %s' % (lib, tgt))
        shutil.copy2(lib, tgt)

def run_buildout(sys_id, buildout_option):
    """
    Executes buildout for the project (upgrades dependencies etc)

    sys_id -- which platform are we on
    buildout_option -- option to pass to buildout
    """
    if sys_id == 'dar':
        # Prevent crashing due to unused arguments.
        os.putenv('CFLAGS', '-Qunused-arguments')
        os.putenv('CPPFLAGS', '-Qunused-arguments')
        msg = 'building for mac'
        logging.info(msg)
        _init_buildout()
        plats = _osdetect()
        #if plats[0] == 'darwin-snowleopard-64':
        if plats[0] in ('10.6', '10.6-64'):
            msg = '  building for 64 bit snow leopard'
            logging.info(msg)
            if conf.EXTRA_CFG:
                extends = 'buildout_osx_10_6.cfg'
                _make_cfg(src=conf.X64_CFG, extends=extends)
                _buildout(os.path.basename(conf.X64_CFG), buildout_option)
            else:
                _buildout('buildout_osx_10_6.cfg', buildout_option)
        elif plats[0] in ('10.7', '10.7-64', '10.8', '10.8-64', '10.9', '10.9-64'):
            msg = '  building for 64 bit mountain lion'
            logging.info(msg)
            if conf.EXTRA_CFG:
                extends = 'buildout_osx_10_8.cfg'
                _make_cfg(src=conf.X64_CFG, extends=extends)
                _buildout(os.path.basename(conf.X64_CFG), buildout_option)
            else:
                _buildout('buildout_osx_10_8.cfg', buildout_option)
        elif len(u'\U00010800') == 1: #is ucs4
            msg = 'building with ucs4'
            logging.info(msg)
            if conf.EXTRA_CFG:
                extends = 'ucs4_buildout.cfg'
                _make_cfg(src=conf.CORE_CFG, extends=extends)
                _buildout(os.path.basename(conf.CORE_CFG), buildout_option)
            else:
                _buildout('ucs4_buildout.cfg', buildout_option)
        else: #is ucs2
            msg = 'building with ucs2'
            logging.info(msg)
            if conf.EXTRA_CFG:
                extends = 'buildout_osx.cfg'
                _make_cfg(src=conf.CORE_CFG, extends=extends)
                _buildout(os.path.basename(conf.CORE_CFG), buildout_option)
            else:
                _buildout('buildout_osx.cfg', buildout_option)
    elif sys_id == 'lin':
        msg = 'building for linux'
        logging.info(msg)
        _init_buildout()
        if len(u'\U00010800') == 1: #is ucs4
            msg = 'building with ucs4'
            logging.info(msg)
            if conf.EXTRA_CFG:
                extends = 'lin_ucs4_buildout.cfg'
                _make_cfg(src=conf.CORE_CFG, extends=extends)
                _buildout(os.path.basename(conf.CORE_CFG), buildout_option)
            else:
                _buildout('lin_ucs4_buildout.cfg', buildout_option)
        else: #is ucs2
            msg = 'building with ucs2'
            logging.info(msg)
            if conf.EXTRA_CFG:
                extends = 'lin_buildout.cfg'
                _make_cfg(src=conf.CORE_CFG, extends=extends)
                _buildout(os.path.basename(conf.CORE_CFG), buildout_option)
            else:
                _buildout('lin_buildout.cfg', buildout_option)

def _buildout(cfg_file, buildout_option):
    msg = 'making temporary copies of config files:'
    logging.info(msg)
    copied_files = []
    if conf.SIMO_ROOT:
        patt = '%s/dev/cfg/*.cfg' % conf.SIMO_ROOT
    else:
        patt = 'dev/cfg/*.cfg'
    if conf.EXTRA_CFG:
        patt2 = '%s/*.cfg' % conf.EXTRA_CFG
        files = glob.glob(patt) + glob.glob(patt2)
    else:
        files = glob.glob(patt)
    for file in files:
        t_file = os.path.basename(file)
        try:
            if not os.path.exists(t_file):
                shutil.copyfile(file, t_file)
                logging.info(msg)
                msg =  '  copied %s' % file
                copied_files.append(t_file)
            else:
                msg = '  already exists: %s, skipping...' % t_file
                logging.info(msg)
        except Exception, e:
            msg = '  copy failed! reason: %s' % str(e)
            logging.warn(msg)
    pathmod = 'PATH=$PATH:%s' % os.path.join(os.getcwd(), 'parts',
                                             'postgresql', 'bin')
    if buildout_option:
        buildout = pathmod + ' bin/buildout -' + buildout_option + \
                   ' -c ' + cfg_file
    else:
        buildout = pathmod + ' bin/buildout -c ' + cfg_file
    os.system(buildout)
    msg = 'removing temporary config files'
    logging.info(msg)
    for file in copied_files:
        try:
            os.remove(file)
            msg = '  removed %s' % file
            logging.info(msg)
        except Exception, e:
            msg = '  remove failed! reason: %s' % str(e)
            logging.warn(msg)

def _compile_ext_et_models(py_exe, dll_ini):
    msg = 'Compiling C extensions and model libraries'
    logging.info(msg)
    if not dll_ini:
        dll_ini=conf.DLL_INI
    if conf.SIMO_ROOT:
        r = '%s %s/dev/utils/build_libraries.py -i %s -r %s'\
            % (py_exe, conf.SIMO_ROOT, dll_ini, conf.SIMO_ROOT)
    else:
        r = '%s dev/utils/build_libraries.py -i %s'\
            % (py_exe, dll_ini)
    res = os.system(r)
    if res > 0:
        msg = 'Compilation failed!'
        logging.error(msg)
        sys.exit(1)
    print '\n' * 5
    print '*' * 20

def _copy_temp_files():
    msg = 'Copying the app specific models...'
    logging.info(msg)
    if conf.EXTRA_MODEL_XML_PATH == '':
        return []
    copied = []
    from_xml = os.path.join(conf.EXTRA_MODEL_XML_PATH, '*.xml')
    files = glob.glob(from_xml)
    if conf.SIMO_ROOT:
        to_xml = os.path.join(conf.SIMO_ROOT, 'simulator', 'xml',
                              'prediction_models')
    else:
        to_xml = os.path.join('simulator', 'xml', 'prediction_models')
    for f in files:
        shutil.copy2(f, to_xml)
        bf = os.path.basename(f)
        copied.append(os.path.join(to_xml, bf))
    from_c = os.path.join(conf.EXTRA_MODEL_CODE_PATH, '*.*')
    files = glob.glob(from_c)
    if conf.SIMO_ROOT:
        to_c = os.path.join(conf.SIMO_ROOT, 'src', 'models', 'Finland', 'c')
    else:
        to_c = os.path.join('src', 'models', 'Finland', 'c')
    for f in files:
        shutil.copy2(f, to_c)
        bf = os.path.basename(f)
        copied.append(os.path.join(to_c, bf))
    return copied

def _init_buildout():
    """
    If not done before, runs bootstrap for buildout
    """
    if not os.path.exists('bin/buildout'):
        if os.path.exists('tmp.cfg'):
            msg = 'Error: tmp.cfg found in root directory, please (re)move it.'
            logging.error(msg)
            sys.exit(1)
        msg = 'creating buildout script...'
        logging.info(msg)
        os.system('touch tmp.cfg')
        os.system(sys.executable + ' bootstrap.py -c tmp.cfg')
        os.system('rm tmp.cfg')
        msg = 'done'
        logging.info(msg)
    else:
        msg = 'buildout script found, moving along...'
        logging.info(msg)

def _make_cfg(src, extends):
    file = open(os.path.basename(src), 'w')
    try:
        file2 = open(src, 'r')
        try:
            source_str = file2.read()
            file.write(source_str % extends)
        finally:
            file2.close()
    finally:
        file.close()

def _osdetect():
    platforms = ['default']
    if sys.platform == 'darwin':
        platforms.insert(0, 'darwin')
        mac_ver = platform.mac_ver()
        mac_ver = re.match(r'^\d+\.\d+', mac_ver[0]).group()
        platforms.insert(0, mac_ver)
        if sys.maxint > 2147483647:
            platforms.insert(0, mac_ver + '-64')
    msg = "Detected these platforms: %s" % ", ".join(platforms)
    logging.info(msg)
    return platforms

def _remove_temp_files(copied):
    msg = 'Removing the app specific models from base SIMO:'
    logging.info(msg)
    for f in copied:
        msg = '    %s' % f
        logging.info(msg)
        os.remove(f)

def _run_builder(py_exe):
    for b_info in conf.BUILD_INIS:
        b_ini, extra_dbs = b_info
        r = ' '.join((py_exe, conf.BUILDER, 'build', b_ini))
        msg = 'Running builder: %s' % r
        logging.info(msg)
        res = os.system(r)
        if res != 0:
            msg = 'Running builder failed!'
            logging.error(msg)
            sys.exit(1)
        if len(extra_dbs) > 0:
            msg = 'Moving and copying databases...'
            logging.info(msg)
            simo_db = os.path.join(conf.DB_PATH, 'simo.db')
            for db_name in extra_dbs:
                db = os.path.join(conf.DB_PATH, db_name)
                shutil.copy2(simo_db, db)
                shutil.copy2(simo_db + '.zip', db + '.zip')
                shutil.copy2(simo_db + '.ver', db + '.ver')
            os.remove(simo_db)
            os.remove(simo_db + '.zip')
            os.remove(simo_db + '.ver')
    msg = 'Done'
    logging.info(msg)

def main(options):
    """
    Build optiops: buildout, model libraries, simo db

    options -- command line options
    """
    sys_id = str.lower(sys.platform[:3])
    if options.only_test:
        conf.run_output_comparison(sys_id)
    else:
        if not options.skip_buildout:
            run_buildout(sys_id, options.buildout_option)
        if 'do_extra_stuff_before_libraries' in dir(conf):
            conf.do_extra_stuff_before_libraries(sys_id)
        build_model_libraries_and_simo_db(sys_id, options.ini)
        if conf.CUSTOM_SIMULATOR:
            model_path = '%s/simulator/models' % conf.SIMO_ROOT
            to_path = '%s/models' % conf.CUSTOM_SIMULATOR
            copy_sources('the SIMO models', model_path, to_path,
                         exclude_dirs=conf.EXCLUDE_MODEL_DIRS)
            if hasattr(conf, 'OVERRIDE_MODEL_DIR'):
                model_path = conf.OVERRIDE_MODEL_DIR
                to_path = '%s/models' % conf.CUSTOM_SIMULATOR
                merge_sources('the OVERRIDING models', model_path, to_path)
            if hasattr(conf, 'EXTRA_MODEL_DIR') \
                                               and conf.EXTRA_MODEL_DIR != '':
                to_path = '%s/models/%s' \
                          % (conf.CUSTOM_SIMULATOR, conf.EXTRA_MODEL_DIR)
                copy_sources('the SIMO models', model_path, to_path,
                             exclude_dirs=conf.EXCLUDE_MODEL_DIRS)
        if conf.CUSTOM_SIMO:
            copy_sources('the SIMO source', 'simo/src/simo', conf.CUSTOM_SIMO)
        conf.do_extra_stuff(sys_id)
        copy_postgis(sys_id)
        if not options.no_test:
            conf.run_output_comparison(sys_id)

if __name__ == '__main__':
    usage = "usage: %prog [options]"
    parser = OptionParser(usage)
    parser.add_option('-b', '--buildout', dest='buildout_option',
                      default='',
                      help='option to pass to buildout run')
    parser.add_option('-i', '--ini', dest='ini', default='',
                      help='alternative ini file used to build libraries, '\
                           'default specified in dev/conf.py')
    parser.add_option('-s', '--skip_buildout', dest='skip_buildout',
                      action='store_true',
                      help='skip buildout run')
    parser.add_option('-n', '--no_test', action='store_true', dest='no_test',
                      help='skip running output tests')
    parser.add_option('-t', '--only_test', action='store_true',
                      dest='only_test', help='only run the output tests')
    (options, args) = parser.parse_args()
    main(options)
