#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# This script will compile C model libraries and generate their Python wrappers
# for SIMO

import sys
import os
from optparse import OptionParser
from subprocess import list2cmdline
import logging
logging.basicConfig(level=logging.INFO)

DLL_INI = 'dev/utils/dll.ini'

def compile(root, ini):
    """
    Compile C and Fortran libraries

    root -- root folder
    ini -- compilation instruction ini
    """
    py_exe = _get_python()

    root = os.path.abspath(root)
    msg = '  using root: %s' % root
    logging.info(msg)

    model_path = os.path.join(root, 'src', 'models')
    model_path = os.path.abspath(model_path)

    comp_info = (py_exe, root, model_path, ini)
    _compile_C_extensions(comp_info)
    _compile_C_models(comp_info)
    _compile_Fortran_models(comp_info)
    _generate_Python_wrappers(comp_info)
    msg = 'Building libraries done.'
    logging.info(msg)

def _compile_C_extensions(comp_info):
    py_exe, root, model_path, dll_ini = comp_info
    cwd = os.getcwd()
    os.chdir(root)
    cmd_path = 'dev/utils/cextensions.py'
    ccmd = list2cmdline((py_exe, cmd_path, 'build_ext'))
    os.system(ccmd)
    os.chdir(cwd)

def _compile_C_models(comp_info):
    py_exe, root, model_path, dll_ini = comp_info
    cmd_path = os.path.join(root, 'dev/utils/dll_build.py')
    dllcmd = list2cmdline((py_exe, cmd_path, root, model_path, py_exe,
                           dll_ini))
    os.system(dllcmd)

def _compile_Fortran_models(comp_info):
    py_exe, root, model_path, dll_ini = comp_info
    cmd_path = os.path.join(root, 'dev/utils/fortran_build.py')
    fortcmd = list2cmdline((py_exe, cmd_path, root, model_path, dll_ini))
    try:
        os.system(fortcmd)
    except Exception, e:
        msg = 'Something failed while building Fortran libraries'
        if sys.version_info[:3] >= (2,6,0):
            msg += ' likely due to bugs in f2py for python2.6'
            msg += '\n' + str(e)
            logging.error(msg)
        else:
            raise

def _generate_Python_wrappers(comp_info):
    msg = 'Building Python wrappers'
    logging.info(msg)
    py_exe, root, model_path, dll_ini = comp_info
    cwd = os.getcwd()
    os.chdir(root)
    sys_id = str.lower(sys.platform[:3])
    if sys_id == 'win':
        list_sep = ';'
    else:
        list_sep = ':'
    cmd_path = os.path.join(root, 'dev/utils/wrapperscript.py')
    wrapcmd = list2cmdline((py_exe, cmd_path, py_exe, root, model_path,
                            dll_ini))
    oldpath = os.getenv('PYTHONPATH') or ''
    try:
        msg = "Modifying PYTHONPATH"
        logging.info(msg)
        os.environ['PYTHONPATH'] = list_sep.join((oldpath, 'src'))
        os.system(wrapcmd)
    except Exception, e:
        msg = 'Something failed while generating Python model wrappers'
        msg += '\n' + str(e)
        logging.error(msg)
    finally:
        msg = "Restoring PYTHONPATH"
        logging.info(msg)
        os.environ['PYTHONPATH'] = oldpath
    os.chdir(cwd)

def _get_python():
    """
    Which Python interpreter are we going to use
    """
    sys_id = str.lower(sys.platform[:3])
    if sys_id == 'win':
        py_exe = sys.executable
        from win32api import GetShortPathName
        py_exe = GetShortPathName(py_exe)
    else:
        try:
            scripts = os.listdir('bin')
            for file in scripts:
                if file.startswith('python'):
                    py_exe = os.path.join('bin', 'python')
                    py_exe = os.path.abspath(py_exe)
                    break
            else:
                py_exe = sys.executable
                py_exe = os.path.abspath(py_exe)
        except:
            py_exe = sys.executable
            if sys_id == 'win':
                from win32api import GetShortPathName
                py_exe = GetShortPathName(py_exe)
            else:
                py_exe = os.path.abspath(py_exe)
    msg = "  using Python: %s" % py_exe
    logging.info(msg)
    return py_exe

def main(options):
    """
    Compile C models and generate wrappers
    """
    if options.root:
        root = options.root
    else:
        root = os.getcwd()
    if options.ini:
        dll_ini = options.ini
    else:
        dll_ini = DLL_INI
    if not os.path.exists(dll_ini):
        msg = 'The given DLL %s does not exist!' % (dll_ini,)
        logging.error(msg)
        raise Exception(msg)
    msg = 'Building models with root: %s, and dll ini: %s' % (root, dll_ini)
    logging.info(msg)
    compile(root, dll_ini)

if __name__ == '__main__':
    usage = "usage: %prog [options]\n"
    parser = OptionParser(usage)
    parser.add_option('-i', '--dll_ini', dest='ini',
                      help='model dll ini file path. Ini file defines the '\
                           'models to compile')
    parser.add_option('-r', '--root', dest='root',
                      help='the root folder for the model build run')
    (options, args) = parser.parse_args()
    main(options)
