#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# For building the SIMO C extensions
#
# © 2009 Simosol Oy

import sys
import os
import tempfile
import shutil
from distutils.core import setup
from distutils.dir_util import remove_tree
#from distutils.extension import Extension
from Cython.Distutils.extension import Extension
from Cython.Distutils import build_ext

EXT_DIR = 'src/cextensions'
OP_EXT_TARGET = os.path.join('simulator', 'models', 'operation')
HANDLER_EXT_TARGET = os.path.join('src', 'simo', 'matrix')
OPT_EXT_TARGET = os.path.join('src', 'simo', 'optimization')
OUT_EXT_TARGET = os.path.join('src', 'simo', 'output')

def get_numpy():
    """
    we need numpy's path explicitly for this, so we're locating it here.
    """
    numpy_path = None
    base = os.getcwd()
    ver = ''
    try:
        for i in os.listdir('eggs'):
            if i.startswith('numpy'):
                this_ver = i.split('-')[1]
                if this_ver > ver:
                    numpy_path = os.path.join(base, 'eggs', i)
                    ver = this_ver
    except OSError:
        import numpy
        dirname = os.path.dirname
        numpy_path = dirname(dirname(numpy.__file__))
    if numpy_path is None:
        import numpy
        dirname = os.path.dirname
        numpy_path = dirname(dirname(numpy.__file__))
    print 'picked Numpy: %s' % numpy_path
    return base, numpy_path

def build(numpy_path):
    """
    Compile extension modules
    """
    print 'proceeding to build the c extensions...'
    os.chdir(EXT_DIR)
    tmp_dir = tempfile.mkdtemp()
    sys_id = str.lower(sys.platform[:3])
    if sys_id == 'win' and sys.version_info[0] == 2 and sys.version_info[1] <= 5:
        # Windows (with python 2.5) must compile using mingw32,
        # so we'll need some additional parameters
        script_args = ['build', '-c', 'mingw32', '--build-base', tmp_dir,
                       'build_ext', '--force', '--inplace']
    else:
        script_args = ['build', '--build-base', tmp_dir,
                       'build_ext', '--force', '--inplace']
    include_dirs = ['%s/numpy/core/include' % (numpy_path,)]

    try:
        # compile the operation model C extensions
        setup(name="simo_c_extensions",
              description = "C extension modules for SIMO operation libraries",
              script_args = script_args,
              ext_modules=[Extension("operations_ext",
                           sources=["operations_ext.c"],
                           include_dirs=include_dirs,)
                          ],
             )
        # build the data handler C extension source files from the Cython
        # files and compile the data handler extensions
        setup(name = 'handler_ext',
              description = "Cythonized C extensions for SIMO data handler",
              script_args = script_args,
              cmdclass = {'build_ext': build_ext},
              ext_modules = [Extension('handler_ext', ['handler_ext.pyx'],
                             include_dirs=include_dirs)
                            ],
             )
        # build the data handler C extension source files from the Cython
        # files and compile the data handler extensions
        setup(name = 'opt_ext',
              description = "Cythonized C extensions for SIMO optimizer",
              script_args = script_args,
              cmdclass = {'build_ext': build_ext},
              ext_modules = [Extension('opt_ext', ['opt_ext.pyx'],
                             include_dirs=include_dirs)
                            ],
             )
        # build the output C extension source files from the Cython
        # files and compile the output extensions
        setup(name = 'out_ext',
              description = "Cythonized C extensions for SIMO output",
              script_args = script_args,
              cmdclass = {'build_ext': build_ext},
              ext_modules = [Extension('out_ext', ['out_ext.pyx'],
                             include_dirs=include_dirs)
                            ],
             )
        
        print 'build complete'
    except Exception, e:
        print e
        return False
    finally:
        remove_tree(tmp_dir)
    return True

def copy_ext(base):
    """
    Copy the built .so files to their target directories.
    """
    sys_id = str.lower(sys.platform[:3])
    if sys_id == 'win':
        fend = '.pyd'
    else:
        fend = '.so'
    # define the extension library names and target directories
    extlibs = ['operations_ext',
               'handler_ext',
               'opt_ext',
               'out_ext']
    targetdirs = [os.path.join(base, OP_EXT_TARGET),
                  os.path.join(base, HANDLER_EXT_TARGET),
                  os.path.join(base, OPT_EXT_TARGET),
                  os.path.join(base, OUT_EXT_TARGET)]
    for i in range(len(extlibs)):
        fromp = os.path.join(extlibs[i] + fend)
        targetdir = targetdirs[i]
        print 'copying:', fromp, 'to', targetdir
        shutil.copy2(fromp, targetdir)
    print '##################################################################'
    print 'C extensions compiled and copied successfully'
    print '##################################################################'


def main():
    base, np_path = get_numpy()
    success = build(np_path)
    if success:
        copy_ext(base)
    else:
        print '###############################################################'
        print 'C extensions build FAILED'
        print '###############################################################'

if __name__=='__main__':
    main()
