#!/usr/bin/env python
# encoding: utf-8
"""
A test module that will compare two output files to see if their contents match

*compare.py*

Usage: compare.py <file1> <file2> <filetype>

Copyright (c) 2010 Simosol Oy.
Distributed under the GNU General Public License 2.0.
"""

import sys
import os
import re
import time
from operator import itemgetter
from difflib import context_diff, unified_diff

EPSILON= 1.0 / (10.0**5) # percentage difference allowed
AGGR_EPSILON = 1.0 / (10.0**5) # percentage difference allowed
ABS_EPSILON = 0.01 # absolute difference allowed (one of these must pass)

def _itemgetter2(*items):
    if len(items) == 1:
        item = items[0]
        def g(obj):
            return obj[item][1]
    else:
        def g(obj):
            return tuple(obj[item][1] for item in items)
    return g

def _pad(vals, length=10):
    vals = [v.rjust(length) for v in vals]
    return vals


def _group_key_values(data):
    keys = []
    results = []
    vals = {}
    for key, val in data:
        if key not in keys:
            keys.append(key)
            vals[key] = []
        vals[key].append(val)
    for key in keys:
        results.append((key, tuple(vals[key])))
    return results

class Comparisor(object):

    def _diff_content(self, key, res, new_res):
        """
        Diff key results and output any differences found
        """
        base = res[key]
        now = new_res[key]
        if base == now:
            return None
        if base[0][0] != now[0][0]:
            print "'%s' has different results" % key
            print 'Key mismatch!'
            print 'exp:', base[0][0]
            print 'got:', now[0][0]
            self.mismatch = True
            return None
        first_diff = True
        keys = [b.strip() for b in base[0][0].split(';')]
        # Compare the data lines
        for i, row in enumerate(base[0][1:]):
            row2 = now[0][1:][i]
            if row != row2:
                rs = [r.strip() for r in row.split(';')]
                rs2 = [r.strip() for r in row2.split(';')]
                diff = False
                for j, val in enumerate(rs):
                    try:
                        diff = abs(float(val) - float(rs2[j])) >= ABS_EPSILON
                        if diff:
                            diff = abs(1.0 - (float(val) / float(rs2[j])))
                            diff = diff >= AGGR_EPSILON
                    except:
                        diff = val != rs2[j]
                    if diff:
                        self.mismatch = True
                        break
                if diff:
                    if first_diff:
                        first_diff = False
                        print key
                        print '    ',base[0][0]
                    print 'exp:', row
                    print 'got:', row2
        # Compare the error lines
        for line in unified_diff(base[1], now[1], fromfile='base',
                                 tofile='now', n=0, lineterm=''):
            print line
            self.mismatch = True


    def _parse_aggr(self, filename):
        f = open(filename, 'rU')
        try:
            lines = f.readlines()
        finally:
            f.close()
        text = ''.join(lines)
        patt = '([A-Za-z]+ for expression \([0-9]+ keys\):.*)'
        results = {}
        m = re.split(patt, text)
        if len(m) > 0:
            if m[0] == '':
                m = m[1:]
        for i in range(0, len(m), 2):
            results[m[i]] = [r.strip('\n\r') for r in m[i+1].split('\n') \
                             if r.strip('\n\r') != '']
        patt = '^Errors \(key: .*\):'
        # find where the error rows begin
        for key, val in results.items():
            for i, row in enumerate(val):
                if re.search(patt, row):
                    break
            else:
                # no errors, move the index to the end 
                # so the error list will be empty
                i = len(val)
            results[key] = (val[:i], val[i:])
        return results

    def _compare_aggr(self, filename1, filename2):
        r_base = self._parse_aggr(filename1)
        r_new = self._parse_aggr(filename2)
        keys_base = set(r_base.keys())
        keys_new = set(r_new.keys())

        if keys_base != keys_new:
            print 'Key mismatch!'
            diff_base = keys_base.difference(keys_new)
            diff_new = keys_new.difference(keys_base)
            print 'Difference:'
            print 'not in the result file :\n%s' % \
                  ('\n'.join(diff_base))
            print 'not in the comparison file:\n%s' % \
                  ('\n'.join(diff_new))
            self.mismatch = True

        common_keys = keys_base.intersection(keys_new)
        for key in common_keys:
            self._diff_content(key, r_base, r_new)

        if self.mismatch:
            print 'Ran into errors!'
            return False
        else:
            print 'Finished without errors'
            return True

    def _parse_simple(self, filename, keys):
        key_dict = {}
        results = []
        f = open(filename, 'rU')
        items = {}
        try:
            header = f.readline()
            for i, item in enumerate(header.split(';')):
                key_dict[i] = item.strip()
            for line in f:
                key = []
                data = []
                for ind, val in enumerate(line.split(';')):
                    val = val.strip()
                    if key_dict[ind] in keys:
                        key.append((key_dict[ind],val))
                    else:
                        data.append((key_dict[ind],val))
                results.append((tuple(key), tuple(data)))
        finally:
            f.close()
        results = _group_key_values(results)
        return results, _pad([k for k, v in key])

    def compare(self, filename2, filename1, type):
        self.mismatch = False
        is_aggregate = False
        if not os.path.exists(filename1):
            print 'ERROR! Comparison file (%s) does not exist!' % filename1
            sys.exit(1)
        if not os.path.exists(filename2):
            print 'ERROR! Result file (%s) does not exist!' % filename2
            sys.exit(1)
        if type == 'by_level':
            keys = ('sim unit','branch','iteration','data level','id','date')
        elif type == 'expres':
            keys = ('level','unit','iteration','branch')
        elif type == 'opres':
            keys = ('op_level','id','op_date','iteration','branch',
                    'op_group','op_name')
            is_aggregate = True
        elif type == 'aggr':
            return self._compare_aggr(filename1, filename2)
        else:
            raise Exception('Invalid type: %s; must be one of: '\
                            'by_level, expres, opres, aggr' % type)


        r_base, rb_head = self._parse_simple(filename1, keys)
        r_new, rn_head = self._parse_simple(filename2, keys)
        keys_base = set([i[0] for i in r_base])
        keys_new = set([i[0] for i in r_new])
        if keys_base != keys_new:
            print 'Key mismatch!'
            diff_base = keys_base.difference(keys_new)
            diff_base = [', '.join(_pad([str(x[1]) for x in d])) \
                         for d in diff_base]
            diff_new = keys_new.difference(keys_base)
            diff_new = [', '.join(_pad([str(x[1]) for x in d])) \
                        for d in diff_new]
            print 'Difference:'
            print 'not in the result file :\n%s\n%s' % \
                  (', '.join(rb_head), '\n'.join(diff_base))
            print 'not in the comparison file:\n%s\n%s' % \
                  (', '.join(rn_head), '\n'.join(diff_new))
            self.mismatch = True
        r_new = dict(r_new)
        for key, res in r_base:
            if key not in r_new:
                continue
            new_res = r_new[key]
            # ensure the values have the same sorting
            for res_col in reversed(range(len(res[0]))):
                res = sorted(res, key=_itemgetter2(res_col))
            for res_col in reversed(range(len(new_res[0]))):
                new_res = sorted(new_res, key=_itemgetter2(res_col))
            exp = []
            got = []
            header = '; '.join([': '.join(k) for k in key])
            first_row = True
            rowcount = max(len(res), len(new_res))
            for i in range(rowcount):
                exp = []
                got = []
                if len(res) <= i:
                    self.mismatch = True
                    row = new_res[i]
                    for k, v in row:
                        got.append(k+': '+v)
                    print '  exp: NOTHING\n  got: %s' % got
                    continue
                if len(new_res) <= i:
                    self.mismatch = True
                    row = res[i]
                    for k, v in row:
                        got.append(k+': '+v)
                    print '  exp: %s\n  got: NOTHING' % got
                    continue
                row = res[i]
                row2 = dict(new_res[i])
                for k, v in row:
                    try:
                        if is_aggregate:
                            eps = AGGR_EPSILON
                        else:
                            eps = EPSILON
                        diff = abs(float(row2[k]) - float(v)) >= ABS_EPSILON
                        if diff:
                            diff = abs(1.0 - (float(row2[k]) / float(v)))
                            diff = diff >= eps
                    except:
                        diff = row2[k] != v
                    if diff:
                        got.append(k+': '+row2[k])
                        exp.append(k+': '+v)
                if len(exp) > 0 and len(got) > 0:
                    if first_row:
                        self.mismatch = True
                        print header
                    print '  exp: %s\n  got: %s' % \
                          ('; '.join(exp), '; '.join(got))
        if self.mismatch:
            print 'Ran into errors!'
            return False
        else:
            print 'Finished without errors'
            return True

def main(file1, file2, type):
    cmp = Comparisor()
    start = time.time()
    res = cmp.compare(file1, file2, type)
    print 'Finished in', time.time()-start, 'seconds'
    return res

if __name__ == '__main__':
    if len(sys.argv) != 4 or sys.argv[1] not in ('by_level', 'expres',
                                                 'opres', 'aggr'):
        print 'Usage: compare_output.py <type> <old> <new>'
        print '<type> one of: by_level, expres, opres, aggr'
        sys.exit(1)
    args = sys.argv
    type = args[1]
    if not os.path.exists(args[2]):
        print '%s does not exist' % args[2]
        sys.exit(1)
    file1 = args[2]
    if not os.path.exists(args[3]):
        print '%s does not exist' % args[3]
        sys.exit(1)
    file2 = args[3]

    main(file1, file2, type)

