#!/usr/bin/env python
# encoding: utf-8
"""
Title: hToPyWrap.py - C Header To Python Wrapper

Reads the SIMO ModelLibrary header and constructs a python
wrapper for the functions

*hToPyWrap.py*

Usage: hToPyWrap.py 
        <Name and full path of the C header file>,
        <Name of the C DLL file>,
        <Directory of xml files>,
        <Directory of the result wrapper>

Copyright (c) 2007 Simosol Oy. Distributed under the GNU General Public License 2.0.
"""
__revision__ = "$Revision: Alpha 0.1 $"

import os
import sys
import time
#from re import *

import xml.etree.cElementTree as ElementTree

# ------------------------------------------------------------------------------
# Functions

def elem_name(elemtag):
    """
    Method: elem_name

    Separates the name from the element namespace, lowercases, and returns the
    name

    Parameters:
        elemtag - the tag of the element whose name should be returned

    Calls:
        None

    Returns:
        The lowercase name of the element
    """
    part = str(elemtag).split("}")
    if len(part) > 1:
        return str.lower(part[1])
    else:
        return str.lower(part[0])
    
def elem_namespace(elemtag):
    """
    Method: elem_namespace

    Separates element's name and namespace, and returns the namespace

    Parameters:
        elemtag - the tag of the element whose namespace should be returned

    Calls:
        None

    Returns:
        The namespace of the element
    """
    part = str(elemtag).split("}")
    if len(part) > 1:
        return str.lower(part[0]) + "}"
    else:
        return ''

def process_modelxml(libname, modelcollection, xmlroot, ns):
    """
    Function: process_modelxml

    Parameters:
        libname
        modelcollection
        xmlroot
        ns

    Returns:
        modelcollection
    """
    # Go through the modelgroups and the models
    for modelgroup in xmlroot:   
        for m in modelgroup:
            # Collect only the models that are implemented in the given library
            if m[1].text != libname:
                continue

            # Add the model to the modelcontainer
            modelname = m[0].text
            modelcollection[modelname] = [[], []]

            # store the input variables, name and level
            for param in m.findall("./"+ns+"variables/"+ns+"variable"):
                modelcollection[modelname][0].append((param[0].text, param[1].text))
            for param in m.findall("./"+ns+"parameters/"+ns+"parameter"):
                modelcollection[modelname][1].append((param[0].text))
    return modelcollection

def process_operationxml(libname, modelcollection, xmlroot, ns):
    """
    Function: process_modelxml

    Parameters:
        libname
        modelcollection
        xmlroot
        ns

    Returns:
        modelcollection
    """  
    # Go through the operationgroups and the operation models
    for operationgroup in xmlroot:
        for oper in operationgroup:
            # Check that operation type is model and it is implemented in the right 
            # library
            if elem_name(oper[1][0].tag) != 'model':
                continue
            if oper[1][0][0].text != libname:
                continue
            modelname = oper[0].text
            modelcollection[modelname] = [[], []]
            # Get the model variables
            for var in oper.findall('./' + ns + 'variables/' + ns + 'variable'):
                modelcollection[modelname][0].append((var[0].text, 'variable'))
            for par in oper.findall('./' + ns + 'parameters/' + ns + 'parameter'):
                modelcollection[modelname][0].append((par.text, 'parameter'))
            for group in oper.findall('./' + ns + 'data/' + ns + 'group'):
                modelcollection[modelname][1].append(group.find('./' + ns +'level').text)
    return modelcollection

def construct_modelcollection(libname, libtype, modelcollection, xmlroot,
                              errorlog):
    """
    Function: construct_modelcollection

    Gather the models that are implemented in a given DLL.

    Parameters:
        libname
        libtype
        modelcollection
        xmlroot

    Returns:
        modelcollection
    """
    # Get the xml namespace
    ns = elem_namespace(xmlroot.tag)

    # create the modelcollection
    if libtype == 'model':
        modelcollection = process_modelxml(libname, modelcollection, xmlroot, ns)
    elif libtype == 'operation':
        modelcollection = process_operationxml(libname, modelcollection, xmlroot, ns)    
    # Return the collection of models
    return modelcollection  

def output_wrapper(wrapperfilename, outfile, header, 
                   libtype, modelcollection, errorlog):
    """
    Function: output_wrapper

    Process the C header, validate that the header and xml definitions are 
    syncronized and create the output for python wrapper

    Parameters:
        wrapperfilename
        header
        libtype
        modelcollection

    """
    if libtype == 'model':
        process_header(wrapperfilename, outfile, 'model', 
                       header, modelcollection, errorlog)
    elif libtype == 'operation':
        process_header(wrapperfilename, outfile, 'operation',
                       header, modelcollection, errorlog)

def divide_model_header(header):
    """
    Function: divide_model_header

    Divide the model header string to modelname and parameterlist

    Parameters:
        header
    """
    paramlist = []
    modelname, parameters = header.split('(')
    # Strip the blanks from the modelname
    modelname = modelname.strip(' ')
    # Split the parameters into list
    paramstr = parameters.split(',')
    # Go through each parameter in the header
    for param in paramstr:
        param = param.split(' ')
        p = []  # Container for a parameter [type, name]
        # Split the paramer into segments
        for seg in param:
            # Check that the segment is not empty string
            if seg != "":
                # Clean unnecessary characters from the segment tring
                for a in ['*', ' ', ')', ';', '\n', '\t']:
                    seg = seg.strip(a)
                # Add the segment to the parameter container
                p.append(seg)
        if len(p) != 2 and p[0] != 'struct':
            pass  # ERROR! Length of parameter container should always be 2
        else:
            paramlist.append(p)
    return modelname, paramlist

def new_model_wrapper(wrapperfilename, outfile):
    """
    Function: new_model_wrapper

    Write the general description, imports etc. of a wrapper file
    """
    timeinfo = time.asctime()
    outfile.write("""
#
# Title: %s
#
# *%s*
#
# Wrapper functions for a SIMO ModelLibrary. Acts as an Python/C DLL calling 
# interface between SIMO Simulator and the model library.
#
# All models are called with a regular function call, which includes the model
# input parameter(s), number of result variables, memory block which is used
# for passing result variables, warning and error memory blocks, error check
# mode, allowed risk level and rectification factor.
# 
# The models are not documented here in detail, as model.xml file is used
# for that purpose.
#
# Wrapper generated automatically %s
#
# Imports
from ctypes import *

# Wrappers for the C functions:
#---------------------------------------------------------------------------"""\
% (wrapperfilename.split('.')[0], wrapperfilename, timeinfo))

def new_operation_wrapper(wrapperfilename, outfile):
    """  
    Function: new_operation_wrapper

    Write the general description, imports etc. of a wrapper file

    """
    timeinfo = time.asctime()
    outfile.write("""
#
# Title: %s
#
# *%s*
#
# Wrapper functions for SIMO OperationLibrary. Acts as an Python/C++ DLL calling 
# interface between SIMO Simulator and the operation library.
#
# Wrapper generated automatically %s 
#
# Imports
# -------
from ctypes import *

# Wrappers for the C functions:
# ------------------------------------------------------------------------- """\
% (wrapperfilename.split('.')[0], wrapperfilename, timeinfo))

def write_model_header(outfile, modelname, modelinputs):
    """
    Function: append_model_header

    Append a single model wrapper in string format to the wrapperfile
    """
    variables, parameters = modelinputs  
    # Create the wrapper for the current model
    modelwrp = '\ndef ' + modelname + '(arg, libc, i):\n\treturn libc.' + \
             modelname + '(\n\t\t'
    p = 0
    for i in range(len(variables)):
        modelwrp += 'arg.variables[i][%s], ' % (str(i))
        p += 1
        if p == 3:
            modelwrp += '\n\t\t'
            p = 0 
    #modelwrp += '\n\t\t'
    for i in range(len(parameters)):
        modelwrp += 'arg.parameters[%s], ' % (str(i))
        p += 1
        if p == 3:
            modelwrp += '\n\t\t'
            p = 0
    if p > 0:
        modelwrp += '\n\t\t'
    modelwrp += 'byref(arg.num_of_res_objs[i]), arg.mem[i], '\
                'byref(arg.errors[i]),\n\t\targ.error_check_mode, '\
                'arg.allowed_risk_level, arg.rect_factor)\n'
    # Write the current model's wrapper to the wrapper file  
    outfile.write(modelwrp)  

def write_operation_header(outfile, modelname, modelcollection):
    """
    Function: write_operation_header

    Append a single operation wrapper in string format to the wrapperfile
    """
    # Create the wrapper for the current operation
    modelwrp = '\ndef ' + modelname + '(operarg, o_libc):\n\treturn o_libc.' + \
             modelname + '(\n\t'
    iparams = 0
    iv, ip = 0, 0
    for variable in modelcollection[0]:
        if variable[1] == 'variable':
            modelwrp += 'operarg.variables[%s], ' % (str(iv))
            iv += 1
        elif variable[1] == 'parameter':
            modelwrp += 'operarg.parameters[%s], ' % (str(ip))
            ip += 1
        iparams += 1
        if iparams == 3:
            modelwrp += '\n\t'
            iparams = 0
    # Add cash flow, cash flow tables, nresobjects and operationresults
    modelwrp += "\n\tbyref(operarg.cash_flow), operarg.cash_flow_table, " \
             "\n\tbyref(operarg.nresobjects), operarg.operationresults,\n\t"
    # Add the operation data levels
    for level in modelcollection[1]:
        modelwrp += "operarg.data['%s'], " % level
    # Add the errors parameters
    modelwrp += "\n\tbyref(operarg.errors))\n"
    outfile.write(modelwrp)

def count_c_variables(libtype, parameters):
    """
    Function: count_c_variables

    Count the number of variables in the C header definition
    """
    nvarsc = 0
    if libtype == 'model':
        stopcondition = ['int', 'nres']
    elif libtype == 'operation':
        stopcondition = ['double', 'cashflow']
    for param in parameters:
        if param == stopcondition:
            return nvarsc
        nvarsc += 1
    return -1

def process_header(wrapperfilename, outfile, libtype, 
                   header, modelcollection, errorlog):
    """
    Function: process_header

    Process a C header of an model/operation library.

    Parameters:
        header
        modelcollection

    Returns:
        wrapper
    """
    # Create a list of modelnames in the C header
    modelnames = []
    # Write the file beginning to the output wrapper
    if libtype == 'model':
        new_model_wrapper(wrapperfilename, outfile)
    elif libtype == 'operation':
        new_operation_wrapper(wrapperfilename, outfile)  
    # Start iterating the lines in the header
    for line in header.readlines():
        # Find the model headers           
        if line.find('DLLEXPORT int') >= 0 or line.find('DLLEXPORT_AGGR int') >=0:
            if line[9] == '_':
                line = line[19:]
            else:
                #remove DLLEXPORT int from the beginning of the line  
                line = line[14:]  
            # Get the model name and the model body
            modelname, parameters = divide_model_header(line)      
            # Add the modelname to the list of models in the C header
            modelnames.append(modelname)      
            # Check that the model is found in the model collection
            if not modelcollection.has_key(modelname):
                errorstr = 'Error! %s %s not found in the model XML file!\n'\
                         % (libtype, modelname)
                errorlog.write(errorstr)
                continue      
            # Count the number of variables in the C header
            nvarsc = count_c_variables(libtype, parameters)
            if nvarsc == -1:
                continue  # ERROR!      
            # Check that number of variables in the header matches the xml definition
            if libtype == 'operation':
                nvarsxml = len(modelcollection[modelname][0])
            else:
                nvarsxml = len(modelcollection[modelname][0]) \
                         + len(modelcollection[modelname][1])
            if nvarsc != nvarsxml:
                # ERROR! The n of vars should match
                errorstr = 'Error in wrapper generation for %s %s, unequal number' \
                         ' of variables (%s variables in C header, '\
                         '%s variables in XML definition\n' % \
                         (libtype, modelname, str(nvarsc), str(nvarsxml))
                errorlog.write(errorstr)        
            else:
                # Write the current model wrapper to the wrapperfile
                if libtype == 'model':
                    write_model_header(outfile, modelname, modelcollection[modelname])
                elif libtype == 'operation':
                    write_operation_header(outfile, modelname, modelcollection[modelname])

    # Crosscheck that all models in the model xml are also defined in the C header
    for modelname, params in modelcollection.iteritems():
        if modelname not in modelnames:
            errorstr = 'Error! %s %s not found in the C header file!\n' \
                     % (libtype, modelname)
            errorlog.write(errorstr)


def main():
    # ------------------------------------------------------------------------------
    # Program main body
    try:

        # Get the program arguments
        inheaderfile = sys.argv[1]  # Name and full path of the C header file
        libname = sys.argv[2]  # Name of the C DLL file
        modelxmlpath = sys.argv[3]  # Directory of xml files
        outwrapperpath = sys.argv[4]  # Directory of the result wrapper

        # Check that the header file exists
        if os.path.isfile(inheaderfile) == 0:
            print "%s is not a valid file." % (inheaderfile)
            sys.exit()
        else:
            inheader = open(inheaderfile, 'r')

        # Check that the xml/models directory exists
        if os.path.isdir(modelxmlpath) == 0:
            print "%s is not a valid file." % (modelxmlpath)
            sys.exit()  

        # Check that the simulator/models (wrappers) directory exists
        if os.path.isdir(outwrapperpath) == 0:
            print "%s is not a valid file." % (outwrapperpath)
            sys.exit()

        wrapperfilename = 'wrapper_' + libname.split('.')[0] + '.py'

        # Create and open the output wrapper file
        outfilepath = os.path.join(outwrapperpath, wrapperfilename)
        outfile = open(outfilepath, 'w')

        # Open errorlog for outputting errors in wrapper generation
        errorlog = open(os.path.join(outwrapperpath,
                                     wrapperfilename[:-3]+'_errors.txt'), 'w')

    except IOError:
        print "%s , %s or %s is not a valid file/path." \
              % (sys.argv[1], sys.argv[2], sys.argv[3])
        sys.exit()
    except IndexError:
        print "Usage: %s headerfile dllfilename modelxmlpath wrapperpath"\
              % (sys.argv[0])
        sys.exit()
    except:
        print "Unspecified error!"
        sys.exit()


    # Create a container for the models
    modelcollection = {}

    # Get the files in the current xml directory
    for root, dirs, files in os.walk(modelxmlpath):  
        # Process only files in the given folder
        if root != modelxmlpath:
            continue  
        # Go through the files in the xml directory
        for fname in files:    
            # Check that the file is xml
            if fname.split('.')[-1] != 'xml':
                continue  
            # Open the current xml file and get the library type ('model' or 'operation')
            xmltree = ElementTree.ElementTree(file = os.path.join(root,fname))
            xmlroot = xmltree.getroot()
            # Get the modelxml namespace
            model_ns = elem_namespace(xmlroot.tag)    
            # Get the librarytype
            if elem_name(xmlroot.tag) == 'modelbase':
                libtype = 'model'
            elif elem_name(xmlroot.tag) == 'operation_modelbase':
                libtype = 'operation'
            else:
                # XML file not of the right type
                sys.exit()      
            # Go through the models in the current model/operation xml file and get
            # all models that should be implemented in the current dll library
            modelcollection = construct_modelcollection(\
                libname, libtype, modelcollection, xmlroot, errorlog)    
    # Process the headerfile and create wrapperfile
    output_wrapper(wrapperfilename, outfile, inheader, 
                   libtype, modelcollection, errorlog)

    # Close the opened files
    outfile.close()
    inheader.close()
    errorlog.close()

if __name__ == '__main__':
    main()