#!/usr/bin/env python
# encoding: utf-8
"""
Title: harvest

Harvesting operations

*harvest.py*

Copyright (c) 2007 Simosol Oy. Distributed under the GNU General Public License
2.0.

NOTE:
    All Python operation models share the same set of input parameters. When
    implementing new operation model, remember to assign the following list
    of parameters as the function parameters:
        data - data by level
        variables - operation variable values
        parameters - operation parameter values
        par_table_val - values retrieved from a parameter table
        cash_flow_table - cash_flow structure
        cash_flow - container for returning cash flow (sum)
        results - container (numpy array) for returning operation results
        warnings - container for returning warnings
        errors - container for returning errors
"""

# Imports
from harvestdata import HarvestData
from harvestop import HarvestOperation
from constants import STANDLABEL, EPSILON
import numpy

############################################################
#  Operation functions, callable from outside of module
############################################################


def bioenergy_thinning(data, variables, parameters, par_table_val,
                       cash_flow_table, cash_flow, results, restype,
                       warnings, errors, price_cache, stem_cache):
    """
    Bioenergy thinning. Implementation of the model is the same as in
    low_thinning, only operation name is different.
    """
    return thinning(data, variables, parameters, par_table_val,
                    cash_flow_table, cash_flow, results, restype,
                    warnings, errors, price_cache, stem_cache)


def clearcut(data, variables, parameters, par_table_val,
             cash_flow_table, cash_flow, results, restype,
             warnings, errors, price_cache, stem_cache):
    """
    Remove all trees except a number of reserve trees from a stand.
    """
    # check that operation data actually exists
    for level, ldata in data.iteritems():
        if ldata is None:
            errors.append("The level %s doesn't have data!")
            return -1
    hd = HarvestData('clearcut', data, variables, parameters,
                     par_table_val, cash_flow_table, price_cache)
    hd.set_stump_inclusion()
    hop = HarvestOperation(hd, stem_cache)
    # Iterate through all trees in the data from smallest to biggest
    for dcind in hd.cdata['removal_order']:
        hop.buck(dcind)
        props = hop.reduce_prop_for_reserve_trees(dcind, 1.0)
        hop.cut_dclass_trees(dcind, props)
    hop.store_harvest_results(results, cash_flow, scale=None,
                              removedvolume=None, restype=restype)
    return 1


def covertree_position(data, variables, parameters, par_table_val,
                       cash_flow_table, cash_flow, results, restype,
                       warnings, errors, price_cache, stem_cache):
    """
    Remove all trees except a number of covertrees from a stand.
    """
    return seedtree_position(data, variables, parameters,
                             par_table_val, cash_flow_table,
                             cash_flow, results, restype, warnings, errors,
                             price_cache, stem_cache)


def estimate_stand_value(data, variables, parameters, par_table_val,
                         cash_flow_table, cash_flow, results, restype,
                         warnings, errors, price_cache, stem_cache):
    """
    buck each tree of a stand and calculate the total value in given currency.
    """
    # check that operation data actually exists
    for level, ldata in data.iteritems():
        if ldata is None:
            errors.append("The level %s doesn't have data!")
            return -1
    hd = HarvestData('standvalue', data, variables, parameters,
                     par_table_val, cash_flow_table, price_cache)
    hd.cdata['min_stump_diam'] = 0.0
    hd.cdata['stump_excl_num'] = 0.0
    hd.cdata['biomass_incl_prop'] = 1.0
    hd.cdata['stump_incl_prop'] = 1.0
    hop = HarvestOperation(hd, stem_cache)
    for dcind in hd.cdata['removal_order']:
        hop.buck(dcind)
        hop.cut_dclass_trees(dcind, (1.0, 1.0), trackchanges=False)
    hop.store_stand_value_results(results)
    return 1


def estimate_tree_value(data, variables, parameters, par_table_val,
                        cash_flow_table, cash_flow, results, restype, warnings,
                        errors, price_cache, stem_cache):
    """
    buck each tree of a stand and calculate the total value in given currency.
    """
    # check that operation data actually exists
    for level, ldata in data.iteritems():
        if ldata is None:
            errors.append("The level %s doesn't have data!")
            return -1
    hd = HarvestData('standvalue', data, variables, parameters,
                     par_table_val, cash_flow_table, price_cache)
    hd.cdata['min_stump_diam'] = 0.0
    hd.cdata['stump_excl_num'] = 0.0
    hd.cdata['biomass_incl_prop'] = 1.0
    hd.cdata['stump_incl_prop'] = 1.0
    hop = HarvestOperation(hd, stem_cache)
    for dcind in hd.cdata['removal_order']:
        hop.buck(dcind)
        hop.cut_dclass_trees(dcind, (1.0, 1.0), trackchanges=False)
    hop.store_tree_value_results(results)
    return 1


def first_thinning(data, variables, parameters, par_table_val,
                   cash_flow_table, cash_flow, results, restype, warnings,
                   errors, price_cache, stem_cache):
    """
    First thinning. Implementation of the model is the same as in low_thinning,
    only operation name is different
    """
    return thinning(data, variables, parameters, par_table_val,
                    cash_flow_table, cash_flow, results, restype, warnings,
                    errors, price_cache, stem_cache)


def remove_seedtrees(data, variables, parameters, par_table_val,
                     cash_flow_table, cash_flow, results, restype, warnings,
                     errors, price_cache, stem_cache):
    """
    remove the seedtrees from a stand.
    """
    # check that operation data actually exists
    for level, ldata in data.iteritems():
        if ldata is None:
            errors.append("The level %s doesn't have data!")
            return -1
    hd = HarvestData('removeseedtree', data, variables, parameters,
                     par_table_val, cash_flow_table, price_cache,
                     keepemptystrata=True)
    hop = HarvestOperation(hd, stem_cache)
    ind = hd.ind
    strata = hd.stratumdata
    for dcind in hd.cdata['removal_order']:
        tree, istratum, sp = hop.get_dclass_info(dcind)
        if strata[istratum, ind['stratum_seedtree']] == 1.0:
            hop.buck(dcind)
            hop.cut_dclass_trees(dcind, (1.0, 1.0), biomass=False)
    rem_vol_merch = hop.store_harvest_results(results, cash_flow, scale=None,
                                              removedvolume=None,
                                              biomass=False, tagmods=False)
    # Abort the operation in case accumulated volume was too small
    # NB! rem_vol_merch is the total merchantable (log+pulpwood)
    # volume removed per hectare
    if rem_vol_merch < hd.cdata['min_vol_accumulation_remove_seedtrees']:
        return -2
    return 1


def seedtree_position(data, variables, parameters, par_table_val,
                      cash_flow_table, cash_flow, results, restype, warnings,
                      errors, price_cache, stem_cache):
    """
    Remove all trees except a number of seedtrees from a stand.
    """
    # check that operation data actually exists
    for level, ldata in data.iteritems():
        if ldata is None:
            errors.append("The level %s doesn't have data!")
            return -1

    hd = HarvestData('seedtree', data, variables, parameters,
                     par_table_val, cash_flow_table, price_cache)
    hd.set_stump_inclusion()
    hop = HarvestOperation(hd, stem_cache)

    # check that seedtreespecies stratum has enough number of big trees
    # and assign seedtree numbers by tree species
    dlim = 0.9 * hd.cdata['ddom']
    hop.number_of_potential_seedtrees(dlim)

    # iteration from smallest to biggest trees
    for dcind in hd.cdata['removal_order']:
        hop.buck(dcind)
        props = hop.reduce_prop_for_seed_and_reserve_trees(dcind, 1.0, dlim)
        hop.cut_dclass_trees(dcind, props)
    rem_vol_merch = hop.store_harvest_results(results, cash_flow, scale=None,
                                              removedvolume=None,
                                              restype=restype)
    # Abort the operation in case accumulated volume was too small
    # NB! rem_vol_merch is the total merchantable (log+pulpwood)
    # volume removed per hectare
    if rem_vol_merch < hd.cdata['min_vol_accumulation_seedtree_position']:
        return -2
    return 1


def sheltertree_position(data, variables, parameters, par_table_val,
                         cash_flow_table, cash_flow, results, restype,
                         warnings, errors, price_cache, stem_cache):
    """
    Remove all trees except a number of sheltertrees from a stand.
    """
    return seedtree_position(data, variables, parameters,
                             par_table_val, cash_flow_table,
                             cash_flow, results, restype, warnings, errors,
                             price_cache, stem_cache)


def strip_cutting(data, variables, parameters, par_table_val,
                  cash_flow_table, cash_flow, results, restype, warnings,
                  errors, price_cache, stem_cache):
    """
    Remove all trees from narrow strip shaped stand. The trees in adjasent
    stands will seed the new tree generation of the strip cutting area.
    """
    # Assign seedtree_density to 0, add the two other parameter table values
    par_table_val = [[[0]], par_table_val[0], par_table_val[1]]
    return seedtree_position(data, variables, parameters,
                             par_table_val, cash_flow_table,
                             cash_flow, results, restype, warnings, errors,
                             price_cache, stem_cache)


def thinning(data, variables, parameters, par_table_val,
             cash_flow_table, cash_flow, results, restype, warnings, errors,
             price_cache, stem_cache):
    """
    Executes a thinning in a stand. Remove a given portion of classes from
    stand's diameter distribution, starting from the smallest or biggest ones.
    The thinning type (low or upper thinning) is defined using parameters.
    """
    # check that operation data actually exists
    for level, ldata in data.iteritems():
        if ldata is None:
            errors.append("The level %s doesn't have data!")
            return -1

    hd = HarvestData('thinning', data, variables, parameters,
                     par_table_val, cash_flow_table, price_cache)
    hd.set_track_removal()
    # Thinning based on basal area
    targetvar = 'BA'
    hd.set_thinning_control_param(targetvar)
    hd.set_stump_inclusion()
    hd.set_d_cut_all_below()
    hop = HarvestOperation(hd, stem_cache)
    # Buck all tree stems and remove stems from tracks
    for dcind in hd.cdata['removal_order']:
        prop = hd.cdata['track_removal']
        hop.buck(dcind)
        hop.cut_dclass_trees(dcind, (prop, prop))
    # Update BA_to_remove when stems from tracks are removed
    hd.update_to_remove()
    # remove diameter classes from both ends of the distribution
    hop.cut_from_top_and_bottom()
    # Finally store the cumulative volumes, values etc. into result data
    rem_vol_merch = hop.store_harvest_results(results, cash_flow, scale=None,
                                              removedvolume=None,
                                              restype=restype)
    # Abort the operation in case accumulated volume was too small
    # NB! rem_vol_merch is the total merchantable (log+pulpwood)
    # volume removed per hectare
    if rem_vol_merch < hd.cdata['min_vol_accumulation']:
        return -2
    return 1


def first_thinning_stem_number(data, variables, parameters, par_table_val,
                               cash_flow_table, cash_flow, results, restype,
                               warnings, errors, price_cache, stem_cache):
    """
    Executes a stem number based first thinning in a stand. Remove a given
    portion of classes from stand's diameter distribution, starting from the
    smallest or biggest ones.
    The thinning type (low or upper thinning) is defined using parameters.
    """
    # check that operation data actually exists
    for level, ldata in data.iteritems():
        if ldata is None:
            errors.append("The level %s doesn't have data!")
            return -1

    hd = HarvestData('first_thinning_stem_number', data, variables, parameters,
                     par_table_val, cash_flow_table, price_cache)
    hd.set_track_removal()
    # Thinning based on stem number
    targetvar = 'N'
    hd.set_thinning_control_param(targetvar)
    hd.set_stump_inclusion()
    hd.set_d_cut_all_below()
    hop = HarvestOperation(hd, stem_cache)
    # Buck all tree stems and remove stems from tracks
    for dcind in hd.cdata['removal_order']:
        prop = hd.cdata['track_removal']
        hop.buck(dcind)
        hop.cut_dclass_trees(dcind, (prop, prop))
    # Update N_to_remove when stems from tracks are removed
    hd.update_to_remove()
    # remove diameter classes from both ends of the distribution
    hop.cut_from_top_and_bottom()
    # Finally store the cumulative volumes, values etc. into result data
    rem_vol_merch = hop.store_harvest_results(results, cash_flow, scale=None,
                                              removedvolume=None,
                                              restype=restype)
    # Abort the operation in case accumulated volume was too small
    # NB! rem_vol_merch is the total merchantable (log+pulpwood)
    # volume removed per hectare
    if rem_vol_merch < hd.cdata['min_vol_accumulation']:
        return -2
    return 1


def thinning_with_removed_volume(data, variables, parameters, par_table_val,
                                 cash_flow_table, cash_flow, results, restype,
                                 warnings, errors, price_cache, stem_cache):
    """
    Executes a thinning in a stand. Remove a given portion of classes from
    stand's diameter distribution, starting from the smallest or biggest ones.
    The thinning type (low or upper thinning) is defined using parameters.
    """
    # Check if V_harvested or remaining_BA are given
    if data[STANDLABEL][0][13] == 0 or (data[STANDLABEL][0][13] > 0 and
                                        data[STANDLABEL][0][14] > 0):
        # Basic low thinning. If removed volume is 0 or both removed volume
        # and remaining basal area are greater then 0, basic low_thinning is
        # used (Remaining_upper and Remaining_lower in thinning model replaced
        # with remaining_ba)
        return thinning(data, variables, parameters, par_table_val,
                        cash_flow_table, cash_flow, results, restype,
                        warnings, errors, price_cache,
                        stem_cache)
    else:
        # check that operation data actually exists
        for level, ldata in data.iteritems():
            if ldata is None:
                errors.append("The level %s doesn't have data!")
                return -1
        hd = HarvestData('thinning_with_removed_volume', data, variables,
                         parameters, par_table_val, cash_flow_table,
                         price_cache)
        hd.set_track_removal()
        # Thinning based on volume of timber (commercially usable wood)
        targetvar = 'V_com'
        hd.set_thinning_control_param(targetvar)
        hd.set_stump_inclusion()
        hd.set_d_cut_all_below()
        hop = HarvestOperation(hd, stem_cache)
        # Calculate current volume
        hop.V_current = hd.get_vol_from_treelist()
        # Buck all tree stems
        for dcind in hd.cdata['removal_order']:
            hop.buck(dcind)
        # Calculate scaling factor and removed volume for results
        svol = hop.hdata.result['stratum_vol']
        scale, removedvolume = hop._set_result_scaling(svol)
        # Remove stems from tracks
        for dcind in hd.cdata['removal_order']:
            prop = hd.cdata['track_removal']
            hop.cut_dclass_trees(dcind, (prop, prop), scale=scale)
        # Update V_to_remove when stems from tracks are removed
        # NB! With wide track_width and narrow track_distance, and small
        # accumulated timber volume  the whole removed volume might be already
        # removed here! Check the thinning model parameters in the operation
        # model chain.
        cut_timber = hop.cut_timber
        hd.update_to_remove(cut_timber)
        # remove diameter classes from both ends of the distribution
        hop.cut_from_top_and_bottom(scale)
        # Finally store the cumulative volumes, values etc. into result data
        hop.store_harvest_results(results, cash_flow, scale, removedvolume,
                                  restype=restype)
        return 1


def selection_cut(data, variables, parameters, par_table_val,
                  cash_flow_table, cash_flow, results, restype, warnings,
                  errors, price_cache, stem_cache):
    """
    Basic selection cutting model derived from basic harvest model.
    """
    # check that operation data actually exists
    for level, ldata in data.iteritems():
        if ldata is None:
            errors.append("The level %s doesn't have data!")
            return -1
    hd = HarvestData('selection_cut', data, variables, parameters,
                     par_table_val, cash_flow_table, price_cache,
                     keepemptystrata=True)
    # Stem number of under storey not included
    hd.set_track_removal_without_under_storey()
    # Thinning based on basal area
    targetvar = 'BA'
    hd.set_selection_cut_control_param(targetvar)
    hd.set_stump_inclusion()
    hd.set_d_cut_all_above()
    hop = HarvestOperation(hd, stem_cache)
    # Buck all tree stems and remove stems from tracks
    for dcind in hd.cdata['removal_order']:
        prop = hd.cdata['track_removal']
        hop.buck(dcind)
        hop.cut_dclass_trees(dcind, (prop, prop))
    # Update BA_to_remove when stems from tracks are removed
    hd.update_to_remove()
    if hd.cdata['slope'] == 0:
        # remove diameter classes evenly
        hop.cut_evenly()
    else:
        # set proportion removed from top and bottom
        hd.set_remove_from_below_and_above_percentage()
        # remove diameter classes from both ends of the distribution
        hop.cut_from_top_and_bottom(harvesttype='selection_cut')
    # Finally store the cumulative volumes, values etc. into result data
    rem_vol_merch = hop.store_harvest_results(results, cash_flow, scale=None,
                                              removedvolume=None,
                                              restype=restype)
    # Abort the operation in case accumulated volume was too small
    # NB! rem_vol_merch is the total merchantable (log+pulpwood)
    # volume removed per hectare
    if rem_vol_merch < hd.cdata['min_vol_accumulation']:
        return -2
    return 1


def selection_cut_with_little_open_areas(data, variables, parameters,
                                         par_table_val, cash_flow_table,
                                         cash_flow, results, restype, warnings,
                                         errors, price_cache, stem_cache):
    """
    Basic selection cutting model derived from basic harvest model.
    """
    # check that operation data actually exists
    for level, ldata in data.iteritems():
        if ldata is None:
            errors.append("The level %s doesn't have data!")
            return -1
    hd = HarvestData('selection_cut', data, variables, parameters,
                     par_table_val, cash_flow_table, price_cache,
                     keepemptystrata=True)
    # Stem number of under storey not included
    hd.set_track_removal_without_under_storey()
    # Thinning based on basal area
    targetvar = 'BA'
    hd.set_selection_cut_control_param(targetvar)
    hd.set_stump_inclusion()
    hd.set_d_cut_all_above()
    hop = HarvestOperation(hd, stem_cache)
    # Buck all tree stems and remove stems from tracks
    for dcind in hd.cdata['removal_order']:
        prop = hd.cdata['track_removal']
        hop.buck(dcind)
        hop.cut_dclass_trees(dcind, (prop, prop))
    # Update BA_to_remove when stems from tracks are removed
    hd.update_to_remove()
    toremove = hd.cdata['to_remove']
    if toremove > EPSILON:
        if toremove - 2 > EPSILON:
            hd.cdata['target'] = hd.cdata['target'] + 2
            # remove diameter classes evenly up to target+2
            hop.cut_evenly()
            hd.cdata['target'] = hd.cdata['target'] - 2
            hd.update_to_remove()
        if hd.cdata['slope'] == 0:
            # remove diameter classes evenly
            hop.cut_evenly()
        else:
            # set proportion removed from top and bottom
            hd.set_remove_from_below_and_above_percentage()
            # remove diameter classes from both ends of the distribution
            hop.cut_from_top_and_bottom(harvesttype='selection_cut')
    # Finally store the cumulative volumes, values etc. into result data
    rem_vol_merch = hop.store_harvest_results(results, cash_flow, scale=None,
                                              removedvolume=None,
                                              restype=restype)
    # Abort the operation in case accumulated volume was too small
    # NB! rem_vol_merch is the total merchantable (log+pulpwood)
    # volume removed per hectare
    if rem_vol_merch < hd.cdata['min_vol_accumulation']:
        return -2
    return 1


def wood_assortments_plot(data, variables, parameters, par_table_val,
                          cash_flow_table, cash_flow, results, restype,
                          warnings, errors, price_cache, stem_cache):
    """
    buck each tree of a stand and calculate the total value in given currency.
    """
    # check that operation data actually exists
    for level, ldata in data.iteritems():
        if ldata is None:
            errors.append("The level %s doesn't have data!")
            return -1

    hd = HarvestData('wood_assortments', data, variables, parameters,
                     par_table_val, cash_flow_table, price_cache)
    hd.cdata['min_stump_diam'] = 0.0
    hd.cdata['stump_excl_num'] = 0.0
    hd.cdata['biomass_incl_prop'] = 0.0
    hd.cdata['stump_incl_prop'] = 0.0
    hop = HarvestOperation(hd, stem_cache)
    for dcind in hd.cdata['removal_order']:
        hop.buck(dcind)
        hop.cut_dclass_trees(dcind, (1.0, 1.0), volres=True, biomass=False,
                             trackchanges=False)
    hop.store_wood_assortment_results(results)
    return 1


def assortment_prices(data, variables, parameters, par_table_val,
                      cash_flow_table, cash_flow, results, restype,
                      warnings, errors, price_cache, stem_cache):
    """
    assortment prices.
    """
    # use 'estimate_stand_value' harvest data parameters here
    hd = HarvestData('assortment_prices', data, variables, parameters,
                     par_table_val, cash_flow_table, price_cache)

    cf = cash_flow_table[0]
    cf_array = cf.active_array.array  # cash_flow_table
    price_region = hd.cdata['price_region']  # price_region value
    species = (1, 2, 3, 9)  # species listing
    assortments = (1, 2)  # assortment listing
    pr_index = cf.classifier_order.index('PRICE_REGION')
    sp_index = cf.classifier_order.index('SP')
    as_index = cf.classifier_order.index('assortment')

    # check that the price region is valid
    if price_region not in set(cf_array[:, pr_index]):
        try:
            msg = "no prices defined for price_region '%d'" % int(price_region)
        except:
            msg = "no price_region when setting prices"
        errors.append(msg)
        return -1

    # get species codes found in the cash flow table
    cf_species = set(numpy.unique(cf_array[:, sp_index]))

    # get assortment values found in cash flow table
    cf_assortments = set(numpy.unique(cf_array[:, as_index]))

    # check the species and assortments
    ok = True
    for sp in species:
        for assortment in assortments:
            if sp not in cf_species or assortment not in cf_assortments:
                msg = "assortment '%d' for species '%d' not "\
                      "found in cash flow table '%s'" % \
                      (int(assortment), int(sp), cf.name)
                errors.append(msg)
                ok = False
    if not ok:
        return -1

    # key locations:
    #   key = (species, assortment) tuple
    #   value = location, ie. mapping from key to result array index
    key_locs = {(1, 1): 0,
                (2, 1): 1,
                (3, 1): 2,
                (9, 1): 3,
                (1, 2): 4,
                (2, 2): 5,
                (3, 2): 6,
                (9, 2): 7}

    # collect the values for each result location
    values = [[] for key in range(max(key_locs.values()) + 1)]
    for row in cf_array:
        preg, sp, assortment, diam, length, price = row
        if preg != price_region:
            continue
        key = (sp, assortment)
        if key not in key_locs:
            continue
        loc = key_locs[key]
        values[loc].append(price)

    # calculate the mean prices and store them to result ar
    for key, loc in key_locs.items():
        prices = values[loc]
        results[loc] = sum(prices) / (1. * len(prices))

    return 1


def main_sp(data, variables, parameters, par_table_val,
            cash_flow_table, cash_flow, results, restype,
            warnings, errors, price_cache, stem_cache):
    """
    calculates mains species. Main species is the growing species of stand
    - not the biggest species calculated as basal area or stem number
    """
    ind = {}
    STANDLABEL = 1
    STRATUMLABEL = 2
    ind['stand_open_area'] = 3
    ind['stratum_storey'] = 4
    ind['stratum_sp'] = 5
    ind['stratum_n'] = 6
    ind['stratum_ba'] = 7
    ind['stratum_artif_ingrowth'] = 8

    # check that strata actually exists
    if data[STANDLABEL][0][ind['stand_open_area']] == 1:
        results[0] = None
        return 1

    sppref = list(par_table_val[0][0])
    stratumdata = data[STRATUMLABEL]
    artif_ingr = False
    growing_storey = False
    artif = [s[ind['stratum_artif_ingrowth']] for s in stratumdata]
    st = [s[ind['stratum_storey']] for s in stratumdata]
    for a in artif:
        if a == 1:
            artif_ingr = True
            break
    for a in st:
        if a in [1, 5]:
            growing_storey = True
            break

    new_stratumdata = {}
    # Go through strata and pick up strata with artificial ingrowth
    # If not artificial inrowth pick up strata with storey 1 and 5
    # Sum strata with the same tree species together
    for stratum in stratumdata:
        sp = stratum[ind['stratum_sp']]
        n = stratum[ind['stratum_n']]
        ba = stratum[ind['stratum_ba']]
        artif_ingrowth = stratum[ind['stratum_artif_ingrowth']]
        storey = stratum[ind['stratum_storey']]
        stratum_sp = sp
        if artif_ingr and growing_storey:
            if stratum_sp not in new_stratumdata and artif_ingrowth \
                    and storey in [1, 5]:
                new_stratumdata[stratum_sp] = 0, 0
                o_ba, o_n = new_stratumdata[stratum_sp]
                new_stratumdata[stratum_sp] = (ba + o_ba, n + o_n)
            elif artif_ingrowth and storey in [1, 5]:
                o_ba, o_n = new_stratumdata[stratum_sp]
                new_stratumdata[stratum_sp] = (ba + o_ba, n + o_n)
        elif growing_storey:
            if stratum_sp not in new_stratumdata and storey in [1, 5]:
                new_stratumdata[stratum_sp] = 0, 0
                o_ba, o_n = new_stratumdata[stratum_sp]
                new_stratumdata[stratum_sp] = (ba + o_ba, n + o_n)
            elif storey in [1, 5]:
                o_ba, o_n = new_stratumdata[stratum_sp]
                new_stratumdata[stratum_sp] = (ba + o_ba, n + o_n)
        else:
            if stratum_sp not in new_stratumdata:
                new_stratumdata[stratum_sp] = 0, 0
                o_ba, o_n = new_stratumdata[stratum_sp]
                new_stratumdata[stratum_sp] = (ba + o_ba, n + o_n)
            else:
                o_ba, o_n = new_stratumdata[stratum_sp]
                new_stratumdata[stratum_sp] = (ba + o_ba, n + o_n)

    found = False
    # Go through strata in preference order and choose species of stratum
    # as comp_unit's main species if stratum's tree number is more than 1200
    # or stratum's basal area more than 10
    for sp in sppref:
        if sp in new_stratumdata:
            for sp_strat, ba_n_strat in new_stratumdata.iteritems():
                if sp == sp_strat:
                    if ba_n_strat[0] > 10 or ba_n_strat[1] > 1200:
                        found = True
                        results[0] = sp_strat
                        return 1
    # If any of strata has more than 1200 trees/ha or 10 m2/ha
    # go through again with lower tree number and basal area restriction
    if not found:
        for sp in sppref:
            if sp in new_stratumdata:
                for sp_strat, ba_n_strat in new_stratumdata.iteritems():
                    if sp == sp_strat:
                        if ba_n_strat[0] > 7 or ba_n_strat[1] > 800:
                            found = True
                            results[0] = sp_strat
                            return 1
    # If any of strata has more than 800 trees/ha or 7 m2/ha
    # go through again without tree number and basal area restriction
    if not found:
        for sp in sppref:
            if sp in new_stratumdata:
                for sp_strat, ba_n_strat in new_stratumdata.iteritems():
                    if sp == sp_strat:
                        if ba_n_strat[0] > 0 or ba_n_strat[1] > 0:
                            found = True
                            results[0] = sp_strat
                            return 1

    return 1
