#!/usr/bin/env python
# encoding: utf-8
"""
Title: silviculturedata

Data object for silviculture operations

*silviculturedata.py*

Copyright (c) 2007 Simosol Oy. Distributed under the GNU General Public License 2.0.
"""

import math
import numpy
from constants import EPSILON, MIN_STEM_NUMBER, TREELABEL, STRATUMLABEL, STANDLABEL
from collections import defaultdict

import constants

class SilvicultureData(object):
    def __init__(self, silviculturetype, data, variables, parameters,
                 ptable_values, cash_flow_table, keepemptystrata=False):
        super(SilvicultureData, self).__init__()
        self.valid = True
        # original data
        self.data = data
        # data indices and
        # controlling data constructed from parameter, variable
        # and data values
        if silviculturetype == 'tending':
            inddict = constants.tendingind
            pardict = constants.tendingpar
            datadict = constants.tendingdata
        elif silviculturetype == 'planting':
            inddict ={}
            pardict = {}
            datadict = constants.plantingdata
        elif silviculturetype == 'supplemental_planting':
            inddict = constants.supplemental_plantingind
            pardict = {}
            datadict = constants.supplemental_plantingdata
        elif silviculturetype == 'fertilization':
            inddict = constants.fertilizationind
            pardict = {}
            datadict = constants.fertilizationdata
        elif silviculturetype == 'cleaning':
            inddict = constants.cleaningind
            pardict = constants.cleaningpar
            datadict = constants.cleaningdata

        self.ind = inddict
        self.cdata = {}
        self.__set_cdata('par', pardict, parameters)
        self.__set_cdata('data', datadict, data)
        if silviculturetype != 'fertilization':
            try:
                self.cdata['targetdensity'] = ptable_values[0][0][0]
                if silviculturetype == 'tending' or silviculturetype == 'cleaning':
                    self.cdata['preferenceorder'] = list(ptable_values[0][1])
            except IndexError:
                self.valid = False
        if silviculturetype == 'tending' or silviculturetype == 'cleaning':
            self.cdata['sp_proportion'] = {}
            sppref = self._modify_sp_pref(self.cdata['preferenceorder'],None)
            pp = self.cdata.get('preference_percentage')
            if pp is None:
                pp = []
            for i in xrange(len(sppref)):
                if i < len(pp):
                    self.cdata['sp_proportion'][sppref[i]] = pp[i]
                else:
                    self.cdata['sp_proportion'][sppref[i]] = 0.0

    def separate_strata_and_remove_treeless(self, treecount):
        """
        Separate the stratums into artificially and naturally created and remove
        stratums without trees.
        """
        pind = self.ind['tree_parentid']
        stpind = self.ind['stratum_parentid']
        spind = self.ind['stratum_sp']
        idind = self.ind['stratum_id']
        Nind = self.ind['stratum_N']
        artind = self.ind['stratum_artif']
        data = self.data
        natstrat = []
        artstrat = []
        for istratum, stratumdata in enumerate(data[STRATUMLABEL]):
            no_trees = False
            sid = stratumdata[idind]
            # Check that stratum has trees in it
            if sid not in data[TREELABEL][:, pind]:
                no_trees = True
                if stratumdata[Nind] < MIN_STEM_NUMBER:
                    # If stratum's N is under minimum, remove the stratum
                    stratumdata[-1] = 2.
                    continue

            if stratumdata[artind] == 0:
                natstrat.append([istratum, stratumdata[stpind], 
                                 stratumdata[spind], no_trees])
            else:
                artstrat.append([istratum, stratumdata[stpind], 
                                 stratumdata[spind], no_trees])

        return natstrat, artstrat, treecount

    def separate_strata(self):
        """
        Separate the stratums into artificially and naturally created and remove
        stratums without trees.
        """
        stpind = self.ind['stratum_parentid']
        spind = self.ind['stratum_sp']
        artind = self.ind['stratum_artif']
        data = self.data
        natstrat = []
        artstrat = []
        for istratum, stratumdata in enumerate(data[STRATUMLABEL]):
            no_trees = True                
            if stratumdata[artind] == 0:
                natstrat.append([istratum, stratumdata[stpind], 
                                 stratumdata[spind], no_trees])
            else:
                artstrat.append([istratum, stratumdata[stpind], 
                                 stratumdata[spind], no_trees])

        return natstrat, artstrat

    def get_real_stem_numbers_for_stratums(self):
        """
        Construct a dictionary, where number key is stratum id and number of
        stems is the value
        """
        pind = self.ind['tree_parentid']
        nind = self.ind['tree_n']
        treelist = self.data[TREELABEL]
        stem_number = {}
        for tree in treelist:
            tid = tree[pind]
            if not stem_number.has_key(tid):
                stem_number[tid] = 0.
            stem_number[tid] += tree[nind]
        return stem_number

    def get_diamclasses_for_stratums(self):
        """
        Construct a dictionary, where number key is stratum id and number of
        diameter classes is the value
        """
        pind = self.ind['tree_parentid']
        treelist = self.data[TREELABEL]
        treenums = {}
        for tree in treelist:
            tid = tree[pind]
            if not treenums.has_key(tid):
                treenums[tid] = 0
            treenums[tid] += 1
        return treenums

    def create_sorted_objectlist(self):
        """
        Construct a sorted list of objects by given sort attribute.
        """
        dind = self.ind['tree_d']
        idind = self.ind['tree_id']
        objectarray = self.data[TREELABEL]
        sortedarray = objectarray[numpy.lexsort((objectarray[:,idind],
                                                 objectarray[:,dind]))]
        return sortedarray

    def __set_cdata(self, dtype, vars, data):
        for varname, vardef in vars.iteritems():
            if dtype == 'data':
                self.cdata[varname] = data[vardef[0]][vardef[1], vardef[2]]
            elif dtype in ('par'):
                vtype, ind = vardef
                if vtype is None:
                    self.cdata[varname] = data[ind]
                elif vtype == 'bool':
                    vals = data[ind]
                    if isinstance(vals, list):
                        self.cdata[varname] = [bool(x) for x in vals]
                    else:
                        self.cdata[varname] = bool(vals)
                elif vardef[0] == 'proportion':
                    vals = data[ind]
                    if isinstance(vals, list):
                        item = [self._constr_prop(x) for x in vals]
                        self.cdata[varname] = item
                    else:
                        self.cdata[varname] = self._constr_prop(vals)
                else:
                    self.cdata[varname] = None

    def update_stem_number(self, treecount, origtreecount):
        """
        Update stem number after supplemental planting and indicate
        modifications
        """
        if abs(treecount - origtreecount) >= EPSILON:
            # update the tree number for the stand and indicate modification
            self.data[STANDLABEL][0, self.ind['stand_N']] = treecount
            self.data[STANDLABEL][0, -1] = 1.
            return True
        return False

    def targetdensity_reached(self, treecount):
        """
        Is target density already reached
        """
        target = self.cdata['targetdensity']
        if treecount < target or abs(treecount - target) < EPSILON:
            # reached target density, exit
            self.data[STANDLABEL][0, self.ind['stand_N']] = treecount
            self.data[STANDLABEL][0, -1] = 1.
            return True
        return False

    def get_site_index(self,sp):
        """
        Calculate site index
        """
        sc = self.cdata['sc']
        ts = self.cdata['ts']
        if ts>1000:
            if sp==1:
                if sc<=2:
                    si_100=30
                elif sc==3:
                    si_100=27
                elif sc==4:
                    si_100=24
                elif sc==5:
                    si_100=21
                elif sc==6:
                    si_100=18
                elif sc>=7:
                    si_100=15
            else:
                #spruce and deciduos species
                if sc==1:
                    si_100=30
                elif sc==2:
                    si_100=27
                elif sc==3:
                    si_100=24
                elif sc==4:
                    si_100=21
                elif sc==5:
                    si_100=18
                elif sc>=6:
                    si_100=15
        elif ts<=1000 and ts>900:
            if sc<=3:
                si_100=24
            elif sc==4:
                si_100=21
            elif sc==5:
                si_100=18
            elif sc==6:
                si_100=15
            elif sc>=7:
                si_100=12
        else:
            #ts<=900
            if sc<=3:
                si_100=21
            elif sc==4:
                si_100=18
            elif sc==5:
                si_100=15
            elif sc==6:
                si_100=12
            elif sc>=7:
                si_100=10

        return si_100

    def get_fertilization_growth(self,sp,hdom,si_100,time_step, ba_stratum):
        """
        Calculate fertilization growth effect
        """
        n = self.cdata['n']
        since_fert = self.cdata['since_fertilization']
        if time_step > 2:
            since_fert = math.floor(since_fert-(time_step/2))
        fert_dose = self.cdata['fertilizer_dose']
        fert = self.cdata['fertilizer_type']
        ba = self.cdata['ba']
        ba_sp = self.cdata['ba_sp']
        ba_ns = self.cdata['ba_ns']
        ts = self.cdata['ts']

        if fert == 1:
            if sp==1:
                a1=0.000003375
                if since_fert==1:
                    a2=1.0
                else:
                    a2=1.0
                a3=3.0
                a4=-2.38
                a5=0.324
                a6=-0.00072
                a7=7.0
                a8=3.0
                a9=0.0
                b1=-0.00005
                b2=1.5
                b3=-0.00564
                b4=0.055
                b5=0.0
                c=0.2
            else:
                #spruce and deciduos species
                a1=0.000000007699
                if since_fert==1:
                    a2=0.8
                else:
                    a2=1.0
                a3=7.0
                a4=-2.02
                a5=0.2
                a6=-0.00045
                a7=9.0
                a8=3.0
                a9=2.2
                b1=-0.00004
                b2=3.0
                b3=-0.2
                b4=0.0
                b5=0.012
                c=0.07
        elif fert == 3:
            a1=0.00000011
            if since_fert==1:
                a2=1.0
            else:
                a2=0.8
            a3=3.0
            a4=-1.4
            a5=0.225
            a6=-0.0002
            a7=22.0
            a8=12.0
            a9=0.0000001
            b1=-0.00004
            b2=3.0
            b3=-0.2
            b4=0.0
            b5=0.012
            c=0.07

        x=since_fert+pow(since_fert/a7,a8)+a9
        #Calculate FA 
        FA=(a1*a2)*pow(x,a3)*math.exp((a4+a5*math.log(fert_dose+1)+a6*fert_dose)*x)
        #Calculate FB
        f1=n*hdom*math.exp(b1*n*hdom)
        if fert == 1:
            if ba_stratum < 0.5:
                coef=0
                f2=0
            else:
                if sp==1:
                    f2=pow(si_100,b2)*math.exp(b3*pow(si_100,(b4*si_100)))
                    coef=ba_sp/ba
                elif sp==2:
                    f2=pow(si_100,b2)*math.exp(b3*si_100)*(1+b5*si_100)
                    coef=ba_ns/ba
                else:
                #deciduos species
                    f2=0.5*(pow(si_100,b2)*math.exp(b3*si_100)*(1+b5*si_100))
                    coef=(ba-(ba_sp+ba_ns))/ba
        elif fert == 3:
            f2=pow(si_100,b2)*math.exp(b3*si_100)*(1+b5*si_100)
            if ba_stratum < 0.5:
                coef=0
            else:
                if sp==1:
                    coef=ba_sp/ba
                elif sp==2:
                    coef=ba_ns/ba
                else:
                #deciduos species
                    coef=(ba-(ba_sp+ba_ns))/ba
                
        FB=f1*f2
        delta_iv=coef*FA*FB*time_step

        #reduction factor for growth effect based on temperature sum
        factor = 1.0
        if ts < 1070:
            factor = -1.88+0.003664*ts-0.0000009143*pow(ts,2)
            if factor < 0.05:
                factor = 0.05
        delta_iv = delta_iv*factor

        return delta_iv

    def get_volume(self,sp,d,h):
        """
        Calculate tree volume for fertilized trees
        """
        if sp in [1,8]:
            if h<3:
                treevol = (math.exp(-4.3115-0.009264*d+0.81849*math.log(h)+2.34597*math.log(2+1.25*d)))/1000
            else:
                treevol = (math.exp(-3.2890 +1.9995*math.log(d) +2.1395*math.log(h)-1.1411*math.log(h-1.3) -0.002847*d))/1000
        elif sp == 2:
            if h<3:
                treevol = (math.exp(-3.9171+1.04666*math.log(h)+1.971*math.log(2+1.25*d)))/1000
            else:
                treevol = (math.exp(-3.7544 +1.8960*math.log(d) +2.8979*math.log(h)-1.6020*math.log(h-1.3) -0.007827*d))/1000
        else: #sp in [3,4,5,6,7,9]
            if h<4:
                treevol = (math.exp(-4.92615-0.010615*d+0.787837*math.log(h)+2.56275*math.log(2+1.25*d)))/1000
            else:
                treevol = (math.exp(-4.4759 +2.0851*math.log(d) +4.0691*math.log(h)-2.7375*math.log(h-1.3) -0.013311*d))/1000

        return treevol

    def get_steps(self,d_iv):
        """
        Calculate iteration steps for diameter and height
        """
        # Scale steps based on delta_iv
        if d_iv > 4:
            d_step=0.05 # about 0,01 % from diameter for an average tree in fertilization
            h_step=0.005 # about 0,001 % from height for an average tree in fertilization
        elif d_iv < 4 and d_iv > 2:
            d_step=0.01 # about 0,005 % from diameter for an average tree in fertilization
            h_step=0.001 # about 0,0005 % from height for an average tree in fertilization
        elif d_iv < 2 and d_iv > 0.5:
            d_step=0.005 # about 0,001 % from diameter for an average tree in fertilization
            h_step=0.0005 # about 0,0001 % from height for an average tree in fertilization
        elif d_iv < 0.5 and d_iv > 0.05:
            d_step=0.0005 # about 0,0001 % from diameter for an average tree in fertilization
            h_step=0.00005 # about 0,00001 % from height for an average tree in fertilization
        else:
            d_step=0.00005 # about 0,00001 % from diameter for an average tree in fertilization
            h_step=0.000005 # about 0,000001 % from height for an average tree in fertilization

        return d_step, h_step

    def store_results(self,d_new,h_new,v_new,tree_ind,s_ind):
        """
        Store results for fertilized trees
        """
        # Scale steps based on delta_iv
        self.data[TREELABEL][tree_ind][self.ind['tree_d']] = d_new
        self.data[TREELABEL][tree_ind][self.ind['tree_h']] = h_new
        self.data[TREELABEL][tree_ind][self.ind['tree_v']] = v_new
        self.data[TREELABEL][tree_ind][-1] = 1.
        self.data[STRATUMLABEL][s_ind][-1] = 1.

        return 1

    def get_height(self,d,D_gM,BA,sp,h,h_before):
        """
        Calculate tree height for fertilized trees
        """
        SC = self.cdata['sc']
        LAT=self.cdata['lat']
        BA_prosb=self.cdata['ba_prosb']
        ALT=self.cdata['alt']
        since_thinning=self.cdata['since_thinning']

        if sp in [1,8]:
            THIN_0_5 = 0
            if since_thinning >= 0 and since_thinning <=5:
                THIN_0_5 = 1

            #SQ24 is categorical variable referring to middle nutrious (mesotrophic, oligo-mesotrophic or oligotrophic) site
            if SC>=2 and SC<=4:
                SQ24=1
            else:
                SQ24=0

            #To avoid extrapolation: Minimum and maximum values
            if d<0.1:
                d=0.1
            if D_gM<3.5:
                if BA<2.7:
                    BA=2.7
                    D_gM=3.5
            if d<=3  and BA>8:
                BA=8
            if d>3 and BA>(8+(d-3)*4):
                BA=8+(d-3)*4
            if BA>36.5:
                BA=36.5

            #Variance correction:
            #If maximum diameter is less than 9 cm, then d=9:
            if d<=9:
                varcor=0.021764931
            else:
                #else variance correction is calculated using maximum diameter:
                varcor=((pow(0.155145,2)+2*0.155145*0.6554*0.07345*(((pow(10,-0.4)-pow(30,-0.4))/(pow(10,-0.4)-pow(30,-0.4)))*-1)+pow(0.07345,2)*
                         pow(((((pow(10,-0.4)-pow(30,-0.4))/(pow(10,-0.4)-pow(30,-0.4)))*-1)),2))+(pow(0.071937,2)+pow(1.389244,2)*(pow(d,-2))))/2

            h_temp =(math.exp(5.369+0.3566*math.log(D_gM)+0.1177*math.log(BA)-0.02638*(LAT/100.0)+0.05115*THIN_0_5+0.0984*SQ24+0.3748*((pow(d,-0.4)-pow(30,-0.4))/(pow(10,-0.4)-pow(30,-0.4)))*-1+
                              0.141*((math.log(D_gM))*((pow(d,-0.4)-pow(30,-0.4))/(pow(10,-0.4)-pow(30,-0.4)))*-1)-0.01263*((BA)*(((pow(d,-0.4)-pow(30,-0.4))/(pow(10,-0.4)-pow(30,-0.4)))*-1))+varcor)+13)/10

        elif sp == 2:
            #To avoid extrapolation: Minimum and maximum values
            if d<0.1:
                d=0.1
            if D_gM<2:
                D_gM=2;
            if d<1:
                if D_gM<3:
                    D_gM=3
                if BA<3:
                    BA=3
            if BA>45.1:
                BA=45.1

            #Variance correction:
            varcor=0.015391832

            h_temp =(math.exp(5.789+0.1889*math.log(D_gM)+0.1665*math.log(BA)-0.02381*(LAT/100.0)+0.4502*((pow(d,-0.4)-pow(30,-0.4))/(pow(10,-0.4)-pow(30,-0.4)))*(-1)+
                              0.1622*((math.log(D_gM))*((pow(d,-0.4)-pow(30,-0.4))/(pow(10,-0.4)-pow(30,-0.4)))*-1)+varcor)+13)/10

        else: #sp in [3,4,5,6,7,9]

            THIN_0_5 = 0;
            if since_thinning >= 0 and since_thinning <=5:
                THIN_0_5 = 1

            #To avoid extrapolation: Minimum and maximum values
            if d<0.1:
                d=0.1
            if D_gM<5:
                D_gM=5
            if d<=3:
                if BA>5:
                    BA=5
            if d>3:
                if BA>(5+(d-3)*5):
                    BA=5+(d-3)*5
            if BA>45:
                BA=45

            #Variance correction:
            #If maximum diameter is less than 9 cm, then d=9:

            if d<=9:
                varcor=0.015597043
            else:
                #else variance correction is calculated using maximum diameter:
                varcor=((pow(0.133116,2)+2*0.133116*0.579*0.004265*(((pow(10,-0.7)-pow(30,-0.7))/(pow(10,-0.7)-pow(30,-0.7)))*-1)+
                         pow(0.004265,2)*pow(((((pow(10,-0.7)-pow(30,-0.7))/(pow(10,-0.7)-pow(30,-0.7)))*-1)),2))+(pow(0.059958,2)+pow(0.923038,2)*(pow(d,-2))))/2;

            h_temp = (math.exp(6.465+0.4743*math.log(D_gM)+0.1339*math.log(BA_prosb/100.0+1)-0.0433*(LAT/100.0)-0.000436*ALT+0.06447*THIN_0_5-0.05942*((pow(d,-0.7)-
                                                                                                                                                        pow(30,-0.7))/(pow(10,-0.7)-pow(30,-0.7)))*-1+0.2123*((math.log(D_gM))*((pow(d,-0.7)-pow(30,-0.7))/(pow(10,-0.7)-pow(30,-0.7)))*-1)-
                               0.004922*((BA)*(((pow(d,-0.7)-pow(30,-0.7))/(pow(10,-0.7)-pow(30,-0.7)))*-1))+varcor)+13)/10

        if h and h_before:
            i_h = h_temp - h_before;

            #Restrict if result is too small
            if i_h<=0.01:
                i_h = 0.01

            #Update h
            h = h + i_h;
        else:
            h=h_temp

        return h
    
    def _constr_prop(self, prop):
        prop = prop / 100.0
        if prop > 1.0:
            prop = 1.0
        elif prop < 0.0:
            prop = 0.0
        return prop
    
    def _modify_sp_pref(self, preforder, forcedprefsp):
        """
        remove non existing species from the preference list and move the
        forced one to first in the list if set
        """
        neworder = []
        allsp = self.data[STRATUMLABEL][:,self.ind['stratum_sp']]
        for sp in preforder:
            if sp in allsp:
                neworder.append(sp)
        if forcedprefsp in neworder:
            popind = neworder.index(forcedprefsp)
            fpsp = neworder.pop(popind)
            neworder.insert(0, fpsp)
        return neworder