#!/usr/bin/env python
# encoding: utf-8
"""
Title: silvicultureop

Silviculture operation implementations

*silvicultureop.py*

Copyright (c) 2007 Simosol Oy. Distributed under the GNU General Public License 2.0.
"""

import math
import numpy
import copy
from collections import defaultdict
from constants import EPSILON, MIN_STEM_NUMBER, TREELABEL, STRATUMLABEL, STANDLABEL
from operations_ext import compute_treestem, bucking_algorithm

class SilvicultureOperation(object):
    def __init__(self, silviculturedata):
        super(SilvicultureOperation, self).__init__()
        self.sdata = silviculturedata
        self.treecount = self.sdata.cdata.get('N')
        self.origtreecount = self.treecount
        self.regen_sp = self.sdata.cdata.get('REGEN_SP')
        self.N_Sp = self.sdata.cdata.get('N_Sp')
        self.N_Ns = self.sdata.cdata.get('N_Ns')
        self.N_sb = self.sdata.cdata.get('N_sb')
    
    def remove_trees(self, stratlist, trees, treenums, sptoremove, treecount, sd):
        """
        Remove trees of given species from a stand. Trees will be removed until
        target density is reached.
    
        stratlist -- list of [stratum indice, parent id and species] -lists
        trees -- ...
        treenums -- ...
        sptoremove -- ...
        treecount -- ...
        targetdensity -- ...
        targetdensity_sp -- ...
        data -- ...
        """
        targetdensity = sd.cdata['targetdensity']
        data = sd.data
        # if the stratums are empty, return without proceeding further
        if stratlist == []:
            return treecount
        # create a list of species in the current stratums
        splist = [s[2] for s in stratlist]
        # If the current species to be removed is not in the list, return
        if not sptoremove in splist:
            return treecount
        
        # construct a list of the indices of the current stratums
        stratums = [s[0] for s in stratlist]
        pind = sd.ind['tree_parentid']
        nind = sd.ind['tree_n']
        tidind = sd.ind['tree_id']
        sidind = sd.ind['stratum_id']
        spind = sd.ind['stratum_sp']
        Nind = sd.ind['stratum_N']
        
        targetdensity_sp_i_stratum = {}
        sp_doubles = {}
        
        # first go through the stratums with small seedlings and no actual
        # diameter classes
        for item in stratlist:
            istratum = item[0]
            sp = item[2]
            # calculate species specific target tree number
            if sp == sptoremove:
                targetdensity_sp, double, sp_doubles = self._set_target(targetdensity,sptoremove,istratum,sp_doubles)
                if double:
                    targetdensity_sp_i_stratum[istratum] = targetdensity_sp
            no_trees = item[3]
            if not no_trees:
                continue  # no small seedlings in the stratum, move along
            if sptoremove != sp:
                continue  # wrong species, move along
            stratumdata = data[STRATUMLABEL][istratum]
            spn = stratumdata[Nind]
            delta = treecount - targetdensity
            if spn > targetdensity_sp + EPSILON:
                if spn < delta and targetdensity_sp == 0:
                    # remove the whole stratum, set tree number as 0 and tag as 2
                    treecount = treecount - spn
                    stratumdata[Nind] = 0
                    stratumdata[-1] = 2
                else:
                    # remove only a portion of the trees in the stratum
                    if spn - delta > targetdensity_sp - EPSILON:
                        stratumdata[Nind] = spn - delta
                        treecount = targetdensity
                    else:
                        stratumdata[Nind] = targetdensity_sp
                        nremoved = abs(spn - targetdensity_sp)
                        treecount = treecount-nremoved
                    # update tree number for the stratum, indicate modification
                    if stratumdata[Nind] < 1:
                        # if stratum's N is less than 1, remove the whole stratum
                        stratumdata[Nind] = 0
                        stratumdata[-1] = 2
                    else: 
                        stratumdata[-1] = 1
                if treecount <= targetdensity or \
                    (stratumdata[Nind] <= targetdensity_sp and not double):
                    return treecount  # we hit the target density, stop the iteration
                
        # tree parent id is the actual simulation indice of parent stratum
        # (used in simulator data matrix)
        # construct a mapping from stratum id to stratum indice to find the
        # correct stratums based on parent info of the trees and mapping from
        # tree id to tree indice
        stratids = list(data[STRATUMLABEL][:,sidind])
        treeids = list(data[TREELABEL][:,tidind])
        stratum_id2ind = dict([(i, stratids.index(i)) for i in stratids])
        tree_id2ind = dict([(i, treeids.index(i)) for i in treeids])
        removed_strata = set([])
        target_set = []
        
        for tree in trees:
            
            # get tree's parent stratum id and indice
            stratum_id = tree[pind]
            stratum_ind = stratum_id2ind[stratum_id]
            
            if stratum_ind not in stratums:
                continue  # skip if stratum is not on the stratum list
            
            # check that current stratum tree species is the one that should
            # be removed during the current iteration
            stratum_sp = data[STRATUMLABEL][stratum_ind, spind]
            if stratum_sp != sptoremove:
                continue
            
            # set targetdensity for double sp strata
            if targetdensity_sp_i_stratum.has_key(stratum_id):
                if stratum_id not in target_set:
                    for i_stratum, targetN in targetdensity_sp_i_stratum.iteritems():
                        if stratum_id == i_stratum:
                            # new targetdensity_sp
                            targetdensity_sp = targetN
                            target_set.append(stratum_id)
                    
            # get stratum's tree number
            spn = data[STRATUMLABEL][stratum_ind, Nind]
            
            # get tree id and indice
            tree_id = tree[tidind]
            tree_ind = tree_id2ind[tree_id]
            
            # if the parent stratum of the current tree has been already
            # removed (N < MIN_STEM_NUMBER), remove the tree and continue
            if stratum_id in removed_strata:
                data[TREELABEL][tree_ind, -1] = 2.
                continue
            
            if abs(treecount - targetdensity) < MIN_STEM_NUMBER:
                continue  # if targetdensity is reached, skip the current tree
            
            if spn > targetdensity_sp + EPSILON:
                if (treecount - targetdensity) - tree[nind] < -MIN_STEM_NUMBER\
                   or spn - tree[nind] < targetdensity_sp - EPSILON:
                    # Remove a part of the current diameter class and tag the tree
                    # as a modified object
                    if (treecount - targetdensity) - tree[nind] < -MIN_STEM_NUMBER:
                        # Remove to target
                        nremoved = treecount - targetdensity
                    else:
                        # Remove to species target 
                        nremoved = abs(spn - targetdensity_sp)
                    data[TREELABEL][tree_ind, -1] = 1.
                else:
                    # Remove the whole diameter class
                    nremoved = tree[nind]
            else:
                continue # if species targetdensity is reached, skip the current tree
            
            # update treecount and both tree- and stratum-level stem numbers
            treecount -= nremoved
            data[TREELABEL][tree_ind, nind] -= nremoved
            data[STRATUMLABEL][stratum_ind, Nind] -= nremoved
            
            # if the diameter class stem number is zero (or close enough),
            # remove the whole diameter class
            if data[TREELABEL][tree_ind, nind] < MIN_STEM_NUMBER:
                data[TREELABEL][tree_ind, nind] = 0.
                data[TREELABEL][tree_ind, -1] = 2.
            
            # if the stratum stem number has been reduced to under minimum
            # or modified, tag the changes to the stratum
            if data[STRATUMLABEL][stratum_ind, Nind] < MIN_STEM_NUMBER:
                data[STRATUMLABEL][stratum_ind][-1] = 2.
                data[STRATUMLABEL][stratum_ind, Nind] = 0.
                removed_strata.add(stratum_id)
            elif nremoved > EPSILON:
                data[STRATUMLABEL][stratum_ind][-1] = 1.
        
        return treecount
    
    def remove_trees_stratum(self, stratlist, sptoremove, treecount, sd):
        """
        Remove trees of given species from a stand. Trees will be removed until
        target density is reached. Model operates on stratum level - tree level 
        not exists yet.
        
        stratlist -- list of [stratum indice, parent id and species] -lists
        sptoremove -- ...
        treecount -- ...
        targetdensity -- ...
        targetdensity_sp -- ...
        data -- ...
        """
        targetdensity = sd.cdata['targetdensity']
        data = sd.data
        # if the stratums are empty, return without proceeding further
        if stratlist == []:
            return treecount
        # create a list of species in the current stratums
        splist = [s[2] for s in stratlist]
        # If the current species to be removed is not in the list, return
        if not sptoremove in splist:
            return treecount
        
        # construct a list of the indices of the current stratums
        Nind = sd.ind['stratum_N']
        
        targetdensity_sp_i_stratum = {}
        sp_doubles = {}
        
        # first go through the stratums with small seedlings and no actual
        # diameter classes
        for item in stratlist:
            istratum = item[0]
            sp = item[2]
            # calculate species specific target tree number
            if sp == sptoremove:
                targetdensity_sp, double, sp_doubles = self._set_target(targetdensity,sptoremove,istratum,sp_doubles)
                if double:
                    targetdensity_sp_i_stratum[istratum] = targetdensity_sp
            no_trees = item[3]
            if sptoremove != sp:
                continue  # wrong species, move along
            stratumdata = data[STRATUMLABEL][istratum]
            spn = stratumdata[Nind]
            delta = treecount - targetdensity
            if spn < delta and targetdensity_sp == 0:
                # remove the whole stratum, set tree number as 0 and tag as 2
                treecount = treecount - spn
                stratumdata[Nind] = 0
                stratumdata[-1] = 2
            else:
                # remove only a portion of the trees in the stratum
                if spn - delta > targetdensity_sp - EPSILON:
                    stratumdata[Nind] = spn - delta
                    treecount = targetdensity
                else:
                    stratumdata[Nind] = targetdensity_sp
                    nremoved = abs(spn - targetdensity_sp)
                    treecount = treecount-nremoved
                # update tree number for the stratum, indicate modification
                if stratumdata[Nind] < 1:
                    # if stratum's N is less than 1, remove the whole stratum
                    stratumdata[Nind] = 0
                    stratumdata[-1] = 2
                else: 
                    stratumdata[-1] = 1
            if treecount <= targetdensity or stratumdata[Nind] <= targetdensity_sp:
                return treecount  # we hit the target density, stop the iteration
        
        return treecount
    
    def define_stem_number_for_supplemental_planting(self, targetdensity):
        """
        Calculate treecount and define planted stem number
        """
        # Tree species (SP) used for regeneration
        SP = self.regen_sp
        # Site class
        SC = self.sdata.cdata['SC']
        # Treecount of regenerated tree species before supplemental planting
        if SP == 1:
            if SC >= 4:
                treecount_regensp = self.N_Sp
            else:
                treecount_regensp = self.N_Ns + self.N_Sp
        elif SP == 2:
            if SC >= 3:
                treecount_regensp = self.N_Ns + self.N_Sp
            else:
                treecount_regensp = self.N_sb + self.N_Ns
        elif SP == 3:
            treecount_regensp = self.N_sb + self.N_Ns
        else:
            if SC >= 4:
                treecount_regensp = self.N_Sp
            else:
                treecount_regensp = self.N_Ns
        
        # Number of seedlings that will be planted
        if treecount_regensp+(targetdensity/4) <= targetdensity:
            # Will fill up to the original planting density
            suppseedlingsnum = targetdensity - treecount_regensp
        else: # Will exceed original planting density
            suppseedlingsnum = targetdensity/4
        return suppseedlingsnum, SP
    
    def _check_sp_restr(self, prop, sp, istratum, pot_removal):
        """
        species specific BA, N or Volume removal constraints
        """
        cdata = self.hdata.cdata
        spprop = cdata['sp_proportion'][sp]
        skipclass = False
        ind = self.hdata.ind
        strata = self.hdata.stratumdata
        spn = strata[istratum, ind['stratum_N']]
        sptargetn = spprop * cdata['target']
        if spn - sptargetn > EPSILON:
            maxnremoval = prop * pot_removal
            if (spn - maxnremoval) - sptargetn < -EPSILON:
                prop = (spn - sptargetn) / maxnremoval
        else:
            skipclass = True
        return prop, skipclass
    
    def _get_double_sp_strata(self,sp,istratum,sp_doubles):
        """
        Check double sp strata.
        """
        spind = self.sdata.ind['stratum_sp']
        Nind = self.sdata.ind['stratum_N']
        data = self.sdata.data
        for istratum2, stratumdata in enumerate(data[STRATUMLABEL]):
            stratum_sp = stratumdata[spind]
            if stratum_sp == sp and istratum2 != istratum:
                if not sp_doubles.has_key(stratum_sp):
                    sp_doubles[stratum_sp] = 0
                sp_doubles[stratum_sp] += stratumdata[Nind]
        if sp_doubles.has_key(sp):
            sp_doubles[sp] += data[STRATUMLABEL][istratum][Nind]
        return sp_doubles
    
    def _set_target(self,targetdensity,sptoremove,istratum,sp_doubles):
        """
        Set target N for species.
        """
        Nind = self.sdata.ind['stratum_N']
        data = self.sdata.data
        # Calculate species scesific target treenumber 
        double = False
        spprop = self.sdata.cdata['sp_proportion'][sptoremove]
        if sptoremove in self.sdata.cdata['sp_proportion']:
            targetdensity_sp = spprop * targetdensity            
            # double sp strata?
            if not sp_doubles.has_key(sptoremove):
                sp_doubles = {}
                sp_doubles = self._get_double_sp_strata(sptoremove,istratum,sp_doubles)
            if sp_doubles.has_key(sptoremove):
                for treenum in sp_doubles.itervalues():
                    # new targetdensity_sp
                    s_N = data[STRATUMLABEL][istratum][Nind]
                    targetdensity_sp = (s_N/treenum) * targetdensity_sp
                    double = True
        else:
            targetdensity_sp = 0
            
        return targetdensity_sp, double, sp_doubles