#!/usr/bin/env python
# encoding: utf-8
"""
*builder.py*

Usage: builder.py <command> <configuration file>
    commands -- build, list, export

Copyright (c) 2009 Simosol Oy.
    Distributed under the GNU General Public License 2.0.
"""

import sys
import os
import traceback
import logging
import glob
import ConfigParser
import pickle
from datetime import datetime
import time
import zipfile
from simo.db.simodb.db import SimoDB
from simo.builder import names
from simo.builder import simoclasses as sc

TRUES = ('1','TRUE','True','true','Yes','yes','y','Y','on')
MODELTYPES = names.mtypes
CHAINTYPES = names.ctypes
XMLORDER = (sc.MESSAGE_TRANSLATION_DEF, sc.LEXICON_TRANSLATION_DEF,
            sc.TEXT2DATA_DEF, sc.OPERATION2MODELCHAINS_DEF,
            sc.PROBLEM_DEFINITION_DEF, sc.OUTPUT_CONSTRAINT_DEF,
            sc.AGGREGATION_DEF, sc.EXPRESSION_DEF)
HELP = "Usage: builder.py <command> <path to ini file>\n" \
       "   commands -- build, list, export"

class Logger(object):
    def __init__(self, path, log_values):
        self.logged_errors = False
        self.logger = logging.getLogger()
        handlers = self.logger.handlers
        for h in handlers:
            self.logger.removeHandler(h)
        console = log_values['console']
        if console:
            clevel = log_values['console_level']
            self._create_handler('console', clevel)
        try:
            file = log_values['file']
            flevel = log_values['file_level']
            self._create_handler('file', flevel, file, path)
        except:
            pass

    def log(self, level, msg):
        if level=='error':
            self.logger.error(msg)
            self.logged_errors = True
        elif level=='warning':
            self.logger.warning(msg)
        elif level=='info':
            self.logger.info(msg)
        elif level=='debug':
            self.logger.debug(msg)

    def _create_handler(self, ltype, llevel, filepath=None, basepath=None):
        formatter = logging.Formatter('%(levelname)s: %(message)s')
        if ltype=='console':
            handler = logging.StreamHandler()
        else:
            dirname = os.path.dirname(filepath)
            dirname = os.path.normpath(os.path.join(basepath, dirname))
            if not os.path.exists(dirname):
                os.makedirs(dirname)
            filepath = os.path.join(dirname, os.path.basename(filepath))
            if os.path.exists(filepath):
                os.remove(filepath)
            handler = logging.FileHandler(filepath)
        handler.setFormatter(formatter)
        loglevel = self._get_logging_level(llevel)
        handler.setLevel(loglevel)
        if self.logger.level > loglevel:
            self.logger.setLevel(loglevel)
        self.logger.addHandler(handler)

    def _get_logging_level(self, level):
            if level=='debug':
                return logging.DEBUG
            elif level=='info':
                return logging.INFO
            elif level=='warning':
                return logging.WARNING
            elif level=='error':
                return logging.ERROR
            else:
                return logging.CRITICAL + 10

class Cfg(object):
    """Processes the ini file for Builder"""
    def __init__(self, inifile):
        """
        inifile -- .ini file to parse
        """
        self.parser = ConfigParser.SafeConfigParser()
        self.parser.read(inifile)
        rawpath = self.parser.get('program_execution', 'base_folder')
        self.path = os.path.expandvars(os.path.expanduser(rawpath))
        log_values = self._parse('logging', False, False)
        self.logger = Logger(self.path, log_values)
        self.db_path = self._prepath(self.parser.get('simo_database', 'path'),
                                    create_path=True)
        self.timestamp = False
        try:
            timestamp = self.parser.get('simo_database', 'timestamp')
            if timestamp in TRUES:
                self.timestamp = True
        except ConfigParser.NoOptionError:
            pass
        self.zip = False
        try:
            zip = self.parser.get('simo_database', 'create_zip')
            if zip in TRUES:
                self.zip = True
        except ConfigParser.NoOptionError:
            pass

    def read(self, item):
        try:
            itm = open(item)
            result = itm.read()
            itm.close()
        except:
            errormsg = "'%s' is not a valid path" % item
            self.logger.log('error', errormsg)
            result = None
        return result

    def _prepath(self, path, create_path=False):
        path = os.path.expandvars(os.path.expanduser(path))
        filepath = os.path.normpath(os.path.join(self.path, path))
        dirpath = os.path.dirname(filepath)
        exists = os.path.exists(dirpath)
        if create_path and not exists:
            os.makedirs(dirpath)
        elif not create_path and not exists:
            filepath = None
        return filepath

    def _parse(self, section, multivalue=True, path_prepend=True,
               create_path=False):
        options = self.parser.options(section)
        results = {}
        for option in options:
            init_value = self.parser.get(section, option)
            if multivalue:
                values = init_value.split(';')
                results[option] = [value for value in values]
            else:
                results[option] = init_value
        if path_prepend:
            for option, result in results.items():
                if multivalue and result[0] != '':
                    results[option] = [self._prepath(item, create_path) \
                                       for item in result]
                elif not multivalue and result != '':
                    results[option] = self._prepath(result, create_path)
        return results

    def _parse_bool(self, section):
        options = self.parser.options(section)
        results = []
        for option in options:
            val = self.parser.get(section, option)
            if val in TRUES:
                results.append(option)
        return results

class BuildCfg(Cfg):
    def __init__(self, inifile):
        super(BuildCfg, self).__init__(inifile)
        self.schemas = self._parse('schema', False)
        self.schemas_modelbase = self._parse('schema.modelbase',
                                             multivalue=False)
        self.xmls = self._parse('xml')
        self.xmls_modelbase = self._parse('xml.modelbase')
        self.xmls_modelchain = self._parse('xml.modelchain')
        self.exec_modelbase = self._parse('executable.modelbase',
                                          path_prepend=False)
        self.typedef = self._prepath(self.parser.get('typedef', 'path'))
        self.new_db = False
        create_new = self.parser.get('simo_database', 'create_new')
        if create_new in TRUES:
            self.new_db = True

class ExportCfg(Cfg):
    def __init__(self, inifile):
        super(ExportCfg, self).__init__(inifile)
        self.db_path = self.parser.get('simo_database', 'path')
        self.general_exports = self._parse('general_export', multivalue=False,
                                           path_prepend=False)
        self.model_exports = self._parse('model_export', multivalue=False,
                                           path_prepend=False)
        self.modelchain_exports = self._parse('modelchain_export',
                                            multivalue=False, path_prepend=False)
        self._check_for_excludes(self.general_exports)
        self._check_for_excludes(self.model_exports)
        self._check_for_excludes(self.modelchain_exports)

    def _check_for_excludes(self, data):
        removes = []
        for name, path in data.items():
            if path == '':
                removes.append(name)
        for name in removes:
            del data[name]

class ListCfg(Cfg):
    def __init__(self, inifile):
        super(ListCfg, self).__init__(inifile)
        self.general_names = self._parse_bool('general')
        self.model_names = self._parse_bool('modelbase')
        self.modelchain_names = self._parse_bool('modelchain')

class Builder(object):
    def __init__(self, command, cfg):
        """
        command -- command builder is to execute
        parameters -- parameters given to the command
                     (an ini filename or a list of types if command==list)
        """
        self.command = command
        self.lexicon = None
        self.cfg = cfg
        if command == 'build':
            self.build()
        elif command == 'export':
            self.export()
        elif command == 'list':
            self.list_names()
        else:
            errormsg = 'Command must be one of build, export or list'
            cfg.logger.log('error', errormsg)
            cfg.logger.log('error', 'Terminating program execution')
            sys.exit(1)

    def build(self):
        """Gets xml files and shoves them into the simodb as xml text and
        associated python objects build from the xml content
        """
        cfg = self.cfg
        cfg.logger.log('info', 'Beginning build process...')
        self.new_db = cfg.new_db
        self.db = SimoDB(cfg.db_path, self.new_db)
        if cfg.timestamp:
            f = open(cfg.db_path + '.ver', 'wb')
            t = datetime.now()
            epoch_secs = int(time.mktime(t.timetuple()))
            pickle.dump(epoch_secs, f)
            f.close()
        try:
            if self.new_db or cfg.typedef is not None:
                typedef = cfg.read(cfg.typedef)
                self.db.typedef = typedef
            self._set_schemas()
            self._set_modelbase_schemas()
            self._add_lexicon()
            self._add_xmls()
            self._add_text2operation()
            self._add_xmls_modelbase()
            self._add_xmls_modelchain()
            self._add_simulation_control()
            self._check_validity(cfg.logger)
            self._check_variable_usage(cfg.logger)
        except Exception, e:
            type, value, tb = sys.exc_info()
            tb = traceback.format_exc(tb)
            msg = '%s: %s' % (e.args[0], tb)
            cfg.logger.log('error', msg)
            sys.exit(1)
        finally:
            self.db.close()
            if cfg.zip:
                self._zip_db()
            cfg.logger.log('info', 'Build finished')

    def export(self):
        """
        Writes out the xml text stored in simodb as xml text files
        """
        self.cfg.logger.log('info', 'Beginning export...')
        if not os.path.exists(self.cfg.db_path):
            self.cfg.logger.log('error', 'No database in %s' % \
                                os.path.abspath(self.cfg.db_path))
            return
        try:
            self.db = SimoDB(self.cfg.db_path, create_new=False,
                             read_only=True)
            generals = {}
            models = {}
            mchains = {}
            genexp = self.cfg.general_exports
            generals['xmls'] = self._get_xmls(genexp)
            generals['schemas'] = self._get_schemas(genexp)
            modelexp = self.cfg.model_exports
            models['xmls'] = self._get_typed_xmls(modelexp, sc.MODELBASE_DEF)
            models['schemas'] = self._get_modelbase_schemas(modelexp)
            mcexp = self.cfg.modelchain_exports
            mchains['xmls'] = self._get_typed_xmls(mcexp, sc.MODELCHAIN_DEF)
            mc = {sc.MODELCHAIN_DEF: sc.MODELCHAIN_DEF}
            mchains['schemas'] = self._get_schemas(mc)
            self._write_files(generals)
            self._write_files(models, sc.MODELBASE_DEF)
            self._write_files(mchains)
        except Exception, e:
            self.cfg.logger.log('error', e.args[0])
            sys.exit(1)
        finally:
            self.db.close()
            self.cfg.logger.log('info', 'Export finished')

    def list_names(self):
        """
        Lists by xml content type all the xml names stored in the simodb.
        The xml names will be used as file names when exporting. These
        names are also used in runner config to refer to specific xml
        content which should be used in the computation
        """
        self.cfg.logger.log('info', 'Beginning listing...')
        if not os.path.exists(self.cfg.db_path):
            self.cfg.logger.log('error', 'No database in %s' % \
                                os.path.abspath(self.cfg.db_path))
            return
        self.db = SimoDB(self.cfg.db_path, create_new=False,
                             read_only=True)
        try:
            self._list_names(self.cfg.general_names)
            self._list_names(self.cfg.model_names, sc.MODELBASE_DEF)
            self._list_names(self.cfg.modelchain_names, sc.MODELCHAIN_DEF)
        except Exception, e:
            self.cfg.logger.log('error', e.args[0])
            sys.exit(1)
        finally:
            self.db.close()
            self.cfg.logger.log('info', 'Listing finished')

    def _add_lexicon(self):
        """Adds xml content of lexicon"""
        cfg = self.cfg
        simoclass = sc.LEXICON_DEF
        xmlfiles = cfg.xmls.get(simoclass)
        if len(xmlfiles) > 1:
            extraxmls = xmlfiles[1:]
            xmlfiles = xmlfiles[:1]
        else:
            extraxmls = None
        if xmlfiles != '':
            data = [None, None]
            self._add_xml(simoclass, data, xmlfiles, True)
        elif self.new_db:
            errormsg = "'%s' not set in ini-file" % simoclass
            cfg.logger.log('error', errormsg)
        if extraxmls not in ('', None):
            data = [None, None]
            self._add_xml(simoclass, data, extraxmls)
        
        self.lexicon = self.db.get_obj(simoclass, 'lexicon')
        if extraxmls not in ('', None):
            # get the extra lexicons and merge them to the master lexicon
            for xml in extraxmls:
                xmlname = os.path.basename(xml).rsplit('.', 1)[0]
                extra = self.db.get_obj(simoclass, xmlname)
                if not self.lexicon.merge(extra, self.cfg.logger):
                    errormsg = "Could not merge extra lexicon '%s' to "\
                               "master lexicon" % xmlname
                    self.cfg.logger.log('error', errormsg)

    def _add_schema(self, simoclass, schema, atype):
        """Writes the schema to the database"""
        try:
            if atype:
                self.db.set_schema(simoclass, schema, atype)
            else:
                self.db.set_schema(simoclass, schema)
        except ValueError, e:
            errormsg = "Schema for simoclass '%s': %s" % (simoclass, e.args[0])
            self.cfg.logger.log('error', errormsg)

    def _add_simulation_control(self):
        """
        Adds xml content of simulation control. This requires that models
        and modelchains have already been set
        """
        cfg = self.cfg
        simoclass = sc.SIMULATION_CONTROL_DEF
        chains = self._get_chains()
        xmlfiles = cfg.xmls.get(simoclass)
        if xmlfiles!='':
            data = [None, None, self.lexicon, chains]
            self._add_xml(simoclass, data, xmlfiles)
        elif self.new_db:
            errormsg = "'%s' not set in ini-file" % simoclass
            cfg.logger.log('error', errormsg)

    def _add_text2operation(self):
        """
        Adds xml content of operation conversion definition.
        This requires that operation2modelchains has already been set
        """
        cfg = self.cfg
        simoclass = sc.TEXT2OPERATION_DEF
        xmlfiles = cfg.xmls.get(simoclass)
        if xmlfiles!='':
            for i in range(len(xmlfiles)):
                if xmlfiles[i]=='':
                    continue
                op2mc = self._get_op2mc(i)
                if op2mc is None:
                    errormsg = "No operation2modelchains mapping object found"\
                               " for '%s'" % xmlfiles[i]
                    self.cfg.logger.log('error', errormsg)
                    continue
                data = [None, None, self.lexicon, op2mc]
                self._add_xml(simoclass, data, (xmlfiles[i],))
        elif self.new_db:
            errormsg = "'%s' not set in ini-file" % simoclass
            cfg.logger.log('error', errormsg)

    def _add_xml(self, simoclass, data, xmlfiles, master_lexicon=False):
        """
        Responsible for actually putting the xml content into simodb
        """
        for xml in xmlfiles:
            if xml in ('', None):
                continue
            if simoclass == sc.LEXICON_DEF and master_lexicon:
                xmlname = 'lexicon'
            else:
                xmlname = os.path.basename(xml).rsplit('.', 1)[0]
            data[0] = xmlname
            data[1] = self.cfg.read(xml)
            try:
                result = self.db.set_xml(simoclass, data)
                if result[0] is not None:
                    errormsg = 'add_xml for %s XML %s errors: %s' % (simoclass,
                            xmlname, result[0])
                    self.cfg.logger.log('error', errormsg)
                if result[2]!=[]:
                    warningmsg = 'add_xml for %s XML %s warnings: %s'\
                            % (simoclass, xmlname, '\n'+'\n'.join(result[2]))
                    self.cfg.logger.log('warning', warningmsg)
                if result[3]!=set([]):
                    errormsg = 'add_xml for %s XML %s errors: %s' % (simoclass,
                            xmlname, '\n'+'\n'.join(result[3]))
                    self.cfg.logger.log('error', errormsg)
            except ValueError, e:
                errormsg = "For simoclass '%s', xml '%s': %s" % (simoclass,
                            xmlname, e.args[0])
                self.cfg.logger.log('error', errormsg)

    def _add_xmls(self):
        """
        Adds xml content in specified order for all XmlObject subclasses
        except for the modelbase and modelchain
        """
        cfg = self.cfg
        for simoclass in XMLORDER:
            xmlfiles = cfg.xmls.get(simoclass)
            if xmlfiles not in  (None, ''):
                data = [None, None, self.lexicon]
                self._add_xml(simoclass, data, xmlfiles)
            elif self.new_db:
                errormsg = "'%s' not set in ini-file" % simoclass
                cfg.logger.log('error', errormsg)

    def _add_xmls_modelbase(self):
        """Adds the xml content for each subclass of the modelbase"""
        cfg = self.cfg
        for modeltype, xmlfiles in cfg.xmls_modelbase.items():
            if modeltype not in MODELTYPES:
                errormsg = "Invalid model type '%s' in ini file. Valid types "\
                           "are: %s" % (modeltype, MODELTYPES)
                self.cfg.logger.log('error', errormsg)
                continue
            if xmlfiles!='':
                dirs = self.cfg.exec_modelbase.get(modeltype)
                if dirs is None:
                    errormsg = "No model implementation directories for model "\
                            "type '%s'" % (modeltype)
                    self.cfg.logger.log('error', errormsg)
                    continue
                data = [None, None, self.lexicon, dirs, modeltype]
                self._add_xml(sc.MODELBASE_DEF, data, xmlfiles)
            elif self.new_db:
                errormsg = "'%s' xml definition not set in ini-file" % modeltype
                cfg.logger.log('error', errormsg)

    def _add_xmls_modelchain(self):
        """Adds the xml content for model chains"""
        cfg = self.cfg
        model_impl = self._get_models()
        if model_impl is None:
            errormsg = "No model implementations could be loaded"
            self.cfg.logger.log('error', errormsg)
            return
        for chaintype, xmldirs in cfg.xmls_modelchain.items():
            if chaintype not in CHAINTYPES:
                errormsg = "Invalid chain type '%s' in ini file. Valid types"\
                           " are: %" % (chaintype, CHAINTYPES)
                self.cfg.logger.log('error', errormsg)
                continue
            if xmldirs!='':
                xmlfiles = []
                for path in xmldirs:
                    if path is None:
                        errormsg = "Invalid path in xml modelchain setting "\
                                "'%s'" % chaintype
                        self.cfg.logger.log('error', errormsg)
                        continue
                    if path != '':
                        xmlfiles += glob.glob(path+'/*.xml')
                data = [None, None, self.lexicon, chaintype, model_impl]
                self._add_xml(sc.MODELCHAIN_DEF, data, xmlfiles)
            elif self.new_db:
                errormsg = "'%s' modelchains not set in ini-file" % chaintype
                cfg.logger.log('error', errormsg)

    def _check_validity(self, logger):
        valid = self.db.get_valid()
        for key, val in valid.items():
            if not val['valid']:
                if ('names' in val and len(val['names']) == 0) or\
                   ('types' in val and len(val['types']) == 0):
                    logger.log('warning', 'The class \''+key+\
                               '\' hasn\'t been built.')
                else:
                    logger.log('error', 'The class \''+key+\
                               '\' hasn\'t been built successfully: '+\
                               str(val))

    def _check_variable_usage(self, logger):
        """
        Check which of the variables defined in the lexicon are actually used
        and log the unused variables
        """
        inactive = self.lexicon.get_inactive_variables()
        for level, variables in inactive.items():
            msg = "Total of %d inactive variables on level '%s': %s" % \
                  (len(variables), level, ', '.join(list(variables)))
            logger.log('debug', msg)

    def _get_chains(self):
        """
        Collects all modelchains out of the simodb into a dictionary keyed
        by chain type
        """
        chains = {}
        for ctype in CHAINTYPES:
            ret = self.db.get_obj(sc.MODELCHAIN_DEF, None, ctype)
            if ret is None:
                warnmsg = "No modelchains for chain type '%s'" % ctype
                self.cfg.logger.log('warning', warnmsg)
                chains[ctype] = {}
            else:
                chains[ctype] = ret
        return chains

    def _get_modelbase_schemas(self, mtypes):
        """
        Collects all modebase schemas out of the simodb into a dictionery
        keyed by the model type
        """
        emptydicts = [{} for i in range(len(mtypes))]
        schemas = dict(zip(mtypes, emptydicts))
        for mtype in mtypes.keys():
            ret = self.db.get_schema(sc.MODELBASE_DEF, mtype)
            if ret is None:
                self.cfg.logger.log('error', "No schema for '%s'" % mtype)
            else:
                schemas[mtype] = ret
        return schemas

    def _get_models(self):
        """
        Collects all model instances from the db into a dictionary keyed
        by model type. Will add the models from different xml names into
        a single dictionary, hence also checking that under two different
        collections of models (xml names) there are no models with identical
        names
        """
        model_impl = {}
        for mtype in MODELTYPES:
            models = {}
            modeldicts = self.db.get_obj(sc.MODELBASE_DEF, None, mtype)
            if modeldicts is None:
                continue
            for xmlname, dict in modeldicts.items():
                for mname, mimpl in dict.items():
                    if mname in models:
                        errormsg = "Duplicate model name in different model "\
                                "libraries: '%s', '%s'" % (mtype, mname)
                        self.cfg.logger.log('error', errormsg)
                    else:
                        models[mname] = mimpl
            model_impl[mtype] = models
        return model_impl

    def _get_op2mc(self, ind):
        """
        Get's the operation2modelchains object from the database. The
        text2operation and operation2modelchains settings in ini are matched
        by location; i.e., the n'th operation2modelchains xml file is
        collected for the n'th text2operation xml file in ini
        """
        xmlfiles = self.cfg.xmls['operation2modelchains']
        if ind>(len(xmlfiles) - 1):
            return None
        xml = xmlfiles[ind]
        xmlname = os.path.basename(xml).rsplit('.', 1)[0]
        op2mc = self.db.get_obj(sc.OPERATION2MODELCHAINS_DEF, xmlname)
        return op2mc

    def _get_typed_xmls(self, types, simoclass):
        """
        Gets the xml dictionaries for the given type (keyed by xml name, having
        value of xml text) from the simodb.
        """
        emptydicts = [{} for i in range(len(types))]
        xmls = dict(zip(types, emptydicts))
        for atype, path in types.items():
            # returns (xml dict, ok dict)
            res = self.db.get_xml(simoclass, None, atype)
            if res==(None, None):
                errormsg = "No xml found for '%s', '%s'"\
                        % (simoclass, atype)
                self.cfg.logger.log('error', errormsg)
                continue
            else:
                xmls[atype] = res
            for name in xmls[atype][0].keys():
                if not xmls[atype][1][name]:
                    errormsg = 'There has been problems with %s' % name
                    self.cfg.logger.log('error', errormsg)
            xmls[atype] = [xmls[atype][0], path]
        return xmls

    def _get_schemas(self, simoclasses):
        """
        Gets the schemas for given simo classes from the simodb
        """
        schemas = {}
        for simoclass in simoclasses.keys():
            ret = self.db.get_schema(simoclass)
            if ret is None:
                errormsg = "No schema found for '%s'" % simoclass
                self.cfg.logger.log('error', errormsg)
            else:
                schemas[simoclass] = ret
        return schemas

    def _get_xmls(self, simoclasses):
        """
        Gets the xml name, xml content dictionaries for the given simo classes
        from the simodb
        """
        xmls = {}
        for simoclass, path in simoclasses.items():
            # returns (xml dict, ok dict)
            ret = self.db.get_xml(simoclass)
            if ret==(None, None):
                errormsg = "No xml found for '%s', '%s'" % (simoclass, atype)
                self.cfg.logger.log('error', errormsg)
                continue
            else:
                xmls[simoclass] = ret
            for name in xmls[simoclass][0].keys():
                if not xmls[simoclass][1][name]:
                    errormsg = 'There has been problems with %s' % name
                    self.cfg.logger.log('error', errormsg)
            xmls[simoclass] = [xmls[simoclass][0], path]
        return xmls

    def _list_names(self, list, simoclass=None):
        """
        Logs the xml names stored in the simodb for the given simo class
        """
        if simoclass is not None:
            for atype in list:
                namelist = self.db.get_names(simoclass, atype)
                self._log_names(namelist, simoclass, atype)
        else:
            for simoclass in list:
                namelist = self.db.get_names(simoclass)
                self._log_names(namelist, simoclass)

    def _log_names(self, namelist, simoclass, atype=None):
        """
        Logs the xmls names, possibly by subtype (for modelbase and modelchain)
        """
        if namelist:
            if atype is None:
                msg = "Names for '%s': %s" % (simoclass, ', '.join(namelist))
                self.cfg.logger.log('info', msg)
            else:
                msg = "Names for '%s' '%s': %s" % \
                      (simoclass, atype, ', '.join(namelist))
                self.cfg.logger.log('info', msg)
        else:
            if atype is None:
                self.cfg.logger.log('info', "No names for '%s'" % simoclass)
            else:
                msg = "No names for '%s' '%s'" % (simoclass, atype)
                self.cfg.logger.log('info', msg)

    def _set_modelbase_schemas(self):
        """Adds schemas for each of the subclass of the modelbase"""
        cfg = self.cfg
        for modeltype, schemafile in cfg.schemas_modelbase.items():
            if schemafile!='':
                schema = cfg.read(schemafile)
                self._add_schema(sc.MODELBASE_DEF, schema, modeltype)
            elif self.new_db:
                errormsg = "Schema file for '%' is not set in ini-file" \
                           % modeltype
                cfg.logger.log('error', errormsg)


    def _set_schemas(self):
        """Adds schemas for all XmlObject subclassess except the modelbase"""
        cfg = self.cfg
        for simoclass, schemafile in cfg.schemas.items():
            if schemafile!='':
                schema = cfg.read(schemafile)
                self._add_schema(simoclass, schema, None)
            elif self.new_db:
                errormsg = "Schema file for '%' is not set in ini-file" \
                           % simoclass
                cfg.logger.log('error', errormsg)

    def _write_files(self, files, extraname=''):
        """
        Writes the xml (schema / xml content) into files
        """
        base = self.cfg.path
        if not os.path.exists(base):
            os.mkdir(base)
        for name, schema in files['schemas'].items():
            path = os.path.join(base, 'schemas')
            if not os.path.exists(path):
                os.mkdir(path)
            if extraname != '':
                name = '_'.join((name,extraname))
            fpath = os.path.join(path, name+'.xsd')
            try:
                file = open(fpath, 'w')
                file.write(schema)
            finally:
                file.close()
        for name, xmldata in files['xmls'].items():
            if xmldata=={}:
                continue
            path = os.path.join(base, xmldata[1])
            if not os.path.exists(path):
                os.makedirs(path)
            for name, xml in xmldata[0].items():
                fpath = os.path.join(path, name+'.xml')
                try:
                    file = open(fpath, 'w')
                    file.write(xml)
                finally:
                    file.close()

    def _zip_db(self):
        """
        Creates a zip file of the simo db file
        """
        db_path = self.cfg.db_path
        zipper = zipfile.ZipFile(db_path + '.zip', 'w', zipfile.ZIP_DEFLATED)
        zipper.write(db_path, os.path.basename(db_path), zipfile.ZIP_DEFLATED)
        zipper.close()

def test():
    import doctest
    doctest.testfile('test/builder-py.rst')

def main():
    options = ('build', 'export', 'list', 'test')
    if len(sys.argv)<2:
        print HELP
        sys.exit(0)
    command = sys.argv[1]
    if command=='test':
        test()
        sys.exit(1)
    if len(sys.argv) < 3:
        errormsg = "Command not given, use one of build, export or list"
        print errormsg
        sys.exit(1)
    if sys.argv[1] not in options:
        errormsg = "Command option '%s' not defined. Use one of build, "\
                   "export or list" % sys.argv[1]
        print errormsg
        sys.exit(1)

    # Check that simulation .ini file exists
    inifile = ''
    if len(sys.argv) == 3:
        inifile = sys.argv[2]
        if os.path.isfile(inifile) == 0:
            print 'Configuration file %s is not a valid file'  % inifile
            sys.exit(1)
    elif len(sys.argv) > 3:
        print 'Too many command line options %s' % ' '.join(sys.argv[3:])
        sys.exit(1)

    if command=='build':
        cfg = BuildCfg(inifile)
    elif command=='export':
        cfg = ExportCfg(inifile)
    elif command=='list':
        cfg = ListCfg(inifile)

    builder = Builder(command, cfg)
    if builder.cfg.logger.logged_errors:
        sys.exit(1)

if __name__ == '__main__':
    main()
