#include <Python.h>
#include <math.h>
#include <string.h>
#include <stdio.h>
#include "numpy/arrayobject.h"

#define PI 3.14159265
#define EPSILON 0.00001


////////////////////////////////////////////////////////////////////////////////
//
//  simooperations_ext - C extensions for computationally heavy routines in
//  operation models, such as computing the tree stem and optimization of tree
//  cutting (bucking)

static char doc[] = "This module provides C implementations for simooperation -module functions.";

static double
truncated_cone_volume(double length, double d1, double d2)
{
	/*
    *	Function: __truncated_cone_volume
    *
    *	Calculate the volume of a stem segment with truncated cone formula.
    *
    *	V = (1/3) * pi * h * (d1**2 + d1 * d2 + d**2)
    */
    double r1 = d1 / 2;
    double r2 = d2 / 2;
    return (PI * length * (pow(r1, 2) + r1 * r2 + pow(r2, 2))) / 3;
}

static double
height_of_stump(int species, double d, double h)
{
	/*
    *	Function: __height_of_stump
    *
    *	Calculate the heigth of stump (the lowest cutting point) of tree stem.
    */
    double hs;
    if (species == 1) {
		// pine
		hs = (0.4456 * d + 0.0952 * h) / 100;
	}
    else if (species == 2) {
		// spruce
		hs = (0.5089 * d + 0.56 * h) / 100;
	}
    else {
		// birch and other species
		hs = (0.4862 * d + 0.4979 * h) / 100;
	}
    if (hs < 0.1) {
		hs = 0.1;
	}
    return hs;
}

static double
x_at_h(int species, double h, double height)
{
    /*
    *	Function: x_at_h
    *
    *	Calculate the factor x at height h with taper curve function.
    */
	double y;
    y = 1 - (h / height);
    if (species == 1 || species == 8) {
		// pine and other coniferous
        return (2.1288*y-0.63157*pow(y,2)-1.6082*pow(y,3)
				+2.4886*pow(y,5)-2.4147*pow(y,8)+2.3619*pow(y,13)
				-1.7539*pow(y,21)+1.0817*pow(y,34));
	}
    else if (species == 2) {
		// spruce
        return (2.3366*y-3.2684*pow(y,2)+3.6513*pow(y,3)
				-2.2608*pow(y,5)+2.1501*pow(y,13)-2.7412*pow(y,21)
				+1.8876*pow(y,34));
	}
    else {
		// birch and other species
        return (0.93838*y+4.1060*pow(y,2)-7.8517*pow(y,3)
				+7.8993*pow(y,5)-7.5018*pow(y,8)+6.3863*pow(y,13)
				-4.3918*pow(y,21)+2.1604*pow(y,34));
	}
}

static char compute_treestem_doc[] = "Compute a tree stem with taper curve algorithm.";

static PyObject *
compute_treestem(PyObject *self, PyObject *args)
{
	/*
	*	Function: compute_tree
	*
	*	Compute a tree with taper curve algorithm.
	*/

	int i, species, piececount;
	double dbh, height, divisor;
	double d_2h, y, hstump, htop, dtop, dbottom, stemvol;
    PyArrayObject *tree;

	/* Parse all input parameters into given pointers */
	if (!PyArg_ParseTuple(args,"O!iiddd:compute_tree",
		&PyArray_Type, &tree, &species, &piececount,
		&dbh, &height, &divisor))
		return PyInt_FromLong(-1);

	
	hstump = height_of_stump(species, dbh, height);
	d_2h = dbh / x_at_h(species, 1.3, height);
	htop = hstump + divisor;
	dbottom = d_2h * x_at_h(species, hstump, height);
	stemvol = 0;

	for (i=0; i<piececount; i++) {
		y = 1.0 - (htop / (double) height);
		dtop = (double)d_2h * x_at_h(species, htop, height);

		stemvol += truncated_cone_volume(divisor, dbottom, dtop);

		*(double *)(tree->data + i * tree->strides[0] + 0 * tree->strides[1]) = (double)htop;
		*(double *)(tree->data + i * tree->strides[0] + 1 * tree->strides[1]) = (double)dtop;
		*(double *)(tree->data + i * tree->strides[0] + 2 * tree->strides[1]) = (double)stemvol;
		htop += (double)divisor;
		if (htop > (double)height) {
			htop = (double)height;  /* last piece of the tree */
		}
		dbottom = dtop;
	}
	return PyFloat_FromDouble(stemvol);
}

static char bucking_algorithm_doc[] = "Optimize a stem for maximized value.";

static PyObject *
bucking_algorithm(PyObject *self, PyObject *args)
{
	/*
	*	Function: bucking_algorithm
	*
	*	Optimize a stem for maximized value.
	*
	*/
	int nsegments, nclasses;
	double divr;
    int i, j, icurseg;
	double dmin, dcur, volume, value, totvol, totval;
	PyArrayObject *tree;        // tree stem structure
	PyArrayObject *P;           // price matrix
	PyArrayObject *vol;         // container for the cumulative volumes
	PyArrayObject *val;         // container for the cumulative values
	PyArrayObject *grd;         // container for the assortment grades
	PyArrayObject *lastlogpos;  // container for last log positions

	/* Parse all input parameters into given pointers */
	if (!PyArg_ParseTuple(args,"iidO!O!O!O!O!O!:bucking_algorithm",
		&nsegments, &nclasses, &divr, &PyArray_Type, &tree,
		&PyArray_Type, &P, &PyArray_Type, &vol,
		&PyArray_Type, &val, &PyArray_Type, &grd,
		&PyArray_Type, &lastlogpos))
		return PyInt_FromLong(-1);

	/* Go through each stem segment and each price matrix class
	*/
	for (i=0; i<nsegments; i++) {
		for (j=0; j<nclasses; j++) {
		    /* Get the index for the current top segment, if the current
		    segment index is less than maximum number of segments, calculate
		    the cumulative volumes, values etc
		    */
			icurseg = (int)(i + (*(double *)(P->data + j * P->strides[0] + 2 * P->strides[1]) / divr + 0.5));
			if (icurseg < nsegments) {
			    /* Get the top diameter of the current tree segment and check
			    if the diameter is greater than the minimum assortment
			    diameter, defined in the price matrix P.
			    */
				dcur = *(double *)(tree->data + icurseg * tree->strides[0] + 1 * tree->strides[1]);
				dmin = *(double *)(P->data + j * P->strides[0] + 1 * P->strides[1]);
				if ((dcur-dmin) >= -EPSILON) {
				    /* Calculate current tree stem's (bottom at i, top at
				    icurseg) volume and the cumulative value.
				    */
					if (i > 0) {
						volume = *(double *)(tree->data + icurseg * tree->strides[0] + 2 * tree->strides[1])
							- *(double *)(tree->data + i * tree->strides[0] + 2 * tree->strides[1]);
					} else {
						volume = *(double *)(tree->data + icurseg * tree->strides[0] + 2 * tree->strides[1]);
					}
					value = volume * *(double *)(P->data + j * P->strides[0] + 3 * P->strides[1]);
					totvol = volume + *(double *)(vol->data + i * vol->strides[0]);
					totval = value + *(double *)(val->data + i * val->strides[0]);
					/* If total value, calculated as the value of the current
					log plus all the logs below i, is higher than the existing
					value in icurseg, overwrite the existing value with totval
					*/
					if ((totval - *(double *)(val->data + icurseg * val->strides[0])) >  EPSILON) {
						*(double *)(val->data + icurseg * val->strides[0]) = (double)totval;
						//*(double *)(vol->data + icurseg * vol->strides[0]) = (double)volume;
						*(double *)(vol->data + icurseg * vol->strides[0]) = (double)totvol;
						*(double *)(grd->data + icurseg * grd->strides[0])
							= *(double *)(P->data + j * P->strides[0] + 0 * P->strides[1]);
						*(int *)(lastlogpos->data + icurseg * lastlogpos->strides[0]) = i;
					}
				}
			}
		}
	}
	return PyInt_FromLong(1);
}


/* Define the methods of the extension module */
static PyMethodDef methods[] = {
	{"compute_treestem", compute_treestem, METH_VARARGS, compute_treestem_doc},
	{"bucking_algorithm", bucking_algorithm, METH_VARARGS, bucking_algorithm_doc},
	{NULL, NULL}
};

/* Define the extension method name etc.*/
PyMODINIT_FUNC
initoperations_ext(void)
{
	Py_Initialize();
	import_array();
	Py_InitModule3("operations_ext", methods, doc);
}
