#!/usr/bin/env python
# encoding: utf-8
"""
Stand (management unit) -level optimizer utilizing Hooke & Jeeves algorithm

*hjopt.py*

Usage: hjopt.py <simo configuration file> <hj configuration file>

Copyright (c) 2013 Simosol Oy. Distributed under the GNU General Public License 2.0.
"""

import sys
import os
#import cProfile
#import pstats
import traceback
import pdb
import datetime
import numpy
from copy import deepcopy as dc
from optparse import OptionParser

from runner import Runner, _get_ini_logger, _get_translators
from simo.utils.runnerconfig import Config
from simo.simulation.sim import Simulator
from simo.builder import simoclasses as sc

#from simo.utils.logger import Logger
#from simo.input.importdata import DataImporter
#from simo.input.importops import OperationImporter
#from simo.optimization.genetic_algorithm import GeneticAlgorithm
#from simo.optimization.simulated_annealing import SimulatedAnnealing
#from simo.optimization.tabusearch import TabuSearch
#from simo.optimization.hero import HERO
#from simo.optimization.jlp import JLP
#from simo.output.out import Out
#import simo.db.simodb.db as simodb
#import simo.db.datadb.db as datadb
#from simo.utils.utils import print_stats


# SIMO simulator-related constants, such as db table, level and variable names
# NB: these should match the names used in lexicon and other simulator XMLs
DATA_LINK = 'data_link'
OP_LINK = 'op_link'
OP_RES = 'op_res'
OPRES_CASH_FLOW = 'cash_flow'
STAND = 'comp_unit'
STAND_AGE = 'Age'
STAND_AREA = 'AREA'
STAND_NUM_OF_STEMS = 'N'
STAND_SPECIES = 'MAIN_SP'
STAND_VOLUME = 'V'


# harvest operation types
THINNING = 'thinning'
CLEARCUT = 'clearcut'
THINNINGS = ('first_thinning', 'thinning')
REGENS = ('clearcut', 'seedtree_position')


# thinning types
LOWER_THINNING = 'lower'
#UPPER_THINNING = 'upper'


# base year
BASE_YEAR = 1000


# harvest ctrl parameters
MIN_YEARS_BTW_HARVESTS = 10


# sqls
STAND_ATTR_SQL = "SELECT %s, %s, %s, %s, D_gM, H_gM, BA FROM %s WHERE id=?" %\
                 (STAND_SPECIES, STAND_AGE, STAND_VOLUME, STAND_NUM_OF_STEMS,
                  STAND)


class Harvest(object):
    def __init__(self, type=THINNING, age=1, proportion=1.0):
        self.type = type
        self.age = age
        self.proportion = proportion
        self.incr_age = None
        self.incr_proportion = None


class Schedule(object):
    def __init__(self):
        self.thinnings = []
        self.clearcut = None
        self.ir = 0.0
        self.npv = 0.0


class Stand(object):
    def __init__(self, id='', sp=None, age=[], dgm=[], hgm=[], ba=[], v=[],
                 n=[], schedule=None):
        self.id = id 
        self.sp = sp
        self.age = age
        self.v = v
        self.n = n
        # NB: these should be maybe parametrized?
        self.dgm = dgm
        self.hgm = hgm
        self.ba = ba
        self.schedule = schedule


class HJOpt(Runner):
    
    def __init__(self, cfg):
        super(HJOpt, self).__init__(cfg)
        self._simulator = None
        self._simulation_ids = None
    
    def run(self):
        """
        Run the whole algorithm
        """
        self._logname = 'H&J optimizer'
        
        # open a file for output
        try:
            self._out_fp = open(self.cfg['hooke_and_jeeves.output_file'], 'w')
        except Exception, e:
            msg = "could not open file for output: '%s'" % str(e)
            self._error(msg)
            return 
        
        # first import the data into SIMO input database 
        try:
            self.import_data()
        except Exception, e:
            msg = "could not import input data: '%s'" % str(e)
            self._error(msg)
            return
        
        # get the optimization parameter sets
        param_sets = self._get_opt_params()
        
        # get stand ids from the input database
        stand_ids = self._get_stand_ids()
        
        # setup after importing the input data
        if self._setup(simtype='data processing'):
            
            # run updating simulations before the actual optimization for all
            # target stands
            self._update(stand_ids)
            
            # setup after running the updating simulations
            if self._setup(simtype='simulation'):
                
                # start going though stands one set at a time
                for id in stand_ids:
                        
                    # start going through the optimization parameters one
                    # parameter set at a time
                    for params in param_sets:
                        
                        # create a stand object, optimize the harvest schedule
                        # of the stand given the set of optimization parameters
                        # and output the optimization results for the current
                        # stand and parameter set
                        stand = self._get_stand(id)
                        self.optimize(stand, params)
                        self.output(stand, params)
        
        # finally, close the databases
        self.close_databases()


    def optimize(self, stand, params):
        """
        Optimize single stand harvest schedule using Hooke & Jeeves algorithm
        """
        max_iterations = self.cfg.get('hooke_and_jeeves.max_iterations')
        max_no_increase_iterations = 20
        iterations = self.cfg.get('simulation.iterations')
        ir = params[0]
        iteration = 0
        no_increase = 0
        
        # a container for n schedules
        # TODO: this should maybe be parametrized (if needed at all?)
        #n = 100 if iterations < 100 else iterations
        #schedules = []
        
        # get an initial schedule
        s_current = self._get_initial_schedule(params)
        s_global = dc(s_current)
        
        # best objective function (npv) values
        x_current = 0.0
        x_global = 0.0
        
        # optimize until the solution does not improve any more or the maximum
        # number of iterations has been reached
        while True:
            
            # get the simulation parameters for the current iteration
            #schedules += self.simulate(stand, params, best, iterations, ir)
            schedules = self.simulate(stand, params, s_current, iterations, ir)
            
            # sort and prune the schedules list
            #self._sort_schedules(schedules, n)
            schedules.sort(key=lambda x: x.npv, reverse=True)
            
            # choose the best schedule from the currently simulated schedules
            s_current = self._get_best_schedule(schedules)
            
            iteration += iterations
            
            if s_current.npv > x_current:
                x_current = s_current.npv
                if x_current > x_global:
                    x_global = x_current
                    s_global = dc(s_current)
            else:
                no_increase += iterations
            
            #DEBUG
            msg = "stand '%s', iter %d, npv %0.2f, current %0.2f, global %0.2f" % \
                  (stand.id, iteration, s_current.npv, x_current, x_global)
            self._out_fp.write(msg + '\n')
            self._info(msg)
            
            # if the objective has not increased in a while, reset the 
            # current best schedule and continue iterations
            if no_increase > max_no_increase_iterations:
                msg = "solution not improving, resetting the schedule..."
                self._info(msg)
                s_current = self._get_initial_schedule(params)
                x_current = 0.0
                no_increase = 0
            
            if iteration >= max_iterations:
                break
        
        # update stand attribute values
        self._update_stand(stand, s_global)


    def simulate(self, stand, params, best, iterations, ir):
        """
        Simulate n harvest schedules for a single stand 
        """
        # generate n new schedules from the currently best schedule,
        # set the scheduled operations to the forced operations db and
        # run the simulation
        schedules = self._get_schedules(params, best, iterations)
        self._set_operations(stand, schedules)
        self._simulator.simulate(iterations, (stand.id,), None)
        
        # collect simulation results and update the schedules with npvs etc.
        schedules = self._collect_results(stand, schedules, ir)
        
        # empty simulation result databases
        self._empty_dbs()
        
        return schedules


    def output(self, stand, params):
        """
        """
        #ir, n_thin, thinnings, clearcut = params
        #parstr = '\tir = %0.1f\n\tnum. of thinnings = %d' % (ir, nthin)
        #for i, thinning in enumerate(thinnings):
        #    parstr += '\n\t%d. thinning '
        parstr = str(params)
        s = 'stand: %s\nparameters: %s' % (stand.id, parstr)
        s += '\n\tspecies = %d' % stand.sp
        s += '\n\tage = %d' % stand.age[0]
        s += '\n\tv = %d' % stand.v[0]
        s += '\n\tn = %d' % stand.n[0]
        s += '\n\tdgm = %d' % stand.dgm[0]
        s += '\n\thgm = %d' % stand.hgm[0]
        s += '\n\tba = %d' % stand.ba[0]
        s += '\n\tNPV = %0.2f' % stand.schedule.npv
        for i, thinning in enumerate(stand.schedule.thinnings):
            s += '\n\t%d. thinning at age %d' % (i+1, thinning.age)
        s += '\n\tclearcut at age %d' % (stand.schedule.clearcut.age,)
        s += '\n\n'
        self._out_fp.write(s)


    def _collect_results(self, stand, schedules, ir):
        """
        Collect simulated NPVs for a number of schedules of a single stand
        """
        cols = [("data_link", "iteration"),
                ("op_link", "op_date"),
                ("op_link", "op_name"),
                ("op_res", "cash_flow"),
                (STAND, STAND_AGE),
                (STAND, STAND_AREA)]
        cs = ', '.join(['.'.join(c) for c in cols])
        sql = "SELECT %s FROM %s LEFT JOIN data_link "\
              "ON %s.data_id=data_link.data_id LEFT JOIN op_link "\
              "ON (data_link.iteration=op_link.iteration "\
              "AND data_link.data_date=op_link.op_date) LEFT JOIN op_res "\
              "ON op_res.op_id=op_link.op_id "\
              "WHERE op_res.cash_flow!=0 "\
              "ORDER BY op_link.iteration, op_link.op_date" % \
              (cs, STAND, STAND)
              
        # get the results and update schedule npv
        results = self.result_db.db.get(sql)
        for iteration, opdate, opname, cashflow, age, area in results:
            npv = self._npv(opdate, cashflow, ir) / area
            schedules[iteration].npv += npv
        
        return schedules


    def _empty_dbs(self):
        """
        Empty the simulation result databases
        """
        self.result_db.clear_data()
        self.fops_db.clear_data()
        self._simulator.any_terminated = False
        self._simulator.release_all()


    def _error(self, msg):
        self.logger.log_message(self._logname, 'error', msg)


    def _info(self, msg):
        self.logger.log_message(self._logname, 'info', msg)


    def _get_best_schedule(self, schedules):
        """
        Get the best schedule
        """
        return schedules[0]


    def _get_initial_schedule(self, params):
        """
        Generate initial (random?) schedule for a single stand
        
        TODO: implementation not ready!!!
        """
        ir, n_thin, thinnings, clearcut = params
        
        # create a new schedule
        schedule = Schedule()
        schedule.ir = ir
        schedule.npv = 0.0
        
        # set thinning timings
        schedule.thinnings = []
        if n_thin > 0:
            type, age, prop = thinnings[0]
            thinning_0 = Harvest(type=LOWER_THINNING,
                                 age=age,
                                 proportion=prop)
            schedule.thinnings.append(thinning_0)
            for i in range(1, n_thin):
                type, yrs, prop = thinnings[i]
                age = schedule.thinnings[i-1].age + yrs
                thinning_i = Harvest(type=LOWER_THINNING,
                                     age=age,
                                     proportion=prop)
            schedule.thinnings.append(thinning_i)
        
        # set clearcut timing
        if n_thin > 0:
            age = schedule.thinnings[-1].age + clearcut[0]
        else:
            age = clearcut[0]
        schedule.clearcut = Harvest(type=CLEARCUT,
                                    age=age)
        
        return schedule


    def _get_opt_params(self):
        """
        Load the H&J optimization parameter sets from a file
        
        TODO: implementation not ready!!!
        """
        # optimization parameters consist of:
        #   interest rate (in pct.)
        #   number of thinnings (max 5)
        #   [thinning params: (type, init_timing, init_proportion),
        #     thinning params: (type, yrs_from_2nd_thinning, init_proportion),
        #     thinning params: (type, yrs_from_3rd_thinning, init_proportion),
        #     thinning params: (type, yrs_from_4th_thinning, init_proportion),
        #     thinning params: (type, yrs_from_5th_thinning, init_proportion),]
        #   clearcut params: (yrs_from_last_thinning, ...)
        # TODO: these should be parsed from a file
        params = [(3., 2, ((LOWER_THINNING, 30, 0.5), (LOWER_THINNING, 20, 0.5),), (20,)),
                  #(3., 2, ((LOWER_THINNING, 30, 0.5), (LOWER_THINNING, 50, 0.5),), (65,)),
                  #(4., 2, ((LOWER_THINNING, 30, 0.5), (LOWER_THINNING, 50, 0.5),), (70,)),
                  #(4., 2, ((LOWER_THINNING, 30, 0.5), (LOWER_THINNING, 50, 0.5),), (65,)),
                  ]
        return params


    def _get_schedules(self, params, schedule, iterations):
        """
        Get a number of alternative harvest schedules
        
        TODO: implementation not ready!!!
        """
        schedules = []
        n_thin = len(schedule.thinnings)
        for i in range(iterations):
            # TODO: here should be the logic for changing the optimization
            # target parameters, such as harvest timings and parameters
            new = dc(schedule)
            new.npv = 0.0
            
            # draw which parameter to change
            n = n_thin + 1
            r = numpy.random.randint(n)
            
            # TODO: at the moment this will only change harvest timings
            if r == (n - 1):
                op = new.clearcut
                if n > 1:
                    neighbours = (new.thinnings[-1], None)
            else:
                op = new.thinnings[r]
                if r > 0:
                    before = new.thinnings[r-1]
                else:
                    before = None
                if r < (n_thin - 1):
                    after = new.thinnings[r+1]
                else:
                    after = None
                neighbours = (before, after)
                
            if op.incr_age is None:
                a = numpy.random.randint(2)
                if a == 0:
                    op.incr_age = -1
                else:
                    op.incr_age = 1
            
            distance = MIN_YEARS_BTW_HARVESTS + 1
            before, after = neighbours
            if before and op.incr_age < 0:
                distance = op.age - before.age
            elif after:
                distance = after.age - op.age
            
            if distance >= MIN_YEARS_BTW_HARVESTS:
                op.age += op.incr_age
                
            schedules.append(new)
            
        return schedules


    def _get_stand(self, id):
        """
        Get a new stand object and update the stand's initial attribute
        values from the input database
        """
        try:
            results = self.input_db.db.get(STAND_ATTR_SQL, (id,))[0]
        except Exception, e:
            msg = "could not get initial stand attributes from input db: "\
                  "'%s'" % str(e)
            self._error(msg)
            return None
        
        sp, age, v, n, dgm, hgm, ba = results
        
        stand = Stand(id=id, sp=sp, age=[age], dgm=[dgm], hgm=[hgm], ba=[ba],
                      v=[v], n=[n])
        
        return stand


    def _get_stand_ids(self):
        """
        Load stand ids from input database and divide the stands into 
        optimization sets, i.e. how many stands will be optimized
        simultaneusly?
        
        NB: at the moment, only one stand is optimized simultaneously
        """
        self._simulation_ids = None
        if 'simulation.id_list' in self.cfg:
            self._simulation_ids = self.cfg['simulation.id_list']
            
        sql = "SELECT DISTINCT id FROM %s" % (STAND,)
        try:
            res = self.input_db.db.get(sql)
        except Exception, e:
            errormsg = "Error getting stand ids from input databas: '%s'" % e
            self.logger.log_message('H&J optimizer', 'error', errormsg)
        if self._simulation_ids is not None:
            stands = [r[0] for r in res if r[0] in self._simulation_ids]
        else:
            stands = [r[0] for r in res]            
        #n = self.cfg['simulation.max_unit_count']
        #i = 0
        #sets = [[]]
        #for s in stands:
        #    sets[i].append(s)
        #    if len(sets[i]) >= n:
        #        sets.append([])
        #        i += 1
        #return sets
        return stands


    def _npv(self, date, value, ir):
        """
        Calculate discount net present value with given interest rate
        """
        return value / (1 + ir / 100.) ** (date.year - BASE_YEAR)


    def _set_operations(self, stand, schedules):
        """
        Set scheduled operations into forced operations db before simulation
        
        TODO: implementation missing
        """
        db = self.fops_db
        dt = datetime.date
        
        branch = 0
        timingtype = 'date'
        timestep = 0
        stepunit = 'year'
        age = stand.age[0]
        ext_col_vals = None  # TODO: this is needed for thinning parameters!!!
        
        for iteration, schedule in enumerate(schedules):
            
            # set thinnings
            for i, thinning in enumerate(schedule.thinnings):
                if age <= thinning.age:
                    date = dt(int(thinning.age - age) + BASE_YEAR, 1, 1)
                    chains = self._op2mc.operation_mapping[THINNING]
                    op_id = "%s-%d-%s-%d" % (stand.id, iteration, THINNING, i)
                    db.add_operation(stand.id, STAND, iteration, branch,
                                     THINNING, timingtype, timestep, date,
                                     stepunit, chains, op_id,
                                     ext_col_vals=ext_col_vals)
            
            # set clearcut
            clearcut = schedule.clearcut
            if age <= clearcut.age:
                date = dt(int(clearcut.age - age) + BASE_YEAR, 1, 1)
                chains = self._op2mc.operation_mapping[CLEARCUT]
                op_id = "%s-%d-%s-%d" % (stand.id, iteration, CLEARCUT, i)
                db.add_operation(stand.id, STAND, iteration, branch,
                                 CLEARCUT, timingtype, timestep, date,
                                 stepunit, chains, op_id,
                                 ext_col_vals=ext_col_vals)


    def _setup(self, simtype='simulation'):
        """
        Setup the optimizer: create a simulator, get required stuff from SIMO
        databases, ...
        """
        # SIMULATOR SETUP
        logname = 'simulation'
        
        msg = 'Initializing simulator...'
        self._info(msg)
        
        add = self.cfg['additional_options.controls']
        if simtype == 'data processing':
            self.ctrl = self.ctrls[add[0]]
        elif simtype == 'simulation':
            self.ctrl = self.ctrls[add[1]]
        
        # get the error handling settings
        max_errors = self.cfg['error_handling.max_error_count_total']
        max_simunit_errors = \
                self.cfg['error_handling.max_error_count_per_unit']
        self.ctrl.use_data_date = False
        self.ctrl.use_today = False
        self.ctrl.initial_date = datetime.date(BASE_YEAR, 1, 1)
        warnings = self.cfg['warning_handling.output_warnings']
        objcount_multiplier = self.cfg['simulation.object_count_multiplier']
        branchcount = 1
        max_unit_count = self.cfg['simulation.max_unit_count']
        noopbranch = False
        branchlimit = 1
        max_res_objs = self.cfg['simulation.max_result_unit_count']
        deterministic = self.cfg['simulation.deterministic']
        track_prices = False
        do_forced_ops = True
        copy2iters = False
        create_branch_desc = False
        cache_dir = self.cfg['simulation.cache_dir']
        
        args = (self.logger, logname, simtype, self.lx, self.ctrl,
                self.input_db, self.result_db, self.fops_db,
                warnings, max_unit_count, objcount_multiplier,
                branchcount, noopbranch, branchlimit, max_res_objs,
                max_errors, max_simunit_errors, deterministic,
                track_prices, do_forced_ops, copy2iters,
                create_branch_desc, cache_dir)
        
        if not self._simulator:
            try:
                self._simulator = Simulator(*args)
            except Exception, e:
                msg = "could not initialize simulator: '%s'" % str(e)
                self._error(msg)
        else:
            try:
                self._simulator.re_init(*args)
            except Exception, e:
                msg = "could not re-initialize simulator: '%s'" % str(e)
                self._error(msg)
        
        
        # OPERATION IMPORT SETUP
        omap = self.cfg['hooke_and_jeeves.operation_modelchains']
        self._op2mc = self.struct_db.get_obj(sc.OPERATION2MODELCHAINS_DEF,
                                             omap)
        
        return True


    def _sort_schedules(self, schedules, n):
        """
        Sort and prune the schedules
        """
        schedules.sort(key=lambda x: x.npv, reverse=True)
        if len(schedules) > n:
            schedules = schedules[:n]


    def _update(self, stand_ids):
        """
        Run the updating simulations ('data processing': update input db
        values) for a set of stands
        """
        self._simulator.simulate(1, stand_ids, None)


    def _update_stand(self, stand, schedule):
        """
        Update stand's attributes using the given schedule
        """
        # set the scheduled operations to the forced operations db and
        # run the simulation
        self._set_operations(stand, (schedule,))
        self._simulator.simulate(1, (stand.id,), None)
        
        for row in self.result_db.db.get(STAND_ATTR_SQL, (stand.id,)):
            sp, age, v, n, dgm, hgm, ba = row
            stand.age.append(age)
            stand.v.append(v)
            stand.n.append(n)
            # NB: these should be maybe parametrized?
            stand.dgm.append(dgm)
            stand.hgm.append(hgm)
            stand.ba.append(ba)
        
        # store the schedule for the stand
        stand.schedule = schedule
        
        # empty simulation
        self._empty_dbs()


class HJConfig(Config):
    
    def _validate_config_content(self):
        """
        Validate H&J optimizer configuration file contents
        """
        self.ini_err=False
        
        
        # program_execution section
        self._check_dir('program_execution.base_folder', setbase=True)
        self._option_exists('program_execution.run_id')
        self._check_boolean('program_execution.no_wipe_import')
        self._check_boolean('program_execution.no_wipe_simulate')
        self._check_boolean('program_execution.no_exec_import')
        self._check_boolean('program_execution.no_exec_simulate')
        self._check_boolean('program_execution.no_exec_logger')
        
        # logging section
        self._check_logging('logging')
        
        
        # translation section
        self._option_exists('translation.message_translation')
        self._option_exists('translation.lexicon_translation')
        self._option_exists('translation.lang')
        
        
        # error_handling section
        self._check_integer('error_handling.max_error_count_total')
        self._check_integer('error_handling.max_error_count_per_unit')
        self._check_boolean('error_handling.suppress_location')
        
        
        # warning_handling section
        self._check_boolean('warning_handling.output_warnings')
        self._check_boolean('warning_handling.suppress_location')
        
        
        # simo_db section
        self._check_file('simo_db.db_path')
        
        
        # data_db section
        self._option_exists('data_db.simulation_level')
        self._check_value('data_db.db_type', ('sqlite', 'postgresql',
                                              'oracle'), True)
        chk = self._option_exists('data_db.threaded',
                                  allowedempty=True)
        if not chk:
            self.cfg['data_db.threaded'] = False
        else:
            self._check_boolean('data_db.threaded')
        dbtype = self.cfg.get('data_db.db_type')
        if dbtype is None:
            return
        self.cfg['data_db.db_type'] = dbtype.upper()
        self._check_boolean('data_db.spatial')
        if self.cfg['data_db.db_type'] == 'SQLITE':
            try:
                from pysqlite2 import dbapi2 as sqlite3
            except:
                self._set_error('SQLITE3 (pysqlite2) not present '\
                                'or found on system!')
            self._check_boolean('data_db.sqlite.in_memory_db')
            if not self.cfg['data_db.sqlite.in_memory_db']:
                self._check_dir('data_db.sqlite.db_dir', create=True)
            else:
                self.cfg['data_db.sqlite.db_dir'] = ''
            self.cfg['data_db.postgresql.db_host'] = ''
            self.cfg['data_db.postgresql.db_port'] = 0
            self.cfg['data_db.postgresql.db_user'] = ''
            self.cfg['data_db.postgresql.db_pw'] = ''

            self.cfg['data_db.postgresql.db_name'] = ''
            self.cfg['data_db.postgresql.db_schema'] = ''

            self.cfg['data_db.oracle.db_host'] = ''
            self.cfg['data_db.oracle.db_user'] = ''
            self.cfg['data_db.oracle.db_pw'] = ''

            self._option_exists('data_db.sqlite.log_db')
            self._option_exists('data_db.sqlite.input_db')
            self._option_exists('data_db.sqlite.result_db')
            self._option_exists('data_db.sqlite.optimized_db')
        elif self.cfg['data_db.db_type'] == 'POSTGRESQL':
            try:
                import psycopg2
            except:
                self._set_error('POSTGRESQL (psycopg2) not present '\
                                'or found on system!')
            self.cfg['data_db.sqlite.db_dir'] = ''
            self.cfg['data_db.sqlite.in_memory_db'] = False
            self.cfg['data_db.oracle.db_host'] = ''
            self.cfg['data_db.oracle.db_user'] = ''
            self.cfg['data_db.oracle.db_pw'] = ''
            self._option_exists('data_db.postgresql.db_host')
            self._check_number('data_db.postgresql.db_port', True, True, True)
            self._option_exists('data_db.postgresql.db_user')
            self._option_exists('data_db.postgresql.db_pw')

            self._check_letters_only('data_db.postgresql.db_name')
            self._check_letters_only('data_db.postgresql.db_schema')

            self._check_letters_only('data_db.postgresql.log_prefix')
            self._check_letters_only('data_db.postgresql.input_prefix')
            self._check_letters_only('data_db.postgresql.result_prefix')
            self._check_letters_only('data_db.postgresql.optimized_prefix')
        elif self.cfg['data_db.db_type'] == 'ORACLE':
            try:
                import cx_Oracle
            except:
                self._set_error('importing cx_Oracle failed! '\
                                'Oracle db_type can not be used')
            self.cfg['data_db.postgresql.db_host'] = ''
            self.cfg['data_db.postgresql.db_port'] = 0
            self.cfg['data_db.postgresql.db_user'] = ''
            self.cfg['data_db.postgresql.db_pw'] = ''
            self.cfg['data_db.postgresql.db_name'] = ''
            self.cfg['data_db.postgresql.db_schema'] = ''
            self.cfg['data_db.sqlite.db_dir'] = ''
            self.cfg['data_db.sqlite.in_memory_db'] = False
            self._option_exists('data_db.oracle.db_host')
            self._option_exists('data_db.oracle.db_user')
            self._option_exists('data_db.oracle.db_pw')
            self._check_letters_only('data_db.oracle.log_prefix')
            self._check_letters_only('data_db.oracle.input_prefix')
            self._check_letters_only('data_db.oracle.result_prefix')
            self._check_letters_only('data_db.oracle.optimized_prefix')
        
        
        # import section
        self._check_value('import.format', self.import_formats)
        self._option_exists('import.mapping_definition')
        self._check_boolean('import.import_date')
        self._check_date('import.data_date')
        if 'import.month_first' in self.cfg.keys():
            self._check_boolean('import.month_first')
        self._check_delimiter('import.data_delimiter')
        self._check_boolean('import.skip_first_line')
        self._check_dir('import.data_dir')
        datadir = self.cfg.get('import.data_dir')
        self._check_multi_file('import.data_file', True, None, datadir)
        
        
        # simulation section
        usedata = True
        usetoday =  False
        t_type = 'simulation'
        self._check_boolean('simulation.deterministic')
        #self._check_boolean('simulation.track_prices')
        #self._check_boolean('simulation.do_forced_ops')
        self._check_dir('simulation.cache_dir', create=True)
        self._check_number('simulation.object_count_multiplier',
                           positive=True, non_zero=True)
        self._check_number('simulation.iterations',
                           integer=True, positive=True, non_zero=True)
        if 'simulation.id_list' in self.cfg.keys():
            self._check_list('simulation.id_list')
        self._check_number('simulation.max_result_unit_count',
                           integer=True, positive=True, non_zero=True)
        
        
        # Hooke & Jeeves options
        self._check_integer('hooke_and_jeeves.max_iterations')
        self._check_dir('hooke_and_jeeves.output_file', create=True, file=True)
        
        
        # ensure additional options are in the resulting cfg as empty lists,
        # even if they do not exist. Strip any '' values.
        self._check_list('additional_options.input_db', True, True)
        self._check_list('additional_options.result_db', True, True)
        self._check_list('additional_options.operation_db', True, True)
        self._check_list('additional_options.optimized_db', True, True)
        self._check_list('additional_options.controls', True, True)
        self._check_list('additional_options.conversions', True, True)
        
        
        # more fixed options
        self.cfg['program_execution.import_data'] = True
        self.cfg['program_execution.simulate'] = True
        self.cfg['program_execution.optimize'] = False
        self.cfg['program_execution.output_data'] = False
        self.cfg['simulation.do_forced_ops'] = False
        self.cfg['simulation.create_branch_desc'] = False
        self.cfg['import.operations.operation_import'] = False
        self.cfg['optimization.track_solution'] = False
        self.cfg['optimization.create_opt_db'] = False


def _parse_ini_file(inilogger, inifile):
    # Parse initialisation file
    cfg = HJConfig(inilogger)
    cfg.load_config(inifile)
    if cfg.ini_err:
        inilogger.error('Terminating execution due to incorrect ' \
                        'initalization parameters')
        sys.exit(0)
    return cfg


def main():
    """
    Runner for HJ optimizer.
    """
    #NB: set Numpy to ignore warnings
    try:
        import numpy
        numpy.seterr(all='ignore')
    except:
        pass

    usage = "usage: %prog [options] simo_ini_file\n"\
            "    ini_file -- text file containing the settings for the "\
            "program run\n"
    parser = OptionParser(usage)
    parser.add_option('-d', '--debug', dest='debug', action='store_true',
                      help='run post mortem debug mode enabled')
    options, args = parser.parse_args()
    if len(args) == 0:
        parser.error('ini file(s) not given')
    inifile = args[0]
    if not os.path.isfile(inifile):
        parser.error('Configuration file %s is not a valid file' % inifile)
    
    inilogger = _get_ini_logger()
    cfg = _parse_ini_file(inilogger, inifile)
    
    # create a H&J Optimizer
    hjopt = HJOpt(cfg.cfg)
    
    # open databases
    hjopt.open_simodb()
    msgttable, lexttable, lang = _get_translators(inilogger, hjopt)
    hjopt.open_logger(msgttable, lexttable, lang)
    try:
        hjopt.open_databases(inilogger)
    except Exception, e:
        type, value, tb = sys.exc_info()
        traceback.print_exc()
        if options.debug:
            pdb.post_mortem(tb)
        hjopt.terminate('Unable to open databases (%s)' % str(e))
    
    # run the H&J optimizer
    hjopt.run()
    

if __name__ == '__main__':
    main()

