#!/usr/bin/env python
# encoding: utf-8
"""
*logger.py*

Usage: 
logger.py [OPTIONS] empty logdb
Clears the logtable
OPTIONS:
-f --from : timestamp as d.m.yyyy or d.m.yyyy:hh:mm
-t --to : timestamp as d.m.yyyy or d.m.yyyy:hh:mm
-i --id : comma delimited list of run ids
Options will limit which values are wiped

logger.py [OPTIONS] print logdb
Prints the logtable's contents to stdout
OPTIONS:
-f --from : timestamp as d.m.yyyy or d.m.yyyy:hh:mm
-t --to : timestamp as d.m.yyyy or d.m.yyyy:hh:mm
-i --id : comma delimited list of run ids
-l --levels : comma delimited list of log levels
-s --source : comma delimited list of modules

If there are no options, the action is done to the entire database
If -f but not -t, from the given timestamp to the last timestamp available
If -t but not -f, from the first timestamp available to the given timestamp
-i, -l and -s; where-conditions for run_id, level- and module-columns

Copyright (c) 2009 Simosol Oy.
    Distributed under the GNU General Public License 2.0.
"""

import optparse
import time
import datetime
import os
from simo.db.datadb.db import LoggerDB

def _parse_timestamp(timestr):
    if timestr is None or timestr == '':
        return None
    try:
        timestamp = datetime.datetime.strptime(timestr, '%d.%m.%Y:%H:%M')
    except ValueError:
        timestamp = datetime.datetime.strptime(timestr, '%d.%m.%Y')
    return timestamp

def _parse_commas(commastr):
    if commastr is None:
        return None
    return commastr.split(',')

def _get_data(options, logdb):
    logdb = LoggerDB(logdb, wipe=False)
    data = logdb.get_log_messages(run_id=options.ids,
                                  level=options.levels, 
                                  module=options.sources,
                                  start_time=options.frm, 
                                  end_time=options.to)
    logdb.close()
    return data
    
def _format_data(data):
    strstruct = '%s - %s, %s(%s, sim_id: %s): %s'
    for i in range(len(data)):
        data[i] = strstruct % (data[i][0].strftime('%Y-%m-%d %H:%M'), 
                               data[i][2], data[i][3], 
                               data[i][1], data[i][4], 
                               data[i][5])
    return data

def _empty_log(options, logdb):
    logdb = LoggerDB(logdb, wipe=False)
    logdb.get_log_messages(run_id=options.ids,
                           start_time=options.frm, 
                           end_time=options.to,
                           empty_log=True)
    logdb.close()
    
def _write_data(data, file):
    if not file:
        for line in data:
            print line
    else:
        try:
            try:
                os.makedirs(os.path.dirname(file))
            except:
                pass
            out = open(file, 'w')
            out.writelines([''.join((line, '\n')) for line in data])
        finally:
            out.close()

def main():
    usage =  'Usage: \n%prog [Options] print <logdb_file>\n'
    usage += 'to print log contents\n'
    usage += '-or-\n'
    usage += '%prog [Options] empty <logdb_file>\n'
    usage += 'to clear out the log database\n'
    usage += 'options will limit what is printed or cleared'
    parser = optparse.OptionParser(usage=usage)
    parser.add_option('-f', '--from', dest='frm', 
                      help='timestamp as d.m.yyyy or d.m.yyyy:hh:mm')
    parser.add_option('-t', '--to', dest='to', 
                      help='timestamp as d.m.yyyy or d.m.yyyy:hh:mm')
    parser.add_option('-i', '--id', dest='ids', 
                      help='comma delimited list of run ids')
    parser.add_option('-l', '--levels', dest='levels', 
                      help='comma delimited list of log levels (printing only,\n'+\
                      'available levels: debug, info, warning, error)')
    parser.add_option('-s', '--source', dest='sources', 
                      help='comma delimited list of source modules (printing only)')
    parser.add_option('-o', '--output', dest='file',
                      help='a file into which the log is written, '+\
                           'instead of printing it out (printing only)')
    (options, args) = parser.parse_args()
    if len(args) != 2 or args[0] not in ('empty', 'print') or not os.path.exists(args[1]):
        if len(args) == 2:
            if not os.path.exists(args[1]):
                print 'File', args[1], 'does not exists!'
        parser.print_help()
        return
    options.frm = _parse_timestamp(options.frm)
    options.to = _parse_timestamp(options.to)
    options.ids = _parse_commas(options.ids)
    options.levels = _parse_commas(options.levels)
    options.sources = _parse_commas(options.sources)
    if args[0] == 'empty':
        _empty_log(options, args[1])
        print 'Operation complete'
    else:
        data = _format_data(_get_data(options, args[1]))
        _write_data(data, options.file)
        print 'Operation complete'

if __name__ == '__main__':
#DEBUG
#    import sys
#    sys.argv.append('-h')
#    sys.argv.append('--from=15.7.2000:15:45')
#    sys.argv.append('-t')
#    sys.argv.append(' 1.1.2051')
#    sys.argv.append('--id=demo')
#    sys.argv.append('--levels=warning,error,info')
#    sys.argv.append('--source=simulation,output')
#    sys.argv.append('-o')
#    sys.argv.append('test/test.log')
#    sys.argv.append('print')
#    sys.argv.append('../simulator/db/log.db')
    main()