// Copyright (c) 2007 Simosol Oy. Distributed under the GNU General Public License 2.0.

//////////////////////////////////////////////////////////////////////////////
//
//  DistributionmodelLibrary.cpp
//
//    DistributionModels - Source file
//
//    Here are the conventional calls of the functions

#include <stdio.h>
#include "DistributionmodelLibrary.h"
#include "InitmodelLibrary.h"
#include "SimoUtils.h"
#include <math.h>
#include <stdlib.h>
#include <string.h>


int MAX_ROUNDS = 5;


double normal_01_cdf ( double x )


//  Purpose:
//
//    NORMAL_01_CDF evaluates the Normal 01 CDF.
//
//  Modified:
//
//    10 February 1999
//
//  Author:
//
//    John Burkardt
//
//  Reference:
//
//    A G Adams,
//    Areas Under the Normal Curve,
//    Algorithm 39,
//    Computer j.,
//    Volume 12, pages 197-198, 1969.
//
//  Parameters:
//
//    Input, double X, the argument of the CDF.
//
//    Output, double CDF, the value of the CDF.
//
{
  double a1 = 0.398942280444;
  double a2 = 0.399903438504;
  double a3 = 5.75885480458;
  double a4 = 29.8213557808;
  double a5 = 2.62433121679;
  double a6 = 48.6959930692;
  double a7 = 5.92885724438;
  double b0 = 0.398942280385;
  double b1 = 3.8052E-08;
  double b2 = 1.00000615302;
  double b3 = 3.98064794E-04;
  double b4 = 1.98615381364;
  double b5 = 0.151679116635;
  double b6 = 5.29330324926;
  double b7 = 4.8385912808;
  double b8 = 15.1508972451;
  double b9 = 0.742380924027;
  double b10 = 30.789933034;
  double b11 = 3.99019417011;
  double cdf;
  double q;
  double y;
//
//  |X| <= 1.28.
//
  if ( fabs ( x ) <= 1.28 )
  {
    y = 0.5 * x * x;

    q = 0.5 - fabs ( x ) * ( a1 - a2 * y / ( y + a3 - a4 / ( y + a5
      + a6 / ( y + a7 ) ) ) );
//
//  1.28 < |X| <= 12.7
//
  }
  else if ( fabs ( x ) <= 12.7 )
  {
    y = 0.5 * x * x;

    q = exp ( - y ) * b0 / ( fabs ( x ) - b1
      + b2 / ( fabs ( x ) + b3
      + b4 / ( fabs ( x ) - b5
      + b6 / ( fabs ( x ) + b7
      - b8 / ( fabs ( x ) + b9
      + b10 / ( fabs ( x ) + b11 ) ) ) ) ) );
//
//  12.7 < |X|
//
  }
  else
  {
    q = 0.0;
  }
//
//  Take account of negative X.
//
  if ( x < 0.0 )
  {
    cdf = q;
  }
  else
  {
    cdf = 1.0 - q;
  }

  return cdf;
}


double normal_cdf ( double x, double a, double b )

//  Purpose:
//
//    NORMAL_CDF evaluates the Normal CDF.
//
//  Modified:
//
//    19 September 2004
//
//  Author:
//
//    John Burkardt
//
//  Parameters:
//
//    Input, double X, the argument of the CDF.
//
//    Input, double A, B, the parameters of the PDF.
//    0.0 < B.
//
//    Output, double CDF, the value of the CDF.
//
{
  double cdf;
  double y;

  y = ( x - a ) / b;

  cdf = normal_01_cdf ( y );

  return cdf;
}



double d_huge ( void )

//  Purpose:
//
//    D_HUGE returns a "huge" real value, usually the largest legal real.
//
//  Discussion:
//
//    HUGE_VAL is the largest representable legal real number, and is usually
//    defined in math.h, or sometimes in stdlib.h.
//
//  Modified:
//
//    08 May 2003
//
//  Author:
//
//    John Burkardt
//
//  Parameters:
//
//    Output, double D_HUGE, a "huge" real value.
//
{
  return HUGE_VAL;
}


double dpoly_value ( int n, double a[], double x )

//  Purpose:
//
//    DPOLY_VALUE evaluates a double precision polynomial.
//
//  Discussion:
//
//    For sanity's sake, the value of N indicates the NUMBER of
//    coefficients, or more precisely, the ORDER of the polynomial,
//    rather than the DEGREE of the polynomial.  The two quantities
//    differ by 1, but cause a great deal of confusion.
//
//    Given N and A, the form of the polynomial is:
//
//      p(x) = a[0] + a[1] * x + ... + a[n-2] * x^(n-2) + a[n-1] * x^(n-1)
//
//  Modified:
//
//    13 August 2004
//
//  Author:
//
//    John Burkardt
//
//  Parameters:
//
//    Input, int N, the order of the polynomial.
//
//    Input, double A[N], the coefficients of the polynomial.
//    A[0] is the constant term.
//
//    Input, double X, the point at which the polynomial is to be evaluated.
//
//    Output, double DPOLY_VALUE, the value of the polynomial at X.
//
{
  int i;
  double value;

  value = 0.0;

  for ( i = n-1; 0 <= i; i-- )
  {
    value = value * x + a[i];
  }

  return value;
}


double normal_01_cdf_inv ( double p )

//
//  Purpose:
//
//    NORMAL_01_CDF_INV inverts the standard normal CDF.
//
//  Discussion:
//
//    The result is accurate to about 1 part in 10*16.
//
//  Modified:
//
//    27 December 2004
//
//  Author:
//
//    Michael Wichura.
//
//    C++ version by John Burkardt.
//
//  Reference:
//
//    Michael Wichura,
//    The Percentage Points of the Normal Distribution,
//    Algorithm AS 241,
//    Applied Statistics,
//    Volume 37, Number 3, pages 477-484, 1988.
//
//  Parameters:
//
//    Input, real ( kind = 8 ) P, the value of the cumulative probability
//    densitity function.  0 < P < 1.  If P is outside this range, an
//    "infinite" value is returned.
//
//    Output, real ( kind = 8 ) NORMAL_01_CDF_INVERSE, the normal deviate value
//    with the property that the probability of a standard normal deviate being
//    less than or equal to this value is P.
//
{
  double a[8] = {
    3.3871328727963666080,     1.3314166789178437745e+2,
    1.9715909503065514427e+3,  1.3731693765509461125e+4,
    4.5921953931549871457e+4,  6.7265770927008700853e+4,
    3.3430575583588128105e+4,  2.5090809287301226727e+3 };
  double b[8] = {
    1.0,                       4.2313330701600911252e+1,
    6.8718700749205790830e+2,  5.3941960214247511077e+3,
    2.1213794301586595867e+4,  3.9307895800092710610e+4,
    2.8729085735721942674e+4,  5.2264952788528545610e+3 };
  double c[8] = {
    1.42343711074968357734,     4.63033784615654529590,
    5.76949722146069140550,     3.64784832476320460504,
    1.27045825245236838258,     2.41780725177450611770e-1,
    2.27238449892691845833e-2,  7.74545014278341407640e-4 };
  double const1 = 0.180625;
  double const2 = 1.6;
  double d[8] = {
    1.0,                        2.05319162663775882187,
    1.67638483018380384940,     6.89767334985100004550e-1,
    1.48103976427480074590e-1,  1.51986665636164571966e-2,
    5.47593808499534494600e-4,  1.05075007164441684324e-9 };
  double e[8] = {
    6.65790464350110377720,     5.46378491116411436990,
    1.78482653991729133580,     2.96560571828504891230e-1,
    2.65321895265761230930e-2,  1.24266094738807843860e-3,
    2.71155556874348757815e-5,  2.01033439929228813265e-7 };
  double f[8] = {
    1.0,                        5.99832206555887937690e-1,
    1.36929880922735805310e-1,  1.48753612908506148525e-2,
    7.86869131145613259100e-4,  1.84631831751005468180e-5,
    1.42151175831644588870e-7,  2.04426310338993978564e-15 };
  double q;
  double r;
  double split1 = 0.425;
  double split2 = 5.0;
  double value;

  if ( p <= 0.0 )
  {
    value = -d_huge ( );
    return value;
  }

  if ( 1.0 <= p )
  {
    value = d_huge ( );
    return value;
  }

  q = p - 0.5;

  if ( fabs ( q ) <= split1 )
  {
    r = const1 - q * q;
    value = q * dpoly_value ( 8, a, r ) / dpoly_value ( 8, b, r );
  }
  else
  {
    if ( q < 0.0 )
    {
      r = p;
    }
    else
    {
      r = 1.0 - p;
    }

    if ( r <= 0.0 )
    {
      value = -1.0;
      exit ( 1 );
    }

    r = sqrt ( -log ( r ) );

    if ( r <= split2 )
    {
      r = r - const2;
      value = dpoly_value ( 8, c, r ) / dpoly_value ( 8, d, r );
     }
     else
     {
       r = r - split2;
       value = dpoly_value ( 8, e, r ) / dpoly_value ( 8, f, r );
    }

    if ( q < 0.0 )
    {
      value = -value;
    }

  }

  return value;
}


double normal_cdf_inv ( double cdf, double a, double b )

//
//  Purpose:
//
//    NORMAL_CDF_INV inverts the Normal CDF.
//
//  Modified:
//
//    19 September 2004
//
//  Author:
//
//    John Burkardt
//
//  Reference:
//
//    Algorithm AS 111,
//    Applied Statistics,
//    Volume 26, pages 118-121, 1977.
//
//  Parameters:
//
//    Input, double CDF, the value of the CDF.
//    0.0 <= CDF <= 1.0.
//
//    Input, double A, B, the parameters of the PDF.
//    0.0 < B.
//
//    Output, double X, the corresponding argument.
//
{
  double x;
  double x2;

  if ( cdf < 0.0 || 1.0 < cdf )
  {
    /*
    cout << "\n";
    cout << "NORMAL_CDF_INV - Fatal error!\n";
    cout << "  CDF < 0 or 1 < CDF.\n";*/
    exit ( 1 );
  }
  x2 = normal_01_cdf_inv ( cdf );
  x = a + b * x2;
  return x;
}


static double get_percent_point_cdf_value ( double class_center, double (*pp)[3][12] )
{
    int ret = 1;

    int i;
    double cum_point, full_interval, point_interval;
    for ( i = 1; i < 12; i++)
    {
        if (  class_center <= (*pp)[0][i] )  // is the class center smaller than the diameter in the percent point cdf
        {
            full_interval = (*pp)[0][i] - (*pp)[0][i-1];    // difference between two successive diameters in the cdf
            point_interval = class_center - (*pp)[0][i-1];    // difference between the smaller cdf diameter and class center

            //Check for fatal errors
            if (full_interval<=0)
            {
                ret = 0;
            }
            if (ret == 0) {
                return 0;
            }

            cum_point = (*pp)[1][i-1] + ( point_interval / full_interval ) * (*pp)[2][i];    // cumulated prob. mass at the class center
            break;
        }
    }
    return cum_point;
}


static int construct_diamdist_using_percent_point(double DIAM_CLASS_WIDTH, double (*pp)[3][12],
                          double N, int class_count, double class_n_min, struct diamclass *dist)
{

    int i;
    double low_boundary, high_boundary, low_mass, high_mass, stem_number, left_out, lower_bound;
#ifdef __cplusplus
    double** dist_2 = NULL;
    dist_2 = new double*[class_count];
    for (int i=0; i < class_count; i++) {
        dist_2[i] = new double[2];
        for (int j=0; j<2; j++) {
            dist_2[i][j] = 0;
        }
    }
#else
    double dist_2[class_count][2];
#endif

    left_out = 0;
    //diameters & stem numbers
    for ( i = 0; i < class_count; i++)  {

        if ( i != class_count - 1)
        {
            // all but the last class that has different class length
            if ( i == 0)
            {
                // first class
                dist_2[i][0] = (*pp)[0][0] + DIAM_CLASS_WIDTH / 2;  // d0 + DIAM_CLASS_WIDTH/2
                lower_bound = dist_2[i][0];
            }
            else
            {
                // intermediate classes
                dist_2[i][0] = dist_2[i-1][0] + DIAM_CLASS_WIDTH;
            }
            // class boundaries for estimation of stem number
            low_boundary = dist_2[i][0] - DIAM_CLASS_WIDTH / 2;
            high_boundary = dist_2[i][0] + DIAM_CLASS_WIDTH / 2;

        }
        else
        {
            // last class
            dist_2[i][0] = dist_2[i-1][0] + DIAM_CLASS_WIDTH / 2 + ((*pp)[0][11] - (dist_2[i-1][0] + DIAM_CLASS_WIDTH / 2)) / 2;
            low_boundary = dist_2[i-1][0] + DIAM_CLASS_WIDTH / 2;
            high_boundary = (*pp)[0][11];  // d100
        }
        // cumulated probabilities at boundaries
        low_mass = get_percent_point_cdf_value(low_boundary, pp);
        high_mass = get_percent_point_cdf_value(high_boundary, pp);
            //stem number
            stem_number = N * (high_mass - low_mass);
        //Calculate stem number of rejected classes
        if (stem_number < class_n_min)
             left_out = left_out + stem_number;

        dist_2[i][1] = stem_number;
    }
    //Calculate new stem numbers to the classes
    int class_count_new;
    double mass_scaler;
    int k = 0;
    //Assign the first class to biggest
    int biggest_class = 0;
    class_count_new = class_count;
    mass_scaler = N / (N - left_out);
    for ( i = 0; i < class_count; i++)
    {
        if (dist_2[i][1] > dist_2[biggest_class][1])
                 biggest_class = i;
        if (dist_2[i][1] < class_n_min)
             class_count_new = class_count_new - 1;
            else
            {
             dist[k].dbh = dist_2[i][0];
             dist[k].n = mass_scaler * dist_2[i][1];
             k++;
        }
    }
    //Simple distribution for the stratum with only few or none classes left
    if (class_count_new < 3)
    {
        //Force the distribution centre lower if too close to the high bounder of distribution
        while (high_boundary < DIAM_CLASS_WIDTH / 2 + dist_2[biggest_class][0] + DIAM_CLASS_WIDTH)
        {
             biggest_class = biggest_class-1;
         }
         while (lower_bound > dist_2[biggest_class][0] - DIAM_CLASS_WIDTH / 2 - DIAM_CLASS_WIDTH)
         {
             biggest_class = biggest_class+1;
         }
        dist[0].dbh = dist_2[biggest_class-1][0];
        dist[0].n = N/3.;
        dist[1].dbh = dist_2[biggest_class][0];
        dist[1].n = N/3.;
        dist[2].dbh = dist_2[biggest_class+1][0];
        dist[2].n = N/3.;
        class_count_new = 3;
    }
#ifdef __cplusplus
    for (int i=0; i < class_count; i++) {
        delete[] dist_2[i];
        dist_2[i] = NULL;
    }
    delete[] dist_2;
    dist_2 = NULL;
#endif
    return class_count_new;
}


static int construct_diamdist_using_Weibull(double a, double b, double c,
                                             double DIAM_CLASS_WIDTH,
                                             double BA, double d0,
                                             double d100, int class_count,
                                             double class_n_min,
                                             struct diamclass *dist)
{
    char errorStr[200] = "";
    int ret = 1;
    int i;
    double low_boundary, high_boundary, low_mass, high_mass, stem_number, left_out, lower_bound;
#ifdef __cplusplus
    double** dist_2 = NULL;
    dist_2 = new double*[class_count];
    for (int i=0; i < class_count; i++) {
        dist_2[i] = new double[2];
        for (int j=0; j<2; j++) {
            dist_2[i][j] = 0;
        }
    }
#else
    double dist_2[class_count][2];
#endif

    left_out = 0;
    //diameters & stem numbers
    for ( i = 0; i < class_count; i++)
    {
        if ( i != class_count - 1)
        {
            // all but the last class that has different class length
            if ( i == 0)
            {
                // first class
                dist_2[i][0] = d0 + DIAM_CLASS_WIDTH/2;
                lower_bound = dist_2[i][0];
            }
            else
            {
                // intermediate classes
                dist_2[i][0] = dist_2[i-1][0] + DIAM_CLASS_WIDTH;
            }
            // class boundaries for estimation of stem number
            low_boundary = dist_2[i][0] - DIAM_CLASS_WIDTH / 2;
            high_boundary = dist_2[i][0] + DIAM_CLASS_WIDTH / 2;

        }
        else
        {
            // last class
            dist_2[i][0] = dist_2[i-1][0] + DIAM_CLASS_WIDTH / 2 + (d100 - (dist_2[i-1][0] + DIAM_CLASS_WIDTH/2)) / 2;
            low_boundary = dist_2[i-1][0] + DIAM_CLASS_WIDTH / 2;
            high_boundary = d100;
        }

        // cumulated probabilities at boundaries
        if ( i == 0)    // first class
            low_mass = 0;
        else
            low_mass = 1 - exp(-pow((( low_boundary - a ) / b ), c ));

        high_mass = 1 - exp(-pow((( high_boundary - a ) / b ), c ));
            //stem number
        stem_number = ( BA * ( high_mass - low_mass )) / ( PI * pow(( dist_2[i][0] / 200 ), 2));
        //Calculate basal area mass of rejected classes
        if (stem_number < class_n_min)
             left_out = left_out + stem_number * PI * pow(( dist_2[i][0] / 200 ), 2);

        dist_2[i][1] = stem_number;
    }
    //Calculate new stem numbers to the classes
    int class_count_new;
    double mass_scaler, ppa_mass, ppa_mass_scaled;
    int j = 0;
    //Assign the first class to biggest
    int biggest_class = 0;
    class_count_new = class_count;
    mass_scaler = BA / (BA - left_out);
    for ( i = 0; i < class_count; i++)
    {
        if (dist_2[i][1] > dist_2[biggest_class][1])
                 biggest_class = i;
        if (dist_2[i][1] < class_n_min)
             class_count_new = class_count_new - 1;
            else
            {
                 dist[j].dbh = dist_2[i][0];
             ppa_mass = PI * pow(( dist_2[i][0] / 2 ), 2) * dist_2[i][1];
             ppa_mass_scaled = ppa_mass * mass_scaler;
             dist[j].n = ppa_mass_scaled/(PI * pow(( dist_2[i][0] / 2 ), 2));
             j++;
        }
    }
    if (class_count_new < 3)
    {
        //Force the distribution centre lower if too close to the high bounder of distribution
        while (high_boundary < DIAM_CLASS_WIDTH / 2 + dist_2[biggest_class][0] + DIAM_CLASS_WIDTH)
        {
             biggest_class = biggest_class-1;
         }
         while (lower_bound > dist_2[biggest_class][0] - DIAM_CLASS_WIDTH / 2 - DIAM_CLASS_WIDTH)
         {
             biggest_class = biggest_class+1;
         }
        dist[0].dbh = dist_2[biggest_class-1][0];
        dist[0].n = (BA/3.)/(PI * pow(( dist_2[biggest_class-1][0] / 200 ), 2));
        dist[1].dbh = dist_2[biggest_class][0];
        dist[1].n = (BA/3.)/(PI * pow(( dist_2[biggest_class][0] / 200 ), 2));
        dist[2].dbh = dist_2[biggest_class+1][0];
        dist[2].n = (BA/3.)/(PI * pow(( dist_2[biggest_class+1][0] / 200 ), 2));
        class_count_new = 3;
    }
#ifdef __cplusplus
    for (int i=0; i < class_count; i++) {
        delete[] dist_2[i];
        dist_2[i] = NULL;
    }
    delete[] dist_2;
    dist_2 = NULL;
#endif
    return class_count_new;
}

static int construct_diamdist_using_Weibull_only_N_measured(double a, double b, double c,
                                             double DIAM_CLASS_WIDTH,
                                             double BA, double N, double d0,
                                             double d100, int class_count,
                                             double class_n_min,
                                             struct diamclass *dist)
{
    char errorStr[200] = "";
    int ret = 1;
    int i;
    double low_boundary, high_boundary, low_mass, high_mass, stem_number, N_calc, left_out, lower_bound;

#ifdef __cplusplus
    double** dist_2 = NULL;
    dist_2 = new double*[class_count];
    for (int i=0; i < class_count; i++) {
        dist_2[i] = new double[2];
        for (int j=0; j<2; j++) {
            dist_2[i][j] = 0;
        }
    }
#else
    double dist_2[class_count][2];
#endif

    N_calc = 0;
    left_out = 0;
    //diameters & stem numbers
    for ( i = 0; i < class_count; i++)
    {
        if ( i != class_count - 1)
        {
            // all but the last class that has different class length
            if ( i == 0)
            {
                // first class
                dist_2[i][0] = d0 + DIAM_CLASS_WIDTH/2;
                lower_bound = dist_2[i][0];
            }
            else
            {
                // intermediate classes
                dist_2[i][0] = dist_2[i-1][0] + DIAM_CLASS_WIDTH;
            }
            // class boundaries for estimation of stem number
            low_boundary = dist_2[i][0] - DIAM_CLASS_WIDTH / 2;
            high_boundary = dist_2[i][0] + DIAM_CLASS_WIDTH / 2;

        }
        else
        {
            // last class
            dist_2[i][0] = dist_2[i-1][0] + DIAM_CLASS_WIDTH / 2 + (d100 - (dist_2[i-1][0] + DIAM_CLASS_WIDTH/2)) / 2;
            low_boundary = dist_2[i-1][0] + DIAM_CLASS_WIDTH / 2;
            high_boundary = d100;
        }

        // cumulated probabilities at boundaries
        if ( i == 0)    // first class
            low_mass = 0;
        else
        low_mass = 1 - exp(-pow((( low_boundary - a ) / b ), c ));

        high_mass = 1 - exp(-pow((( high_boundary - a ) / b ), c ));
            //stem number
        stem_number = ( BA * ( high_mass - low_mass )) / ( PI * pow(( dist_2[i][0] / 200 ), 2));
        dist_2[i][1] = stem_number;

        N_calc = N_calc+stem_number;
    }
    //Simple scaling for stem numbers
    double init_n;
    for ( i = 0; i < class_count; i++)
    {
            init_n = dist_2[i][1];
        dist_2[i][1] = (N/N_calc)*init_n;
        //Calculate stem number of rejected classes
        if (dist_2[i][1] < class_n_min)
             left_out = left_out + stem_number;
        }
        //Calculate new stem numbers to the classes
        int class_count_new;
    double mass_scaler;
    int j = 0;
    //Assign the first class to biggest
    int biggest_class = 0;
    class_count_new = class_count;
    mass_scaler = N_calc / (N_calc - left_out);
    for ( i = 0; i < class_count; i++)
    {
        if (dist_2[i][1] > dist_2[biggest_class][1])
                 biggest_class = i;
        if (dist_2[i][1] < class_n_min)
             class_count_new = class_count_new - 1;
            else
            {
             dist[j].dbh = dist_2[i][0];
             dist[j].n = mass_scaler * dist_2[i][1];
             j++;
        }
    }
    //Simple distribution for the stratum with only few or none classes left
    if (class_count_new < 3)
    {
        //Force the distribution centre lower if too close to the high bounder of distribution
        while (high_boundary < DIAM_CLASS_WIDTH / 2 + dist_2[biggest_class][0] + DIAM_CLASS_WIDTH)
        {
             biggest_class = biggest_class-1;
         }
         while (lower_bound > dist_2[biggest_class][0] - DIAM_CLASS_WIDTH / 2 - DIAM_CLASS_WIDTH)
         {
             biggest_class = biggest_class+1;
         }
        dist[0].dbh = dist_2[biggest_class-1][0];
        dist[0].n = N_calc/3;
        dist[1].dbh = dist_2[biggest_class][0];
        dist[1].n = N_calc/3;
        dist[2].dbh = dist_2[biggest_class+1][0];
        dist[2].n = N_calc/3;
        class_count_new = 3;
    }
#ifdef __cplusplus
    for (int i=0; i < class_count; i++) {
        delete[] dist_2[i];
        dist_2[i] = NULL;
    }
    delete[] dist_2;
    dist_2 = NULL;
#endif
    return class_count_new;

}

double normal_cdf_sb ( double x, double c, double d, double upper_bound, double lower_bound )

//    Itse tehty muunnos yll‰ olevasta functiosta normal_cdf (...)!
//  Purpose:
//
//    NORMAL_CDF evaluates the Normal CDF.
//
//  Modified:
//
//    19 September 2004
//
//  Author:
//
//    John Burkardt
//
//  Parameters:
//
//    Input, double X, the argument of the CDF.
//
//    Input, double A, B, the parameters of the PDF.
//    0.0 < B.
//
//    Output, double CDF, the value of the CDF.
//
{
  double cdf;
  double y;

  y = c + d *log((x-lower_bound)/(upper_bound+lower_bound-x));
  cdf = normal_01_cdf ( y );

  return cdf;
}


static int construct_diamdist_using_SB (double upper_bound, double g, double y, double lower_bound, double DIAM_CLASS_WIDTH, double N, double BA,
                             double d0, double d100, int class_count, double class_n_min, struct diamclass *dist, int *nres) {
    int i;
    double low_boundary, high_boundary, low_mass, high_mass, stem_number, left_out;

#ifdef __cplusplus
    double** dist_2 = NULL;
    dist_2 = new double*[class_count];
    for (int i=0; i < class_count; i++) {
        dist_2[i] = new double[2];
        for (int j=0; j<2; j++) {
            dist_2[i][j] = 0;
        }
    }
#else
    double dist_2[class_count][2];
#endif

    left_out = 0;
    //diameters & stem numbers
    for ( i = 0; i < class_count; i++)
    {
        if ( i != class_count - 1)
        {    // all but the last class that has different class length
            if ( i == 0)
            {    // first class
                dist_2[i][0] = d0 + DIAM_CLASS_WIDTH/2;
            }
            else
            {    // intermediate classes
                dist_2[i][0] = dist_2[i-1][0] + DIAM_CLASS_WIDTH;
            }
            // class boundaries for estimation of stem number
            low_boundary = dist_2[i][0] - DIAM_CLASS_WIDTH / 2;
            high_boundary = dist_2[i][0] + DIAM_CLASS_WIDTH / 2;

        }
        else
        {    // last class
            dist_2[i][0] = dist_2[i-1][0] + DIAM_CLASS_WIDTH/2 + (d100- (dist_2[i-1][0] + DIAM_CLASS_WIDTH/2)) / 2;
            low_boundary = dist_2[i-1][0] + DIAM_CLASS_WIDTH / 2;
            high_boundary = d100;
        }
        // cumulated probabilities at boundaries
        low_mass = normal_cdf_sb(low_boundary, y, g, upper_bound, lower_bound);
        high_mass = normal_cdf_sb(high_boundary, y, g, upper_bound, lower_bound);

            //stem number
        if ( N != 0 )
        {
            stem_number = N * ( high_mass - low_mass );
        }
        else
        {
            stem_number = ( BA * ( high_mass - low_mass )) / ( PI * pow(( dist_2[i][0] / 200 ), 2));
        }
        //Calculate basal area mass of rejected classes
        if (stem_number < class_n_min)
             left_out = left_out + stem_number * PI * pow(( dist_2[i][0] / 200 ), 2);

        dist_2[i][1] = stem_number;

    }
    //Calculate new stem numbers to the classes
    int class_count_new;
    double mass_scaler, ppa_mass, ppa_mass_scaled;
    int j = 0;
    //Assign the first class to biggest
    int biggest_class = 0;
    class_count_new = class_count;
    mass_scaler = BA / (BA - left_out);
    for ( i = 0; i < class_count; i++)
    {
        if (dist_2[i][1] > dist_2[biggest_class][1])
                 biggest_class = i;
        if (dist_2[i][1] < class_n_min)
             class_count_new = class_count_new - 1;
            else
            {
                 dist[j].dbh = dist_2[i][0];
             ppa_mass = PI * pow(( dist_2[i][0] / 2 ), 2) * dist_2[i][1];
             ppa_mass_scaled = ppa_mass * mass_scaler;
             dist[j].n = ppa_mass_scaled/(PI * pow(( dist_2[i][0] / 2 ), 2));
             j++;
        }
    }
    if (class_count_new < 3)
    {
        //Force the distribution centre lower if too close to the high bounder of distribution
        while (high_boundary < DIAM_CLASS_WIDTH / 2 + dist_2[biggest_class][0] + DIAM_CLASS_WIDTH)
        {
             biggest_class = biggest_class-1;
         }
         while (lower_bound > dist_2[biggest_class][0] - DIAM_CLASS_WIDTH / 2 - DIAM_CLASS_WIDTH)
         {
             biggest_class = biggest_class+1;
         }
        dist[0].dbh = dist_2[biggest_class-1][0];
        dist[0].n = (BA/3.)/(PI * pow(( dist_2[biggest_class-1][0] / 200 ), 2));
        dist[1].dbh = dist_2[biggest_class][0];
        dist[1].n = (BA/3.)/(PI * pow(( dist_2[biggest_class][0] / 200 ), 2));
        dist[2].dbh = dist_2[biggest_class+1][0];
        dist[2].n = (BA/3.)/(PI * pow(( dist_2[biggest_class+1][0] / 200 ), 2));
        class_count_new = 3;
    }
#ifdef __cplusplus
    for (int i=0; i < class_count; i++) {
        delete[] dist_2[i];
        dist_2[i] = NULL;
    }
    delete[] dist_2;
    dist_2 = NULL;
#endif
    return class_count_new;
}



int Diameter_distribution_pine_percent_point_KangasMaltamo(double BA, double D_gM, double Age, double N,
                             double DIAM_CLASS_WIDTH, double MAX_CLASS_COUNT, double CLASS_N_MIN,
                            int *nres, struct diamclass *dist, char *errors,
                            int errorCheckMode, double allowedRiskLevel, double rectFactor)
{

    int ret, class_count, rounds;
    double d0, d10, d20, d30, d40, d60, d70, d80, d90, d95, d100;
    double pp[3][12] =    {    { 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0},  // {diameter at percent point}
                            {  0, .1, .2, .3, .4, .5, .6, .7, .8, .9, .95, 1},    // {cum. prob. mass at the percent point}
                            {  0, .1, .1, .1, .1, .1, .1, .1, .1, .1, .05, .05} };    // {prob. mass that the class cumulates}
    char errorStr[200] = "";

    ret = 1;
    //Check for fatal errors
    if (D_gM<=0)
    {
        ret = 0;
        constructErrorMessage (errorStr, "D_gM<=0(D_gM=",D_gM);
    }
    if (BA<=0)
    {
        ret = 0;
        constructErrorMessage (errorStr, "BA<=0(BA=",BA);
    }
    if (N/BA<0)
    {
        ret = 0;
        constructErrorMessage2val (errorStr, "N/BA<0(N=",N, ",BA=",BA);
    }
    if (Age<0)
    {
        ret = 0;
        constructErrorMessage (errorStr, "Age<0(Age=",Age);
    }
    if (ret == 0) {
        strcat(errors, errorStr);
        return 0;
    }

    d0 =  exp(8.606-0.8606*log(D_gM)-0.9777*log(N/BA)-0.1158*log(BA));
    d10 = exp(6.134-0.2992*log(D_gM)-0.7091*log(N/BA));
    d20 = exp(3.262+0.3124*log(D_gM)-0.3852*log(N/BA));
    d30 = exp(1.6192+0.6628*log(D_gM)-0.2012*log(N/BA));
    d40 = exp(1.0015+0.7891*log(D_gM)-0.1167*log(N/BA));
    d60 = exp(0.1612+0.9298*log(D_gM)+0.0265*log(Age));
    d70 = exp(0.3378+0.8542*log(D_gM)+0.0535*log(Age));
    d80 = exp(0.4671+0.8037*log(D_gM)+0.0747*log(Age));
    d90 = exp(0.7097+0.7411*log(D_gM)+0.0837*log(Age));
    d95 = exp(0.8148+0.7246*log(D_gM)+0.0842*log(Age));
    d100 = exp(1.0392+0.6288*log(D_gM)+0.1139*log(Age)+0.0249*log(BA));
    //double pp[3][12] =    {    { d0, d10, d20, d30, d40, D_gM, d60, d70, d80, d90, d95, d100 },  // {diameter at percent point}
    //                        {  0, .1, .2, .3, .4, .5, .6, .7, .8, .9, .95, 1},                // {cum. prob. mass at the percent point}
    //                        {  0, .1, .1, .1, .1, .1, .1, .1, .1, .1, .05, .05} };                // {prob. mass that the class cumulates}
    pp[0][0] = d0;
    pp[0][1] = d10;
    pp[0][2] = d20;
    pp[0][3] = d30;
    pp[0][4] = d40;
    pp[0][5] = D_gM;
    pp[0][6] = d60;
    pp[0][7] = d70;
    pp[0][8] = d80;
    pp[0][9] = d90;
    pp[0][10] = d95;
    pp[0][11] = d100;

    // how many diameter classes
    class_count = (int) floor((d100 - d0) / DIAM_CLASS_WIDTH);

    if (class_count<=0)
    {
        ret = 0;
        constructErrorMessage (errorStr, "class_count<0(class_count=",class_count);
    }
    if (ret == 0) {
        strcat(errors, errorStr);
        return 0;
    }

    rounds = 0;
    while (class_count < 15 && rounds <= MAX_ROUNDS)
    {  //If too few classes
        DIAM_CLASS_WIDTH = DIAM_CLASS_WIDTH / 2;
        class_count = (int) floor((d100 - d0) / DIAM_CLASS_WIDTH);
        rounds = rounds + 1;
    }
    if (class_count < 15)
    {
        ret = 0;
        constructErrorMessage (errorStr, "Too few generated diameter classes! Check D_gM and BA!", 15);
    }
    if (ret == 0) {
        strcat(errors, errorStr);
        return 0;
    }

    rounds = 0;
    while (class_count > 50 && rounds <= MAX_ROUNDS)
    {  //If too many classes
        DIAM_CLASS_WIDTH = DIAM_CLASS_WIDTH * 2;
        class_count = (int) floor((d100 - d0) / DIAM_CLASS_WIDTH);
        rounds = rounds + 1;
    }
    if (class_count > 50)
    {
        ret = 0;
        constructErrorMessage (errorStr, "Too many generated diameter classes! Check D_gM and BA!", 50);
    }
    if (class_count > MAX_CLASS_COUNT)
    {
        ret = 0;
        constructErrorMessage (errorStr, "Too many generated diameter classes! Check D_gM and BA! (MAX_CLASS_COUNT=", MAX_CLASS_COUNT);
    }
        if (ret == 0) {
        strcat(errors, errorStr);
        return 0;
    }

    // diameter distribution vector, elements contain diameter at breast height and number of stems / ha in the class
    class_count = construct_diamdist_using_percent_point( DIAM_CLASS_WIDTH, &pp, N, class_count, CLASS_N_MIN, dist );

    *nres = class_count;

    return ret;
}




int Diameter_distribution_spruce_percent_point_KangasMaltamo ( double BA, double BA_total,
     double D_gM, double Age, double N, double SC, double DIAM_CLASS_WIDTH, double MAX_CLASS_COUNT, double CLASS_N_MIN,
    int *nres, struct diamclass *dist, char *errors,
    int errorCheckMode, double allowedRiskLevel, double rectFactor)
{

    int SC4_6, ret, class_count, rounds;
    double d0, d10, d20, d30, d40, d60, d70, d80, d90, d95, d100;
    double pp[3][12] =    {    { 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0},  // {diameter at percent point}
                            {  0, .1, .2, .3, .4, .5, .6, .7, .8, .9, .95, 1},    // {cum. prob. mass at the percent point}
                            {  0, .1, .1, .1, .1, .1, .1, .1, .1, .1, .05, .05} };    // {prob. mass that the class cumulates}
    char errorStr[200] = "";

    int iSC = (int) SC;
    ret = 1;
    switch ( iSC )
     {
            case 1:
            SC4_6 = 0;
            break;
        case 2:
            SC4_6 = 0;
            break;
        case 3:
            SC4_6 = 0;
            break;
        case 4:
            SC4_6 = 1;
            break;
        case 5:
            SC4_6 = 1;
            break;
        case 6:
            SC4_6 = 1;
            break;
        // SC 7 and 8 not present in original publication, treated here as SC 4/6
        case 7:
            SC4_6 = 1;
            break;
            case 8:
            SC4_6 = 1;
            break;
        // defaults to SC==1 or 2 or 3, sensible?
        default:
            SC4_6 = 0;
            break;
    }

    //Check for fatal errors
    if (D_gM<=0)
    {
        ret = 0;
        constructErrorMessage (errorStr, "D_gM<=0(D_gM=",D_gM);
    }
    if (BA<=0)
    {
        ret = 0;
        constructErrorMessage (errorStr, "BA<=0(BA=",BA);
    }
    if (N/BA<0)
    {
        ret = 0;
        constructErrorMessage2val (errorStr, "N/BA<0(N=",N, ",BA=",BA);
    }
    if (N<0)
    {
        ret = 0;
        constructErrorMessage (errorStr, "N<0(N=",N);
    }
    if (N/BA_total<0)
    {
        ret = 0;
        constructErrorMessage2val (errorStr, "N/BA_total<0(N=",N, ",BA_total=",BA_total);
    }
    if (Age<0)
    {
        ret = 0;
        constructErrorMessage (errorStr, "Age<0(Age=",Age);
    }
    if (ret == 0) {
        strcat(errors, errorStr);
        return 0;
    }

    d0 =  exp(6.7355-0.6355*log(D_gM)-0.6862*log(N/BA)-0.1075*log(N/BA_total));
    d10 = exp(5.9686-0.3039*log(D_gM)-0.6656*log(N/BA));
    d20 = exp(2.9714+0.3504*log(D_gM)-0.3454*log(N/BA));
    d30 = exp(1.2553+0.7431*log(D_gM)-0.1728*log(N/BA));
    d40 = exp(0.8795+0.8251*log(D_gM)-0.112*log(N/BA));
    d60 = exp(0.1766+0.9688*log(D_gM));
    d70 = exp(0.3432+0.9013*log(D_gM)+0.0269*log(Age));
    d80 = exp(0.5222+0.8496*log(D_gM)+0.0421*log(Age));
    d90 = exp(0.8332+0.7694*log(D_gM)+0.0529*log(Age)-0.0343*(SC4_6));
    d95 = exp(0.9762+0.728*log(D_gM)+0.0656*log(Age)-0.0337*(SC4_6));
    d100 = exp(1.0844+0.746*log(D_gM)+0.0433*log(N)-0.045*(SC4_6));

    //double pp[3][12] =    {    { d0, d10, d20, d30, d40, D_gM, d60, d70, d80, d90, d95, d100 },  // {diameter at percent point}
    //                        {  0, .1, .2, .3, .4, .5, .6, .7, .8, .9, .95, 1},                // {cum percent point}
    //                        {  0, .1, .1, .1, .1, .1, .1, .1, .1, .1, .05, .05} };                // {class length}

    pp[0][0] = d0;
    pp[0][1] = d10;
    pp[0][2] = d20;
    pp[0][3] = d30;
    pp[0][4] = d40;
    pp[0][5] = D_gM;
    pp[0][6] = d60;
    pp[0][7] = d70;
    pp[0][8] = d80;
    pp[0][9] = d90;
    pp[0][10] = d95;
    pp[0][11] = d100;

    // how many diameter classes (??????)
    class_count = (int) floor((d100 - d0) / DIAM_CLASS_WIDTH);

    if (class_count<=0)
    {
        ret = 0;
        constructErrorMessage (errorStr, "class_count<0(class_count=",class_count);
    }
    if (ret == 0) {
        strcat(errors, errorStr);
        return 0;
    }

    rounds = 0;
    while (class_count < 15 && rounds <= MAX_ROUNDS)
    {  //If too few classes
        DIAM_CLASS_WIDTH = DIAM_CLASS_WIDTH / 2;
        class_count = (int) floor((d100 - d0) / DIAM_CLASS_WIDTH);
        rounds = rounds + 1;
    }
    if (class_count < 15)
    {
        ret = 0;
        constructErrorMessage (errorStr, "Too few generated diameter classes! Check D_gM and BA!", 15);
    }
    if (ret == 0) {
        strcat(errors, errorStr);
        return 0;
    }

        rounds = 0;
    while (class_count > 50 && rounds <= MAX_ROUNDS)
    {  //If too many classes
        DIAM_CLASS_WIDTH = DIAM_CLASS_WIDTH * 2;
        class_count = (int) floor((d100 - d0) / DIAM_CLASS_WIDTH);
        rounds = rounds + 1;
    }
    if (class_count > 50)
    {
        ret = 0;
        constructErrorMessage (errorStr, "Too many generated diameter classes! Check D_gM and BA!", 50);
    }
    if (class_count > MAX_CLASS_COUNT)
    {
        ret = 0;
        constructErrorMessage (errorStr, "Too many generated diameter classes! Check D_gM and BA! (MAX_CLASS_COUNT=", MAX_CLASS_COUNT);
    }
        if (ret == 0) {
        strcat(errors, errorStr);
        return 0;
    }

    // diameter distribution vector, elements contain diameter at breast height and number of stems / ha in the class
    class_count = construct_diamdist_using_percent_point( DIAM_CLASS_WIDTH, &pp, N, class_count, CLASS_N_MIN, dist );

    *nres = class_count;

    return ret;
}



int Diameter_distribution_birch_percent_point_KangasMaltamo ( double BA, double D_gM, double Age, double N,
                                    double SC, double DIAM_CLASS_WIDTH, double MAX_CLASS_COUNT, double CLASS_N_MIN,
                                    int *nres, struct diamclass *dist, char *errors,
                                    int errorCheckMode, double allowedRiskLevel, double rectFactor)
{
    int SC4_6, ret, class_count, rounds;
    double d0, d10, d20, d30, d40, d60, d70, d80, d90, d95, d100;
    double pp[3][12] =    {    { 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0},  // {diameter at percent point}
                            {  0, .1, .2, .3, .4, .5, .6, .7, .8, .9, .95, 1},    // {cum. prob. mass at the percent point}
                            {  0, .1, .1, .1, .1, .1, .1, .1, .1, .1, .05, .05} };    // {prob. mass that the class cumulates}
    char errorStr[200] = "";

    int iSC = (int) SC;
    ret = 1;
    switch ( iSC )
     {
            case 1:
            SC4_6 = 0;
            break;
        case 2:
            SC4_6 = 0;
            break;
        case 3:
            SC4_6 = 0;
            break;
        case 4:
            SC4_6 = 1;
            break;
        case 5:
            SC4_6 = 1;
            break;
        case 6:
            SC4_6 = 1;
            break;
        // SC 7 and 8 not present in original publication, treated here as SC 4/6
        case 7:
            SC4_6 = 1;
            break;
        case 8:
            SC4_6 = 1;
            break;
        // defaults to SC==1 or 2 or 3, sensible?
        default:
            SC4_6 = 0;
            break;
    }

    //Check for fatal errors
    if (D_gM<=0)
    {
        ret = 0;
        constructErrorMessage (errorStr, "D_gM<=0(D_gM=",D_gM);
    }
    if (BA<=0)
    {
        ret = 0;
        constructErrorMessage (errorStr, "BA<=0(BA=",BA);
    }
    if (N/BA<0)
    {
        ret = 0;
        constructErrorMessage2val (errorStr, "N/BA<0(N=",N, ",BA=",BA);
    }
    if (Age<0)
    {
        ret = 0;
        constructErrorMessage (errorStr, "Age<0(Age=",Age);
    }
    if (ret == 0) {
        strcat(errors, errorStr);
        return 0;
    }

    d0 =  exp(7.5465-0.7072*log(D_gM)-0.8683*log(N/BA));
    d10 = exp(5.1096-0.1207*log(D_gM)-0.5997*log(N/BA));
    d20 = exp(1.9001+0.5752*log(D_gM)-0.2585*log(N/BA));
    d30 = exp(0.4795+0.8865*log(D_gM)-0.0946*log(N/BA));
    d40 = exp(-0.2672+1.052*log(D_gM));
    d60 = exp(0.2257+0.95*log(D_gM));
    d70 = exp(0.3717+0.8749*log(D_gM)+0.0397*log(Age));
    d80 = exp(0.5119+0.7485*log(D_gM)+0.1126*log(Age));
    d90 = exp(0.8064+0.6471*log(D_gM)+0.1465*log(Age)-0.0487*(SC4_6));
    d95 = exp(1.0168+0.592*log(D_gM)+0.1541*log(Age)-0.0778*(SC4_6));
    d100 = exp(1.1898+0.5457*log(D_gM)+0.1787*log(Age)-0.0758*(SC4_6));

    //double pp[3][12] =    {    { d0, d10, d20, d30, d40, D_gM, d60, d70, d80, d90, d95, d100 },  // {diameter at percent point}
    //                        {  0, .1, .2, .3, .4, .5, .6, .7, .8, .9, .95, 1},                // {cum percent point}
    //                        {  0, .1, .1, .1, .1, .1, .1, .1, .1, .1, .05, .05} };                // {class length}

    pp[0][0] = d0;
    pp[0][1] = d10;
    pp[0][2] = d20;
    pp[0][3] = d30;
    pp[0][4] = d40;
    pp[0][5] = D_gM;
    pp[0][6] = d60;
    pp[0][7] = d70;
    pp[0][8] = d80;
    pp[0][9] = d90;
    pp[0][10] = d95;
    pp[0][11] = d100;

    // how many diameter classes (??????)
    class_count = (int) floor((d100 - d0) / DIAM_CLASS_WIDTH);

    if (class_count<=0)
    {
        ret = 0;
        constructErrorMessage (errorStr, "class_count<0(class_count=",class_count);
    }
    if (ret == 0) {
        strcat(errors, errorStr);
        return 0;
    }

    rounds = 0;
    while (class_count < 15 && rounds <= MAX_ROUNDS)
    {  //If too few classes
        DIAM_CLASS_WIDTH = DIAM_CLASS_WIDTH / 2;
        class_count = (int) floor((d100 - d0) / DIAM_CLASS_WIDTH);
        rounds = rounds + 1;
    }
    if (class_count < 15)
    {
        ret = 0;
        constructErrorMessage (errorStr, "Too few generated diameter classes! Check D_gM and BA!", 15);
    }
    if (ret == 0) {
        strcat(errors, errorStr);
        return 0;
    }

    rounds = 0;
    while (class_count > 50 && rounds <= MAX_ROUNDS)
    {  //If too many classes
        DIAM_CLASS_WIDTH = DIAM_CLASS_WIDTH * 2;
        class_count = (int) floor((d100 - d0) / DIAM_CLASS_WIDTH);
        rounds = rounds + 1;
    }
    if (class_count > 50)
    {
        ret = 0;
        constructErrorMessage (errorStr, "Too many generated diameter classes! Check D_gM and BA!", 50);
    }
    if (class_count > MAX_CLASS_COUNT)
    {
        ret = 0;
        constructErrorMessage (errorStr, "Too many generated diameter classes! Check D_gM and BA! (MAX_CLASS_COUNT=", MAX_CLASS_COUNT);
    }
        if (ret == 0) {
        strcat(errors, errorStr);
        return 0;
    }

    // diameter distribution vector, elements contain diameter at breast height and number of stems / ha in the class
    class_count = construct_diamdist_using_percent_point( DIAM_CLASS_WIDTH, &pp, N, class_count, CLASS_N_MIN, dist );

    *nres = class_count;

    return ret;
}


int Diameter_distribution_spruce_Weibull_Kilkki ( double BA, double D_gM, double DIAM_CLASS_WIDTH, double MAX_CLASS_COUNT,
                        double CLASS_N_MIN,
                        int *nres, struct diamclass *dist, char *errors,
                        int errorCheckMode, double allowedRiskLevel, double rectFactor)
{
    int ret, class_count, rounds;
    double a, b, c, d0, d100;
    char errorStr[200] = "";

    ret = 1;
    //Check for fatal errors
    if (D_gM<=0)
    {
        ret = 0;
        constructErrorMessage (errorStr, "D_gM<=0(D_gM=",D_gM);
    }
    if (BA<=0)
    {
        ret = 0;
        constructErrorMessage (errorStr, "BA<=0(BA=",BA);
    }
    if (ret == 0) {
        strcat(errors, errorStr);
        return 0;
    }

    // Weibull distribution cumulative distribution function parameters
    a = 0.001389+0.517444*D_gM;
    b = exp(-0.346223+0.934993*log(D_gM)-0.000925*BA);

    //Check for fatal errors
    if (b<=0)
    {
        ret = 0;
        constructErrorMessage (errorStr, "b<=0(b=",b);
    }
    if ((log((D_gM-a)/b))==0)
    {
        ret = 0;
        constructErrorMessage3val (errorStr, "(log((D_gM-a)/b))<=0(D_gM=", D_gM, ",a=",a, ",b=",b);
    }
    if (ret == 0) {
        strcat(errors, errorStr);
        return 0;
    }


    c = (log(-log(0.5)))/(log((D_gM-a)/b));
    d0 = a;   // low boundary
    //The right tail of distribution is very long with almost zero tree number.
    d100 = a + b * pow( -log( 1.0 - 0.999999999999 ), (1 / c)); // HUOM! Weibull-jakauman prosenttipistejakauman yht‰lˆ?

    // how many diameter classes
    class_count = (int) floor((d100 - d0) / DIAM_CLASS_WIDTH);

        rounds = 0;
    while (class_count < 14 && rounds <= MAX_ROUNDS)
    {  //If too few classes
        DIAM_CLASS_WIDTH = DIAM_CLASS_WIDTH / 2;
        class_count = (int) floor((d100 - d0) / DIAM_CLASS_WIDTH);
        rounds = rounds + 1;
    }

    if (class_count < 14)
    {
        ret = 0;
        constructErrorMessage (errorStr, "Too few generated diameter classes! Check D_gM and BA!", 15);
    }
    if (ret == 0) {
        strcat(errors, errorStr);
        return 0;
    }

        rounds = 0;
    while (class_count > 60 && rounds <= MAX_ROUNDS)
    {  //If too many classes
        DIAM_CLASS_WIDTH = DIAM_CLASS_WIDTH * 2;
        class_count = (int) floor((d100 - d0) / DIAM_CLASS_WIDTH);
        rounds = rounds + 1;
    }

    if (class_count > 60)
    {
        ret = 0;
        constructErrorMessage (errorStr, "Too many generated diameter classes! Check D_gM and BA!", 50);
    }
    if (class_count > MAX_CLASS_COUNT)
    {
        ret = 0;
        constructErrorMessage (errorStr, "Too many generated diameter classes! Check D_gM and BA! (MAX_CLASS_COUNT=", MAX_CLASS_COUNT);
    }
        if (ret == 0) {
        strcat(errors, errorStr);
        return 0;
        }

    // diameter distribution vector, elements contain diameter at breast height and number of stems / ha in the class
    class_count = construct_diamdist_using_Weibull( a, b, c, DIAM_CLASS_WIDTH, BA, d0, d100, class_count, CLASS_N_MIN, dist );

    *nres = class_count;

    return ret;
}

int Diameter_distribution_spruce_Weibull_only_N_measured_Kilkki ( double BA, double D_gM, double N, double DIAM_CLASS_WIDTH, double MAX_CLASS_COUNT,
                        double CLASS_N_MIN,
                        int *nres, struct diamclass *dist, char *errors,
                        int errorCheckMode, double allowedRiskLevel, double rectFactor)
{
    int ret, class_count, rounds;
    double a, b, c, d0, d100;
    char errorStr[200] = "";

    ret = 1;
    //Check for fatal errors
    if (D_gM<=0)
    {
        ret = 0;
        constructErrorMessage (errorStr, "D_gM<=0(D_gM=",D_gM);
    }
    if (BA<=0)
    {
        ret = 0;
        constructErrorMessage (errorStr, "BA<=0(BA=",BA);
    }
    if (BA>100)
    {
        ret = 0;
        constructErrorMessage (errorStr, "BA>100(BA=",BA);
    }
    if (ret == 0) {
        strcat(errors, errorStr);
        return 0;
    }

    // Weibull distribution cumulative distribution function parameters
    a = 0.001389+0.517444*D_gM;
    b = exp(-0.346223+0.934993*log(D_gM)-0.000925*BA);

    //Check for fatal errors
    if (b<=0)
    {
        ret = 0;
        constructErrorMessage (errorStr, "b<=0(b=",b);
    }
    if ((log((D_gM-a)/b))==0)
    {
        ret = 0;
        constructErrorMessage3val (errorStr, "(log((D_gM-a)/b))<=0(D_gM=", D_gM, ",a=",a, ",b=",b);
    }
    if (ret == 0) {
        strcat(errors, errorStr);
        return 0;
    }


    c = (log(-log(0.5)))/(log((D_gM-a)/b));
    d0 = a;   // low boundary
    //The right tail of distribution is very long with almost zero tree number.
    d100 = a + b * pow( -log( 1.0 - 0.999999999999 ), (1 / c)); // HUOM! Weibull-jakauman prosenttipistejakauman yht‰lˆ?

    // how many diameter classes
    class_count = (int) floor((d100 - d0) / DIAM_CLASS_WIDTH);
        rounds = 0;
    while (class_count < 14 && rounds <= MAX_ROUNDS)
    {  //If too few classes
        DIAM_CLASS_WIDTH = DIAM_CLASS_WIDTH / 2;
        class_count = (int) floor((d100 - d0) / DIAM_CLASS_WIDTH);
        rounds = rounds + 1;
    }
    if (class_count < 14)
    {
        ret = 0;
        constructErrorMessage (errorStr, "Too few generated diameter classes! Check D_gM and BA!", 15);
    }
    if (ret == 0) {
        strcat(errors, errorStr);
        return 0;
    }

        rounds = 0;
    while (class_count > 60 && rounds <= MAX_ROUNDS)
    {  //If too many classes
        DIAM_CLASS_WIDTH = DIAM_CLASS_WIDTH * 2;
        class_count = (int) floor((d100 - d0) / DIAM_CLASS_WIDTH);
        rounds = rounds + 1;
    }
    if (class_count > 60)
    {
        ret = 0;
        constructErrorMessage (errorStr, "Too many generated diameter classes! Check D_gM and BA!", 60);
    }
    if (class_count > MAX_CLASS_COUNT)
    {
        ret = 0;
        constructErrorMessage (errorStr, "Too many generated diameter classes! Check D_gM and BA! (MAX_CLASS_COUNT=", MAX_CLASS_COUNT);
    }
    if (ret == 0) {
        strcat(errors, errorStr);
        return 0;
    }

    // diameter distribution vector, elements contain diameter at breast height and number of stems / ha in the class
    class_count = construct_diamdist_using_Weibull_only_N_measured( a, b, c, DIAM_CLASS_WIDTH, BA, N, d0, d100, class_count, CLASS_N_MIN, dist );

    *nres = class_count;

    return ret;
}

int Diameter_distribution_pine_Weibull_Mykkanen ( double BA, double D_gM, double DIAM_CLASS_WIDTH, double MAX_CLASS_COUNT,
                            double CLASS_N_MIN,
                            int *nres, struct diamclass *dist, char *errors,
                            int errorCheckMode, double allowedRiskLevel, double rectFactor)
{
    int ret, class_count, rounds;
    double a, b, c, d0, d100;
    char errorStr[200] = "";

    ret = 1;
    //Check for fatal errors
    if (D_gM<=0)
    {
        ret = 0;
        constructErrorMessage (errorStr, "D_gM<=0(D_gM=",D_gM);
    }
    if (BA<=0)
    {
        ret = 0;
        constructErrorMessage (errorStr, "BA<=0(BA=",BA);
    }
    if (ret == 0) {
        strcat(errors, errorStr);
        return 0;
    }

    // Weibull distribution cumulative distribution function parameters
    a = exp(-1.306454+1.154433*log(D_gM));
    c = exp(0.647888+0.025530*D_gM-0.005558*BA);
    b = (D_gM-a)/(pow((-log(0.5)),(1/c)));

    // how many diameter classes (??????)
    // distribution boundaries
    d0 = a;   // low boundary
    d100 = a + b * pow( -log( 1.0 - 0.999999999999999 ), (1 / c)); // HUOM! Weibull-jakauman prosenttipistejakauman yht‰lˆ?

    // how many diameter classes
    class_count = (int) floor((d100 - d0) / DIAM_CLASS_WIDTH);
        rounds = 0;
    while (class_count < 14 && rounds <= MAX_ROUNDS)
    {  //If too few classes
        DIAM_CLASS_WIDTH = DIAM_CLASS_WIDTH / 2;
        class_count = (int) floor((d100 - d0) / DIAM_CLASS_WIDTH);
        rounds = rounds + 1;
    }
    if (class_count < 14)
    {
        ret = 0;
        constructErrorMessage (errorStr, "Too few generated diameter classes! Check D_gM and BA!", 15);
    }
    if (ret == 0) {
        strcat(errors, errorStr);
        return 0;
    }

        rounds = 0;
    while (class_count > 50 && rounds <= MAX_ROUNDS)
    {  //If too many classes
        DIAM_CLASS_WIDTH = DIAM_CLASS_WIDTH * 2;
        class_count = (int) floor((d100 - d0) / DIAM_CLASS_WIDTH);
        rounds = rounds + 1;
    }
    if (class_count > 50)
    {
        ret = 0;
        constructErrorMessage (errorStr, "Too many generated diameter classes! Check D_gM and BA!", 50);
    }
    if (class_count > MAX_CLASS_COUNT)
    {
        ret = 0;
        constructErrorMessage (errorStr, "Too many generated diameter classes! Check D_gM and BA! (MAX_CLASS_COUNT=", MAX_CLASS_COUNT);
    }
    if (b<=0)
    {
        ret = 0;
        constructErrorMessage (errorStr, "b<=0(b=",b);
    }
        if (ret == 0) {
        strcat(errors, errorStr);
        return 0;
    }

    // diameter distribution vector, elements contain diameter at breast height and number of stems / ha in the class
    class_count = construct_diamdist_using_Weibull( a, b, c, DIAM_CLASS_WIDTH, BA, d0, d100, class_count, CLASS_N_MIN, dist );

    *nres = class_count;

    return ret;
}

int Diameter_distribution_pine_Weibull_only_N_measured_Mykkanen ( double BA, double D_gM, double N, double DIAM_CLASS_WIDTH, double MAX_CLASS_COUNT,
                            double CLASS_N_MIN,
                            int *nres, struct diamclass *dist, char *errors,
                            int errorCheckMode, double allowedRiskLevel, double rectFactor)
{
    int ret, class_count, rounds;
    double a, b, c, d0, d100;
    char errorStr[200] = "";

    ret = 1;
    //Check for fatal errors
    if (D_gM<=0)
    {
        ret = 0;
        constructErrorMessage (errorStr, "D_gM<=0(D_gM=",D_gM);
    }
    if (BA<=0)
    {
        ret = 0;
        constructErrorMessage (errorStr, "BA<=0(BA=",BA);
    }
    if (BA>100)
    {
        ret = 0;
        constructErrorMessage (errorStr, "BA>100(BA=",BA);
    }
    if (ret == 0) {
        strcat(errors, errorStr);
        return 0;
    }

    // Weibull distribution cumulative distribution function parameters
    a = exp(-1.306454+1.154433*log(D_gM));
    c = exp(0.647888+0.025530*D_gM-0.005558*BA);
    b = (D_gM-a)/(pow((-log(0.5)),(1/c)));

    // how many diameter classes (??????)
    // distribution boundaries
    d0 = a;   // low boundary
    d100 = a + b * pow( -log( 1.0 - 0.999999999999999 ), (1 / c)); // HUOM! Weibull-jakauman prosenttipistejakauman yht‰lˆ?

    // how many diameter classes
    class_count = (int) floor((d100 - d0) / DIAM_CLASS_WIDTH);
        rounds = 0;
    while (class_count < 14 && rounds <= MAX_ROUNDS)
    {  //If too few classes
        DIAM_CLASS_WIDTH = DIAM_CLASS_WIDTH / 2;
        class_count = (int) floor((d100 - d0) / DIAM_CLASS_WIDTH);
        rounds = rounds + 1;
    }
    if (class_count < 14)
    {
        ret = 0;
        constructErrorMessage (errorStr, "Too few generated diameter classes! Check D_gM and BA!", 15);
    }
    if (ret == 0) {
        strcat(errors, errorStr);
        return 0;
    }

        rounds = 0;
    while (class_count > 50 && rounds <= MAX_ROUNDS)
    {  //If too many classes
        DIAM_CLASS_WIDTH = DIAM_CLASS_WIDTH * 2;
        class_count = (int) floor((d100 - d0) / DIAM_CLASS_WIDTH);
        rounds = rounds + 1;
    }
    if (class_count > 50)
    {
        ret = 0;
        constructErrorMessage (errorStr, "Too many generated diameter classes! Check D_gM and BA!", 50);
    }
    if (class_count > MAX_CLASS_COUNT)
    {
        ret = 0;
        constructErrorMessage (errorStr, "Too many generated diameter classes! Check D_gM and BA! (MAX_CLASS_COUNT=", MAX_CLASS_COUNT);
    }
    if (b<=0)
    {
        ret = 0;
        constructErrorMessage (errorStr, "b<=0(b=",b);
    }
        if (ret == 0) {
        strcat(errors, errorStr);
        return 0;
    }

    // diameter distribution vector, elements contain diameter at breast height and number of stems / ha in the class
    class_count = construct_diamdist_using_Weibull_only_N_measured( a, b, c, DIAM_CLASS_WIDTH, BA, N, d0, d100, class_count, CLASS_N_MIN, dist );

    *nres = class_count;

    return ret;
}


int Diameter_distribution_birch_Weibull_Siipilehto (double BA, double D_gM, double A, double DIAM_CLASS_WIDTH,
                            double MAX_CLASS_COUNT,    double CLASS_N_MIN,
                            int *nres, struct diamclass *dist, char *errors,
                            int errorCheckMode, double allowedRiskLevel, double rectFactor)
{
    int ret, class_count, rounds;
    double a, b, c, d0, d100;
    char errorStr[200] = "";

    ret = 1;

    // Weibull distribution cumulative distribution function parameters
    a = 0;    // lower boundary not estimated, it is planned to be 0
    c = 3.7062+0.08632*D_gM+0.02185*A-0.1126*BA;

    //Check for fatal errors
    if (c<=0)
    {
        ret = 0;
        constructErrorMessage (errorStr, "c<=0(c=",c);
    }
    if (pow((-log(0.5)),(1/c))<=0)
    {
        ret = 0;
        constructErrorMessage (errorStr, "pow((-log(0.5)),(1/c))<=0(c=",c);
    }
    if (BA<=0)
    {
        ret = 0;
        constructErrorMessage (errorStr, "BA<=0(BA=",BA);
    }
    if (ret == 0) {
        strcat(errors, errorStr);
        return 0;
    }


    b = D_gM/pow((-log(0.5)),(1/c));

    // how many diameter classes (??????)
    // distribution boundaries
    d0 = a;   // low boundary
    d100 = a + b * pow( -log( 1.0 - 0.999999999999999 ), (1 / c)); // HUOM! Weibull-jakauman prosenttipistejakauman yht‰lˆ?

    // how many diameter classes
    class_count = (int) floor((d100 - d0) / DIAM_CLASS_WIDTH);
        rounds = 0;
    while (class_count < 14 && rounds <= MAX_ROUNDS)
    {  //If too few classes
        DIAM_CLASS_WIDTH = DIAM_CLASS_WIDTH / 2;
        class_count = (int) floor((d100 - d0) / DIAM_CLASS_WIDTH);
        rounds = rounds + 1;
    }
    if (class_count < 14)
    {
        ret = 0;
        constructErrorMessage (errorStr, "Too few generated diameter classes! Check D_gM and BA!", 15);
    }
    if (ret == 0) {
        strcat(errors, errorStr);
        return 0;
    }

        rounds = 0;
    while (class_count > 50 && rounds <= MAX_ROUNDS)
    {  //If too many classes
        DIAM_CLASS_WIDTH = DIAM_CLASS_WIDTH * 2;
        class_count = (int) floor((d100 - d0) / DIAM_CLASS_WIDTH);
        rounds = rounds + 1;
    }
    if (class_count > 50)
    {
        ret = 0;
        constructErrorMessage (errorStr, "Too many generated diameter classes! Check D_gM and BA!", 50);
    }
    if (class_count > MAX_CLASS_COUNT)
    {
        ret = 0;
        constructErrorMessage (errorStr, "Too many generated diameter classes! Check D_gM and BA! (MAX_CLASS_COUNT=", MAX_CLASS_COUNT);
    }
    if (b<=0)
    {
        ret = 0;
        constructErrorMessage (errorStr, "b<=0(b=",b);
    }
        if (ret == 0) {
        strcat(errors, errorStr);
        return 0;
    }

    // diameter distribution vector, elements contain diameter at breast height and number of stems / ha in the class
    class_count = construct_diamdist_using_Weibull( a, b, c, DIAM_CLASS_WIDTH, BA, d0, d100, class_count, CLASS_N_MIN, dist );

    *nres = class_count;

    return ret;
}

int Diameter_distribution_birch_Weibull_only_N_measured_Siipilehto (double BA, double D_gM, double A, double N, double DIAM_CLASS_WIDTH,
                            double MAX_CLASS_COUNT,    double CLASS_N_MIN,
                            int *nres, struct diamclass *dist, char *errors,
                            int errorCheckMode, double allowedRiskLevel, double rectFactor)
{
    int ret, class_count, rounds;
    double a, b, c, d0, d100;
    char errorStr[200] = "";

    ret = 1;

    // Weibull distribution cumulative distribution function parameters
    a = 0;    // lower boundary not estimated, it is planned to be 0
    c = 3.7062 + 0.08632 * D_gM + 0.02185 * A - 0.1126 * BA;

    //Check for fatal errors
    if (c<=0)
    {
        ret = 0;
        constructErrorMessage (errorStr, "c<=0(c=",c);
    }
    if (pow((-log(0.5)),(1/c))<=0)
    {
        ret = 0;
        constructErrorMessage (errorStr, "pow((-log(0.5)),(1/c))<=0(c=",c);
    }
    if (BA<=0)
    {
        ret = 0;
        constructErrorMessage (errorStr, "BA<=0(BA=",BA);
    }
    if (BA>50)
    {
        ret = 0;
        constructErrorMessage (errorStr, "BA>50(BA=",BA);
    }
    if (ret == 0) {
        strcat(errors, errorStr);
        return 0;
    }

    b = D_gM / pow((-log(0.5)), (1 / c));

    // how many diameter classes (??????)
    // distribution boundaries
    d0 = a;   // low boundary
    d100 = a + b * pow(-log( 1.0 - 0.999999999999999 ), (1 / c)); // HUOM! Weibull-jakauman prosenttipistejakauman yht‰lˆ?

    // how many diameter classes
    class_count = (int) floor((d100 - d0) / DIAM_CLASS_WIDTH);
        rounds = 0;
    while (class_count < 14 && rounds <= MAX_ROUNDS)
    {  //If too few classes
        DIAM_CLASS_WIDTH = DIAM_CLASS_WIDTH / 2;
        class_count = (int) floor((d100 - d0) / DIAM_CLASS_WIDTH);
        rounds = rounds + 1;
    }
    if (class_count < 14)
    {
        ret = 0;
        constructErrorMessage (errorStr, "Too few generated diameter classes! Check D_gM and BA!", 15);
    }
    if (ret == 0) {
        strcat(errors, errorStr);
        return 0;
    }

        rounds = 0;
    while (class_count > 50 && rounds <= MAX_ROUNDS)
    {  //If too many classes
        DIAM_CLASS_WIDTH = DIAM_CLASS_WIDTH * 2;
        class_count = (int) floor((d100 - d0) / DIAM_CLASS_WIDTH);
        rounds = rounds + 1;
    }
    if (class_count > 50)
    {
        ret = 0;
        constructErrorMessage (errorStr, "Too many generated diameter classes! Check D_gM and BA!", 50);
    }
    if (class_count > MAX_CLASS_COUNT)
    {
        ret = 0;
        constructErrorMessage (errorStr, "Too many generated diameter classes! Check D_gM and BA! (MAX_CLASS_COUNT=", MAX_CLASS_COUNT);
    }
    if (b<=0)
    {
        ret = 0;
        constructErrorMessage (errorStr, "b<=0(b=",b);
    }
    if (ret == 0) {
        strcat(errors, errorStr);
        return 0;
    }

    // diameter distribution vector, elements contain diameter at breast height and number of stems / ha in the class
    class_count = construct_diamdist_using_Weibull_only_N_measured( a, b, c, DIAM_CLASS_WIDTH, BA, N, d0, d100, class_count, CLASS_N_MIN, dist );

    *nres = class_count;

    return ret;
}

int Diameter_distribution_grey_alder_Weibull_Karkiym (double BA, double D_gM, double DIAM_CLASS_WIDTH,
                            double MAX_CLASS_COUNT,    double CLASS_N_MIN,
                            int *nres, struct diamclass *dist, char *errors,
                            int errorCheckMode, double allowedRiskLevel, double rectFactor)
{
    int ret, class_count, rounds;
    double a, b, c, d0, d100;
    char errorStr[200] = "";

    ret = 1;

    //Check for fatal errors
    if (BA <= 0)
    {
        ret = 0;
        constructErrorMessage (errorStr, "BA<=0(BA=",BA);
    }
    if (D_gM<=0)
    {
        ret = 0;
        constructErrorMessage (errorStr, "D_gM<=0(D_gM=",D_gM);
    }
    if (ret == 0) {
        strcat(errors, errorStr);
        return 0;
    }

    // Weibull distribution cumulative distribution function parameters
    a = 0;    // lower boundary not estimated, it is planned to be 0
    c = 9.4529 - 1.3841 * log(pow(BA, 2)/D_gM);

    //Check for fatal errors
    if (c<=0)
    {
        ret = 0;
        constructErrorMessage (errorStr, "c<=0(c=",c);
    }
    if (ret == 0) {
        strcat(errors, errorStr);
        return 0;
    }


    b = exp(0.2915+0.9242*log(D_gM));

    // how many diameter classes (??????)
    // distribution boundaries
    d0 = a;   // low boundary
    d100 = a + b * pow( -log( 1.0 - 0.999999999999999 ), (1 / c)); // HUOM! Weibull-jakauman prosenttipistejakauman yht‰lˆ?

    // how many diameter classes
    class_count = (int) floor((d100 - d0) / DIAM_CLASS_WIDTH);
        rounds = 0;
    while (class_count < 14 && rounds <= MAX_ROUNDS)
    {  //If too few classes
        DIAM_CLASS_WIDTH = DIAM_CLASS_WIDTH / 2;
        class_count = (int) floor((d100 - d0) / DIAM_CLASS_WIDTH);
        rounds = rounds + 1;
    }
    if (class_count < 14)
    {
        ret = 0;
        constructErrorMessage (errorStr, "Too few generated diameter classes! Check D_gM and BA!", 15);
    }
    if (ret == 0) {
        strcat(errors, errorStr);
        return 0;
    }

        rounds = 0;
    while (class_count > 50 && rounds <= MAX_ROUNDS)
    {  //If too many classes
        DIAM_CLASS_WIDTH = DIAM_CLASS_WIDTH * 2;
        class_count = (int) floor((d100 - d0) / DIAM_CLASS_WIDTH);
        rounds = rounds + 1;
    }
    if (class_count > 50)
    {
        ret = 0;
        constructErrorMessage (errorStr, "Too many generated diameter classes! Check D_gM and BA!", 50);
    }
    if (class_count > MAX_CLASS_COUNT)
    {
        ret = 0;
        constructErrorMessage (errorStr, "Too many generated diameter classes! Check D_gM and BA! (MAX_CLASS_COUNT=", MAX_CLASS_COUNT);
    }
    if (b<=0)
    {
        ret = 0;
        constructErrorMessage (errorStr, "b<=0(b=",b);
    }
        if (ret == 0) {
        strcat(errors, errorStr);
        return 0;
    }

    // diameter distribution vector, elements contain diameter at breast height and number of stems / ha in the class
    class_count = construct_diamdist_using_Weibull( a, b, c, DIAM_CLASS_WIDTH, BA, d0, d100, class_count, CLASS_N_MIN, dist );

    *nres = class_count;

    return ret;
}

int Diameter_distribution_grey_alder_Weibull_only_N_measured_Karkiym (double BA, double D_gM, double N, double DIAM_CLASS_WIDTH,
                            double MAX_CLASS_COUNT,    double CLASS_N_MIN,
                            int *nres, struct diamclass *dist, char *errors,
                            int errorCheckMode, double allowedRiskLevel, double rectFactor)
{
    int ret, class_count, rounds;
    double a, b, c, d0, d100;
    char errorStr[200] = "";

    ret = 1;

    //Check for fatal errors
    if (BA <= 0)
    {
        ret = 0;
        constructErrorMessage (errorStr, "BA<=0(BA=",BA);
    }
    if (D_gM<=0)
    {
        ret = 0;
        constructErrorMessage (errorStr, "D_gM<=0(D_gM=",D_gM);
    }
    if (ret == 0) {
        strcat(errors, errorStr);
        return 0;
    }

    // Weibull distribution cumulative distribution function parameters
    a = 0;    // lower boundary not estimated, it is planned to be 0
    c = 9.4529 - 1.3841 * log(pow(BA, 2)/D_gM);

    //Check for fatal errors
    if (c<=0)
    {
        ret = 0;
        constructErrorMessage (errorStr, "c<=0(c=",c);
    }
    if (ret == 0) {
        strcat(errors, errorStr);
        return 0;
    }


    b = exp(0.2915+0.9242*log(D_gM));

    // how many diameter classes (??????)
    // distribution boundaries
    d0 = a;   // low boundary
    d100 = a + b * pow( -log( 1.0 - 0.999999999999999 ), (1 / c)); // HUOM! Weibull-jakauman prosenttipistejakauman yht‰lˆ?

    // how many diameter classes
    class_count = (int) floor((d100 - d0) / DIAM_CLASS_WIDTH);
        rounds = 0;
    while (class_count < 14 && rounds <= MAX_ROUNDS)
    {  //If too few classes
        DIAM_CLASS_WIDTH = DIAM_CLASS_WIDTH / 2;
        class_count = (int) floor((d100 - d0) / DIAM_CLASS_WIDTH);
        rounds = rounds + 1;
    }
    if (class_count < 14)
    {
        ret = 0;
        constructErrorMessage (errorStr, "Too few generated diameter classes! Check D_gM and BA!", 15);
    }
    if (ret == 0) {
        strcat(errors, errorStr);
        return 0;
    }

    rounds = 0;
    while (class_count > 50 && rounds <= MAX_ROUNDS)
    {  //If too many classes
        DIAM_CLASS_WIDTH = DIAM_CLASS_WIDTH * 2;
        class_count = (int) floor((d100 - d0) / DIAM_CLASS_WIDTH);
        rounds = rounds + 1;
    }
    if (class_count > 50)
    {
        ret = 0;
        constructErrorMessage (errorStr, "Too many generated diameter classes! Check D_gM and BA!", 50);
    }
    if (class_count > MAX_CLASS_COUNT)
    {
        ret = 0;
        constructErrorMessage (errorStr, "Too many generated diameter classes! Check D_gM and BA! (MAX_CLASS_COUNT=", MAX_CLASS_COUNT);
    }
    if (b<=0)
    {
        ret = 0;
        constructErrorMessage (errorStr, "b<=0(b=",b);
    }
        if (ret == 0) {
        strcat(errors, errorStr);
        return 0;
    }

    // diameter distribution vector, elements contain diameter at breast height and number of stems / ha in the class
    class_count = construct_diamdist_using_Weibull_only_N_measured( a, b, c, DIAM_CLASS_WIDTH, BA, N, d0, d100, class_count, CLASS_N_MIN, dist );

    *nres = class_count;

    return ret;
}

int Diameter_distribution_spruce_Weibull_integer_class_Kilkki ( double BA, double D_gM, double DIAM_CLASS_WIDTH, double MAX_CLASS_COUNT,
                        double CLASS_N_MIN,
                        int *nres, struct diamclass *dist, char *errors,
                        int errorCheckMode, double allowedRiskLevel, double rectFactor)
{
    int ret, class_count, rounds;
    double a, b, c, d0, d100;
    char errorStr[200] = "";

    ret = 1;
    //Check for fatal errors
    if (D_gM<=0)
    {
        ret = 0;
        constructErrorMessage (errorStr, "D_gM<=0(D_gM=",D_gM);
    }
    if (BA<=0)
    {
        ret = 0;
        constructErrorMessage (errorStr, "BA<=0(BA=",BA);
    }
    if (ret == 0) {
        strcat(errors, errorStr);
        return 0;
    }

    // Weibull distribution cumulative distribution function parameters
    a = 0.001389+0.517444*D_gM;
    b = exp(-0.346223+0.934993*log(D_gM)-0.000925*BA);

    //Check for fatal errors
    if (b<=0)
    {
        ret = 0;
        constructErrorMessage (errorStr, "b<=0(b=",b);
    }
    if ((log((D_gM-a)/b))==0)
    {
        ret = 0;
        constructErrorMessage3val (errorStr, "(log((D_gM-a)/b))<=0(D_gM=", D_gM, ",a=",a, ",b=",b);
    }
    if (ret == 0) {
        strcat(errors, errorStr);
        return 0;
    }


    c = (log(-log(0.5)))/(log((D_gM-a)/b));
    d0 = a;   // low boundary
    //The right tail of distribution is very long with almost zero tree number.
    d100 = a + b * pow( -log( 1.0 - 0.999999999999 ), (1 / c)); // HUOM! Weibull-jakauman prosenttipistejakauman yht‰lˆ?

    // how many diameter classes
    class_count = (int) floor((d100 - d0) / DIAM_CLASS_WIDTH);
        rounds = 0;
    while (class_count < 14 && rounds <= MAX_ROUNDS && DIAM_CLASS_WIDTH >= 2)
    {  //If too few classes
        DIAM_CLASS_WIDTH = DIAM_CLASS_WIDTH / 2;
        class_count = (int) floor((d100 - d0) / DIAM_CLASS_WIDTH);
        rounds = rounds + 1;
    }
    if (class_count < 14)
    {
        ret = 0;
        constructErrorMessage (errorStr, "Too few generated diameter classes! Check D_gM and BA!", 15);
    }
    if (ret == 0) {
        strcat(errors, errorStr);
        return 0;
    }

    rounds = 0;
    while (class_count > 60 && rounds <= MAX_ROUNDS)
    {  //If too many classes
        DIAM_CLASS_WIDTH = DIAM_CLASS_WIDTH * 2;
        class_count = (int) floor((d100 - d0) / DIAM_CLASS_WIDTH);
        rounds = rounds + 1;
    }
    if (class_count > 60)
    {
        ret = 0;
        constructErrorMessage (errorStr, "Too many generated diameter classes! Check D_gM and BA!", 50);
    }
    if (class_count > MAX_CLASS_COUNT)
    {
        ret = 0;
        constructErrorMessage (errorStr, "Too many generated diameter classes! Check D_gM and BA! (MAX_CLASS_COUNT=", MAX_CLASS_COUNT);
    }
    if (ret == 0) {
        strcat(errors, errorStr);
        return 0;
    }

     // Define the nearest integer dbh classes for lower and upper diameters
        d0 = floor(d0);
        d100 = ceil(d100);

    // diameter distribution vector, elements contain diameter at breast height and number of stems / ha in the class
    class_count = construct_diamdist_using_Weibull( a, b, c, DIAM_CLASS_WIDTH, BA, d0, d100, class_count, CLASS_N_MIN, dist );

    *nres = class_count;

    return ret;
}

int Diameter_distribution_pine_Weibull_integer_class_Mykkanen ( double BA, double D_gM, double DIAM_CLASS_WIDTH, double MAX_CLASS_COUNT,
                            double CLASS_N_MIN,
                            int *nres, struct diamclass *dist, char *errors,
                            int errorCheckMode, double allowedRiskLevel, double rectFactor)
{
    int ret, class_count, rounds;
    double a, b, c, d0, d100;
    char errorStr[200] = "";

    ret = 1;
    //Check for fatal errors
    if (D_gM<=0)
    {
        ret = 0;
        constructErrorMessage (errorStr, "D_gM<=0(D_gM=",D_gM);
    }
    if (BA<=0)
    {
        ret = 0;
        constructErrorMessage (errorStr, "BA<=0(BA=",BA);
    }
    if (ret == 0) {
        strcat(errors, errorStr);
        return 0;
    }

    // Weibull distribution cumulative distribution function parameters
    a = exp(-1.306454+1.154433*log(D_gM));
    c = exp(0.647888+0.025530*D_gM-0.005558*BA);
    b = (D_gM-a)/(pow((-log(0.5)),(1/c)));

    // how many diameter classes (??????)
    // distribution boundaries
    d0 = a;   // low boundary
    d100 = a + b * pow( -log( 1.0 - 0.999999999999999 ), (1 / c)); // HUOM! Weibull-jakauman prosenttipistejakauman yht‰lˆ?

    // how many diameter classes
    class_count = (int) floor((d100 - d0) / DIAM_CLASS_WIDTH);
        rounds = 0;
    while (class_count < 14 && rounds <= MAX_ROUNDS && DIAM_CLASS_WIDTH >= 2)
    {  //If too few classes
        DIAM_CLASS_WIDTH = DIAM_CLASS_WIDTH / 2;
        class_count = (int) floor((d100 - d0) / DIAM_CLASS_WIDTH);
        rounds = rounds + 1;
    }
    if (class_count < 14)
    {
        ret = 0;
        constructErrorMessage (errorStr, "Too few generated diameter classes! Check D_gM and BA!", 15);
    }
    if (ret == 0) {
        strcat(errors, errorStr);
        return 0;
    }

        rounds = 0;
    while (class_count > 50 && rounds <= MAX_ROUNDS)
    {  //If too many classes
        DIAM_CLASS_WIDTH = DIAM_CLASS_WIDTH * 2;
        class_count = (int) floor((d100 - d0) / DIAM_CLASS_WIDTH);
        rounds = rounds + 1;
    }
    if (class_count > 50)
    {
        ret = 0;
        constructErrorMessage (errorStr, "Too many generated diameter classes! Check D_gM and BA!", 50);
    }
    if (class_count > MAX_CLASS_COUNT)
    {
        ret = 0;
        constructErrorMessage (errorStr, "Too many generated diameter classes! Check D_gM and BA! (MAX_CLASS_COUNT=", MAX_CLASS_COUNT);
    }
    if (b<=0)
    {
        ret = 0;
        constructErrorMessage (errorStr, "b<=0(b=",b);
    }
    if (ret == 0) {
        strcat(errors, errorStr);
        return 0;
    }

        // Define the nearest integer dbh classes for lower and upper diameters
        d0 = floor(d0);
        d100 = ceil(d100);

    // diameter distribution vector, elements contain diameter at breast height and number of stems / ha in the class
    class_count = construct_diamdist_using_Weibull( a, b, c, DIAM_CLASS_WIDTH, BA, d0, d100, class_count, CLASS_N_MIN, dist );

    *nres = class_count;

    return ret;
}


int Diameter_distribution_birch_Weibull_integer_class_Siipilehto (double BA, double D_gM, double A, double DIAM_CLASS_WIDTH,
                            double MAX_CLASS_COUNT,    double CLASS_N_MIN,
                            int *nres, struct diamclass *dist, char *errors,
                            int errorCheckMode, double allowedRiskLevel, double rectFactor)
{
    int ret, class_count, rounds;
    double a, b, c, d0, d100;
    char errorStr[200] = "";

    ret = 1;

    // Weibull distribution cumulative distribution function parameters
    a = 0;    // lower boundary not estimated, it is planned to be 0
    c = 3.7062+0.08632*D_gM+0.02185*A-0.1126*BA;

    //Check for fatal errors
    if (c<=0)
    {
        ret = 0;
        constructErrorMessage (errorStr, "c<=0(c=",c);
    }
    if (pow((-log(0.5)),(1/c))<=0)
    {
        ret = 0;
        constructErrorMessage (errorStr, "pow((-log(0.5)),(1/c))<=0(c=",c);
    }
    if (BA<=0)
    {
        ret = 0;
        constructErrorMessage (errorStr, "BA<=0(BA=",BA);
    }
    if (ret == 0) {
        strcat(errors, errorStr);
        return 0;
    }

    b = D_gM/pow((-log(0.5)),(1/c));

    // how many diameter classes (??????)
    // distribution boundaries
    d0 = a;   // low boundary
    d100 = a + b * pow( -log( 1.0 - 0.999999999999999 ), (1 / c)); // HUOM! Weibull-jakauman prosenttipistejakauman yht‰lˆ?

    // how many diameter classes
    class_count = (int) floor((d100 - d0) / DIAM_CLASS_WIDTH);
        rounds = 0;
    while (class_count < 14 && rounds <= MAX_ROUNDS && DIAM_CLASS_WIDTH >= 2)
    {  //If too few classes
        DIAM_CLASS_WIDTH = DIAM_CLASS_WIDTH / 2;
        class_count = (int) floor((d100 - d0) / DIAM_CLASS_WIDTH);
        rounds = rounds + 1;
    }
    if (class_count < 14)
    {
        ret = 0;
        constructErrorMessage (errorStr, "Too few generated diameter classes! Check D_gM and BA!", 15);
    }
    if (ret == 0) {
        strcat(errors, errorStr);
        return 0;
    }

        rounds = 0;
    while (class_count > 50 && rounds <= MAX_ROUNDS)
    {  //If too many classes
        DIAM_CLASS_WIDTH = DIAM_CLASS_WIDTH * 2;
        class_count = (int) floor((d100 - d0) / DIAM_CLASS_WIDTH);
        rounds = rounds + 1;
    }
    if (class_count > 50)
    {
        ret = 0;
        constructErrorMessage (errorStr, "Too many generated diameter classes! Check D_gM and BA!", 50);
    }
    if (class_count > MAX_CLASS_COUNT)
    {
        ret = 0;
        constructErrorMessage (errorStr, "Too many generated diameter classes! Check D_gM and BA! (MAX_CLASS_COUNT=", MAX_CLASS_COUNT);
    }
    if (b<=0)
    {
        ret = 0;
        constructErrorMessage (errorStr, "b<=0(b=",b);
    }
        if (ret == 0) {
        strcat(errors, errorStr);
        return 0;
    }

    // Define the nearest integer dbh classes for lower and upper diameters
        d0 = floor(d0);
        d100 = ceil(d100);

    // diameter distribution vector, elements contain diameter at breast height and number of stems / ha in the class
    class_count = construct_diamdist_using_Weibull( a, b, c, DIAM_CLASS_WIDTH, BA, d0, d100, class_count, CLASS_N_MIN, dist );

    *nres = class_count;

    return ret;
}


int Diameter_distribution_grey_alder_Weibull_integer_class_Karkiym (double BA, double D_gM, double DIAM_CLASS_WIDTH,
                            double MAX_CLASS_COUNT,    double CLASS_N_MIN,
                            int *nres, struct diamclass *dist, char *errors,
                            int errorCheckMode, double allowedRiskLevel, double rectFactor)
{
    int ret, class_count, rounds;
    double a, b, c, d0, d100;
    char errorStr[200] = "";

    ret = 1;

    //Check for fatal errors
    if (BA <= 0)
    {
        ret = 0;
        constructErrorMessage (errorStr, "BA<=0(BA=",BA);
    }
    if (D_gM<=0)
    {
        ret = 0;
        constructErrorMessage (errorStr, "D_gM<=0(D_gM=",D_gM);
    }
    if (ret == 0) {
        strcat(errors, errorStr);
        return 0;
    }

    // Weibull distribution cumulative distribution function parameters
    a = 0;    // lower boundary not estimated, it is planned to be 0
    c = 9.4529 - 1.3841 * log(pow(BA, 2)/D_gM);

    //Check for fatal errors
    if (c<=0)
    {
        ret = 0;
        constructErrorMessage (errorStr, "c<=0(c=",c);
    }
    if (ret == 0) {
        strcat(errors, errorStr);
        return 0;
    }

    b = exp(0.2915+0.9242*log(D_gM));

      // how many diameter classes (??????)
    // distribution boundaries
    d0 = a;   // low boundary
    d100 = a + b * pow( -log( 1.0 - 0.999999999999999 ), (1 / c)); // HUOM! Weibull-jakauman prosenttipistejakauman yht‰lˆ?

    // how many diameter classes
    class_count = (int) floor((d100 - d0) / DIAM_CLASS_WIDTH);
        rounds = 0;
    while (class_count < 14 && rounds <= MAX_ROUNDS && DIAM_CLASS_WIDTH >= 2)
    {  //If too few classes
        DIAM_CLASS_WIDTH = DIAM_CLASS_WIDTH / 2;
        class_count = (int) floor((d100 - d0) / DIAM_CLASS_WIDTH);
        rounds = rounds + 1;
    }
    if (class_count < 14)
    {
        ret = 0;
        constructErrorMessage (errorStr, "Too few generated diameter classes! Check D_gM and BA!", 15);
    }
    if (ret == 0) {
        strcat(errors, errorStr);
        return 0;
    }

        rounds = 0;
    while (class_count > 50 && rounds <= MAX_ROUNDS)
    {  //If too many classes
        DIAM_CLASS_WIDTH = DIAM_CLASS_WIDTH * 2;
        class_count = (int) floor((d100 - d0) / DIAM_CLASS_WIDTH);
        rounds = rounds + 1;
    }
    if (class_count > 50)
    {
        ret = 0;
        constructErrorMessage (errorStr, "Too many generated diameter classes! Check D_gM and BA!", 50);
    }
    if (class_count > MAX_CLASS_COUNT)
    {
        ret = 0;
        constructErrorMessage (errorStr, "Too many generated diameter classes! Check D_gM and BA! (MAX_CLASS_COUNT=", MAX_CLASS_COUNT);
    }
    if (b<=0)
    {
        ret = 0;
        constructErrorMessage (errorStr, "b<=0(b=",b);
    }
        if (ret == 0) {
        strcat(errors, errorStr);
        return 0;
    }

        // Define the nearest integer dbh classes for lower and upper diameters
        d0 = floor(d0);
        d100 = ceil(d100);

    // diameter distribution vector, elements contain diameter at breast height and number of stems / ha in the class
    class_count = construct_diamdist_using_Weibull( a, b, c, DIAM_CLASS_WIDTH, BA, d0, d100, class_count, CLASS_N_MIN, dist );

    *nres = class_count;

    return ret;
}

int Diameter_distribution_pine_with_stem_number_JohnsonSB_Siipilehto ( double BA, double D_gM, double N, double DIAM_CLASS_WIDTH,
                                    double MAX_CLASS_COUNT, double CLASS_N_MIN, double D_MIN, int *nres, struct diamclass *dist, char *errors,
                                    int errorCheckMode, double allowedRiskLevel, double rectFactor)
{
    int ret, class_count, rounds;
    double BA_gM, Shape, upper_bound, lower_bound, g, y, d0, d100;
    char errorStr[200] = "";

    ret = 1;

    // Johnson SB distribution cumulative distribution function parameters

    BA_gM = (PI/4)*pow((D_gM/100),2);
    Shape = BA/(BA_gM*N);
    upper_bound = -15.4051+0.8758*D_gM+39.4779*Shape;
    g = -1.8339+2.8417*Shape;
    //Set lower bound for distribution!!
    if (D_MIN > 0.5 * D_gM)
        lower_bound = 0.5 * D_gM;
    else
        lower_bound = D_MIN;

    y = g*log(upper_bound+lower_bound-D_gM)-g*log(D_gM-lower_bound);

    //Check for fatal errors
    if (BA_gM<=0)
    {
        ret = 0;
        constructErrorMessage (errorStr, "BA_gM<=0(BA_gM=",BA_gM);
    }
    if (BA<=0)
    {
        ret = 0;
        constructErrorMessage (errorStr, "BA<=0(BA=",BA);
    }
    if (N<=0)
    {
        ret = 0;
        constructErrorMessage (errorStr, "N<=0(N=",N);
    }
    if (g<0)
    {
        ret = 0;
        constructErrorMessage (errorStr, "g<0(g=",g);
    }
    if (D_gM>=upper_bound)
    {
        ret = 0;
        constructErrorMessage2val (errorStr, "D_gM>=upper_bound(D_gM=",D_gM ,"upper_bound=",upper_bound);
    }
    if (ret == 0) {
        strcat(errors, errorStr);
        return 0;
    }

    // distribution boundaries (to avoid negative logarithms 0.000001 was added to lower bound)
    d0 = lower_bound+0.000000000000001;
    d100 = upper_bound;

    // how many diameter classes
    class_count = (int) floor((d100 - d0) / DIAM_CLASS_WIDTH);
        rounds = 0;
    while (class_count > 50 && rounds <= MAX_ROUNDS)
    {  //If too many classes
        DIAM_CLASS_WIDTH = DIAM_CLASS_WIDTH * 2;
        class_count = (int) floor((d100 - d0) / DIAM_CLASS_WIDTH);
        rounds = rounds + 1;
    }
    if (class_count > 50)
    {
        ret = 0;
        constructErrorMessage (errorStr, "Too many generated diameter classes! Check D_gM and BA!", 50);
    }
    if (class_count > MAX_CLASS_COUNT)
    {
        ret = 0;
        constructErrorMessage (errorStr, "Too many generated diameter classes! Check D_gM and BA! (MAX_CLASS_COUNT=", MAX_CLASS_COUNT);
    }
    if (ret == 0) {
        strcat(errors, errorStr);
        return 0;
    }

    // diameter distribution vector, elements contain diameter at breast height and number of stems / ha in the class
    class_count = construct_diamdist_using_SB (upper_bound, g, y, lower_bound, DIAM_CLASS_WIDTH, N, BA, d0, d100, class_count, CLASS_N_MIN, dist, nres );

    *nres = class_count;

    return ret;
}


int Diameter_distribution_spruce_with_stem_number_JohnsonSB_Siipilehto ( double BA, double D_gM, double N, double DIAM_CLASS_WIDTH,
                                    double MAX_CLASS_COUNT, double CLASS_N_MIN, double D_MIN, int *nres, struct diamclass *dist, char *errors,
                                    int errorCheckMode, double allowedRiskLevel, double rectFactor)
{
    int ret, class_count, rounds;
    double BA_gM, Shape, upper_bound, lower_bound, g, y, d0, d100;
    char errorStr[200] = "";

    ret = 1;

    // Johnson SB distribution cumulative distribution function parameters
    BA_gM = (PI/4)*pow((D_gM/100),2);
    Shape = BA/(BA_gM*N);
    upper_bound = 5.3617+0.954*D_gM+18.5939*Shape;
    g = -0.423+3.5742*log(1+Shape);
    //Set lower bound for distribution!!
    if (D_MIN > 0.5 * D_gM)
        lower_bound = 0.5 * D_gM;
    else
        lower_bound = D_MIN;

    y = g*log(upper_bound+lower_bound-D_gM)-g*log(D_gM-lower_bound);

    //Check for fatal errors
    if (BA_gM<=0)
    {
        ret = 0;
        constructErrorMessage (errorStr, "BA_gM<=0(BA_gM=",BA_gM);
    }
    if (BA<=0)
    {
        ret = 0;
        constructErrorMessage (errorStr, "BA<=0(BA=",BA);
    }
    if (N<=0)
    {
        ret = 0;
        constructErrorMessage (errorStr, "N<=0(N=",N);
    }
    if (g<0)
    {
        ret = 0;
        constructErrorMessage (errorStr, "g<0(g=",g);
    }
    if (D_gM>=upper_bound)
    {
        ret = 0;
        constructErrorMessage2val (errorStr, "D_gM>=upper_bound(D_gM=",D_gM ,"upper_bound=",upper_bound);
    }
    if (ret == 0) {
        strcat(errors, errorStr);
        return 0;
    }

    // distribution boundaries (to avoid negative logarithms 0.000001 was added to lower bound)
    d0 = lower_bound+0.000000000000001;
    d100 = upper_bound;

    // how many diameter classes
    class_count = (int) floor((d100 - d0) / DIAM_CLASS_WIDTH);
        rounds = 0;
    while (class_count > 50 && rounds <= MAX_ROUNDS)
    {  //If too many classes
        DIAM_CLASS_WIDTH = DIAM_CLASS_WIDTH * 2;
        class_count = (int) floor((d100 - d0) / DIAM_CLASS_WIDTH);
        rounds = rounds + 1;
    }
    if (class_count > 50)
    {
        ret = 0;
        constructErrorMessage (errorStr, "Too many generated diameter classes! Check D_gM and BA!", 50);
    }
    if (class_count > MAX_CLASS_COUNT)
    {
        ret = 0;
        constructErrorMessage (errorStr, "Too many generated diameter classes! Check D_gM and BA! (MAX_CLASS_COUNT=", MAX_CLASS_COUNT);
    }
        if (ret == 0) {
        strcat(errors, errorStr);
        return 0;
    }

    // diameter distribution vector, elements contain diameter at breast height and number of stems / ha in the class
    class_count = construct_diamdist_using_SB (upper_bound, g, y, lower_bound, DIAM_CLASS_WIDTH, N, BA, d0, d100, class_count, CLASS_N_MIN, dist, nres );

    *nres = class_count;

    return ret;
}

int Diameter_distribution_birch_with_stem_number_JohnsonSB_Siipilehto ( double BA, double D_gM, double N, double DIAM_CLASS_WIDTH,
                                    double MAX_CLASS_COUNT, double CLASS_N_MIN, double D_MIN, int *nres, struct diamclass *dist, char *errors,
                                    int errorCheckMode, double allowedRiskLevel, double rectFactor)
{
    int ret, class_count, rounds;
    double BA_gM, Shape, upper_bound, lower_bound, g, y, d0, d100;
    char errorStr[200] = "";

    ret = 1;

    // Johnson SB distribution cumulative distribution function parameters
    BA_gM = (PI/4)*pow((D_gM/100),2);
    Shape = BA/(BA_gM*N);
    upper_bound = -0.763+1.1181*D_gM+18.3167*Shape;
    g = -5.0019-0.01897*D_gM+5.0148*Shape+1.4746*(1/Shape);
    //Set lower bound for distribution!!
    if (D_MIN > 0.5 * D_gM)
        lower_bound = 0.5 * D_gM;
    else
        lower_bound = D_MIN;

    y = g*log(upper_bound+lower_bound-D_gM)-g*log(D_gM-lower_bound);

    //Check for fatal errors
    if (BA_gM<=0)
    {
        ret = 0;
        constructErrorMessage (errorStr, "BA_gM<=0(BA_gM=",BA_gM);
    }
    if (BA<=0)
    {
        ret = 0;
        constructErrorMessage (errorStr, "BA<=0(BA=",BA);
    }
    if (N<=0)
    {
        ret = 0;
        constructErrorMessage (errorStr, "N<=0(N=",N);
    }
    if (g<0)
    {
        ret = 0;
        constructErrorMessage (errorStr, "g<0(g=",g);
    }
    if (D_gM>=upper_bound)
    {
        ret = 0;
        constructErrorMessage2val (errorStr, "D_gM>=upper_bound(D_gM=",D_gM ,"upper_bound=",upper_bound);
    }
    if (ret == 0) {
        strcat(errors, errorStr);
        return 0;
    }

    // distribution boundaries (to avoid negative logarithms 0.000001 was added to lower bound)
    d0 = lower_bound+0.000000000000001;
    d100 = upper_bound;

    // how many diameter classes
    class_count = (int) floor((d100 - d0) / DIAM_CLASS_WIDTH);
        rounds = 0;
    while (class_count > 50 && rounds <= MAX_ROUNDS)
    {  //If too many classes
        DIAM_CLASS_WIDTH = DIAM_CLASS_WIDTH * 2;
        class_count = (int) floor((d100 - d0) / DIAM_CLASS_WIDTH);
        rounds = rounds + 1;
    }
    if (class_count > 50)
    {
        ret = 0;
        constructErrorMessage (errorStr, "Too many generated diameter classes! Check D_gM and BA!", 50);
    }
    if (class_count > MAX_CLASS_COUNT)
    {
        ret = 0;
        constructErrorMessage (errorStr, "Too many generated diameter classes! Check D_gM and BA! (MAX_CLASS_COUNT=", MAX_CLASS_COUNT);
    }
        if (ret == 0) {
        strcat(errors, errorStr);
        return 0;
    }

    // diameter distribution vector, elements contain diameter at breast height and number of stems / ha in the class
    class_count = construct_diamdist_using_SB (upper_bound, g, y, lower_bound, DIAM_CLASS_WIDTH, N, BA, d0, d100, class_count, CLASS_N_MIN, dist, nres );

    *nres = class_count;

    return ret;
}


int Diameter_distribution_pine_JohnsonSB_Siipilehto ( double D_gM, double A, double H_gM, double BA, double DIAM_CLASS_WIDTH,
                                double MAX_CLASS_COUNT, double CLASS_N_MIN, double D_MIN, int *nres, struct diamclass *dist, char *errors,
                                int errorCheckMode, double allowedRiskLevel, double rectFactor)
{
    int ret, class_count, rounds;
    double F, upper_bound, lower_bound, g, y, d0, d100;
    char errorStr[200] = "";

    ret = 1;

    // Johnson SB distribution cumulative distribution function parameters
    F = H_gM/D_gM;
    upper_bound = -17.5244+14.667*log(A);
    g = exp(-1.4617+0.9928*F+0.3063*log(A));
    //Set lower bound for distribution!!
    if (D_MIN > 0.5 * D_gM)
        lower_bound = 0.5 * D_gM;
    else
        lower_bound = D_MIN;

    y = g*log(upper_bound+lower_bound-D_gM)-g*log(D_gM-lower_bound);

    //Check for fatal errors
    if (D_gM<=0)
    {
        ret = 0;
        constructErrorMessage (errorStr, "D_gM<=0(D_gM=",D_gM);
    }
    if (BA<=0)
    {
        ret = 0;
        constructErrorMessage (errorStr, "BA<=0(BA=",BA);
    }
    if (A<=0)
    {
        ret = 0;
        constructErrorMessage (errorStr, "A<=0(A=",A);
    }
    if (D_gM>=upper_bound)
    {
        ret = 0;
        constructErrorMessage2val (errorStr, "D_gM>=upper_bound(D_gM=",D_gM ,"upper_bound=",upper_bound);
    }
    if (ret == 0) {
        strcat(errors, errorStr);
        return 0;
    }

    // distribution boundaries (to avoid negative logarithms 0.000001 was added to lower bound)
    d0 = lower_bound+0.000000000000001;
    d100 = upper_bound;

    // how many diameter classes
    class_count = (int) floor((d100 - d0) / DIAM_CLASS_WIDTH);
        rounds = 0;
    while (class_count > 50 && rounds <= MAX_ROUNDS)
    {  //If too many classes
        DIAM_CLASS_WIDTH = DIAM_CLASS_WIDTH * 2;
        class_count = (int) floor((d100 - d0) / DIAM_CLASS_WIDTH);
        rounds = rounds + 1;
    }
    if (class_count > 50)
    {
        ret = 0;
        constructErrorMessage (errorStr, "Too many generated diameter classes! Check D_gM and BA!", 50);
    }
    if (class_count > MAX_CLASS_COUNT)
    {
        ret = 0;
        constructErrorMessage (errorStr, "Too many generated diameter classes! Check D_gM and BA! (MAX_CLASS_COUNT=", MAX_CLASS_COUNT);
    }
        if (ret == 0) {
        strcat(errors, errorStr);
        return 0;
    }

    // diameter distribution vector, elements contain diameter at breast height and number of stems / ha in the class
    class_count = construct_diamdist_using_SB (upper_bound, g, y, lower_bound, DIAM_CLASS_WIDTH, 0, BA, d0, d100, class_count, CLASS_N_MIN, dist, nres );

    *nres = class_count;

    return ret;
}



int Diameter_distribution_spruce_JohnsonSB_Siipilehto (double D_gM, double A, double H_gM, double BA, double DIAM_CLASS_WIDTH,
                            double MAX_CLASS_COUNT, double CLASS_N_MIN, double D_MIN,
                            int *nres, struct diamclass *dist, char *errors,
                            int errorCheckMode, double allowedRiskLevel, double rectFactor)
{
    int ret, class_count, rounds;
    double F, upper_bound, lower_bound, g, y, d0, d100;
    char errorStr[200] = "";

    ret = 1;

    // Johnson SB distribution cumulative distribution function parameters
    F = H_gM/D_gM;
    upper_bound = 14.658+0.6519*D_gM+0.2089*A;
    g = exp(1.5046+0.9989*F+0.0227*A-0.8177*log(A));
    //Set lower bound for distribution!!
    if (D_MIN > 0.5 * D_gM)
        lower_bound = 0.5 * D_gM;
    else
        lower_bound = D_MIN;

    y = g*log(upper_bound+lower_bound-D_gM)-g*log(D_gM-lower_bound);

    //Check for fatal errors
    if (D_gM<=0)
    {
        ret = 0;
        constructErrorMessage (errorStr, "D_gM<=0(D_gM=",D_gM);
    }
    if (BA<=0)
    {
        ret = 0;
        constructErrorMessage (errorStr, "BA<=0(BA=",BA);
    }
    if (A<=0)
    {
        ret = 0;
        constructErrorMessage (errorStr, "A<=0(A=",A);
    }
    if (D_gM>=upper_bound)
    {
        ret = 0;
        constructErrorMessage2val (errorStr, "D_gM>=upper_bound(D_gM=",D_gM ,"upper_bound=",upper_bound);
    }
    if (ret == 0) {
        strcat(errors, errorStr);
        return 0;
    }

    // distribution boundaries (to avoid negative logarithms 0.000001 was added to lower bound)
    d0 = lower_bound+0.000000000000001;
    d100 = upper_bound;

    // how many diameter classes
    class_count = (int) floor((d100 - d0) / DIAM_CLASS_WIDTH);
        rounds = 0;
    while (class_count > 50 && rounds <= MAX_ROUNDS)
    {  //If too many classes
        DIAM_CLASS_WIDTH = DIAM_CLASS_WIDTH * 2;
        class_count = (int) floor((d100 - d0) / DIAM_CLASS_WIDTH);
        rounds = rounds + 1;
    }
    if (class_count > 50)
    {
        ret = 0;
        constructErrorMessage (errorStr, "Too many generated diameter classes! Check D_gM and BA!", 50);
    }
    if (class_count > MAX_CLASS_COUNT)
    {
        ret = 0;
        constructErrorMessage (errorStr, "Too many generated diameter classes! Check D_gM and BA! (MAX_CLASS_COUNT=", MAX_CLASS_COUNT);
    }
        if (ret == 0) {
        strcat(errors, errorStr);
        return 0;
    }

    // diameter distribution vector, elements contain diameter at breast height and number of stems / ha in the class
    class_count = construct_diamdist_using_SB (upper_bound, g, y, lower_bound, DIAM_CLASS_WIDTH, 0, BA, d0, d100, class_count, CLASS_N_MIN, dist, nres );

    *nres = class_count;

    return ret;
}


int Diameter_distribution_birch_JohnsonSB_Siipilehto ( double D_gM, double A, double BA, double DIAM_CLASS_WIDTH,
                                double MAX_CLASS_COUNT, double CLASS_N_MIN, double D_MIN, int *nres, struct diamclass *dist, char *errors,
                                int errorCheckMode, double allowedRiskLevel, double rectFactor)
{
    int ret, class_count, rounds;
    double upper_bound, lower_bound, g, y, d0, d100;
    char errorStr[200] = "";

    ret = 1;

    // Johnson SB distribution cumulative distribution function parameters
    upper_bound = 13.1531+1.1977*D_gM;
    g = exp(0.03029+0.1482*log(A));
    //Set lower bound for distribution!!
    if (D_MIN > 0.5 * D_gM)
        lower_bound = 0.5 * D_gM;
    else
        lower_bound = D_MIN;

    y = g*log(upper_bound+lower_bound-D_gM)-g*log(D_gM-lower_bound);

    //Check for fatal errors
    if (A<=0)
    {
        ret = 0;
        constructErrorMessage (errorStr, "A<=0(A=",A);
    }
    if (BA<=0)
    {
        ret = 0;
        constructErrorMessage (errorStr, "BA<=0(BA=",BA);
    }
    if (D_gM>=upper_bound)
    {
        ret = 0;
        constructErrorMessage2val (errorStr, "D_gM>=upper_bound(D_gM=",D_gM ,"upper_bound=",upper_bound);
    }
    if (ret == 0) {
        strcat(errors, errorStr);
        return 0;
    }

    // distribution boundaries (to avoid negative logarithms 0.000001 was added to lower bound)
    d0 = lower_bound+0.000000000000001;
    d100 = upper_bound;

    // how many diameter classes
    class_count = (int) floor((d100 - d0) / DIAM_CLASS_WIDTH);
        rounds = 0;
    while (class_count > 50 && rounds <= MAX_ROUNDS)
    {  //If too many classes
        DIAM_CLASS_WIDTH = DIAM_CLASS_WIDTH * 2;
        class_count = (int) floor((d100 - d0) / DIAM_CLASS_WIDTH);
        rounds = rounds + 1;
    }
    if (class_count > 50)
    {
        ret = 0;
        constructErrorMessage (errorStr, "Too many generated diameter classes! Check D_gM and BA!", 50);
    }
    if (class_count > MAX_CLASS_COUNT)
    {
        ret = 0;
        constructErrorMessage (errorStr, "Too many generated diameter classes! Check D_gM and BA! (MAX_CLASS_COUNT=", MAX_CLASS_COUNT);
    }
        if (ret == 0) {
        strcat(errors, errorStr);
        return 0;
    }

    // diameter distribution vector, elements contain diameter at breast height and number of stems / ha in the class
    class_count = construct_diamdist_using_SB (upper_bound, g, y, lower_bound, DIAM_CLASS_WIDTH, 0, BA, d0, d100, class_count, CLASS_N_MIN, dist, nres );

    *nres = class_count;

    return ret;
}

static void construct_heightdist_using_normaldist ( double H_gM, double sd, double N, double HEIGHT_CLASS_WIDTH,
                        double h0, double h100, int class_count, struct heightclass *dist)
{

    int i;
    double low_boundary, high_boundary, low_mass, high_mass;
    double low_mass_left_out, high_mass_left_out, mass_scaler;

    //probability masses not included in distribution classes
    low_mass_left_out = normal_cdf(h0, H_gM, sd);
    high_mass_left_out = 1 - (normal_cdf(h100, H_gM, sd));
    mass_scaler = 1 / (1 - (low_mass_left_out + high_mass_left_out));

    //heights & stem numbers
    for ( i = 0; i < class_count; i++)  {
        if ( i != class_count - 1) {    // all but the last class that has different class length
            if ( i == 0) {    // first class
                dist[i].h = h0 + HEIGHT_CLASS_WIDTH/2;
            }
            else {    // intermediate classes
                dist[i].h = dist[i-1].h + HEIGHT_CLASS_WIDTH;
            }
            // class boundaries for estimation of stem number
            low_boundary = dist[i].h - HEIGHT_CLASS_WIDTH / 2;
            high_boundary = dist[i].h + HEIGHT_CLASS_WIDTH / 2;

        }
        else {    // last class
            dist[i].h = dist[i-1].h + HEIGHT_CLASS_WIDTH/2 + (h100- (dist[i-1].h + HEIGHT_CLASS_WIDTH/2)) / 2;
            low_boundary = dist[i-1].h + HEIGHT_CLASS_WIDTH / 2;
            high_boundary = h100;
        }

        // cumulated probabilities at boundaries
        low_mass = normal_cdf(low_boundary, H_gM, sd);
        high_mass = normal_cdf(high_boundary, H_gM, sd);
        //stem number
        dist[i].n = N * (( high_mass - low_mass ) * mass_scaler);
    }
}


int Normal_height_distribution (double H_gM, double N, double HEIGHT_CLASS_NUM, double CULT,
                                double max_class_count, int *nres, struct heightclass *dist, char *errors,
                                int errorCheckMode, double allowedRiskLevel, double rectFactor)
{
    int ret;
    double sd, h0, h100, HEIGHT_CLASS_WIDTH;
    char errorStr[100] = "";

    ret = 1;

    //Check for fatal errors
    if (N<=0)
    {
        ret = 0;
        constructErrorMessage (errorStr, "N<=0(N=",N);
    }
    if (H_gM<=0)
    {
        ret = 0;
        constructErrorMessage (errorStr, "H_gM<=0(H_gM=",H_gM);
    }
    if (HEIGHT_CLASS_NUM<=0)
    {
        ret = 0;
        constructErrorMessage (errorStr, "HEIGHT_CLASS_NUM<=0(HEIGHT_CLASS_NUM=",HEIGHT_CLASS_NUM);
    }
    if (ret == 0) {
        strcat(errors, errorStr);
        return 0;
    }

    // Normal distribution cumulative distribution function height in lower and upper limits
    if(CULT==1)
        sd=0.08*H_gM;
    else
        sd=0.13*H_gM;

    h0 = normal_cdf_inv (0.000000000000001, H_gM, sd);

    h100 = normal_cdf_inv (0.999999999999999, H_gM, sd);


    // How width one height class is (??????)
    HEIGHT_CLASS_WIDTH = floor((h100*100000-h0*100000)/HEIGHT_CLASS_NUM)/100000;

    // height distribution vector, elements contain height and number of stems / ha in the class
    construct_heightdist_using_normaldist ( H_gM, sd, N, HEIGHT_CLASS_WIDTH, h0, h100, HEIGHT_CLASS_NUM, dist);

    *nres = HEIGHT_CLASS_NUM;

    return ret;
}



static int construct_heightdist_using_normaldist_with_diameters ( double H_gM, double sd, double N, double HEIGHT_CLASS_WIDTH,
                        double h0, double h100, double SP, double D_gM, int class_count, double class_n_min, struct heightclass *dist)
{

    int i, j;
    double low_boundary, high_boundary, low_mass, high_mass, stem_number, left_out;
    double sum_ppa, sum_ppa_d, ppa, ppa_d, D_g, step, counter1, counter2, counter3;
#ifdef __cplusplus
    double** dist_2 = NULL;
    dist_2 = new double*[class_count];
    for (int i=0; i < class_count; i++) {
        dist_2[i] = new double[3];
        for (int j=0; j<3; j++) {
            dist_2[i][j] = 0;
        }
    }
#else
    double dist_2[class_count][3];
#endif
    left_out = 0;
    counter1 = 0;
    counter2 = 0;
    counter3 = 0;

    //heights & stem numbers
    for ( i = 0; i < class_count; i++)
    {
        if ( i != class_count - 1)
        {    // all but the last class that has different class length
            if ( i == 0)
            {    // first class
                dist_2[i][0] = h0 + HEIGHT_CLASS_WIDTH/2;
            }
            else
            {    // intermediate classes
                dist_2[i][0] = dist_2[i-1][0] + HEIGHT_CLASS_WIDTH;
            }
            // class boundaries for estimation of stem number
            low_boundary = dist_2[i][0] - HEIGHT_CLASS_WIDTH / 2;
            high_boundary = dist_2[i][0] + HEIGHT_CLASS_WIDTH / 2;

        }
        else
        {    // last class
            dist_2[i][0] = dist_2[i-1][0] + HEIGHT_CLASS_WIDTH/2 + (h100- (dist_2[i-1][0] + HEIGHT_CLASS_WIDTH/2)) / 2;
            low_boundary = dist_2[i-1][0] + HEIGHT_CLASS_WIDTH / 2;
            high_boundary = h100;
        }
        //diameter using Ojansuu's models for little trees
        double epsilon = 0.0001;
        if(dist_2[i][0]<1.3-epsilon)
            dist_2[i][1]=0;
        else if(dist_2[i][0]>=1.3-epsilon&&dist_2[i][0]<=1.3)
        {
                dist_2[i][0]=1.3;
            dist_2[i][1]=0.1+(counter1/100);
            counter1++;
            }
        else if (dist_2[i][0]>1.3&&dist_2[i][0]<=1.3+epsilon)
        {
                dist_2[i][0]=1.3002;
                dist_2[i][1]=0.1+(counter2/100);
                counter2++;
            }
        else
        {
            if (SP==1||SP==8) dist_2[i][1] = exp(0.7305+0.8973*log(dist_2[i][0]-1.3)+pow(0.0539,2)/2);
            if (SP==2) dist_2[i][1] = exp(0.7204+0.8352*log(dist_2[i][0]-1.3)+pow(0.0385,2)/2);
            if ((SP>2&&SP<8)||SP==9) dist_2[i][1] = exp(-0.6607+1.6042*log(dist_2[i][0]-1.3)+pow(0.07637,2)/2);
        }

        // cumulated probabilities at boundaries
        low_mass = normal_cdf(low_boundary, H_gM, sd);
        high_mass = normal_cdf(high_boundary, H_gM, sd);
            //stem number
        stem_number = N * ( high_mass - low_mass );
        //Calculate basal area mass of rejected classes
        if (stem_number < class_n_min)
             left_out = left_out + stem_number;
         dist_2[i][2] = stem_number;
    }

    //Calculate new stem numbers to the classes
    int class_count_new;
    double mass_scaler;
    int k = 0;
    //Assign the first class to biggest
    int biggest_class = 0;
    class_count_new = class_count;
    mass_scaler = N / (N - left_out);
    for ( i = 0; i < class_count; i++)
    {
        if (dist_2[i][2] > dist_2[biggest_class][2])
                 biggest_class = i;
        if (dist_2[i][2] < class_n_min)
             class_count_new = class_count_new - 1;
            else
            {
             dist[k].h = dist_2[i][0];
             dist[k].d = dist_2[i][1];
             dist[k].n = mass_scaler * dist_2[i][2];
             //seedling_below13 dummy for little trees
             if(dist_2[i][0]<1.3)
             {
                 dist[k].seedling_below13=1;
                 dist[k].seedling_to_mature13=0;
             }
             else
             {
                 dist[k].seedling_below13=0;
                 dist[k].seedling_to_mature13=1;
             }
             k++;
        }
    }
    //Simple distribution for the stratum with only few or none classes left
    if (class_count_new < 3)
    {
        dist[0].h = dist_2[biggest_class-1][0];
        dist[0].d = dist_2[biggest_class-1][1];
        dist[0].n = N/3.;
        if(dist_2[biggest_class-1][0]<1.3)
        {
             dist[0].seedling_below13=1;
             dist[0].seedling_to_mature13=0;
        }
        else
         {
             dist[0].seedling_below13=0;
             dist[0].seedling_to_mature13=1;
        }
        dist[1].h = dist_2[biggest_class-1][0];
        dist[1].d = dist_2[biggest_class][1];
        dist[1].n = N/3.;
        if(dist_2[biggest_class][0]<1.3)
        {
             dist[1].seedling_below13=1;
             dist[1].seedling_to_mature13=0;
        }
        else
         {
             dist[1].seedling_below13=0;
             dist[1].seedling_to_mature13=1;
        }
        dist[2].h = dist_2[biggest_class-1][0];
        dist[2].d = dist_2[biggest_class+1][1];
        dist[2].n = N/3.;
        if(dist_2[biggest_class+1][0]<1.3)
        {
             dist[2].seedling_below13=1;
             dist[2].seedling_to_mature13=0;
        }
        else
         {
             dist[2].seedling_below13=0;
             dist[2].seedling_to_mature13=1;
        }
        class_count_new = 3;
    }

    //Scale tree diameters so that D_gM is the same as measured
    //First check if D_gM is too little:

    if(SP==1||SP==8||SP==2) {
        if(D_gM<0.6)
            D_gM=0.6;
    }
    if((SP>2&&SP<8)||SP==9) {
        if(D_gM<0.15)
            D_gM=0.15;
    }

    sum_ppa=0;
    sum_ppa_d=0;
    for ( i = 0; i < class_count_new; i++)  {
        ppa=dist[i].n*PI*pow((dist[i].d/2),2);
        ppa_d=dist[i].n*PI*pow((dist[i].d/2),2)*dist[i].d;
        sum_ppa=sum_ppa+ppa;
        sum_ppa_d=sum_ppa_d+ppa_d;
    }
    D_g = sum_ppa_d / sum_ppa;
    j = 1;
    step = 0.01;

    do {
        if (D_g<D_gM&&fabs(D_g-D_gM)>step)
            for ( i = 0; i < class_count; i++)  {
                if(dist[i].d!=0)
                    dist[i].d=dist[i].d+step;
            }
        if (D_g>D_gM&&fabs(D_g-D_gM)>step)
            for ( i = 0; i < class_count; i++)  {
                if(dist[i].d>step)
                    dist[i].d=dist[i].d-step;
            }
        sum_ppa=0;
        sum_ppa_d=0;
        for ( i = 0; i < class_count_new; i++)  {
            ppa = dist[i].n*PI*pow((dist[i].d/2),2);
            ppa_d = dist[i].n*PI*pow((dist[i].d/2),2)*dist[i].d;
            sum_ppa=sum_ppa+ppa;
            sum_ppa_d=sum_ppa_d+ppa_d;
        }
        D_g=sum_ppa_d/sum_ppa;
        j++;
    }
    while(fabs(D_g-D_gM)>step);

    //To avoid too little diameter values
    for ( i = 0; i < class_count_new; i++)  {
        if(dist[i].d>0&&dist[i].d<=0.1) {
            dist[i].d=0.1+(counter3/100);
            counter3++;
        }
    }
#ifdef __cplusplus
    for (int i=0; i < class_count; i++) {
        delete[] dist_2[i];
        dist_2[i] = NULL;
    }
    delete[] dist_2;
    dist_2 = NULL;
#endif
    return class_count_new;

}


int Normal_height_distribution_with_diameters (double H_gM, double CULT, double N, double HEIGHT_CLASS_NUM,
                        double SP, double D_gM, double MAX_CLASS_COUNT, double CLASS_N_MIN,
                        int *nres, struct heightclass *dist, char *errors,
                        int errorCheckMode, double allowedRiskLevel, double rectFactor)
{
    int ret;
    double sd, h0, h100, HEIGHT_CLASS_WIDTH;
    char errorStr[100] = "";

    ret = 1;

    //Check for fatal errors
    if (N<=0)
    {
        ret = 0;
        constructErrorMessage (errorStr, "N<=0 (N=",N);
    }
    if (H_gM<=0)
    {
        ret = 0;
        constructErrorMessage (errorStr, "H_gM<=0 (H_gM=",H_gM);
    }
    if (HEIGHT_CLASS_NUM<=0)
    {
        ret = 0;
        constructErrorMessage (errorStr, "HEIGHT_CLASS_NUM<=0 (HEIGHT_CLASS_NUM=",HEIGHT_CLASS_NUM);
    }
    if (ret == 0) {
        strcat(errors, errorStr);
        return ret;
    }

    // Normal distribution cumulative distribution function height in lower and upper limits
    if(CULT==1)
        sd=0.08*H_gM;
    else
        sd=0.13*H_gM;

    h0 = normal_cdf_inv (0.0000000001, H_gM, sd);

    h100 = normal_cdf_inv (0.9999999999, H_gM, sd);


    // Calculate the width of class
    HEIGHT_CLASS_WIDTH = floor((h100*100000-h0*100000)/HEIGHT_CLASS_NUM)/100000;

    // height distribution vector, elements contain height and number of stems / ha in the class
    //Height class number will have a new value if minimum stem number per class > 0
    HEIGHT_CLASS_NUM = construct_heightdist_using_normaldist_with_diameters(H_gM, sd, N, HEIGHT_CLASS_WIDTH, h0, h100, SP, D_gM, HEIGHT_CLASS_NUM, CLASS_N_MIN, dist);

    *nres = HEIGHT_CLASS_NUM;

    return ret;
}


int Diameter_distribution_seedtree_JKK (double N, double D_gM, double H_gM, double MAX_CLASS_COUNT,
                                        int *nres, struct diamclass_seedtree *dist, char *errors,
                                        int errorCheckMode, double allowedRiskLevel, double rectFactor)
{
    int ret = 1;

    dist[0].d = D_gM*0.95;
    dist[0].h = H_gM*0.95;
    dist[0].n = N/3.;
    dist[1].d = D_gM;
    dist[1].h = H_gM;
    dist[1].n = N/3.;
    dist[2].d = D_gM*1.05;
    dist[2].h = H_gM*1.05;
    dist[2].n = N/3.;

    *nres = 3;

    return ret;
}

int Fix_D_gM_for_JohnsonSB_pine ( double D_gM, double A, int *nres,
                   double *modelresult, char *errors, int errorCheckMode,
                   double allowedRiskLevel, double rectFactor)
{
    int ret;
    double upper_bound;
    char errorStr[200] = "";

    ret = 1;

    //Check for fatal errors
    if (A<=0)
    {
        ret = 0;
        constructErrorMessage (errorStr, "A<=0(A=",A);
    }
    if (ret == 0) {
        strcat(errors, errorStr);
        return 0;
    }
    // NB: this is here to counter negative upper_bounds!
    if (A<4) 
    { 
    	A=4; 
    }
    upper_bound = -17.5244+14.667*log(A);
    if (D_gM >= upper_bound) 
    { 
    	D_gM=upper_bound-.5; 
    }
    *nres = 1;
    *modelresult = rectFactor*D_gM;
    return ret;
}

static int construct_heightdist_with_diameters_using_Weibull_2_parameters ( double b, double c, double HEIGHT_CLASS_WIDTH,
                        double N, double D_gM, double H_gM, double H_dom, double h0, double h100,
                        int class_count, double class_n_min, struct heightclass_weibull_with_diameters *dist)
{

    int ret = 1;

    int i, j;
    double low_boundary, high_boundary, low_mass, high_mass, stem_number, left_out;
    double sum_ppa, sum_ppa_d, ppa, ppa_d, D_g, step;
    //
#ifdef __cplusplus
    double** dist_2 = NULL;
    dist_2 = new double*[class_count];
    for (int i=0; i < class_count; i++) {
        dist_2[i] = new double[3];
        for (int j=0; j<3; j++) {
            dist_2[i][j] = 0;
        }
    }
#else
    double dist_2[class_count][3];
#endif

    left_out = 0;
    //heights & stem numbers
    for ( i = 0; i < class_count; i++)  {
        if ( i != class_count - 1) {    // all but the last class that has different class length
            if ( i == 0) {    // first class
                dist_2[i][0] = h0 + HEIGHT_CLASS_WIDTH/2;
            }
            else {    // intermediate classes
                dist_2[i][0] = dist_2[i-1][0] + HEIGHT_CLASS_WIDTH;
            }
            // class boundaries for estimation of stem number
            low_boundary = dist_2[i][0] - HEIGHT_CLASS_WIDTH / 2;
            high_boundary = dist_2[i][0] + HEIGHT_CLASS_WIDTH / 2;

        }
        else {    // last class
            dist_2[i][0] = dist_2[i-1][0] + HEIGHT_CLASS_WIDTH/2 + (h100- (dist_2[i-1][0] + HEIGHT_CLASS_WIDTH/2)) / 2;
            low_boundary = dist_2[i-1][0] + HEIGHT_CLASS_WIDTH / 2;
            high_boundary = h100;
        }

        // Calculate diameter
        double h = dist_2[i][0];
        double epsilon = 0.0001;
        if(h<1.3-epsilon)
                dist_2[i][1] = 0;
            else
            dist_2[i][1] = exp(0.3904+0.9119*log(h-1)+0.0532*h-1.0845*log(D_gM+1)+0.9468*log(H_gM)-0.0311*H_dom+(pow(0.0219,2)+pow(0.0175,2)+pow(0.1789,2))/2);

        // cumulated probabilities at boundaries
        low_mass = 1 - exp(-pow(( low_boundary / b ), c ));
        high_mass = 1 - exp(-pow(( high_boundary / b ), c ));
            //stem number
        stem_number = N * ( high_mass - low_mass );
        //Calculate basal area mass of rejected classes
        if (stem_number < class_n_min)
             left_out = left_out + stem_number;
         dist_2[i][2] = stem_number;
    }
    //Calculate new stem numbers to the classes
    int class_count_new;
    double mass_scaler;
    int k = 0;
    //Assign the first class to biggest
    int biggest_class = 0;
    class_count_new = class_count;
    mass_scaler = N / (N - left_out);
    for ( i = 0; i < class_count; i++)
    {
        if (dist_2[i][2] > dist_2[biggest_class][2])
                 biggest_class = i;
        if (dist_2[i][2] < class_n_min)
             class_count_new = class_count_new - 1;
            else
            {
             dist[k].h = dist_2[i][0];
             dist[k].d = dist_2[i][1];
             dist[k].n = mass_scaler * dist_2[i][2];
             k++;
        }
    }
    //Simple distribution for the stratum with only few or none classes left
    if (class_count_new < 3)
    {
        dist[0].h = dist_2[biggest_class-1][0];
        dist[0].d = dist_2[biggest_class-1][1];
        dist[0].n = N/3.;
        dist[1].h = dist_2[biggest_class-1][0];
        dist[1].d = dist_2[biggest_class][1];
        dist[1].n = N/3.;
        dist[2].h = dist_2[biggest_class-1][0];
        dist[2].d = dist_2[biggest_class+1][1];
        dist[2].n = N/3.;
        class_count_new = 3;
    }

    //Scale tree diameters so that D_gM is the same as measured
    //First check if D_gM is too little:

    if(D_gM<0.6)
        D_gM=0.6;

    sum_ppa=0;
    sum_ppa_d=0;
    for ( i = 0; i < class_count; i++)  {
        ppa=dist[i].n*PI*pow((dist[i].d/2),2);
        ppa_d=dist[i].n*PI*pow((dist[i].d/2),2)*dist[i].d;
        sum_ppa=sum_ppa+ppa;
        sum_ppa_d=sum_ppa_d+ppa_d;
    }
    D_g = sum_ppa_d / sum_ppa;
    j = 1;
    step = 0.01;

    do {
        if (D_g<D_gM&&fabs(D_g-D_gM)>step)
            for ( i = 0; i < class_count; i++)  {
                if(dist[i].d!=0)
                    dist[i].d=dist[i].d+step;
            }
        if (D_g>D_gM&&fabs(D_g-D_gM)>step)
            for ( i = 0; i < class_count; i++)  {
                if(dist[i].d>step)
                    dist[i].d=dist[i].d-step;
            }
        sum_ppa=0;
        sum_ppa_d=0;
        for ( i = 0; i < class_count; i++)  {
            ppa = dist[i].n*PI*pow((dist[i].d/2),2);
            ppa_d = dist[i].n*PI*pow((dist[i].d/2),2)*dist[i].d;
            sum_ppa=sum_ppa+ppa;
            sum_ppa_d=sum_ppa_d+ppa_d;
        }
        D_g=sum_ppa_d/sum_ppa;
        j++;
    }
    while(fabs(D_g-D_gM)>step);

    //To avoid too little diameter values
    for ( i = 0; i < class_count; i++)  {
        if(dist[i].d>0&&dist[i].d<0.1) {
            dist[i].d=0.1;
        }
    }

#ifdef __cplusplus
    for (int i=0; i < class_count; i++) {
        delete[] dist_2[i];
        dist_2[i] = NULL;
    }
    delete[] dist_2;
    dist_2 = NULL;
#endif

    return class_count_new;

}

static void construct_heightdist_using_Weibull_2_parameters ( double b, double c, double HEIGHT_CLASS_WIDTH,
                        double N, double h0, double h100, int class_count, struct heightclass_weibull *dist)
{

    int ret = 1;

    int i;
    double low_boundary, high_boundary, low_mass, high_mass;
    double low_mass_left_out, high_mass_left_out, mass_scaler;

    //probability masses not included in distribution classes
    low_mass_left_out = 1 - exp(-pow(( h0 / b ), c ));
    high_mass_left_out = 1 - (1 - exp(-pow(( h100 / b ), c )));
    mass_scaler = 1 / (1 - (low_mass_left_out + high_mass_left_out));

    //heights & stem numbers
    for ( i = 0; i < class_count; i++)  {
        if ( i != class_count - 1) {    // all but the last class that has different class length
            if ( i == 0) {    // first class
                dist[i].h = h0 + HEIGHT_CLASS_WIDTH/2;
            }
            else {    // intermediate classes
                dist[i].h = dist[i-1].h + HEIGHT_CLASS_WIDTH;
            }
            // class boundaries for estimation of stem number
            low_boundary = dist[i].h - HEIGHT_CLASS_WIDTH / 2;
            high_boundary = dist[i].h + HEIGHT_CLASS_WIDTH / 2;

        }
        else {    // last class
            dist[i].h = dist[i-1].h + HEIGHT_CLASS_WIDTH/2 + (h100- (dist[i-1].h + HEIGHT_CLASS_WIDTH/2)) / 2;
            low_boundary = dist[i-1].h + HEIGHT_CLASS_WIDTH / 2;
            high_boundary = h100;
        }

        // cumulated probabilities at boundaries
        low_mass = 1 - exp(-pow(( low_boundary / b ), c ));
        high_mass = 1 - exp(-pow(( high_boundary / b ), c ));

        //stem number
        dist[i].n = N * (( high_mass - low_mass ) * mass_scaler);
    }
}

int Height_distribution_young_pine_with_diameters_Weibull_Siipilehto (double H_gM, double H_dom, double N, double D_gM,
                        double HEIGHT_CLASS_WIDTH, double CLASS_N_MIN,
                        double max_class_count, int *nres, struct heightclass_weibull_with_diameters *dist, char *errors,
                        int errorCheckMode, double allowedRiskLevel, double rectFactor) {
    int ret = 1;
    int rounds;


    //Check for fatal errors
    char errorStr[100] = "";
    if (H_gM<=0)
    {
        ret = 0;
        constructErrorMessage (errorStr, "H_gM<=0(H_gM=",H_gM);
    }
    if (N<=0)
    {
        ret = 0;
        constructErrorMessage (errorStr, "N<=0(N=",N);
    }
    if (ret == 0) {
        strcat(errors, errorStr);
        return ret;
    }

    // Weibull distribution cumulative distribution function parameters
    double b = exp(0.1556+0.997*log(H_gM)-0.0385/log(H_dom/H_gM+0.4));
    double c = exp(-2.3205+0.0744*H_gM-0.0469*H_dom+0.2296*log(N)+1.2869/log(H_dom/H_gM+0.4));
    double h0 = b * pow( -log( 1.0 - 0.000000000000001 ), (1 / c));
    double h100 = b * pow( -log( 1.0 - 0.999999999999999 ), (1 / c)); // HUOM! Weibull-jakauman prosenttipistejakauman yht�l�?

    // how many height classes (??????)
    int class_count = (int) floor((h100-h0)/HEIGHT_CLASS_WIDTH) + 1;

    rounds = 0;
    while (class_count < 14 && rounds <= MAX_ROUNDS) {    //If too few classes
        HEIGHT_CLASS_WIDTH = HEIGHT_CLASS_WIDTH / 2;
        class_count = (int) floor((h100 - h0) / HEIGHT_CLASS_WIDTH);
    }
    if (class_count < 14)
    {
        ret = 0;
        constructErrorMessage (errorStr, "Too few generated height classes! Check D_gM and BA!", 14);
    }
    if (ret == 0) {
        strcat(errors, errorStr);
        return 0;
    }

    rounds = 0;
    while (class_count > 40 && rounds <= MAX_ROUNDS) {    //If too many classes
        HEIGHT_CLASS_WIDTH = HEIGHT_CLASS_WIDTH * 2;
        class_count = (int) floor((h100 - h0) / HEIGHT_CLASS_WIDTH);
    }
    if (class_count > 40)
    {
        ret = 0;
        constructErrorMessage (errorStr, "Too many generated height classes! Check D_gM and BA!", 40);
    }
    if (class_count > max_class_count)
    {
        ret = 0;
        constructErrorMessage (errorStr, "Too many generated height classes! Check D_gM and BA! (MAX_CLASS_COUNT=",
                                          max_class_count);
    }
    if (ret == 0) {
        strcat(errors, errorStr);
        return 0;
    }

    // height distribution vector, elements contain height and number of stems / ha in the class
    class_count = construct_heightdist_with_diameters_using_Weibull_2_parameters( b, c, HEIGHT_CLASS_WIDTH, N, D_gM, H_gM, H_dom, h0, h100, class_count, CLASS_N_MIN, dist );

    *nres = class_count;

    return ret;
}

int Height_distribution_young_pine_Weibull_Siipilehto (double H_gM, double H_dom, double N,
                        double HEIGHT_CLASS_WIDTH,
                        double max_class_count, int *nres, struct heightclass_weibull *dist, char *errors,
                        int errorCheckMode, double allowedRiskLevel, double rectFactor) {
    int ret = 1;
    int rounds;


    //Check for fatal errors
    char errorStr[100] = "";
    if (H_gM<=0)
    {
        ret = 0;
        constructErrorMessage (errorStr, "H_gM<=0(H_gM=",H_gM);
    }
    if (N<=0)
    {
        ret = 0;
        constructErrorMessage (errorStr, "N<=0(N=",N);
    }
    if (ret == 0) {
        strcat(errors, errorStr);
        return ret;
    }

    // Weibull distribution cumulative distribution function parameters
    double b = exp(0.1556+0.997*log(H_gM)-0.0385/log(H_dom/H_gM+0.4));
    double c = exp(-2.3205+0.0744*H_gM-0.0469*H_dom+0.2296*log(N)+1.2869/log(H_dom/H_gM+0.4));
    double h0 = b * pow( -log( 1.0 - 0.000000000000001 ), (1 / c));
    double h100 = b * pow( -log( 1.0 - 0.999999999999999 ), (1 / c)); // HUOM! Weibull-jakauman prosenttipistejakauman yht�l�?

    // how many height classes (??????)
    int class_count = (int) floor((h100-h0)/HEIGHT_CLASS_WIDTH) + 1;

        rounds = 0;
    while (class_count < 14 && rounds <= MAX_ROUNDS) {    //If too few classes
        HEIGHT_CLASS_WIDTH = HEIGHT_CLASS_WIDTH / 2;
        class_count = (int) floor((h100 - h0) / HEIGHT_CLASS_WIDTH);
    }
    if (class_count < 14)
    {
        ret = 0;
        constructErrorMessage (errorStr, "Too few generated height classes! Check D_gM and BA!", 14);
    }
    if (ret == 0) {
        strcat(errors, errorStr);
        return 0;
    }

    rounds = 0;
    while (class_count > 40 && rounds <= MAX_ROUNDS) {    //If too many classes
        HEIGHT_CLASS_WIDTH = HEIGHT_CLASS_WIDTH * 2;
        class_count = (int) floor((h100 - h0) / HEIGHT_CLASS_WIDTH);
    }
    if (class_count > 40)
    {
        ret = 0;
        constructErrorMessage (errorStr, "Too many generated height classes! Check D_gM and BA!", 40);
    }
    if (class_count > max_class_count)
    {
        ret = 0;
        constructErrorMessage (errorStr, "Too many generated height classes! Check D_gM and BA! (MAX_CLASS_COUNT=",
                                          max_class_count);
    }
        if (ret == 0) {
        strcat(errors, errorStr);
        return 0;
    }

    *nres = class_count;

    // height distribution vector, elements contain height and number of stems / ha in the class
    construct_heightdist_using_Weibull_2_parameters( b, c, HEIGHT_CLASS_WIDTH, N, h0, h100, class_count, dist );

    return ret;
}

static void construct_heightdist_using_Weibull_3_parameters ( double a, double b, double c, double HEIGHT_CLASS_WIDTH, double N,
                        double h0, double h100, int class_count, struct heightclass_weibull *dist)
{

    int ret = 1;

    int i;
    double low_boundary, high_boundary, low_mass, high_mass;
    double low_mass_left_out, high_mass_left_out, mass_scaler;

    //probability masses not included in distribution classes
    low_mass_left_out = 1 - exp(-pow((( h0 - a ) / b ), c ));
    high_mass_left_out = 1 - (1 - exp(-pow((( h100 - a ) / b ), c )));
    mass_scaler = 1 / (1 - (low_mass_left_out + high_mass_left_out));

    //heights & stem numbers
    for ( i = 0; i < class_count; i++)  {
        if ( i != class_count - 1) {    // all but the last class that has different class length
            if ( i == 0) {    // first class
                dist[i].h = h0 + HEIGHT_CLASS_WIDTH/2;
            }
            else {    // intermediate classes
                dist[i].h = dist[i-1].h + HEIGHT_CLASS_WIDTH;
            }
            // class boundaries for estimation of stem number
            low_boundary = dist[i].h - HEIGHT_CLASS_WIDTH / 2;
            high_boundary = dist[i].h + HEIGHT_CLASS_WIDTH / 2;

        }
        else {    // last class
            dist[i].h = dist[i-1].h + HEIGHT_CLASS_WIDTH/2 + (h100- (dist[i-1].h + HEIGHT_CLASS_WIDTH/2)) / 2;
            low_boundary = dist[i-1].h + HEIGHT_CLASS_WIDTH / 2;
            high_boundary = h100;
        }

        // cumulated probabilities at boundaries
        low_mass = 1 - exp(-pow((( low_boundary - a ) / b ), c ));
        high_mass = 1 - exp(-pow((( high_boundary - a ) / b ), c ));
        //stem number
        dist[i].n = N * (( high_mass - low_mass ) * mass_scaler);
    }
}



int Height_distribution_young_plantation_spruce_Valkonen ( double N, double H_ari, double HEIGHT_CLASS_WIDTH,
                        double max_class_count, int *nres, struct heightclass_weibull *dist, char *errors,
                        int errorCheckMode, double allowedRiskLevel, double rectFactor) {
    int ret = 1;


    //Check for fatal errors
    char errorStr[100] = "";
    if (N<=0)
    {
        ret = 0;
        constructErrorMessage (errorStr, "N<=0(N=",N);
    }
    if (H_ari<=0)
    {
        ret = 0;
        constructErrorMessage (errorStr, "H_ari<=0(H_ari=",H_ari);
    }
    if (ret == 0) {
        strcat(errors, errorStr);
        return 0;
    }

    // Weibull distribution cumulative distribution function parameters
    double a = 0.9279*H_ari/log(N);

    if ((H_ari-a)<=0)
    {
        ret = 0;
        constructErrorMessage (errorStr, "(H_ari-a)<=0((H_ari-a)=",(H_ari-a));
    }

    double b = exp(1.0438*log(H_ari-a));
    double c = 1.883+0.127*H_ari;
    double h0;
    if ( a != -1 ) {
        h0 = a;
    }
    else {
        h0 = b * pow( -log( 0 + 0.000000000000001 ), (1 / c));
    }
    double h100 = a + b * pow( -log( 1.0 - 0.999999999999999 ), (1 / c)); // HUOM! Weibull-jakauman prosenttipistejakauman yht�l�?

    // how many height classes (??????)
    int class_count = (int) floor((h100-h0)/HEIGHT_CLASS_WIDTH) + 1;
    if (class_count > max_class_count)
    {
        ret = 0;
        constructErrorMessage (errorStr, "Too many generated height classes! Check D_gM and BA! (MAX_CLASS_COUNT=",max_class_count);
    }
    //Check for fatal errors
    if (b<=0)
    {
        ret = 0;
        constructErrorMessage (errorStr, "b<=0(b=",b);
    }
    if (ret == 0) {
        strcat(errors, errorStr);
        return 0;
    }

    // height distribution vector, elements contain height and number of stems / ha in the class
    construct_heightdist_using_Weibull_3_parameters( a, b, c, HEIGHT_CLASS_WIDTH, N, h0, h100, class_count, dist );

    *nres = class_count;

    return ret;

    //Skaalaus siten ett� mitattu H_ari toteutuu: kerroin=mitattu H_ari/jakaumasta laskettu H_ari. Siis jokaista pituutta
    //korjataan kertoimella!
}



int Height_distribution_understorey_spruce_Valkonen (double N, double H_ari, double HEIGHT_CLASS_WIDTH, double MAX_CLASS_COUNT,
                        int *nres, struct heightclass_weibull *dist, char *errors,
                        int errorCheckMode, double allowedRiskLevel, double rectFactor) {
    int ret = 1;


    //Check for fatal errors
    char errorStr[100] = "";
    if (H_ari<=0)
    {
        ret = 0;
        constructErrorMessage (errorStr, "H_ari<=0(H_ari=",H_ari);
    }
    if (ret == 0) {
        strcat(errors, errorStr);
        return 0;
    }

    // Weibull distribution cumulative distribution function parameters
    double a = 0.1251*H_ari;
    double b = 0.9412*H_ari;
    double c = 1.226;
    double h0;
    if ( a != -1 ) {
        h0 = a;
    }
    else {
        h0 = b * pow( -log( 0 + 0.000000000000001 ), (1 / c));
    }
    double h100 = a + b * pow( -log( 1.0 - 0.999999999999999 ), (1 / c)); // HUOM! Weibull-jakauman prosenttipistejakauman yht�l�?

    // how many height classes (??????)
    int class_count = (int) floor((h100-h0)/HEIGHT_CLASS_WIDTH) + 1;
    if (class_count > MAX_CLASS_COUNT)
    {
        ret = 0;
        constructErrorMessage (errorStr, "Too many generated height classes! Check D_gM and BA! (MAX_CLASS_COUNT=",MAX_CLASS_COUNT);
    }
        if (ret == 0) {
        strcat(errors, errorStr);
        return 0;
    }

    // height distribution vector, elements contain height and number of stems / ha in the class
    construct_heightdist_using_Weibull_3_parameters( a, b, c, HEIGHT_CLASS_WIDTH, N, h0, h100, class_count, dist );

    *nres = class_count;

    return ret;
}

int Diameter_distribution_JKK (double BA, double BA_total, double D_gM, double H_gM, double Age,
                        double A, double N, double SP, double N_measured, double PEAT, double SC,
                        double DIAM_CLASS_WIDTH, double MAX_CLASS_COUNT,
                        double CLASS_N_MIN, double D_MIN,
                        int *nres, struct diamclass *dist, char *errors,
                        int errorCheckMode, double allowedRiskLevel, double rectFactor)
{
    int ret = 1;
    //Generate diameter distribution
    //When number of stems is measured
    if (N>0&&N_measured==1)
    {
        //Mineral soil
        if (PEAT==0)
        {
            //Scots pine and other conifers
            if (SP==1||SP==8)
            {
                ret = Diameter_distribution_pine_percent_point_KangasMaltamo(BA, D_gM, Age, N, DIAM_CLASS_WIDTH, MAX_CLASS_COUNT, CLASS_N_MIN,
                        nres, dist, errors, errorCheckMode, allowedRiskLevel, 1);
            }
            //Norway spruce
            else if (SP==2)
            {
                ret = Diameter_distribution_spruce_percent_point_KangasMaltamo(BA, BA_total, D_gM, Age, N, DIAM_CLASS_WIDTH, SC, MAX_CLASS_COUNT, CLASS_N_MIN,
                        nres, dist, errors, errorCheckMode, allowedRiskLevel, 1);
            }
            //Birches and other deciduous species
            else
            {
                ret = Diameter_distribution_birch_percent_point_KangasMaltamo(BA, D_gM, Age, N, DIAM_CLASS_WIDTH, SC, MAX_CLASS_COUNT, CLASS_N_MIN,
                        nres, dist, errors, errorCheckMode, allowedRiskLevel, 1);
            }

            if (ret == 0)
                return ret;

        }
        //Peatland (PEAT==1)
        else
        {
            //Scots pine and other conifers
            if (SP==1||SP==8)
            {
                ret = Diameter_distribution_pine_with_stem_number_JohnsonSB_Siipilehto(BA, D_gM, N, DIAM_CLASS_WIDTH, MAX_CLASS_COUNT, CLASS_N_MIN, D_MIN,
                        nres, dist, errors, errorCheckMode, allowedRiskLevel, 1);
            }
            //Norway spruce
            else if (SP==2)
            {
                ret = Diameter_distribution_spruce_with_stem_number_JohnsonSB_Siipilehto(BA, D_gM, N, DIAM_CLASS_WIDTH, MAX_CLASS_COUNT, CLASS_N_MIN, D_MIN,
                        nres, dist, errors, errorCheckMode, allowedRiskLevel, 1);
            }
            //Birches and other deciduous species
            else
            {
                ret = Diameter_distribution_birch_with_stem_number_JohnsonSB_Siipilehto(BA, D_gM, N, DIAM_CLASS_WIDTH, MAX_CLASS_COUNT, CLASS_N_MIN, D_MIN,
                        nres, dist, errors, errorCheckMode, allowedRiskLevel, 1);
            }

            if (ret == 0)
                return ret;

        }
    //When number of stems is not measured
    }
    else if (N_measured==0)
    {
        if (D_gM<13||D_gM>45)
        {
            //Scots pine and other conifers
            if (SP==1||SP==8)
            {
                Diameter_distribution_pine_Weibull_Mykkanen(BA, D_gM, DIAM_CLASS_WIDTH, MAX_CLASS_COUNT, CLASS_N_MIN,
                        nres, dist, errors, errorCheckMode, allowedRiskLevel, 1);
            }
            //Norway spruce
            else if (SP==2)
            {
                ret = Diameter_distribution_spruce_Weibull_Kilkki(BA, D_gM, DIAM_CLASS_WIDTH, MAX_CLASS_COUNT, CLASS_N_MIN,
                        nres, dist, errors, errorCheckMode, allowedRiskLevel, 1);
            }
            //Birches and other deciduous species
            else
            {
                if (BA>20||(BA>15&&Age<30))
                {
                    Diameter_distribution_pine_Weibull_Mykkanen(BA, D_gM, DIAM_CLASS_WIDTH, MAX_CLASS_COUNT, CLASS_N_MIN,
                                nres, dist, errors, errorCheckMode, allowedRiskLevel, 1);
                }
                else
                {
                        double *Age_at_breast_height = (double *) malloc(1*sizeof(double));
                    if (A==0)
                    {
                        ret = Fix_age_at_breast_height_JKK(SC, D_gM, Age, nres, Age_at_breast_height, errors, errorCheckMode, allowedRiskLevel, 1);
                        A = *Age_at_breast_height;
                        if (ret == 0)
                            return ret;
                            free(Age_at_breast_height);
                    }
                    ret = Diameter_distribution_birch_Weibull_Siipilehto(BA, D_gM, A, DIAM_CLASS_WIDTH, MAX_CLASS_COUNT, CLASS_N_MIN,
                                nres, dist, errors, errorCheckMode, allowedRiskLevel, 1);
                }
            }

            if (ret == 0)
                return ret;

        }
        //D_gM>=13&&D_gM<45, problems with too big trees
        else
        {
            //Scots pine and other conifers
            if (SP==1||SP==8)
            {
                        double *Age_at_breast_height = (double *) malloc(1*sizeof(double));
                if (A<=10)
                {
                    ret = Fix_age_at_breast_height_JKK(SC, D_gM, Age, nres, Age_at_breast_height, errors, errorCheckMode, allowedRiskLevel, 1);
                    A = *Age_at_breast_height;

                    if (ret == 0)
                        return ret;
                                free(Age_at_breast_height);
                }
                ret = Diameter_distribution_pine_JohnsonSB_Siipilehto(D_gM, A, H_gM, BA, DIAM_CLASS_WIDTH, MAX_CLASS_COUNT, CLASS_N_MIN, D_MIN,
                        nres, dist, errors, errorCheckMode, allowedRiskLevel, 1);
            }
            else
            {
                    double *Age_at_breast_height = (double *) malloc(1*sizeof(double));
                if (A==0)
                {
                    ret = Fix_age_at_breast_height_JKK(SC, D_gM, Age, nres, Age_at_breast_height, errors, errorCheckMode, allowedRiskLevel, 1);
                    A = *Age_at_breast_height;

                    if (ret == 0)
                        return ret;
                            free(Age_at_breast_height);
                }
                        //Norway spruce
                        if (SP==2)
                {
                        ret = Diameter_distribution_spruce_JohnsonSB_Siipilehto(D_gM, A, H_gM, BA, DIAM_CLASS_WIDTH, MAX_CLASS_COUNT, CLASS_N_MIN, D_MIN,
                            nres, dist, errors, errorCheckMode, allowedRiskLevel, 1);
                }
                    //Birches and other deciduous species
                        else
                    {
                        ret = Diameter_distribution_birch_JohnsonSB_Siipilehto(D_gM, A, BA, DIAM_CLASS_WIDTH, MAX_CLASS_COUNT, CLASS_N_MIN, D_MIN,
                        nres, dist, errors, errorCheckMode, allowedRiskLevel, 1);
                    }
                        }
        }
    }

    return ret;

}

int Diameter_distribution_only_N_measured_JKK (double BA, double D_gM, double Age,
                        double A, double N, double SP, double N_measured, double SC,
                        double DIAM_CLASS_WIDTH, double MAX_CLASS_COUNT,
                        double CLASS_N_MIN,
                        int *nres, struct diamclass *dist, char *errors,
                        int errorCheckMode, double allowedRiskLevel, double rectFactor)
{
    int ret = 1;
    //Generate diameter distribution
    //Scale N after distribution generation to be same as before
    //Scots pine and other conifers
    if (SP==1||SP==8)
    {
        Diameter_distribution_pine_Weibull_only_N_measured_Mykkanen(BA, D_gM, N, DIAM_CLASS_WIDTH, MAX_CLASS_COUNT, CLASS_N_MIN,
                nres, dist, errors, errorCheckMode, allowedRiskLevel, 1);
    }
    //Norway spruce
    else if (SP==2)
    {
        ret = Diameter_distribution_spruce_Weibull_only_N_measured_Kilkki(BA, D_gM, N, DIAM_CLASS_WIDTH, MAX_CLASS_COUNT, CLASS_N_MIN,
                nres, dist, errors, errorCheckMode, allowedRiskLevel, 1);
    }
    //Birches and other deciduous species
    else
    {
        if (BA>20||(BA>15&&Age<30))
        {
            Diameter_distribution_pine_Weibull_only_N_measured_Mykkanen(BA, D_gM, N, DIAM_CLASS_WIDTH, MAX_CLASS_COUNT, CLASS_N_MIN,
                    nres, dist, errors, errorCheckMode, allowedRiskLevel, 1);
        }
        else
        {
        double *Age_at_breast_height = (double *) malloc(1*sizeof(double));
        if (A==0)
        {
            ret = Fix_age_at_breast_height_JKK(SC, D_gM, Age, nres, Age_at_breast_height, errors, errorCheckMode, allowedRiskLevel, 1);
            A = *Age_at_breast_height;
            if (ret == 0)
                return ret;
            free(Age_at_breast_height);
        }
        ret = Diameter_distribution_birch_Weibull_only_N_measured_Siipilehto(BA, D_gM, A, N, DIAM_CLASS_WIDTH, MAX_CLASS_COUNT, CLASS_N_MIN,
                    nres, dist, errors, errorCheckMode, allowedRiskLevel, 1);
    }
    }

    if (ret == 0)
        return ret;
    return ret;

}

int Diameter_distribution_weibull_JKK (double BA, double D_gM, double Age,
                        double A, double SP, double SC,
                        double DIAM_CLASS_WIDTH, double MAX_CLASS_COUNT,
                        double CLASS_N_MIN,
                        int *nres, struct diamclass *dist, char *errors,
                        int errorCheckMode, double allowedRiskLevel, double rectFactor)
{
    int ret = 1;
    //Generate diameter distribution
    //Scots pine and other conifers
    if (SP==1||SP==8)
    {
        Diameter_distribution_pine_Weibull_Mykkanen(BA, D_gM, DIAM_CLASS_WIDTH, MAX_CLASS_COUNT, CLASS_N_MIN,
                    nres, dist, errors, errorCheckMode, allowedRiskLevel, 1);
    }
    //Norway spruce
    else if (SP==2)
    {
        ret = Diameter_distribution_spruce_Weibull_Kilkki(BA, D_gM, DIAM_CLASS_WIDTH, MAX_CLASS_COUNT, CLASS_N_MIN,
                        nres, dist, errors, errorCheckMode, allowedRiskLevel, 1);
    }
    //Birches and other deciduous species
    else
    {
        if (BA>20||(BA>15&&Age<30))
        {
            Diameter_distribution_pine_Weibull_Mykkanen(BA, D_gM, DIAM_CLASS_WIDTH, MAX_CLASS_COUNT, CLASS_N_MIN,
                        nres, dist, errors, errorCheckMode, allowedRiskLevel, 1);
        }
        else
        {
                double *Age_at_breast_height = (double *) malloc(1*sizeof(double));
            if (A==0)
            {
                ret = Fix_age_at_breast_height_JKK(SC, D_gM, Age, nres, Age_at_breast_height, errors, errorCheckMode, allowedRiskLevel, 1);
                A = *Age_at_breast_height;
                if (ret == 0)
                    return ret;
                    free(Age_at_breast_height);
            }
            ret = Diameter_distribution_birch_Weibull_Siipilehto(BA, D_gM, A, DIAM_CLASS_WIDTH, MAX_CLASS_COUNT, CLASS_N_MIN,
                        nres, dist, errors, errorCheckMode, allowedRiskLevel, 1);
        }
    }

    if (ret == 0)
        return ret;

    return ret;

}

int Diameter_distribution_weibull_integer_class_JKK (double BA, double D_gM, double Age,
                        double A, double SP, double SC,
                        double DIAM_CLASS_WIDTH, double MAX_CLASS_COUNT,
                        double CLASS_N_MIN,
                        int *nres, struct diamclass *dist, char *errors,
                        int errorCheckMode, double allowedRiskLevel, double rectFactor)
{
    int ret = 1;
    //Generate diameter distribution
    //Scots pine and other conifers
    if (SP==1||SP==8)
    {
        Diameter_distribution_pine_Weibull_integer_class_Mykkanen(BA, D_gM, DIAM_CLASS_WIDTH, MAX_CLASS_COUNT, CLASS_N_MIN,
                    nres, dist, errors, errorCheckMode, allowedRiskLevel, 1);
    }
    //Norway spruce
    else if (SP==2)
    {
        ret = Diameter_distribution_spruce_Weibull_integer_class_Kilkki(BA, D_gM, DIAM_CLASS_WIDTH, MAX_CLASS_COUNT, CLASS_N_MIN,
                        nres, dist, errors, errorCheckMode, allowedRiskLevel, 1);
    }
    //Birches and other deciduous species
    else
    {
        if (BA>20||(BA>15&&Age<30))
        {
            Diameter_distribution_pine_Weibull_integer_class_Mykkanen(BA, D_gM, DIAM_CLASS_WIDTH, MAX_CLASS_COUNT, CLASS_N_MIN,
                        nres, dist, errors, errorCheckMode, allowedRiskLevel, 1);
        }
        else
        {
        double *Age_at_breast_height = (double *) malloc(1*sizeof(double));
        if (A==0)
        {
            ret = Fix_age_at_breast_height_JKK(SC, D_gM, Age, nres, Age_at_breast_height, errors, errorCheckMode, allowedRiskLevel, 1);
            A = *Age_at_breast_height;
            if (ret == 0)
                return ret;
            free(Age_at_breast_height);
        }
        ret = Diameter_distribution_birch_Weibull_integer_class_Siipilehto(BA, D_gM, A, DIAM_CLASS_WIDTH, MAX_CLASS_COUNT, CLASS_N_MIN,
                    nres, dist, errors, errorCheckMode, allowedRiskLevel, 1);
        }
    }

    if (ret == 0)
        return ret;

    return ret;

}

int Diameter_distribution_spruce_Weibull_MELA_Kilkki ( double BA, double D_gM, double CLASS_N_MIN,
                        int *nres, struct diamclass *dist, char *errors,
                        int errorCheckMode, double allowedRiskLevel, double rectFactor)
{
    int ret, class_count;
    double a, b, c, d0, d100;
    char errorStr[200] = "";

    ret = 1;
    //Check for fatal errors
    if (D_gM<=0)
    {
        ret = 0;
        constructErrorMessage (errorStr, "D_gM<=0(D_gM=",D_gM);
    }
    if (BA<=0)
    {
        ret = 0;
        constructErrorMessage (errorStr, "BA<=0(BA=",BA);
    }
    if (ret == 0) {
        strcat(errors, errorStr);
        return 0;
    }

    // Weibull distribution cumulative distribution function parameters
    a = 0.001389+0.517444*D_gM;
    b = exp(-0.346223+0.934993*log(D_gM)-0.000925*BA);

    //Check for fatal errors
    if (b<=0)
    {
        ret = 0;
        constructErrorMessage (errorStr, "b<=0(b=",b);
    }
    if ((log((D_gM-a)/b))==0)
    {
        ret = 0;
        constructErrorMessage3val (errorStr, "(log((D_gM-a)/b))<=0(D_gM=", D_gM, ",a=",a, ",b=",b);
    }
    if (ret == 0) {
        strcat(errors, errorStr);
        return 0;
    }


    c = (log(-log(0.5)))/(log((D_gM-a)/b));
    d0 = a;   // low boundary
    //The right tail of distribution is very long with almost zero tree number.
    //MELA right tail truncated using 0.99
    d100 = a + b * pow( -log( 1.0 - 0.999999999999999 ), (1 / c)); // HUOM! Weibull-jakauman prosenttipistejakauman yht‰lˆ?

    // 10 diameter classes for MELA distribution
    class_count = 10;
    //Class width
    double DIAM_CLASS_WIDTH = (d100-d0)/10;
    //Minimum number of trees in one class
    CLASS_N_MIN = 0;

        if (ret == 0) {
        strcat(errors, errorStr);
        return 0;
        }

    // diameter distribution vector, elements contain diameter at breast height and number of stems / ha in the class
    class_count = construct_diamdist_using_Weibull( a, b, c, DIAM_CLASS_WIDTH, BA, d0, d100, class_count, CLASS_N_MIN, dist );

    *nres = class_count;

    return ret;
}

int Diameter_distribution_pine_Weibull_MELA_Mykkanen ( double BA, double D_gM, double CLASS_N_MIN,
                            int *nres, struct diamclass *dist, char *errors,
                            int errorCheckMode, double allowedRiskLevel, double rectFactor)
{
    int ret, class_count;
    double a, b, c, d0, d100;
    char errorStr[200] = "";

    ret = 1;
    //Check for fatal errors
    if (D_gM<=0)
    {
        ret = 0;
        constructErrorMessage (errorStr, "D_gM<=0(D_gM=",D_gM);
    }
    if (BA<=0)
    {
        ret = 0;
        constructErrorMessage (errorStr, "BA<=0(BA=",BA);
    }
    if (ret == 0) {
        strcat(errors, errorStr);
        return 0;
    }

    // Weibull distribution cumulative distribution function parameters
    a = exp(-1.306454+1.154433*log(D_gM));
    c = exp(0.647888+0.025530*D_gM-0.005558*BA);
    b = (D_gM-a)/(pow((-log(0.5)),(1/c)));

    // how many diameter classes (??????)
    // distribution boundaries
    d0 = a;   // low boundary
    //MELA right tail truncated using 0.99
    d100 = a + b * pow( -log( 1.0 - 0.999999999999999 ), (1 / c)); // HUOM! Weibull-jakauman prosenttipistejakauman yht‰lˆ?

    // 10 diameter classes for MELA distribution
    class_count = 10;
    //Class width
    double DIAM_CLASS_WIDTH = (d100-d0)/10;
    //Minimum number of trees in one class
    CLASS_N_MIN = 0;
    
    //Check for fatal errors
    if (b<=0)
    {
        ret = 0;
        constructErrorMessage (errorStr, "b<=0(b=",b);
    }
    if (ret == 0) {
        strcat(errors, errorStr);
        return 0;
    }

        // diameter distribution vector, elements contain diameter at breast height and number of stems / ha in the class
    class_count = construct_diamdist_using_Weibull( a, b, c, DIAM_CLASS_WIDTH, BA, d0, d100, class_count, CLASS_N_MIN, dist );

    *nres = class_count;

    return ret;
}

int Diameter_distribution_MELA_JKK (double BA, double D_gM, double SP, double CLASS_N_MIN,
                        int *nres, struct diamclass *dist, char *errors,
                        int errorCheckMode, double allowedRiskLevel, double rectFactor)
{
    int ret = 1;
    //Generate diameter distribution
    //Norway spruce
    if (SP==2)
    {
           ret = Diameter_distribution_spruce_Weibull_MELA_Kilkki(BA, D_gM, CLASS_N_MIN,
                            nres, dist, errors, errorCheckMode, allowedRiskLevel, 1);
    }
    //Scots pine and other conifers and all deciduos
    else
    {
        ret = Diameter_distribution_pine_Weibull_MELA_Mykkanen(BA, D_gM, CLASS_N_MIN,
                    nres, dist, errors, errorCheckMode, allowedRiskLevel, 1);
    }

    return ret;

}

int Height_distribution_young_pine_Siipilehto (double N, double H_gM, double H_dom, double HEIGHT_CLASS_WIDTH,
                        double max_class_count, int *nres, struct heightclass_weibull *dist, char *errors,
                        int errorCheckMode, double allowedRiskLevel, double rectFactor) {
    int ret = 1;


    //Check for fatal errors
    char errorStr[100] = "";
    if (N<=0)
    {
        ret = 0;
        constructErrorMessage (errorStr, "N<=0(N=",N);
    }
    if (H_gM<=0)
    {
        ret = 0;
        constructErrorMessage (errorStr, "H_gM<=0(H_gM=",H_gM);
    }
    if (H_dom<=0)
    {
        ret = 0;
        constructErrorMessage (errorStr, "H_dom<=0(H_dom=",H_dom);
    }
    if (ret == 0) {
        strcat(errors, errorStr);
        return 0;
    }

    // Weibull distribution cumulative distribution function parameters

    double b = exp(0.1942+0.9971*log(H_gM)-0.058/log(H_dom/H_gM+0.4));
    double c = exp(-2.4203+0.0895*H_gM-0.0637*H_dom+0.251*log(N)+1.2707/log(H_dom/H_gM+0.4));
    double h0 = b * pow( -log( 0 + 0.000000000000001 ), (1 / c));
    double h100 = b * pow( -log( 1.0 - 0.999999999999999 ), (1 / c)); // HUOM! Weibull-jakauman prosenttipistejakauman yht�l�?

    // how many height classes (??????)
    int class_count = (int) floor((h100-h0)/HEIGHT_CLASS_WIDTH) + 1;
    if (class_count > max_class_count)
    {
        ret = 0;
        constructErrorMessage (errorStr, "Too many generated height classes! Check D_gM and BA! (MAX_CLASS_COUNT=",max_class_count);
    }
        if (ret == 0) {
        strcat(errors, errorStr);
        return 0;
    }

    // height distribution vector, elements contain height and number of stems / ha in the class
    //täällä pitäisi lisäksi laskea uusi arvo parametrille b ja sen jälkeen pyöräyttää jakauma vielä uusiksi
    //ja ehkä myös lpm:t puille
    construct_heightdist_using_Weibull_2_parameters( b, c, HEIGHT_CLASS_WIDTH, N, h0, h100, class_count, dist );

    *nres = class_count;

    return ret;
}

int Diameter_distribution_little_ba_and_n_JKK (double N, double D_gM, double H_gM, double BA, double BA_measured, double MAX_CLASS_COUNT,
                                        int *nres, struct diamclass_seedtree *dist, char *errors,
                                        int errorCheckMode, double allowedRiskLevel, double rectFactor)
{
    int ret = 1;
#ifdef __cplusplus
    double** dist2 = NULL;
    dist2 = new double*[3];
    for (int i=0; i<3; i++) {
        dist2[i] = new double[3];
        for (int j=0; j<3; j++) {
            dist2[i][j] = 0;
        }
    }
#else
    double dist2[3][3];
#endif

    dist2[0][0] = D_gM*0.95;
    dist2[0][1] = H_gM*0.95;
    dist2[0][2] = N/3.;
    dist2[1][0] = D_gM;
    dist2[1][1] = H_gM;
    dist2[1][2] = N/3.;
    dist2[2][0] = D_gM*1.05;
    dist2[2][1] = H_gM*1.05;
    dist2[2][2] = N/3.;

    //If basal area measured
    //Calculate basal area and scale stem number for sample trees
    if (BA_measured == 1)
    {
        double ba_sum=0;
        double ba;
        int i;
        for (i = 0; i < 3; i++)
        {
                    ba = PI * pow(( dist2[i][0] / 200 ), 2) * dist2[i][2];
             ba_sum = ba+ba_sum;
        }
        double scale = BA/ba_sum;
        N = N*scale;
    }

    dist[0].d = dist2[0][0];
    dist[0].h = dist2[0][1];
    dist[0].n = N/3.;
    dist[1].d = dist2[1][0];
    dist[1].h = dist2[1][1];
    dist[1].n = N/3.;
    dist[2].d = dist2[2][0];
    dist[2].h = dist2[2][1];
    dist[2].n = N/3.;

    *nres = 3;

#ifdef __cplusplus
    for (int i=0; i<3; i++) {
        delete[] dist2[i];
        dist2[i] = NULL;
    }
    delete[] dist2;
    dist2 = NULL;
#endif

    return ret;
}
