#!/usr/bin/env python
# encoding: utf-8
"""
Title: Testbench for SIMO prediction model implementations in Python

*model_testbench.py*

Each model definition must be of the form:
    def model_name(arg, oind)
in which
    arg - contains all the data being passed between the simulator
          and the prediction model
    oind - is the index number for the object for which the prediction is
           being made

The arg data object has the following structure:
    arg.variables[object_index, variable_index] 
        - the values of the explaining variables of the model. 
          The number and order of these must match with the variable 
          declarations in the variables section of the model definition in
          the model xml file
    arg.parameters[parameter_index]
        - the model parameter values. The order in which these are used must 
          match the order given in the parameters section of the model xml
          file
    arg.nres[object_index]
      - number of result objects. This is always 1 unless the prediction
        model really creates new objects; e.g. in stem distribution models
        each diameter or height class is a new object
    arg.mem[variable_index, result_index]
      - stores the model results. Again the results must be stored here
        in the same order as defined in the result/variables section of
        the model xml file
    arg.errors[object_index] 
        - for storing possible error messages. Before actually 
          evaluating the model, you should test for error situations and
          set error messages detailing why the model can't be evalueated and
          then exit
When getting or setting the attribute values of arg, oind is always used for
the object_index; e.g. the value of the first explaining variable of the 
model is accessed by arg.variables[0][oind]. As can be seen from the example
indices always start from 0.

The model must return an integer as its return value. 1 if there were no
errors and 0 if errors occurred.

Copyright (c) 2007 Simosol Oy. Distributed under the GNU General Public License 2.0.
"""

import math

def Test_Biomass_stump_spruce_Repola(arg, oind):
    """
    Method: Biomass_stump_spruce_Repola
    Calculates biomass of stump for Norway spruce 
    """
    ret = 1
    # model variables
    d = arg.variables[oind, 0]
    
    # the actual model
    d_s = 2+1.25*d
    if d_s-26.0 == 0.0:
        arg.errors[oind] = 'd=-26'
        return 0
    res = math.exp(-3.962+11.725*(d_s/(d_s+26.))+(0.065+0.058/2.))
    
    # data back to SIMO
    arg.nres[oind] = 1
    arg.mem[oind, 0] = arg.rect_factor * res
    return ret

def main():
    pass

if __name__ == '__main__':
    main()

