#!/usr/bin/env python
# encoding: utf-8
"""
The main module of SIMO forest management planning framework providing
access points to data import, simulation, optimization and data output.

*runner.py*

Usage: runner.py <configuration file>

Copyright (c) 2007 Simosol Oy. Distributed under the GNU General Public License 2.0.
"""

import sys
import os
import shutil
import logging
import cProfile
import pstats
import traceback
import pdb
from optparse import OptionParser

from simo.utils.runnerconfig import Config
from simo.utils.logger import Logger
from simo.builder import simoclasses as sc
from simo.input.importdata import DataImporter
from simo.input.importops import OperationImporter
from simo.simulation.sim import Simulator
from simo.optimization.genetic_algorithm import GeneticAlgorithm
from simo.optimization.simulated_annealing import SimulatedAnnealing
from simo.optimization.tabusearch import TabuSearch
from simo.optimization.hero import HERO
from simo.optimization.jlp import JLP
from simo.output.out import Out
import simo.db.simodb.db as simodb
import simo.db.datadb.db as datadb
from simo.utils.utils import print_stats

class Runner(object):

    def __init__(self, cfg):
        """
        Initialize Runner class instance

        cfg -- Config instance
        logger - Logger instance
        logname - log name as string
        """
        self.cfg = cfg
        self.initial_dir = os.getcwd()
        self.input_dbs = {}
        self.result_dbs = {}
        self.fops_dbs = {}
        self.opt_dbs = {}
        self.ctrls = {}
        self.mappings = {}
        self.remove_on_del = []
        self.remove_pidfiles = []
        os.chdir(cfg['program_execution.base_folder'])

    def remove_temp_paths(self):
        pop_paths = []
        for path in self.remove_on_del:
            print 'removing %s' % path
            try:
                if os.path.exists(path):
                    shutil.rmtree(path)
            except Exception:
                pass
            if os.path.exists(path):
                print 'failed to remove %s' % path
            pop_paths.append(path)
        for path in pop_paths:
            self.remove_on_del.pop(self.remove_on_del.index(path))
        for f in self.remove_pidfiles:
            try:
                if os.path.exists(f):
                    os.remove(f)
            except Exception:
                print "couldn't remove pidfile %s" % (f,)


    def _make_new_db(self, option):
        if option:
            make_new = True
        else:
            make_new = False
        return make_new

    def _load_sim_control(self, inilogger, sc_name):
        try:
            ctrl = self.struct_db.get_obj(sc.SIMULATION_CONTROL_DEF,
                                          sc_name)
        except ValueError, e:
            errormsg = "When loading simulation control '%s': %s" \
                        % (sc_name, e.args[0])
            inilogger.error(errormsg)
            self.terminate()
        if ctrl is None:
            errormsg = "Could not load simulation control with name '%s'" \
                     % sc_name
            inilogger.error(errormsg)
            self.terminate()
        return ctrl

    def open_databases(self, inilogger, memory_import=False):
        """
        Open DataDB databases
        """
        # First get the needed definitions from lexicon and simulation control
        try:
            self.lx = self.struct_db.get_obj(sc.LEXICON_DEF, 'lexicon')
        except ValueError, e:
            errormsg = "When loading lexicon: %s" % e.args[0]
            inilogger.error(errormsg)
            self.terminate()
        self.sim_level_name = self.lx.get_level_name(0)
        content_def = self.lx.get_content_def()
        hierarchy = self.lx.hierarchy
        level_meta = self.lx.levels
        sc_name = self.cfg['simulation.simulation_control']
        self.ctrl = self._load_sim_control(inilogger, sc_name)
        for sc_name in self.cfg['additional_options.controls']:
            self.ctrls[sc_name] = self._load_sim_control(inilogger, sc_name)
        classifiers = self.ctrl.opres_cfiers
        opres_vars = self.ctrl.opres_vars
        constraints = self.ctrl.output_constraints
        opres_def = opres_vars.union(classifiers)
        cf_cfiers = self.ctrl.cf_cfiers

        # checks for simulation level for database, and the one in
        # simulation control
        sim_level = self.cfg['data_db.simulation_level']
        ok = False
        all_levels = []
        for level in content_def.keys():
            all_levels.append(level[0])
            if level[0] == sim_level:
                ok = True
                sim_ord = level[1]
                break
        if not ok:
            errormsg = "Simulation level definition for database is not valid"\
                    " (data_db.simulation_level = %s). I only have: %s" \
                    % (sim_level, all_levels)
            inilogger.error(errormsg)
            self.terminate()
        if self.ctrl.main_level != sim_ord:
            errormsg = "Simulation level definition for database and the "\
                       "simulation control main level don't match: %s != %s"\
                       % (self.cfg['data_db.simulation_level'],
                          self.lx.get_level_name(self.ctrl.main_level))
            inilogger.error(errormsg)
            self.terminate()

        db_type = self.cfg['data_db.db_type']
        dbdir = self.cfg['data_db.sqlite.db_dir']
        in_memory = self.cfg['data_db.sqlite.in_memory_db']
        spatial = self.cfg['data_db.spatial']
        threaded = self.cfg['data_db.threaded']
        unlogged = self.cfg['data_db.unlogged']
        if db_type == 'SQLITE':
            input_db = os.path.join(dbdir, self.cfg['data_db.sqlite.input_db'])
            operation_db = os.path.join(dbdir,
                                    self.cfg['data_db.sqlite.operation_db'])
            result_db = os.path.join(dbdir,
                                     self.cfg['data_db.sqlite.result_db'])
            opt_db = os.path.join(dbdir,
                                  self.cfg['data_db.sqlite.optimized_db'])
        elif db_type == 'POSTGRESQL':
            input_db = self.cfg['data_db.postgresql.input_prefix']
            operation_db = self.cfg['data_db.postgresql.operation_prefix']
            result_db = self.cfg['data_db.postgresql.result_prefix']
            opt_db = self.cfg['data_db.postgresql.optimized_prefix']
        elif db_type == 'ORACLE':
            input_db = self.cfg['data_db.oracle.input_prefix']
            operation_db = self.cfg['data_db.oracle.operation_prefix']
            result_db = self.cfg['data_db.oracle.result_prefix']
            opt_db = self.cfg['data_db.oracle.optimized_prefix']

        make_new = self._make_new_db(self.cfg['program_execution.import_data'])
        make_new = make_new and not \
                self._make_new_db(self.cfg['program_execution.no_wipe_import'])
        no_exec = self.cfg['program_execution.no_exec_import']
        db_host = self.cfg['data_db.postgresql.db_host']
        db_port = self.cfg['data_db.postgresql.db_port']
        db_user = self.cfg['data_db.postgresql.db_user']
        db_pw = self.cfg['data_db.postgresql.db_pw']
        if db_type == 'ORACLE':
            db_host = self.cfg['data_db.oracle.db_host']
            db_user = self.cfg['data_db.oracle.db_user']
            db_pw = self.cfg['data_db.oracle.db_pw']
        db_database = self.cfg['data_db.postgresql.db_name']
        db_schema = self.cfg['data_db.postgresql.db_schema']

        # Set the memory import settings if applicable
        if memory_import:
            orig_db_type = db_type
            orig_in_memory = in_memory
            db_type = 'SQLITE'
            in_memory=True

        self.input_db = datadb.DataDB('read', (sim_ord, sim_level),
                      content_def, hierarchy, level_meta,
                      opres_def, cf_cfiers, self.logger, db_name=input_db,
                      make_new_db=make_new, memory=in_memory, spatial=spatial,
                      db_host=db_host, db_port=db_port, user=db_user,
                      pw=db_pw, db_class=db_type, no_exec=no_exec,
                      db_database=db_database, db_schema=db_schema,
                      threaded=threaded, unlogged=unlogged)
        self.input_dbs[input_db] = self.input_db
        for item in self.cfg['additional_options.input_db']:
            item = os.path.join(dbdir, item)
            self.input_dbs[item] = datadb.DataDB('read', (sim_ord, sim_level),
                      content_def, hierarchy, level_meta,
                      opres_def, cf_cfiers, self.logger, db_name=item,
                      make_new_db=make_new, memory=in_memory, spatial=spatial,
                      db_host=db_host, db_port=db_port, user=db_user,
                      pw=db_pw, db_class=db_type, no_exec=no_exec,
                      db_database=db_database, db_schema=db_schema,
                      threaded=threaded, unlogged=unlogged)
            self.input_dbs[item].close()
        ext_cols=[]
        if self.cfg['import.operations.operation_import']:
            ext_cols = self._get_forced_op_extra_cols(\
                        'import.operations.operation_conversion')
        self.fops_db = datadb.OperationDB(operation_db, make_new, in_memory,
                       db_host=db_host, db_port=db_port, user=db_user,
                       pw=db_pw, db_class=db_type,
                       db_database=db_database, db_schema=db_schema,
                       extra_columns=ext_cols,
                       threaded=threaded)
        self.fops_dbs[operation_db] = self.fops_db
        for item in self.cfg['additional_options.operation_db']:
            item = os.path.join(dbdir, item)
            self.fops_dbs[item] = datadb.OperationDB(item, make_new, in_memory,
                              db_host=db_host, db_port=db_port, user=db_user,
                              pw=db_pw, db_class=db_type,
                              db_database=db_database, db_schema=db_schema,
                              threaded=threaded)
            self.fops_dbs[item].close()

        # Revert memory import settings if applicable
        if memory_import:
            db_type = orig_db_type
            in_memory = orig_in_memory
        make_new = self._make_new_db(self.cfg['program_execution.simulate'])
        make_new = make_new and not \
              self._make_new_db(self.cfg['program_execution.no_wipe_simulate'])
        no_exec = self.cfg['program_execution.no_exec_simulate']
        create_branch_desc = self.cfg['simulation.create_branch_desc']
        track_solution = self.cfg['optimization.track_solution']
        self.result_db = datadb.DataDB('write', (sim_ord, sim_level),
                         content_def, hierarchy, level_meta,
                         opres_def, cf_cfiers, self.logger, db_name=result_db,
                         make_new_db=make_new, memory=in_memory,
                         spatial=spatial, constraints=constraints,
                         db_host=db_host, db_port=db_port, user=db_user,
                         pw=db_pw, db_class=db_type,
                         create_branch_desc=create_branch_desc,
                         track_solution=track_solution, no_exec=no_exec,
                         db_database=db_database, db_schema=db_schema,
                         threaded=threaded, unlogged=unlogged)
        self.result_dbs[result_db] = self.result_db
        for item in self.cfg['additional_options.result_db']:
            item = os.path.join(dbdir, item)
            self.result_dbs[item] = \
                        datadb.DataDB('write', (sim_ord, sim_level),
                            content_def, hierarchy, level_meta,
                            opres_def, cf_cfiers, self.logger, db_name=item,
                            make_new_db=make_new, memory=in_memory,
                            spatial=spatial, constraints=constraints,
                            db_host=db_host, db_port=db_port, user=db_user,
                            pw=db_pw, db_class=db_type,
                            create_branch_desc=create_branch_desc,
                            track_solution=track_solution, no_exec=no_exec,
                            db_database=db_database, db_schema=db_schema,
                            threaded=threaded, unlogged=unlogged)
            self.result_dbs[item].close()

        make_new = self._make_new_db(self.cfg['program_execution.optimize'])
        make_new = make_new and not \
              self._make_new_db(self.cfg['program_execution.no_wipe_optimize'])
        no_exec = self.cfg['program_execution.no_exec_optimize']
        create_opt_db = self.cfg['optimization.create_opt_db']
        if not create_opt_db:
            make_new = False
        try:

            # TODO: HACK: FIXME: this is a hack for a situation where the
            # simulated op_res structure differs from optimized (different
            # simulation controls)
            if self.result_db.opres_def != opres_def:
                opres_def = list(self.result_db.opres_def)
                if 'cash_flow' in opres_def:
                    opres_def.remove('cash_flow')

            self.opt_db = datadb.DataDB('write', (sim_ord, sim_level),
                              content_def, hierarchy, level_meta,
                              opres_def, cf_cfiers, self.logger,
                              db_name=opt_db, make_new_db=make_new,
                              memory=in_memory, spatial=spatial,
                              constraints=constraints,
                              db_host=db_host, db_port=db_port, user=db_user,
                              pw=db_pw, db_class=db_type, no_exec=no_exec,
                              db_database=db_database, db_schema=db_schema,
                              threaded=threaded, unlogged=unlogged)
            self.opt_dbs[opt_db] = self.opt_db
            for item in self.cfg['additional_options.optimized_db']:
                item = os.path.join(dbdir, item)
                self.opt_dbs[item] = \
                          datadb.DataDB('write', (sim_ord, sim_level),
                              content_def, hierarchy, level_meta,
                              opres_def, cf_cfiers, self.logger, db_name=item,
                              make_new_db=make_new, memory=in_memory,
                              spatial=spatial, constraints=constraints,
                              db_host=db_host, db_port=db_port, user=db_user,
                              pw=db_pw, db_class=db_type, no_exec=no_exec,
                              db_database=db_database, db_schema=db_schema,
                              threaded=threaded, unlogged=unlogged)
                self.opt_dbs[item].close()
        except ValueError:
            if not make_new:
                # schema change detected, but since not doing optimization on
                # this database:
                self.opt_db = None

    def _get_forced_op_extra_cols(self, option):
        """
        Get all forced operation parameters and variables as they have to be
        set as columns in forced operation database
        """
        conv = self.cfg[option]
        conv_def = self.struct_db.get_obj(sc.TEXT2OPERATION_DEF, conv)
        if not conv_def:
            return []
        ext_cols = set([])
        if conv_def.operation_conv.variables:
            for key, item in conv_def.operation_conv.variables.items():
                if item.var_type == 'categorical':
                    for key, items in item.map.items():
                        for var in items:
                            ext_cols.add((var.level, var.to_var))
                else:
                    for var in item.map:
                        ext_cols.add((var.level, var.to_var))
        for key, op in conv_def.operation_conv.operations.items():
            if op.parameters:
                for p in op.parameters:
                    ext_cols.add((p.level, p.variable))
        return list(ext_cols)

    def open_logger(self, msgttable, lexttable, lang):
        wipe = self.cfg['logging.wipe_log']
        db_type = self.cfg['data_db.db_type']
        in_memory = self.cfg['data_db.sqlite.in_memory_db']
        dbdir = self.cfg['data_db.sqlite.db_dir']
        if db_type == 'SQLITE':
            dbpath = os.path.join(dbdir, self.cfg['data_db.sqlite.log_db'])
        else:
            dbpath = self.cfg['data_db.postgresql.log_prefix']
        db_host = self.cfg['data_db.postgresql.db_host']
        db_port = self.cfg['data_db.postgresql.db_port']
        db_user = self.cfg['data_db.postgresql.db_user']
        db_pw = self.cfg['data_db.postgresql.db_pw']
        db_database = self.cfg['data_db.postgresql.db_name']
        db_schema = self.cfg['data_db.postgresql.db_schema']
        no_exec = self.cfg['program_execution.no_exec_logger']
        if db_type == 'ORACLE':
            db_host = self.cfg['data_db.oracle.db_host']
            db_user = self.cfg['data_db.oracle.db_user']
            db_pw = self.cfg['data_db.oracle.db_pw']
            dbpath = self.cfg['data_db.oracle.log_prefix']
        self.log_db = datadb.LoggerDB(dbpath, wipe, in_memory,
                       db_host=db_host, db_port=db_port, user=db_user,
                       pw=db_pw, db_class=db_type, no_exec=no_exec,
                       db_database=db_database, db_schema=db_schema)
        self.logger = Logger(self.log_db,
                     self.cfg['program_execution.run_id'],
                     msgttable.translation, lexttable, lang,
                     self.cfg['logging.console'],
                     self.cfg['logging.console_level'],
                     self.cfg['warning_handling.output_warnings'])

    def open_simodb(self):
        """
        Opens the SIMO DB containing all the XmlObject instances
        """
        simo_db = self.cfg['simo_db.db_path']
        self.struct_db = simodb.SimoDB(simo_db, create_new=False,
                                       read_only=True)

    def _close_db(self, db):
        try:
            db.close()
        except:
            pass

    def close_databases(self):
        """Close DataDB, SimoDB and LoggerDB databases"""
        try: # just in case the log_db is closed.
            infomsg = 'Closing databases...'
            self.logger.log_message('runner', 'info', infomsg)
        except Exception:
            pass
        self._close_db(self.input_db)
        self._close_db(self.fops_db)
        self._close_db(self.result_db)
        self._close_db(self.opt_db)
        self._close_db(self.struct_db)
        self._close_db(self.log_db)

    def import_data(self):
        """Import data and forced operations into input dbs from data source"""
        log_name = 'data import'
        max_errors = self.cfg['error_handling.max_error_count_total']
        import_date = self.cfg['import.import_date']
        initial_date = self.cfg['import.data_date']
        month_first = self.cfg.get('import.month_first')
        if month_first is None: month_first = False
        encodings = self.cfg.get('import.encodings')
        mappingdef = self.cfg['import.mapping_definition']
        format = self.cfg['import.format']
        folder = self.cfg['import.data_dir']
        datafiles = self.cfg['import.data_file']
        separator = self.cfg['import.data_delimiter']
        skipfirst = self.cfg['import.skip_first_line']
        forced_ops = self.cfg['import.operations.operation_import']
        if forced_ops:
            # Import forced operations into input database
            oformat = self.cfg['import.operations.operation_format']
            ofile = self.cfg['import.operations.operation_file']
            omap = self.cfg['import.operations.operation_modelchains']
            oconv = self.cfg['import.operations.operation_conversion']
        infomsg = 'Initializing SIMO importer!'
        self.logger.log_message(log_name, 'info', infomsg)
        simlevel = self.sim_level_name
        mapping = self.struct_db.get_obj(sc.TEXT2DATA_DEF, mappingdef)
        if mapping is None:
            errormsg = "Data mapping definition '%s' couldn't be loaded" \
                       % mappingdef
            self.logger.log_message(log_name, 'error', errormsg)
            return False
        imp = DataImporter(self.input_db, mapping, import_date, self.logger,
                           log_name, self.lx, max_errors)
        buffers = []
        for file in datafiles:
            fpath = os.path.join(folder, file)
            buffers.append(open(fpath))
        imp.import_data(format, buffers, separator=separator,
                        skip_first=skipfirst, sim_level=simlevel,
                        data_date=initial_date, month_first=month_first,
                        encodings=encodings)
        cont = not imp.terminate_import
        if not imp.terminate_import and forced_ops:
            conv_def, op2mc = None, None
            if oformat in ('text',):
                conv_def = self.struct_db.get_obj(sc.TEXT2OPERATION_DEF, oconv)
                if conv_def is None and oformat:
                    errormsg = "Operation mapping definition '%s' couldn't "\
                               "be loaded"  % oconv
                    self.logger.log_message(log_name, 'error', errormsg)
                    return False
            elif oformat in ('xml', 'db'):
                op2mc = self.struct_db.get_obj(sc.OPERATION2MODELCHAINS_DEF,
                                               omap)
            imp = OperationImporter(self.input_db, self.fops_db, oformat,
                                    conv_def, op2mc, self.logger, log_name,
                                    max_errors)
            if oformat != 'db':
                op_buffer = open(ofile)
            else:
                op_buffer = datadb.SQLiteDB(ofile, wipe=False, commit=False)
            imp.import_operations(op_buffer, separator)
            cont = not imp.terminate_import
        return cont

    def simulate(self):
        """
        Constructs the simulator according to the configuration and simulation
        definitions and runs the simulation for the data in the input db.

        Simulation definitions are collected into SimInit object, which is then
        used to initialize the simulator object.
        """
        simtype = self.cfg['simulation.task_type']
        logname = 'simulation'
        msg = 'Initializing simulator...'
        self.logger.log_message(logname, 'info', msg)
        if self.ctrl is None:
            errormsg = "Simulation control definition couldn't be loaded"
            self.logger.log_message(logname, 'error', errormsg)
            return False
        # get the error handling settings
        max_errors = self.cfg['error_handling.max_error_count_total']
        max_simunit_errors = \
                self.cfg['error_handling.max_error_count_per_unit']
        # Initialise the simulator with the simulation definitions and DBs
        self.ctrl.use_data_date = \
                self.cfg['simulation.use_data_date_as_init_date']
        self.ctrl.use_today = self.cfg['simulation.use_today_as_init_date']
        self.ctrl.initial_date = self.cfg['simulation.fixed_init_date']
        if simtype == 'data processing':
            opt = 'simulation.fix_data_processing_date_to_init'
            self.ctrl.fixed_dp_date = self.cfg[opt]
        warnings = self.cfg['warning_handling.output_warnings']
        objcount_multiplier = self.cfg['simulation.object_count_multiplier']
        branchcount = self.cfg['simulation.estimated_branch_count']
        max_unit_count = self.cfg['simulation.max_unit_count']
        noopbranch = self.cfg['simulation.leave_no_op_branch']
        separatebgs = self.cfg['simulation.separate_branching_groups']
        branchlimit = self.cfg['simulation.max_branch_limit']
        max_res_objs = self.cfg['simulation.max_result_unit_count']
        deterministic = self.cfg['simulation.deterministic']
        track_prices = self.cfg['simulation.track_prices']
        do_forced_ops = self.cfg['simulation.do_forced_ops']
        copy2iters = self.cfg['simulation.copy_ops_to_all_iterations']
        create_branch_desc = self.cfg['simulation.create_branch_desc']
        cache_dir = self.cfg['simulation.cache_dir']
        simulator = Simulator(self.logger, logname, simtype, self.lx,
                   self.ctrl, self.input_db, self.result_db, self.fops_db,
                   warnings, max_unit_count, objcount_multiplier, branchcount,
                   noopbranch, separatebgs, branchlimit, max_res_objs,
                   max_errors, max_simunit_errors, deterministic, track_prices,
                   do_forced_ops, copy2iters, create_branch_desc, cache_dir)
        idlist = self.cfg.get('simulation.id_list')
        indseq = self.cfg.get('simulation.index_sequence')
        iterations = self.cfg.get('simulation.iterations')
        run_profiler = self.cfg['program_execution.code_profiling']
        try:
            if not run_profiler:
                simulator.simulate(iterations, idlist, indseq)
            else:
                self._profile_sim(simulator, iterations, idlist, indseq)
        finally:
            simulator.cleanup()

        cont = not simulator.terminate
        return cont

    def _get_opt_settings(self, opmethod, logname):
        """
        Get settings for selected optimisation method from configuration file
        
        opmethod -- optimization method, string
        """
        sect = 'optimization.%s.' % opmethod
        if opmethod == 'ts':
            settings = {
                'repeats':              self.cfg[sect+'repeats'],
                'initial_solutions':    self.cfg[sect+'initial_solutions'],
                'maximum_iterations':   self.cfg[sect+'maximum_iterations'],
                'entrytenure':          self.cfg[sect+'entry_tenure'],
                'exittenure':           self.cfg[sect+'exit_tenure'],
                }
        elif opmethod == 'ta':
            # NB: Optimization method not yet implemented!
            settings = {}
        elif opmethod == 'hero':
            settings = {
                'repeats':              self.cfg[sect+'repeats'],
                'initial_solutions':    self.cfg[sect+'initial_solutions'],
                'maximum_iterations':   self.cfg[sect+'maximum_iterations'],
                }
        elif opmethod == 'sa':
            settings = {
                'repeats':              self.cfg[sect+'repeats'],
                'initial_solutions':    self.cfg[sect+'initial_solutions'],
                'maximum_iterations':   self.cfg[sect+'maximum_iterations'],
                'starting_temperature': self.cfg[sect+'starting_temperature'],
                'freezing_temperature': self.cfg[sect+'freezing_temperature'],
                'cooling_factor':       self.cfg[sect+'cooling_factor'],
                'iterations_per_temperature': \
                    self.cfg[sect+'iterations_per_temperature'],
                'iteration_multiplier': self.cfg[sect+'iteration_multiplier'],
                }
        elif opmethod == 'ga':
            settings = {
                'repeats':              self.cfg[sect+'repeats'],
                'maximum_iterations':   self.cfg[sect+'maximum_iterations'],
                'population_size':      self.cfg[sect+'population_size'],
                'breeding_rate':        self.cfg[sect+'breeding_rate'],
                'mutation_probability': self.cfg[sect+'mutation_probability'],
                'swapping_probability': self.cfg[sect+'swapping_probability'],
                }
        elif opmethod == 'jlp':
            settings = {
                'maxobjects':           self.cfg[sect+'max_objects'],
                'jfolder':              self.cfg[sect+'j_folder'],
                'prefix':               self.cfg[sect+'file_prefix'],
                'timeout':              self.cfg[sect+'timeout'],
                'do_split':             self.cfg[sect+'do_split'],
                'skip_weight_splits':   self.cfg[sect+'skip_weight_splits'],
                }
            # JLP can have a fallback heuristic method for situations with 
            # severe constraint violations due to LP relaxation
            settings['do_fallback'] = self.cfg[sect+'do_fallback']
            if settings['do_fallback']:
                fb_method = self.cfg[sect+'fallback_method'].lower()
                fb_settings = self._get_opt_settings(fb_method, logname)
                settings['fallback_method'] = fb_method
                settings['fallback_settings'] = fb_settings
            else:
                settings['fallback_method'] = None
                settings['fallback_settings'] = None
        
        return settings

    def optimize(self):
        """
        Find an optimal solution from the result database with an optimization
        algorithm.

        At the moment there are two heuristic optimization methods and an
        interface to JLP linear optimization library.
        """
        logname = 'optimization'
        opmethod = self.cfg['optimization.method'].lower()
        taskdef = self.cfg['optimization.problem_definition']
        limit2branches = self.cfg['optimization.limit2branches']
        keep_data = self.cfg['optimization.keep_opt_data']
        reuse_data = self.cfg['optimization.reuse_opt_data']
        track_solution = self.cfg['optimization.track_solution']
        create_opt_db = self.cfg['optimization.create_opt_db']
        chart_path = self.cfg['optimization.chart_path']

        opt_task = self.struct_db.get_obj(sc.PROBLEM_DEFINITION_DEF, taskdef)
        if opt_task is None:
            errormsg = "Optimization task definition '%s' couldn't be loaded"\
                    % taskdef
            self.logger.log_message(logname, 'error', errormsg)
            return False
        
        # Set optimization task arguments that are common to all methods
        args = (self.logger, logname, opt_task, self.result_db, self.opt_db,
                keep_data, reuse_data, limit2branches, chart_path,
                track_solution, taskdef, create_opt_db,)
        
        # Get optimization method-specific settings from configuration file
        settings = self._get_opt_settings(opmethod, logname)
        
        if opmethod == 'ts':
            opmethod_name = 'Tabu Search'
            optimizer = TabuSearch(*args, **settings)
        elif opmethod == 'ta':
            opmethod_name = "Threshold accepting"
            msg = "Threshold accepting not yet implemented!"
            self.logger.log_message(logname, 'error', msg)
            return False
        elif opmethod == 'hero':
            opmethod_name = 'HERO'
            optimizer = HERO(*args, **settings)
        elif opmethod == 'sa':
            opmethod_name = 'Simulated Annealing'
            optimizer = SimulatedAnnealing(*args, **settings)
        elif opmethod == 'ga':
            opmethod_name = 'Genetic Algorithm'
            optimizer = GeneticAlgorithm(*args, **settings)
        elif opmethod == 'jlp':
            opmethod_name = 'J Linear Programming'
            optimizer = JLP(*args, **settings)
        # Run the optimization algorithm and close the databases
        if optimizer:
            run_profiler = self.cfg['program_execution.code_profiling']
            infomsg = "Starting %s optimization" % opmethod_name
            self.logger.log_message(logname, 'INFO', infomsg)
            if not run_profiler:
                optimizer.optimize()
            else:
                self._profile_opt(optimizer, 'opt')
            optimizer.close()

    def output_data(self):
        """
        Output simulated or optimized results from 'result' and
        'operation_result' databases.
        """
        logname = 'output'
        infomsg = 'Initializing data output...'
        self.logger.log_message(logname, 'info', infomsg)
        # Set databases
        resdb = self._out_setdbs()
        dbtype = self.cfg['output.data_type']
        formats = self.cfg['output.output_format']
        restype = self.cfg['output.result_type']
        # Get other output definitions
        resdir = self.cfg['output.output_dir']
        resfiles = self.cfg['output.output_filename']
        archiving = self.cfg.get('output.archiving')
        decimals = self.cfg['output.default_decimal_places']
        start_date = self.cfg.get('output.start_date', None)
        end_date = self.cfg.get('output.end_date', None)
        start_date = None if start_date == '' else start_date
        end_date = None if end_date == '' else end_date
        full_date = self.cfg['output.full_date']
        dates = (start_date, end_date)
        if dates == (None, None):
            dates = None
        result_padding = self.cfg['output.inlined.result_padding']
        id_list = self.cfg.get('output.id_list')
        if id_list == ['']:
            id_list = None
        format = list(formats)
        main_level = self.ctrl.main_level
        constraint = None
        aggr_def = None
        expr_def = None
        outputdef = None
        outputvars = None
        group_by_vars = None
        constr_init = False
        # Go through all defined output formats and create output for each of
        # the result types
        for i in xrange(len(formats)):
            if not constr_init and format[i] in ['inlined', 'by_level', 'smt']:
                constr_init = True
                constraint = self.cfg['output.output_constraint']
                outputdef = self.struct_db.get_obj(sc.OUTPUT_CONSTRAINT_DEF,
                                                   constraint)
                if outputdef is None:
                    errormsg = "Output constraint definition '%s' couldn't"\
                               " be loaded" % constraint
                    self.logger.log_message(logname, 'error', errormsg)
                    return False
            elif format[i] == 'aggregation':
                aggrdef = self.cfg['output.aggregation.aggregation_definition']
                aggr_def = self.struct_db.get_obj(sc.AGGREGATION_DEF, aggrdef)
                if aggr_def is None:
                    errormsg = "Aggregation definition '%s' couldn't"\
                               " be loaded" % aggrdef
                    self.logger.log_message(logname, 'error', errormsg)
                    return False
            elif format[i] == 'expression':
                exprdef = self.cfg['output.expression.expression_definition']
                expr_def = self.struct_db.get_obj(sc.EXPRESSION_DEF, exprdef)
                if expr_def is None:
                    errormsg = "Expression definition '%s' couldn't"\
                               " be loaded" % exprdef
                    self.logger.log_message(logname, 'error', errormsg)
                    return False
            elif format[i] == 'operation_result':
                outputvars = self.cfg.get('output.operation_result.vars')
                group_by_vars = self.cfg.get('output.operation_result.group_by')
        # check if output should be profiled
        run_profiler = self.cfg['program_execution.code_profiling']
        if run_profiler:
            self._profile_out(resdb, dbtype, id_list, main_level, restype,
                              formats, resdir, resfiles, self.logger,
                              outputdef, decimals, archiving, result_padding,
                              aggr_def, expr_def, outputvars, group_by_vars,
                              full_date, dates)
        else:
            Out(resdb, dbtype, id_list, main_level,
                restype, formats, resdir, resfiles, self.logger, outputdef,
                decimals, archiving, result_padding, aggr_def, expr_def,
                outputvars, group_by_vars, full_date, dates)


    def _out_setdbs(self):
        """Assign the databases for output based on the setting in ini
        """
        dbtype = self.cfg['output.data_type']
        if dbtype == 'input':
            resdb = self.input_db
        elif dbtype in ['result', 'operation_result']:
            if self.cfg['output.result_type']=='simulated':
                resdb = self.result_db
            else:
                resdb = self.opt_db
        return resdb

    def _profile_sim(self, simulator, iterations, idlist, indseq):
        # try to import guppy for memory use profiling
        profile_heap = True
        try:
            from guppy import hpy
        except:
            profile_heap = False
        globals()['simulator'] = simulator
        globals()['iterations'] = iterations
        globals()['idlist'] = idlist
        globals()['indseq'] = indseq
        # set the default heap, everything allocated before this call will not
        # be in the data sets you get later.
        if profile_heap:
            hp = hpy()
            hp.setrelheap()
        cProfile.run('simulator.simulate(iterations, idlist, indseq)',
                     'simprofile.pro')
        stats = pstats.Stats('simprofile.pro')
        h = hp.heap() if profile_heap else None
        print_stats(stats, None, 'simulator', h)

    def _profile_opt(self, optimizer, ptype):
        globals()['optimizer'] = optimizer
        globals()['data_ok'] = data_ok = False
        if ptype=='data':
            cProfile.run('data_ok = optimizer.set_data()', 'opt_data.pro')
            stats = pstats.Stats('opt_data.pro')
            data_ok = globals()['data_ok']
        else:
            cProfile.run('optimizer.optimize()', 'opt.pro')
            stats = pstats.Stats('opt.pro')
        stats.strip_dirs()
        stats.sort_stats('cumulative')
        print_stats(stats, None, 'opt %s' % ptype)
        return data_ok

    def _profile_out(self, resdb, dbtype, id_list, main_level, restype,
                     formats, resdir, resfiles, logger, outputdef, decimals,
                     archiving, result_padding, aggr_def, expr_def, outputvars,
                     group_by_vars, full_date, dates):
        globals()['Out'] = Out
        globals()['resdb'] = resdb
        globals()['dbtype'] = dbtype
        globals()['id_list'] = id_list
        globals()['main_level'] = main_level
        globals()['restype'] = restype
        globals()['formats'] = formats
        globals()['resdir'] = resdir
        globals()['resfiles'] = resfiles
        globals()['logger'] = self.logger
        globals()['outputdef'] = outputdef
        globals()['decimals'] = decimals
        globals()['archiving'] = archiving
        globals()['result_padding'] = result_padding
        globals()['aggr_def'] = aggr_def
        globals()['expr_def'] = expr_def
        globals()['outputvars'] = outputvars
        globals()['group_by_vars'] = group_by_vars
        globals()['full_date'] = full_date
        globals()['dates'] = dates
        cProfile.run("Out(resdb, dbtype, id_list, main_level, restype, "\
                     "formats, resdir, resfiles, logger, outputdef, "\
                     "decimals, archiving, result_padding, aggr_def, "\
                     "expr_def,outputvars, group_by_vars, full_date, dates)",
                     "output.pro")
        stats = pstats.Stats('output.pro')
        stats.strip_dirs()
        stats.sort_stats('cumulative')
        print_stats(stats, None, 'output')


    def use_temp_cache(self, register_to_path=None):
        if register_to_path and not os.path.exists(register_to_path):
            raise Exception('register_to_path %s does not exist' \
                            % (register_to_path,))
        import tempfile
        tempdir = tempfile.mkdtemp()
        self.cfg['simulation.cache_dir'] = tempdir
        self.remove_on_del.append(self.cfg['simulation.cache_dir'])
        msg = 'Now using temp path for cache: %s' \
              % self.cfg['simulation.cache_dir']
        self.logger.log_message('', 'debug', msg)
        if register_to_path is not None:
            my_pid = os.getpid()
            pidfile = os.path.join(register_to_path, str(my_pid))
            f = open(pidfile, 'w')
            try:
                f.write(tempdir)
            finally:
                f.close()
            self.remove_pidfiles.append(pidfile)


    def terminate(self, errormsg=''):
        if errormsg == '':
            self.logger.log_message('', 'error', 'Terminated!')
        else:
            self.logger.log_message('', 'error', errormsg)
        self.struct_db.close()
        os.chdir(self.initial_dir)
        self.remove_temp_paths()
        sys.exit(0)

def _parse_ini_file(inilogger, inifile, replace_cfg=None):
    # Parse initialisation file
    cfg = Config(inilogger)
    cfg.load_config(inifile, replace_cfg=replace_cfg)
    if cfg.ini_err:
        inilogger.error('Terminating execution due to incorrect ' \
                        'initalization parameters')
        sys.exit(0)
    return cfg

def _get_ini_logger():
    inilogger = logging.getLogger('INI')
    handler = logging.StreamHandler()
    handler.setLevel(logging.ERROR)
    formatter = logging.Formatter('%(levelname)s: %(message)s')
    handler.setFormatter(formatter)
    inilogger.addHandler(handler)
    return inilogger

def _get_translators(inilogger, runner):
    # Need to open the simodb to get the translation tables
    struct_db = runner.struct_db
    msgtrans = runner.cfg['translation.message_translation']
    msgtable = _get_translation_table(sc.MESSAGE_TRANSLATION_DEF, msgtrans,
                                      struct_db, inilogger)
    lextrans =  runner.cfg['translation.lexicon_translation']
    lextable = _get_translation_table(sc.LEXICON_TRANSLATION_DEF, lextrans,
                                      struct_db, inilogger)
    if None in (msgtable, lextable):
        inilogger.error('Terminating execution due to incorrect ' \
                        'translation table definition')
        runner.terminate()
    lang = runner.cfg['translation.lang'].lower()
    return msgtable, lextable, lang

def _get_translation_table(ttype, tdef, struct_db, inilogger):
    ttable = struct_db.get_obj(ttype, tdef)
    if ttable is None:
        if ttype==sc.MESSAGE_TRANSLATION_DEF:
            msg = "Terminating execution due to incorrect message "\
                  "translation table definition '%s'" % tdef
        else:
            msg = "Terminating execution due to incorrect lexicon "\
                  "translation table definition '%s'" % tdef
        inilogger.error(msg)
    return ttable

def main():
    #DEBUG
    #import logging
    #logging.basicConfig()
    """
    Runner for SIMO computation. Reads initialization file and does data
    import, simulation, optimization and reporting according to the
    initialization definitions.
    """
    #NB: set Numpy to ignore warnings
    try:
        import numpy
        numpy.seterr(all='ignore')
    except:
        pass

    usage = "usage: %prog [options] ini_file\n"\
            "    ini_file -- text file containing the settings for the "\
            "program run\n"
    parser = OptionParser(usage)
    parser.add_option('-d', '--debug', dest='debug', action='store_true',
                      help='run post mortem debug mode enabled')
    options, args = parser.parse_args()
    if len(args) == 0:
        parser.error('Ini file not given')
    inifile = args[0]
    if os.path.isfile(inifile) == 0:
        parser.error('Configuration file %s is not a valid file' % inifile)
    inilogger = _get_ini_logger()
    cfg = _parse_ini_file(inilogger, inifile)
    runner = Runner(cfg.cfg)
    runner.open_simodb()
    msgttable, lexttable, lang = _get_translators(inilogger, runner)
    runner.open_logger(msgttable, lexttable, lang)
    try:
        runner.open_databases(inilogger)
    except Exception, e:
        tp, value, tb = sys.exc_info()
        traceback.print_exc()
        if options.debug:
            pdb.post_mortem(tb)
        runner.terminate('Unable to open databases (%s)' % str(e))
    cont = True
    try:
        if runner.cfg['program_execution.import_data']:
            cont = runner.import_data()
        if cont and runner.cfg['program_execution.simulate']:
            cont = runner.simulate()
        if cont and runner.cfg['program_execution.optimize']:
            runner.optimize()
        if cont and runner.cfg['program_execution.output_data']:
            runner.output_data()
    except Exception, e:
        tp, value, tb = sys.exc_info()
        traceback.print_exc()
        if options.debug:
            pdb.post_mortem(tb)
    finally:
        runner.close_databases()
        runner.remove_temp_paths()

if __name__ == '__main__':
    main()
