import sys, os
import time
import traceback
import pickle
import runner
from cStringIO import StringIO
from simo.builder import simoclasses as sc
from simo.input.importops import OperationImporter
from simo.input.importdata import DataImporter
from simo.simulation.sim import Simulator
import simo.db.datadb.db as datadb


def _get_ini_file(inilogger, filename):
# Check that runner .ini file exists
    if os.path.isfile(filename) == 0:
        inilogger.error('Configuration file %s is not a valid file',
                        os.path.join(os.getcwd(), filename))
        inilogger.error('Terminating program execution')
        sys.exit(0)
    return filename

class CeleryRunner(runner.Runner):
    def __init__(self, initial_setup, inilogger=None, cfg=None, db_info=None,
                 fully_in_memory=False):
        """
        Do the initialization, but do not change dirs as with original runner.

        initial_setup -- if this is the master setting things up, we have a
                         parsed ini configuration
        inilogger -- logger used with initialisation
        cfg -- parsed ini configuration, or None in case of worker setup
        db_info -- db connection details for worker setup
        fully_in_memory -- running fully in memory (no cfg dump and db_info)
        """
        self.inilogger = inilogger
        self.initial_dir = os.getcwd()
        # set or get the ini configuration
        if fully_in_memory:
            self.cfg = cfg.cfg
        elif initial_setup:
            self.cfg = cfg.cfg
            db_type = cfg.cfg['data_db.db_type'].upper()
            if db_type != 'SQLITE':
                host = cfg.cfg['data_db.%s.db_host' % db_type.lower()]
                user = cfg.cfg['data_db.%s.db_user' % db_type.lower()]
                pw = cfg.cfg['data_db.%s.db_pw' % db_type.lower()]
                if db_type == 'ORACLE':
                    port = None
                    schema = None
                    name = None
                elif db_type == 'POSTGRESQL':
                    port = cfg.cfg['data_db.%s.db_port' % db_type.lower()]
                    schema = cfg.cfg['data_db.%s.db_schema' % db_type.lower()]
                    name = cfg.cfg['data_db.%s.db_name' % db_type.lower()]
                self.db_info = (db_type, host, port, user, pw, name, schema)
                self.dump_config(cfg.cfg)
        else:
            self.db_info = db_info
            self.get_config(inilogger)
        self.struct_db = None
        self.input_dbs = {}
        self.result_dbs = {}
        self.fops_dbs = {}
        self.opt_dbs = {}
        self.ctrls = {}
        self.mappings = {}
        self.remove_on_del = []
        self.remove_pidfiles = []
        self.data_imp = None
        self.op_imp = None

    def open_databases(self, inilogger, memory_import=False):
        """
        Get the runner specific options

        WARNING!
        if there are differences in the controls used, specifically:
        ctrl.opres_cfiers
        ctrl.opres_vars
        ctrl.output_constraints
        ctrl.cf_cfiers
        need to be the same (or at least compatible) when swapping controls.
        The main simulation control is used to initialize all alt. databases.
        """
        super(CeleryRunner, self).open_databases(inilogger, memory_import)
        self.sim_ctrl = self.ctrl
        base_level = self.ctrl.main_level
        self.input_db.base_level = base_level
        self.result_db.base_level = base_level
        self.opt_db.base_level = base_level

    def _get_db(self, db_class, db_name, in_memory, override_no_exec,
                force_prefix=True, make_new=True):
        """
        Open a DataDB as a memory database
        db_class - input, result, optimal, operation or log
        db_name - name of the database (for table naming convention if
                  in-memory) or filepath to the database
        in_memory - return an in-memory database?
        override_no_exec - set no_exec to False regardless of ini setting
                           (necessary if you want to do any operations)
        force_prefix - name tables <db_name>_<table_name> regardless of db type
        make_new - make a new database instead of opening an old one
        """
        types = ('input', 'result', 'optimal', 'operation', 'log')
        if db_class not in types:
            raise Exception('Bad db_type: %s, expected one of %s' \
                            % (db_class, types))
        # First get the needed definitions from lexicon and simulation control
        try:
            self.lx = self.struct_db.get_obj(sc.LEXICON_DEF, 'lexicon')
        except ValueError, e:
            errormsg = "When loading lexicon: %s" % e.args[0]
            self.inilogger.error(errormsg)
            self.terminate()
        self.sim_level_name = self.lx.get_level_name(0)
        content_def = self.lx.get_content_def()
        hierarchy = self.lx.hierarchy
        level_meta = self.lx.levels
        classifiers = self.ctrl.opres_cfiers
        opres_vars = self.ctrl.opres_vars
        constraints = self.ctrl.output_constraints
        opres_def = opres_vars.union(classifiers)
        cf_cfiers = self.ctrl.cf_cfiers

        no_exec_imp = self.cfg['program_execution.no_exec_import']
        no_exec_sim = self.cfg['program_execution.no_exec_simulate']
        no_exec_opt = self.cfg['program_execution.no_exec_optimize']
        no_exec_log = self.cfg['program_execution.no_exec_logger']
        if override_no_exec:
            # This is what would usually be done here, unless you want to
            # create sql data to be put into some other database
            no_exec_imp = False
            no_exec_sim = False
            no_exec_opt = False
            no_exec_log = False

        # checks for simulation level for database, and the one in
        # simulation control
        sim_level = self.cfg['data_db.simulation_level']
        ok = False
        for level in content_def.keys():
            if level[0] == sim_level:
                ok = True
                sim_ord = level[1]
                break
        if not ok:
            errormsg = "Simulation level definition for database is not valid"\
                    " (data_db.simulation_level = %s)" % sim_ord
            self.inilogger.error(errormsg)
            self.terminate()
        if self.ctrl.main_level != sim_ord:
            errormsg = "Simulation level definition for database and the "\
                       "simulation control main level don't match: %s != %s"\
                       % (self.cfg['data_db.simulation_level'],
                          self.lx.get_level_name(self.ctrl.main_level))
            self.inilogger.error(errormsg)
            self.terminate()

        db_type = self.cfg['data_db.db_type']
        spatial = self.cfg['data_db.spatial']
        threaded = self.cfg['data_db.threaded']
        db_type = 'SQLITE'

        if db_class == 'input':
            return datadb.DataDB('read', (sim_ord, sim_level),
                          content_def, hierarchy, level_meta,
                          opres_def, cf_cfiers, self.logger, db_name=db_name,
                          make_new_db=make_new, memory=in_memory,
                          spatial=spatial, db_class=db_type,
                          no_exec=no_exec_imp, threaded=threaded,
                          force_prefix=force_prefix)


        ext_cols=[]
        if self.cfg['import.operations.operation_import']:
            ext_cols = self._get_forced_op_extra_cols(\
                        'import.operations.operation_conversion')
        if db_class == 'operation':
            return datadb.OperationDB(db_name, make_new, in_memory,
                               db_class=db_type, extra_columns=ext_cols,
                               threaded=threaded, force_prefix=force_prefix)
        if db_class == 'result':
            create_branch_desc = self.cfg['simulation.create_branch_desc']
            track_solution = self.cfg['optimization.track_solution']
            return datadb.DataDB('write', (sim_ord, sim_level),
                          content_def, hierarchy, level_meta,
                          opres_def, cf_cfiers, self.logger, db_name=db_name,
                          make_new_db=make_new, memory=in_memory,
                          spatial=spatial,
                          constraints=constraints, db_class=db_type,
                          create_branch_desc=create_branch_desc,
                          track_solution=track_solution, no_exec=no_exec_sim,
                          threaded=threaded, force_prefix=force_prefix)

        if db_class == 'optimal':
            return datadb.DataDB('write', (sim_ord, sim_level),
                          content_def, hierarchy, level_meta,
                          opres_def, cf_cfiers, self.logger,
                          db_name=db_name, make_new_db=make_new,
                          spatial=spatial,
                          memory=in_memory, constraints=constraints,
                          db_class=db_type, no_exec=no_exec_opt,
                          threaded=threaded, force_prefix=force_prefix)
        if db_class == 'log':
            return datadb.LoggerDB(db_name, make_new, in_memory,
                            db_class=db_type, no_exec=no_exec_log,
                            force_prefix=force_prefix)

        raise Exception('THIS SHOULD NEVER HAPPEN')


    def get_memory_db(self, db_class, db_name, override_no_exec=True,
                      force_prefix=True):
        """
        Open a DataDB as a memory database
        db_class - input, result, optimal, operation or log
        db_name - name of the database (for table naming convention)
        override_no_exec - force no_exec to be False regardless of ini
                           (necessary if you want to do any operations)
        force_prefix - name tables <db_name>_<table_name> regardless of db type
        """
        return self._get_db(db_class, db_name, True, override_no_exec,
                            force_prefix)


    def get_file_db(self, db_class, filepath, override_no_exec=True,
                    force_prefix=True, make_new=True):
        """
        Open a DataDB as a SQLite based database to the given file
        db_class - input, result, optimal, operation or log
        filepath - full path to database file
        override_no_exec - force no_exec to be False regardless of ini
                           (necessary if you want to do any operations)
        force_prefix - name tables <db_name>_<table_name> regardless of db type
        make_new - make a new database instead of opening an old one
        """
        return self._get_db(db_class, filepath, False, override_no_exec,
                            force_prefix, make_new)


    def get_config(self, inilogger):
        """
        Load configuration from the database
        """
        db_type, host, port, user, pw, name, schema = self.db_info
        if db_type.upper() == 'POSTGRESQL':
            db = runner.datadb.PostgreSQL(db_name='config', db_host=host,
                                   db_port=int(port), user=user, pw=pw,
                                   wipe=False, spatial=False,
                                   db_database=name, db_schema=schema)
        elif db_type.upper() == 'ORACLE':
            db = runner.datadb.Oracle(db_name='config', db_host=host,
                            user=user, pw=pw, wipe=False, spatial=False)
        else:
            raise Exception('Unsupported database engine %s' % db_type)
        try:
            # wait 10 seconds to assure the server has time to upload config
            time.sleep(10)
            # give the server 15 seconds to get data into configuration.
            data = None
            for i in range(5):
                if not db.table_exists('configuration'):
                    inilogger.warning('No configuration table, waiting...')
                    time.sleep(3)
                    continue
                data = db.get('SELECT data FROM configuration')
                if not data:
                    inilogger.warning('No data in configuration table, '\
                                      'waiting...')
                    time.sleep(3)
                else:
                    break
            if not data:
                inilogger.error('Terminating due to no configuration '\
                                'data in the database!')
                raise Exception('No configuration data in the database')
            cfg = pickle.loads(str(data[0][0]))
            self.cfg = cfg
            return cfg
        finally:
            db.close()

    def dump_config(self, cfg):
        """
        Dump the configuration into the database (so clients can read it)

        cfg -- parsed ini file
        """
        db_type, host, port, user, pw, name, schema = self.db_info
        self.cfg = cfg
        if db_type.upper() == 'POSTGRESQL':
            db = runner.datadb.PostgreSQL(db_name='config', db_host=host,
                                   db_port=int(port), user=user, pw=pw,
                                   wipe=True, spatial=False,
                                   db_database=name, db_schema=schema)
        elif db_type.upper() == 'ORACLE':
            db = runner.datadb.Oracle(db_name='config', db_host=host,
                            user=user, pw=pw, wipe=True, spatial=False)
        else:
            raise Exception('Unsupported database engine %s' % db_type)
        try:
            if db.table_exists('configuration'):
                sql = 'DELETE FROM configuration'
            else:
                sql = 'CREATE TABLE configuration (data TEXT)'
                if db_type.upper() == 'ORACLE':
                    sql = 'CREATE TABLE configuration (data NCLOB)'
            db.execute(sql)
            db.execute('INSERT INTO configuration VALUES (?)',
                       (pickle.dumps(cfg),))
        finally:
            db.close()


    def set_dbs(self, input=None, result=None, optimal=None, operation=None):
        """
        For each type given (name of the database), swap the database in as
        active
        """
        if input and not self.input_db == self.input_dbs[input]:
            self.input_db.db.close()
            self.input_db = self.input_dbs[input]
            self.input_db.db.open()
            self.input_db.base_level = self.ctrl.main_level
        if result and not self.result_db == self.result_dbs[result]:
            self.result_db.db.close()
            self.result_db = self.result_dbs[result]
            self.result_db.db.open()
            self.result_db.base_level = self.ctrl.main_level
        if optimal and not self.optimal_db == self.optimal_dbs[optimal]:
            self.opt_db.db.close()
            self.opt_db = self.opt_dbs[optimal]
            self.opt_db.db.open()
            self.opt_db.base_level = self.ctrl.main_level
        if operation and not self.fops_dbs==self.fops_dbs[operation]:
            self.fops_db.db.close()
            self.fops_db = self.fops_dbs[operation]
            self.fops_db.db.open()

    def set_sim_ctrl(self, sc_name):
        """
        Set the active simulation control
        """
        self.ctrl = self.ctrls[sc_name]

    def set_data_mapping(self, mapping_name):
        if self.cfg['import.mapping_definition'] != mapping_name:
            self.cfg['import.mapping_definition'] = mapping_name
            self._set_data_importer()

    def set_op_mapping(self, mapping_name):
        if self.cfg['import.operations.operation_conversion'] != mapping_name:
            self.cfg['import.operations.operation_conversion'] = mapping_name
            self._set_op_importer()


    def _set_data_importer(self):
        """
        Create the data importer based on the configuration and store it to
        self.data_imp
        """
        log_name = 'data import'
        max_errors = self.cfg['error_handling.max_error_count_total']
        import_date = self.cfg['import.import_date']
        mappingdef = self.cfg['import.mapping_definition']

        infomsg = 'Initializing SIMO importer!'
        self.logger.log_message(log_name, 'info', infomsg)
        mapping = self.struct_db.get_obj(sc.TEXT2DATA_DEF, mappingdef)
        if mapping is None:
            errormsg = "Data mapping definition '%s' "\
                       "couldn't be loaded" % mappingdef
            self.logger.log_message(log_name, 'error', errormsg)
            return False
        imp = DataImporter(self.input_db, mapping, import_date,
                           self.logger, log_name, self.lx, max_errors)
        self.data_imp = imp


    def _set_op_importer(self):
        """
        Create the operation importer based on the configuration and store
        it to self.op_imp
        """
        log_name = 'op import'
        # Import forced operations into input database
        max_errors = self.cfg['error_handling.max_error_count_total']
        oformat = self.cfg['import.operations.operation_format']
        omap = self.cfg['import.operations.operation_modelchains']
        oconv = self.cfg['import.operations.operation_conversion']

        conv_def, op2mc = None, None
        if oformat in ('text',):
            conv_def = self.struct_db.get_obj(\
                                        sc.TEXT2OPERATION_DEF,
                                        oconv)
            if conv_def is None and oformat:
                errormsg = "Operation mapping definition '%s' "\
                           "couldn't be loaded"  % oconv
                self.logger.log_message(log_name, 'error',
                                        errormsg)
                return False
        elif oformat in ('xml', 'db'):
            op2mc = self.struct_db.get_obj(\
                                    sc.OPERATION2MODELCHAINS_DEF,
                                    omap)
        imp = OperationImporter(self.input_db, self.fops_db,
                                oformat, conv_def, op2mc,
                                self.logger, log_name,
                                max_errors)
        self.op_imp = imp


    def import_data_from_strings(self, data_strs, ops):
        """
        Import data and forced operations into input dbs from data source

        data_strs - list of string data to be imported
        ops - operation data to be imported
        """
        if self.data_imp is None:
            self._set_data_importer()
        data = [StringIO(data_str) for data_str in data_strs]
        forced_ops = self.cfg['import.operations.operation_import']
        if forced_ops:
            if self.op_imp is None:
                self._set_op_importer()
            ops = StringIO(ops)
        try:
            format = self.cfg['import.format']
            separator = self.cfg['import.data_delimiter']
            skipfirst = self.cfg['import.skip_first_line']
            simlevel = self.sim_level_name
            initial_date = self.cfg['import.data_date']
            month_first = self.cfg.get('import.month_first')
            if month_first is None:
                month_first = False
            encodings = self.cfg.get('import.encodings')

            self.data_imp.import_data(format, data, separator=separator,
                            skip_first=skipfirst, sim_level=simlevel,
                            data_date=initial_date, month_first=month_first,
                            encodings=encodings)
            ids = self.data_imp.import_ids
            cont = not self.data_imp.terminate_import
            if not self.data_imp.terminate_import and forced_ops:
                self.op_imp.import_operations(ops, separator)
                cont = not self.op_imp.terminate_import
            return cont, ids
        finally:
            if not data[0].closed:
                data[0].close()
            if forced_ops and not ops.closed:
                ops.close()
            del self.data_imp
            self.data_imp = None

    def get_data_importer(self, definition):
        max_errors = self.cfg['error_handling.max_error_count_total']
        return self._load_data_import(definition, max_errors)

    def get_op_importer(self, text2operation, operation2modelchains=None,
                        oformat='text'):
        return self._load_op_importer(text2operation, operation2modelchains,
                                      oformat)

    def _load_data_import(self, datamapping_def, max_errors):
        """
        Create a DataImporter and return it.
        """
        max_errors = self.cfg['error_handling.max_error_count_total']
        import_date = self.cfg['import.import_date']
        basedir = os.getcwd()
        sim_dir = self.cfg['program_execution.base_folder']
        os.chdir(sim_dir)
        try:
            dm = self.struct_db.get_obj(sc.TEXT2DATA_DEF, datamapping_def)
            if dm is None:
                msg = "Data conversion definition '%s' not in SIMO DB" \
                      % datamapping_def
                raise Exception(msg)
            data_imp = DataImporter(self.input_db, dm, import_date,
                                    self.logger, 'data import', self.lx,
                                    max_errors)
            return data_imp
        finally:
            os.chdir(basedir)

    def _load_op_importer(self, text2operation=None,
                          operation2modelchains=None, oformat='text'):
        log_name = 'data import'
        max_errors = self.cfg['error_handling.max_error_count_total']
        basedir = os.getcwd()
        sim_dir = self.cfg['program_execution.base_folder']
        os.chdir(sim_dir)
        try:
            conv_def, op2mc = None, None
            if oformat in ('text',):
                conv_def = self.struct_db.get_obj(sc.TEXT2OPERATION_DEF,
                                                  text2operation)
                if conv_def is None:
                    errormsg = "Operation mapping definition '%s' couldn't "\
                               "be loaded"  % text2operation
                    raise Exception(errormsg)
            elif oformat in ('xml', 'db'):
                op2mc = self.struct_db.get_obj(sc.OPERATION2MODELCHAINS_DEF,
                                               operation2modelchains)
            return OperationImporter(self.input_db, self.fops_db,
                                     oformat, conv_def, op2mc,
                                     self.logger, log_name, max_errors)
        finally:
            os.chdir(basedir)


    def init_data_import(self):
        """
        Loads the data conversion mapping, operation conversion mapping and
        operation to model chains conversion mapping from SIMO DB
        """
        max_errors = self.cfg['error_handling.max_error_count_total']
        # Initialise data import
        dm_def = self.cfg['import.mapping_definition']
        self.data_imp = self._load_data_import(dm_def, max_errors)


    def simulate(self):
        """
        Constructs the simulator according to the configuration and simulation
        definitions and runs the simulation for the data in the input db.

        Simulation definitions are collected into SimInit object, which is then
        used to initialize the simulator object.
        """
        simtype = self.cfg['simulation.task_type']
        logname = 'simulation'
        msg = 'Initializing simulator...'
        self.logger.log_message(logname, 'info', msg)
        if self.ctrl is None:
            errormsg = "Simulation control definition '%s' couldn't be "\
                       "loaded"  % self.cfg['simulation.simulation_control']
            self.logger.log_message(logname, 'error', errormsg)
            return False
        # get the error handling settings
        max_errors = self.cfg['error_handling.max_error_count_total']
        max_simunit_errors = \
                self.cfg['error_handling.max_error_count_per_unit']
        # Initialise the simulator with the simulation definitions and DBs
        self.ctrl.use_data_date = \
                self.cfg['simulation.use_data_date_as_init_date']
        self.ctrl.use_today = self.cfg['simulation.use_today_as_init_date']
        self.ctrl.initial_date = self.cfg['simulation.fixed_init_date']
        if simtype == 'data processing':
            opt = 'simulation.fix_data_processing_date_to_init'
            self.ctrl.fixed_dp_date = self.cfg[opt]
        warnings = self.cfg['warning_handling.output_warnings']
        objcount_multiplier = self.cfg['simulation.object_count_multiplier']
        branchcount = self.cfg['simulation.estimated_branch_count']
        max_unit_count = self.cfg['simulation.max_unit_count']
        noopbranch = self.cfg['simulation.leave_no_op_branch']
        separatebgs = self.cfg['simulation.separate_branching_groups']
        branchlimit = self.cfg['simulation.max_branch_limit']
        max_res_objs = self.cfg['simulation.max_result_unit_count']
        deterministic = self.cfg['simulation.deterministic']
        track_prices = self.cfg['simulation.track_prices']
        do_forced_ops = self.cfg['simulation.do_forced_ops']
        copy2iters = self.cfg['simulation.copy_ops_to_all_iterations']
        create_branch_desc = self.cfg['simulation.create_branch_desc']
        cache_dir = self.cfg['simulation.cache_dir']

        simulator = Simulator(self.logger, logname, simtype, self.lx,
                   self.ctrl, self.input_db, self.result_db, self.fops_db,
                   warnings, max_unit_count, objcount_multiplier,
                   branchcount, noopbranch, separatebgs, branchlimit, max_res_objs,
                   max_errors, max_simunit_errors, deterministic,
                   track_prices, do_forced_ops, copy2iters,
                   create_branch_desc, cache_dir)

        idlist = self.cfg.get('simulation.id_list')
        indseq = self.cfg.get('simulation.index_sequence')
        iterations = self.cfg.get('simulation.iterations')
        run_profiler = self.cfg['program_execution.code_profiling']
        try:
            if not run_profiler:
                simulator.simulate(iterations, idlist, indseq)
            else:
                self._profile_sim(simulator, iterations, idlist, indseq)
        finally:
            simulator.cleanup()
        cont = not simulator.terminate

        if not self.cfg.get('program_execution.no_exec_simulate'):
            return cont
        else:
            result = self.result_db.db.store, self.log_db.db.store
            self.result_db.db.clear_store()
            self.log_db.db.clear_store()
            return result

def do_import(runner_obj):
    """
    Run import on the database _if_ there is nothing imported yet.
    """
    basedir = os.getcwd()
    try:
        try:
            os.chdir(runner_obj.cfg['program_execution.base_folder'])
            add = runner_obj.input_db._add
            rows = runner_obj.input_db.db.get('SELECT count(*) '\
                                              'FROM %scomp_unit' % (add,))
            if not rows[0][0]:
                runner_obj.logger.log_message('celery db building', 'info',
                                          'No pre-existing import, importing.')
                return runner_obj.import_data()
            else:
                runner_obj.logger.log_message('celery db building', 'info',
                                              'Pre-existing import, skipping.')
                return True
        except:
            traceback.print_exc()
        return False
    finally:
        os.chdir(basedir)

def do_memory_import(runner_obj, data_strs, op_str):
    """
    Run import on the database _if_ there is nothing imported yet.

    data_strs -- list of data strings to import
    op_str -- operation data string to import
    """
    basedir = os.getcwd()
    try:
        try:
            os.chdir(runner_obj.cfg['program_execution.base_folder'])
            return runner_obj.import_data_from_strings(data_strs, op_str)
        except:
            traceback.print_exc()
    finally:
        os.chdir(basedir)

def get_runner(initial_setup, inifile, db_info, do_wipe=False,
               no_exec_import=False, no_exec_simulate=False,
               no_exec_optimize=False, no_exec_logger=False, sim_dir=None,
               replace_cfg=None, memory_import=False, fully_in_memory=False):
    inilogger = runner._get_ini_logger()
    cfg = None
    if initial_setup or fully_in_memory:
        inifile = _get_ini_file(inilogger, inifile)
        cfg = runner._parse_ini_file(inilogger, inifile,
                                     replace_cfg=replace_cfg)
    runner_obj = CeleryRunner(initial_setup, inilogger, cfg, db_info,
                              fully_in_memory)
    if not do_wipe:
        runner_obj.cfg['program_execution.no_wipe_import'] = True
        runner_obj.cfg['program_execution.no_wipe_simulate'] = True
        runner_obj.cfg['program_execution.no_wipe_optimize'] = True
    else:
        runner_obj.cfg['program_execution.no_wipe_import'] = False
        runner_obj.cfg['program_execution.no_wipe_simulate'] = False
        runner_obj.cfg['program_execution.no_wipe_optimize'] = False
    if no_exec_import:
        runner_obj.cfg['program_execution.no_exec_import'] = True
    if no_exec_simulate:
        runner_obj.cfg['program_execution.no_exec_simulate'] = True
    if no_exec_optimize:
        runner_obj.cfg['program_execution.no_exec_optimize'] = True
    if no_exec_logger:
        runner_obj.cfg['program_execution.no_exec_logger'] = True
    runner_obj.cfg['program_execution.output_data'] = False
    runner_obj.cfg['logging.wipe_log'] = False
    basedir = os.getcwd()
    if not sim_dir:
        sim_dir = runner_obj.cfg['program_execution.base_folder']
    sim_dir = os.path.abspath(sim_dir)
    runner_obj.cfg['program_execution.base_folder'] = sim_dir
    os.chdir(sim_dir)
    try:
        runner_obj.open_simodb()
        msgttable, lexttable, lang = runner._get_translators(inilogger,
                                                             runner_obj)

        runner_obj.open_logger(msgttable, lexttable, lang)
        try:
            runner_obj.open_databases(inilogger, memory_import)
        except Exception, e:
            traceback.print_exc()
            runner_obj.terminate('Unable to open databases (%s)' % str(e))
        runner_obj.init_data_import()
    finally:
        os.chdir(basedir)
    runner_obj.reconnected = False
    return runner_obj

def run_ids(runner_obj, ids):
    runner_obj.cfg['simulation.id_list'] = ids
    return runner_obj.simulate()

def main():
    """
    Note: use celery_runner.py to runs utilising this module
    """
    pass

if __name__ == '__main__':
    main()
