#!/usr/bin/env python
# encoding: utf-8
"""
*text2data.py*

Process the data conversion XML definitions for data import

Copyright (c) 2008 Simosol Oy. Distributed under the GNU General Public
License 2.0.
"""
# these are for running the doctest from src/simo
#import sys
#sys.path.append('..')
from simo.builder.xmlobject import XmlObject

class ConversionMappingDef(XmlObject):
    """Data conversion definitions"""

    def __init__(self, typedef):
        super(ConversionMappingDef, self).__init__(typedef)
        self.from_module = 'conversion mapping'

    def xml_to_obj(self, xmlname, root, lexicon):
        """data conversion XML definition to object instance"""
        self.lexicon = lexicon
        ns = self.elem_namespace(root.tag)
        cm = ConversionMapping(ns, root, self)
        self.obj[xmlname] = cm
        self.lexicon = None

class ValueConv(object):
    """Value conversion definition for imported data values"""
    def __init__(self, varind, level, tovar, toval, convfact=None,
                 epochyear=None):
        self.var_ind = varind
        self.level = level
        self.to_var = tovar
        self.to_val = toval
        self.conv_fact = convfact
        self.epoch_year = epochyear

class Mapping(object):
    """Mapping definition for imported data."""
    def __init__(self, vartype, fromdt):
        self.var_type = vartype
        if vartype == 'categorical':
            self.map = {}
        elif vartype == 'text':
            self.map = []
        elif vartype == 'numerical':
            self.map = []
        elif vartype == 'date':
            self.map = []
        elif vartype == 'geometry':
            self.map = []
        self.from_data_type = fromdt

class LevelDef(object):
    """
    Data level definition for imported data.
    """
    def __init__(self, levelname, rowtypevalue, idpos, delim, linkidpos,
                 rowtypepos, datepos):
        self.level_name = levelname
        self.rowtype_value = rowtypevalue
        self.id_pos = idpos
        self.id_delimiter = delim
        self.linkid_pos = linkidpos
        self.rowtype_pos = rowtypepos
        self.new_object_row = min(rowtypevalue)
        self.date_pos = datepos

class ConversionMapping(object):
    """Data conversion definition for data import"""
    def __init__(self, ns, root, validator):
        super(ConversionMapping, self).__init__()
        self.validator = validator
        self.level_list = []
        self.none_val = []
        self.object_rejection = {}
        self.comment_prefix = None
        self.attributes = {}
        self.mapping = {}
        self.defaults = {}
        for elem in root:
            if elem.tag == ns+'data_levels':
                self._extract_levels(ns, elem[0], first_level=True)
                if self.level_list==[]:
                    self.validator = None
                    return
                self._check_rowtype_pos()
            elif elem.tag == ns+'none_value_indicator':
                nonevals = elem.text.split()
                for noneval in nonevals:
                    if noneval in ("''", '""'):
                        self.none_val.append('')
                    else:
                        self.none_val.append(noneval)
            elif elem.tag == ns+'object_rejection':
                self._set_rejection_criteria(elem, ns)
            elif elem.tag == ns+'comment_prefix':
                if elem.text is not None:
                    self.comment_prefix = elem.text.split()
            elif elem.tag == ns+'defaults':
                self._convert_defaults(elem, ns)
            else:
                self._convert_var(elem, ns)
        # we don't want to store these two for every object
        self.validator = None

    def _check_rowtype_pos(self):
        """sanity check for the rowtypepos definitions"""
        rowtypepos = self.level_list[0].rowtype_pos
        for i in range(1, len(self.level_list)):
            if self.level_list[i].rowtype_pos != rowtypepos:
                warning = 'Different row type indicator positions given for '\
                          'different data levels. This may cause data '\
                          'import errors'
                self.validator.add_warning(warning)

    def _convert_defaults(self, elem, ns):
        for var in elem:
            level = var.find(ns+'level').text
            name = var.find(ns+'name').text
            value = var.find(ns+'value').text
            level_found = False
            for leveldef in self.level_list:
                if leveldef.level_name == level:
                    level_found = True
                    break
            if not level_found:
                errormsg = "Invalid level for default '%s'" % (name,)
                self.validator.add_error(errormsg,)
                return
            varind = self.validator.variable_ind(level, name)
            try:
                self.defaults[level][name] = {'value':value, 'varind':varind}
            except KeyError, e:
                self.defaults[level] = {}
                self.defaults[level][name] = {'value':value, 'varind':varind}

    def _convert_var(self, elem, ns):
        """Construct mapping definitions for single variable"""
        fromname = elem.find(ns+'name').find(ns+'from').text  # NB! unused
        toname = elem.find(ns+'name').find(ns+'to').text
        rowtype = int(elem.find(ns+'row_type').text)
        try:
            self.mapping[rowtype]
        except:
            self.mapping[rowtype] = {}
        level = None
        for leveldef in self.level_list:
            if rowtype in leveldef.rowtype_value:
                level = leveldef.level_name
        # Check rowtype is valid and level name was found
        if level is None:
            errormsg = "invalid rowtype for variable '%s'" % (toname)
            self.validator.add_error(errormsg)
            return
        try:
            self.attributes[level]
        except:
            self.attributes[level] = []
        if not toname in self.attributes[level]:
            self.attributes[level].append(toname)
        rowpos = int(elem.find(ns+'row_position').text)
        # fromdt used when importing categorical values
        fromdt = elem.find(ns+'from_datatype').text
        if elem.find(ns+'none_to_value').text is not None:
            if fromdt == 'int':
                defval = int(float(elem.find(ns+'none_to_value').text))
            else:
                defval = float(elem.find(ns+'none_to_value').text)
        else:
            defval = None
        catvar = elem.find(ns+'categorical')
        if catvar is not None:
            vartype = 'categorical'
        else:
            numvar = elem.find(ns+'numerical')
            if numvar is not None:
                vartype = 'numerical'
            else:
                datevar = elem.find(ns+'date')
                if datevar is not None:
                    vartype = 'date'
                else:
                    geomvar = elem.find(ns+'geometry')
                    if geomvar is not None:
                        vartype = 'geometry'
                    else:
                        vartype = 'text'
        if vartype == 'geometry':
            varind = None
        else:
            varind = self.validator.variable_ind(level, toname)

        mappings, ind = self._get_mappings(rowtype, rowpos, vartype, fromdt)
        if vartype == 'categorical':
            self._create_categorical_mapping(catvar, ns, mappings[ind],
                           fromdt, defval, varind, level, fromname, toname)
        elif vartype == 'numerical':
            convfact = float(numvar[0].text)
            mappings[ind].map.append(\
                (ValueConv(varind, level, toname, defval, convfact)))
        elif vartype == 'date':
            # The epoch year is either 'current' or 'gregorian'. This is used
            # when converting a date into ordinal date
            epochyear = datevar[0].text
            if not epochyear in ['current', 'gregorian']:
                self.validator.add_error("Invalid epoch year '%s'" % epochyear)
                return
            mappings[ind].map.append(\
                (ValueConv(varind, level, toname, defval, None, epochyear)))
        elif vartype == 'text':
            mappings[ind].map.append(\
                (ValueConv(varind, level, toname, defval)))
        elif vartype == 'geometry':
            mappings[ind].map.append(\
                (ValueConv(varind, level, toname, defval)))

    def _extract_levels(self, ns, elem, first_level=False):
        """Parse level information from input conversion xml document"""
        # Get level name and check that level is defined in the lexicon
        levelname = elem.find(ns+'name').text
        levelind = self.validator.level_ind(levelname)
        if first_level and levelind!=1:
            errormsg = "Data level '%s' is not the child level of simulation "\
                    "level in lexicon" % levelname
            self.validator.add_error(errormsg)
        if levelind is None:
            return
        # Get level row positions, id positions and linking id positions and
        # construct level definition
        idpos, delimiter = self._extract_id_const(ns,
                                                  elem.find(ns+'id_rowpos'))
        lirp = elem.find(ns+'link_id_rowpos')
        if lirp is not None:
            linkidpos = self._parse_linkid(ns, lirp, first_level)
            rowtypepos = None
        else:
            rowtypepos = int(elem.find(ns+'rowtype_rowpos').text)
            linkidpos = None
        rtv = elem.find(ns+'rowtype_value')
        rowtypevalue = [int(num) for num in rtv.text.split()]
        drpos = elem.find(ns+'date_rowpos')
        if drpos is not None:
            date_pos = int(drpos.text)
        else:
            date_pos = None
        if len(idpos)==1 and linkidpos is not None and idpos[0]==linkidpos[1]:
            errormsg = "Link id position (%d) is the same as data id "\
                       "position. This conflict will result in "\
                       "import errors and data corruption" % (linkidpos[1],)
            self.validator.add_error(errormsg)
        leveldef = LevelDef(levelname, rowtypevalue, idpos, delimiter,
                            linkidpos, rowtypepos, date_pos)
        self.level_list.append(leveldef)
        sublevel = elem.find(ns+'sublevel')
        if sublevel is not None:
            self._extract_levels(ns, sublevel)

    def _set_rejection_criteria(self, elem, ns):
        """Set criteria for rejecting input for an object"""
        # the rejection variable level must be the top level
        rlevel = self.level_list[0].level_name
        rejectvars = elem.findall(ns+'SIMO_variable')
        for var in rejectvars:
            rvar = var.find(ns+'name').text
            varind = self.validator.variable_ind(rlevel, rvar)
            if varind!=(None, None):
                # Objects can be either rejected by rules, similar to condition
                # elements in SIMO model chains, but much more simplistic
                rcrit = var.find(ns+'reject_criterion')
                oper = rcrit.attrib.get('oper')
                rval = rcrit.find(ns+'value')
                renum = rcrit.find(ns+'enum')
                if rval is not None:
                    crit = float(rval.text)
                elif renum is not None:
                    crit = self._format_rejection_criterium(renum, oper)
                if crit is not None:
                    rejection = {'oper': oper, 'criteria': crit}
                    if rvar in self.object_rejection:
                        self.object_rejection[rvar].append(rejection)
                    else:
                        self.object_rejection[rvar] = [rejection]

    def _format_rejection_criterium(self, renum, oper):
        srenum = renum.text.split()
        if oper=='in':
            crit = [int(x) for x in srenum]
        else:
            if len(srenum)>1:
                msg = 'Data conversion definition: error converting rejection '\
                      'criteria %s for operator %s'  % (renum.text, oper)
                self.validator.add_error(msg)
                crit = None
            else:
                crit = float(srenum[0])
        return crit

    def _create_categorical_mapping(self, elem, ns, mapping, fromdt, defval,
                                    varind, level, fromname, toname):
        """import data value to SIMO data value mappings"""
        valid_nonevals = set()
        for noneval in self.none_val:
            try:
                int(noneval)
                if fromdt == 'int':
                    valid_nonevals.add(int(noneval))
                    continue
            except ValueError:
                if fromdt == 'int':
                    continue
            try:
                float(noneval)
                if fromdt == 'double':
                    valid_nonevals.add(float(noneval))
                    continue
            except ValueError:
                if fromdt == 'double':
                    continue
            valid_nonevals.add(noneval)

        for noneval in valid_nonevals:
            try:
                mapping.map[noneval]
            except:
                mapping.map[noneval] = []
            mapping.map[noneval].append(ValueConv(varind, level, toname,
                                                  defval))
        for m in elem.getiterator(ns+'value'):
            # do not split string when fromdt is 'string' in order to use
            # strings with space characters (eg. for tree species names, place
            # names and such)
            if fromdt == 'string':
                fromvalstr = [m.find(ns+'from').text.strip()]
            else:
                fromvalstr = m.find(ns+'from').text.strip().split(' ')
            try:
                if fromdt == 'int':
                    fromval = [int(val) for val in fromvalstr]
                elif fromdt == 'double':
                    fromval = [float(val) for val in fromvalstr]
                else:
                    fromval = fromvalstr
            except ValueError:
                msg = 'Data conversion definition: error converting ' \
                      'values %s for variable %s to data type %s' \
                      % (fromvalstr, fromname, fromdt)
                self.validator.add_error(msg)
                return
            toval = int(m.find(ns+'to').text)
            for fval in fromval:
                try:
                    mapping.map[fval]
                except:
                    mapping.map[fval] = []
                mapping.map[fval].append(\
                                     (ValueConv(varind, level, toname, toval)))

    def _extract_id_const(self, ns, elem):
        """
        Parse level id_rowpos element information. This information is used
        for constructing object ids
        """
        posstr = elem.find(ns+'pos')
        posindx = map(lambda x: int(x), posstr.text.split(' '))
        delimiter = elem.find(ns+'delimiter')
        if delimiter is not None:
            delimiter = delimiter.text
        return posindx, delimiter

    def _get_mappings(self, rowtype, rowpos, vartype, fromdt):
        """
        Get the variable mapping definition for the given row type, position
        and variable type

        rowtype -- row type
        rowpos -- position of the variable on the row
        vartype -- variable type (categorical, numerical, date, text, geometry)
        """
        if not rowtype in self.mapping:
            self.mapping[rowtype] = {}
        if not rowpos in self.mapping[rowtype]:
            self.mapping[rowtype][rowpos] = []
        mappings = self.mapping[rowtype][rowpos]
        found = False
        for ind in range(len(mappings)):
            if mappings[ind].var_type == vartype:
                found = True
                break
        if not found:
            ind = -1
            mappings.append(Mapping(vartype, fromdt))
        return mappings, ind

    def _parse_linkid(self, ns, elem, first_level):
        """
        Parse link_id_rowpos element into link structure which links individual
        objects to a single top level object.
        For first level this is not needed.
        """
        if first_level:
            # top level object, no link id defined
            return None
        # find the parent level position in the levellist
        for level in self.level_list:
            if level.level_name==elem.find(ns+'level').text:
                break
        link = (level.level_name, int(elem.find(ns+'pos').text))
        return link

def main():
    import doctest
    doctest.testfile('./test/text2data-py.rst')

if __name__ == '__main__':
        main()
