#!/usr/bin/env python
# encoding: utf-8
"""
*OperationConversion.py*

Process the data conversion XML definitions for data import

Copyright (c) 2008 Simosol Oy. Distributed under the GNU General Public
License 2.0.
"""
# these are for running the doctest from src/simo
import sys
sys.path.append('..')
from lxml import etree
from simo.builder.xmlobject import XmlObject
from simo.builder.importers.text2data import Mapping, ValueConv

class OperationConversionDef(XmlObject):
    """Data conversion definitions"""

    def __init__(self, typedef):
        super(OperationConversionDef, self).__init__(typedef)
        self.from_module = 'operation conversion'

    def xml():
        doc = "dictionary of text representation of XML documents"
        def fget(self):
            return self._xml
        def fset(self, xmldata):
            """parse the xml text and validate against the schema, create
               the object representation, and set state"""
            self.errors = set()
            self.warnings = []
            if self._schema is not None:
                try:
                    xmlname, xmltext, lexicon, operationmapping = xmldata
                    self.xml_name = xmlname
                except ValueError:
                    errormsg = 'name, xml content and lexicon instance must '\
                               'be passed'
                    raise ValueError(errormsg)
                try:
                    schematext = self._schema
                    if self._typedef_schema is not None:
                        modschema = self._include_typedefs(schematext)
                    else:
                        modschema = schematext
                    schemaroot = etree.XML(modschema)
                    s = etree.XMLSchema(schemaroot)
                    parser = etree.XMLParser(schema=s, remove_comments=True)
                    root = etree.fromstring(xmltext, parser)
                except etree.XMLSyntaxError, e:
                    raise ValueError(e.args[0])
                else:
                    self._xml[xmlname] = xmltext
                    self.xml_to_obj(xmlname, root, lexicon, operationmapping)
                    if self.errors != set([]):
                        self._ok[xmlname] = False
                        self._all_ok = False
                        errormsg = "errors in xml to object conversion"
                        raise ValueError(errormsg)
                    else:
                        self._ok[xmlname] = True
                        self._all_ok = True
                        for xmlok in self._ok.values():
                            if not xmlok:
                                self._all_ok = False
            else:
                raise ValueError('No schema set')
        return locals()
    xml = property(**xml())

    def xml_to_obj(self, xmlname, root, lexicon, op2mc):
        """operation conversion XML definition to object instance"""
        self.lexicon = lexicon
        ns = self.elem_namespace(root.tag)
        cm = OperationConversion(ns, root, self, op2mc)
        self.obj[xmlname] = cm
        self.lexicon = None

class OperationConversion(object):
    """Data conversion definition for data import"""
    def __init__(self, ns, root, validator, op2mc):
        super(OperationConversion, self).__init__()
        self.validator = validator
        self.op2mc = op2mc
        self.errors = set([])
        self.none_val = []
        commentprefix = None
        typerowpos = None
        noopval = None  # 'no operation' indicator

        oplevel = root.find(ns + 'operation_level').text
        elem = root.find(ns + 'none_value_indicator')
        if elem is not None:
            nonevals = elem.text.split()
            for noneval in nonevals:
                if elem.text in ("''", '""'):
                    self.none_val.append('')
                else:
                    self.none_val.append(noneval)
        elem = root.find(ns + 'comment_prefix')
        if elem is not None:
            commentprefix = elem.text

        idpos, delimiter = self._extract_id_const(ns,
                                            root.find(ns+'id_row_position'))
        opid = None
        elem = root.find(ns + 'op_id_row_position')
        if elem is not None:
            opid = int(elem.text)
        typeelem = root.find(ns + 'operation_type')
        opelem = root.find(ns + 'operation_name')
        timeelem = root.find(ns + 'operation_timing')
        varelem = root.find(ns + 'variable_mapping')
        if typeelem is not None:
            typerowpos = int(typeelem[0].text)
        if opelem.find(ns+'no_operation_value') is not None:
            noopval = opelem.find(ns+'no_operation_value').text.split()
        oprowpos = int(opelem[0].text)
        timerowpos = int(timeelem[0].text)
        # initialize operation converter
        self.operation_conv = OperationConverter(\
            oplevel, idpos, delimiter, opid, typerowpos, oprowpos,
            timerowpos, noopval, self.none_val, commentprefix)
        # parse the operation type, name and timing conversions
        self._create_operation_types(ns, typeelem)
        self._convert_operations(ns, opelem.find(ns+'mapping'))
        self._create_operation_timing(ns, timeelem[1])
        if varelem is not None:
            self._map_operation_variables(ns, varelem)

        # we don't want to store these two for every object
        self.validator = None

    def _convert_operations(self, ns, elem):
        """
        Construct operation mappings.

        ns - xml namespace as string
        elem - operation_name element
        """
        self.operation_conv.operations = {}
        for e in elem.findall(ns+'operation'):
            type = e.find(ns + 'type')
            if type is not None:
                type = type.text
            fromname = e.find(ns+'from').text
            toname = e.find(ns+'to').text
            if self.operation_conv.types is not None:
                if not type in self.operation_conv.types:
                    errorstr = "Unknown operation type code '%s'" % type
                    self.validator.add_error(errorstr)
                    continue
            if toname not in self.op2mc.operation_mapping:
                errorstr = "Operation '%s' not defined in operation mapping" \
                         % toname
                self.validator.add_error(errorstr)
                continue
            params = None
            pelem = e.find(ns + 'parameters')
            if pelem is not None:
                params = self._parse_operation_params(ns, pelem)
            opdef = OperationDef(self.operation_conv.types[type], fromname,
                                 toname, params)
            self.operation_conv.operations[(type, fromname)] = opdef

    def _extract_id_const(self, ns, elem):
        """
        Parse level id_rowpos element information. This information is used
        for constructing object ids
        """
        posstr = elem.find(ns+'pos')
        posindx = map(lambda x: int(x), posstr.text.split(' '))
        delimiter = elem.find(ns+'delimiter')
        if delimiter is not None:
            delimiter = delimiter.text
        return posindx, delimiter

    def _parse_operation_params(self, ns, elem):
        """
        Parse forced operation parameters, eg. variables and their values
        associated with given forced operation.

        ns - xml namespace as string
        elem - parameters element
        """
        parameters = []
        for e in elem.findall(ns + 'parameter'):
            param = Parameter()
            var_name = e.find(ns + 'name').text
            level_name = e.find(ns + 'level').text
            var_ind = self.validator.variable_ind(level_name, var_name)
            if var_ind==(None, None):
                errorstr = "Variable '%s' not found at level '%s' in "\
                           "lexicon" % (var_name, level_name)
                self.validator.add_error(errorstr)
                continue
            param.variable = var_name
            param.level = level_name
            param.variable_ind = var_ind
            param.value = float(e.find(ns + 'value').text)
            parameters.append(param)
        return parameters


    def _create_operation_timing(self, ns, elem):
        """
        Construct operation timing conversion definitions.

        ns -- xml namespace as string
        elem -- 'operation_timing' element
        """
        if elem.tag == ns+'date':
            timing = OperationTiming('date')
        elif elem.tag == ns+'time_step':
            timing = OperationTiming('step')
            timing.step_unit = elem[0].text
            timing.step_conversion = {}
            for e in elem.find(ns+'mapping'):
                fromval = e.find(ns+'from').text
                tostep = int(e.find(ns+'to').text) - 1
                timing.step_conversion[fromval] = tostep
        self.operation_conv.timing = timing

    def _create_operation_types(self, ns, typeelem):
        """
        Construct optional operation type mappings.

        ns -- xml namespace as string
        elem -- operation_type element
        """
        self.operation_conv.types = {None: None}
        if typeelem is not None:
            elem = typeelem.find(ns+'mapping')
            for e in elem.findall(ns+'type'):
                fromtype = e[0].text.split(' ')
                totype = e[1].text
                for a in fromtype:
                    self.operation_conv.types[a] = totype

    def _map_operation_variables(self, ns, elem):
        """
        Create mappings for operation variable conversions.

        ns -- xml namespace as string
        elem -- operation variables element
        """
        self.operation_conv.variables = {}
        for e in elem.findall(ns + 'variable'):
            self._convert_var(ns, e)

    def _convert_var(self, ns, elem):
        """Construct mapping definitions for single variable"""
        variables = self.operation_conv.variables
        fromname = elem.find(ns+'name').find(ns+'from').text
        toname = elem.find(ns+'name').find(ns+'to').text
        level = elem.find(ns+'level').text
        varind = self.validator.variable_ind(level, toname)
        rowpos = int(elem.find(ns+'row_position').text)
        # fromdt used when importing categorical values
        fromdt = elem.find(ns+'from_datatype').text
        if elem.find(ns+'none_to_value').text is not None:
            defval = int(elem.find(ns+'none_to_value').text)
        else:
            defval = None
        catvar = elem.find(ns+'categorical')
        numvar = elem.find(ns+'numerical')
        datevar = elem.find(ns+'date')
        if catvar is not None:
            self._create_categorical_mapping(catvar, ns, rowpos, fromdt,
                                         fromname, defval, varind, level, toname)
        elif numvar is not None:
            if rowpos not in variables:
                variables[rowpos] = Mapping('numerical', fromdt)
            convfact = float(numvar[0].text)
            variables[rowpos].map.append(\
                ValueConv(varind, level, toname, defval, convfact))
        elif datevar is not None:
            # The epoch year is either 'current' or 'gregorian'. This is used
            # when converting a date into ordinal date
            epochyear = datevar[0].text
            if not epochyear in ['current', 'gregorian']:
                self.validator.add_error("Invalid epoch year '%s'" % epochyear)
                return
            if rowpos not in variables:
                variables[rowpos] = Mapping('date', fromdt)
            variables[rowpos].map.append(\
                ValueConv(varind, level, toname, defval, None, epochyear))

    def _create_categorical_mapping(self, elem, ns, rowpos, fromdt, fromname,
                                    defval, varind, level, toname):
        """import data value to SIMO data value mappings"""
        variables = self.operation_conv.variables
        if rowpos not in variables:
            variables[rowpos] = Mapping('categorical', fromdt)
        if defval is not None:
            for noneval in self.none_val:
                try:
                    variables[rowpos].map[noneval]
                except:
                    variables[rowpos].map[noneval] = []
                variables[rowpos].map[noneval].append(\
                    ValueConv(varind, level, toname, defval))
        for m in elem.getiterator(ns+'value'):
            fromvalstr = m.find(ns+'from').text.strip().split(' ')
            try:
                if fromdt == 'int':
                    fromval = [int(val) for val in fromvalstr]
                elif fromdt == 'double':
                    fromval = [float(val) for val in fromvalstr]
                else:
                    fromval = fromvalstr
            except ValueError:
                msg = 'Data conversion definition: error converting ' \
                      'values %s for variable %s to data type %s' \
                      % (fromvalstr, fromname, fromdt)
                self.validator.add_error(msg)
            toval = int(m.find(ns+'to').text)
            for fval in fromval:
                try:
                    variables[rowpos].map[fval]
                except:
                    variables[rowpos].map[fval] = []
                variables[rowpos].map[fval].append(\
                                     (ValueConv(varind, level, toname, toval)))

class OperationConverter:
    """Forced operation import conversion definition."""
    def __init__(self, oplevel, idpos, delim, opid, optype, op, timing,
                 noopval=None, noneval=None, commentprefix=None):
        """
        Initialize operation converter.

        oplevel -- operation level as sstring
        oid -- target unit id row position as int
        opid -- operation id row position as int
        optype -- operation type row position as int
        op -- operation definition row position as int
        timing -- timing indicator row position as int
        noopval -- 'no operation' indicator value
        noneval -- none value indicator as string
        commentefix -- text file comment line prefix as string
        """
        self.operation_level = oplevel
        self.no_operation_indicator = noopval
        self.none_value_indicator = noneval
        self.comment_prefix = commentprefix
        self.id_rowpos = idpos
        self.id_delim = delim
        self.op_id_rowpos = opid
        self.type_rowpos = optype
        self.operation_rowpos = op
        self.timing_rowpos = timing
        self.types = None
        self.timing = None
        self.variables = None

class Parameter:
    """Parameter definition"""
    def __init__(self):
        self.variable_ind = (None, None)
        self.value = None
        self.variable = None
        self.level = None

class OperationDef:
    """Conversion definition for a single forced operation"""
    def __init__(self, type, fromname, toname, param=None):
        """
        Initialize operation conversion.

        type -- operation type as string
        fromname -- text file operation name as string
        toname -- SIMO operation name as string
        param -- operation parameters NB: Not implemented!
        """
        self.type = type
        self.from_name = fromname
        self.to_name = toname
        self.parameters = param

class OperationTiming:
    """Operation timing conversion definition."""
    def __init__(self, type):
        """
        Initialize operation timing conversion.

        type -- timing type as string, either 'date' or 'step'
        """
        self.type = type
        self.step_conversion = None
        self.step_unit = None

def main():
    import doctest
    doctest.testfile('test/text2operation-py.rst')

if __name__ == '__main__':
    main()

