#!/usr/bin/env python
# encoding: utf-8
"""
Title: SIMO lexicon

Classes for storing and handling simulation lexicon

*lexicon.py*

Copyright (c) 2007 Simosol Oy.
Distributed under the GNU General Public License 2.0.
"""
# these are for running the doctest from src/simo
# import sys
# sys.path.append('..')

from lxml import etree
from copy import deepcopy as copy
from simo.builder.xmlobject import XmlObject
from simo.builder.lexicon.lexiconlevel import LexiconLevel
from simo.builder import names


class LexiconDef(XmlObject):
    """Lexicon definition"""

    def __init__(self, typedef):
        super(LexiconDef, self).__init__(typedef)
        self.from_module = "lexicon"

    def xml():
        doc = "text representation of XML document"

        def fget(self):
            return self._xml

        def fset(self, xmldata):
            """parse the xml text and validate against the schema, create
               the object representation, and set state"""
            self.errors = set()
            self.warnings = []
            if self._schema is not None:
                try:
                    xmlname, xmltext = xmldata
                    self.xml_name = xmlname
                except ValueError:
                    errmsg = 'xml name and xml content must be passed'
                    raise ValueError(errmsg)
                try:
                    schematext = self._schema
                    if self._typedef_schema is not None:
                        modschema = self._include_typedefs(schematext)
                    else:
                        modschema = schematext
                    schemaroot = etree.XML(modschema)
                    s = etree.XMLSchema(schemaroot)
                    parser = etree.XMLParser(schema=s, remove_comments=True)
                    root = etree.fromstring(xmltext, parser)
                except etree.XMLSyntaxError, e:
                    raise ValueError(e.args[0])
                else:
                    self._xml[xmlname] = xmltext
                    self.xml_to_obj(xmlname, root)
                    self._ok[xmlname] = True
                    self._all_ok = True
                    for xmlok in self._ok.values():
                        if not xmlok:
                            self._all_ok = False
            else:
                raise ValueError('No schema set')
        return locals()
    xml = property(**xml())

    def xml_to_obj(self, xmlname, root):
        """lexicon XML definition to object instance"""
        ns = self.elem_namespace(root.tag)
        lex = Lexicon(ns, root, self)
        self.obj[xmlname] = lex


class Lexicon(object):

    """ Class for building and accessing SIMO lexicon """
    def __init__(self, ns, root, validator):
        """Parse a lexicon xml document into Lexicon object"""
        super(Lexicon, self).__init__()
        self._level_var_length = {}
        self._variable_check_list = {}
        self._levels = {}
        self._variable_ind = {}
        self._level_name = {}
        self._variable_name = {}
        self._level_ind = {}
        self._operation_ind = {}
        self._hierarchy = {}
        self._model_check_list = {}
        self.from_module = 'lexicon'
        self._validator = validator
        self._ns = ns
        # parse root level and all sublevels with recursion
        self._parse_level(root, 0, 0, 0, None)
        # construct and update variable checklist
        for level, varind in self._variable_ind.iteritems():
            self._variable_check_list[level] = \
                dict([(x, False) for x in varind.keys()])
        # calculate the maximum number of variables on any level
        mx = max([len(x.keys()) for x in self._variable_check_list.values()])
        self._max_num_of_vars = mx
        self._validator = None

    def _parse_level(self, elem, levelind, ordinal, line, parent):
        """
        Parse a single level into lexicon
        """
        if elem is None:
            return
        level = LexiconLevel(self._ns, elem, self._validator)
        subelems = elem.find(self._ns + 'sublevels')
        if subelems is not None:
            n = len(subelems)
            lineage = range(line, line + n)
            children = []
        else:
            lineage = [line]
            children = None
        self._store_level(level, levelind, ordinal, lineage, parent, children)
        pind = levelind
        if subelems is not None:
            ordinal += 1
            for i in range(n):
                children.append(levelind+1)
                levelind = self._parse_level(subelems[i], levelind + 1,
                                             ordinal, lineage[i], pind)
        return levelind

    def _store_level(self, level, levelind, ordinal, lineage, parent,
                     children):
        """
        Store variable indices, level indices etc. from a single level
        """
        variables = list(level.numerical) + list(level.categorical) + \
            list(level.textual)
        self._levels[level.name] = level
        self._level_ind[level.name] = levelind
        self._variable_ind[level.name] = dict([(v, None) for v in variables])
        for v in variables:
            if len(v) > 30:
                self._validator.errors.add(v)
        self._level_name[levelind] = level.name
        self._level_var_length[levelind] = 0
        self._variable_name[levelind] = {}
        self.hierarchy[levelind] = {'ordinal': ordinal,
                                    'lineage': set(lineage),
                                    'parent': parent,
                                    'children': children}
        # make sure the lineage is set for the parent levels too
        self._set_lineage_for_parents(levelind)

    def add_model(self, mtype, mname):
        if mtype not in self._model_check_list:
            self._model_check_list[mtype] = set([])
        self._model_check_list[mtype].update([mname])
        if mtype == names.OPERATION:
            oi = self._operation_ind
            if mname not in oi:
                if len(oi) == 0:
                    ind = 0
                else:
                    ind = len(oi.values())
                self._operation_ind[mname] = ind
            else:
                ind = self._operation_ind[mname]
            return ind
        else:
            return None

    def get_content_def(self):
        """
        Returns a dictionary that has (level name, level index) tuples as keys
        and a list of (variable name, variable index) tuples as data.
        This is used by the DataDB to create needed database tables.
        """
        content_def = {}
        active_var = {}
        for key, val in self._level_ind.items():
            content_def[(key, val)] = []
            active_var[key] = []
        for key, val in self._variable_check_list.items():
            for item, active in val.items():
                if active:
                    vartuple = (item, self._variable_ind[key][item])
                    active_var[key].append(vartuple)
        for key in content_def.keys():
            content_def[key] = active_var[key[0]]
        # weed out levels with no active variables
        for k, v in content_def.items():
            if len(v) == 0:
                content_def.pop(k)
        return content_def

    def get_level_ind(self, level):
        """Validate level against lexicon"""
        if level not in self._level_ind:
            return None
        return self._level_ind[level]

    def get_operation_ind(self, operation):
        """Validate operation against lexicon"""
        if operation not in self._operation_ind:
            return None
        return self._operation_ind[operation]

    def get_variable_ind(self, level, variable, already_active=False):
        """Validate variable name and level against lexicon.
        Will create a new index if one doesn't exist already."""
        if level not in self._level_ind:
            return None, None
        if variable not in self._variable_ind[level]:
            return None, None
        if already_active:
            active = self._variable_check_list[level][variable]
            if not active:
                return None, None
        levind = self._level_ind[level]
        if self._variable_ind[level][variable] is None:
            varind = self._level_var_length[levind]
            self._variable_ind[level][variable] = varind
            self._variable_name[levind][varind] = variable
            self._level_var_length[levind] += 1
        varind = self._variable_ind[level][variable]
        self._variable_check_list[level][variable] = True
        return (levind, varind)

    def get_level_name(self, levelind):
        """Get level name"""
        if levelind not in self._level_name:
            return None
        return self._level_name[levelind]

    def get_variable_name(self, levelind, varind):
        """Get variable name"""
        if levelind not in self._level_name:
            return None
        if varind not in self._variable_name[levelind]:
            return None
        return self._variable_name[levelind][varind]

    def get_inactive_variables(self):
        """Get inactive variables"""
        inactive = {}
        for level, variables in self._variable_check_list.items():
            inactive[level] = set([])
            for variable, active in variables.items():
                if not active:
                    inactive[level].add(variable)
        return inactive

    def is_parent(self, parentcandidate, childcandidate):
        """Check whether parentcandidate is a parent level of childcandidate
        level"""
        if childcandidate == parentcandidate:
            return False
        child = self._hierarchy[childcandidate]
        parent = self._hierarchy[parentcandidate]
        return child['ordinal'] > parent['ordinal'] and \
            child['lineage'].issubset(parent['lineage'])

    def is_child(self, childcandidate, parentcandidate):
        """Check whether parentcandidate is a parent level of childcandidate
        level"""
        if childcandidate == parentcandidate:
            return False
        child = self._hierarchy[childcandidate]
        parent = self._hierarchy[parentcandidate]
        return child['ordinal'] > parent['ordinal'] and \
            child['lineage'].issubset(parent['lineage'])

    def reset_active_variables_and_models(self):
        """
        Sets the variable and model check lists globally to False
        """
        self._model_check_list = {}
        self._operation_ind = {}
        for level in self._variable_check_list:
            for variable in self._variable_check_list[level]:
                self._variable_check_list[level][variable] = False

    def _match_hierarchy(self, lexicon, mylevel, lexlevel):
        """Make sure that self and another lexicon are compatible. This
        requires that the level names and the parent-child relations should
        match between the two lexicons.

        lexicon -- lexicon instance to be merged to self
        mylevel -- current level indice for self
        lexlevel -- current level indice for lexicon
        """
        if not self.level_name[mylevel] == lexicon.level_name[lexlevel]:
            return False
        if lexicon._hierarchy[lexlevel]['children'] is None:
            return True
        for lexchild in lexicon._hierarchy[lexlevel]['children']:
            childname = lexicon.level_name[lexchild]
            # check that the child level exists in self
            if childname not in self.level_ind:
                return False
            mychild = self.level_ind[childname]
            # make sure that the current child level is really a child of
            # the current level in self
            if mychild not in self._hierarchy[mylevel]['children']:
                return False
            # process the child level recursively
            if not self._match_hierarchy(lexicon, mychild, lexchild):
                return False
        return True

    def merge(self, lexicon, logger):
        """Merge self with another lexicon. This way lexicons can be
        built from a master lexicon and a number of additional lexicons.

        lexicon -- Lexicon instance to be joined with self
        logger -- logger instance
        """
        # make sure that the hierarchies are compatible
        if not self._match_hierarchy(lexicon, 0, 0):
            errmsg = "mismatching data hierarchies, cannot merge lexicons"
            logger.log('error', errmsg)
            return False
        # walk through the levels in random order and copy variables
        # from the additional lexicon to self
        levels = lexicon.level_ind.keys()
        for level in levels:
            variables = lexicon.variable_ind[level].keys()
            for variable in variables:
                if variable in self._variable_ind[level]:
                    warnmsg = "variable '%s' defined in both master and "\
                              "extra lexicons; master lexicon definition "\
                              "will be overwritten." % variable
                    logger.log('warning', warnmsg)
                # set the variable name as a key in variable index
                self._variable_ind[level][variable] = None
                # copy the current variable information from lexicon to self
                self._levels[level].variables[variable] = \
                    copy(lexicon.levels[level].variables[variable])
                # variable type
                llevel = lexicon.levels[level]
                if variable in llevel.categorical:
                    if len(self._levels[level]._categorical) == 0:
                        self._levels[level]._categorical = set([variable])
                    else:
                        self._levels[level]._categorical.add(variable)
                elif variable in llevel.numerical:
                    if len(self._levels[level]._numerical) == 0:
                        self._levels[level]._numerical = set([variable])
                    else:
                        self._levels[level]._numerical.add(variable)
                elif variable in llevel.textual:
                    if len(self._levels[level]._textual) == 0:
                        self._levels[level]._textual = set([variable])
                    else:
                        self._levels[level]._textual.add(variable)
        return True

    def _set_lineage_for_parents(self, lind):
        """
        Checks that the lineage information for lind level is also set for the
        parent levels

        lind -- level index for which we do the check
        """
        parent = self.hierarchy[lind]['parent']
        lineage = self.hierarchy[lind]['lineage']
        if parent is not None:
            self.hierarchy[parent]['lineage'].update(lineage)
            self._set_lineage_for_parents(parent)

    def level_ind():
        doc = "Level indices keyd by level names"

        def fget(self):
            return self._level_ind
        return locals()
    level_ind = property(**level_ind())

    def operation_ind():
        doc = "Operation indices"

        def fget(self):
            return self._operation_ind
        return locals()
    operation_ind = property(**operation_ind())

    def variable_ind():
        doc = "Variable index structure"

        def fget(self):
            return self._variable_ind
        return locals()
    variable_ind = property(**variable_ind())

    def level_name():
        doc = "Level names structure"

        def fget(self):
            return self._level_name
        return locals()
    level_name = property(**level_name())

    def variable_name():
        doc = "Variable names structure"

        def fget(self):
            return self._variable_name
        return locals()
    variable_name = property(**variable_name())

    def hierarchy():
        doc = "Lexicon hierarchy definition"

        def fget(self):
            return self._hierarchy
        return locals()
    hierarchy = property(**hierarchy())

    def active_variables():
        doc = "Variable checklist"

        def fget(self):
            return self._variable_check_list
        return locals()
    active_variables = property(**active_variables())

    def max_num_of_vars():
        doc = "Maximum number of variables on any level"

        def fget(self):
            return max(self._level_var_length.values())
        return locals()
    max_num_of_vars = property(**max_num_of_vars())

    def models():
        doc = "Model checklist"

        def fget(self):
            return self._model_check_list
        return locals()
    models = property(**models())

    def levels():
        doc = "level definitions"

        def fget(self):
            return self._levels
        return locals()
    levels = property(**levels())


def main():
    import doctest
    doctest.testfile('test/lexicon-py.rst')

if __name__ == '__main__':
    main()
