#!/usr/bin/env python
# encoding: utf-8
"""
*aggregationmodel.py*

Process the aggregation model XML definitions

Copyright (c) 2008 Simosol Oy. Distributed under the GNU General Public
License 2.0.
"""

# these are for running the doctest from src/simo
#import sys
#sys.path.append('..')

from simo.builder.modelbase.model import Model
from simo.builder.modelchain.conditionparser import ConditionParser
from simo.builder import names
import numpy

SKIP = set(['', '\t', '\n'])

def txt2indices(strv, level, validator):
    """
    Split variable string into variable names and get indice for
    each variable
    
    Returns (True indicating scalar, (<level indice>, <variable indice>))
    if txt contains only a single variable, i.e. scalar
    
    Returns (False indicating vector, (<level indice>, <list of indices>))
    if txt contains multiple variables, i.e. variable vector)
    """
    varnames = [v for v in strv.split(' ') if v not in SKIP]
    if len(varnames) == 1:
        return True, validator.variable_ind(level, varnames[0])
    else:
        indices = []
        level_ind = validator.level_ind(level)
        for varname in varnames:
            il, iv = validator.variable_ind(level, varname)
            indices.append(iv)
        return False, (level_ind, indices)


class AggregationModel(Model):
    """
    Aggregation model definition
    """
    def __init__(self, ns, elem, validator, dirs):
        super(AggregationModel, self).__init__()
        self._type = names.AGGREGATION
        self._name = elem.find(ns+'name').text
        self._lib = elem.find(ns+'implemented_at').text
        self.vector_implementation = \
            bool(elem.find(ns+'vector_implementation').text)
        self._lang = 'python'
        self._dirs = dirs
        validator.add_model(self._type, self._name)


class AggregationModelParam(object):
    def __init__(self, ns, elem, task, model):
        """
        Extracts aggregation model parameters defined in the model chain
        """
        super(AggregationModelParam, self).__init__()
        self.target_type = None
        self.scalar = None
        self.target_ind = None
        self.through_level = None
        self.operands = []
        self.condition = None
        self.lowest_operand_level = None
        
        # get aggregation target level and if the level is different than
        # chain evaluation level, add a warning message
        levelem = elem.find(ns+'level')
        level = levelem.text if levelem is not None else task.chain_level
        if level != task.chain_level:
            modind = task.validator.level_ind(level)
            chainind = task.validator.level_ind(task.chain_level)
            is_parent = task.validator.lexicon.is_parent(modind, chainind)
            rel = 'parent' if is_parent else 'child'
            msg = "invalid target level '%s' - %s of evaluation level '%s' "\
                  "in task '%s'" % (level, rel, task.chain_level, task.path)
            task.validator.add_warning(msg)
        
        telem = elem.find(ns+'target')
        scalar, tin = txt2indices(telem.text, level, task.validator)
        self.scalar = scalar
        self.target_ind = tin 
        if not self.scalar and not model.vector_implementation:
            errormsg = "multiple targets not allowed for aggregation model "\
                "'%s', model lacks vector implementation!" % model.name
            task.validator.add_error(errormsg)
            return
        target_size = 1 if self.scalar else len(self.target_ind[1])
        thlelem = elem.find(ns+'through_level')
        if thlelem is not None:
            self.through_level = task.validator.level_ind(thlelem.text)
            
        # keep track of operand levels and the lowest operand level (the
        # lowest level is the level where values are aggregated from)
        oper_level = None
        lowest = None
        for operand in elem.find(ns+'operands'):
            oper = AggregationModelOperand(ns, operand, task.validator,
                                           self.scalar, target_size)
            if oper.type=='variable':
                if oper_level is None:
                    oper_level = oper.level
                if oper_level != oper.level:
                    error = True
                    # if the other operand level equals chain level and the
                    # other is it's parent level, it's ok
                    chainlevel = task.validator.level_ind(task.chain_level)
                    opl1 = task.validator.level_ind(oper_level)
                    opl2 = task.validator.level_ind(oper.level)
                    if opl1 < opl2 and oper.level == task.chain_level:
                        error = False
                    elif opl2 < opl1 and oper_level == task.chain_level:
                        error = False
                    if error:
                        errormsg = "aggregation operands at different levels"\
                                   " for task '%s'" % task.name
                        task.validator.add_error(errormsg)
                if lowest is None or oper.variable[0] > lowest:
                    lowest = oper.variable[0]
            self.operands.append(oper)
        # store the lowest operand level
        self.lowest_operand_level = lowest
        c = elem.find(ns+'condition')
        if c is not None:
            cp = ConditionParser(task.validator, task.chain_level)
            self.condition = cp.parse(c.text)

class AggregationModelOperand(object):
    def __init__(self, ns, elem, validator, scalar=True,
                 target_size=1):
        """
        Parse aggregation model operand definition in a model chain into an
        operand object
        """
        super(AggregationModelOperand, self).__init__()
        self.type = validator.elem_name(elem[0].tag)
        self.level = None
        self.variable = None
        self.weight = None
        self.default_value = None
        self.default_weight = None
        self.value = None
        if self.type == 'variable':
            self._process_variable_op(ns, elem, validator, scalar, target_size)
        else:
            self._process_value_op(ns, elem, validator, scalar, target_size)
    
    def _process_variable_op(self, ns, elem, validator, scalar, target_size):
        """
        Parse variable type aggregation operand
        """
        varelem = elem.find(ns+'variable')
        levelem = elem.find(ns+'level')
        self.level = levelem.text
        scalar_var, vin = txt2indices(varelem.text, self.level, validator)
        self.variable = vin
        if scalar_var != scalar:
            errormsg = "target variable and operands must be either "\
                       "both scalar or both vector!"
            validator.add_error(errormsg)
        if varelem.get('default') is not None:
            self.default_value = float(varelem.get('default'))
        welem = elem.find(ns+'weight')
        if welem is not None:
            scalar_weight, win = txt2indices(welem.text, self.level, validator)
            self.weight = win
            if scalar_weight != scalar:
                errormsg = "aggregation variable and weight must be "\
                           "either both scalar or both vector!"
                validator.add_error(errormsg)
            if welem.get('default') is not None:
                self.default_weight = float(welem.get('default'))
            elif self.default_value is not None:
                # if variable has a default value, weight should also
                # have a default value
                self.default_weight = 0.0
            if not scalar and len(self.weight[1]) != len(self.variable[1]):
                errormsg = "variable vector and weight vector must "\
                           "contain equal number of variables!"
                validator.add_error(errormsg)
        if not scalar and len(self.variable[1]) != target_size:
            errormsg = "unequal operand variable vector and target "\
                       "variable vector lengths!"
            validator.add_error(errormsg)
                
    def _process_value_op(self, ns, elem, validator, scalar, target_size):
        """
        Parse value type aggregation operand
        """
        try:
            items = elem[0].text.split(' ')
            vvector = [float(val) for val in items if val not in SKIP]
        except:
            msg = "invalid value vector '%s'" % elem[0].text
            validator.add_error(msg)
            return
        if len(vvector) == 1:
            scalar_value = True
            self.value = vvector[0]
        else:
            scalar_value = False
            self.value = numpy.array(vvector, dtype=float)
        if scalar_value != scalar:
            msg = "target variable and operands must be either both "\
                  "scalar or both vector!"
            validator.add_error(msg)
            return
        if not scalar and len(self.value) != target_size:
            msg = "unequal value vector and target variable vector lengths!"
            validator.add_error(msg)


def main():
    import doctest
    doctest.testfile('./test/aggregationmodel-py.rst')

if __name__ == '__main__':
    main()
