    #!/usr/bin/env python
# encoding: utf-8
"""
*cashflowtable.py*

Process the cashflow XML definitions

Copyright (c) 2008 Simosol Oy. Distributed under the GNU General Public
License 2.0.
"""

import re
import copy
import numpy

import simo.utils.utils as utils

from simo.builder.modelbase.table import table2array
from simo.builder.modelbase.table import table2dict
from simo.builder.modelbase.table import parse_classifiers
from simo.builder.modelbase.table import parse_tables
from simo.builder import names


class CashFlowTable(object):
    """Class for cash flow tables used in operations"""
    def __init__(self, ns, elem, validator, dirs):
        super(CashFlowTable, self).__init__()
        self._ns = ns
        self._validator = validator
        self.name = elem.get('name')
        tp = elem.get('logged')
        if tp=='true':
            self.persist = True
        else:
            self.persist = False
        self.type = names.CASH_FLOW_TABLE
        self.classifiers = None
        self.classifier_order = None
        self.db_cfiers = None
        self.db_cfier_ind = None
        self.keycount = None
        self.arrays = None
        self.Narrays = None
        self.active_array = None
        self.trends = None
        self._reduced_table = {}
        baseelem = elem.find(ns + 'base')
        self._parse_base(baseelem)
        # TODO: implement cash flow variables!
        #varelem = baseelem.find(ns + 'cash_flow_variable')
        trendelem = elem.find(ns + 'trends')
        if trendelem is not None:
            self.trends = [Trend(ns, elem, validator) for elem in trendelem]
        validator.add_model(names.CASH_FLOW_TABLE, self.name)
        self._validator = None

    def _parse_base(self, elem):
        """Parses the classifiers used in the cash flow table as well as the
        table data"""
        celem = elem.find(self._ns + 'classifiers')
        if celem is not None:
            res = parse_classifiers(self._ns, celem, self._validator)
            self.classifiers, self.db_cfiers, self.db_cfier_ind,\
                self.classifier_order, self.keycount = res
        telem = elem.find(self._ns + 'tables')
        self.arrays, self.Narrays = parse_tables(telem)

    def _get_trend(self, date):
        """Set currently active trend, None if no active trends"""
        if self.trends is None:
            return None
        for trend in self.trends:
            if trend.start <= date and date <= trend.end:
                return trend
        return None

    def _adjust_table(self, trend, date):
        """Adjust cash flow table values with the current trend"""
        order = trend.classifier_order
        arr = self.active_array.array
        Ntb = len(self.classifiers)
        where = numpy.where
        adjust = trend.adjust
        if order is not None:
            Ntr = len(trend.classifiers)
            for row in trend.factor_array:
                key, factor = row[:Ntr], row[Ntr]
                targets = arr[:,order] == key
                arr[targets,Ntb:] = adjust(date, arr[targets,Ntb:], factor)
        else:
            factor = trend.factor_array[0,0]
            arr[:,Ntb:] = trend.adjust(date, arr[:,Ntb:], factor)

    def set_active_array(self, date):
        """Set active array and adjust table values with trend"""
        if self.Narrays > 1:
            found = False
            for a_ind in range(len(self.arrays)):
                a = self.arrays[a_ind]
                if date > a.start_date and \
                    (a.end_date is None or date <= a.end_date):
                    self.active_array = self._deepcopy(a)
                    found = True
                    break
            if not found:
                # use the last value array
                a_ind = -1
                self.active_array = self._deepcopy(self.arrays[a_ind])
        else:
            # only one value array for the current table
            a_ind = 0
            self.active_array = self._deepcopy(self.arrays[a_ind])
        trend = self._get_trend(date)
        if trend is not None:
            self._adjust_table(trend, date)
            if a_ind in self._reduced_table:
                self._reduced_table.pop(a_ind)
        if self.persist and a_ind not in self._reduced_table:
            self._reduce_prices_for_db(a_ind)
        elif self.persist:
            self.active_reduced = self._reduced_table[a_ind]
        else:
            self.active_reduced = None

    def get_values(self, classifiers, labels):
        """Get values from the active array"""
        arr = self.active_array.array
        targets = numpy.all(arr[:,classifiers] == labels, 1)
        return arr[targets, -1]

    def set_values(self, classifiers, labels, values):
        """Set values of active array"""
        arr = self.active_array.array
        targets = numpy.all(arr[:,classifiers] == labels, 1)
        arr[targets, -1] = values
        if self.persist:
            self._reduce_prices_for_db()

    def _reduce_prices_for_db(self, a_ind=None):
        """
        Averages the table prices over the classifying variables that are
        found in the price db table
        """
        reduced = {}
        if a_ind is not None:
            arr = self.arrays[a_ind].array
        else:
            arr = self.active_array.array
        cfiers = set([tuple(row) for row in arr[:,self.db_cfier_ind]])
        for cfier in cfiers:
            target = numpy.all(arr[:, self.db_cfier_ind]==cfier, 1)
            reduced[cfier] = arr[target, -1].mean()
        if a_ind is not None:
            self._reduced_table[a_ind] = reduced
        self.active_reduced = reduced

    def _copy(self, data):
        return copy.copy(data)
    
    def _deepcopy(self, data):
        # NB: this deepcopy takes around 5-8% of the total simulation time!!!
        return copy.deepcopy(data)


class CashFlowTableParam(object):
    """docstring for CashFlowTableParam"""
    def __init__(self):
        super(CashFlowTableParam, self).__init__()


class Trend(object):
    def __init__(self, ns, elem, validator):
        """Cash flow table time trend class"""
        super(Trend, self).__init__()
        self.start = None
        self.end = None
        self.classifiers = None
        self.classifier_order = None
        self.keycount = None
        self.factor_array = None
        self.factor_dict = None
        self._parse_trend(ns, elem, validator)

    def _parse_trend(self, ns, elem, validator):
        """Parse trend element into Trend object"""
        pelem = elem.find(ns + 'time_period')
        if pelem is not None:
            self.start = utils.str2date(pelem[0].text)
            self.end = utils.str2date(pelem[1].text)
        celem = elem.find(ns + 'classifiers')
        if celem is not None:
            res = parse_classifiers(ns, celem, validator)
            self.classifiers, dummy, dummy, self.classifier_order,\
                self.keycount = res
        felem = elem.find(ns + 'cumulative_factors')
        self.factor_array = table2array(felem)
        self.factor_dict = table2dict(felem)

    def adjust(self, date, values, factor):
        """Interpolate values at given point in time
        """
        y_start, y_end = values, values + values * factor
        t_start, t_end = float(self.start.year), float(self.end.year)
        t = float(date.year)
        y_t = y_start + (y_end - y_start) * ((t - t_start) / (t_end - t_start))
        return y_t


def main():
    pass

if __name__ == '__main__':
    main()
