#!/usr/bin/env python
# encoding: utf-8
"""
*geotable.py*

Processes the geo table XML definitions, provides handling of geo tables.

Copyright (c) 2008 Simosol Oy. Distributed under the GNU General Public
License 2.0.
"""
# these are for running the doctest from src/simo
#import sys
#sys.path.append('..')

import os
import numpy
from simo.builder import names

class GeoTable(object):
    def __init__(self, ns, root, validator, dirs):
        """Construct geo table from the parsed geo table xml
        """
        super(GeoTable, self).__init__()
        self.validator = validator
        self.name = root.get('name')
        self.type = names.GEO_TABLE
        self.targets = []
        self.sim = None
        self.filename = root.find(ns+'filename').text
        self.ulx = float(root.find(ns+'ulx').text)
        self.uly = float(root.find(ns+'uly').text)
        self.nrow = int(root.find(ns+'nrow').text)
        self.ncol = int(root.find(ns+'ncol').text)
        self.xdim = float(root.find(ns+'xdim').text)
        self.ydim = float(root.find(ns+'ydim').text)
        self.nvar = int(root.find(ns+'nvar').text)
        cvar = root.find(ns+'coordinate_variables').find(ns+'longitude')
        self.longitude_var = self.__get_coord_var(ns, cvar)
        cvar = root.find(ns+'coordinate_variables').find(ns+'latitude')
        self.latitude_var = self.__get_coord_var(ns, cvar)
        for target in root.find(ns+'targets'):
            variable = target.find(ns+'variable').text
            level = target.find(ns+'level').text
            vardef = self.validator.variable_ind(level, variable)
            self.targets.append(vardef)
        if len(self.targets) != self.nvar:
            errormsg = "The number of target definitions (%s) doesn't "\
                       "match nvar definition (%s)"\
                       % (len(self.targets), self.nvar)
            self.validator.add_error(errormsg)
        if len(dirs) != 1:
            errormsg = "Only a single directory can be passed as array data "\
                       "directory"
            self.validator.add_error(errormsg)
        else:
            self.dirs = dirs
        validator.add_model(names.GEO_TABLE, self.name)
        # we don't want to store this for every object
        self.validator = None

    def load_function(self, validator):
        self.validator = validator
        self.__read_geo_array(os.path.expanduser( \
                              os.path.expandvars(self.dirs[0])))
        self.validator = None

    def __get_coord_var(self, ns, celem):
        """Coordinate variable definition to data matrix indices"""
        variable = celem.find(ns+'variable').text
        level = celem.find(ns+'level').text
        return self.validator.variable_ind(level, variable)

    def __read_geo_array(self, dirpath):
        """Reads the geo array data from disk into a numpy array"""
        afile = os.path.join(os.path.abspath(dirpath),
                             self.filename)
        try:
            self.arr = numpy.fromfile(afile, dtype=numpy.float32)
            self.arr.resize((self.nrow, self.ncol, self.nvar))
        except IOError:
            errormsg = "no such file or directory '%s'" % (afile)
            self.validator.add_error(errormsg)
            self.arr = None
        except ValueError:
            errormsg = "the size of the array and nrow, ncol and nvar "\
                       "settings don't match"
            self.validator.add_error(errormsg)
            self.arr = None


class GeoTableParam(object):
    def __init__(self, ns, elem, task, model):
        """Extracts geo table model parameters defined in the model chain"""
        super(GeoTableParam, self).__init__()
        self.result_variables = None
        self.get_all = None
        self.result_variables = []
        getallelem = elem.find(ns+'all_vars')
        if getallelem is not None:
            self.get_all = True
        else:
            self.get_all = False
            for resvar in elem.find(ns+'only_vars'):
                var = resvar.find(ns+'name').text
                level = resvar.find(ns+'level').text
                varind = task.validator.variable_ind(level, var)
                if varind in model.targets:
                    self.result_variables.append(varind)
                else:
                    errormsg = "variable '%s' at level '%s' is not a target "\
                               "variable for geo table '%s'"\
                               % (var, level, model.name)
                    task.validator.add_error(errormsg)

def main():
    import doctest
    doctest.testfile('./test/geotable-py.rst')

if __name__ == '__main__':
    main()
