#!/usr/bin/env python
# encoding: utf-8
"""
*managementmodel.py*

Process the management model XML definitions

Copyright (c) 2008 Simosol Oy. Distributed under the GNU General Public
License 2.0.
"""

# these are for running the doctest from src/simo
#import sys
#sys.path.append('..')

import re
from simo.builder.modelbase.model import Model
from simo.builder import names

class ManagementModel(Model):
    """Simo built-in data management model definition"""
    def __init__(self, ns, elem, validator, dirs):
        super(ManagementModel, self).__init__()
        self._validator = validator
        self._type = names.MANAGEMENT
        self._name = elem.find(ns+'name').text
        self._lib = elem.find(ns+'implemented_at').text
        self._lang = 'python'
        self._dirs = dirs
        self.parameters = []
        #Not all models have parameters, so we need to check
        if elem.find(ns+'parameters') is not None:
            for pelem in elem.find(ns+'parameters'):
                ptype = pelem.find(ns+'type').text
                desc = pelem.find(ns+'description').text
                self.parameters.append({'type': ptype, 'desc':desc})
        validator.add_model(self._type, self._name)
        self._validator = None


class ManagementModelParam(object):
    """Parameter value for management model, defined in modelchain"""
    def __init__(self, ns, elem, task, model):
        super(ManagementModelParam, self).__init__()
        self.parameters = []
        self._ns = ns
        self._validator = task.validator
        self._task = task
        self._model = model
        parelem = elem.find(ns+'parameters')
        if parelem is not None:
            for p in parelem:
                self._parse_param(p)
        if len(model.parameters) != len(self.parameters):
            errormsg = "wrong number of parameters defined for management "\
                    "model '%s' in task '%s'" % (model.name, task.name)
            self._validator.add_error(errormsg)
        else:
            for i in range(len(model.parameters)):
                self._check_param(i, model)
        self._validator = None
        self._task = None
        self._model = None

    def _parse_param(self, elem):
        """docstring for _parse_param"""
        if elem[0].tag == (self._ns + 'level'):
            levelind = self._validator.level_ind(elem[0].text)
            self.parameters.append({'type': 'level', 'level': levelind})
        elif elem[0].tag == (self._ns+'variable'):
            level = elem[0].find(self._ns + 'level').text
            var = elem[0].find(self._ns + 'name').text
            vardef = self._validator.variable_ind(level, var)
            self.parameters.append({'type': 'variable', 'vardef': vardef})
        elif elem[0].tag == (self._ns + 'value'):
            self.parameters.append({'type': 'value', 'value': elem[0].text})

    def _check_param(self, i, model):
        """docstring for _check_param"""
        mtype = model.parameters[i]['type']
        mctype = self.parameters[i]['type']
        if  mtype != mctype:
            errormsg = "parameter at position %s should be of type "\
                "'%s' instead of '%s' for management model '%s' in task '%s'"\
                       % (i, mtype, mctype, self._model.name, self._task.name)
            self._validator.add_error(errormsg)


def main():
    import doctest
    doctest.testfile('./test/managementmodel-py.rst')

if __name__ == '__main__':
    main()
