#!/usr/bin/env python
# encoding: utf-8
"""
Title: SIMO model super class

Super classes for different model classes in modelbase

*models.py*

Usage:

Copyright (c) 2007 Simosol Oy. Distributed under the GNU General Public License 2.0.

TODO:
"""
import os
import sys
import imp
import ctypes
import numpy


class Model(object):
    """Super class for all modelbase model classes"""
    def __init__(self):
        """Construct Model object"""
        super(Model, self).__init__()
        self._type = None
        self._name = None
        self._lang = None
        self._lib = None
        self._dirs = None
        self._v_mlib = None
        self._v_mwrap = None
        self._v_func = None

    def type():
        doc = "Get model type"
        def fget(self):
            return self._type
        return locals()
    type = property(**type())

    def name():
        doc = "Get model name"
        def fget(self):
            return self._name
        return locals()
    name = property(**name())

    def language():
        doc = "Get model language"
        def fget(self):
            return self._lang
        return locals()
    language = property(**language())

    def dirs():
        doc = "Get model directories"
        def fget(self):
            return self._dirs
        return locals()
    dirs = property(**dirs())

    def function():
        doc = "Get model function object"
        def fget(self):
            return self._v_func
        return locals()
    function = property(**function())

    def library():
        doc = "Get model library object"
        def fget(self):
            return self._v_mlib
        return locals()
    library = property(**library())

    def wrapper():
        doc = "Get model library wrapper object"
        def fget(self):
            return self._v_mwrap
        return locals()
    wrapper = property(**wrapper())

    def load_function(self, validator):
        """Load model function object from model library"""
        # check that all needed attributes are set
        if self._name is None or self._lib is None or self._dirs is None:
            errorstr = "Cannot load %s model function. Model name, library "\
                "name and library directory required!" % self._type
            validator.add_error(errorstr)
            return False
        if self._load_library(validator):
            return self._load_function_object(validator)

    def _load_c_library(self, library, validator):
        try:
            mlib = ctypes.CDLL(os.path.join(os.getcwd(), self._lib))
        except ImportError, errorstr:
            errormsg = 'Library %s not found: %s' % (library, errorstr)
            validator.add_error(errormsg)
            mlib = None
        except OSError, errorstr:
            errormsg = "Library depency error: '%s': %s" \
                       % (library, errorstr)
            validator.add_error(errormsg)
            mlib = None
        wrappername = 'wrapper_' + self._lib.split('.')[0]
        try:
            # Changed to work with Python 2.5
            modfile, modpathname, moddesc = imp.find_module(wrappername)
            mwrap = imp.load_module(wrappername, modfile,
                                    modpathname, moddesc)
        except ImportError:
            validator.add_error('Wrapper module for the library '\
                                '%s not found' % (library))
            mwrap = None
        return mlib, mwrap

    def _load_python_library(self, library, validator):
        # Changed to work with Python 2.5
        pylib = self._lib.split('.')[0]
        try:
            m_file, m_path, m_desc = imp.find_module(pylib)
            mlib = imp.load_module(pylib, m_file, m_path, m_desc)
        except ImportError, errorstr:
            errormsg = 'Library %s not found: %s' % (library, errorstr)
            validator.add_error(errormsg)
            mlib = None
        except OSError, errorstr:
            errormsg = "Library depency error: '%s': %s" % (library, errorstr)
            validator.add_error(errormsg)
            mlib = None
        finally:
            if m_file:
                m_file.close()
        # no need for the wrapper when using Python
        if mlib is not None:
            mwrap = mlib
        else:
            mwrap = None
        return mlib, mwrap

    def _load_library(self, validator):
        """Load the model library and wrapper module from the given path."""
        dirs = self._dirs
        if not isinstance(self._dirs, list):
            dirs = [self._dirs]
        lib_exists = False
        for d in dirs:
            d = os.path.abspath(os.path.expanduser(os.path.expandvars(d)))
            # go through possible model library directories and try to find
            # the library file
            # get all the files from the current directory and try to match
            # library name to the file names
            files = set([s.split('.')[0] for s in os.listdir(d)])
            if self._lib.split('.')[0] in files:
                library = os.path.join(d, self._lib)
                lib_exists = True
                break
        if not lib_exists:
            validator.add_error('Library %s not found' % self._lib)
            return False
        if not os.path.abspath(d) in sys.path:
            sys.path.append(os.path.abspath(d))
        lang = self._lang.lower()
        curdir = os.path.abspath(os.curdir)
        # Change the current working directory to the library path
        os.chdir(d)
        if lang == 'c':
            mlib, mwrap = self._load_c_library(library, validator)
        elif lang == 'python':
            mlib, mwrap = self._load_python_library(library, validator)
        # Change the working dir back to curdir
        os.chdir(curdir)
        self._v_mlib = mlib
        self._v_mwrap = mwrap
        return True

    def _load_function_object(self, validator):
        """Load model function object from model wrapper."""
        if self._v_mwrap is None:
            return False
        mwrap = self._v_mwrap
        func = None
        code = "func = mwrap." + self._name
        try:
            exec code
        except NameError:
            validator.add_error('%s %s is not defined in the %s library '\
                    'wrapper %s' % (self._type, self._name, self._type,\
                                mwrap.__name__))
        except AttributeError:
            validator.add_error('%s %s is not defined in the %s library '\
                    'wrapper %s' % (self._type, self._name, self._type,\
                                mwrap.__name__))
        self._v_func = func
        return True


class POModel(Model):
    """Super class for prediction and operation models"""
    def __init__(self, ns, elem, validator, dirs):
        """docstring for __init__"""
        super(POModel, self).__init__()
        self.validator = validator
        self.n_vars = None
        self.vars = None
        self.n_params = 0
        self.param_names = None
        self._parse_variables(ns, elem)
        self.validator = None

    def _parse_variables(self, ns, elem):
        """docstring for __parse_variables"""
        varelem = elem.find(ns+'variables')
        if varelem is None:
            return
        self.vars = {}
        self.n_vars = 0
        for e in varelem.findall(ns+"variable"):
            name = e.find(ns+'name').text
            level = e.find(ns+'level').text
            limits = None
            le = e.find(ns+'limits')
            if le is not None:
                limits = Limit(ns, le)
            levind, varind = self.validator.variable_ind(level, name)
            if levind is None or varind is None:
                continue
            if not levind in self.vars:
                self.vars[levind] = {'index': [], 'order': [], 'limits': []}
            self.vars[levind]['index'].append(varind)
            self.vars[levind]['order'].append(self.n_vars)
            self.vars[levind]['limits'].append(limits)
            self.n_vars+= 1
        for key in self.vars.keys():
            item = self.vars[key]['index']
            self.vars[key]['index'] = numpy.array(item, dtype=int)


class Limit(object):
    """Class for handling model variable limit"""
    def __init__(self, ns, elem):
        """Construct variable/parameter limits from XML element"""
        super(Limit, self).__init__()
        self.lower = None
        self.upper = None
        self.enum = None
        lower = elem.find(ns+"lower_limit")
        upper = elem.find(ns+"upper_limit")
        enum = elem.find(ns+"enum")
        if lower is not None:
            self.lower = float(lower.text)
        if upper is not None:
            self.upper = float(upper.text)
        if enum is not None:
            self.enum = [float(i) for i in enum.text.split()]

    def check(self, values):
        """Check if a value (or an array of values) is inside limits"""
        if self.enum is not None:
            result = numpy.array(map(lambda x: x in self.enum, values))
        elif self.lower is not None and self.upper is not None:
            result = (self.lower <= values) * (values <= self.upper)
        elif self.lower is not None and self.upper is None:
            result = self.lower <= values
        else:
            result = values <= self.upper
        return result

    def limitstr(self, variable):
        """Return a string representation of the limits"""
        if self.enum is not None:
            limitstr = "'%s' @ %s" % (variable, str(self.enum))
        elif self.lower is not None and self.upper is not None:
            limitstr = "%s <= '%s' <= %s" % (self.lower, variable, self.upper)
        elif self.lower is not None and self.upper is None:
            limitstr = "%s <= '%s'" % (self.lower, variable)
        else:
            limitstr = "%s <= '%s'" % (variable, self.upper)
        return limitstr


def main():
    pass

if __name__ == '__main__':
    main()
