#!/usr/bin/env python
# encoding: utf-8
"""
*modelbase.py*

Provides a central repository for all models: definitions and implementations
alike

Copyright (c) 2008 Simosol Oy. Distributed under the GNU General Public
License 2.0.
"""

import re
from lxml import etree

from simo.builder.xmlobject import XmlObject
from simo.builder.modelbase.aggregationmodel import AggregationModel
from simo.builder.modelbase.geotable import GeoTable
from simo.builder.modelbase.cashflowmodel import CashFlowModel
from simo.builder.modelbase.cashflowtable import CashFlowTable
from simo.builder.modelbase.managementmodel import ManagementModel
from simo.builder.modelbase.operationmodel import OperationModel
from simo.builder.modelbase.parametertable import ParameterTable
from simo.builder.modelbase.predictionmodel import PredictionModel

class ModelbaseDef(XmlObject):
    """Repository of all kind of models used in SIMO"""

    def __init__(self, typedef):
        super(ModelbaseDef, self).__init__(typedef)
        self._schema = {}
        self._xml = {}
        self.obj = {}
        self._ok = {}

    def schema():
        doc = "dictionary of text representations of XML Schema documents"
        def fget(self):
            return self._schema
        def fset(self, schemadata):
            """parse the schema text to a schema instance and create a parser
               that uses it
            """
            try:
                mtype, schematext = schemadata
            except ValueError:
                errmsg = 'schema type and schema text must be passed'
                raise ValueError(errmsg)
            if schematext!=self._schema.get(mtype) and schematext is not None:
                if self._typedef_schema is not None:
                    modschema= self._include_typedefs(schematext)
                else:
                    modschema = schematext
                try:
                    schemaroot = etree.XML(modschema)
                    s = etree.XMLSchema(schemaroot)
                    self._schema[mtype] = schematext
                except etree.XMLSyntaxError, e:
                    self._schema[mtype] = None
                    raise ValueError(e.args[0])
                finally:
                    self._all_ok = False
                    for x in self._ok:
                        self._ok[x] = False
        return locals()
    schema = property(**schema())

    def xml():
        doc = "text representation of XML document"
        def fget(self):
            return self._xml
        def fset(self, xmldata):
            """parse the xml text and validate against the schema, create
               the object representation, and set state"""
            self.errors = set()
            self.warnings = []
            try:
                xmlname, xmltext, lexicon, dirpath, mtype = xmldata
                self.xml_name = xmlname
            except ValueError:
                errormsg = 'xml name, xml content, lexicon instance, model '\
                        'directory path and model type must be passed'
                raise ValueError(errormsg)
            if self._schema[mtype]!='':
                try:
                    schematext = self._schema[mtype]
                    if self._typedef_schema is not None:
                        modschema = self._include_typedefs(schematext)
                    else:
                        modschema = schematext
                    schemaroot = etree.XML(modschema)
                    s = etree.XMLSchema(schemaroot)
                    parser = etree.XMLParser(schema=s, remove_comments=True)
                    root = etree.fromstring(xmltext, parser)
                except etree.XMLSyntaxError, e:
                    raise ValueError(e.args[0])
                else:
                    try:
                        self._xml[mtype]
                    except:
                        self._xml[mtype] = {}
                    self._xml[mtype][xmlname] = xmltext
                    self.xml_to_obj(xmlname, root, lexicon, mtype, dirpath)
                    if self.errors != set([]):
                        self._ok[xmlname] = False
                        self._all_ok = False
                        errormsg = "errors in xml to object conversion"
                        raise ValueError(errormsg)
                    else:
                        self._ok[xmlname] = True
                        self._all_ok = True
                        for xmlok in self._ok.values():
                            if not xmlok:
                                self._all_ok = False
            else:
                raise ValueError('No schema set')
        return locals()
    xml = property(**xml())

    def xml_to_obj(self, xmlname, root, lexicon, mtype, dirs):
        """
        Constructs a dictionary of model types having dictionary of
        model name:model instance as data
        """
        self.lexicon = lexicon
        self.model_type = mtype
        ns = self.elem_namespace(root.tag)
        if self.model_type == 'aggregation':
            moduleinfo = ('aggregation model collection', AggregationModel)
        elif self.model_type == 'cash_flow':
            moduleinfo = ('cash flow model collection', CashFlowModel)
        elif self.model_type == 'cash_flow_table':
            moduleinfo = ('cash flow table collection', CashFlowTable)
        elif self.model_type == 'geo_table':
            moduleinfo = ('geo table collection', GeoTable)
        elif self.model_type == 'management':
            moduleinfo = ('management model collection', ManagementModel)
        elif self.model_type == 'operation':
            moduleinfo = ('operation model collection', OperationModel)
        elif self.model_type == 'parameter_table':
            moduleinfo = ('parameter table collection', ParameterTable)
        elif self.model_type == 'prediction':
            moduleinfo = ('prediction model collection', PredictionModel)
        self._set_models(xmlname, ns, root, dirs, moduleinfo)

    def _set_models(self, xmlname, ns, root, dirs, moduleinfo):
        """Construct models from the parsed xml for a specific model type"""
        fm, ModelClass = moduleinfo
        self.from_module = fm
        if self.model_type not in self.obj:
            self.obj[self.model_type] = {}
        if xmlname not in self.obj[self.model_type]:
            self.obj[self.model_type][xmlname] = {}
        poset = ('operation model collection', 'prediction model collection')
        if moduleinfo[0] not in poset:
            for elem in root:
                m = ModelClass(ns, elem, self, dirs)
                self.obj[self.model_type][xmlname][m.name] = m
        else:
            for group in root:
                for elem in group:
                    m = ModelClass(ns, elem, self, dirs, group.attrib['name'])
                    self.obj[self.model_type][xmlname][m.name] = m


def main():
    pass

if __name__ == '__main__':
    main()

