#!/usr/bin/env python
# encoding: utf-8
"""
*operationmodel.py*

Process the operation model XML definitions

Copyright (c) 2008 Simosol Oy. Distributed under the GNU General Public
License 2.0.
"""

# these are for running the doctest from src/simo
#import sys
#sys.path.append('..')

from simo.builder.modelbase.model import POModel
from simo.builder.modelbase.model import Limit
from simo.builder import names


class OperationModel(POModel):
    """docstring for OperationModel"""
    def __init__(self, ns, elem, validator, dirs, group=None):
        super(OperationModel, self).__init__(ns, elem, validator, dirs)
        self._validator = validator
        self._type = names.OPERATION
        self._dirs = dirs
        self._name = None
        self.group = group
        self.optype = None
        self.operation_result_weight = None
        self.data_vars = None
        self.data_limits = None
        self.cash_flow_handler = None
        self.interactive = True
        self.n_params = 0
        self.parameters = None
        self.parameter_tables = None
        self.result_type = None
        self.result_level = None
        self.result_target = None
        self.result_variables = None
        self.result_classifiers = {}
        self.text_classifiers = {}
        self._parse_operation(ns, elem)
        self.index = validator.add_model(self._type, self._name)
        self._validator = None

    def _parse_operation(self, ns, elem):
        """Parse operation model structures"""
        self._name = elem.find(ns+'name').text
        typeelem = elem.find(ns+'type')
        self.optype = self._validator.elem_name(typeelem[0].tag)
        if self.optype=='model':
            self._lib = typeelem[0][0].text
            self._lang = typeelem[0][1].text.lower()
            self.cash_flow_handler = typeelem[0][2].text.lower()
            if typeelem[0][3].text.lower() == 'off':
                self.interactive = False
        self._parse_data(ns, elem)
        self._parse_parameters(ns, elem)
        self._parse_param_tables(ns, elem)
        self._parse_weights(ns, elem)
        self._parse_results(ns, elem)

    def _parse_data(self, ns, elem):
        """Parse operation model input data structures"""
        dataelem = elem.findall("./"+ns+'data/'+ns+'group')
        if len(dataelem) == 0:
            return
        self.data_vars = {}
        self.data_limits = {}
        for d in dataelem:
            level = d.find(ns+'level').text
            varelem = d.find(ns+'variables')
            if varelem is None:
                continue
            variables, limits = [], []
            for e in varelem:
                variable = e.find(ns+'name').text
                lim = None
                le = e.find(ns+'limits')
                if le is not None:
                    lim = Limit(ns, le)
                levelind, variableind = \
                    self._validator.variable_ind(level, variable)
                variables.append(variableind)
                limits.append(lim)
            self.data_vars[levelind] = variables
            self.data_limits[levelind] = limits
        if self.data_vars == {}:
            self.data_vars = None
            self.data_limits = None

    def _parse_parameters(self, ns, elem):
        """Parse operation parameter names"""
        pelem = elem.find(ns+'parameters')
        if pelem is None:
            return
        self.param_names = []
        self.param_defs = []
        for p in pelem:
            name = p.find(ns+"name").text
            optional = p.find(ns+"optional").text[0].upper()=='T'
            le = p.find(ns+'limits')
            if le is not None:
                limits = Limit(ns, le)
            else:
                limits = None
            self.param_names.append(name)
            self.param_defs.append((optional, limits))
        self.n_params = len(self.param_names)

    def _parse_param_tables(self, ns, elem):
        """Parse operation parameter table names"""
        telem = elem.find(ns+'parameter_tables')
        if telem is None:
            return
        self.parameter_tables = []
        for t in telem:
            self.parameter_tables.append(t.text)

    def _parse_weights(self, ns, elem):
        """Parse operation model result weight variable"""
        welem = elem.find(ns+"operation_result_weight")
        if welem is None:
            return
        name = welem.find(ns+"name").text
        level = welem.find(ns+"level").text
        ind = self._validator.variable_ind(level, name)
        self.operation_result_weight = ind

    def _parse_results(self, ns, elem):
        """Parse operation model result structures"""
        re = elem.find(ns+"results")
        if re is None:
            return
        self.result_type = []
        self.result_variables = []
        for res in re:
            res_type = self._validator.elem_name(res[0].tag)
            self.result_type.append(res_type)
            if res_type == 'data':
                res_level = res[0][1].text
                self.result_level = self._validator.level_ind(res_level)
                self.result_target = res[0][0].text
            for e in res.find(ns+'variables'):
                variable = e.find(ns+'name').text
                var_type = e.get('type')
                var_type = 'scalar' if var_type is None else var_type

                ce = e.find(ns+'classifiers')
                key = None
                if ce is not None and res_type == 'operation_result':
                    key, cfiers, text_cfiers = self._parse_classifiers(ns, ce)
                    if not key in self.result_classifiers:
                        self.result_classifiers[key] = cfiers
                        self.text_classifiers[key] = text_cfiers
                if res_type == 'data':
                    key = self._validator.variable_ind(res_level, variable)

                weight = True
                we = e.find(ns+'weight')
                if we is not None:
                    weight = True if we.text == 'true' else False

                result_variable = (res_type, variable, key, var_type, weight)
                self.result_variables.append(result_variable)

    def _parse_classifiers(self, ns, elem):
        """Parse operation model result classifiers"""
        classifiers = []
        text_classifiers = []
        key = []
        for c in elem:
            ctype = self._validator.elem_name(c[0].tag)
            if ctype == 'variable':
                name = c.find(ns+'variable').text
                level = c.find(ns+'level').text
                if name in self._validator.lexicon.levels[level].textual:
                    text_classifiers.append((level, name))
                else:
                    levind, varind = self._validator.variable_ind(level, name)
                    cfier = Classifier(ctype, name, varind, levind)
                    classifiers.append(cfier)
            elif ctype == 'cashflow_classifier':
                ctype = 'cashflow'
                name = c[0][0].text
                cfier = Classifier(ctype, name, None, None)
                classifiers.append(cfier)
            key.append(name)
        return tuple(key), classifiers, text_classifiers


class Classifier(object):
    def __init__(self, ctype, name, ind, level):
        """Class for operation model result classifiers"""
        super(Classifier, self).__init__()
        self.type = ctype
        self.name = name
        self.ind = ind
        self.level = level
        self.labels = None


class OperationModelParam(object):
    """Operation model parameters"""
    def __init__(self, ns, elem, task, model, cftables, cfmodels,
                 ptables):
        super(OperationModelParam, self).__init__()
        self._validator = task.validator
        self.task_name = task.name
        self.model = model
        self.name = model.name
        self.group = model.group
        self.implementation = model.name
        self.all_cf_tables = cftables
        self.all_cf_models = cfmodels
        self.all_p_tables = ptables
        self.parameters = None
        self.long_term = None
        self.cash_flow_table = None
        self.cash_flow_model = None
        self.parameter_tables = []
        self.simulation_effects = None
        self.memory_level = None
        self.notes = ''
        self.materials = []
        self._parse_params(ns, elem)
        self._parse_long_term(ns, elem)
        self._get_cash_flow_tables(ns, elem)
        self._get_cash_flow_model(ns, elem)
        self._get_param_tables(ns, elem)
        self._get_sim_effects(ns, elem)
        self._get_mem_level(ns, elem)
        self._get_db_names(ns, elem)
        self._get_notes(ns, elem)
        self._get_materials(ns, elem)
        self.model = None
        self.all_cf_tables = None
        self.all_cf_models = None
        self.all_p_tables = None
        self._validator = None

    def _parse_params(self, ns, elem):
        """Parse operation model parameters"""
        params = elem.find(ns + 'parameters')
        if params is None and self.model.n_params == 0:
            return
        if self.model.n_params != 0 and params is None:
            errormsg = "No parameters defined for model '%s', %s "\
                    "expected, task '%s'" % (self.model.name,
                    self.model.n_params, self.task_name)
            self._validator.add_error(errormsg)
            return
        if self.model.n_params > 0:
            self.parameters = [None for i in range(self.model.n_params)]
            paramset = set(self.model.param_names)
        else:
            paramset = set([])
        for p in params:
            pname = p[0].text
            if pname in paramset:
                paramset.remove(pname)
                self._add_param(p, pname)
            else:
                errormsg = "Parameter '%s' is not a valid parameter for "\
                        "operation model '%s', task '%s'" % (pname,
                                self.model.name, self.task_name)
                self._validator.add_error(errormsg)
        # check the remaining parameters that didn't get values
        for param in paramset:
            ind = self.model.param_names.index(param)
            optional = self.model.param_defs[ind][0]
            if not optional:
                errormsg = "Non optional parameter '%s' is not defined for" \
                        " operation model '%s', task '%s'" % (param,
                                self.model.name, self.task_name)
                self._validator.add_error(errormsg)

    def _add_param(self, pelem, pname):
        """
        Adds the parameter value to parameters dictionary if it passes validity
        checks

        pelem -- parameter element
        pname -- parameter name
        """
        ind = self.model.param_names.index(pname)
        optional = self.model.param_defs[ind][0]
        limits = self.model.param_defs[ind][1]
        if pelem[1].text is not None:
            pvals = [float(val) for val in pelem[1].text.split()]
        elif optional:
            pvals = [None]
        else:
            errormsg = "Non optional parameter '%s' has no value for "\
                    "operation model '%s', task '%s'" % (pname,
                            self.model.name, self.task_name)
            self._validator.add_error(errormsg)
        if limits is not None:
            self._check_parameter_limits(pname, pvals, limits)
        if len(pvals) == 1:
            pvals = pvals[0]
        self.parameters[ind] = pvals

    def _check_parameter_limits(self, pname, pvals, limits):
        """
        Checks that defined parameter values follow the limit definitions

        pname -- parameter name
        pvals -- parameter values in a list
        limits -- Limit object
        """
        if limits.enum is not None:
            for value in pvals:
                if value not in limits.enum and value is not None:
                    errormsg = "Value %s for parameter '%s' is not in the"\
                        " accepted value list %s for operation model '%s',"\
                        " task '%s'" % (value, pname, limits.enum,
                                self.model.name, self.task_name)
                    self._validator.add_error(errormsg)
        if limits.lower is not None:
            for value in pvals:
                if value < limits.lower and value is not None:
                    errormsg = "Value %s for parameter '%s' is smaller than"\
                        " the lower limit %s for operation model '%s', task " \
                        "'%s'" % (value, pname, limits.lower, self.model.name,
                                self.task_name)
                    self._validator.add_error(errormsg)
        if limits.upper is not None:
            for value in pvals:
                if value > limits.upper and value is not None:
                    errormsg = "Value %s for parameter '%s' is bigger than"\
                        " the upper limit %s for operation model '%s', task "\
                        "'%s'" % (value, pname, limits.upper, self.model.name,
                                self.task_name)
                    self._validator.add_error(errormsg)

    def _parse_long_term(self, ns, elem):
        """Parse operation model long time effects"""
        longterm = elem.find(ns + 'long_term')
        if longterm is None:
            return
        self.long_term = []
        for e in longterm:
            level = e[0].text
            var = e[1].text
            level, var = self._validator.variable_ind(e[0].text, e[1].text)
            effects = []
            for eff in e[2]:
                effects.append([[int(eff[0][0].text), int(eff[0][1].text)],
                                float(eff[1].text)])
            self.long_term.append({'level': level, 'variable': var,
                                    'effects': effects})

    def _get_cash_flow_tables(self, ns, elem):
        """Get operation model cash flow table"""
        e = elem.find(ns + 'cash_flow_tables')
        ind = 0
        if e is not None:
            self.cash_flow_table = [None for i in e]
            for cf in e:
                tabledef = cf.text
                if self.all_cf_tables is None:
                    errormsg = "No cash flow tables defined, could not get "\
                               "table '%s' for operation model '%s' in task "\
                               "'%s'" % (tabledef, self.model.name,
                                         self.task_name)
                    self._validator.add_error(errormsg)
                    continue
                if tabledef in self.all_cf_tables:
                    cft = self.all_cf_tables[tabledef]
                    if cft.classifiers is not None:
                        key = cft.classifier_order[:cft.keycount]
                        for val in key:
                            classifier = cft.classifiers[val]
                            lev = classifier.level
                            i = classifier.ind
                            if lev not in self.model.data_vars:
                                levname = self._validator.level_name(lev)
                                if levname == None:
                                    levname = str(lev)
                                errormsg = "Cash flow table key classifier "\
                                           "level '%s' not found for operati"\
                                           "on model '%s', task '%s'" %\
                                           (levname, self.model.name,
                                            self.task_name)
                                self._validator.add_error(errormsg)
                            elif i not in self.model.data_vars[lev]:
                                levname = self._validator.level_name(lev)
                                if levname == None:
                                    levname = str(lev)
                                errormsg = "Cash flow table key classifier '%s' "\
                                           "not found on level '%s' "\
                                           "for operation model '%s', task '%s'" %\
                                           (val, levname, self.model.name,
                                            self.task_name)
                                self._validator.add_error(errormsg)
                    self.cash_flow_table[ind] = cft
                    ind = ind + 1
                else:
                    errormsg = "Cash flow table '%s' not found for operation "\
                            "model '%s', task '%s'" % (tabledef, self.model.name,
                                    self.task_name)
                    self._validator.add_error(errormsg)

    def _get_cash_flow_model(self, ns, elem):
        """Get operation model cash flow model"""
        e = elem.find(ns + 'cash_flow_model')
        if e is not None:
            if self.all_cf_models is None:
                errormsg = "No cash flow models defined, could not get "\
                           "model '%s' for operation model '%s' in task "\
                           "'%s'" % (e.text, self.model.name, self.task_name)
                self._validator.add_error(errormsg)
            elif e.text in self.all_cf_models:
                m = self.all_cf_models[e.text]
                self.cash_flow_model = m
                m.table = [None for i in m.table_names]
                ind = 0
                for name in m.table_names:
                    if name not in self.all_cf_tables:
                        errormsg = "Cash flow table '%s' not found for cash flow "\
                                 "model '%s' of operation model '%s', task '%s'"\
                                 % (name, m.name, self.model.name,
                                    self.task_name)
                        self._validator.add_error(errormsg)
                    else:
                        m.table[ind] = self.all_cf_tables[name]
                        ind = ind + 1

            else:
                errormsg = "Cash flow model '%s' not found for operation "\
                        "model '%s', task '%s'" % (e.text, self.model.name,
                                self.task_name)
                self._validator.add_error(errormsg)

    def _get_db_names(self, ns, elem):
        """
        Get result db names for operation model name and group
        """
        e = elem.find(ns + 'db_names')
        if e is not None:
            self.name = e.find(ns + 'name').text
            self.group = e.find(ns + 'group').text

    def _get_notes(self, ns, elem):
        """
        Get result db names for operation model name and group
        """
        e = elem.find(ns + 'notes')
        if e is not None:
            self.notes = e.text

    def _get_materials(self, ns, elem):
        """
        Get result db names for operation model name and group
        """
        e = elem.find(ns + 'materials')
        if e is not None:
            for material in e.findall(ns + 'material'):
                self.materials.append(material.text)

    def _get_param_tables(self, ns, elem):
        """Get operation model parameter tables"""
        e = elem.find(ns + 'parameter_tables')
        if e is not None and self.model.parameter_tables is not None:
            self.parameter_tables = [None for i in self.model.parameter_tables]
            for pt in e:
                tabledef = pt[0].text
                ind = None
                if tabledef in self.model.parameter_tables:
                    ind = self.model.parameter_tables.index(tabledef)
                if ind is None:
                    errormsg = "Parameter table '%s' is not valid for "\
                        "operation model '%s', task '%s'" % (tabledef,
                                self.model.name, self.task_name)
                    self._validator.add_error(errormsg)
                    continue
                valuetable = pt[1].text
                if valuetable in self.all_p_tables:
                    self.parameter_tables[ind] = self.all_p_tables[valuetable]
                else:
                    errormsg = "Parameter table '%s' not found for operation "\
                            "model '%s', task '%s'" % (valuetable,
                                    self.model.name, self.task_name)
                    self._validator.add_error(errormsg)
            if None in self.parameter_tables:
                errormsg = "All parameter tables for operation model '%s' not"\
                        " found, task '%s'" % (self.model.name, self.task_name)
                self._validator.add_error(errormsg)
        elif self.model.parameter_tables is not None:
            errormsg = "No parameter tables defined, %s expected, task '%s'" \
                    % (self.model.parameter_tables, self.task_name)
            self._validator.add_error(errormsg)
        elif e is not None:
            errormsg = "Operation model '%s' has no parameter tables defined,"\
                    " task '%s' has parameter tables" % (self.model.name,
                    self.task_name)

    def _get_sim_effects(self, ns, elem):
        """Get operation model simulation effects"""
        e = elem.find(ns + 'simulation_effects')
        if e is None:
            return
        effects = {'erase_memory': False, 'force_step': None}
        # Get the non-optional erase memory effect
        if e[0].text.lower() == 'true':
            effects['erase_memory'] = True
        # Check if the optional force_step element is defined
        if len(e) > 1:
            # Parse force step definition
            step_count = int(e[1][0].text)
            time_step = int(e[1][1][0].text)
            time_unit = e[1][1][1].text
            effects['force_step'] = {'stepcount': step_count,
                                     'timestep': time_step,
                                     'unit': time_unit}
        self.simulation_effects = effects

    def _get_mem_level(self, ns, elem):
        """Get operation model memory level"""
        e = elem.find(ns + 'memory_level')
        if e is not None:
            level = self._validator.level_ind(e.text)
            if level is None:
                errorstr = "Operation model memory level '%s' not found in "\
                    "lexicon for operation model '%s'"\
                    % (e.text, self.model.name)
                self._validator.add_error(errorstr)
            else:
                self.memory_level = level

def main():
    import doctest
    doctest.testfile('./test/operationmodel-py.rst')

if __name__ == '__main__':
    main()
