#!/usr/bin/env python
# encoding: utf-8
"""
*predictionmodel.py*

Process the prediction model XML definitions

Copyright (c) 2008 Simosol Oy. Distributed under the GNU General Public
License 2.0.
"""

# these are for running the doctest from src/simo
#import sys
#sys.path.append('..')

from simo.builder import names
from simo.builder.modelbase.model import POModel
from simo.builder.modelbase.model import Limit

from operator import itemgetter


class PredictionModel(POModel):
    """Class for prediction model definitions"""
    def __init__(self, ns, elem, validator, dirs, group=None):
        super(PredictionModel, self).__init__(ns, elem, validator, dirs)
        self.validator = validator
        self._type = names.PREDICTION
        self.group = group
        self._name = elem.find(ns+'name').text
        self._lib = elem.find(ns+'implemented_at').text
        self._lang = elem.find(ns+'implemented_in').text.lower()
        self._dirs = dirs
        self._parse_parameters(ns, elem)
        self.result_level = None
        self.result_vars = None
        self.memory_model = False
        self.__parse_results(ns, elem)
        validator.add_model(self._type, self._name)
        self.validator = None

    def _parse_parameters(self, ns, elem):
        """Parse parameter definitions"""
        pelem = elem.find(ns+'parameters')
        if pelem is None:
            return
        self.param_limits = []
        self.param_names = []
        for p in pelem:
            name = p.find(ns+"name").text
            le = p.find(ns+'limits')
            if le is not None:
                limits = Limit(ns, le)
            else:
                limits = None
            self.param_limits.append(limits)
            self.param_names.append(name)
        self.n_params = len(self.param_names)

    def __parse_results(self, ns, elem):
        """Parse prediction model result variables"""
        reselem = elem.find(ns+"result")
        self.result_vars = []
        level = reselem[0].text
        if self.validator.elem_name(reselem[1].tag) == 'timespan':
            timespan = reselem[1]
            self.memory_model = True
        else:
            timespan = None
        for e in reselem.find(ns+"variables"):
            name = e.find(ns+"name").text
            levind, varind = self.validator.variable_ind(level, name)
            if timespan is not None:
                steps = float(timespan[0].text)
                unit = timespan[1].text
                cumulation = e[1].text
                rv = ResultVariable(varind, steps, unit, cumulation)
            else:
                rv = ResultVariable(varind)
            self.result_vars.append(rv)
        self.result_level = levind


class ResultVariable(object):
    """Class for prediction model result variables"""
    def __init__(self, variable, timespan=None, unit=None, cumul=None):
        super(ResultVariable, self).__init__()
        self.variable = variable
        self.time_span = timespan
        self.time_unit = unit
        self.cumulation = cumul


class PredictionModelParam(object):
    """Class for prediction model parameters"""
    def __init__(self, ns, elem, task, model):
        super(PredictionModelParam, self).__init__()
        self._validator = task.validator
        self.model = model
        self.parameters = None
        self.rect_factor = 1.0
        self.risk_level = 1
        for e in elem:
            t = self._validator.elem_name(e.tag)
            if t == 'parameters':
                self._parse_params(e)
            elif t == 'rect_factor':
                self.rect_factor = float(e.text)
            elif t == 'risk_level':
                self.risk_level = float(e.text)
        if model.n_params != 0 and self.parameters is None:
            errormsg = "No parameters defined in model chain for prediction "\
                    "model '%s', %d parameters expected"\
                    % (model.name, model.n_params)
            self._validator.add_error(errormsg)
        self.model = None
        self._validator = None

    def _parse_params(self, pelem):
        """
        Matches the parameter values defined in the model chain task against
        the model definition

        pelem -- parameters element
        """
        if self.model.n_params != 0:
            self.parameters = [None for i in range(self.model.n_params)]
            paramset = set(self.model.param_names)
        else:
            paramset = set([])
        for p in pelem:
            pname = p[0].text
            if pname in paramset:
                paramset.remove(pname)
                self._add_param(p, pname)
            else:
                errormsg = "Parameter '%s' is not a valid parameter for "\
                        "prediction model '%s'" % (pname, self.model.name)
                self._validator.add_error(errormsg)
        # check the remaining parameters that didn't get values
        for param in paramset:
            errormsg = "Parameter '%s' is not defined for" \
                        " prediction model '%s'" % (param, self.model.name)
            self._validator.add_error(errormsg)

    def _add_param(self, pelem, pname):
        """
        Adds the parameter value to parameters dictionary if it passes validity
        checks

        pelem -- parameter element
        pname -- parameter name
        """
        ind = self.model.param_names.index(pname)
        limits = self.model.param_limits[ind]
        if pelem[1].text is not None:
            pvals = [float(val) for val in pelem[1].text.split()]
        else:
            errormsg = "Parameter '%s' has no value for prediction model '%s'"\
                       % (pname, self.model.name)
            self._validator.add_error(errormsg)
        if limits is not None:
            self._check_parameter_limits(pname, pvals, limits)
        if len(pvals) == 1:
            pvals = pvals[0]
        self.parameters[ind] = pvals

    def _check_parameter_limits(self, pname, pvals, limits):
        """
        Checks that defined parameter values follow the limit definitions

        pname -- parameter name
        pvals -- parameter values in a list
        limits -- Limit object
        """
        if limits.enum is not None:
            for value in pvals:
                if value not in limits.enum and value is not None:
                    errormsg = "Value %s for parameter '%s' is not in the"\
                        " accepted value list %s for prediction model '%s'"\
                        % (value, pname, limits.enum, self.model.name)
                    self._validator.add_error(errormsg)
        if limits.lower is not None:
            for value in pvals:
                if value < limits.lower and value is not None:
                    errormsg = "Value %s for parameter '%s' is smaller than"\
                        " the lower limit %s for prediction model '%s'"\
                        % (value, pname, limits.lower, self.model.name)
                    self._validator.add_error(errormsg)
        if limits.upper is not None:
            for value in pvals:
                if value > limits.upper and value is not None:
                    errormsg = "Value %s for parameter '%s' is bigger than"\
                        " the upper limit %s for prediction model '%s'"\
                        % (value, pname, limits.upper, self.model.name)
                    self._validator.add_error(errormsg)

def main():
    import doctest
    doctest.testfile('./test/predictionmodel-py.rst')

if __name__ == '__main__':
    main()
