#!/usr/bin/env python
# encoding: utf-8
"""
*table.py*

Common classes and functions for cash flow and parameter tables

Copyright (c) 2008 Simosol Oy. Distributed under the GNU General Public
License 2.0.
"""

import numpy
import copy

import simo.utils.utils as utils

def table2array(table):
    """Create a single Numpy array from XML cash_flow table

    NB: the numpy array MUST be sorted or otherwise value search algorithms
    needed by table caller classes will not work!
    """
    tpl = map(lambda x: tuple([float(y) for y in x.text.split()]), table)
    tpl.sort()
    try:
        return numpy.array(tpl, dtype=float)
    except ValueError, e:
        if str(e) == 'setting an array element with a sequence.':
            width = len(tpl[0])
            for row in tpl:
                if len(row) != width:
                    try:
                        name = table.getparent().getparent().getparent()
                        name = name.get('name')
                    except: 
                        break #failsafe, raise original exception instad.
                    raise ValueError('The table rows have inconsistent ' \
                                     'width in %s' % (name,))
        raise e

def table2dict(table):
    """Create a dictionary with the classifying values joined together as the
    dictionary key and the cash flow as the value; i.e., a row
    '1 2 3 4 5 6 7 100' translates to cdict['1:2:3:4:5:6:7']=100"""
    cols = len(table[0].text.split())
    cdict = {}
    for row in table:
        values = row.text.split()
        if len(values) == 1:
            cdict[0] = float(values[0])
        elif len(values) == 2:
            cdict[values[0]] = float(values[1])
        else:
            # Create the key by joining the classifying values and
            # then the last classifying variable and the actual cash_flow
            cdict[':'.join(values[:-1])] = float(values[cols-1])
    return cdict

def parse_classifiers(ns, elem, validator):
    """Parse cash flow table and/or trend classifiers"""
    cfiers = {}
    db_cfiers = []
    keycount = 0
    i = 0
    order = []
    db_cfier_ind = []
    key_complete = False
    for c in elem.findall(ns + 'classifier'):
        name = c.find(ns + 'variable').text
        level = c.find(ns + 'level').text
        to_db = c.get('to_db')
        if to_db=='true':
            db_cfiers.append(name)
            db_cfier_ind.append(i)
        ind = None
        if level != 'within-operation':
            if key_complete:
                err = 'A key variable is set after a within-operation '\
                        'variable.'
                validator.add_error(err)
            level, ind = validator.variable_ind(level, name)
            keycount += 1
        else:
            key_complete = True
        if len(c) > 2 and validator.model_type == 'cash_flow_table':
            labs = [TableLabel(x[1].text, int(x[0].text)) for x in c[2]]
            variable_type = None
        elif len(c) > 2 and validator.model_type == 'parameter_table':
            variable_type = c.find(ns + 'variable_type').text
            labs = None
        else:
            labs = None
            variable_type = None
        cfiers[name] = TableClassifier(level, name, ind, labs, i,
                                       variable_type)
        order.append(name)
        i += 1
    return cfiers, db_cfiers, db_cfier_ind, order, keycount

def parse_tables(elem):
    """Parse <tables> element contents into ctypes array, value
    dictionary and activation time objects for each table"""
    arrays = []
    N = 0
    for t in elem:
        array = table2array(t)
        dictn = table2dict(t)
        start = t.get("start")
        end = t.get("end")
        if start is not None:
            start = utils.str2date(start)
        if end is not None:
            end = utils.str2date(end)
        arrays.append(Array(array, dictn, start, end))
        N += 1
    return arrays, N


class TableLabel(object):
    def __init__(self, label, value):
        """docstring for __init__"""
        super(TableLabel, self).__init__()
        self.label = label
        self.value = value


class TableClassifier(object):
    def __init__(self, level, name, ind, labels, order, variable_type):
        """docstring for __init__"""
        super(TableClassifier, self).__init__()
        self.level = level
        self.name = name
        self.ind = ind
        self.labels = labels
        self.order = order
        self.variable_type = variable_type
        self.label_by_val = None
        if labels is not None:
            self.label_by_val = dict([(i.value, i.label) for i in labels])


class Array(object):
    def __init__(self, array, dictionary, start=None, end=None):
        """docstring for __init__"""
        super(Array, self).__init__()
        self.array = array
        self.dict = dictionary
        self.start_date = start
        self.end_date = end


def main():
    pass

if __name__ == '__main__':
    main()
