#!/usr/bin/env python
# encoding: utf-8
"""
*conditionparser.py*

Class for condition elements of SIMO simulation logic

Copyright (c) 2007 Simosol Oy. Distributed under the GNU General Public
License 2.0.
"""
import re
from pyparsing import Word, Literal, Combine, Optional, nums, ZeroOrMore, \
                      alphanums, delimitedList, Group, Suppress, Forward, \
                      ParseException, ParseSyntaxException, oneOf, ZeroOrMore
from chainfunc import *

from simo.builder.names import EXPR_DISCOUNT, EXPR_DTYPE_COMB,\
                               EXPR_DTYPE_DATA, EXPR_DTYPE_OP,\
                               EXPR_DTYPE_VALUE, EXPR_OPERATION,\
                               EXPR_DTYPE_FUNC, EXPR_FUNCTION,\
                               EXPR_DTYPE_LEVEL

#grammar
eqop = oneOf(['gt', 'ge', 'eq', 'ue', 'le', 'lt'])
boolop = oneOf(['and', 'or'])
boolset = set(('and', 'or'))
timeop = oneOf(['times_eq', 'times_lt', 'times_gt', 'since_eq', 'since_ue',
                'since_lt', 'since_gt'])
timeset = set(('times_eq', 'times_lt', 'times_gt', 'since_eq', 'since_ue',
               'since_lt', 'since_gt'))
existop = oneOf(['exists', 'not exists'])
existset = set(('exists', 'not exists'))
inop = oneOf(['not in', 'in'])
inset = set(('not in', 'in'))
indef = 'in'
number = Combine(Optional('-') + ('0' | Word('123456789',nums)) +
                 Optional('.' + Word(nums)))
number.setParseAction(float_)
inlist = Group(Suppress('[') + delimitedList(number) + Suppress(']'))
lpar  = Literal("(").suppress()
rpar  = Literal(")").suppress()
level = Word(alphanums + '_')
var = Combine(Literal(':').suppress() + Word(alphanums + '_'))
levelvar = Group(level + var)
eqexpr = Group(levelvar + eqop + (number | levelvar))
inexpr = Group(levelvar + inop + inlist)
timeexpr = Group(levelvar + timeop + (number | levelvar))
existexpr = Group((levelvar | level) + existop)
expr = eqexpr | inexpr | timeexpr | existexpr
predicatelogic = Forward()
atom = expr | (lpar + predicatelogic + rpar)
predicatelogic << Group(atom + ZeroOrMore(boolop - atom))

epsilon = 0.0001
opn = {'gt': ('eq', gte),
       'ge': ('eq', gee),
       'eq': ('eq', eee),
       'ue': ('eq', nee),
       'le': ('eq', lee),
       'lt': ('eq', lte),
       'in': ('eq', in_),
       'not in': ('eq', nin),
       'times_eq': ('times', eee),
       'times_lt': ('times', lte),
       'times_gt': ('times', gte),
       'since_eq': ('since', seq),
       'since_ue': ('since', sue),
       'since_lt': ('since', slt),
       'since_gt': ('since', sgt),
       'exists': ('ex', ext),
       'not exists': ('ex', n_ext),
       'and': ('group', and_),
       'or': ('group', or_)
      }

def is_exist_cond(cond):
    if not isinstance(cond, list):
        if cond in existset:
            return True
        else:
            return False
    else:
        return False

class ConditionParser(object):
    """Class for condition parsing for model chains"""

    def __init__(self, validator, mc_evaluation_level):
        super(ConditionParser, self).__init__()
        self.validator = validator
        self.mc_evaluation_level = mc_evaluation_level
        self._vars_to_indices = True
        self._no_children = False

    def parse(self, expr, vars_to_indices=True, no_children=False):
        # remove newlines and tabs
        expr = expr.replace('\n', ' ').replace('\r', ' ').replace('\t', ' ')
        # remove double spaces
        while '  ' in expr:
            expr = expr.replace('  ', ' ')
        self.rpn = []
        self._vars_to_indices = vars_to_indices
        self._no_children = no_children
        try:
            cond = predicatelogic.parseString(expr, parseAll=True).asList()
            self._construct_condition(cond)
        except (ParseException, ParseSyntaxException), e:
            self.rpn = None
            try:
                bad_bit = expr[e.loc:].split()[0]
            except IndexError:
                bad_bit = '<end of file>'

            # Create a mask to mark the current location of the error
            subber = re.compile('\S')
            mask = subber.sub('a', expr)
            mask = 'A'.join((mask[:e.loc], mask[e.loc+1:]))
            # Remove excess whitespace
            whitespace = re.compile('\s+')
            mask = whitespace.sub(' ', mask)
            expr = whitespace.sub(' ', expr)
            # Correct e.loc based on the mask
            e.loc = mask.find('A')
            errormsg = 'Error parsing expression "%s", %s %s, got "%s"'\
                            % (expr, e.msg, 'at char '+str(e.loc), bad_bit)
            self.validator.add_error(errormsg)
        return self.rpn

    def _construct_condition(self, cond):
        """turn the parsed list into Reverse Polish notation / Postfix list"""
        if len(cond) == 1:
            # this is the case for a single atomic condition of
            # [variable, operator, value]
            self._construct_condition(cond[0])
        elif len(cond) == 3 or is_exist_cond(cond[1]):
            # this is an atomic condition including atomic condition with
            # exists/not exists operator which has the length of 2
            self._process_triplet(cond)
        elif len(cond) == 2:
            # if this wasn't the exists/not exists condition then this is
            # a combination of a boolean operator and an atomic condition
            # if size of cond[1] is 3, the second item is a triplet, but it
            # can also be longer, in which case recursively call
            # self._construct_condition
            if len(cond[1]) == 3:
                self._process_triplet(cond[1])
            else:
                self._construct_condition(cond[1])
            self.rpn.append(opn[cond[0]])
        elif len(cond) > 3:
            # a string of "atomic boolop atomic boolop atomic ..."
            self._construct_condition(cond[:3])
            for i in range(3, len(cond), 2):
                self._construct_condition(cond[i:i+2])

    def _process_triplet(self, cond):
        """Processess the atomic condition of [variable, operator, value] or
           [variable, operator, variable] or [level, exists operator] or
           [operation, operator, value] or [operation, operator, variable]
           Also recursively calls the _construct_condition to divide the
           [condition, boolean operator, condition] cases.
        """
        if cond[1] in boolset:
            # subcondition of cond1, boolean operator, cond2
            self._construct_condition(cond[0])
            self._construct_condition(cond[2])
            self.rpn.append(opn[cond[1]])
        else:
            operator = cond[1]
            if operator in timeset:
                dtype = EXPR_DTYPE_OP
            elif operator in existset and not isinstance(cond[0], list):
                dtype = EXPR_DTYPE_LEVEL
                cond[0] = [cond[0], None]
            else:
                # if it's a list, it's a variable or a function
                if cond[0][0] == EXPR_FUNCTION:
                    dtype = EXPR_DTYPE_FUNC
                else:
                    dtype = EXPR_DTYPE_DATA
            if dtype == EXPR_DTYPE_FUNC:
                self._set_function(cond[0][1])
            else:
                self._set_variable(dtype, cond[0])
            if operator in existset:
                self.rpn.append(None)
            else:
                # 'in test' here to differentiate between a list of
                # values for 'in' operator and a list containing a
                # variable definition
                if operator in inset:
                    self.rpn.append((EXPR_DTYPE_VALUE, cond[2]))
                else:
                    if isinstance(cond[2], list):
                        self._set_variable(EXPR_DTYPE_DATA, cond[2])
                    else:
                        self.rpn.append((EXPR_DTYPE_VALUE, cond[2]))
            self.rpn.append(opn[operator])

    def _set_function(self, funcname):
        """appends a function from bfunc into the rpn
        """
        self.rpn.append(('function', functions[funcname]))

    def _set_variable(self, dtype, vardef):
        """Converts the variable, level and operation names into indices
        """
        obey_throughlevel = True
        level, variable = vardef[0:2]
        if level == 'self':
            level = self.mc_evaluation_level
            obey_throughlevel = False
        elif self._no_children and variable is not None:
            # make sure that condition variables are either on evaluation
            # level or on a level above evaluation level (parent level), but
            # never on a child level
            eval_at = self.mc_evaluation_level
            level_1 = self.validator.level_ind(level)
            level_2 = self.validator.level_ind(eval_at)
            if level_1 is None:
                errormsg = "invalid level '%s'" % level_1
                self.validator.add_error(errormsg)
                return
            if level_2 is None:
                errormsg = "invalid level '%s'" % level_2
                self.validator.add_error(errormsg)
                return
            if self.validator.lexicon.is_child(level_1, level_2):
                errormsg = "variable '%s:%s' cannot be in '%s' level "\
                           "condition" % (level, variable, eval_at)
                self.validator.add_error(errormsg)
        if self._vars_to_indices:
            if dtype == EXPR_DTYPE_OP:
                varind = self.validator.operation_ind(variable)
            elif dtype == EXPR_DTYPE_DATA:
                levelind = self.validator.level_ind(level)
                varind = self.validator.variable_ind(level, variable)
                varind += (obey_throughlevel, )
            else:
                varind = self.validator.level_ind(level)
        else:
            varind = (level, variable)
        self.rpn.append((dtype, varind))

def main():
    import doctest
    doctest.testfile('./test/conditionparser-py.rst')

if __name__ == '__main__':
    main()


