#!/usr/bin/env python
# encoding: utf-8
"""
*modelchain.py*

Processes the model chain XML definitions, provides handling of model chains.

Copyright (c) 2008 Simosol Oy. Distributed under the GNU General Public License 2.0.
"""
# these are for running the doctest from src/simo
#import sys
#sys.path.append('..')

from lxml import etree
from simo.builder.xmlobject import XmlObject
from conditionparser import ConditionParser

from task import Task

class ModelChainDef(XmlObject):
    """Model chain definitions"""

    def __init__(self, typedef):
        super(ModelChainDef, self).__init__(typedef)
        self.from_module = "model chain collection"
        self.chainpath = ''

    def xml():
        doc = "dictionary of text representation of XML documents"
        def fget(self):
            return self._xml
        def fset(self, xmldata):
            """parse the xml text and validate against the schema, create
               the object representation, and set state"""
            self.errors = set()
            self.warnings = []
            if self._schema is not None:
                try:
                    xmlname, xmltext, lexicon, chaintype, \
                            models = xmldata
                    self.xml_name = xmlname
                except ValueError:
                    errormsg = 'name, xml content, lexicon instance, chain '\
                            'type and model implementation must be passed'
                    raise ValueError(errormsg)
                try:
                    schematext = self._schema
                    if self._typedef_schema is not None:
                        modschema = self._include_typedefs(schematext)
                    else:
                        modschema = schematext
                    schemaroot = etree.XML(modschema)
                    s = etree.XMLSchema(schemaroot)
                    parser = etree.XMLParser(schema=s, remove_comments=True)
                    root = etree.fromstring(xmltext, parser)
                except etree.XMLSyntaxError, e:
                    raise ValueError(e.args[0])
                else:
                    try:
                        self._xml[chaintype]
                    except:
                        self._xml[chaintype] = {}
                    self._xml[chaintype][xmlname] = xmltext
                    self.xml_to_obj(xmlname, root, lexicon, chaintype, models)
                    self.chainpath = ''
                    if self.errors != set([]):
                        self._ok[xmlname] = False
                        self._all_ok = False
                        errormsg = "errors in xml to object conversion"
                        raise ValueError(errormsg)
                    else:
                        self._ok[xmlname] = True
                        self._all_ok = True
                        for xmlok in self._ok.values():
                            if not xmlok:
                                self._all_ok = False
            else:
                raise ValueError('No schema set')
        return locals()

    xml = property(**xml())

    def add_error(self, msg):
        errormsg = msg + " for %s at '%s'%s" % (self.from_module,
                                                self.xml_name, self.chainpath)
        self.errors.add(errormsg)

    def xml_to_obj(self, xmlname, root, lexicon, chaintype, models):
        """model chain XML definition to object instance"""
        self.lexicon = lexicon
        ns = self.elem_namespace(root.tag)
        mcc = ModelChainCollection(ns, root, self, xmlname, chaintype, models)
        if chaintype not in self.obj:
            self.obj[chaintype] = {}
        self.obj[chaintype][xmlname] = mcc
        self.lexicon = None

class ModelChainCollection(object):
    """A bunch of model chain groups originally in an XML file"""
    def __init__(self, ns, root, validator, xmlname, chaintype, models):
        super(ModelChainCollection, self).__init__()
        self.name = xmlname
        validator.chainpath = ', root'
        try:
            validator.chainpath += '/'+self.name
        except TypeError:
            pass
        chainpath = validator.chainpath
        self.chain_type = chaintype
        self.chains = []
        self.depth = 0
        self.chain_groups = {}
        self.branching_groups = {}
        self.memory_models = set([])
        self.opres_cfiers = set([])
        self.opres_vars = set([])
        self.cf_cfiers = set([])
        for chaingroup in root:
            chainnames = set()
            groupname = chaingroup.get('name')
            if groupname is not None:
                validator.chainpath = chainpath + '/' + groupname
            else:
                validator.chainpath = chainpath
            gchainpath = validator.chainpath
            for mc in chaingroup:
                chainname = mc.get('name')
                validator.chainpath = gchainpath+'/'+chainname
                if chainname in chainnames:
                    errormsg = "Duplicate model chain name '%s'" % (chainname)
                    validator.add_error(errormsg) #self.validator...
                else:
                    chainnames.add(chainname)
                    try:
                        self.chain_groups[groupname]
                    except:
                        self.chain_groups[groupname] = []
                    self.chain_groups[groupname].append(chainname)
                    chain = ModelChain(ns, mc, validator, chaintype, models,
                           self.branching_groups, self.opres_cfiers,
                           self.opres_vars, self.cf_cfiers, self.memory_models)
                    if self.depth < chain.depth:
                        self.depth = chain.depth
                    self.chains.append(chain)
        validator.chainpath = ''


class ModelChain(object):
    def __init__(self, ns, root, validator, chaintype, models, bgroups,
                 opres_cfiers, opres_vars, cf_cfiers, memory_models):
        """Construct model chain from the parsed model chain xml"""
        super(ModelChain, self).__init__()
        self.name = root.get("name")
        self.evaluate_at = root.get("evaluate_at")
        self.condition = None
        self.condition_str = None
        self.tasks = []
        cp = ConditionParser(validator, self.evaluate_at)
        self.depth = 0
        maxdepth = self.depth
        for elem in root:
            if elem.tag == ns+'condition':
                self.condition_str = elem.text
                self.condition = cp.parse(elem.text, no_children=True)
            elif elem.tag == ns+'task' or elem.tag == ns+'branch_task':
                task = Task(ns, elem, validator, cp, self.name,
                            self.evaluate_at, self.depth, memory_models,
                            models)
                self.tasks.append(task)
                opres_cfiers.update(task.opres_cfiers)
                cf_cfiers.update(task.cf_cfiers)
                opres_vars.update(task.opres_vars)
                if maxdepth < task.depth:
                    maxdepth = task.depth
            elif elem.tag == ns + 'branching_group':
                # parse the current branching_group element
                bg = elem.get('name')
                fb = elem.get('free_branching')
                if fb in (None, 'true'):
                    fb = True
                else:
                    fb = False
                bg_cond = elem.find(ns+'condition')
                if bg_cond is not None:
                    bg_cond = cp.parse(bg_cond.text, no_children=True)
                if not bg in bgroups:
                    bgroups[bg] = [[(bg_cond, {})], fb]
                    part_ind = 0
                else:
                    if fb != bgroups[bg][1]:
                        errormsg = "conflicting 'free_branching' values "\
                                   "in branching_group '%s'" % bg
                        validator.add_error(errormsg)
                    bgroups[bg][0].append((bg_cond, {}))
                    part_ind = len(bgroups[bg]) - 1
                # branching group can contain multiple branch tasks
                for e in elem.findall(ns+'branch_task'):
                    task = Task(ns, e, validator, cp, self.name,
                                self.evaluate_at, self.depth, memory_models,
                                models, bg, part_ind)
                    self.tasks.append(task)
                    opres_cfiers.update(task.opres_cfiers)
                    cf_cfiers.update(task.cf_cfiers)
                    opres_vars.update(task.opres_vars)
                    if maxdepth < task.depth:
                        maxdepth = task.depth
                    tname = task.name
                    # lists branching condition names
                    branches = [i[0] for i in task.branch_conditions]
                    if None in branches:
                        errormsg = "missing 'name' attribute in branching "\
                                   "condition of branching group '%s'" % bg
                        validator.add_error(errormsg)
                    bgroups[bg][0][-1][1][tname] = branches
        self.depth = maxdepth

def main():
    import doctest
    doctest.testfile('./test/modelchain-py.rst')

if __name__ == '__main__':
    main()
