#!/usr/bin/env python
# encoding: utf-8
"""
*task.py*

Processes the task XML definition in model chain

Copyright (c) 2008 Simosol Oy. Distributed under the GNU General Public
License 2.0.
"""
from copy import deepcopy

from simo.builder import names
from simo.builder.modelbase.aggregationmodel import AggregationModelParam
from simo.builder.modelbase.geotable import GeoTableParam
from simo.builder.modelbase.managementmodel import ManagementModelParam
from simo.builder.modelbase.operationmodel import OperationModelParam
from simo.builder.modelbase.predictionmodel import PredictionModelParam


class Task(object):
    def __init__(self, ns, elem, validator, cp, path, chain_level,
                 depth, memory_models, models=None, bgroup=None, bg_part_ind=0):
        self.validator = validator
        self.condition_parser = cp
        self.name = elem.get("name")
        basechain = validator.chainpath
        validator.chainpath += '/' + self.name
        self.path = '/'.join([path, self.name])
        self.chain_level = chain_level
        self.depth = depth + 1
        self.type = None
        self.condition = None
        self.branch_conditions = []
        self.subtasks = []
        self.models = models
        self.memory_models = memory_models
        self.model = None
        self.model_type = None
        self.model_obj = None
        self.model_param = None
        self.branching_ops = None
        self.branching_group = bgroup
        self.branching_group_part = bg_part_ind
        self.opres_cfiers = set([])
        self.opres_vars = set([])
        self.cf_cfiers = set([])
        if elem.get("value_fixer") in ('true', '1'):
            self.value_fixer = True
        else:
            self.value_fixer = False
        self._construct_chain_task(ns, elem)
        validator.chainpath = basechain
        self.validator = None
        self.condition_parser = None
        self.models = None
        self.memory_models = None

    def _check_cash_flow_classifier(self, tables, item, mimpl):
        if tables is None:
            errorstr = "Operationmodel '%s' has neither cash_flow_table "\
                           "nor cash_flow_model!" % (mimpl._name,)
            self.validator.add_error(errorstr)
            return
        errors = []
        for table in tables:
            if table is None:
                errorstr = "Operationmodel '%s' has neither cash_flow_table "\
                           "nor cash_flow_model!" % (mimpl._name,)
                errors.append(errorstr)
            else:
                if not item.name in table.classifiers:
                    errorstr = "classifier '%s' of Operation model '%s' "\
                               "not defined in cash flow "\
                               "table '%s' classifiers!" % (item.name,
                                                            mimpl._name,
                                                            table.name)
                    errors.append(errorstr)
                else:
                    item.labels = table.classifiers[item.name].label_by_val
                    return
        for errorstr in errors:
            self.validator.add_error(errorstr)

    def _check_model_levels(self, type, models):
        if type not in ('operation', 'prediction', 'parameter_table'):
            return
        eval_ind = self.validator.lexicon.get_level_ind(self.chain_level)
        for name, model in models.items():
            if type in ('operation', 'prediction'):
                try:
                    for key in model.vars.keys():
                        if key > eval_ind:
                            klname = self.validator.lexicon.get_level_name(key)
                            ealname = self.chain_level
                            errorstr = "%s model (%s) input variables cannot be on "\
                           "level '%s' when model evaluation level is '%s' "\
                           "(variables must be on evaluation level or on one"\
                           " of it's parent levels)" % (type, name, klname,
                                                        ealname)
                            self.validator.add_error(errorstr)
                            break
                except TypeError: #hit a None
                    pass
                except AttributeError: #no vars here or hit a None (again)
                    pass
            elif type == 'parameter_table':
                for target in model.targets:
                    if target.level is None:
                        continue
                    if target.level != eval_ind:
                        tlname = self.validator.lexicon.get_level_name(target.level)
                        ealname = self.chain_level
                        errorstr = "%s model (%s) targets cannot be on "\
                           "level '%s' when model evaluation level is '%s' "\
                           "(must be on the same level)" % (type, name, tlname,
                                                            ealname)
                        self.validator.add_error(errorstr)
                        break

    def _construct_chain_task(self, ns, elem):
        """Constructs a nested task-subtasks structure"""
        if elem.tag.lower() == ns+'task':
            self.type = 'task'
        else:
            self.type = 'branch_task'
            bops = elem.get('branching_operations').split(';')
            # convert the branching operation names to operation indices
            ibops = []
            for bop in bops:
                bopind = self.validator.operation_ind(bop)
                if bopind is not None:
                    ibops.append(bopind)
            self.branching_ops = ibops
        for c in elem.findall(ns+'condition'):
            self.condition_str = c.text
            self.condition = self.condition_parser.parse(c.text,
                                                         no_children=True)
        for bc in elem.findall(ns+'branch_conditions'):
            for c in bc.findall(ns+'condition'):
                branchcond = self.condition_parser.parse(c.text,
                                                         no_children=True)
                self.branch_conditions.append((c.get('name'), branchcond))
        for m in elem.findall(ns+'model'):
            self._set_model_and_param(ns, m)
        maxdepth = self.depth
        for st in elem.findall(ns+'task'):
            subtask = Task(ns, st, self.validator, self.condition_parser,
                           self.path, self.chain_level, self.depth,
                           self.memory_models, self.models)
            if maxdepth < subtask.depth:
                maxdepth = subtask.depth
            self.subtasks.append(subtask)
            self.opres_cfiers.update(subtask.opres_cfiers)
            self.opres_vars.update(subtask.opres_vars)
            self.cf_cfiers.update(subtask.cf_cfiers)
        self.depth = maxdepth

    def _get_model_impl(self, mtype, model=None, all=False):
        """
        Checks that the implementation really exists for the task's model.
        Optionally returns all implementations of the given type

        mtype -- model type
        model -- model name
        all -- boolean for returning all implementations
        """
        mdict = self.models.get(mtype)
        if mdict is None:
            warnmsg = "No models for model type '%s'" % mtype
            self.validator.add_warning(warnmsg)
            return None
        if all:
            return mdict
        mimpl = mdict.get(model)
        if mimpl is None:
            errormsg = "Model implementation for model '%s' of type '%s' "\
                       "is missing" % (model, mtype)
            self.validator.add_error(errormsg)
            return None
        self._check_model_levels(mtype, {model:mimpl})
        # Model instances should be unique to tasks, as they can have different
        # parameterisation from each other (like cashflow tables)
        # Not at all sure deepcopy is the proper way to do this.
        # Seems to work though...
        return deepcopy(mimpl)

    def _set_model_and_param(self, ns, elem):
        """
        Extract the model parameter definitions from an XML element if the
        model exists
        Stores the model object and creates the model parameter object
        instance according to the model type
        """
        mimpl = None
        model = elem.find(ns+'name').text
        for e in elem.findall(ns+'aggregation'):
            mimpl = self._get_model_impl(names.AGGREGATION, model)
            if mimpl is not None:
                self.model_type = names.AGGREGATION
                self.model_param = AggregationModelParam(ns, e, self, mimpl)
        for e in elem.findall(ns+'geo_table'):
            mimpl = self._get_model_impl(names.GEO_TABLE, model)
            if mimpl is not None:
                self.model_type = names.GEO_TABLE
                self.model_param = GeoTableParam(ns, e, self, mimpl)
        for e in elem.findall(ns+'parameter_table'):
            mimpl = self._get_model_impl(names.PARAMETER_TABLE, model)
            if mimpl is not None:
                self.model_type = names.PARAMETER_TABLE
        for e in elem.findall(ns+'prediction'):
            mimpl = self._get_model_impl(names.PREDICTION, model)
            if mimpl is not None:
                self.model_type = names.PREDICTION
                self.model_param = PredictionModelParam(ns, e, self, mimpl)
                if mimpl.memory_model:
                    self.memory_models.add(mimpl.name)
        for e in elem.findall(ns+'management'):
            mimpl = self._get_model_impl(names.MANAGEMENT, model)
            if mimpl is not None:
                self.model_type = names.MANAGEMENT
                self.model_param = ManagementModelParam(ns, e, self, mimpl)
        for e in elem.findall(ns+'operation'):
            mimpl = self._get_model_impl(names.OPERATION, model)
            cftables = self._get_model_impl(names.CASH_FLOW_TABLE, all=True)
            cfmodels = self._get_model_impl(names.CASH_FLOW, all=True)
            ptables = self._get_model_impl(names.PARAMETER_TABLE, all=True)
            if mimpl is not None:
                self.model_type = names.OPERATION
                self.model_param = OperationModelParam(ns, e, self, mimpl,
                                                       cftables, cfmodels,
                                                       ptables)
                tables = None
                if self.model_param.cash_flow_table is not None:
                    tables = self.model_param.cash_flow_table
                elif self.model_param.cash_flow_model is not None:
                    tables = self.model_param.cash_flow_model.table
                if tables is None:
                    warnmsg = "Operation model '%s' has neither cash flow "\
                              "table nor cash flow model!" % model
                    self.validator.add_warning(warnmsg)
                else:
                    for table in tables:
                        if table is not None and table.db_cfiers is not None:
                            self.cf_cfiers.update(set(table.db_cfiers))
                if mimpl.result_classifiers is not None:
                    for key, cfiers in mimpl.result_classifiers.items():
                        for var in key:
                            self.opres_cfiers.add(var)
                        for item in cfiers:
                            if item.type == 'cashflow':
                                self._check_cash_flow_classifier(tables, item,
                                                                 mimpl)

                if mimpl.result_variables is not None and \
                   'operation_result' in mimpl.result_type:
                    for item in mimpl.result_variables:
                        res_type, var_name, ind, var_type, weight = item
                        if res_type == 'data':
                            continue
                        self.opres_vars.add(var_name)
        if mimpl is not None:
            self.model = mimpl

def main():
    pass

if __name__ == '__main__':
    main()
