#!/usr/bin/env python
# encoding: utf-8
"""
Title: SIMO optimization expression parser

*exprparser.py*

Copyright (c) 2007 Simosol Oy. Distributed under the GNU General Public License 2.0.
"""
# these are for running the doctest from src/simo
#import sys
#sys.path.append('..')

from collections import deque
from pyparsing import Word, Literal, Combine, Optional, nums, ZeroOrMore, \
                      alphanums, delimitedList, Group, Suppress, Forward, \
                      ParseException, ParseSyntaxException, oneOf, OneOrMore
import numpy
from pprint import pprint

from simo.utils import utils
from simo.builder.names import EXPR_DISCOUNT, EXPR_DTYPE_COMB,\
                               EXPR_DTYPE_DATA, EXPR_DTYPE_OP,\
                               EXPR_DTYPE_VALUE, EXPR_OPERATION
from exprfunc import *


# ----------------------------------------------------------------------------
# GRAMMAR
# -------
# operator definitions
aggregators = oneOf(['sum', 'mean', 'min', 'max', 'avg'])
ariop = oneOf(['+', '-', '*', '/'])  # arithmetic operators
eqop = oneOf(['gt', 'ge', 'eq', 'ue', 'le', 'lt', 'in'])
boolop = oneOf(['and', 'or'])  # boolean operators

# operator sets
ariset = set(['+', '-', '*', '/'])
eqset = set(['gt', 'ge', 'eq', 'ue', 'le', 'lt', 'in'])
boolset = set(('and', 'or'))
aggregatorset = set(['sum', 'mean', 'min', 'max', 'avg'])

# operator precedence
precedence = {'*': 2, '/': 2, '+': 1, '-': 1, '(': 0}

# numbers, parentheses, etc.
number = Combine(Optional('-') + ('0' | Word('123456789', nums)) +
                 Optional('.' + Word(nums)))
integer = Combine(Optional('-') + ('0' | Word('123456789', nums)))
word = Word(alphanums + '_')
number.setParseAction(float_)
lpar  = Literal("(").suppress()
rpar  = Literal(")").suppress()
lbrt = Literal("[").suppress()
rbrt = Literal("]").suppress()
colon = Literal(":").suppress()
intlist = Group(lpar + integer + ZeroOrMore(Literal(',').suppress() + integer) + rpar)
wordlist = Group(lpar + word + ZeroOrMore(Literal(',').suppress() + word) + rpar)

# level (or type), variable, period, condition, discount...
level = Word(alphanums + '_')
var = Combine(colon + Word(alphanums + '_'))
condexpr =  Group(Group(word + colon + word) + eqop + (integer | intlist | word | wordlist))
#condexpr =  Group(word + eqop + (integer | intlist | word))
condition = Optional(Group(lbrt + condexpr + ZeroOrMore(boolop + condexpr) + rbrt))
period = Combine(lbrt + integer + Literal(":") + integer + rbrt)
discount = Optional(Combine(colon + Literal('discount')))

# level:variable[period][condition]:discount
levelvar = Group(level + var + condition + discount)
nestedvars = Literal("(") + levelvar + ZeroOrMore(ariop - levelvar) + Literal(")")
aggrexpr = (levelvar | nestedvars) + ZeroOrMore(ariop - (levelvar | nestedvars))
aggr = Group(aggregators + period + lpar + aggrexpr + rpar)

# expressions
ariexpr = Group(aggr + ZeroOrMore(ariop + (number | aggr)))
eqexpr = Group((ariexpr | aggr) + eqop + (number | aggr | ariexpr))
expr = eqexpr | ariexpr | aggr | number

predicatelogic = Forward()
atom = expr | (lpar + predicatelogic + rpar)
op = ariop | boolop
predicatelogic << Group(atom + ZeroOrMore(op + atom))

# operator functions for evaluating expressions with numerical values
epsilon = 0.0001
opn = {'gt': ('eq', gte, 'gt'),
       'ge': ('eq', gee, 'ge'),
       'eq': ('eq', eee, 'eq'),
       'ue': ('eq', nee, 'ue'),
       'le': ('eq', lee, 'le'),
       'lt': ('eq', lte, 'lt'),
       '+': ('ari', plus, '+'),
       '-': ('ari', minus, '-'),
       '*': ('ari', multi, '*'),
       '/': ('ari', divide, '/'),
       'and': ('eq', and_, 'and'),
       'or': ('eq', or_, 'or')
      }

# operator functions for evaluating variable conditions, where a and b can b
# strings
condopn = {
    'gt': ('eq', gt, 'gt'),
    'ge': ('eq', ge, 'ge'),
    'eq': ('eq', ee, 'eq'),
    'ue': ('eq', ne, 'ue'),
    'le': ('eq', le, 'le'),
    'lt': ('eq', lt, 'lt'),
    'in': ('eq', in_, 'in'),
    'and': ('eq', and_, 'and'),
    'or': ('eq', or_, 'or')
}

# custom nanmean aggregation function
def nanmean(a, ax=0):
    return numpy.nansum(a, axis=ax) / (numpy.isnan(a) == False).sum()

# aggregator functions
aggrfuncs = {
    'sum': ('aggr', numpy.sum),
    'mean': ('aggr', nanmean),
    'min': ('aggr', numpy.nanmin),
    'max': ('aggr', numpy.nanmax),
    'avg': ('aggr', numpy.average)
}




class ExpressionParser(object):
    """Optimization task expression parser"""
    def __init__(self):
        super(ExpressionParser, self).__init__()

    def parse(self, expr, initdate, step, aggregator_constraint=None):
        """Parse SIMO optimization subobjective or constraint expression into
        a nested list and further into a Reverse Polish notation / Postfix
        stack"""
        # make sure that aggregator constraint contains only valid aggregator
        # function names
        if aggregator_constraint is not None:
            for ac in aggregator_constraint:
                if ac not in aggregatorset:
                    raise Exception
            self.aggregators = set(aggregator_constraint)
        else:
            self.aggregators = aggregatorset
        self.rpn = []
        self.errors = None
        self.init_date = initdate
        self.step = step
        try:
            expr = predicatelogic.parseString(expr, parseAll=True).asList()
            self._construct_expression(expr)
        except (ParseException, ParseSyntaxException), e:
            if e.msg == "Expected Re:('gt|ge|eq|ue|le|lt')":
                e.msg = "Expected one of 'gt', 'ge', 'eq', 'ue', 'le', 'lt'"
            elif e.msg == "Expected Re:('sum|mean|min|max|avg')":
                e.msg = "Expected one of 'sum', 'mean', 'min', 'max', 'avg'"
            elif e.msg == "Expected Re:('[+\\\\-*/]')":
                e.msg = "Expected one of '+', '-', '*', '/'"
            self.errors = (e.loc, e.msg)
        except Exception, e:
            self.errors = (None, e.args[0])
        return self.rpn, self.errors

    def _construct_expression(self, expr):
        """Process the parsed list into Reverse Polish notation / Postfix list
        """
        if len(expr) == 1:
            # go deeper in the nested structure
            self._construct_expression(expr[0])
        elif len(expr) == 2:
            # [op, var|val]
            self._process_item(expr[1])
            self.rpn.append(opn[expr[0]])
        elif len(expr) == 3:
            if str(expr[0]) in aggregatorset:
                if str(expr[0]) not in self.aggregators:
                    raise ParseException("Invalid aggregator function '%s'" \
                                         % str(expr[0]))
                # [aggregator, period, [level, variable, condition, discount]]
                self._set_variable(expr[1], expr[2])
                self.rpn.append(aggrfuncs[expr[0]])
            else:
                # [expr, boolop|eqop, expr]
                self._process_triplet(expr)
        else:
            # [aggr, var|val, op, var|val, op, var|val, ...] or
            # [expr, ariop|eqop|boolop, expr, ariop|eqop|boolop, expr, ...]
            if str(expr[0]) in aggregatorset:
                if str(expr[0]) not in self.aggregators:
                    raise ParseException("Invalid aggregator function '%s'" \
                                         % str(expr[0]))
                # aggregation definition. Parse the contents into postfix
                # with Shunting-Yard algorithm
                self._parse_aggr_infix(expr)
            else:
                # expression, with possibly multiple operators
                self._parse_expr_infix(expr)

    def _get_dtype(self, dtype, level):
        """Get data type for a constraint operand. This data type will be used
        in the DB interface for determining from which data table the operand 
        values should be searched for.
        """
        # make sure that if the actual variable data type is 'op', the
        # condition variable can be from operation or data results, but if
        # variable data type is 'data', the condition data can come only
        # from data results
        # if the level is 'op' and original data type is 'data',
        # then the data type should be 'combined' as the DB interface needs 
        # to use both operation and data results for searching the correct 
        # values
        if dtype == EXPR_DTYPE_DATA and level == EXPR_DTYPE_OP:
            raise ValueError("Invalid condition level '%s' for data type "\
                             "'%s'" % (level, EXPR_DTYPE_DATA))
        elif dtype == EXPR_DTYPE_OP and level != EXPR_DTYPE_OP:
            dtype = EXPR_DTYPE_COMB
        return dtype

    def _parse_aggr_infix(self, expr):
        """Parse infix expression inside an aggregation. Infix expression can
        consist of arithmetic operators, parentheses and variables.
        """
        opstack = []
        for i in range(2, len(expr)):
            if isinstance(expr[i], list):
                self._set_variable(expr[1], expr[i])
            elif expr[i] == '(':
                opstack.append(expr[i])
            elif expr[i] == ')':
                while opstack[-1] != '(':
                    self.rpn.append(opn[opstack.pop()])
                if opstack.pop() != '(':
                    raise ValueError("No matching (")
            elif expr[i] in ariset:
                pr = precedence[expr[i]]
                while opstack and pr < precedence[opstack[-1]]:
                    self.rpn.append(opn[opstack.pop()])
                opstack.append(expr[i])
        while opstack:
            self.rpn.append(opn[opstack.pop()])
        self.rpn.append(aggrfuncs[expr[0]])
    
    def _parse_cond(self, expr):
        """
        Parse condition expression
        """
        lv, op, val = expr
        level, var = lv
        if level == EXPR_OPERATION:
            level = EXPR_DTYPE_OP
        return level, var, op, val
    
    def _parse_expr_infix(self, expr):
        """Parse infix expression inside an expression. Infix expression can
        consist of arithmetic operators, parentheses, variables and numbers.
        """
        opstack = []
        for e in expr:
            if isinstance(e, list):
                self._parse_aggr_infix(e)
            elif isinstance(e, float):
                self.rpn.append((EXPR_DTYPE_VALUE, e))
            elif e == '(':
                opstack.append(e)
            elif e == ')':
                while opstack[-1] != '(':
                    self.rpn.append(opn[opstack.pop()])
                if opstack.pop() != '(':
                    raise ValueError('No matching (')
            elif e in ariset:
                pr = precedence[e]
                while opstack and pr <= precedence[opstack[-1]]:
                    self.rpn.append(opn[opstack.pop()])
                opstack.append(e)
        while opstack:
            self.rpn.append(opn[opstack.pop()])

    def _process_triplet(self, expr):
        """Processess the atomic condition of [variable, operator, value] or
           [variable, operator, variable] or [operation, operator, value] or
           [operation, operator, variable]. Also recursively calls the
           __construct_expression to divide the
           [expr, operator, expr] cases.
        """
        self._process_item(expr[0])
        self._process_item(expr[2])
        self.rpn.append(opn[expr[1]])

    def _process_item(self, item):
        if isinstance(item, float):
            self.rpn.append((EXPR_DTYPE_VALUE, item))
        else:
            self._construct_expression(item)

    def _set_variable(self, period, var):
        """Converts the variable, level and operation names"""
        level, name = var[:2]
        if level == EXPR_OPERATION:
            dtype = EXPR_DTYPE_OP
            variable = unicode(name)
        else:
            dtype = EXPR_DTYPE_DATA
            variable = (level, unicode(name))
        period = self._set_period(period)
        condition = None
        if len(var) > 2 and var[2] != EXPR_DISCOUNT:
            condition, dtype = self._set_condition(var[2], dtype)
        discount = False
        if var[-1] == EXPR_DISCOUNT:
            discount = True
        self.rpn.append((dtype, variable, period, condition, discount))

    def _set_condition(self, condexpr, origdtype):
        """Convert condition expression of a single variable into simplified
        reverse polish notation
        """
        gd = self._get_dtype
        
        level, var, op, val = self._parse_cond(condexpr[0])
        
        # Try to cast the value into integer, if possible. If the value is
        # a word (like operation name etc.), pass silently.
        try:
            val = int(val)
        except:
            pass
        if op == 'in':
            # if the operator is 'in', cast the list items into integers
            try:
                val = tuple(map(int_, val))
            except:
                val = tuple(map(str_, val))
        dtype = gd(origdtype, level)
        condition = [(dtype, level, var,), 
                     (EXPR_DTYPE_VALUE, val,),
                     condopn[op]]
        fix = False
        if len(condexpr) > 1:
            for i in range(1, len(condexpr), 2):
                boolop = condexpr[i]
                level, var, op, val = self._parse_cond(condexpr[i+1])
                if op == 'in':
                    try:
                        val = tuple(map(int_, val))
                    except:
                        val = tuple(map(str_, val))
                else:
                    try:
                        val = int(val)
                    except:
                        pass
                dtype = gd(origdtype, level)
                if dtype != origdtype:
                    fix = True
                condition.append((dtype, level, var,))
                condition.append((EXPR_DTYPE_VALUE, val))
                condition.append(condopn[op])
                condition.append(condopn[boolop])
            if fix:
                for i, item in enumerate(condition):
                    if item[0] in (EXPR_DTYPE_DATA, EXPR_DTYPE_OP):
                        condition[i] = (EXPR_DTYPE_COMB,) + item[1:]
        return tuple(condition), dtype

    def _set_period(self, perioddef):
        """Convert period definition into actual dates"""
        s, e = [int(x) for x in perioddef.split(':')]
        origo = self.init_date
        step = self.step
        if s == 0:
            begin = origo
        elif s == -1:
            begin = None  # None indicates last time step
        else:
            begin = utils.update_date(origo, s-1, step)
        if e == 0:
            end = origo
        elif e == -1:
            end = None  # None indicates last time step
        else:
            # end will be set to the last day of the given period
            end = utils.update_date(origo, e, step, deduceaday=True)
        return (begin, end)


def main():
    import doctest
    doctest.testfile('./test/exprparser-py.rst')

if __name__ == '__main__':
    main()
