# encoding: utf-8
"""
Title: Objective functions

SIMO optimization objective, subobjective and utility function definitions

*objfunc.py*

Copyright (c) 2007 Simosol Oy. Distributed under the GNU General Public License 2.0.
"""

class ObjectiveFunction(object):
    """Optimization task objective function definition"""
    def __init__(self, ftype, target, scope):
        """docstring for __init__"""
        super(ObjectiveFunction, self).__init__()
        self._type = ftype
        self._target = target
        self._scope = scope
    
    def type():
        doc = "Objective function type"
        def fget(self):
            return self._type
        return locals()
    type = property(**type())
    
    def target():
        doc = "Objective function target"
        def fget(self):
            return self._target
        return locals()
    target = property(**target())

    def scope():
        doc = "Objective function scope"
        def fget(self):
            return self._scope
        return locals()
    scope = property(**scope())


class SubObjective(object):
    """Optimization task subobjective definition"""
    def __init__(self, weight, expr, ufunc, optlevel=None):
        """docstring for __init__"""
        super(SubObjective, self).__init__()
        self._weight = weight
        self._expression = expr
        self._utility_func = ufunc
        self._optimal_level = optlevel
    
    def weight():
        doc = "Subobjective weight"
        def fget(self):
            return self._weight
        return locals()
    weight = property(**weight())
    
    def expr():
        doc = "Subobjective expression"
        def fget(self):
            return self._expression
        def fset(self, expr):
            self._expression = expr
        return locals()
    expr = property(**expr())
    
    def ufunc():
        doc = "Subobjective utility function"
        def fget(self):
            return self._utility_func
        return locals()
    ufunc = property(**ufunc())
    
    def optlevel():
        doc = "Subobjective optimal level"
        def fget(self):
            return self._optimal_level
        return locals()
    optlevel = property(**optlevel())


class Constraint(object):
    """Optimization task constraint definition"""
    def __init__(self, expr, max_violation=None, violation_metric=None):
        """Constraint object"""
        super(Constraint, self).__init__()
        self._expression = expr
        self._max_violation = max_violation
        self._violation_metric = violation_metric
    
    def expr():
        doc = "Constraint expression"
        def fget(self):
            return self._expression
        def fset(self, expr):
            self._expression = expr
        return locals()
    expr = property(**expr())
    
    def max_violation():
        doc = "Constraint maximum violation"
        def fget(self):
            return self._max_violation
        return locals()
    max_violation = property(**max_violation())
    
    def violation_metric():
        doc = "Constraint violation metric: 'percentage' or 'absolute'"
        def fget(self):
            return self._violation_metric
        return locals()
    violation_metric = property(**violation_metric())


class UFunc(object):
    """Subobjective utility function"""
    def __init__(self, ftype, x, y):
        """docstring for __init__"""
        super(UFunc, self).__init__()
        self._type = ftype
        if x is None and y is None:
            self._midpoint = False
            self._x = None
            self._y = None
        else:
            self._midpoint = True
            self._x = x
            self._y = y
    
    def type():
        doc = "Subobjective utility function type"
        def fget(self):
            return self._type
        return locals()
    type = property(**type())
    
    def midpoint():
        doc = "Subobjective utility function midpoint"
        def fget(self):
            return self._midpoint
        return locals()
    midpoint = property(**midpoint())
    
    def coords():
        doc = "Subobjective utility function midpoint coordinates"
        def fget(self):
            return (self._x, self._y,)
        return locals()
    coords = property(**coords())


def main():
    pass

if __name__ == '__main__':
    main()