#!/usr/bin/env python
# encoding: utf-8
"""
Title: SIMO Optimization task definition

...

*opttask.py*

Copyright (c) 2007 Simosol Oy. Distributed under the GNU General Public License 2.0.
"""

import datetime
from dateutil.relativedelta import relativedelta
from lxml import etree

from simo.builder.xmlobject import XmlObject
from simo.builder.optimization.exprparser import ExpressionParser
from simo.builder.optimization.objfunc import ObjectiveFunction, SubObjective, \
    UFunc, Constraint
from simo.builder.modelchain.conditionparser import ConditionParser

from simo.builder.names import EXPR_DISCOUNT, EXPR_DTYPE_COMB,\
                               EXPR_DTYPE_DATA, EXPR_DTYPE_OP,\
                               EXPR_DTYPE_VALUE, EXPR_OPERATION



class OptimizationTaskDef(XmlObject):
    """Optimization task definition"""

    def __init__(self, typedef):
        """docstring for __init__"""
        super(OptimizationTaskDef, self).__init__(typedef)
        self.from_module = "optimization"

    def xml_to_obj(self, xmlname, root, lexicon):
        """lexicon XML definition to object instance"""
        self.lexicon = lexicon
        ns = self.elem_namespace(root.tag)
        opt = OptimizationTask(ns, root, self)
        self.obj[xmlname] = opt


class OptimizationTask(object):
    """Optimization task"""

    def __init__(self, ns, root, validator):
        """docstring for __init__"""
        self._ns = ns
        self._validator = validator
        self._expr_parser = ExpressionParser()
        self._cond_parser = None
        self._cond_parser = None
        self._init_date = None
        self._period_step = None
        self._data_filter = None
        self._discount_rate = None
        self._conditional_discounting = None
        self._main_level = None
        self._objective_func = None
        self._subobjectives = None
        self._num_of_subobjs = 0
        self._constraints = None
        self._num_of_consts = 0
        self._unique_operands = None
        self._operand_index = None
        self._one_aggr_level = None
        self.date_ranges = None
        self._subobj_strings = None  # in original SIMO syntax
        self._constr_strings = None  # in original SIMO syntax
        self._parse(root)
        self._validator = None
        self._expr_parser = None
        self._cond_parser = None

    def _parse(self, root):
        """
        Parse optimization task definition XML
        
        root -- XML root element
        """
        self._operand_index = {}
        self._unique_operands = {}
        self._one_aggr_level = {}
        self.date_ranges = {}
        self._subobj_strings = []  # in original SIMO syntax
        self._constr_strings = []  # in original SIMO syntax
        self._init_date = self._parse_date(root.find(self._ns+'initial_date'))
        self._period_step = root.find(self._ns + 'period_step').text
        self._discount_rate = float(root.find(self._ns + 'discount_rate').text)
        objf = root.find(self._ns + 'objective_function')
        self._main_level = objf.find(self._ns + 'level').text
        self._parse_objective_func(objf.find(self._ns + 'type'))
        self._parse_subobjectives(objf.find(self._ns + 'sub_objectives'))
        self._parse_constraints(objf.find(self._ns + 'constraints'))
        self._unify_operands()
        
        self._cond_parser = ConditionParser(self._validator, self._main_level)
        
        df = root.find(self._ns + 'data_filter')
        if not self._parse_data_filter(df):
            errormsg = "Could not parse optimization data filter!"
            self._validator.add_error(errormsg)
            
        cd = root.find(self._ns + 'conditional_discount')
        if not self._parse_conditional_discount(cd):
            errormsg = "Could not parse optimization discount rate!"
            self._validator.add_error(errormsg)

    def _parse_conditional_discount(self, elem):
        """
        Parse conditional discounting rates
        """
        if elem is None:
            return True
        self._conditional_discounting = []
        for e in elem:
            condition = e[0].text
            value = float(e[1].text)
            try:
                expr = self._cond_parser.parse(condition, False)
            except:
                errormsg = "error encountered when parsing optimization "\
                           "task discount rate condition '%s'" % condition
                self._validator.add_error(errormsg)
                return False
            self._conditional_discounting.append((condition, expr, value))
        return True

    def _parse_data_filter(self, elem):
        """
        Parse optimization data filter expression
        """
        if elem is None:
            return True
        try:
            self._data_filter = self._cond_parser.parse(elem.text, False)
        except:
            errormsg = "error encountered when parsing optimization task "\
                       "data filter '%s'" % elem.text
            self._validator.add_error(errormsg)
            return False
        return True

    def _parse_date(self, elem):
        """
        Parse date expression from xs:date format into datetime.date object
        """
        if elem.text == 'today':
            date = datetime.date.today()
        else:
            d = [int(x) for x in elem.text.split('-')]
            date = datetime.date(d[0], d[1], d[2])
        return date

    def _parse_objective_func(self, elem):
        """
        Parse objective function definition
        """
        try:
            ftype = elem[0].text
            target = elem[1].text
            scope = elem[2].text
            self._objective_func = ObjectiveFunction(ftype, target, scope)
        except:
            errormsg = "Error parsing objective function!"
            self._validator.add_error(errormsg)

    def _parse_subobjectives(self, elem):
        """
        Parse objective function subobjectives
        """
        self._subobjectives = []
        for counter, subelem in enumerate(elem):
            weight = float(subelem.find(self._ns+'weight').text)
            estr = subelem.find(self._ns+'expression').text
            expr, e = self._expr_parser.parse(estr, self._init_date,
                                              self._period_step, ['sum'])
            if e is not None:
                errormsg = "encountered '%s' when parsing optimization task "\
                           "subobjective number %d." % (e, counter)
                self._validator.add_error(errormsg)
                self._subobjectives = None
                return
            ufunc = self._parse_utility_function(\
                subelem.find(self._ns+'utility_function'))
            if subelem.find(self._ns+'optimal_level') is not None:
                optlevel = float(subelem.find(self._ns+'optimal_level').text)
            else:
                if self._objective_func.type == 'distance':
                    errormsg = "If objective function type is 'distance', "\
                               "optimal level must be defined!"
                    self._validator.add_error(errormsg)
                    return
                optlevel = None
            so = SubObjective(weight, expr, ufunc, optlevel)
            self._subobjectives.append(so)
            # store the subobjective expression in the SIMO syntax for J files
            s = "%0.3f * (%s)" % (weight, estr.strip())
            s = s.replace('\n', '')
            s = s.replace('\t', '')
            if optlevel:
                s += ' = %0.3f' % optlevel
            self._subobj_strings.append(s)
        self._num_of_subobjs = len(self._subobjectives)

    def _parse_constraints(self, elem):
        """
        Parse optimization task constraint expressions
        """
        if elem is None:
            return
        self._constraints = []
        for counter, subelem in enumerate(elem):
            expr, e = self._expr_parser.parse(subelem.text, self._init_date,
                                              self._period_step)
            if e is not None:
                details = "at: %s, error: %s" % (e[0], e[1])
                errormsg = "error encountered when parsing optimization task"\
                           " constraint number %d.: %s" % (counter, details)
                self._validator.add_error(errormsg)
                self._constraints = None
                return
            
            # Process constraint attributes related to constraint violations in
            # LP solutions
            max_violation = subelem.get('max_violation')
            violation_metric = subelem.get('violation_metric')
            if max_violation is not None:
                max_violation = float(max_violation)
            
            self._constraints.append(Constraint(expr, max_violation, 
                                                violation_metric))
            
            # store the constraint expression in the SIMO syntax for J files
            s = subelem.text.strip()
            s = s.replace('\n', '')
            s = s.replace('\t', '')
            self._constr_strings.append(s)
        self._num_of_consts = len(self._constraints)

    def _parse_utility_function(self, elem):
        """
        Parse utility function element
        """
        ftype = elem[0].text
        if len(elem) > 1:
            x = elem[1][0].text
            y = elem[1][1].text
        else:
            x, y = None, None
        ufunc = UFunc(ftype, x, y)
        return ufunc
    
    def _unify_operands(self):
        """
        Go through each subobjective constraint expression and all operands 
        of each expression and construct index of unique operands.
        
        All subobjective and constraint operands are mapped to unique operands 
        and in optimization data matrix, all data processing is done to the 
        unique operands which saves a lot of memory with large data sets.
        
        Unique operands contain variable name, level, condition and 
        discounting info (no date information)
        """
        if not self._subobjectives:
            return
        for so in self._subobjectives:
            so.expr = self._process_expression(so.expr)
        if self._constraints:
            for i, co in enumerate(self._constraints):
                self._constraints[i].expr = self._process_expression(co.expr)
    
    def _process_expression(self, expression):
        """
        Process a single subobjective or constraint expression and update
        unique operand index
        
        expression -- parsed expression in postfix form
        """
        modded_exp = []  # construct a modified expression list
        aggr_stack = []  # temporary container for aggregation groups
        for oper in expression:
            if oper[0] in ('eq', 'ari', 'value'):
                modded_exp.append(oper)
                continue
            elif oper[0] == 'aggr':
                # keep track of which unique operands are in aggregation
                # groups that have more than one unique operand
                if len(aggr_stack) > 0:
                    dtype = aggr_stack[0][0]
                    all_same_type = True
                    for dt, ui in aggr_stack[1:]:
                        if dt != dtype:
                            all_same_type = False
                            break
                    for dt, ui in aggr_stack:
                        if ui in self._one_aggr_level:
                            if not all_same_type:
                                self._one_aggr_level[ui] = all_same_type
                        else:
                            self._one_aggr_level[ui] = all_same_type
                aggr_stack = []
                modded_exp.append(oper)
                continue
            
            # unique operands contain variable name, level, condition and
            # discounting info (no date information)
            uop = oper[:2] + oper[3:]
            
            # get the current operand date limits
            start, end = oper[2]
            
            # get unique operand key (unique operand indice) if the current
            # operand is not yet in the unique operand mapping
            if uop not in self._unique_operands:
                ui = len(self._unique_operands)
                self._unique_operands[uop] = ui
                smallest, biggest = start, end
            else:
                ui = self._unique_operands[uop]
                smallest, biggest = self.date_ranges[ui]
            
            # get a key (operand indice) for mapping current operand to unique
            # operand and do the mapping
            oi = len(self._operand_index)
            self._operand_index[oi] = ui
            
            # add the current operand indice to aggregation stack
            dtype = EXPR_DTYPE_OP if oper[0] == EXPR_OPERATION else EXPR_DTYPE_DATA
            aggr_stack.append((dtype, ui))
            
            # update also date ranges, the aim is to find the maximum and
            # minimum date limits for each unique operand
            if not smallest and start:
                smallest = start
            elif smallest and start:
                if start < smallest:
                    smallest = start
            if not end:
                biggest = end
            elif end and biggest:
                if end > biggest:
                    biggest = end
            self.date_ranges[ui] = (smallest, biggest)
            
            # add both the operand indice and the unique operand indice to the
            # actual operand tuple and store the modified operand to the
            # modified expression
            modoper = oper + ((oi, ui),)
            modded_exp.append(modoper)
        
        return modded_exp
    
    def _change_opt_def_date(self, item, delta):
        """
        Recursively parse through the item, adding the delta to all date 
        objects.
        
        item -- an item from a subobjective or constraint
        delta -- dateutil.relativedelta object"""
        if not isinstance(item, (tuple, list)):
            if isinstance(item, datetime.date):
                item += delta
            return item
        elif isinstance(item, tuple):
            new_item = []
            for val in item:
                new_item.append(self._change_opt_def_date(val, delta))
            return tuple(new_item)
        new_item = []
        for val in item:
            new_item.append(self._change_opt_def_date(val, delta))
        return new_item
    
    def update_init_date(self, new_date):
        """
        Give new initial date, and fix all subobjectives and constraints 
        to match.
        
        new_date -- a datetime.date object"""
        delta = relativedelta(new_date, self._init_date)
        # update init_date so we don't keep changing on subsequent calls
        self._init_date = new_date
        for subobj in self._subobjectives:
            for i, row in enumerate(subobj.expr):
                new_row = []
                for item in row:
                    new_row.append(self._change_opt_def_date(item, delta))
                subobj.expr[i] = tuple(new_row)
        if self._constraints:
            for i, co in enumerate(self._constraints):
                self._constraints[i].expr = self._change_opt_def_date(co.expr, 
                                                                      delta)
        for i, dt in self.date_ranges.items():
            self.date_ranges[i] = (dt[0]+delta, dt[1]+delta)

    def init_date():
        def fget(self):
            return self._init_date
        return locals()
    init_date = property(**init_date())

    def period_step():
        def fget(self):
            return self._period_step
        return locals()
    period_step = property(**period_step())

    def discount_rate():
        def fget(self):
            return self._discount_rate
        return locals()
    discount_rate = property(**discount_rate())
    
    def conditional_discounting():
        def fget(self):
            return self._conditional_discounting
        return locals()
    conditional_discounting = property(**conditional_discounting())
    
    def data_filter():
        def fget(self):
            return self._data_filter
        return locals()
    data_filter = property(**data_filter())

    def objective_func():
        def fget(self):
            return self._objective_func
        return locals()
    objective_func = property(**objective_func())

    def main_level():
        def fget(self):
            return self._main_level
        return locals()
    main_level = property(**main_level())

    def num_of_consts():
        def fget(self):
            return self._num_of_consts
        return locals()
    num_of_consts = property(**num_of_consts())

    def num_of_subobjs():
        def fget(self):
            return self._num_of_subobjs
        return locals()
    num_of_subobjs = property(**num_of_subobjs())

    def constraints():
        def fget(self):
            return self._constraints
        return locals()
    constraints = property(**constraints())

    def subobjectives():
        def fget(self):
            return self._subobjectives
        return locals()
    subobjectives = property(**subobjectives())
    
    def subobj_strings():
        def fget(self):
            return self._subobj_strings
        return locals()
    subobj_strings = property(**subobj_strings())
    
    def constr_strings():
        def fget(self):
            return self._constr_strings
        return locals()
    constr_strings = property(**constr_strings())

    def operand_index():
        def fget(self):
            return self._operand_index
        return locals()
    operand_index = property(**operand_index())

    def unique_operands():
        def fget(self):
            return self._unique_operands
        return locals()
    unique_operands = property(**unique_operands())
    
    def one_aggr_level():
        def fget(self):
            return self._one_aggr_level
        return locals()
    one_aggr_level = property(**one_aggr_level())


def main():
    pass

if __name__ == '__main__':
    main()
