#!/usr/bin/env python
# encoding: utf-8
"""
Title: SIMO simulation control structure

Classes for storing and handling SIMO simulation controls defined in
simulation XML

*simcontrol.py*

Usage:

Copyright (c) 2007 Simosol Oy. Distributed under the GNU General Public License 2.0.
"""

# these are for running the doctest from src/simo
#import sys
#sys.path.append('..')

from simo.builder import names
from simo.utils import utils
from simo.builder.xmlobject import XmlObject
from simo.builder.modelchain.conditionparser import ConditionParser
from lxml import etree

class SimControlDef(XmlObject):
    """Simulation control for SIMO"""
    def _init__(self, typedef):
        super(SimControlDef, self).__init__(typedef)

    def xml():
        doc = "dictionary of text representation of XML documents"
        def fget(self):
            return self._xml
        def fset(self, xmldata):
            """parse the xml text and validate against the schema, create
               the object representation, and set state"""
            self.errors = set()
            self.warnings = []
            if self._schema is not None:
                try:
                    xmlname, xmltext, lexicon, chains = xmldata
                    self.xml_name = xmlname
                except ValueError:
                    errormsg = 'name, xml content and lexicon instance must '\
                               'be passed'
                    raise ValueError(errormsg)
                try:
                    schematext = self._schema
                    if self._typedef_schema is not None:
                        modschema = self._include_typedefs(schematext)
                    else:
                        modschema = schematext
                    schemaroot = etree.XML(modschema)
                    s = etree.XMLSchema(schemaroot)
                    parser = etree.XMLParser(schema=s, remove_comments=True)
                    root = etree.fromstring(xmltext, parser)
                except etree.XMLSyntaxError, e:
                    raise ValueError(e.args[0])
                else:
                    self._xml[xmlname] = xmltext
                    self.xml_to_obj(xmlname, root, lexicon, chains)
                    if self.errors != set([]):
                        self._ok[xmlname] = False
                        self._all_ok = False
                        errormsg = "errors in xml to object conversion"
                        raise ValueError(errormsg)
                    else:
                        self._ok[xmlname] = True
                        self._all_ok = True
                        for xmlok in self._ok.values():
                            if not xmlok:
                                self._all_ok = False
            else:
                raise ValueError('No schema set')
        return locals()
    xml = property(**xml())

    def xml_to_obj(self, xmlname, root, lexicon, chains):
        """
        Construct lexicon translations tables from the parsed translation
        table xml. Data level names and variable names get their own
        translation tables
        """
        self.from_module = 'Simulation control'
        self.lexicon = lexicon
        ns = self.elem_namespace(root.tag)
        sc = SimControl(ns, root, self, chains)
        self.obj[xmlname] = sc
        self.lexicon = None


class SimControl(object):
    """
    Class for storing simulation control structures
    """

    def __init__(self, ns, root, validator, chains):
        """docstring for __init__"""
        self.ns = ns
        self._chain_coll = chains
        self._validator = validator
        self.main_level = None
        self.built_ins = {}
        self.growth_season_start_month = None
        self.growth_season_start_day = None
        self.growth_season_end_month = None
        self.growth_season_end_day = None
        self.stop_logic = None
        self.time_spans = None
        self.forced_operations = None
        self.init_variables = None
        self.output_constraints = None
        self.max_model_chain_depth = 0
        self.opres_cfiers = set([])
        self.opres_vars = set([])
        self.cf_cfiers = set([])
        self.all_branching_groups = None
        self.memory_models = set([])
        # parse optimization task xml document
        self._parse(root)
        self._validator = None
        self._condition_parser = None
        self._chain_coll = None
        self.use_data_date = None
        self.use_today = None
        self.initial_date = None
        self.fixed_dp_date = False

    def _parse(self, root):
        mainlevel = root.find(self.ns+'main_level').text
        self._condition_parser = ConditionParser(self._validator, mainlevel)
        levelind = self._validator.level_ind(mainlevel)
        self.main_level = levelind
        # process the built-in attribute mappings
        main_lev_built_ins = (names.TIME_STEP,
                              names.YEAR,
                              names.MONTH,
                              names.DAY)
        built_ins = root.find(self.ns+'built_in_attributes')
        for name in main_lev_built_ins:
            bivar = built_ins.find(self.ns + name).text
            ind = self._validator.variable_ind(mainlevel, bivar)
            if ind == (None, None):
                errormsg = "built-in attribute '%s' set to point to "\
                           "attribute '%s' which is not defined in lexicon "\
                           "for level simulation" % (name, bivar)
                self._validator.add_error(errormsg)
            else:
                self.built_ins[name] = ind
        iteration = built_ins.find(self.ns+names.ITERATION).text
        # NB: make sure that iteration name is NOT 'iteration' as that would
        # lead to duplicate column names in result databases!!!
        if iteration == 'iteration':
            errormsg = "built-in attribute 'iteration' name cannot be "\
                       "'iteration', try for example 'iter' instead!"
            self._validator.add_error(errormsg)
        ind = self._validator.variable_ind('simulation', iteration)
        if ind == (None, None):
            errormsg = "built-in attribute 'iteration' set to point to "\
                       "attribute '%s' which is not defined in lexicon for "\
                       "level '%s'" % (iteration, mainlevel)
            self._validator.add_error(errormsg)
        else:
            self.built_ins[names.ITERATION] = ind
        # Check if simulation has a stop logic defined and construct the
        # stop logic condition if yes
        stoplogic = None
        stopelem = root.find(self.ns+'stop_logic')
        if stopelem is not None:
            self.cur_eval_level = None
            stoplogic = self._condition_parser.parse(stopelem.text)
        self.stop_logic = stoplogic
        # Process the simulation control
        control = root.find(self.ns + 'control')
        growthstart = control[0].text
        growthend = control[1].text
        # Go through the time spans defined in the simulation control and
        # construct the time span control structures
        # Initialize empty set for identifying duplicate branching groups
        bgnames = {}
        all_branching_groups = {}
        spans = []
        forcedops = False
        for i, span in enumerate(control[2:]):
            # Parse the current time span structure
            timespan = TimeSpan()
            timespan.parse_timespan(self.ns, span[0])
            # Which steps should be stored to the result db?
            timespan.save_steps_to_db = span[1].text
            # Process the model chain files in the current timespan
            mccs = self._process_mccs(span[2:])
            if mccs is None:
                break
            # check if forced operations are on
            if mccs['forced_operation'] == []:
                timespan.forced_operations = False
            else:
                timespan.forced_operations = True
            timespan.chain_collections = mccs
            if names.INIT in mccs:
                for mcc in mccs[names.INIT]:
                    self.memory_models.update(mcc.memory_models)
                    self.cf_cfiers.update(mcc.cf_cfiers)
            if names.SIMULATION in mccs:
                for mcc in mccs[names.SIMULATION]:
                    self.memory_models.update(mcc.memory_models)
                    self.cf_cfiers.update(mcc.cf_cfiers)
            if names.OPERATION in mccs:
                for mcc in mccs[names.OPERATION]:
                    self.opres_cfiers.update(mcc.opres_cfiers)
                    self.opres_vars.update(mcc.opres_vars)
                    self.cf_cfiers.update(mcc.cf_cfiers)
                    self.memory_models.update(mcc.memory_models)
                    if mcc.branching_groups is None:
                        continue
                    for bgname, bitem in mcc.branching_groups.items():
                        if bgname in bgnames and mcc.name != bgnames[bgname]:
                            errormsg = "Duplicate branching group name '%s' "\
                                    "in operation model chains '%s' and '%s'"\
                                    " (%d. timespan)" % (bgname,
                                    bgnames[bgname], mcc.name, i+1)
                            self._validator.add_error(errormsg)
                        else:
                            bgnames[bgname] = mcc.name
                            timespan.branching_groups[bgname] = bitem
                        # duplicate branching groups will overwrite existing
                        # branching groups
                        all_branching_groups[bgname] = bitem
            if names.FORCED_OPERATION in mccs:
                for mcc in mccs[names.FORCED_OPERATION]:
                    self.opres_cfiers.update(mcc.opres_cfiers)
                    self.opres_vars.update(mcc.opres_vars)
                    self.cf_cfiers.update(mcc.cf_cfiers)
                    self.memory_models.update(mcc.memory_models)
                    if mcc.branching_groups is None:
                        continue
                    for bgname, bitem in mcc.branching_groups.items():
                        if bgname in bgnames and mcc.name != bgnames[bgname]:
                            errormsg = "Duplicate branching group name '%s' "\
                                    "in forded operation model chains '%s' "\
                                    "and '%s' (%d. timespan)" % (bgname,
                                    bgnames[bgname], mcc.name, i+1)
                            self._validator.add_error(errormsg)
                        else:
                            bgnames[bgname] = mcc.name
                            timespan.branching_groups[bgname] = bitem
                        # duplicate branching groups will overwrite existing
                        # branching groups
                        all_branching_groups[bgname] = bitem
            spans.append(timespan)
        # Store the simulation control definitions into SimControl
        self._process_init_variables(root.find(self.ns + 'init_variables'))
        constraints = root.find(self.ns + 'output_constraints')
        if constraints is not None:
            self._process_output_constraints(constraints)
        gstart = growthstart.split('-')
        self.growth_season_start_month = int(gstart[-2])
        self.growth_season_start_day = int(gstart[-1])
        gend = growthend.split('-')
        self.growth_season_end_month = int(gend[-2])
        self.growth_season_end_day = int(gend[-1])
        self.time_spans = spans
        self.all_branching_groups = all_branching_groups
        self.forced_operations = timespan.forced_operations

    def _process_init_variables(self, varnodes):
        init_variables = []
        if varnodes is not None:
            for i in range(len(varnodes)):
                name = varnodes[i][0].text
                level = varnodes[i][1].text
                value = float(varnodes[i][2].text)
                nl = self._validator.variable_ind(level, name)
                if nl==(None, None):
                    continue
                init_variables.append((nl, value))
        self.init_variables = init_variables

    def _process_output_constraints(self, constnodes):
        if constnodes.get('active') == 'off':
            return
        constraints = {}
        numnodes = len(constnodes)
        for i in xrange(numnodes):
            level = constnodes[i][0].text
            levelind = self._validator.level_ind(level)
            if levelind is None:
                continue
            numnodes2 = len(constnodes[i][1])
            if numnodes2 == 1:
                if constnodes[i][1][0].text.strip() == '*':
                    constraints[levelind] = None
                    continue
            variables = []
            for j in xrange(numnodes2):
                var = constnodes[i][1][j].text
                varind = self._validator.variable_ind(level, var)
                if varind==(None, None):
                    continue
                variables.append(varind[1])
            constraints[levelind] = variables
        self.output_constraints = constraints

    def _process_mccs(self, chainnodes):
        """
        Gets the model chain collections referenced in the simulation control
        xml and stores them by model chain type
        """
        elem_name = self._validator.elem_name
        mccs = {names.INIT:[], names.SIMULATION:[], names.OPERATION: [],
                  names.FORCED_OPERATION: []}
        numnodes = len(chainnodes)
        for i in xrange(numnodes):
            if elem_name(chainnodes[i].tag) == 'init_chains':
                chaintype = names.INIT
            elif elem_name(chainnodes[i].tag) == 'simulation_chains':
                chaintype = names.SIMULATION
            elif elem_name(chainnodes[i].tag) == 'operation_chains':
                chaintype = names.OPERATION
            elif elem_name(chainnodes[i].tag) == 'forced_operation_chains':
                chaintype = names.FORCED_OPERATION
            for chainfile in chainnodes[i]:
                cname = chainfile.text
                if cname is not None:
                    chaindict = self._chain_coll.get(chaintype)
                    if chaindict is None:
                        errormsg = "No model chain collection of type '%s'"\
                                   % chaintype
                        self._validator.add_error(errormsg)
                        break
                    chain_coll = chaindict.get(cname)
                    if chain_coll is not None:
                        mccs[chaintype].append(chain_coll)
                        if self.max_model_chain_depth < chain_coll.depth:
                            self.max_model_chain_depth = chain_coll.depth
                    else:
                        errormsg = "modelchain collection not found for"\
                                   " chain collection name '%s'" % (cname)
                        self._validator.add_error(errormsg)
        return mccs

class TimeSpan(object):
    def __init__(self):
        """
        Method: __init__

        Initialize TimeSpan object.
        """
        self.time_step = None
        self.operation_step = None
        self.unit = None
        self.ending = None
        self.save_steps_to_db = None
        self.forced_operations = None
        self.chain_collections = None
        self.branching_groups = {}

    def parse_timespan(self, ns, elem):
        """
        Parse timespan time definitions.

        ns -- xml namespace as string
        elem -- timespan xml structure as ElementTree object
        """
        # Get the timestep of the timespan
        self.time_step = int(elem.find(ns + 'time_step').text)
        # Get the operation step
        operstep = elem.find(ns + 'operation_step')
        if operstep is not None:
            operstep = int(operstep.text)
        else:
            # If operationstep is not defined by user, it is assumed
            # to be same as the timestep
            operstep = self.time_step
        self.operation_step = operstep
        # Get the simulation time unit (year, month, day)
        self.unit = elem.find(ns + 'time_unit').text
        # Get the timespan length which is given either as a number of
        # steps or as an ending date
        steps = elem.find(ns + 'steps')
        if steps is not None:
            endtype = 'steps'
            endtarget =  int(steps.text)
        enddate = elem.find(ns + 'end_date')
        if enddate is not None:
            endtype = 'end_date'
            endtarget = utils.str2date(enddate.text)
        self.ending = Ending(endtype, endtarget)


class Ending(object):
    def __init__(self, endtype, target):
        self.type = endtype
        self.target = target


def main():
    import doctest
    doctest.testfile('test/simcontrol-py.rst')

if __name__ == '__main__':
    main()
