#!/usr/bin/env python
# encoding: utf-8
"""
*xmlobject.py*

XML definition base class

Copyright (c) 2008 Simosol Oy. Distributed under the GNU General Public
License 2.0.
"""
# these are for running the doctest from src/simo
#import sys
#sys.path.append('..')
import re
from lxml import etree

class XmlObject(object):

    """The second base class for SIMOs xml definitions, adds the schema and
       xml properties to the class as well as the object placeholder"""

    def __init__(self, typedef):
        super(XmlObject, self).__init__()
        self._xml = {}
        self.obj = {}
        self._ok = {}
        self.errors = set()
        self.warnings = []
        self._typedef_schema = None
        self._schema = None
        self._all_ok = False
        self.lexicon = None
        self.typedef_schema = typedef

    def ok():
        doc = "Dictionary of booleans indicating whether typedef, schema and "\
              "xml are set successfully by XML document name"
        def fget(self):
            return self._ok
        return locals()
    ok = property(**ok())

    def all_ok():
        doc = "Boolean indicating whether typedef, schema and all xml are set "\
              "successfully"
        def fget(self):
            return self._all_ok
        return locals()
    all_ok = property(**all_ok())

    def schema():
        doc = "text representation of XML Schema document"
        def fget(self):
            return self._schema
        def fset(self, schematext):
            """parse the schema text to a schema instance and create a parser
               that uses it
            """
            if schematext != self._schema and schematext is not None:
                if self._typedef_schema is not None:
                    modschema = self._include_typedefs(schematext)
                else:
                    modschema = schematext
                try:
                    schemaroot = etree.XML(modschema)
                    s = etree.XMLSchema(schemaroot)
                    self._schema = schematext
                except etree.XMLSyntaxError, e:
                    self._schema = None
                    raise ValueError(e.args[0])
                finally:
                    self._all_ok = False
                    for x in self._ok:
                        self._ok[x] = False
        return locals()
    schema = property(**schema())

    def typedef_schema():
        doc = "text representation of SIMO specific type definitions XML "\
              "Schema document which will be merged to theactual schema "\
              "document when parsing it"
        def fget(self):
            return self._typedef_schema
        def fset(self, schematext):
            """parses the schema text to validate it and sets the property
               value to the text representation if succesfull."""
            try:
                schemaroot = etree.XML(schematext)
                s = etree.XMLSchema(schemaroot)
                self._typedef_schema = schematext
            except (etree.XMLSyntaxError, ValueError), e:
                self._typedef_schema = None
                raise ValueError(e.args[0])
            finally:
                self._all_ok = False
                for x in self._ok:
                    self._ok[x] = False
        return locals()
    typedef_schema = property(**typedef_schema())

    def xml():
        doc = "dictionary of text representation of XML documents"
        def fget(self):
            return self._xml
        def fset(self, xmldata):
            """parse the xml text and validate against the schema, create
               the object representation, and set state"""
            self.errors = set()
            self.warnings = []
            if self._schema is not None:
                try:
                    xmlname, xmltext, lexicon = xmldata
                    self.xml_name = xmlname
                except ValueError:
                    errormsg = 'name, xml content and lexicon instance must '\
                               'be passed'
                    raise ValueError(errormsg)
                try:
                    schematext = self._schema
                    if self._typedef_schema is not None:
                        modschema = self._include_typedefs(schematext)
                    else:
                        modschema = schematext
                    schemaroot = etree.XML(modschema)
                    s = etree.XMLSchema(schemaroot)
                    parser = etree.XMLParser(schema=s, remove_comments=True)
                    root = etree.fromstring(xmltext, parser)
                except etree.XMLSyntaxError, e:
                    raise ValueError(e.args[0])
                else:
                    self._xml[xmlname] = xmltext
                    self.xml_to_obj(xmlname, root, lexicon)
                    if self.errors != set([]):
                        self._ok[xmlname] = False
                        self._all_ok = False
                        errormsg = "errors in xml to object conversion"
                        raise ValueError(errormsg)
                    else:
                        self._ok[xmlname] = True
                        self._all_ok = True
                        for xmlok in self._ok.values():
                            if not xmlok:
                                self._all_ok = False
            else:
                raise ValueError('No schema set')
        return locals()
    xml = property(**xml())

    def elem_name(self, elemtag):
        """
        Separates the name from the element namespace, lowercases, and returns
        the name
        """
        part = str(elemtag).split("}")
        if len(part) > 1:
            return str.lower(part[1])
        else:
            return str.lower(part[0])

    def elem_namespace(self, elemtag):
        """
        Separates element's name and namespace, and returns the namespace
        """
        part = str(elemtag).split("}")
        if len(part) > 1:
            return str.lower(part[0]) + "}"
        else:
            return ''

    def reparse(self, lexicon=None):
        """used to explicitly parse the schema and xml content again after
           typedef or schema changes
        """
        self.schema = self.schema
        for name, xml in self.xml.items():
            self.xml = (name, xml, lexicon)

    def _include_typedefs(self, schematext):
        """embeds the typedef schema into another schema before it's parsed"""
        tdstr = '<\s*xs:include\s*schemaLocation\s*=\s*'\
                '"[/\w\s\d\.\\\]*Typedefs_SIMO.xsd"\s*/>'
        patt = re.compile(tdstr)
        if patt.search(schematext):
            schematext = patt.sub('', schematext)
            startpatt = re.compile('<xs:schema[\s\w:/\n="\.-]+>')
            typedefs = etree.XML(self._typedef_schema)
            for tdef in typedefs:
                startmatch = startpatt.search(schematext)
                matchstr = startmatch.group()
                incl = matchstr + '\n' + etree.tostring(tdef)
                schematext = startpatt.sub(incl, schematext)
        return schematext

    def xml_to_obj(self, xmlname, root, lexicon):
        """placeholder for subclass implementations of xml to object
           conversion
        """
        pass

    def add_model(self, mtype, model):
        """
        Adds the model to the lexicon or logs an error if it already exists
        """
        if self.lexicon is None:
            errormsg = "No lexicon set when adding model"
            self.add_error(errormsg)
            return
        if mtype in self.lexicon.models and model in self.lexicon.models[mtype]:
            errormsg = "Model '%s' of type '%s' already defined in in lexicon" \
                       % (model, mtype)
            self.add_error(errormsg)
            return
        else:
            return self.lexicon.add_model(mtype, model)

    def level_ind(self, level):
        if self.lexicon is None:
            errormsg = "No lexicon set when validating level"
            self.add_error(errormsg)
            return None
        ind = self.lexicon.get_level_ind(level)
        if ind is None:
            errormsg = "Level '%s' not found in lexicon" % (level)
            self.add_error(errormsg)
        return ind

    def level_name(self, levelind):
        if self.lexicon is None:
            errormsg = "No lexicon set when validating level"
            self.add_error(errormsg)
            return None
        name = self.lexicon.get_level_name(levelind)
        if name is None:
            if levelind is None:
                return None
            errormsg = "Level index '%d' not found in lexicon" % (levelind)
            self.add_error(errormsg)
        return name

    def operation_ind(self, operation):
        if self.lexicon is None:
            errormsg = "No lexicon set when validating operation model"
            self.add_error(errormsg)
            return None
        ind = self.lexicon.get_operation_ind(operation)
        if ind is None:
            errormsg = "Operation '%s' not found in lexicon" % (operation)
            self.add_error(errormsg)
        return ind

    def variable_ind(self, level, var, already_active=False):
        if self.lexicon is None:
            errormsg = "No lexicon set when validating variable"
            self.add_error(errormsg)
            return (None, None)
        ind = self.lexicon.get_variable_ind(level, var, already_active)
        if ind == (None, None):
            if already_active:
                errormsg = "Variable '%s' not found at level '%s' in lexicon"\
                           " or it's not set to be computed during simulation"\
                           % (var, level)
            else:
                errormsg = "Variable '%s' not found at level '%s' in lexicon"\
                           % (var, level)
            self.add_error(errormsg)
        return ind

    def add_warning(self, msg):
        warning = msg + " for %s '%s'" % (self.from_module, self.xml_name)
        self.warnings.append(warning)

    def add_error(self, msg):
        errormsg = msg + " for %s '%s'" % (self.from_module, self.xml_name)
        self.errors.add(errormsg)

    def clear_warnings_and_errors(self):
        self.warnings = []
        self.errors = set()

def main():
    import doctest
    doctest.testfile('./test/xmlobject-py.rst')

if __name__ == '__main__':
    main()
