import os
import glob

# these are for running the doctest from src/simo
#import sys
#sys.path.append('..')

from simo.builder import simoclasses as sc
from simo.builder import names

import cPickle

###############################################################################
#
# class SimoDB
#
###############################################################################
class Validator(object):
    def __init__(self):
        self.errors = set([])

    def add_error(self, errormsg):
        self.errors.add(errormsg)

class SimoDB(object):
    """
    Class: SimoDB

    This class handels a database containing modelbase and xmlobject instances.

    Attributes:
        file - the file for the database (path/to/something.fs)
        create_new - if True, the existing database will be wiped out.
    """
    def __init__(self, file, create_new=False, read_only=False):
        #This import is here to retain regular run/debug.
        from simo.builder.names import mtypes
        if create_new:
            try:
                [os.remove(i) for i in glob.glob(file+'*')]
            except OSError:
                pass
        self.mtypes = mtypes
        self.typenames = {}
        self.read_only = read_only
        self.always_pack = False #Set to true if needed, but it shouldn't be.
        self.file = file
        if os.path.exists(file):
            with open(file, 'rb') as f:
                self.dbroot = cPickle.load(f)
        else:
            self.dbroot = {}
        if not read_only:
            if 'valid' not in self.dbroot:
                self.dbroot['valid'] = {}
                for sclass in sc.simoclasses:
                    self.dbroot['valid'][sclass] = {'valid':False}
                    if sclass in (sc.MODELBASE_DEF, sc.MODELCHAIN_DEF):
                        self.dbroot['valid'][sclass]['types']={}
                    else:
                        self.dbroot['valid'][sclass]['names']={}

    def typedef():
        doc = "The commen type definitions XML Schema used collectively by "\
              "several other XML Schemas"
        def fset(self, typedef):
            self.dbroot['typedef'] = typedef
        def fget(self):
            return self.dbroot.get('typedef')
        return locals()
    typedef = property(**typedef())

    def close(self):
        if self.read_only:
            return
        else:
            with open(self.file, 'wb') as f:
                cPickle.dump(self.dbroot, f, -1)

    def commit(self):
        if self.read_only:
            return
        else:
            with open(self.file, 'wb') as f:
                cPickle.dump(self.dbroot, f, -1)

    def get_names(self, simoclass, atype=None):
        """
        Returns a list of names defined for the given SIMO XmlObject subclass

        simoclass -- SIMO XmlObject subclass
        atype -- model/chain type must be specified for modelbase and
                 modelchain objects
        """
        if simoclass not in self.dbroot:
            return None
        xmlobj = self.dbroot[simoclass]
        if atype is not None:
            if atype in xmlobj.xml:
                return xmlobj.xml[atype].keys()
            else:
                return None
        else:
            return xmlobj.xml.keys()

    def get_obj(self, simoclass, name=None, atype=None):
        """
        Returns the stored class instance for the given SIMO class and xml
        name.  For modelbase and modelchain the model/chain type must be
        specified. If no name is specified, returns all name, obj combinations

        simoclass -- the XmlObject subclass
        name -- the name of the object containing the XML
        atype -- model/chain type
        """
        xmlobj = self.dbroot.get(simoclass)
        if xmlobj is None:
            return None
        if atype is None:
            if name is None:
                return xmlobj.obj
            else:
                obj = xmlobj.obj.get(name)
                if obj is not None and simoclass==sc.SIMULATION_CONTROL_DEF:
                    for ts in obj.time_spans:
                        for mccs in ts.chain_collections.values():
                            for mcc in mccs:
                                self._load_model_implementations(mcc)
                return obj
        else:
            if atype not in xmlobj.obj:
                return None
            else:
                if name is None:
                    return xmlobj.obj[atype]
                else:
                    obj = xmlobj.obj[atype].get(name)
                    # special case for modelchains, they must load the model
                    # implementations as well
                    if obj is not None and simoclass==sc.MODELCHAIN_DEF:
                        self._load_model_implementations(obj)
                    return obj

    def get_schema(self, simoclass, mtype=None):
        """
        Returns the schema associated with the given SIMO class. For models
        the model type is needed as well
        """
        if mtype is not None and mtype not in self.mtypes:
            raise ValueError(mtype, 'not in allowed model types:', self.mtypes)
        if simoclass in self.dbroot:
            if mtype is not None:
                allschema = self.dbroot[simoclass].schema
                return allschema[mtype]
            else:
                return self.dbroot[simoclass].schema
        else:
            return None

    def get_xml(self, simoclass, name=None, atype=None):
        """
        Returns the stored xml text for the given SIMO class and xml name.
        For modelbase and modelchain the model/chain type must be specified.
        If no name is given, returns all name, xml combinations and the all_ok
        property.

        simoclass -- the XmlObject subclass
        name -- the name of the object containing the XML
        atype -- model/chain type
        """
        xmlobj = self.dbroot.get(simoclass)
        if xmlobj is None:
            return None, None
        if atype is None:
            if name is None:
                return (xmlobj.xml, xmlobj.ok)
            else:
                return (xmlobj.xml.get(name), xmlobj.ok.get(name))
        else:
            if atype not in xmlobj.obj:
                return None, None
            else:
                if name is None:
                    return (xmlobj.xml[atype], xmlobj.ok)
                else:
                    return (xmlobj.xml[atype].get(name), xmlobj.ok.get(name))
                
    def get_valid(self, simoclass=None):
        if simoclass is None:
            return self.dbroot['valid']
        elif simoclass not in self.dbroot['valid']:
            return None
        else:
            return self.dbroot['valid'][simoclass]

    def set_schema(self, simoclass, schema, mtype=None):
        """
        Sets the schema used to validate the given SIMO class

        simoclass -- name of the SIMO class for which the schema is stored
        schema -- XML Schema text
        mtype -- for models a model type is required as well
        """
        if mtype is not None and mtype not in self.mtypes:
            raise ValueError(mtype,'not in allowed model types:', self.mtypes)
        if simoclass not in self.dbroot:
            pythonclass = self._get_python_class(simoclass)
            if self.typedef is None:
                raise ValueError('Typedef not set')
            self.dbroot[simoclass] = pythonclass(self.typedef)
        if mtype is not None:
            self.dbroot[simoclass].schema = (mtype, schema)
        else:
            self.dbroot[simoclass].schema = schema
        self._flag_depends(simoclass)
        self._set_valids(simoclass)

    def set_xml(self, simoclass, xml):
        """
        Sets the XML for the given SIMO class instance. That triggers the
        class instance creation as well

        simoclass -- the SIMO class for the XML
        xml -- a tuple containing the name used for the xml/class instance
               pair, xml text and lexicon (see exceptions below)
        """
        if not simoclass in self.dbroot:
            errormsg = "Schema has not been set for '%s'" % (simoclass)
            raise ValueError(errormsg)
        model_type = self._check_xml_content(simoclass, xml)
        xmlname = xml[0]
        xmlobj = self.dbroot[simoclass]
        toperror = None
        try:
            xmlobj.xml = xml
            self._set_valids(simoclass, xmlname, model_type, True)
        except ValueError, e:
            toperror = e.args[0]
            xmlobj.ok[xmlname] = False
            self._set_valids(simoclass, xmlname, model_type, False)
        self._flag_depends(simoclass)
        return (toperror, xmlobj.ok[xmlname], xmlobj.warnings, xmlobj.errors)

    def _check_xml_content(self, simoclass, xml):
        errormsg = ''
        checked = False
        model_type = None
        if simoclass=='lexicon':
            if len(xml)!=2:
                errormsg = 'To set the lexicon the name and xml must be given'
            checked = True
        elif simoclass=='modelbase':
            if len(xml)!=5:
                errormsg = 'To set modelbase xml these must be given: '\
                        'name, xml text, lexicon instance, '\
                        'directory path and model type'
            model_type = xml[4]
            checked = True
        elif simoclass=='modelchain':
            if len(xml)!=5:
                errormsg = 'To set modelchain xml these must be given: '\
                        'name, xml text, lexicon instance, chain type and '\
                        'model implementation'
            model_type = xml[3]
            checked = True
        elif simoclass=='text2operation':
            if len(xml)!=4:
                errormsg = 'To set text2operation xml these must be given: '\
                    'name, xml text, lexicon instance, operation2modelchains'
            checked = True
        elif simoclass=='simulation_control':
            if len(xml)!=4:
                errormsg = 'To set modelchain xml these must be given: '\
                        'name, xml text, lexicon instance, chains'
            checked = True
        elif not checked and len(xml)!=3:
            errormsg = 'To set xml these must be given: name, xml text and '\
                       'lexicon instance, simoclass %s' % (simoclass)
        if errormsg!='':
            raise ValueError(errormsg)
        
        return model_type

    def _get_python_class(self, simoclass):
        pythonclass = sc.simoclasses.get(simoclass)
        if pythonclass is None:
            errormsg = "unsupported SIMO class type '" + simoclass + "'"
            raise ValueError(errormsg)
        return pythonclass

    def _load_model_implementations(self, mcc):
        """
        Loads the model implementations which can't be stored in a pickle for
        all models in the modelchains

        mcc -- ModelChainCollection instance
        """
        validator = Validator()
        for chain in mcc.chains:
            for task in chain.tasks:
                self._load_task_model(task, validator)
        if len(validator.errors)>0:
            combined = ''
            for error in validator.errors:
                combined += error +'\n'
            combined += 'Trying to load from %s' % os.getcwd()
            raise ValueError(combined)

    def _load_task_model(self, task, validator):
        """
        task -- Task instance
        """
        model = task.model
        if model is not None:
            nomodelset = set([names.CASH_FLOW_TABLE,
                              names.PARAMETER_TABLE])
            if model.type in nomodelset:
                return
            elif model.type==names.OPERATION and model.optype=='cash_flow':
                return
            else:
                model.load_function(validator)
            if model.type==names.OPERATION and \
               task.model_param.cash_flow_model is not None:
                task.model_param.cash_flow_model.load_function(validator)
            teststr = "Cannot load operation model function. Model name, "\
                    "library name and library directory required!"
            if teststr in validator.errors:
                errormsg = 'Cannot load operation model function. Model name,'\
                    ' library name and library directory required\n'\
                    'name: %s, lib: %s, dirs: %s'\
                    % (task.model._name, task.model._lib, task.model._dirs)
                validator.errors.add(errormsg)
                validator.errors -= set([teststr])
        else:
            for stask in task.subtasks:
                self._load_task_model(stask, validator)

    def _flag_depends(self, simoclass, depth=0):
        if depth==100:
            raise Exception('Endless recursion in depends (%s)!' % simoclass)
        if simoclass in sc.simodepends:
            for sclass in sc.simodepends[simoclass]:
                self._set_valids(sclass)
                self._flag_depends(sclass, depth+1)
                
    def _set_valids(self, simoclass, name=None, model_type=None, valid=False):
        valbase = self.dbroot['valid'].get(simoclass)
        all_valid = True
        if simoclass in (sc.MODELBASE_DEF, sc.MODELCHAIN_DEF):
            if name is not None:
                if model_type not in valbase['types'].keys():
                    valbase['types'][model_type] = {}
                valbase['types'][model_type][name] = valid
                for mtype in valbase['types'].values():
                    if False in mtype.values():
                        all_valid = False
                        break
            else:
                all_valid = False
                for mtype in valbase['types'].values():
                    for key in mtype.keys():
                        mtype[key] = False
        else:
            if name is not None:
                valbase['names'][name] = valid
                if False in valbase['names'].values():
                    all_valid = False
            else:
                for key in valbase['names'].keys():
                    valbase[key] = False
                all_valid = False
                
        if all_valid:
            valbase['valid'] = True
        else:
            valbase['valid'] = False

def main():
    import doctest
    doctest.testfile('./test/db-py.rst')

if __name__ == '__main__':
    main()
