#!/usr/bin/env python
# encoding: utf-8
"""
Title: SIMO data importing

Module for importing external data from text files to the internal database.

*importdata.py*

Copyright (c) 2007 Simosol Oy. Distributed under the GNU General Public License 2.0.
"""

import sys
import time
import datetime
from collections import defaultdict
from operator import itemgetter

CALMER = 200
SIMULATION_OBJ_ID = 'simulation'



class DataImporter(object):
    """
    Class: DataImporter

    Class for importing data into SIMO input database
    """
    def __init__(self, inputdb, mapping, import_date, logger, logname,
                 lexicon, max_errors):
        """
        Initialize SimoImporter for importing different data formats into
        SIMO simulation environment.

        inputdb -- input database, DataDB instance
        import_date -- boolean value indicating whether initial date should
                      be get from data or from initial date definition
        logger -- SimoLogger instance
        logname -- import log name as string
        """
        self.input_db = inputdb
        self.mapping = mapping
        self.import_date = import_date
        self.logger = logger
        self.log_name = logname
        self.error_count = 0
        self.max_errors = max_errors
        self.id_counter = 1
        self.id2lev = {}
        self.oids = {}
        self.lexicon = lexicon

    def import_data(self, format, buffers, sim_level, separator=None,
                    skip_first=False, level_ind=None, data_date=None,
                    clear_old=True, parent_id=None, commit=True,
                    month_first=False, encodings=('utf8',)):
        self.clear_old = clear_old
        self.curr_obj = {'level': None, 'id': None, 'pid': None}
        self.sim_level = sim_level
        self.main_level = self.lexicon.get_level_name(\
                            self.input_db.get_main_level())
        if parent_id is None:
            self.parent_id = {}
        else:
            self.parent_id = parent_id
        self.global_data_date = data_date
        self.month_first = month_first
        self.encodings = encodings
        self.initial_date = defaultdict(dict)
        self.object_queue = defaultdict(list)
        self.rejected = set([])
        self.counter = 0
        self.reject_count = 0
        self.error_count = 0
        self.id_counter = 1
        self.abort_object = False
        self.errors = set([])
        self.warnings = set([])
        self.terminate_import = False
        self.format = format
        self.import_ids = set([])
        self.oids = {}
        # sanity check for the simulation level
        if self.import_date:
            if self.mapping.level_list[0].date_pos == None:
                msg = "If 'import_date' is set to True, date row position " \
                    "must be given in conversion definition"
                self._add_error(msg)
                self.terminate_import = True
        # Import the data
        if not self.terminate_import:
            msg = 'Importing data...'
            self.logger.log_message(self.log_name, 'info', msg)
            if format == 'by_level':
                # Input data is in multiple files, one data level per file
                # if level_ind is given, only data for those levels is imported
                # otherwise it's assumed that there's a buffer for each level
                if level_ind is None:
                    level_ind = range(len(self.mapping.level_list))
                bind = 0
                for lind in level_ind:
                    self.curr_level_ind = lind
                    self._import_buffer(buffers[bind], separator, skip_first)
                    bind += 1
            elif format == 'inlined':
                # Data is in single (or multiple) file(s), all data levels are
                # in a single file
                for buffer in buffers:
                    self._import_buffer(buffer, separator, skip_first)
            # Log user info
            msg = 'In total %d simulation units processed' % self.counter
            self.logger.log_message(self.log_name, 'info', msg)
            msg = 'In total %d simulation units imported' % \
                (self.counter - self.reject_count)
            self.logger.log_message(self.log_name, 'info', msg)
        if commit:
            self.input_db.db.commit()
        try:
            self.input_db.db.vacuum_analyze()
        except Exception:
            pass
        return self.terminate_import


    def decode(self, text):
        """
        Decode the given text with the encodings set on the import_data call.

        text -- text to be decoded
        """
        text = text.strip()
        for encoding in self.encodings:
            try:
                return unicode(text, encoding)
            except UnicodeDecodeError:
                pass

        msg = ("Failed to decode import value %s"
               % (text.decode(errors='replace').encode(errors='replace'),))
        self._add_error(msg)
        # Even though we failed, we'll still get the text as well as we can
        return text.decode(errors='replace')


    def _add_object(self, sim_level, mainrow, levelnum, levelname, rowtype,
                    item):
        if levelnum == 0 and mainrow:
            self.counter += 1
            if self.counter % CALMER == 0:
                msg = 'importing simulation unit %d ...' % (self.counter,)
                self.logger.log_message(self.log_name,'info', msg)
        oid = self._construct_id(item, levelnum, mainrow)
        if oid is None:
            if sim_level:
                self.reject_count += 1
            return
        lind = self.lexicon.level_ind[levelname]

        if sim_level and self.clear_old:
            self.input_db.drop_id(oid)
        pid = self._get_parent_id(levelnum, item)

        plind = None
        if self.mapping.level_list[levelnum].linkid_pos is not None:
            parent_level = self.mapping.level_list[levelnum].linkid_pos[0]
            plind = self.lexicon.level_ind[parent_level]

        leaf_level = levelnum == (len(self.mapping.level_list) - 1)
        uid = self._construct_unique_id(lind, oid, pid, leaf_level)

        if uid is None:
            if sim_level:
                self.reject_count += 1
            return

        if pid in self.rejected:
            self.rejected.add(uid)
            return

        nowoid = self.curr_obj['id']
        nowpid = self.curr_obj['pid']
        nowlevel = self.curr_obj['level']
        if nowoid != oid or nowlevel != levelname or nowpid != pid:
            if nowoid is not None:
                self._insert_object()
            datadict = {'id': uid,
                        'oid': oid,
                        'parent id': pid,
                        'values': []}
            if plind is not None:
                datadict['parent level'] = plind
            self.curr_obj = {'level': levelname,
                             'id': oid,
                             'pid': pid,
                             'ddate': None,
                             'data': datadict}
            self.parent_id[levelnum] = uid
        self.curr_obj['ddate'] = self._get_data_date(mainrow, levelnum, uid,
                                                     pid, item)
        values = self._convert_data(levelname, rowtype, item)
        self.curr_obj['data']['values'] += values

    def _add_error(self, msg):
        """
        Adds the error message to the data import error list and writes the
        message to error log
        """
        self.error_count += 1
        if msg not in self.errors:
            self.errors.add(msg)
            try:
                self.logger.log_message(self.log_name, 'error', msg)
            except UnicodeEncodeError:
                self.logger.log_message(self.log_name, 'error',\
                                        'Invalid alphabets in error string')
        # Check if maximum error counts are exceeded
        if self.max_errors >= 0 and self.error_count > self.max_errors:
            msg = "Maximum number of errors (%s) exceeded!" \
                % str(self.max_errors)
            self._terminate_data_import(msg)

    def _add_warning(self, msg):
        """
        Adds the warning message to the data import error list and writes the
        message to error log
        """
        if msg not in self.warnings:
            self.warnings.add(msg)
            try:
                self.logger.log_message(self.log_name, 'warning', msg)
            except UnicodeEncodeError:
                self.logger.log_message(self.log_name, 'warning',\
                                        'Invalid alphabets in error string')

    def _add_simulation_object(self, datadate):
        """
        Adds the root object for data, all the top level objects in the
        mapping definition are added as child for this object

        datadate - date for the simulation object
        """
        data = {
            'values': [],
            'id': SIMULATION_OBJ_ID,
            'oid': SIMULATION_OBJ_ID,
            'parent id': None}
        valuetuple = (datadate, data)
        datadict = {self.sim_level: [valuetuple]}
        self.input_db.add_data_from_dictionary(datadict, 0, 0)

    def _check_for_rejection(self, var, val):
        """
        Check if current variable value is in the rejection list

        var -- variable name as string
        val -- variable value as string
        """
        if var in self.mapping.object_rejection:
            # If object has been aborted already, no need to check further
            # values
            if self.abort_object:
                return
            for rejection in self.mapping.object_rejection[var]:
                oper = rejection['oper']
                criteria = rejection['criteria']
                if oper=='eq':
                    self.abort_object = abs(val - criteria) < 0.001
                elif oper=='ue':
                    self.abort_object = abs(val - criteria) > 0.001
                elif oper=='gt':
                    self.abort_object = val > criteria
                elif oper=='lt':
                    self.abort_object = val < criteria
                elif oper=='in':
                    self.abort_object = val in criteria
                if self.abort_object:
                    self._add_error('REJECTING: %s %s; %s (%s) %s %s' \
                                     % (self.curr_obj['level'],
                                        self.curr_obj['id'], var, val,
                                        oper, criteria))
                    self.rejected.add(self.curr_obj['id'])
                    return True
        return False

    def _construct_id(self, item, level, mainrow):
        """
        Construct id string for an input object.

        item -- object's input variables
        level -- level index
        mainrow -- whether the row being imported is the main data row
        """
        level_list = self.mapping.level_list
        try:
            idelems = map(lambda x: item[x], level_list[level].id_pos)
        except:
            idelems = [str(self.id_counter)]
            self.id_counter += 1
        for ind in range(len(idelems)):
            if idelems[ind] in self.mapping.none_val:
                idelems[ind] = str(self.id_counter)
                self.id_counter += 1
        if level_list[level].id_delimiter is not None:
            idstr = level_list[level].id_delimiter.join(idelems)
        else:
            idstr = '-'.join(idelems)
        if len(idstr) == 0:
            msg = "could not construct valid id from line '%s'" \
                  % ' '.join(item)
            self._add_error(msg)
            return None
        lname = level_list[level].level_name
        if lname == self.main_level and mainrow:
            if idstr in self.import_ids:
                msg = "duplicate simulation level id '%s', skipping import" \
                      % idstr
                self._add_error(msg)
                self.rejected.add(idstr)
                return None
            self.import_ids.add(idstr)
        return idstr

    def _construct_unique_id(self, lind, oid, pid, leaf_level):
        """
        Construct unique id for current object

        lind -- level index
        oid -- object id
        pid -- parent object id
        leaf_level -- boolean indicating whether level is leaf level
        """
        create_unique_id = False
        if (pid is not None) and pid != SIMULATION_OBJ_ID:
            # Root level objects (main simulation units) must always be
            # unique already in the input data. If input format is by_level,
            # all levels except leaf level must be unique already in the data.
            # In the case of inlined data, only root level objects are
            # required to be unique and thus unique ids should be generated
            # for all levels below the root level.
            if self.format == 'by_level' and leaf_level:
                create_unique_id = True
            elif self.format == 'inlined':
                create_unique_id = True

        if not create_unique_id:
            if not lind in self.oids:
                self.oids[lind] = set([])
            elif self.format == 'by_level' and oid in self.oids[lind]:
                # in by_level data, all ids should be unique already,
                # especially if there's more than two data levels!!!
                levname = self.lexicon.level_name[lind]
                errmsg = "duplicate object id for %s '%s'" % (levname, oid,)
                self._add_error(errmsg)
                return None
            self.oids[lind].add(oid)
            self.id2lev[oid] = lind
            return oid
        else:
            if pid not in self.id2lev:
                errmsg = "could not find parent object '%s' for %s '%s'. "\
                         "Make sure that the parent object IDs are correct"
                levname = self.lexicon.level_name[lind]
                self._add_error(errmsg % (pid, levname, oid))
                return None
            plevel = self.id2lev[pid]
            distance = self._get_mid_level_length(lind, plevel)
            if distance is None:
                errmsg = "no parent path available from '%s' to '%s'!"
                st = self.lexicon.level_name[lind]
                ed = self.lexicon.level_name[plevel]
                self._add_error(errmsg % (st, ed))
            if distance > 0:
                oid = '-'.join(['X'] * distance + [oid])
            uid = '-'.join([pid, oid])
            self.id2lev[uid] = lind
            if not lind in self.oids:
                self.oids[lind] = set([])
            self.oids[lind].add(oid)
            return uid

    def _get_mid_level_length(self, levind1, levind2, dist=0):
        parent = self.lexicon.hierarchy[levind1]['parent']
        if parent == levind2:
            return dist
        elif parent is None:
            return None
        return self._get_mid_level_length(parent, levind2, dist+1)

    def _convert_data(self, levelname, rowtype, item):
        converted = []
        mapping = self.mapping.mapping
        if rowtype not in mapping:
            return converted
        cont = True
        for rowpos in mapping[rowtype].keys():
            try:
                impval = item[rowpos]
            except IndexError:
                msg = "row position mapping for position %d doesn't match the "\
                      " data for '%s' (data: %s)" \
                      % (rowpos, self.curr_obj['id'], item)
                self._add_error(msg)
                continue
            mappers = mapping[rowtype][rowpos]
            for mapper in mappers:
                if mapper.var_type=='numerical':
                    vals, cont = self._convert_numerical(\
                                mapper, levelname, rowtype, rowpos, impval)
                elif mapper.var_type=='categorical':
                    vals, cont = self._convert_categorical(\
                                mapper, levelname, rowtype, rowpos, impval)
                elif mapper.var_type=='date':
                    vals, cont = self._convert_date_variable(\
                                mapper, levelname, impval)
                elif mapper.var_type=='text':
                    vals, cont = self._convert_text(mapper, levelname, impval)
                elif mapper.var_type=='geometry':
                    vals, cont = self._convert_geom(mapper, levelname, impval)
                for val in vals:
                    converted.append(val)
                if not cont:
                    break  # break both loops due to object rejection
            if not cont:
                break
        return converted

    def _convert_categorical(self, mapper, levelname, rowtype, rowpos, impval):
        vals = []
        fromtype = mapper.from_data_type
        try:
            if  fromtype == 'int':
                keyval = int(impval)
            elif fromtype == 'double':
                keyval = float(impval)
            else:
                keyval = impval
        except ValueError:
            msg = "error converting imported value '%s' to data type "\
                  "'%s' for object '%s' at level '%s'" % \
                  (str(impval), fromtype, self.curr_obj['id'], levelname)
            self._add_error(msg)
            keyval = None
        catval = mapper.map.get(keyval)
        if catval is not None and catval != '':
            for m in catval:
                if levelname == self.mapping.level_list[0].level_name:
                    if self._check_for_rejection(m.to_var, m.to_val):
                        return vals, False
                vals.append((m.to_var, m.to_val))
        else:
            msg = "could not convert categorical value '%s' at row position "\
                  "%d for object '%s' at level '%s'" % \
                  (keyval, rowpos, self.curr_obj['id'], levelname)
            self._add_error(msg)
        return vals, True

    def _convert_date_variable(self, mapper, levelname, impval):
        vals = []
        ddate = self._parse_date(impval)
        for m in mapper.map:
            if m.epoch_year == 'current':
                beginning = datetime.date(ddate.year, 1, 1).toordinal()
                val = ddate.toordinal() - beginning
            elif m.epoch_year == 'gregorian':
                val =  ddate.toordinal()
            if levelname == self.mapping.level_list[0].level_name:
                if self._check_for_rejection(m.to_var, val):
                    return vals, False
            vals.append((m.to_var, val))
        return vals, True

    def _convert_numerical(self, mapper, levelname, rowtype, rowpos, impval):
        errormsg = "data at row type %s and row position %s can't be "\
            "converted to %s for object '%s' at level '%s'; value %s is not "\
            "a number, check the number_format_locale setting in ini file"
        vals = []
        for m in mapper.map:
            if impval is None or str(impval) in self.mapping.none_val:
                val = m.to_val  # to_val equals default value for numericals
            else:
                try:
                    val = m.conv_fact * float(impval)
                except:
                    msg = errormsg % (rowtype, rowpos, m.to_var,
                                      self.curr_obj['id'], m.level, impval)
                    self._add_error(msg)
                    val = None
            if val is not None:
                if levelname == self.mapping.level_list[0].level_name:
                    if self._check_for_rejection(m.to_var, val):
                        return vals, False
                vals.append((m.to_var, val))
        return vals, True

    def _convert_text(self, mapper, levelname, impval):
        vals = []
        for m in mapper.map:
            val = impval
            vals.append((m.to_var, val))
        return vals, True

    def _convert_geom(self, mapper, levelname, impval):
        vals = []
        for m in mapper.map:
            val = impval
            if m.to_var != 'geom':
                msg = 'Data of type geometry must be mapped to variable '\
                        'geom'
                self._add_error(msg)
            # TODO: This only works for PG at the moment. It supports
            # direct inserting of WKT and WKB to the geometry column
            vals.append((m.to_var, val))
        return vals, True

    def _set_default_values(self):
        """
        Set the default values to objects where applicable
        """
        defaults = self.mapping.defaults
        if self.curr_obj['level'] not in defaults.keys():
            return
        level_defaults = defaults[self.curr_obj['level']]
        values = self.curr_obj['data']['values']
        keys = map(itemgetter(0), values)
        for key, val in level_defaults.items():
            if key not in keys:
                values.append((key, val['value']))

    def _insert_object(self):
        """
        Store the object in the queue into the input database, empty
        the object queue.
        """
        # in case all objects were rejected, the object queue can be empty
        if self.curr_obj == {'level': None, 'id': None, 'pid': None}:
            return
        sim_data = self.input_db.row_count(self.sim_level)
        if not sim_data:
            self._add_simulation_object(self.curr_obj['ddate'])
        self._set_default_values()
        add2db = self.input_db.add_data_from_dictionary
        obj = (self.curr_obj['ddate'], self.curr_obj['data'])
        leveldict = {self.curr_obj['level']: [obj]}
        if not self.abort_object:
            add2db(leveldict, 0, 0)
        else:
            self.reject_count += 1
            self.abort_object = False
        self.curr_obj = {'level': None, 'id': None, 'pid': None}

    def _get_data_date(self, mainrow, levelnum, oid, parentid, item):
        date_pos = self.mapping.level_list[levelnum].date_pos
        levelname = self.mapping.level_list[levelnum].level_name
        if mainrow and self.import_date and date_pos != None:
            if item[date_pos] in self.mapping.none_val:
                ddate = self.global_data_date
            else:
                try:
                    datestr = item[date_pos]
                except IndexError:
                    msg = "invalid 'date_rowpos' indice for level '%s'" % levelname
                    self._add_error(msg)
                    return
                ddate = self._parse_date(datestr)
            if ddate is None:
                msg = "could not convert value '%s' to data date for object "\
                      "'%s' at level '%s'" % (item[date_pos], oid, levelname)
                self._add_error(msg)
                # set the global data date as the unit's date as DataDB
                # requires that each main level object has a date
                ddate = self.global_data_date
        else:
            ddate = self.initial_date[levelnum].get(oid, None)
            if ddate is None:
                ddate = self.initial_date[levelnum - 1].get(parentid, None)
            if ddate is None:
                ddate = self.global_data_date
        self.initial_date[levelnum][oid] = ddate
        return ddate

    def _get_parent_id(self, levelnum, item):
        parent_id = None
        if self.format == 'inlined':
            parentlevel = levelnum - 1
            if parentlevel == -1:
                parent_id = SIMULATION_OBJ_ID
            else:
                try:
                    parent_id = self.parent_id[parentlevel]
                except:
                    parent_id = None
        elif self.format == 'by_level':
            if levelnum == 0:
                parent_id = SIMULATION_OBJ_ID
            else:
                leveldef = self.mapping.level_list[levelnum]
                linkid_pos = leveldef.linkid_pos[1]
                try:
                    parent_id = item[linkid_pos]
                except:
                    oid = self.curr_obj['id']
                    lname = leveldef.level_name
                    msg= "The parent level id couldn't be found for object "\
                        "'%s' at level '%s', either the link_id_rowpos hasn't"\
                        " been defined in the conversion definition or this "\
                        "is a data error" % (oid, lname)
                    self._add_error(msg)
        return parent_id

    def _get_rowtype(self, item):
        """
        For the given format gets the row type, level index and level name.
        """
        mainrow_val = min(self.mapping.level_list[0].rowtype_value)
        if self.format == 'inlined':
            for levelind in range(len(self.mapping.level_list)):
                level = self.mapping.level_list[levelind]
                typepos = level.rowtype_pos
                try:
                    rowtype = int(item[typepos])
                except:
                    rowtype = None
                if rowtype in level.rowtype_value:
                    if rowtype == mainrow_val:
                        mainrow = True
                        sim_level = True
                    else:
                        mainrow = False
                        sim_level = False
                    return sim_level, mainrow, rowtype, levelind,\
                           level.level_name
        elif self.format == 'by_level':
            levelind = self.curr_level_ind
            level = self.mapping.level_list[levelind]
            rowtype = level.rowtype_value[0]
            if rowtype == mainrow_val:
                sim_level = True
            else:
                sim_level = False
            return sim_level, True, rowtype, levelind, level.level_name
        return None, None, None, None, None

    def _import_buffer(self, buffer, separator, skipfirst):
        """
        Import data from a data file, map the data values according to the
        mapping definition and store the data into input database.

        buffer -- StringIO buffer containing the data to import
        separator -- separator as string
        skipfirst -- skipfirst as boolean
        """
        firstline = self._skip_first(skipfirst)
        for line in buffer:
            if self.terminate_import:
                break
            if self._is_comment(line):
                continue
            item = self._split_row(line, separator)
            if not firstline and item is not None:
                # what kind of row are we dealing here
                sim_level, mainrow, rowtype, lind, lname = \
                        self._get_rowtype(item)
                if rowtype is None:
                    continue
                self._add_object(sim_level, mainrow, lind, lname, rowtype,
                                 item)
            firstline = False
        # flush the last object into the database
        self._insert_object()
        buffer.close()

    def _is_comment(self, line):
        """
        Method: is_comment
        """
        cp = self.mapping.comment_prefix
        if cp is not None:
            for pf in cp:
                if line[:len(pf)] == pf:
                    return True
        return False

    def _split_row(self, line, sep):
        """
        Checks that the current line is valid and then splits the line with
        the given separator

        line -- input data line, string
        sep -- column separator, string or None
        """
        if line.strip() == '':
            return None
        if sep is not None:
            if sep not in line:
                return None
        item = map(self.decode, line.split(sep))
        return item

    def _parse_date(self, datestr):
        """
        Parse a date string into a datetime object.

        datestr -- date as string
        """
        if datestr == '':
            return None
        if type(datestr)==datetime.date:
            return datestr
        valid = False
        # List of allowed date formats
        if self.month_first:
            formats = ['%m%d%y', '%m.%d.%y', '%m-%d-%y', '%m/%d/%y',\
                       '%m%d%Y', '%m.%d.%Y', '%m-%d-%Y', '%m/%d/%Y',\
                       '%Y-%m-%d']
        else:
            formats = ['%d%m%y', '%d.%m.%y', '%d-%m-%y', '%d/%m/%y',\
                       '%d%m%Y', '%d.%m.%Y', '%d-%m-%Y', '%d/%m/%Y',\
                       '%Y-%m-%d']
        for fmt in formats:
            try:
                dt = datetime.datetime.strptime(datestr, fmt)
            except:
                # date format could not be parsed
                continue
            valid = True
            break
        if not valid:
            self._add_error("invalid date format '%s'" % datestr)
            return None
        date = datetime.date(dt.year, dt.month, dt.day)
        return date

    def _skip_first(self, skipfirst):
        if skipfirst:
            return True
        else:
            return False

    def _terminate_data_import(self, msg):
        """
        Terminate the data import and program execution

        msg -- The reason for simulation termination
        """
        newmsg = '%s - Program set to terminate at %s' % (msg,time.asctime())
        self.logger.log_message(self.log_name, 'error', newmsg)
        self.terminate_import = True

def main():
    import doctest
    import sys
    import os
    os.chdir('..')
    sys.path.append('..')
    doctest.testfile('./input/test/importdata-py.rst')

if __name__ == '__main__':
    main()

