#!/usr/bin/env python
# encoding: utf-8
"""
Title: SIMO data importing

Module for importing external data from text files to the internal database.

*importdata.py*

Copyright (c) 2007 Simosol Oy. Distributed under the GNU General Public License 2.0.
"""

import time
from simo.utils import utils

CALMER = 200

class OperationImporter:
    """
    Class: OperationImporter

    Class for importing operations into SIMO input database
    """
    def __init__(self, inputdb, fopsdb, oformat, convdef, op2mc, logger,
                 logname, max_errors):
        """
        Initialize SimoImporter for importing different data formats into SIMO
        simulation environment.

        inputdb -- input database, SimoDB instance
        fopsdb -- database for forced operations, SimoDB instance
        oformat -- operation format as string; one of: 'text', 'xml', 'db'
        convdef -- OperationConversion instance
        op2mc -- Operation2Modelchain instance
        logger -- SimoLogger instance
        logname -- import log name as string
        """
        self.input_db = inputdb
        self.fops_db = fopsdb
        oformat = oformat.lower()
        if oformat in ('text'):
            # if operation format is text, operation conversion definition is
            # needed, operation2modelchain mapping is also stored in the
            # operation conversion instance
            self.op_conv = convdef.operation_conv
            self.none_val = convdef.operation_conv.none_value_indicator
            self.op_mapping = convdef.op2mc.operation_mapping
        elif oformat in ('xml', 'db'):
            # operation conversion definition is not needed for xml and db
            # formats
            self.op_conv = None
            self.none_val = None
            self.op_mapping = op2mc.operation_mapping
        self.format = oformat
        self.comment_prefix = None
        self.counter = 0
        self.logger = logger
        self.log_name = logname
        self.max_errors = max_errors
        self.errors = set([])
        self.error_count = 0
        self.terminate_import = False

    def import_operations(self, opbuffer, separator):
        """
        Import forced operations into input database. Currently only available
        format is an xml format for defining forced operation.

        opbuffer -- operation data as an open buffer
        separator -- value separator as string
        """
        self.error_count = 0
        self.terminate_import = False
        msg = 'Importing forced operations...'
        self.logger.log_message(self.log_name, 'info', msg)
        if self.format == 'xml':
            self._import_ops_xml(opbuffer)
        elif self.format == 'text':
            self._import_ops_text(opbuffer, separator)
        elif self.format == 'db':
            self._import_ops_db(opbuffer)
        msg = 'Forced operations imported!'
        self.logger.log_message(self.log_name, 'info', msg)
        opbuffer.close()
        self.input_db.db.commit()
        self.fops_db.db.commit()
        try:
            self.fops_db.db.vacuum_analyze()
        except Exception:
            pass

    def _add_error(self, msg):
        """
        Adds the error message to the data import error list and writes the
        message to error log

        msg -- errormessage as string
        """
        self.error_count += 1
        error='(first instance of) '+msg
        if error not in self.errors:
            self.errors.add(error)
            try:
                self.logger.log_message(self.log_name,'error',error)
            except UnicodeEncodeError:
                self.logger.log_message(self.log_name,'error',\
                                        'Invalid alphabets in error string')
        # Check if maximum error counts are exceeded
        if self.max_errors >= 0 and self.error_count > self.max_errors:
            msg = "Maximum number of errors (%s) exceeded!" \
                % str(self.max_errors)
            self._terminate_data_import(msg)

    def _add_numerical_variable(self, value, mapping):
        for m in mapping.map:
            if value is None or str(value) in self.none_val:
                val = m.toval # toval equals default value for numericals
            else:
                try:
                    val = m.convfact * float(value)
                except:
                    errorstr = "Data at row position %s can't "\
                       "be converted to %s at level %s; value %s is not "\
                       "a number" % (rowpos, m.tovar, m.level, value)
                    self._add_error(errorstr)
                    val = None
            if val is not None:
                self._set_data_value(m.level, m.to_var, val)

    def _add_operation(self, oid, level, name, chains, timeval, op_id=None,
                       force_date=False, iteration=0, branch=0,
                       variables=None, parameters=None):
        """
        Add a single forced operation into current unit db queue.

        oid -- operation unit id as string
        level -- operation level name as string
        name -- operation name as string
        chains -- model chain names in which the operations are implemented
        timeval -- operation timing value, date or time step
        op_id -- operation id
        force_date -- if True, timingtype = date
        iteration -- iteration of the operation
        branch -- branch of the operation
        """
        if force_date:
            timingtype = 'date'
        else:
            timingtype = self.op_conv.timing.type
        timestep, date, stepunit = None, None, None
        if timingtype == 'step':
            timestep = timeval
            stepunit = self.op_conv.timing.step_unit
        else:
            date = timeval
        ext_col_vals = {}
        if variables:
            ext_col_vals.update(variables)
        if parameters:
            ext_col_vals.update(parameters)
        if not ext_col_vals:
            ext_col_vals = None
        self.fops_db.add_operation(oid, level, iteration, branch, name,
                                   timingtype, timestep, date, stepunit,
                                   chains, op_id, ext_col_vals=ext_col_vals)

    def _construct_id(self, item):
        """
        Construct id string for the operation object.

        item -- object's input data
        """
        idelems = map(lambda x: item[x], self.op_conv.id_rowpos)
        if self.op_conv.id_delim is not None:
            idstr = self.op_conv.id_delim.join(idelems)
        else:
            idstr = ''.join(idelems)
        if len(idstr) == 0:
            return None
        return idstr

    def _get_operation_timing(self, value):
        """
        Convert operation timing from text format into a date or timestep.
        """
        ret = None
        if self.op_conv.timing.type == 'date':
            ret = utils.str2date(value)
            if ret is None:
                ret = utils.str2date(value, sep='.', order='day first')
        elif self.op_conv.timing.type == 'step':
            if self.op_conv.timing.step_unit == 'year':
                if not value in self.op_conv.timing.step_conversion:
                    errorstr = "Unknown operation time step indicator '%s'"\
                               % value
                    self._add_error(errorstr)
                else:
                    ret = self.op_conv.timing.step_conversion[value]
        return ret

    def _get_simulation_data(self, level, oid):
        """
        The data will come from the db in this format:
        {'comp_unit':
            [(ddate,
              {'id':'stand1', 'parent id': 'sim1',
               'values': [('AREA', 2.1), ('SC', 1.2), ('DEV_CLASS', 2.3)]}
             ),
            ]
        }
        """
        self.data_updated = False
        self.sim_data = {}
        data = self.input_db.fill_dictionary(level, oid)
        if data is None:
            return False
        self.sim_data = data
        self.existing_vars = {}
        for dlevel, vlist in data.items():
            self.existing_vars[dlevel] = {}
            ldata = vlist[0][1]['values']
            for ind, vartuple in enumerate(ldata):
                self.existing_vars[dlevel][vartuple[0]] = ind
        return True

    def _import_ops_text(self, opbuffer, sep):
        """
        Import forced operations into operation database from a text format.

        Parameters:
            opbuffer - operation file/StringIO buffer
            sep - value separator as string
        """
        prfx = self.op_conv.comment_prefix
        noopind = self.op_conv.no_operation_indicator
        level = self.op_conv.operation_level
        oid = None
        counter = 0
        if self.op_conv.op_id_rowpos is not None:
            op_id_set = set([])
        for row in opbuffer:
            counter += 1
            if counter % CALMER==0:
                msg = 'Importing operation %d ...' % counter
                self.logger.log_message(self.log_name,'info', msg)
            if prfx is not None:
                if row[:len(prfx)] == prfx:
                    continue
            items = self._split_row(row, sep)
            if items is None:
                continue
            oid = self._construct_id(items)
            if oid is None:
                continue
            if self.op_conv.op_id_rowpos is not None:
                op_id = items[self.op_conv.op_id_rowpos]
                if (oid, op_id) in op_id_set:
                    msg = "Duplicate operation id '%s' for simulation "\
                          "unit '%s'" % (oid, op_id)
                    self._add_error(msg)
                    continue
                op_id_set.add((oid, op_id))
            else:
                op_id = None

            if self.op_conv.type_rowpos is not None:
                optype = items[self.op_conv.type_rowpos]
            else:
                optype = None
            try:
                timval = items[self.op_conv.timing_rowpos]
            except:
                msg = "Invalid timing value on row '%s', cannot import "\
                      "operation" % ', '.join(items)
                self._add_error(msg)
                continue
            # get the operation code
            try:
                opcode = items[self.op_conv.operation_rowpos]
            except:
                msg = "Invalid operation code on row '%s', cannot import "\
                      "operation" % ', '.join(items)
                self._add_error(msg)
                continue
            if noopind is not None:
                if opcode in noopind:
                    continue
            if not optype in self.op_conv.types:
                errorstr = "Unknown operation type '%s', cannot import "\
                         "operation!" % optype
                self._add_error(errorstr)
                continue
            if not (optype, opcode) in self.op_conv.operations:
                errorstr = "Unknown operation code '%s', cannot import "\
                         "operation!" % opcode
                self._add_error(errorstr)
                continue
            name = self.op_conv.operations[(optype, opcode)].to_name
            if name not in self.op_mapping:
                errorstr = "Operation '%s' not defined in the operation "\
                         "mapping definition!" % name
                self._add_error(errorstr)
                continue
            timing = self._get_operation_timing(timval)
            variables = self._process_data_variables(items, oid)
            parameters = self._process_operation_parameters(items, optype, opcode)
            self._add_operation(oid, level, name, self.op_mapping[name],
                                timing, op_id=op_id, variables=variables,
                                parameters=parameters)

    def _import_ops_xml(self, opfile):
        """
        Import forced operations into input database from native xml format.

        opfile -- operation filename as string
        """
        root = ElementTree.ElementTree(file=opfile).getroot()
        ns = utils.elem_namespace(root.tag)
        counter = 0
        for elem in root:
            counter += 1
            if counter % CALMER==0:
                msg = 'importing operation %d ...'
                self.logger.log_message(self.log_name,'info', msg, counter)
            id = elem.find(ns+'id').text
            level = elem.find(ns+'level').text
            for operation in elem.find(ns+'operations'):
                ids = None
                targetelem = operation.find(ns+'target')
                targetlevel = targetelem.find(ns+'level').text
                idelem = targetelem.find(ns+'ids')
                if idelem is not None:
                    ids = [e.text for e in idelem]
                else:
                    # if there is no ids-element, the simulation unit is
                    # assumed to be the operation target
                    ids = [id]
                name = operation.find(ns+'name').text
                group = operation.find(ns+'group').text
                date = utils.str2date(operation.find(ns+'date').text)
                if date is None:
                    msg = "Invalid date format '%s' for operation '%s', "\
                        "unit '%s'!" \
                        % (operation.find(ns+'date').text, name, id)
                    self._add_error(msg)
                    continue
                if name not in self.op_mapping:
                    msg = "Operation '%s' not defined in the operation "\
                        "mapping definition!" % name
                    self._add_error(msg)
                    continue
                for id in ids:
                    self._add_operation(id, targetlevel, name, group,
                                            self.op_mapping[name], date)

    def _import_ops_db(self, opdb):
        """Import forced operations from SIMO result db operation result table
        """
        sql = 'SELECT id, iteration, op_date, op_level, op_name FROM '\
              'op_link ORDER BY id, op_date'
        data = opdb.conn.execute(sql)
        for item in data:
            oid, iterr, date, level, name = item
            if name not in self.op_mapping:
                msg = "Operation '%s' not defined in the operation "\
                    "mapping definition!" % name
                self._add_error(msg)
                continue
            self._add_operation(oid, level, name, self.op_mapping[name], date,
                                force_date=True, iteration=iterr)

    def _split_row(self, line, sep):
        """
        Checks that the current line is valid and then splits the line with
        the given separator

        line -- input data line, string
        sep -- column separator, string or None
        """
        if line.strip() == '':
            return None
        if sep is not None:
            if sep not in line:
                return None
        try:
            item = map(lambda x: x.strip(), line.split(sep))
        except:
            return None
        return item

    def _process_data_variables(self, item, oid):
        """
        Import data variables from operation transfer file.
        """
        oc = self.op_conv
        if oc.variables is None:
            return None  # no variables defined in the conversion
        variable_values = []
        for rowpos, mapping in oc.variables.iteritems():
            try:
                impval = item[rowpos]
            except IndexError:
                msg = "Mapping for row position %d doesn't match the data "\
                      "content for object '%s'" % (rowpos, oid)
                self._add_error(msg)
                break
            if mapping.var_type=='numerical':
                values = self._convert_numerical(mapping, rowpos, impval)
            elif mapping.var_type=='categorical':
                values = self._convert_categorical(mapping, impval)
            elif mapping.var_type=='date':
                values = self._convert_date_variable(mapping, impval)
            variable_values += values
        if variable_values:
            values_by_var = dict(variable_values)
        else:
            values_by_var = None
        return values_by_var

    def _convert_categorical(self, mapper, impval):
        fromtype = mapper.from_data_type
        try:
            if  fromtype == 'int':
                keyval = int(impval)
            elif fromtype == 'double':
                keyval = float(impval)
            else:
                keyval = impval
        except ValueError:
            msg = 'Error converting imported value %s to data type %s'\
                  % (str(impval), fromtype)
            self._add_error(msg)
            keyval = None
        vals = []
        value_convs = mapper.map.get(keyval)
        if value_convs is not None:
            for conv in value_convs:
                vals.append((conv.to_var, conv.to_val))
        return vals

    def _convert_date_variable(self, mapper, impval):
        vals = []
        ddate = self._parse_date(impval)
        for m in mapper.map:
            if m.epoch_year == 'current':
                beginning = datetime.date(ddate.year, 1, 1).toordinal()
                val = ddate.toordinal() - beginning
            elif m.epoch_year == 'gregorian':
                val =  ddate.toordinal()
            vals.append((m.to_var, val))
        return vals

    def _convert_numerical(self, mapper, rowpos, impval):
        errormsg = "Data at row position %s can't be converted to %s at "\
            "level %s; value %s is not a number, check the "\
            "number_format_locale setting in ini file"
        vals = []
        for m in mapper.map:
            if impval is None or str(impval) in self.none_val:
                val = m.to_val # to_val equals default value for numericals
            else:
                try:
                    val = m.conv_fact * float(impval)
                except:
                    msg =  errormsg % (rowpos, m.to_var, m.level, impval)
                    self._add_error(msg)
                    val = None
            vals.append((m.to_var, val))
        return vals

    def _process_operation_parameters(self, items, optype, opcode):
        """
        Import data parameters related to a given operation.

        items -- text file row as list of strings
        optype --
        opcode --
        """
        values = {}
        parameters = self.op_conv.operations[(optype, opcode)].parameters
        if parameters is None:
            return None
        for p in parameters:
            values[p.variable] = p.value
        return values

    def _terminate_data_import(self, msg):
        """
        Terminate the data import and program execution

        msg -- The reason for simulation termination
        """
        newmsg = '%s - Program set to terminate at %s' % (msg,time.asctime())
        self.logger.log_message(self.log_name, 'error', newmsg)
        self.terminate_import = True

def main():
    import doctest
    import sys
    import os
    os.chdir('..')
    sys.path.append('..')
    doctest.testfile('./input/test/importops-py.rst')

if __name__ == '__main__':
    main()

