#!/usr/bin/env python
# encoding: utf-8
"""
*brancher.py*

Simulator branch construction handler

Copyright (c) 2008 Simosol Oy. Distributed under the GNU General Public
License 2.0.
"""

import numpy

from collections import defaultdict, deque
from copy import deepcopy as dc
from datetime import date

from simo.utils.utils import update_dates
from constants import *


class Brancher(object):
    """Branching handler"""
    def __init__(self, iterations, bgroups, no_op_branch,
                 separate_branching_groups, branch_limit):
        """Initialize a branching handler
        """
        super(Brancher, self).__init__()
        self._no_op_branch = no_op_branch
        self._separate_branching_groups = separate_branching_groups
        self._iterations = iterations
        self._branching_groups = bgroups
        self._task_index = defaultdict(dict)
        self._group_index = dict()
        self._branch_limit = branch_limit
        self._active_tasks = None  # active branching task mappping
        self._active_bgs = None # active branching group mapping
        self._bg_bind = None  # fixed branching group of each (it, br, obj)
        self._parent_branches = None  # parent branch target index
        self._new_branches = None  # new branch target index
        self._blocked = None  # objects blocked by the brancher
        self._task = None  # current branching task
        self._gokeys = None # (operation group, operation name) keys
        self._is_branching = False  # is branching on at the moment?
        self._aborted = None  # branching aborted for targets?
        self._branching_info = None  # storage for branching history
        self._stored_branches = set([])  # branches already stored into binfo
        self._set_branch_groups()

    def _resize(self, inc, ax):
        """Resize active branching operation matrix and initialized branching
        matrix
        """
        at_shp = list(self._active_tasks.shape)
        at_shp[ax] = inc
        at = numpy.ones(at_shp, dtype=DTYPE_BOOL)
        self._active_tasks = numpy.concatenate((self._active_tasks, at), axis=ax)
        abg_shp = list(self._active_bgs.shape)
        abg_shp[ax] = inc
        abg = numpy.ones(abg_shp, dtype=DTYPE_BOOL)
        self._active_bgs = numpy.concatenate((self._active_bgs, abg), axis=ax)
        bgb_shp = list(self._bg_bind.shape)
        bgb_shp[ax] = inc
        bg_bind = numpy.zeros(bgb_shp, dtype=DTYPE_INT) - 1
        self._bg_bind = numpy.concatenate((self._bg_bind, bg_bind), axis=ax)

    def add_object(self, inc):
        """Add a number of main level objects to the brancher
        """
        self._resize(inc, 2)

    def add_branch(self, iteration, frombranch, tobranch, obj):
        """Add a new branch
        """
        if tobranch > (self._active_tasks.shape[1] - 1):
            inc = tobranch - frombranch
            if inc > 0:
                self._resize(inc, 1)
        self._active_tasks[iteration,tobranch,obj,:] = \
            dc(self._active_tasks[iteration,frombranch,obj,:])

    def remove_branch(self, iteration, branch, obj, bg_ind):
        """Remove a branch
        """
        self._active_tasks[iteration, branch, obj, :] = True
        self._active_bgs[iteration, branch, obj, bg_ind] = True

    def start(self, task, branch_name, data, lexicon, sim_level, depth,
              pred_mem, oper_mem):
        """Start branching for objects defined in the target index (tind)
        """
        # get target index of the currently active objects
        tind, torem = data.get_tind(sim_level, depth)
        
        # get branching group and branching task indices
        bg_name = task.branching_group
        bg_ind, bg_tasks, free_branching = self._group_index[bg_name]
        task_ind = self._task_index[bg_name][(task.name, branch_name)]
        
        # first check if any of the units (it,br,obj) are already associated to
        # a specific branching group
        bg_bind = self._bg_bind[tind[:,0], tind[:,1], tind[:,2]]
        bb = (bg_bind == -1) | (bg_bind == bg_ind)
        
        # then check if the current branching group still has any active
        # branching tasks and count the number of active tasks for each object
        at = self._active_tasks[tind[:,0], tind[:,1], tind[:,2], :]
        at_count = at[:, bg_tasks].sum(axis=1)
        
        # check which units have only single branching task active and remove
        # those units from the target index (no branching for those units)
        targets = self._active_tasks[tind[:,0], tind[:,1], tind[:,2], task_ind]
        targets[at_count == 0] = False
        targets *= bb
        
        # make sure that locked tasks are not done
        if numpy.all(targets == False):
            # if there are no branching targets, return without initializing
            # branching
            return
        
        if numpy.any(targets == False):
            # block those targets that should not be evaluated as the current
            # branching task has been already done and is not active
            self._blocked = tind[targets==False,:]
            data.block(sim_level, depth, self._blocked)
            
        # select target objects from the target index, for which the
        # branching task is still active, and start branching for the
        # selected objects
        self._parent_branches = tind[targets,:]
        
        # only targets with count more than 1 should produce new branches if
        # free branching is not allowed
        threshold = 0 if free_branching else 1
        create_new = at_count[targets] > threshold
        self._new_branches = tind[targets,:]
        
        # new branches should be generated from zero branch if there are
        # still active branching groups and if separate_branching_groups
        # is False
        if numpy.any(create_new==False) and numpy.any(tind[targets][:,1]==0) \
            and not self._separate_branching_groups:
            abg = self._active_bgs[tind[:,0], tind[:,1], tind[:,2], :]
            threshold = 0 if self._no_op_branch else 1
            active = abg[targets].sum(axis=1) > threshold
            zero_branch = tind[targets][:,1] == 0
            create_new = create_new | (active * zero_branch)
            
        # construct the new branches
        if self._parent_branches[create_new,:].shape[0] > 0:
            nbr = data.add_branch(self._parent_branches[create_new,:])
            self._new_branches[create_new,:] = nbr
            
            # if max branch limit has been set, check that none of the new
            # branches exceeds the branch limit
            if self._branch_limit is not None:
                abort = nbr[:,1] >= self._branch_limit
                if numpy.any(abort):
                    # delete branches that were aborted due to maximum
                    # branch limit
                    data.del_branch(self._new_branches[abort,:], bg_ind)
                    if numpy.all(abort):
                        # if all branches were aborted due to maximum branch
                        # limit, abort the whole branching process
                        self._parent_branches = None
                        self._new_branches = None
                        if self._blocked is not None:
                            data.release(sim_level, depth, self._blocked)
                            self._blocked = None
                        return
                    ok = abort == False
                    self._new_branches = self._new_branches[ok,:]
                    self._parent_branches = self._parent_branches[ok,:]
                    create_new = create_new[ok]
                    
            # block parent branch evaluation
            data.block(sim_level, depth, self._parent_branches[create_new,:])
            
            # update prediction and operation memories
            # go through target objects and all their relatives and update the
            # model memories of both the targets and the offspring
            n = self._parent_branches[create_new,:].shape[0]
            par_tind = self._parent_branches[create_new,:]
            new_tind = self._new_branches[create_new,:]
            for i in range(n):
                par_it, par_br, par_ind = par_tind[i,0:3]
                new_it, new_br, new_ind = new_tind[i,0:3]
                par_links = data.linkage.links[(par_it,par_br)][(sim_level,par_ind)]
                new_links = data.linkage.links[(new_it,new_br)][(sim_level,new_ind)]
                pred_mem.add_branch(par_tind[i,0:3], new_tind[i,0:3], sim_level)
                oper_mem.add_branch(par_tind[i,0:3], new_tind[i,0:3], sim_level)
                for level in par_links.keys():
                    # check from lexicon that level is a child level of
                    # simulation level (or the current evaluation level)
                    if not lexicon.is_child(level, sim_level):
                        continue
                    par_rels = par_links[level]
                    new_rels = new_links[level]
                    n_rel = len(par_rels)
                    for j in range(n_rel):
                        par_obj = (par_it, par_br, par_rels[j])
                        new_obj = (new_it, new_br, new_rels[j])
                        pred_mem.add_branch(par_obj, new_obj, level)
                        oper_mem.add_branch(par_obj, new_obj, level)
        self._task = task
        self._gokeys = [None for i in range(self._new_branches.shape[0])]
        self._is_branching = True
        self._aborted = numpy.ones(self._new_branches.shape[0], dtype=DTYPE_BOOL)

    def check_operation(self, model, data, level, depth):
        """Check if the current operation is a branching operation and if a
        new branch should be created or not

        Objects can be dropped from branching, e.g. branching will be aborted
        for the objects, if branching operations are aborted

        model -- model object
        data -- data handler
        level -- level indice
        depth -- model chain depth indice
        """
        if not self._is_branching or \
            model.index not in self._task.branching_ops:
            return
        gokey = (model.group, model.name)
        tind, torem = data.get_tind(level, depth)
        
        if model.dropped is not None and len(model.dropped) > 0:
            if numpy.all(tind[:,0:3]==model.dropped[:,0:3]):
                return
            newtind = []
            oldtind = tind[:,0:3]
            dropped = model.dropped[:,0:3]
            
            for i in oldtind:
                if True not in ((i == dropped).sum(axis=1) == 3):
                    newtind.append(i)
            tind = numpy.array(newtind, dtype=DTYPE_INT)
        
        if len(tind) == 0:
            return
        
        for i in range(self._new_branches.shape[0]):
            # check if the current new iteration-branch-object combination
            # is in the current target index and if yes add the indice of the
            # current combination to the done operations deque
            comb = self._new_branches[i,0:3]
            if numpy.any((comb == tind[:,0:3]).sum(axis=1) == 3):
                self._aborted[i] = False
                self._gokeys[i] = gokey

    def stop(self, branchname, data, timespan, sim_level, depth, pred_mem,
             oper_mem, lexicon, time_step_level, time_step_ind, opres_queue,
             opres_before_branching):
        """Stop branching, store branching information, reset initialized
        branching targets and branching task

        branchname -- new branch name as string
        data -- data handler
        timespan -- timespan control object
        sim_level -- simulation level indice
        depth -- model chain depth indice
        pred_mem -- prediction model memory structure
        oper_mem -- operation model memory structure
        time_step_level -- time step variable level, int
        time_step_ind -- time step variable indice, int
        opres_queue -- queued operation results, list
        opres_before_branching -- queued operations before branching, list
        """
        if not self._is_branching:
            return []
        
        bg_name = self._task.branching_group
        bg_ind, bg_tasks, free_branching = self._group_index[bg_name]
        
        # make sure that only new branches (child != parent) are deleted due
        # to branch abortion
        new = numpy.ones(self._new_branches.shape[0], dtype=DTYPE_BOOL)
        for i in range(self._new_branches.shape[0]):
            if numpy.all(self._new_branches[i] == self._parent_branches[i]):
                new[i] = False
        aborted = self._new_branches[self._aborted * new,:]

        # get date and timestep information
        cdates = data.get_date(self._new_branches)
        tind, torem = data.get_tind(time_step_level, depth)
        timesteps, err = data.get_value(tind, time_step_ind)
        unit = timespan.unit
        ndates = update_dates(cdates, timesteps, unit, deduceaday=True)
        if timespan.ending.type == 'end_date':
            end_month = timespan.ending.target.month
            end_day = timespan.ending.target.day
            end_year = timespan.ending.target.year
            end_date = date(end_year, end_month, end_day)
            for ind in range(len(ndates)):
                if ndates[ind] > end_date:
                    ndates[ind] = end_date

        # delete aborted objects from prediction and operation memories
        # go through target objects and all their relatives and update the
        # model memories of both the targets and the offspring
        for i in range(aborted.shape[0]):
            it, br, o = aborted[i,:3]
            pred_mem.del_objects(it, br, [o], sim_level)
            oper_mem.del_objects(it, br, [o], sim_level)
            links = data.linkage.links[(it,br)][(sim_level,o)]
            for level in links.keys():
                # check from lexicon that level is a child level of
                # simulation level (or the current evaluation level)
                if not lexicon.is_child(level, sim_level):
                    continue
                rels = links[level]
                pred_mem.del_objects(it, br, rels, level)
                oper_mem.del_objects(it, br, rels, level)

        # inactivate branches for which branching operations were not done
        data.del_branch(aborted, bg_ind)

        # store branching information
        ab = self._aborted
        pb = self._parent_branches
        nb = self._new_branches
        N = pb.shape[0]
        
        # create a dictionary of the new branches so that keys are:
        # (iteration, new branch, object ind) and values are:
        # (parent branch, operation group, operation name, branch name,
        #  branching group name, operation date) and add that
        # dictionary to branching information (history)
        go = self._gokeys

        # update active branching task map and
        # the branching dates up to the end date of the current period
        task_ind = self._task_index[bg_name][(self._task.name, branchname)]
        if self._branching_info is None:
            self._branching_info = {}
        pairs, dkeys, fromto = [], set([]), {}
        for i in range(N):
            if not ab[i]:
                key = tuple(nb[i,0:3])
                self._active_tasks[pb[i,0],pb[i,1],pb[i,2],task_ind] = False
                
                # bind the object to the current branching group
                # NB: setting the separate branching groups to False prevents  
                # the binding of objects into a specific branching groups, 
                # i.e. single object can have branches from multiple 
                # branching groups
                if not self._separate_branching_groups:
                    self._bg_bind[key[0],key[1],key[2]] = bg_ind
                
                # if the new branch has already been stored into the branching
                # history, skip the new branch from branching info
                if key in self._stored_branches:
                    continue
                val = (pb[i,1], go[i][0], go[i][1], branchname, bg_name, ndates[i])
                dkeys.add(key)
                pairs.append((key, val))
                pkey = tuple(pb[i,:3])
                fromto[pkey] = key
        
        # copy the queued operation results from the parent branch to
        # the new branch
        abkeys = set([tuple(item[0:3]) for item in aborted])
        copies = []
        for i, opres in enumerate(opres_before_branching):
            orkey = tuple(opres[1][0,:3])
            if orkey in fromto and orkey != fromto[orkey]:
                key = fromto[orkey]
                orcopy = list(dc(opres))
                orcopy[1][0,:3] = key
                orcopy[2] = (key[0], key[1], orcopy[2][2], orcopy[2][3])
                copies.append(tuple(orcopy))
        opres_queue += copies
        
        # remove all operation results from aborted branches
        remove = []
        for i, opres in enumerate(opres_queue):
            orkey = tuple(opres[1][0,:3])
            if orkey in abkeys:
                remove.append(i)
        for i in reversed(remove):
            opres_queue.pop(i)
        

        # update active branching groups
        active_tasks = self._active_tasks[pb[:,0], pb[:,1], pb[:,2], :]
        active_count = active_tasks[:,bg_tasks].sum(axis=1)
        if numpy.any(active_count == 0):
            m = active_count == 0
            self._active_bgs[pb[m,0], pb[m,1], pb[m,2], bg_ind] = False
        new = dict(pairs)
        self._branching_info.update(new)
        self._stored_branches.update(dkeys)
        
        # release parent branches
        data.release(sim_level, depth, self._parent_branches)
        
        # release objects blocked by the brancher (if there are any)
        if self._blocked is not None:
            data.release(sim_level, depth, self._blocked)
            self._blocked = None
            
        # reset the active branching task map and active branching group map
        # of the newly created branch
        for i in bg_tasks:
            self._active_tasks[nb[:,0],nb[:,1],nb[:,2],i] = True
        
        # get the parent branches that were not aborted and return them to
        # the caller
        parents = self._parent_branches[self._aborted == False,:]
        new = self._new_branches[self._aborted == False,:]
        blocked = numpy.concatenate((parents, new))
        
        # reset brancher, except for the branching history
        self._parent_branches = None
        self._new_branches = None
        self._task = None
        self._cur_group = None
        self._cur_operation = None
        self._done = None
        self._is_branching = False
        
        return blocked

    def is_blocked(self, task, branchname):
        """Check if branching group / task combination is blocked, i.e. it has
        been done for all units already
        """
        group = task.branching_group
        ind = self._task_index[group][(task.name, branchname)]
        if numpy.all(self._active_tasks[:,:,:,ind] == False):
            return True
        return False

    def reset_binfo(self):
        """Reset branching info container
        """
        self._branching_info = None

    def _set_branch_groups(self):
        """Construct branching task and group indices from the branching
        group dictionaries
        """
        n = 0
        if self._branching_groups is None:
            raise ValueError('There are no branching groups!')
        if self._iterations < 0:
            raise ValueError('Iterations is negative!')
        bg_ind = 0
        for bg_name, bg_item in self._branching_groups.items():
            self._task_index[bg_name] = defaultdict(dict)
            tasks = deque([])
            bg_tasks, free_branching = bg_item
            
            # no_op_branch from ini overrides the free_branching attribute
            # from model chain
            if not self._no_op_branch:
                free_branching = False
            for bg_part in bg_tasks:
                for task, branches in bg_part[1].items():
                    for branch in branches:
                        self._task_index[bg_name][(task, branch)] = n
                        tasks.append(n)
                        n += 1
            self._group_index[bg_name] = (bg_ind, tasks, free_branching)
            bg_ind += 1
            
        # construct two mappings (Numpy arrays) for storing information about
        # active branching tasks (_active_tasks), active branching groups
        # (_active_bgs) and branching groups associated to object (_bg_bind)
        # _amap is boolean array and _bg_bind is integer array where value -1
        # means that object is not associated with any branching group and
        # positive integer is specific branching group indice
        # _active_tasks dimensions: (iterations, branches, objects, bg tasks)
        # _active_bgs dimensions: (iterations, branches, objects, bgroups)
        # _bg_bind dimensions: (iterations, branches, object)
        self._active_tasks = numpy.ones((self._iterations, 1, 0, n),
                                        dtype=DTYPE_BOOL)
        self._active_bgs = numpy.ones((self._iterations, 1, 0, bg_ind),
                                      dtype=DTYPE_BOOL)
        self._bg_bind = numpy.zeros((self._iterations, 1, 0),
                                    dtype=DTYPE_INT) - 1


    ##########################################################################
    ## PROPERTIES
    def is_branching():
        doc = "Check if branching is on at the moment"
        def fget(self):
            return self._is_branching
        return locals()
    is_branching = property(**is_branching())

    def branching_info():
        doc = "Branching information"
        def fget(self):
            return self._branching_info
        return locals()
    branching_info = property(**branching_info())


def main():
    pass

if __name__ == '__main__':
    main()

