#!/usr/bin/env python
# encoding: utf-8
"""
Title: SIMO simulation; simulator in-memory data matrix handler

*handler.py*

Copyright (c) 2008 Simosol Oy. Distributed under the GNU General Public
License 2.0.
"""
import numpy
import copy
from collections import defaultdict
from linkage import Linkage
from ind2id import Ind2Id
from matrix import Matrix
from brancher import Brancher

from handler_ext import c_get_tind, c_get_value, c_block_children
from constants import *

# constants
BRANCH_INC = 10  # how many branches should be added at a time?


class Handler(object):
    """Data handler class for accessing data during simulation"""

    def __init__(self, iterations, lexicon, attributes, mcdepth, mainlevel,
                 branchgroups, noopbranch, separate_branching_groups, sim,
                 objcount=50, branchcount=1, branch_limit=None):
        """
        Initialize a data handler object

        iterations -- number of iterations, int
        lexicon -- Lexicon instance
        attributes -- simulation attribute container
        mcdepth -- maximum model chain depth, int
        mainlevel -- simulation main level, int
        branchgroups -- branching group dictionary
        noopbranch -- leave no operation branch or not, boolean
        separate_branching_groups -- should "no_op_branch" affect all branching
                                     groups simultaneously, or each one
                                     separately
        sim -- Simulator instance
        objcount -- number of simulation level objects, int
        branchcount -- initial number of branches, int
        branch_limit -- maximum number of branches, int
        """
        self._lexicon = lexicon
        self.hierarchy = lexicon.hierarchy
        self._main_level = mainlevel
        self._levels = len(self.hierarchy.keys())
        self.matrix = Matrix(iterations, self._levels, attributes, mcdepth,
                             sim=sim, objcount=objcount,
                             branchcount=branchcount)
        self.linkage = Linkage()
        self.ind2id = Ind2Id()
        self._iterations = iterations
        self._branchgroups = branchgroups
        self._no_op_branch = noopbranch
        self._separate_branching_groups = separate_branching_groups
        self._branch_limit = branch_limit
        self.reset()
        self.error_count = 0

    def reset(self):
        """
        Reset data handler: empty matrix, create new linkage,
        ind2id and brancher
        """
        self.matrix.reset()
        self.linkage = Linkage()
        self.ind2id = Ind2Id()
        if self._branchgroups  in (None, {}):
            self.brancher = None
        else:
            self.brancher = Brancher(self._iterations, self._branchgroups,
                                     self._no_op_branch,
                                     self._separate_branching_groups,
                                     self._branch_limit)
        self.use_active_tind = False
        self.active_level = None
        self.active_tind = None
        self.active_data_level = None
        self.active_dataobjects = None
        self.active_value_level = None
        self.text_data = {}

    def add_objects(self, iteration, branch, level, num, parlev=None,
                    parind=None, obj_ids=None):
        """
        Add number of objects to given level in the data matrix

        iteration -- target iteration, int
        branch -- target branch, int
        level -- target level, int
        num -- number of objects, int
        parlev -- parent level, int
        parind -- parent indice, int
        obj_ids -- list of object ids, if known
        """
        tind = numpy.zeros((num, 5), dtype=DTYPE_INT)
        tind[:,0] = iteration
        tind[:,1] = branch
        freeind = self._squeeze_objects(iteration, branch, level, num)
        for i in range(num):
            ind = freeind[i]
            tind[i,2] = ind
            self.matrix.omap[iteration, branch, level, ind] = True
            self.matrix.emap[iteration, branch, level, :, ind] = True
            self.matrix.mx[iteration, branch, level, ind, :] = numpy.NaN
            # NB! If a main level object is added, the date is not reset
            self.linkage.add_object(self.hierarchy, iteration, branch,
                                    (level, ind), (parlev, parind))
            # add id for the added object
            obj_id = (None, None) if obj_ids is None else obj_ids[i]
            self.ind2id.add_id(iteration, branch, level, ind, obj_id, parlev,
                               parind)
        # if objects are added to the simulation level, add brancher objects
        # also
        if level == self._main_level and \
           self.brancher is not None:
            self.brancher.add_object(num)
        return tind

    def del_objects(self, iteration, branch, level, objs):
        """
        Delete objects from data by nullifying data and
        removing the relationship links

        iteration -- target iteration, int
        branch -- target branch, int
        level -- target level, int
        objs -- target object indices, list or numpy array
        """
        self.matrix.mx[iteration, branch, level, objs, :] = numpy.NaN
        self.matrix.omap[iteration, branch, level, objs] = False
        self.matrix.emap[iteration, branch, level, :, objs] = False
        for o in objs:
            self.linkage.remove_object(iteration, branch, (level, o))
            self.ind2id.remove_id(iteration, branch, level, o)

    def add_branch(self, tind):
        """
        Copy simulation level objects with their attribute values, object
        and evaluation maps, links and ids to new branch. Increase matrix
        branch dimension when necessary.

        tind -- target index, numpy array
        """
        num = tind.shape[0]  # number of simulation level objects
        if num == 0:
            return
        # construct target index for the simulation level objects which
        # are added to new branches
        newtind = numpy.zeros((num, 5), dtype=DTYPE_INT)
        newtind[:,(0,2)] = tind[:,(0,2)]
        level = self._main_level
        iteration, frombranch, objs = tind[:,0], tind[:,1], tind[:,2]
        # find target branches for each simulation level object
        tobranch = self._find_empty_branches(iteration, level, objs)
        # update the new target index with the branch locations
        newtind[:,1] = tobranch
        # copy matrix values and object and evaluation maps of simulation
        # level objects from parent branches to the new branches
        self._copy_matrices(iteration, frombranch, tobranch, level, objs, objs)
        # go through each of the parent objects, copy values from parent
        # branch to new branch and update linkage and id information on
        # all levels
        for i in range(num):
            it, fb, tb, ob = iteration[i], frombranch[i], tobranch[i], objs[i]
            # add new top level object linkage and ids and copy matrix values,
            # if we're in a new branch and the new branch doesn't have a top
            # level object yet
            if not (it,tb) in self.linkage.links:
                self.linkage.links[(it,tb)] = {}
            if (0,0) not in self.linkage.links[(it,tb)]:
                self.linkage.add_object(self.hierarchy, it, tb, (0, 0),
                                        (None, None))
                obj_id = ('simulation', 'simulation')
                self.ind2id.add_id(it, tb, 0, 0, obj_id, None, None)
                self._copy_matrices(it, fb, tb, 0, 0, 0)
            # add a new branch to the brancher
            self.brancher.add_branch(it, fb, tb, ob)
            # construct iteration-branch and level-object keys
            ibkey = (it, fb)
            tibkey = (it, tb)
            objkey = (level, ob)
            ind2id = self.ind2id.ids[ibkey]
            # go through each level where parent branch object has a link to
            levels = self.linkage.links[ibkey][objkey].keys()
            # make sure that main level is included in the levels, or else the
            # linkage will break
            if not level in levels:
                levels.append(level)
            if levels == [0]: levels = [1]
            prevind, prevfree = {0: [0,]}, {0: [0,]}
            for lev in range(1, max(levels)+1):
                if lev == level:
                    # if we're on simulation main level, there's only one
                    # object and we know it's original and new indices (they
                    # are the same)
                    uid, oid = ind2id[objkey]
                    obj_ids = [(uid, oid)]
                    inds, free, n = [ob,], [ob,], 1
                else:
                    # get the indices of the objects in the parent branch
                    inds = self.linkage.links[ibkey][objkey][lev]
                    obj_ids = [ind2id[(lev,k)] for k in inds]
                    n = len(inds)
                    # find empty space for objects in the current
                    # iteration-branch-level combination
                    freeind = self._squeeze_objects(it, tb, lev, n)
                    free = list(freeind[:n])
                    # copy data matrix and object and evaluation map values
                    # from parent branch to the new one on the current level
                    self._copy_matrices(it, fb, tb, lev, inds, free)
                # go through the current level objects and add the objects to
                # the linkage and id handler
                for j in range(n):
                    # get the current object's parent indice in the original
                    # branch and update
                    plev = self.hierarchy[lev]['parent']
                    p = self.linkage.links[ibkey][(lev,inds[j])][plev][0]
                    pind = prevfree[plev][prevind[plev].index(p)]
                    self.linkage.add_object(self.hierarchy, it, tb,
                                            (lev, free[j]), (plev, pind))
                    self.ind2id.add_id(it, tb, lev, free[j], obj_ids[j], plev,
                                       pind)
                # copy the original indices and free indices
                prevind[lev], prevfree[lev] = copy.copy(inds), copy.copy(free)
        return newtind

    def del_branch(self, tind, bg_ind):
        """
        Delete branches defined in the target index

        tind -- target index, numpy array
        bg_ind -- branching group indice, int
        """
        level = self._main_level
        self.matrix.mx[tind[:,0], tind[:,1], level, tind[:,2], :] = numpy.NaN
        self.matrix.omap[tind[:,0], tind[:,1], level, tind[:,2]] = False
        self.matrix.emap[tind[:,0], tind[:,1], level, :, tind[:,2]] = False
        self.matrix.errormx[tind[:,0], tind[:,1], tind[:,2]] = 0
        # remove iteration-branch-object from ind2id, linkage and brancher
        # data structures at every level down from simulation level
        for it, br, obj in tind[:,:3]:
            ibkey = (it, br)
            objkey = (level, obj)
            # go through each level where parent branch object has a link to
            levels = self.linkage.links[ibkey][objkey].keys()
            for lev in levels:
                if lev<level:
                    # leave the parent levels alone
                    continue
                # get the indices of the objects in the parent branch
                ind = self.linkage.links[ibkey][objkey][lev]
                n = len(ind)
                self.matrix.mx[it, br, lev, ind, :] = numpy.NaN
                self.matrix.omap[it, br, lev, ind] = False
                self.matrix.emap[it, br, lev, :, ind] = False
            removed = self.linkage.remove_object(it, br, (level, obj),
                                                 children=True)
            self.ind2id.remove_a_bunch(it, br, removed)
            self.brancher.remove_branch(it, br, obj, bg_ind)

    def add_iteration(self, fromit):
        """
        Add number of iterations by copying the values of a single
        iteration to all other iterations

        Method does not resize matrix as the maximum number of iterations is
        already fixed in class constructor

        fromit -- source iteration, int
        """
        dc = self._deepcopy
        links = self.linkage.links
        ids = self.ind2id.ids
        inds = self.ind2id.inds
        n = self.matrix.shape[0]
        # go through each iterations and copy matrices, linkages, ids etc.
        for i in range(1,n):
            # copy matrices
            self.matrix.mx[i,...] = self.matrix.mx[fromit,...]
            self.matrix.omap[i,...] = self.matrix.omap[fromit,...]
            self.matrix.emap[i,...] = self.matrix.emap[fromit,...]
            self.matrix.errormx[i,...] = self.matrix.errormx[fromit,...]
            # copy linkages
            for ib in links.keys():
                it, br = ib
                if it == fromit:
                    links[(i,br)] = dc(links[ib])
            # copy ids and inds
            for ib in ids.keys():
                it, br = ib
                if it == fromit:
                    ids[(i,br)] = dc(ids[ib])
                    inds[(i,br)] = dc(inds[ib])

    def add_error(self, tind, max_simunit_errors):
        """
        Increase the error counter for objects defined in target index

        tind -- simulation main level target index, numpy array
        max_simunit_errors -- maximum number of errors per simulation unit
        """
        self.matrix.errormx[tind[:,0], tind[:,1], tind[:,2]] += 1
        self.error_count += len(tind)
        errormx = self.matrix.errormx[tind[:,0], tind[:,1], tind[:,2]]
        return errormx == (max_simunit_errors + 1)

    def check_errors(self, max_simunit_errors):
        """
        Check if error counter has passed the maximum number of errors and
        return the objects that are past the error count

        max_simunit_errors -- maximum number of errors per simulation unit
        """
        exc = (self.matrix.errormx > max_simunit_errors)
        exc = exc & self.matrix.omap[:,:,self._main_level,:]
        if numpy.all(exc == False):
            return None
        ind = numpy.where(exc)
        tind = numpy.zeros((len(ind[0]), 5), dtype=DTYPE_INT)
        tind[:,0] = ind[0]
        tind[:,1] = ind[1]
        tind[:,2] = ind[2]
        return tind

    def set_value(self, level, tind, attrs, values,
                  modelname=None, indicatechange=False):
        """
        Store attribute value(s) into the data matrix
        """
        if isinstance(attrs, (int, float)):
            setter = self.matrix.mx[:,:,level,:,attrs]
            setter[tind[:,0],tind[:,1],tind[:,2]] = values
        else:
            for i in range(len(attrs)):
                setter = self.matrix.mx[:,:,level,:,attrs[i]]
                setter[tind[:,0],tind[:,1],tind[:,2]] = values[i]

    def get_value(self, tind, attrs, ignore_missing=False):
        """Get attribute value(s) from the data matrix. If any of the target
        objects are missing values, return also error messages, an error
        object target index (for error logging), and a set of object indices
        that were missing values.
        """
        if tind.size == 0:
            return None, None
        level = self.active_level
        datalevel = self.active_data_level
        origtind = tind
        if datalevel is not None:
            level = datalevel
            if self.active_value_level is not None:
                level = self.active_value_level
            tind = self.active_dataobjects
            if tind.size == 0:
                return None, None
        # call the C extension (implemented using Cython)
        return c_get_value(tind, origtind, self.matrix.mx, self._lexicon,
                           level, datalevel, attrs, ignore_missing)

    def get_exists(self, tind, attrs):
        """Get attribute value(s) from the data matrix. If the target object
        has a value, return True(s) and False(s) for those which are missing
        a value
        """
        level = self.active_level
        datalevel = self.active_data_level
        origtind = tind
        if datalevel is not None:
            level = datalevel
            tind = self.active_dataobjects
            if tind == []:
                return None
        # NB: This is a fix needed for NumPy > 1.7.0, where the indexing
        # behaviour has changed a bit.
        multiattr = False
        if isinstance(attrs, int):
            attrs = (attrs,)
            multiattr = True
        getter = self.matrix.mx[:,:,level,:,attrs]
        values = getter[:,tind[:,0],tind[:,1],tind[:,2]]
        # NB: Related also to Numpy > 1.7.0: If a single attribute is in
        # question, the 2d array needs to be changed into 1d array.
        if multiattr:
            values = values[0]
        # check if any of the values are missing
        result = (numpy.isnan(values) == False)
        return result

    def get_tind(self, level, depth, datalevel=None, throughlevel=None,
                 vals4all=True, valuelevel=None, order=None):
        """
        Get index of target iterations, branches and objects. If datalevel
        differs from level, also start and stop indices for the get_value
        result matrix are returned. The same applies if throughlevel is set
        together with the datalevel. Parameter valuelevel can be defined in
        situations when datalevel differs from level, but the actual values
        should be from a level other than datalevel.
        """
        if self.use_active_tind and level == self.active_level:
            # if previously constructed target index is on same level and
            # emap hasn't changed, use the existing tind as constructing
            # a new one is costly, but erase possible values in the two
            # last columns
            tind = self.active_tind
            tind[:,3:] = 0
        else:
            #  find the active data object target indices
            indices = self._get_indices(level, depth)
            tind = numpy.zeros((indices[0].size, 5), dtype=DTYPE_INT)
            if order is None:
                # tind; at 0: iteration index, at 1: branch index, at 2: object index
                tind[:,0] = indices[0]
                tind[:,1] = indices[1]
                tind[:,2] = indices[2]
            else:
                # use pre-defined order for the target objects
                if order.shape[0] != tind.shape[0]:
                    raise Exception("Invalid order indicator!")
                tind[:,0:3] = order[:,0:3]
            self.active_tind = tind

        # store active levels, target indices etc.
        self.use_active_tind = True
        self.active_level = level
        self.active_data_level = datalevel
        self.active_value_level = valuelevel
        self.active_dataobjects = None

        # if datalevel is defined (different from level), construct an active
        # data objects structure
        toremove = set()
        if datalevel is not None:
            # call the C extension (implemented using Cython)
            throughlevel = -1 if throughlevel is None else throughlevel
            self.active_dataobjects = c_get_tind(tind, self.matrix.emap,
                                                 self.linkage.links, toremove,
                                                 level, depth, datalevel,
                                                 throughlevel, vals4all)

        return tind, toremove


    def _get_indices(self, level, depth):
        """Get active target object indices

        NB: this is a _very_ costly method with large data matrices!
        """
        return numpy.where(self.matrix.emap[:,:,level,depth,:])

    def get_active(self):
        """Get active level, data level and value level, and active data
        objects
        """
        return (self.active_level,
                self.active_data_level,
                self.active_value_level,
                self.active_dataobjects)

    def set_active(self, act):
        """Set active level, data level and value level, and active data
        objects
        """
        self.active_level = act[0]
        self.active_data_level = act[1]
        self.active_value_level = act[2]
        self.active_dataobjects = act[3]

    def set_date(self, tind, dates, set_init_date=False):
        """Store date(s) into the date matrix
        """
        self.matrix.dmx[tind[:,2]] = dates
        if set_init_date:
            self.matrix.init_dmx[tind[:,2]] = dates

    def get_date(self, tind):
        """Get date(s) from the date matrix
        """
        return self.matrix.dmx[tind[:,2]]

    def get_init_date(self, tind):
        """Get initial date(s) from the date matrix
        """
        return self.matrix.init_dmx[tind[:,2]]

    def get_id(self, level, tind):
        """Get id of objects defined in target index
        """
        get = self.ind2id.get_id
        it, br, obj = tind[:,0], tind[:,1], tind[:,2]
        return [get(it[i],br[i],level,obj[i]) for i in range(tind.shape[0])]

    def get_parent(self, level, tind, parentlevel, filterby=None):
        """
        Get parent target index for objects in the target index, possibly
        only for the objects in the filter by set
        """
        links = self.linkage.links
        do_filter = False if filterby is None else True
        if do_filter:
            filtershape = (len(filterby), tind.shape[1])
        ptind = []
        torem = set([])
        for i in range(tind.shape[0]):
            if do_filter:
                if tind[i,2] not in filterby:
                    continue
            it, br, obj = tind[i,0:3]
            link = links[(it,br)][(level,obj)]
            if parentlevel not in link or link[parentlevel] == set([]):
                torem.add(i)
                continue
            parent = link[parentlevel][0]
            ptind.append((it, br, parent, 0, 0))
        return numpy.array(ptind, dtype=DTYPE_INT), torem

    def get_children(self, iteration, branch, obj, level, childlevel):
        """Get the children of an object from child level in given iteration
         and branch
        """
        ibkey = (iteration, branch)
        lokey = (level, obj)
        if not childlevel in self.linkage.links[ibkey][lokey]:
            return None
        return self._copy(self.linkage.links[ibkey][lokey][childlevel])

    def get_object_map(self, level, tind=None):
        """Get object mapping for given level
        """
        if tind is not None:
            it, br, obj = tind[:,0], tind[:,1], tind[:,2]
            omap = self.matrix.omap[it,br,level,obj]
        else:
            omap = self.matrix.omap[:,:,level,:]
        return omap

    def block(self, level, depth, tind, block_children=False):
        """Block evaluation of objects at given model chain depth level and
        all levels below the given depth level (prevents evaluation of blocked
        objects)
        """
        if tind.size == 0:
            return
        it, br, obj = tind[:,0], tind[:,1], tind[:,2]
        self.matrix.emap[it,br,level,depth:,obj] = False
        if block_children:
            # all child levels can be blocked also if necessary
            for it, br, obj in tind[:,:3]:
                ibkey = (it, br)
                objkey = (level, obj)
                # go through each level where parent branch object has a link to
                levels = self.linkage.links[ibkey][objkey].keys()
                for lev in levels:
                    if lev <= level:
                        continue
                    # get the indices of the objects in the parent branch
                    ind = self.linkage.links[ibkey][objkey][lev]
                    self.matrix.emap[it,br,lev,depth:,ind] = False
        # emap changed, don't use the active target index anymore
        self.use_active_tind = False
        self.active_tind = None

    def block_children(self, depth, level, clevels):
        """
        Find the blocked objects from 'level' and block all previously
        non-blocked relatives from all descendant levels
        """
        blocked_children = c_block_children(depth, level, clevels, 
                                            self.matrix.emap, self.matrix.omap,
                                            self.linkage.links)
        # emap changed, don't use the active target index anymore
        self.use_active_tind = False
        self.active_tind = None
        return blocked_children

    def release(self, level, depth, tind, release_children=False,
                release_all=False):
        """Release objects at given level and model chain depth for
        evaluation. If level is not defined, release all levels. If depth is
        not defined release objects from all depths and if target index is not
        defined, release all objects.
        """
        if tind is not None and tind.size == 0:
            return
        if release_all:
            maxdepth = self.matrix.emap.shape[3]
            for i in range(maxdepth):
                self.matrix.emap[:,:,:,i,:] = self.matrix.omap[:,:,:,:]
        elif level is None or depth is None:
            raise Exception, "Cannot release objects: level and depth must "\
                "be defined if release_all is False!"
        elif tind is None:
            self.matrix.emap[:,:,level,depth,:] = self.matrix.omap[:,:,level,:]
        else:
            if depth is not None:
                it, br, obj = tind[:,0], tind[:,1], tind[:,2]
                maxdepth = self.matrix.emap.shape[3]
                for i in range(depth, maxdepth):
                    self.matrix.emap[it,br,level,i,obj] = \
                        self.matrix.omap[it,br,level,obj]
            if release_children:
                # release also the children of target objects
                for it, br, obj in tind[:,:3]:
                    ibkey = (it, br)
                    objkey = (level, obj)
                    # go through each level where parent branch object has a link to
                    levels = self.linkage.links[ibkey][objkey].keys()
                    for lev in levels:
                        if lev <= level:
                            continue
                        # get the indices of the objects in the parent branch and
                        # release the objects
                        ind = self.linkage.links[ibkey][objkey][lev]
                        maxdepth = self.matrix.emap.shape[3]
                        for i in range(depth, maxdepth):
                            self.matrix.emap[it, br, lev, i, ind] = \
                                self.matrix.omap[it, br, lev, ind]
        # emap changed, don't use the active target index anymore
        self.use_active_tind = False
        self.active_tind = None

    def _squeeze_objects(self, iteration, branch, level, num):
        """Find empty space for number of objects in given iteration, branch
        and level. Increase matrix size to make space if needed.
        """
        where = numpy.where
        freeind = where(self.matrix.omap[iteration,branch,level,:]==False)[0]
        inc = num - len(freeind)
        if inc > 0:
            if not self.matrix.modify_matrix(3, inc):
                raise Exception('Error modifying matrix!')
            freeind = where(self.matrix.omap[iteration, branch,level,:] == \
                      False)[0]
        return freeind

    def _copy_matrices(self, iteration, frombranch, tobranch, level, fromobj,
                       toobj):
        """
        Copy values in all matrices from parent branch/object to target
        branch/object
        """
        self.matrix.mx[iteration,tobranch,level,toobj,:] = \
            self.matrix.mx[iteration,frombranch,level,fromobj,:]
        self.matrix.omap[iteration,tobranch,level,toobj] = \
            self.matrix.omap[iteration,frombranch,level,fromobj]
        self.matrix.emap[iteration,tobranch,level,:,toobj] = \
            self.matrix.emap[iteration,frombranch,level,:,fromobj]
        if level == self._main_level:
            self.matrix.errormx[iteration,tobranch,toobj] = \
                self.matrix.errormx[iteration,frombranch,fromobj]

    def _find_empty_branches(self, iteration, level, objs):
        """
        Find empty branches for objects on given level. Resize matrices
        in branch dimension if necessary.
        """
        # find the first empty locations for the new branches and check for
        # which simulation level objects there is still empty space
        omap = self.matrix.omap[iteration,:,level,objs]
        locs = numpy.argmin(omap, axis=1)
        empty = self.matrix.shape[1] - numpy.sum(omap, axis=1)
        # if there is at least a single object without any empty branches,
        # grow matrix branch dimension by one
        if numpy.any(empty == 0):
            # put the objects without empty branches into the first new branch
            locs[empty == 0] = self.matrix.shape[1]
            if not self.matrix.modify_matrix(1, BRANCH_INC):
                raise Exception('Error modifying matrix!')
        # when all object indices are not unique, find the locations by
        # increasing the location index
        N = len(objs)
        if len(set(objs)) != N:
            objs = list(objs)
            for i in range(N):
                if objs.count(objs[i]) > 1:
                    it = iteration[i]
                    loc = locs[i]
                    obj = objs[i]
                    if not self.matrix.omap[it,loc,level,obj]:
                        # original location is ok, occupy it
                        self.matrix.omap[it,loc,level,obj] = True
                    else:
                        # try to find an unoccupied location by testing
                        # locations with larger indices
                        loc = locs[i] + 1
                        while True:
                            if loc > (self.matrix.omap.shape[1] - 1):
                                # in case the current indice is out of bounds,
                                # increase matrix size in branch dimension
                                self.matrix.modify_matrix(1, BRANCH_INC)
                            if not self.matrix.omap[it,loc,level,obj]:
                                self.matrix.omap[it,loc,level,obj] = True
                                locs[i] = loc
                                break
                            loc += 1
        # return the new object locations (branches)
        return locs

    def _copy(self, data):
        return copy.copy(data)

    def _deepcopy(self, data):
        return copy.deepcopy(data)


def main():
    pass

if __name__ == '__main__':
    main()
