#!/usr/bin/env python
# encoding: utf-8
"""
*linkage.py*

the complete object linking for the data matrix

Copyright (c) 2008 Simosol Oy. Distributed under the GNU General Public
License 2.0.
"""

#from collections import defaultdict
from copy import copy

class Linkage(object):
    """links between all objects"""
    def __init__(self):
        super(Linkage, self).__init__()
        self.links = {}

    def add_object(self, hierarchy, iteration, branch, objkey, parentkey,
                   update_existing=True):
        """
        Copies the parents links to the new object and adds the link to the
        new object for all the existing objects linked to it through the
        parent objkey is a tuple of level and object indices, optional
        parentkey is a tuple of parent level and parent object indices
        """
        if not (iteration,branch) in self.links:
            self.links[(iteration,branch)] = {}
        ib = self.links[(iteration,branch)]
        if objkey not in ib:
            ib[objkey] = {}
        if parentkey != (None, None):
            if parentkey not in ib:
                ib[parentkey] = {}
            level, obj = objkey
            myord = hierarchy[level]['ordinal']
            mylin = hierarchy[level]['lineage']
            plevel, pobj = parentkey
            if not plevel in ib[objkey]:
                ib[objkey][plevel] = []
            ib[objkey][plevel].append(pobj)
            ib[objkey][plevel].sort()
            # copy the links from parent's links
            for tolevel, toobjs in ib[parentkey].items():
                if tolevel == plevel:
                    continue
                toord = hierarchy[tolevel]['ordinal']
                tolin = hierarchy[tolevel]['lineage']
                # if the link for the parent is higher than added object
                # in the hierarchy and they belong to the same lineage, add
                if toord < myord and mylin.intersection(tolin) != set([]):
                    ib[objkey][tolevel] = copy(toobjs)
            # update the existing links
            for tolevel, toobjs in ib[objkey].items():
                if not update_existing and tolevel != plevel:
                    continue
                for to in toobjs:
                    if not level in ib[(tolevel,to)]:
                        ib[(tolevel,to)][level] = []
                    ib[(tolevel,to)][level].append(obj)
                ib[(tolevel,to)][level].sort()

    def remove_object(self, iteration, branch, obj, children=False):
        """
        Remove the references to the object from all the connected nodes,
        and the object from the links structure. If children is set,
        remove also all the object's children.
        """
        ib = self.links[(iteration,branch)]
        level, o = obj
        tonodes = ib[obj]
        removed = {}
        levels = tonodes.keys()
        for tolevel, toobjs in tonodes.items():
            for to in toobjs:
                if children and tolevel>level:
                    for lev in levels:
                        if lev<level:
                            # remove children from levels higher in the hierarchy
                            # than the parent to be removed
                            top_obj = (lev, tonodes[lev][0])
                            ib[top_obj][tolevel].remove(to)
                            if len(ib[top_obj][tolevel]) == 0:
                                del ib[top_obj][tolevel]
                    del ib[(tolevel, to)]
                    if not tolevel in removed:
                        removed[tolevel] = []
                    removed[tolevel].append(to)
                else:
                    ib[(tolevel,to)][level].remove(o)
                    if len(ib[(tolevel,to)][level]) == 0:
                        del ib[(tolevel,to)][level]
        del ib[obj]
        if not level in removed:
            removed[level] = []
        removed[level].append(o)
        return removed

def main():
    pass

if __name__ == '__main__':
    main()

