#!/usr/bin/env python
# encoding: utf-8
"""
Title: SIMO simulation; simulator in-memory data matrix

*matrix.py*

Copyright (c) 2008 Simosol Oy. Distributed under the GNU General Public
License 2.0.
"""

import numpy
import datetime

from constants import *

class Matrix(object):
    """Data matrix class for storing data during simulation"""

    def __init__(self, iterations, levels, attributes, mcdepth, sim, 
                 objcount=50, branchcount=1):
        """Initialize a data matrix
        mx: simulation data core, which contains data attribute values as
        floats

        omap: (object map) indicates object existence for the particular
        index combination in the data matrix as booleans

        emap: (evaluation map) indicates the current model chain task
        evaluation status at the given model chain depth as booleans

        dmx: (date matrix) contains current date for each simulation unit as
        datetime.date objects
        
        errormx: (error matrix) is an error counter for simulation main level
        objects

        mx dimensions: (iterations, branches, levels, objects, attributes)
        omap dimensions: (iterations, branches, levels, objects)
        emap dimensions: (iterations, branches, levels, depth, objects)
        dmx dimensions: (objects)
        errormx dimensions: (iterations, branches, objects)
        """
        self.mx = None
        self.omap = None
        self.emap = None
        self.dmx = None
        self.init_dmx = None
        self.errormx = None
        self.shape = (None, None, None, None, None)
        self.sim = sim
        try:
            shp = (iterations, branchcount, levels, objcount, attributes)
            self.mx = numpy.empty(shp, dtype=DTYPE_FLOAT)
            self.mx.fill(numpy.NaN)
            shp = (iterations, branchcount, levels, objcount)
            self.omap = numpy.zeros(shp, dtype=DTYPE_BOOL)
            shp = (iterations, branchcount, levels, mcdepth, objcount)
            self.emap = numpy.zeros(shp, dtype=DTYPE_BOOL)
            self.dmx = numpy.zeros((objcount), dtype=DTYPE_DT)
            self.init_dmx = numpy.zeros((objcount), dtype=DTYPE_DT)
            self.dmx.fill(datetime.date(1,1,1))
            self.init_dmx.fill(datetime.date(1,1,1))
            shp = (iterations, branchcount, objcount)
            self.errormx = numpy.zeros(shp, dtype=DTYPE_INT)
            self.shape = self.mx.shape
        except Exception, e:
            sim.add_error('Matrix initialization failed, '+str(e), None, None)
            raise  # NB! This is a critical failure; out of memory or broken
                   # initialization parameters, so let it raise.
    
    def reset(self):
        """Reset all matrix values
        """
        self.mx[...] = numpy.NaN
        self.omap[...] = False
        self.emap[...] = False
        self.dmx[:] = numpy.zeros((self.omap.shape[-1]), dtype=DTYPE_DT)
        self.init_dmx[:] = numpy.zeros((self.omap.shape[-1]), dtype=DTYPE_DT)
        self.dmx.fill(datetime.date(1,1,1))
        self.init_dmx.fill(datetime.date(1,1,1))
        self.errormx[...] = 0

    def modify_matrix(self, axis, inc):
        """Modify data matrix by increasing array size along given axis
        """
        # sanity check for axis: only branches and objects can be added,
        # number of iterations, levels and attributes is fixed when matrix is
        # created
        if axis not in (1,3):
            self.sim.add_error('Bad axis (expected 1 or 3, got %d), '\
                               'the size of the matrix can not be modified.' %\
                               (axis,), None, None)
            return False  # NB! This shouldn't happen and in fact currently 
                          # can't, but it's handy around for future changes
        shp = list(self.shape)
        # construct an array of correct shape and size, fill data matrix with
        # NaNs, omap and emap with Falses, and date matrix with zeros
        shp[axis] = inc
        mx = numpy.empty(shp, dtype=DTYPE_FLOAT)
        mx.fill(numpy.NaN)
        oshp = shp[:-1]
        omap = numpy.zeros(oshp, dtype=DTYPE_BOOL)
        eshp = shp[:3] + [self.emap.shape[3]] + [shp[3]]
        emap = numpy.zeros(eshp, dtype=DTYPE_BOOL)
        eaxis = 1 if axis == 1 else 4  # emap object dimension index is 4
        erraxis = 1 if axis == 1 else 2  # errormx object dimension index is 2
        errshp = shp[0:2] + [self.shape[3]]
        errshp[erraxis] = inc
        emx = numpy.zeros(errshp, dtype=DTYPE_INT)
        self.mx = numpy.concatenate((self.mx, mx), axis=axis)
        self.omap = numpy.concatenate((self.omap, omap), axis=axis)
        self.emap = numpy.concatenate((self.emap, emap), axis=eaxis)
        self.errormx = numpy.concatenate((self.errormx, emx), axis=erraxis)
        if axis == 3:
            # increase date matrix dimensions only when adding new objects
            d1 = numpy.zeros(shp[3], dtype=DTYPE_DT)
            d2 = numpy.zeros(shp[3], dtype=DTYPE_DT)
            d1.fill(datetime.date(1,1,1))
            d2.fill(datetime.date(1,1,1))
            self.dmx = numpy.concatenate((self.dmx, d1), axis=0)
            self.init_dmx = numpy.concatenate((self.init_dmx, d2), axis=0)
        self.shape = self.mx.shape
        return True

    # Data class properties
    def shape():
        doc = 'list of current matrix dimensions for branches, levels, '\
              'objects and attributes'
        def fget(self):
            return self._shape
        def fset(self, shape):
            self._shape = shape
        return locals()
    shape = property(**shape())

    def iterations():
        doc = 'number of iterations in the matrix'
        def fget(self):
            return self._shape[0]
        return locals()
    iterations = property(**iterations())

    def branches():
        doc = 'number of data branches in the matrix'
        def fget(self):
            return self._shape[1]
        return locals()
    branches = property(**branches())

    def levels():
        doc = 'number of data levels in the matrix'
        def fget(self):
            return self._shape[2]
        return locals()
    levels = property(**levels())

    def objects():
        doc = 'number of data objects in the matrix'
        def fget(self):
            return self._shape[3]
        return locals()
    objects = property(**objects())

    def attributes():
        doc = 'number of data attributes in the matrix'
        def fget(self):
            return self._shape[4]
        return locals()
    attributes = property(**attributes())


def main():
    pass

if __name__ == '__main__':
    main()
