#!/usr/bin/env python
# encoding: utf-8
"""
Title: SIMO metaheuristic genetic algorithm (GA) optimization algorithm

...

*genetic_algorithm.py*

Copyright (c) 2007 Simosol Oy. Distributed under the GNU General Public License 2.0.
"""

# these are for running the doctest from src/simo
#import sys
#sys.path.append('..')

import time
import math
import numpy
from optimizer import Optimizer


ARGUMENTS = ('repeats',
             'maximum_iterations',
             'population_size',
             'breeding_rate',
             'mutation_probability',
             'swapping_probability')


class GeneticAlgorithm(Optimizer):
    """Genetic algorithm implementation
    """
    def _process_arguments(self, args):
        """Process keyword arguments
        """
        self._abort = False
        for arg in ARGUMENTS:
            if not arg in args:
                self._add_error('Parameter "%s" missing!' % arg)
                self._abort = True
        if self._abort:
            return
        try:
            self._repeats = int(args["repeats"])
            self._max_iterations = int(args["maximum_iterations"])
            self._population_size = int(args["population_size"])
            self._breeding_rate = float(args["breeding_rate"])
            self._mutation_probability = float(args["mutation_probability"])
            self._swapping_probability = float(args["swapping_probability"])
        except:
            self._add_error('Invalid parameter values!')
            self._abort = True
    
    def _sort_by_fitness(self, population):
        """Sort a population by the fitness of individuals, best first
        
        population -- population to be sorted
        """
        def _cmp(x, y):
            delta = x[0]-y[0]
            if delta < 0:
                return -1
            elif delta == 0:
                return 0
            else:
                return 1
                
        target = self._data.task_def.objective_func.target
        reverse = True if target == 'max' else False
        population.sort(_cmp, reverse=reverse)
    
    def _populate(self):
        """Generate a population from scratch
        """
        msg = "  generating initial population..."
        self._add_info(msg)
        
        start = time.time()
        feas = self._solution_feasibility
        rand = self._random_solution
        util = self._solution_utility
        population = []
        maxloops = self._population_size * 10000
        loops = 0
        feas_count = 0
        while len(population) < self._population_size:
            # generate random individual, make sure that it is feasible and
            # evaluate the fitness of the individual
            individual = rand()
            if feas(individual):
                fitness = util(individual)
                population.append((fitness,individual))
                feas_count += 1
                
            loops += 1
            if loops > maxloops:
                msg = "Cannot generate initial population due to too strict "\
                      "constraints! A total of %d iterations produced %d/%d "\
                      "feasible individuals." % (maxloops, feas_count,
                                                 self._population_size)
                self._add_error(msg)
                self._best_U = None
                return None
            
        # sort the population by the fitness
        self._sort_by_fitness(population)
        
        msg = "  initial population of %d individuals found in %0.1f secs" \
              % (feas_count, time.time() - start)
        self._add_info(msg)
        
        return population
    
    def _mutate(self, p, individual):
        """Mutate (randomly change treatment) a proportion p of the genes
        of given individual
        
        p -- mutation probability
        individual -- individual (solution) as numpy array
        """
        copy = numpy.copy
        move = self._lambda_opt_move
        feas = self._solution_feasibility
        
        N_u = self._num_of_units
        mylambda = int(math.floor(N_u * p))
        feasible = False
        loops = 0
        while not feasible:
            mutated = copy(individual)
            if not move(mutated, mylambda):
                continue
            feasible = feas(mutated)
            loops += 1
            if loops > N_u:
                return None
        return mutated
    
    def _np_crossover(self, n, parent1, parent2):
        """Do n point crossover for two individuals
        
        n -- number of split points
        parent1 -- parent individual
        parent2 -- parent individual
        """
        copy = numpy.copy
        rand = numpy.random.randint
        feas = self._solution_feasibility
        
        N_u = self._num_of_units
        last = [N_u - 1]
        feasible = False
        loops = 0
        while not feasible:
            child1 = copy(parent1)
            child2 = copy(parent2)
            points = list(rand(1, N_u, n)) + last
            start = 0
            for i, end in enumerate(points):
                # switch the gene sequences on even rounds
                if (i % 2) == 0:
                    child1[start:end] = parent2[start:end]
                    child2[start:end] = parent1[start:end]
                start = end
            # check that both of the offspring are feasible, continue
            # looping if not
            feasible = feas(child1) and feas(child2)
            loops += 1
            if loops > N_u:
                return None, None
        return child1, child2
    
    def _unif_crossover(self, p, parent1, parent2):
        """Do uniform crossover for two individuals
        
        p -- swapping probability
        parent1 -- parent individual
        parent2 -- parent individual
        """
        copy = numpy.copy
        rand = numpy.random.random
        feas = self._solution_feasibility
        N_u = self._num_of_units
        feasible = False
        loops = 0
        while not feasible:
            child1 = copy(parent1)
            child2 = copy(parent2)
            # generate a random vector, check which genes are swapped
            # and do the swapping
            X = rand(N_u)
            swap = X > p
            child1[swap] = parent2[swap]
            child2[swap] = parent1[swap]
            # check that both of the offspring are feasible, continue
            # looping if not
            feasible = feas(child1) and feas(child2)
            loops += 1
            if loops > N_u:
                return None, None
        return child1, child2
    
    def _breed(self, parents):
        """Breed a new generation of individuals
        
        parents -- parent population
        """
        uniform = False
        if uniform:
            p_cross = self._swapping_probability
            crossover = self._unif_crossover
        else:
            p_cross = 5
            crossover = self._np_crossover
        
        rand = numpy.random.randint
        feas = self._solution_feasibility
        util = self._solution_utility
        mutate = self._mutate
        p_mutate = self._mutation_probability
        
        N_parents = len(parents)
        F = self._breeding_rate
        N_offspring = int(N_parents * F)
        offspring = []
        maxloops = N_offspring * 100
        loops = 0
        while len(offspring) < N_offspring:
            # randomly select two parents (and make sure that the two parents
            # are not the same individual)
            i_p1, i_p2 = rand(0, N_parents, 2)
            while i_p1 == i_p2:
                i_p1, i_p2 = rand(0, N_parents, 2)
            parent1, parent2 = parents[i_p1][1], parents[i_p2][1]
            
            # breed two children by crossover and mutate the genes
            # of both children with mutation probability p_mutation
            child1, child2 = crossover(p_cross, parent1, parent2)
            if (child1, child2) != (None, None):
                child1 = mutate(p_mutate, child1)
                child2 = mutate(p_mutate, child2)
            
            # add feasible children to the offspring together with
            # their fitness values
            if child1 is not None and feas(child1):
                fitness = util(child1)
                offspring.append((fitness, child1))
            if child2 is not None and feas(child2):
                fitness = util(child2)
                offspring.append((fitness, child2))
            loops += 1
            if loops > maxloops:
                msg = "Cannot breed new generation due to too strict "\
                      "constraints"
                self._add_error(msg)
                break
        # join the parents and offspring together into a new population
        # and sort the new population by the fitness value
        population = parents + offspring
        self._sort_by_fitness(population)
        return population
    
    def _select_fittest(self, population):
        """Select the fittest individuals
        
        population -- whole population including parents and offspring
        """
        N_pop = self._population_size
        return population[:N_pop]

    def _run_GA(self,):
        """
        Run genetic algorithm search for a single iteration/repeat
        
        Vocalbulary:
        
            population -- set of alternative solutions
            individual -- single solution
            generation -- single iterative round
            gene -- single unit (eg. forest stand)
            allele -- one branch of a single unit
            crossover -- an exchange of parts between two parent solutions
            mutation -- random change of branch (ie. change of treatment)
            fitness -- utility function value
            offspring -- new generation of solutions from breeding
        """
        comp = self._compare_utilities
        
        # generate initial population with predefined number of individuals
        # and evaluate the fitness for each individual in the population
        population = self._populate()
        if population is None:
            return
        
        # draw random individual and assign it as the best so far
        b = population[numpy.random.randint(0, len(population))]
        f_best, s_best = b  # so far best fitness and individual
        f_curr = f_best  # currently best fitness
        s_curr = s_best  # currently best individual (solution)
        # the search is considered converged after one tenth of the maximum
        # generations have been bred without improvement in best fitness
        # TODO: come up with a better convergence criteria
        convergence_limit = int(self._max_iterations / 10.)
        if convergence_limit < 5:
            convergence_limit = 5
        
        hang_count = 0
        for generation in range(self._max_iterations):
            start = time.time()
            # breed a new generation of individuals and select the fittest
            # of the parent population and the offspring to form a new
            # population
            population = self._breed(population)
            population = self._select_fittest(population)
            f_curr, s_curr = population[0]
            if comp(f_best, f_curr):
                f_best = f_curr
                s_best[:] = s_curr[:]
                hang_count = 0
            else:
                hang_count += 1
            
            msg = "  %d. generation bred in %0.1f secs, utility = %0.3f" \
                  % (generation + 1, time.time() - start, f_curr)
            self._add_info(msg)
            
            # check if stopping conditions are met; has the increase of the
            # fitness converged?
            if hang_count > convergence_limit:
                msg = "  converged at %d iterations" % (hang_count,)
                self._add_info(msg)
                break
        
        # store the best individual (solution) and fitness (utility)
        self._best_sol[:] = s_best[:]
        self._best_U = f_best

    def optimize(self):
        """Run GA algorithm
        """
        runf = self._run_GA
        repeats = self._repeats
        self._run_algorithm(runf, repeats, None)


def main():
    import doctest
    doctest.testfile('./test/genetic_algorithm-py.rst')

if __name__ == '__main__':
    main()
