#!/usr/bin/env python
# encoding: utf-8
"""
Title: SIMO metaheuristic HERO optimization algorithm

...

*hero.py*

Copyright (c) 2007 Simosol Oy. Distributed under the GNU General Public License
2.0.
"""
# these are for running the doctest from src/simo
#import sys
#sys.path.append('..')

import copy
import numpy
from optimizer import Optimizer


ARGUMENTS = ('repeats',
             'initial_solutions',
             'maximum_iterations')
MAX_NO_IMPROVEMENT = 2


class HERO(Optimizer):
    """HERO optimization algorithm implementation
    """
    def _process_arguments(self, args):
        """Process keyword arguments
        """
        self._abort = False
        for arg in ARGUMENTS:
            if not arg in args:
                self._add_error('Parameter "%s" missing!' % arg)
                self._abort = True
        if self._abort:
            return
        try:
            self._repeats = int(args["repeats"])
            self._initial_solutions = int(args["initial_solutions"])
            self._max_iterations = int(args["maximum_iterations"])
        except:
            self._add_error('Invalid parameter values!')
            self._abort = True

    def _get_best_solution(self, unit):
        """Get the best feasible random solution for the current unit

        unit -- unit index as int
        """
        feas = self._solution_feasibility
        util = self._solution_utility
        comp = self._compare_utilities
        b_best = self._best_sol[unit]
        U_best = self._best_U
        candidate = numpy.copy(self._best_sol)
        for b in self._branches[unit]:
            candidate[unit] = b
            feasible = feas(candidate)
            if feasible:
                U_new = util(candidate)
                if comp(U_best, U_new):
                    b_best = b
                    U_best = U_new
                    self._solution_improved = True
        self._best_sol[unit] = b_best
        self._best_U = U_best

    def _run_HERO(self):
        """
        Run HERO algorithm for a single iteration/repeat

        iteration -- current simulated iteration as int
        """
        # iterate until stop conditions are met, stop after maximum iterations
        # or after the solution has not improved during n last iterations
        noi = 0  # iterations without improvement in utility
        for i in range(self._max_iterations):
            self._solution_improved = False
            for unit in range(self._num_of_units):
                self._get_best_solution(unit)
            self._stat_logger.add_round(self._best_sol,
                                        self._best_U,
                                        self._best_U)
            if self._solution_improved is False:
                noi += 1
                if noi > MAX_NO_IMPROVEMENT:
                    break

    def optimize(self):
        """Run HERO optimization algorithm
        """
        runf = self._run_HERO
        repeats = self._repeats
        initial_solutions = self._initial_solutions
        self._run_algorithm(runf, repeats, initial_solutions)


def main():
    import doctest
    doctest.testfile('./test/hero-py.rst')

if __name__ == '__main__':
    main()
