#!/usr/bin/env python
# encoding: utf-8
"""
Title: SIMO interface to JLP linear programming library

...

*jlp.py*

Copyright (c) 2007 Simosol Oy. Distributed under the GNU General Public License
2.0.
"""

###############################################################################
# TODO: syntax checking for optimistation task definition:
#           objective function type: additive or distance
#               if distance, optimal_level must be given
#           utility function type: must be linear
#           constraint: ge or le are not allowed due to jlp!
#       add goal programming formulation for distance objective function
#           see jlp manual chp 2.3.2
#       domain specifications in xml & in opt data generation

import os
import shutil
import numpy
import subprocess
import threading
import traceback
import re
from collections import defaultdict
from dateutil.relativedelta import relativedelta

from optimizer import Optimizer
from genetic_algorithm import GeneticAlgorithm
from simulated_annealing import SimulatedAnnealing
from tabusearch import TabuSearch
from hero import HERO
from simo.builder.names import EXPR_DISCOUNT, EXPR_DTYPE_COMB,\
                               EXPR_DTYPE_DATA, EXPR_DTYPE_OP,\
                               EXPR_DTYPE_VALUE, EXPR_OPERATION


class RunCmd(threading.Thread):
    def __init__(self, cmd, timeout, logger, log_name, path):
        threading.Thread.__init__(self)
        self.cmd = cmd
        self.timeout = timeout
        self.path = path
        self.logger = logger
        self.log_name = log_name

    def run(self):
        try:
            self.p = subprocess.Popen(self.cmd, cwd=self.path, 
                                      stdout=subprocess.PIPE, 
                                      stderr=subprocess.PIPE)
            stdout, stderr = self.p.communicate()
            stdout = stdout.replace('\r\n', '\n').replace('\r', '\n')
            stderr = stderr.replace('\r\n', '\n').replace('\r', '\n')
            if stderr:
                for row in stderr.split('\n'):
                    self.logger.log_message(self.log_name,\
                            'error', unicode(row.rstrip().decode('utf-8')))
            if stdout:
                for row in stdout.split('\n'):
                    self.logger.log_message(self.log_name,\
                                'info', unicode(row.rstrip().decode('utf-8')))
        except Exception:
            error = traceback.format_exc()
            self.logger.log_message(self.log_name, 'error', error.strip())
            # This is in a thread, exceptions are not propagated

    def Run(self):
        self.start()
        self.join(self.timeout)

        if self.is_alive():
            self.p.terminate()
            self.join()

# equality operator conversion from SIMO to J
OPCONV = {'eq': '.eq.',
          'ue': '<>',
          'gt': '>',
          'lt': '<',}
EQOPS = set(['gt','lt','eq','ue'])
JOPS = set(['+','-','*'])
TRANSOPS = set(['*'])


class JLPConstraints(object):
    def __init__(self):
        """
        Initialize a data structure for storing JLP constraints
        """
        self.infix = []
        self.names = []
        self.order = []


class JLP(Optimizer):
    """
    JLP implementation of LP optimization algorithm. Needs a J installation
    to work (see: http://www.metla.fi/products/J/)
    """

    def _process_arguments(self, args):
        """Process JLP arguments
        """
        parameters = ('maxobjects', 
                      'jfolder', 
                      'prefix',)
        for parameter in parameters:
            if not parameter in args:
                self._add_error("Parameter '%s' missing" % parameter)
                self._abort = True
        if self._abort:
            return
        pf = args['prefix']
        self.j_prefix = pf
        timeout = args['timeout']
        self.j_max_objects = str(args['maxobjects'])
        self.j_folder = os.path.abspath(args['jfolder'])
        self.j_par = os.path.join(self.j_folder, 'j.par')
        self.j_problem_name = pf + '_problem.txt'
        self.j_unit_data = os.path.join(self.j_folder, pf + '_unitdata.txt')
        self.j_var_data = os.path.join(self.j_folder, pf + '_vardata.txt')
        self.j_problem = os.path.join(self.j_folder, self.j_problem_name)
        self.j_branch_result = os.path.join(self.j_folder, pf + '_bresult.txt')
        self.j_weight_result = os.path.join(self.j_folder, pf + '_wresult.txt')
        self.j_jlp_output = os.path.join(self.j_folder, pf + '_jlp_output.txt')
        self.res_folder = os.path.join(self.j_folder, 'j_results')
        if 'do_split' not in args:
            self.do_split = False
        else:
            self.do_split=args['do_split']
        if not self.do_split:
            self.skip_weight_splits = None
        else:
            self.skip_weight_splits=args['skip_weight_splits']
        if 'do_fallback' in args:
            self.do_fallback = args['do_fallback']
        else:
            self.do_fallback = False
        if 'fallback_method' in args:
            self.fallback_method = args['fallback_method']
        else:
            self.fallback_method = None
        if 'fallback_settings' in args:
            self.fallback_settings = args['fallback_settings']
        else:
            self.fallback_settings = None
        self.j_consts = None
        self.j_gp_rhs = None
        self._surplus_vars = None
        self._slack_vars = None
        self.do_transform = False
        cmd = os.path.join(self.j_folder, 'j')
        # run in thread, wait max timeout seconds and terminate
        #  this because J can drop to sit> in case of internal error
        self.j_runner = RunCmd([cmd], timeout, self._logger, self._log_name,
                               self.j_folder)

    def _process_branch(self, unitsol, weights):
        """
        Process unit that was split by the JLP algorithm
        """
        parts = len(unitsol)
        if parts > 1:
            for i in range(1, parts):
                unitid, branchid = unitsol[i]
                # grow the sizes of self._best_sol & self._best_weights arrays
                self._best_sol = numpy.insert(self._best_sol,
                                              self._best_sol.size, 0)
                self._best_weights = numpy.insert(self._best_weights,
                                                  self._best_weights.size, 1)
                self.nunits += 1
                # store the branchid for the new item in self._best_sol
                newid = ''.join(['lp',str(i).decode('utf-8'), '_', unitid])
                self._data.unit_index[newid] = self.nunits - 1
                self._data.orig_uids[newid] = unitid
                self._data.unit_ids.append(newid)
                unitsol[i] = (newid, branchid)
        for i in range(parts):
            unitid, branchid = unitsol[i]
            ind = self._data.unit_index[unitid]
            self._best_sol[ind] = branchid
            if parts > 1:
                weight = weights[i]
            else:
                weight = 1.
            self._best_weights[ind] = weight

    def _process_infeasible(self):
        """
        Process JLP output when the problem is infeasible
        """
        jlpfile = open(self.j_jlp_output, 'r')
        trapped_constraint = False
        inf_at_start = False
        for line in jlpfile:
            if 'infeasible at start' in line:
                inf_at_start = True
            part = line.split()
            c_num = None
            if inf_at_start:
                if part[0] == 'row':
                    try:
                        c_num = int(part[1]) - 1
                    except:
                        # invalid jlp output format for this code
                        pass
            else:
                if part[-1] in ('U', 'L'):
                    try:
                        c_num = int(part[0].split(')')[0]) - 1
                    except:
                        # invalid jlp output format for this code
                        pass
            if c_num is not None:
                msg = 'Infeasible constraint number %s (starting from 0): \n'\
                        '   output from jlp: \n'% c_num + line
                self._logger.log_message(self._log_name, 'error', msg)
                trapped_constraint = True
        if not trapped_constraint:
            msg = "Couldn't extract the limiting constraint(s) from J output"\
                    "\n  J version mismatch between SIMO and your J"\
                    "\n  check file jlp_output.txt in your J folder manually"
            self._logger.log_message(self._log_name, 'error', msg)

    def _copy_result_file(self, respath, iteration):
        """
        Copy J result file to result folder
        """
        fname = respath.split(os.path.sep)[-1]
        left, right = fname.split('.')
        target = os.path.join(self.res_folder, left + '_i%d.' % iteration + \
                              right)
        if not os.path.isfile(respath):
            errormsg = "J result file '%s' not found, results could not be "\
                       "copied" % respath
            self._add_error(errormsg)
        else:
            shutil.copy(respath, target)

    def _check_violations(self, solution, cur_dir):
        """
        Check that the jlp solution constraints do not have severe violations
        due to LP relaxation
        """
        use_jlp_solution = True
        
        if not self.do_fallback:
            # If fallback is not enabled, always use the jlp solution
            return use_jlp_solution
                
        has_violations = False
        omatrix = self._data
        nconst = omatrix.task_def.num_of_consts
        if nconst > 0:
            msg = "Checking JLP solution for constraint violations..."
            self._logger.log_message(self._log_name, 'INFO', msg)
            evaluate = omatrix._evaluator.evaluate
            dlocs = omatrix.data['date_locs']
            data = omatrix.data['operands']
            for i, c in enumerate(omatrix.task_def.constraints):
                if c.violation_metric is None:
                    msg = "  skipping violation check for constraint %d. as "\
                          "violation metric is missing" % (i + 1)
                    self._logger.log_message(self._log_name, 'INFO', msg)
                    continue
                evalres, violations = evaluate(omatrix._unit_range, solution, 
                                        c.expr, dlocs, data, 
                                        violation_metric=c.violation_metric)
                # If the constraint is valid, the violations list
                # should contain single value
                if max(violations) > c.max_violation:
                    has_violations = True
                    msg = "  constraint %d. violated by %0.1f (max "\
                          "violation %0.1f)" % (i + 1, max(violations), 
                                               c.max_violation)
                    self._logger.log_message(self._log_name, 'INFO', msg)
                
        # If there were violations greater than the max_violation 
        # value, fall back to selected heuristic.
        if has_violations:
            use_jlp_solution = False
            
            # Set the heuristic optimizer settings
            args = (self._logger, self._log_name, self._data.task_def,
                    self._db_in, self._db_out,
                    self._data._keep_data, self._data._use_existing, 
                    self._data.limit2branches, None,
                    self._track_solution, self._run_id, self._create_opt_db)
        
            # Get optimization method-specific settings from configuration file
            settings = self.fallback_settings
            fb_method = self.fallback_method
            
            # Initialize the fallback optimizer
            optimizer = None
            if fb_method == 'ts':
                opmethod_name = 'Tabu Search'
                optimizer = TabuSearch(*args, **settings)
            elif fb_method == 'ta':
                opmethod_name = "Threshold accepting"
                msg = "Threshold accepting not yet implemented!"
                self.logger.log_message(self._log_name, 'error', msg)
                return False
            elif fb_method == 'hero':
                opmethod_name = 'HERO'
                optimizer = HERO(*args, **settings)
            elif fb_method == 'sa':
                opmethod_name = 'Simulated Annealing'
                optimizer = SimulatedAnnealing(*args, **settings)
            elif fb_method == 'ga':
                opmethod_name = 'Genetic Algorithm'
                optimizer = GeneticAlgorithm(*args, **settings)
            
            # Run the optimization algorithm and close the databases
            if optimizer:
                msg = "-" * 50
                self._logger.log_message(self._log_name, 'INFO', msg)
                msg = "Reverting from JLP to %s due to constraint "\
                      "violations" % opmethod_name
                self._logger.log_message(self._log_name, 'INFO', msg)
                optimizer.optimize()
                optimizer.close()
            
        return use_jlp_solution

    def _read_solution(self, iteration):
        """
        Read and process solution from JLP output files
        """
        # copy the result files into a result folder
        if not os.path.isdir(self.res_folder):
            os.makedirs(self.res_folder)
        self._copy_result_file(self.j_branch_result, iteration)
        self._copy_result_file(self.j_jlp_output, iteration)
        if not os.path.isfile(self.j_branch_result):
            errormsg = "J result branch file '%s' not found, could not read "\
                       "solution" % self.j_branch_result
            self._add_error(errormsg)
            return False
        bresfile = open(self.j_branch_result, 'r')
        if os.path.isfile(self.j_weight_result):
            self._copy_result_file(self.j_weight_result, iteration)
            wresfile = open(self.j_weight_result, 'r')
        prevunit = None
        unitsol, weights = [], []
        for line in bresfile:
            if line == 'infeasible problem\n':
                self._process_infeasible()
                return False
            else:
                items = [float(item) for item in line.split()]
                unitind, branchid, unitshadowprice = items
                # J id is the index of the SIMO id
                unit_id = self._data.unit_ids[int(unitind)]
                branch_ind = int(branchid-1)
                branchid = self._data.branches[unitind][branch_ind]
                if prevunit != unit_id and prevunit is not None:
                    branchcount = len(unitsol)
                    if branchcount > 1:
                        # read partial weights for lp-split units
                        for i in range(branchcount):
                            wline = wresfile.readline()
                            winfo = [float(item) for item in wline.split()]
                            weights.append(winfo[2])
                    self._process_branch(unitsol, weights)
                    unitsol, weights = [], []
                branch = (unit_id, branchid)
                unitsol.append(branch)
                prevunit = unit_id
        # process the last unitsol
        branchcount = len(unitsol)
        if branchcount > 1:
            # read partial weights for lp-split units
            for i in range(branchcount):
                wline = wresfile.readline()
                winfo = [float(item) for item in wline.split()]
                weights.append(winfo[2])
        self._process_branch(unitsol, weights)
        # for debugging purposes, write a file containing all input variable
        # values for the optimal solution
        vardata = defaultdict(list)
        fp = open(self.j_var_data, 'r')
        n = self.orig_nunits
        for i, unitid in enumerate(self._data.unit_ids[:n]):
            try:
                branches = self._branches[i]
            except:
                # if simulation is terminated, it may be that the unit is not
                # in the optimisation data at all
                continue
            for j in branches:
                row = fp.readline()
                vardata[unitid].append(row.strip())
        fp.close()
        target = os.path.join(self.res_folder, '%s_vardata_i%d.txt' % \
                              (self.j_prefix, iteration))
        fp = open(target, 'w')
        # regex for getting the original id from split stand id
        p = re.compile('lp\d+_(.+)')
        for i, unitid in enumerate(self._data.unit_ids):
            origid = unitid
            # now we are processing stands split by LP
            if i >= n:
                m = p.search(unitid)
                origid = m.groups()[0]
            branch = self._best_sol[i]
            weight = self._best_weights[i]
            try:
                row = vardata[origid][branch]
            except:
                # if simulation is terminated, it may be that the unit is not
                # in the optimisation data at all
                continue
            uid = unitid  # this should be unicode
            oid = origid  # this should be unicode
            b = ("%d" % branch).decode('utf-8')
            w = ("%0.6f" % weight).decode('utf-8')
            r = row.decode('utf-8')
            unic = ','.join([uid, oid, b, w, r])
            fp.write((unic + u'\n').encode('utf-8'))
        fp.close()
        return True

    def _remove_j_files(self):
        """
        Remove J input and output files
        """
        if os.path.exists(self.j_par):
            os.remove(self.j_par)
        if os.path.exists(self.j_problem):
            os.remove(self.j_problem)
        if os.path.exists(self.j_unit_data):
            os.remove(self.j_unit_data)
        if os.path.exists(self.j_var_data):
            os.remove(self.j_var_data)
        if os.path.exists(self.j_branch_result):
            os.remove(self.j_branch_result)
        if os.path.exists(self.j_weight_result):
            os.remove(self.j_weight_result)
        if os.path.exists(self.j_jlp_output):
            os.remove(self.j_jlp_output)

    def _write_j_data(self):
        """
        Write the optimisation data to two data files for J. The unit data
        contains the simulation unit id and number of simulated alternatives
        for each unit. The variable contains the objective and constraint
        variable values for each alternative in each simulation unit.

        Simulation unit ids for J are taken from an integer sequence and a
        mapping between J ids and original ids is generated. The same applies
        for branch identifiers.
        """
        unitfile = open(self.j_unit_data, 'w')
        varfile = open(self.j_var_data, 'w')
        subs = self._data.task_def.subobjectives
        consts = self._data.task_def.constraints
        dlocs = self._data.data['date_locs']
        data = self._data.data['operands']
        evaluate = self._data._evaluator.evaluate
        dtypes = (EXPR_DTYPE_COMB, EXPR_DTYPE_DATA, EXPR_DTYPE_OP)

        # float formatter
        fmt = "%0.6f"

        # localize numpy functions
        _sum = numpy.sum
        _isnan = numpy.isnan

        for ind, oid in enumerate(self._data.unit_ids):
            try:
                branches = self._data.branches[ind]
            except:
                # if simulation has been terminated for the unit, it may be
                # that it's not in the optimisation data
                continue
            rowdata = []
            for b_ind, br in enumerate(branches):
                rowdata.append([])

                # subobjective variables first
                for sub in subs:
                    value = evaluate(ind, br, sub.expr, dlocs, data)
                    if numpy.isnan(value):
                        value = 0.0
                    s = (fmt % value).rstrip('0')
                    rowdata[b_ind].append(s)

                # then the constraint variables
                if consts:
                    for const in consts:
                        for operand in const.expr:
                            if operand[0] not in dtypes:
                                continue
                            oi, ui = operand[-1]
                            # If either of date locs are None, the value is 
                            # considered to be missing.
                            start, end = dlocs[oi]
                            if None in (start, end):
                                value = 0.0
                            else:
                                value = _sum(data[ui][ind, br, start:end+1])
                            if _isnan(value):
                                value = 0.0
                            s = (fmt % value).rstrip('0')
                            rowdata[b_ind].append(s)

            # check that the number of rows matches branches
            if len(rowdata) != len(branches):
                errormsg = "Error writing data for unit '%s'" % str(oid)
                self._add_error(errormsg)
                continue

            # write the unit and variable data
            unitfile.write('%d,%d\n' % (ind, len(branches)))
            for row in rowdata:
                varfile.write(','.join(row) + '\n')

        unitfile.close()
        varfile.close()

    def _update_stack(self, stack):
        """
        Update infix stack
        """
        # if there's multiplication with a constant, stack won't be needing 
        # updating
        if len(stack) > 1:
            a = stack.pop()
            b = stack.pop()
            stack.append(b + a)

    def _process_const_val(self, postfix, expr):
        """
        Process constraint value
        """
        postfix.append("%0.2f" % (expr[1]))

    def _process_const_var(self, postfix, namestack, orderstack, expr, ind):
        """
        Process constraint variable (i.e. data or operation variable)
        """
        if expr[0] == EXPR_DTYPE_DATA:
            name = expr[1][1]
        else:
            name = expr[1]
        
        from_date, to_date = expr[2]
        step = self._data.task_def.period_step
        init_date = self._data.task_def.init_date
        
        from_delta = relativedelta(from_date, init_date)
        to_delta = None
        if to_date:
            to_delta = relativedelta(to_date, init_date)
        
        if step == 'year':
            start = from_delta.years + 1
            end = to_delta.years + 1 if to_delta else 'end'
        elif step == 'month':
            add = from_delta.years * 12
            start = add + from_delta.months + 1
            end = add + to_delta.months + 1 if to_delta else 'end'
        
        start, end = str(start), str(end)
              
        jname = "c%d" % ind
        truename = "%s, steps %s-%s" % (name, start, end,)
        self._c_var_names[jname] = truename
        postfix.append(jname)
        namestack.append([jname])
        orderstack.append([ind])

    def _process_const_op(self, postfix, namestack, orderstack, expr, ind,
                          last):
        """
        Process constraint operator
        """
        simoop = expr[2]
        if simoop in ('and', 'or'):
            errorstr = "Cannot handle operator '%s' in J "\
                       "constraints" % simoop
            self._add_error(errorstr)
            self._init_ok = False
        # arithmetic operators do not need to be converted, but eq
        # operators do
        if expr[0] == 'ari':
            if simoop not in JOPS:
                errorstr = "Invalid operator for J '%s'" % simoop
                self._add_error(errorstr)
                self._init_ok = False
            jop = simoop
        else:
            if simoop not in OPCONV:
                errorstr = "Invalid operator for J '%s'" % simoop
                self._add_error(errorstr)
                self._init_ok = False
                jop = None
            else:
                jop = OPCONV[simoop]
        # multiplying variables is illegal in JLP constraints (such as
        # X1*Y1 - X2*Y2 > 0), but in SIMO these types of constraints are
        # perfectly ok
        # in order to solve problems with these type of constraints, the
        # constraints have to be transformed in J
        a = postfix.pop()
        b = postfix.pop()
        if jop in TRANSOPS:
            tvar = 'tr%d' % (len(self.transformations) + 1)
            transf = "%s = %s %s %s" % (tvar, b, jop, a)
            self.transformations.append(transf)
            self.do_transform = True
            postfix.append(tvar)
        else:
            postfix.append("%s %s %s" % (b, jop, a))
            
        # check that the operands are valid if processing the last
        # operands as J constraints must be of form: ... > 0
        if last:
            if simoop not in EQOPS:
                errorstr = "Invalid JLP constraint definition"
                self._add_error(errorstr)
                self._init_ok = False
            try:
                float(a)
            except:
                errorstr = "JLP condition right hand side must "\
                           "be a numerical value"
                self._add_error(errorstr)
                self._init_ok = False
        else:
            # update name and order stacks
            self._update_stack(namestack)
            self._update_stack(orderstack)

    def _process_const_expr(self, expressions):
        """
        Process constraint postfix expression back to infix form and try to
        figure out if the expression is good for JLP.

        expressions -- SIMO constraint postfix expressions in a list
        """
        num_of_expr = len(expressions)
        self.j_consts = JLPConstraints()
        self.transformations = []
        self._c_var_names = {}
        ind = 0
        for i in range(num_of_expr):
            # initialize empty stacks
            postfix, namestack, orderstack = [], [], []
            for j, expr in enumerate(expressions[i].expr):
                last = j == (len(expressions[i].expr) - 1)
                etype = expr[0]
                if etype == EXPR_DTYPE_VALUE:
                    self._process_const_val(postfix, expr)
                elif etype in set([EXPR_DTYPE_COMB,
                                   EXPR_DTYPE_OP,
                                   EXPR_DTYPE_DATA]):
                    self._process_const_var(postfix, namestack, orderstack,
                                            expr, ind)
                    ind += 1
                elif etype in set(['ari', 'eq']):
                    self._process_const_op(postfix, namestack, orderstack,
                                           expr, ind, last)
            # store the processed expression infix string and the current
            # constraint variable name and order stacks
            if len(postfix) > 0:
                self.j_consts.infix.append(postfix[0])
                self.j_consts.names += namestack[0]
                self.j_consts.order += orderstack[0]

    def _process_gp_vars(self, no):
        """
        Construct slack and surplus variables needed in JLP goal
        programming tasks
        """
        subobjs = self._data.task_def.subobjectives
        self.j_gp_rhs = [s.optlevel for s in subobjs]
        self._surplus_vars = ["sp%d" % x for x in range(no)]
        self._slack_vars = ["sl%d" % x for x in range(no)]

    def _write_j_par(self, partype):
        """
        Write J parameter file (j.par)
        """
        if partype == 'init':
            par = "!" + self.j_max_objects + "\n;incl('" + \
                  self.j_problem_name + "')\n"
        elif partype == 'close':
            par = "*" + self.j_max_objects + '\n'
        jpar = open(self.j_par, 'w')
        jpar.write(par)
        jpar.close()

    def _write_j_problem(self):
        """
        Write the optimisation problem definition to a file for J.
        """
        probfile = open(self.j_problem, 'w')

        # write out some comments on variable names etc.
        legend = '! SUBOBJECTIVE VARIABLES\n'
        for i, ss in enumerate(self._data.task_def.subobj_strings):
            legend += '!   x%d = %s\n' % (i, ss)

        # write out constraint variable labels
        if self.j_consts:
            cvars = self._c_var_names.keys()
            cvars.sort()
            legend += '\n! CONSTRAINT VARIABLES (see SIMO lexicon for '\
                      'details)\n'
            for cvar in cvars:
                legend += "!   %s = %s\n" % (cvar, self._c_var_names[cvar])
        probfile.write(legend + '\n')

        # data
        d = "! READ OBJECTIVE AND CONSTRAINT VARIABLE VALUES FROM FILE\n"
        d += "udat=data(read->(id,nbranch),in->'"
        d += os.path.split(self.j_unit_data)[1] + "')\n"
        d += 'sdat=data(read->('
        rowl = len(d)
        # (sub)objective variables
        first_one = True
        for ov in self._obj_vars:
            rowl += (len(ov) + 1)
            if first_one:
                first_one = False
            else:
                d += ','
            if rowl >= 200:
                d += '\n'
                rowl = len(ov)
            d += ov
        # constraints
        if self.j_consts:
            for cn in self.j_consts.names:
                rowl += (len(cn) + 1)
                d += ','
                if rowl >= 200:
                    d += '\n'
                    rowl = len(cn)
                d += cn
        d += "), in->'"
        d += os.path.split(self.j_var_data)[1] + "')\n"
        d += 'linkdata(data->udat,subdata->sdat,nobsw->nbranch)\n'

        # possible transformations
        if self.do_transform:
            d += "\n! TRANSFORM ILLEGAL CONSTRAINTS\n"
            d += "strans=trans()\n"
            for row in self.transformations:
                d += "%s\n" % row
            d += "/\n"
        probfile.write(d)

        # the problem
        probfile.write('\n! OPTIMISATION PROBLEM DEFINITION\n')
        probfile.write('pr=problem()\n')

        # objective function
        probfile.write('\n! OBJECTIVE FUNCTION DEFINITION\n')
        # the objective function for 'distance' problems is a bit different
        # as we must utilize the GP functionalities of J
        if self._data.task_def.objective_func.type == 'distance':
            slack_and_surplus = ''
            opers = []
            for i, so in enumerate(self._data.task_def.subobjectives):
                weight = str(so.weight)
                row = ''.join([weight, '*',
                               self._obj_vars[i], '-',
                               self._surplus_vars[i], '+',
                               self._slack_vars[i], '=',
                               str(self.j_gp_rhs[i])])
                slack_and_surplus += row + '\n'
                opers.append(''.join([weight, '*',
                                      self._surplus_vars[i], '+',
                                      weight, '*',
                                      self._slack_vars[i]]))
            orow = '+'.join(opers)
            if len(orow) >= 200:
                items = orow.split('+')
                n, orow = len(items[0]), '%s+' % items[0]
                for i, item in enumerate(items[1:]):
                    if (n + len(item)) >= 200:
                        orow += '\n'
                        n = 0
                    orow += '%s+' % item
                    n += (len(item) + 1)
            if orow[-1] == '+':
                orow = orow[:-1]
            # TODO: should this be always 'min' in J goal programming?
            orow += '==' + self._data.task_def.objective_func.target + '\n'
            ofun = orow + slack_and_surplus
        else:
            opers = []
            for i, so in enumerate(self._data.task_def.subobjectives):
                weight = str(so.weight)
                opers.append(''.join([weight, '*', self._obj_vars[i]]))
            ofun = '+'.join(opers)
            # max or min
            ofun += '==' + self._data.task_def.objective_func.target + '\n'

        # write the objective function into a file
        probfile.write(ofun)

        # constraints
        if self.j_consts:
            probfile.write("\n! CONSTRAINT DEFINITIONS\n")
            for crow in self.j_consts.infix:
                probfile.write(crow + "\n")
        probfile.write('/\n')

        # optimise
        bresultfile = self.j_branch_result
        wresultfile = self.j_weight_result
        if self.do_transform:
            jlp = "jlp(problem->pr,data->udat,subtrans->strans"
        else:
            jlp = "jlp(problem->pr,data->udat"
        if self._data.task_def.objective_func.type == 'distance':
            jlp += ',z->'
        jlp += ",report->'"+self.j_jlp_output+"')\n" \
            'branch_report=trans()\n' \
            'do(i,1,weights())\n' \
            "write('"+bresultfile+"',$,matrix%udat(unit(i),1),schedw(i),price%unit(unit(i)))\n" \
            'enddo\n' \
            '/\n' \
            'weight_report=trans()\n' \
            "do(i,1,partweights())\n" \
            "write('"+wresultfile+"',$,partunit(i),partschedw(i),partweight(i))\n" \
            "enddo\n" \
            '/\n' \
            ';if(Feasible);then\n'\
            'call(branch_report)\n' \
            'call(weight_report)\n' \
            ';else\n'\
            "write('"+bresultfile+"','(~infeasible problem~)')\n" \
            ';endif\n'\
            'end\n'
        probfile.write(jlp)
        probfile.close()

    def optimize(self):
        """
        Construct J input files, run J program as a subprocess and process
        J output
        """
        cur_dir = os.path.abspath(os.path.curdir)
        # first analyze the optimization input databases
        if not self._analyze_data():
            errorstr = 'Invalid input database, optimization aborted'
            self._add_error(errorstr)
            return
        
        self._init_ok = True
        self.nunits = self.orig_nunits = self._data.num_of_units
        self._best_weights = numpy.ones(self.nunits, dtype=float)
        self.j_consts = None
        self.j_gp_rhs = None
        
        # process the (sub)objective variables and construct names for the
        # variables
        # NB: (sub)objective variables are x0, x1, x2, ..., xN as the
        # variables are actually whole expressions that have been evaluated
        # already when constructing the data
        no = self._data.task_def.num_of_subobjs
        self._obj_vars = ["x%d" % x for x in range(no)]
        
        # if objective function type is 'distance', collect also the target
        # levels, which will then be used as RHS (right hand side) value in
        # GP (goal programming) J problem, and construct also the z and slack
        # variables
        if self._data.task_def.objective_func.type == 'distance':
            self._process_gp_vars(no)
            
        # process the constraints and construct names for the constraint
        # variables
        self.do_transform = False
        if self._data.task_def.constraints:
            self._process_const_expr(self._data.task_def.constraints)
            
        # write J control files
        os.chdir(self.j_folder)
        self._remove_j_files()
        self._write_j_par('init')
        self._write_j_problem()
        os.chdir(cur_dir)
        
        if not self._init_ok:
            msg = "Errors in JLP initialization - aborting"
            self._add_error(msg)
            return
        
        use_jlp_solution = True
        
        # the optimization process is repeated for each iteration
        for iteration in range(self._data.max_iterations):
            # first, set construct dataset for the current iteration
            os.chdir(cur_dir)
            if not self.set_data(iteration):
                msg = "Could not construct data set - skipping iteration %d" \
                      % iteration
                self._add_error(msg)
                continue
            os.chdir(self.j_folder)
            
            # write J data file
            msg = 'Writing JLP data files for iteration %d' % iteration
            self._add_info(msg)
            self._write_j_data()
            if not self._init_ok:
                msg = "Error constructing J input files for iteration %d"\
                      " - aborting optimization" % iteration
                self._add_error(msg)
                continue
            self._add_info('Running JLP for iteration %s' % iteration)
            try:
                self.j_runner.Run()
            except Exception:
                msg = 'Error running J software for iteration %d' % iteration
                self._add_error(msg)
                continue
            feasible = self._read_solution(iteration)
            if feasible:
                # Use only the original units when evaluating the solution
                S_best = self._best_sol[:self.orig_nunits]
                status = self._data.get_status(S_best)
                for sub in status['subobjectives']:
                    utility = sub[1]
                    if self._best_U is None:
                        self._best_U = utility
                    else:
                        self._best_U += utility
                
                # If do_fallback is True, check the constraints for violation 
                # settings and if any of the constraints are violated. If so, 
                # then fall back to a heuristic method.
                use_jlp_solution = self._check_violations(S_best, cur_dir)
                
                if use_jlp_solution:
                    # NB: Use now only the original units for calculating the 
                    # utility value.
                    limit = self.orig_nunits
                    #limit = None if self.do_split else self.orig_nunits
                    self._store_solution(limit, self.do_split)
                    # Store optimal solution for the current iteration
                    self._store_results(None, iteration, False,
                                        self.skip_weight_splits)
        
        self._write_j_par('close')
        
        # change back to current directory
        os.chdir(cur_dir)


def main():
    pass

if __name__ == '__main__':
    main()
