#!/usr/bin/env python
# encoding: utf-8
"""
Title: Super class for optimization algorithms

...

*optimizer.py*

Copyright (c) 2007 Simosol Oy. Distributed under the GNU General Public
License 2.0.
"""
# these are for running the doctest from src/simo
#import sys
#sys.path.append('..')

import numpy
import time

from simo.optimization.tools.omatrix import OMatrix
from simo.optimization.tools.ologger import OLogger

MAX_LOOP_FACTOR = 10

class Optimizer(object):
    """Parent class for all heuristic optimization algorithms.
    """
    def __init__(self, logger, logname, taskdef, dbin, dbout, keep_data,
                 use_existing, limit2branches, chart_path, track_solution,
                 run_id, create_opt_db, **keywords):
        """docstring for __init__"""
        super(Optimizer, self).__init__()
        self._abort = False
        self._logger = logger
        self._log_name = logname
        self._stat_logger = OLogger(self._logger, self._log_name, chart_path)
        self._data = OMatrix(taskdef, run_id, use_existing, keep_data,
                             limit2branches, self._add_error)
        self._db_in = dbin  # input database handle
        self._db_out = dbout  # output database handle
        self._ids = None  # unit ids
        self._max_loop_count = None  # maximum number of loops
        self._best_sol = None  # best solution so far
        self._best_U = None  # best utility so far
        self._best_weights = None  # best solution weights
        self._curr_sol = None  # current solution
        self._curr_U = None  # current utility
        self._global_best_sol = None  # global best solution
        self._global_best_U = None  # global best utility
        self._track_solution = track_solution # store solution to db_in?
        self._run_id = run_id # run_id for storing solution to db_in
        self._create_opt_db = create_opt_db # write optimized database?
        self.do_split = False  # Split LP results?
        self._process_arguments(keywords)

    def set_data(self, iteration):
        """Set optimization data for current simulated iteration: collect
        values from data and operation result databases into data structure
        """
        if not self._data.construct_data(iteration):
            errorstr = "Could not construct optimization data matrix"
            self._add_error(errorstr)
            self._data = None
            return False
        # NB: TEMPORARY SOLUTION!
        self._max_loop_count = self._data.num_of_units * MAX_LOOP_FACTOR
        # create a container for storing optimal solutions
        n = self._data.num_of_units
        self._best_sol = numpy.zeros(n, dtype=int)
        self._curr_sol = numpy.zeros(n, dtype=int)
        self._global_best_sol = numpy.zeros(n, dtype=int)
        # assign initial utility values
        target = self._data.task_def.objective_func.target
        U_init = None
        self._best_U = U_init
        self._curr_U = U_init
        self._global_best_U = U_init
        return True

    def _analyze_data(self):
        """Analyze optimization input data before the actual data processing
        """
        if not self._data.analyze_data(self._db_in, self._add_info):
            return False
        return True

    def close(self):
        """Close the data file
        """
        if self._data is not None:
            self._data.close()

    def optimize(self):
        """Placeholder for algorithm -specific implementations
        """
        pass

    def _process_arguments(self, keywords):
        """Placeholder for processing algorithm -specific keyword arguments
        """
        pass

    def _add_error(self, msg):
        """Log error message
        """
        self._logger.log_message(self._log_name, 'ERROR', msg)
    
    def _add_debug(self, msg):
        """
        Log debug message
        """
        self._logger.log_message(self._log_name, 'DEBUG', msg)

    def _add_info(self, msg):
        """Log info message
        """
        self._logger.log_message(self._log_name, 'INFO', msg)

    def _generate_random_solution(self, num_of_solutions=10):
        """Choose a feasible random solution by randomly choosing a branch for
        each unit in the data.

        num_of_solutions -- number of initial solutions as int
        """
        start = time.time()
        
        if num_of_solutions is None:
            num_of_solutions = 10  # default to 10
        
        rand = self._data.get_random_solution
        feas = self._data.solution_feasibility
        util = self._data.solution_utility
        comp = self._compare_utilities
        if not self._data.ok:
            errorstr = "Cannot generate random solution: optimization data " \
                "was not constructed successfully"
            self._add_error(errorstr)
            return False
        U_best = None  # best utility so far
        U_new = None   # candidate utility
        S_best = None  # best solution so far
        S_new = None   # candidate solution
        msg = "  finding random initial solution"
        self._add_info(msg)
        # counter for calculating the number of
        # self._data.solution_feasibility method calls
        feas_tests = 0
        # initial solution should be the initial best solution, 'zero-solution'
        S_new = self._best_sol[:]
        for i in range(num_of_solutions):
            feasible = False
            count = 0
            while not feasible:
                count += 1
                feas_tests += 1
                if count > self._max_loop_count:
                    # if feasible solution could not be found for round i,
                    # break the while loop and move to next solution
                    # search round
                    break
                S_new = rand()
                feasible = feas(S_new, first_seek=True)
            if feasible:
                U_new = util(S_new)
                if U_best is None or S_best is None:
                    S_best = S_new
                    U_best = U_new
                if comp(U_best, U_new):
                    S_best = S_new
                    U_best = U_new

        # if not a single feasible solution was found, report an error message
        # to the user and return False
        if U_best is None or S_best is None:
            errorstr = 'Feasible random solution not found in %s ' \
                       'iterations' % str(self._max_loop_count)
            self._add_error(errorstr)
            self._log_constraint_feasibility_ratios(feas_tests)
            return False

        # set initial solution value for the current and elite solutions and
        # best solution of current simulated iteration (global best)
        self._curr_sol[:] = S_best
        self._curr_U = U_best
        self._best_sol[:] = S_best
        self._best_U = U_best
        # random initial solution found successfully, inform the user
        msg = "  feasible initial solution found in %0.1f secs" \
              % (time.time() - start)
        self._add_info(msg)
        return True

    def _log_constraint_feasibility_ratios(self, feas_tests):
        """
        Logs the feasible/tested ratio for each constraint in the optimization
        task
        """
        nfcounter = self._data._non_feasible_counter
        msg = 'Constraint feasibility ratios (% feasible solutions):'
        self._add_info(msg)
        for i in range(len(nfcounter)):
            ratio = (feas_tests - nfcounter[i]) / feas_tests * 100.
            msg = "   constraint %s: %.0f %%" % (i, abs(ratio))
            self._add_info(msg)

    def _lambda_opt_move(self, solution, mylambda):
        """Do a random lambda opt move, ie. change the treatment of lambda
        units randomly

        solution -- current solution as numpy array
        mylambda -- the number of units to be changed as int
        """
        rand = numpy.random.randint
        nu = self._data.num_of_units
        branches = self._data.branches
        used = set([])
        for i in range(mylambda):
            unit = rand(0, nu)
            B = len(branches[unit])
            counter = 0
            while B == 1 or unit in used:
                unit = rand(0, nu)
                B = len(branches[unit])
                counter += 1
                if counter > (nu * 10):
                    msg = "Could not make %d-opt move!" % mylambda
                    self._add_error(msg)
                    return False
            used.add(unit)
            b = rand(0, B)
            while b == solution[unit]:
                b = rand(0, B)
            solution[unit] = b
        return True

    def _store_solution(self, limit=None, do_split=False):
        """Store the optimal (elite) solution of a single iteration

        limit -- maximum unit number as int
        """
        U_new = self._best_U
        U_cur = self._global_best_U
        
        if U_new is None:
            msg = "  could not calculate utility as solution was not found"
            self._add_error(msg)
            return
        
        if U_new and self._compare_utilities(U_cur, U_new):
            self._global_best_U = U_new
            # NB: If limit is not None and we have split stands, the 
            # global best solution array must be resized.
            if do_split:
                self._global_best_sol = numpy.copy(self._best_sol)
            elif limit is not None:
                self._global_best_sol[:] = self._best_sol[:limit]
            else:
                self._global_best_sol[:] = self._best_sol[:]
                
        # get optimal solution status (total utility and subobjective values)
        # and output the status info to user
        U_global = self._global_best_U
        if U_global is None:
            U_global = 0.0
        S_best = self._best_sol if limit is None else self._best_sol[:limit]
        status = self._data.get_status(S_best)
                
        msg = "  utility: %0.3f\tbest so far: %0.3f" % (U_new, U_global)
        self._add_info(msg)
        
        for sub in status['subobjectives']:
            msg = "  %s: %0.3f" % sub
            self._add_info(msg)

    def _store_results(self, repeats, iteration, log_util=True, 
                       skip_weight_splits=None):
        """
        Store the optimum treatment schedules to result databases
        """
        if log_util:
            U_global = self._global_best_U
            S_global = self._global_best_sol
            status = self._data.get_status(S_global)
            if U_global is None:
                U_global = 0.0
            msg = "Best utility out of %d repeats: %0.3f" % (repeats, U_global)
            self._add_info(msg)
            for sub in status['subobjectives']:
                msg = "  %s: %0.3f" % sub
                self._add_info(msg)
        msg = 'Writing optimized treatment schedule to result database(s)...'
        self._add_info(msg)
        starttime = time.time()
        
        # If we have weights for each unit and we are splitting the main units
        # with the weights, store the weight together with the key.
        weights = None
        if hasattr(self, 'do_split') and hasattr(self, '_best_weights'):
            if self.do_split and self._best_weights is not None:
                weights = self._best_weights
        
        units = self._global_best_sol.size
        level = self._data.data_level
        keys = []
        origunits = units
        if hasattr(self, 'orig_nunits'):
            origunits = self.orig_nunits
        for j in range(units):
            uid = self._data.unit_ids[j]
            origuid = self._data.orig_uids[uid]
            branch = self._global_best_sol[j]
            # For the original units, do not restrict the copying of 
            # data to the optimal db. For the split units, copy only 
            # the operation data.
            weight = weights[j] if weights is not None else 1.
            if j < origunits:
                keys.append((iteration, uid, origuid, branch, weight, False))
            else:
                keys.append((iteration, uid, origuid, branch, weight, True))
        if self._track_solution:
            use_orig_uids = True if self.do_split else False
            self._db_in.store_solution(keys, self._run_id, use_orig_uids)
        if self._create_opt_db:
            self._db_in.copy_to_db(keys, level, self._db_out, iteration, 
                                   skip_weight_splits=skip_weight_splits)
        tottime = time.time()-starttime
        msg = 'Writing completed in %s min' % str(tottime/60)
        self._add_info(msg)

    def _run_algorithm(self, runf, repeats, initial_solutions):
        """Run an optimization algorithm

        runf -- algorithm specific runner method
        repeats -- number of repeats as int
        initial_solutions -- number of initial random solutions as int
        """
        # first analyze the optimization input databases
        if not self._analyze_data():
            errorstr = 'Invalid input database, optimization aborted!'
            self._add_error(errorstr)
            return

        rands = self._generate_random_solution
        
        # optimization is done for a single iteration at a time
        # NB: here 'iteration' refers to simulation iteration (as in Monte
        # Carlo type of simulation) and 'optimization_round' or 'round' to
        # a single round in iterative heuristic optimization
        self.any_feasible = False
        for iteration in range(self._num_of_iterations):
            # first, set construct dataset for the current iteration
            if not self.set_data(iteration):
                msg = "Could not construct data set - skipping iteration %d" \
                      % iteration
                continue
            # start optimization logger
            target = self._data.task_def.objective_func.target
            self._stat_logger.start(iteration, self._branches, target)
            any_feasible = False
            for repeat in range(repeats):
                # Find initial solution for the problem
                if rands(initial_solutions):
                    # run the algorithm for the single repeat
                    runf()
                    # update stat log and store the best solution of the 
                    # current repeat
                    self._stat_logger.add_repeat()
                    self._store_solution()
                    any_feasible = True
            
            # Stop the logger
            self._stat_logger.stop(iteration)
            # Store the optimum solution(s) to databases, if a feasible 
            # solution was found
            if any_feasible:
                self._store_results(repeats, iteration)
                self.any_feasible = True
            else:
                msg = "Feasible solution not found!"
                self._add_error(msg)


    ##########################################################################
    ## WRAPPERS FOR OMATRIX METHODS
    def _solution_feasibility(self, solution):
        """Wrapper for OMatrix.solution_feasibility
        """
        return self._data.solution_feasibility(solution)

    def _solution_utility(self, solution):
        """Wrapper for OMatrix.solution_utility
        """
        return self._data.solution_utility(solution)

    def _compare_utilities(self, best, candidate):
        """Wrapper for OMatrix.solution_utility
        """
        return self._data.compare_utilities(best, candidate)

    def _random_solution(self):
        """Wrapper for OMatrix.get_random_solution
        """
        return self._data.get_random_solution()


    ##########################################################################
    ## WRAPPERS FOR OMATRIX ATTRIBUTES
    def _num_of_units():
        def fget(self):
            return self._data.num_of_units
        return locals()
    _num_of_units = property(**_num_of_units())

    def _branches():
        def fget(self):
            return self._data.branches
        return locals()
    _branches = property(**_branches())

    def _num_of_iterations():
        def fget(self):
            return self._data.max_iterations
        return locals()
    _num_of_iterations = property(**_num_of_iterations())



def main():
    import doctest
    doctest.testfile('./test/optimizer-py.rst')

if __name__ == '__main__':
    main()
