#!/usr/bin/env python
# encoding: utf-8
"""
Title: SIMO metaheuristic simulated annealing (SA) optimization algorithm

...

*simulated_annealing.py*

Copyright (c) 2007 Simosol Oy. Distributed under the GNU General Public License 2.0.
"""

# these are for running the doctest from src/simo
#import sys
#sys.path.append('..')

import copy
import numpy
import math
from optimizer import Optimizer


ARGUMENTS = ('repeats',
             'initial_solutions',
             'maximum_iterations',
             'starting_temperature',
             'freezing_temperature',
             'cooling_factor',
             'iterations_per_temperature',
             'iteration_multiplier')


class SimulatedAnnealing(Optimizer):
    """Simulated annealing optimization algorithm implementation
    """
    def _process_arguments(self, args):
        """Process keyword arguments
        """
        self._abort = False
        for arg in ARGUMENTS:
            if not arg in args:
                self._add_error('Parameter "%s" missing!' % arg)
                self._abort = True
        if self._abort:
            return
        try:
            self._repeats = int(args["repeats"])
            self._initial_solutions = int(args["initial_solutions"])
            self._maximum_iterations = int(args["maximum_iterations"])
            self._starting_temperature = float(args["starting_temperature"])
            self._freezing_temperature = float(args["freezing_temperature"])
            self._cooling_factor = float(args["cooling_factor"])
            self._iterations_per_temperature = int(args["iterations_per_temperature"])
            self._iteration_multiplier = float(args["iteration_multiplier"])
        except e:
            self._add_error('Invalid parameter values!')
            self._abort = True
    
    def _run_SA(self):
        """
        Run SA algorithm for a single iteration/repeat
        """
        move = self._lambda_opt_move
        util = self._solution_utility
        feas = self._solution_feasibility
        comp = self._compare_utilities
        addr = self._stat_logger.add_round
        copy = numpy.copy
        exp = math.exp
        rand = numpy.random.random
        
        target = self._data.task_def.objective_func.target
        maximize = True if target == 'max' else False
        
        T_curr = self._starting_temperature
        T_freeze = self._freezing_temperature
        cf = self._cooling_factor
        ipt = self._iterations_per_temperature
        im = self._iteration_multiplier
        U_curr = self._curr_U
        U_best = self._best_U
        U_new = None
        S_curr = self._curr_sol
        S_best = self._best_sol
        S_new = None  # candidate solution
        N_curr = 0  # number of iterations in current temperature
        for i in range(self._maximum_iterations):
            # get new candidate solution by making a 1-opt move
            S_new = copy(self._curr_sol) 
            move(S_new, 1)
            U_new = util(S_new)
            if not feas(S_new):
                continue
            
            # the calculation of probability p depends on the objective target
            if maximize:
                p = exp((U_new - U_curr) / T_curr)
            else:
                p = exp(-(U_new - U_curr) / T_curr)
            
            # approve the new solution if either the utility is higher than
            # current utility or [0,1] uniform random number is less than p
            if comp(U_curr, U_new) or rand() < p:
                U_curr = self._curr_U = U_new
                S_curr = S_new
                if comp(U_best, U_curr):
                    U_best = self._best_U = U_curr
                    S_best[:] = S_curr[:]
            
            # log the current iteration and update current temperature
            # iteration counter
            addr(S_curr, U_curr, U_best)
            N_curr += 1
            # check if the temperature should be dropped
            if N_curr > ipt:
                N_curr = 0
                ipt = math.floor(ipt * im)
                T_curr = cf * T_curr
            # check the stopping condition
            if T_curr <= T_freeze:
                break

    def optimize(self):
        """Run SA algorithm
        """
        runf = self._run_SA
        repeats = self._repeats
        initial_solutions = self._initial_solutions
        self._run_algorithm(runf, repeats, initial_solutions)


def main():
    import doctest
    doctest.testfile('./test/simulated_annealing-py.rst')

if __name__ == '__main__':
    main()
