#!/usr/bin/env python
# encoding: utf-8
"""
Title: SIMO metaheuristic TABUSEARCH algorithm

...

*tabusearch.py*

Copyright (c) 2007 Simosol Oy. Distributed under the GNU General Public License 2.0.
"""
# these are for running the doctest from src/simo
#import sys
#sys.path.append('..')

import numpy
import random
from optimizer import Optimizer


ARGUMENTS = ('repeats',
             'initial_solutions',
             'maximum_iterations',
             'entrytenure',
             'exittenure')


class TabuSearch(Optimizer):
    """TabuSearch heuristic optimization algorithm implementation. Initialize
    with keyword arguments eg::

        repeats=10
        initial_solutions=100
        maximum_iterations=1000
        entrytenure=10
        exittenure=20
    """
    def _process_arguments(self, args):
        """Process keyword arguments

        args -- keyword arguments in a dict
        """
        self._abort = False
        for arg in ARGUMENTS:
            if not arg in args:
                self._add_error('Parameter "%s" missing!' % arg)
                self._abort = True
        if self._abort:
            return
        try:
            self._repeats = int(args['repeats'])
            self._initial_solutions = int(args['initial_solutions'])
            self._max_iterations = int(args['maximum_iterations'])
            self._entry_tenure = int(args['entrytenure'])
            self._exit_tenure = int(args['exittenure'])
        except:
            self._add_error('Invalid parameter values!')
            self._abort = True

    def _get_new_solution(self):
        """
        Identify current solution neighbourhood and choose the best solution
        from the neighbourhood. New solution can be inferior to the elite
        (global best) solution. Neighbourhood consists of all possible
        treatment schedules for a single stand.

        iteration -- simulated iteration indices as int
        """
        # get a random unit and it's feasible neighbourhood
        unit, neighbourhood = self._get_random_unit()
        if unit is None or neighbourhood is None:
            # Feasible solution cannot be found with current constraints
            # and/or tenures
            return False
        # find move exit and entry points (b_from and b_to) and update the
        # tabu set of the curren unit
        b_from = self._curr_sol[unit]
        b_to, U = self._choose_best_solution(unit, neighbourhood)
        self._update_tabu_set(unit, b_from, b_to)
        return True

    def _get_random_unit(self):
        """Get randomly a unit (stand) from the data. Stand needs to have more
        than 1 treatment schedules (branches) in order to get picked. Also,
        stand's treatment schedule in the current solution (exit point) must
        not be taboo.
        """
        feas = self._solution_feasibility
        rand = random.randint
        getn = self._get_neighbourhood
        canex = self._can_exit
        nany = numpy.any
        nu = self._num_of_units - 1
        branches = self._branches
        feasible, count, freezecount = False, 0, 0

        # get random solutions until conditions are met or solution freezes
        while not feasible:
            # draw random unit and check that it can exit from current
            # solution (exit map at (unit,from) should be False)
            unit = rand(0, nu)
            b_from = self._curr_sol[unit]
            if unit in branches:
                if canex and len(branches[unit]) > 1:
                    candidates = getn(unit, b_from)
                    for i in branches[unit]:
                        if not candidates[i]:
                            continue
                        self._curr_sol[unit] = i
                        if not feas(self._curr_sol):
                            candidates[i] = False
                    if nany(candidates):
                        feasible = True
            # store the original branch to the current solution
            self._curr_sol[unit] = b_from
            count += 1
            if freezecount > self._max_loop_count:
                # If optimizer freezes without finding feasible solution for
                # too long time, abort optimisation task
                errorstr = 'Computational load for finding feasible '\
                           'solution is too heavy! Decrease tenures!'
                self._add_error(errorstr)
                self._abort = True
                return None, None
            if count > self._max_loop_count:
                # In some cases the search can get stuck in a solution where it
                # is difficult to find a feasible solution by changing only
                # single unit solution. If so, initialize random solutions for
                # all units.
                freezecount += 1
                innercount = 0
                while True:
                    new_solution = numpy.zeros(len(branches), dtype=int)
                    for unit_ind, ub in branches.iteritems():
                        pick = random.randint(0, len(ub) - 1)
                        try:
                            new_solution[unit_ind] = branches[unit_ind][pick]
                        except:
                            ##DEBUG START
                            msg = 'length of branches = %d\n' % len(branches)
                            msg = 'new solution size = %d\n' % new_solution.size
                            msg += 'picked indice = %d\n' % pick
                            msg += 'unit indice = %d\n' % unit_ind
                            if unit_ind in branches:
                                msg += 'branches[unit_ind] = %s' % str(branches[unit_ind])
                            self._add_debug(msg)
                            self._abort = True
                            return None, None
                            ##DEBUG END
                    if feas(new_solution):
                        break
                    innercount += 1
                    if innercount > self._max_loop_count:
                        errorstr = 'Computational load for finding '\
                                   'feasible solution is too heavy! '\
                                   'Decrease tenures!'
                        self._add_error(errorstr)
                        self._abort = True
                        return None, None
                self._curr_sol = new_solution
        neighbourhood = numpy.where(candidates)[0]
        return unit, neighbourhood

    def _choose_best_solution(self, unit, neighbourhood):
        """
        Choose the best solution from the current neighbourhood

        unit -- unit indice as int
        neighbourhood -- current unit neighbourhood
        """
        util = self._solution_utility
        tabu = self._is_move_tabu
        comp = self._compare_utilities

        S_curr = self._curr_sol
        # make sure that initial best solution is not tabu
        b_from = S_curr[unit]
        for b_to in neighbourhood:
            if not tabu(unit, b_from, b_to):
                break
        # Set first neighbour as the current best solution
        S_curr[unit] = b_best = b_to
        U_best = util(S_curr)
        neighbourcount = neighbourhood.size  # how many neighbours?
        tabucount = 0  # how many taboo neighbours?
        mintenure = self._exit_tenure  # neighbour with smallest tenure

        # Go through all neighbours and update best solution if current
        # neighbour is the best so far.
        # If neighbour is taboo but better than the global
        # best, it can be considered as a feasible solution
        for b_to in neighbourhood[1:]:
            # set the current solution and get current solution value
            S_curr[unit] = b_to
            U_curr = util(S_curr)
            if tabu(unit, b_from, b_to):
                tabucount += 1  # increase the number of taboo neighbours
                if self._entry_timer[unit,b_to] < mintenure:
                    mintenure = self._entry_timer[unit,b_to]
                    mintenureneighbour = b_to
                # new move is taboo, check if aspiration criteria fulfills
                if not comp(self._best_U, U_curr):
                    # if currvalue is better than global best, allow tabu move
                    # otherwise continue searching best move in neighbourhood
                    continue
            if comp(U_best, U_curr):
                # update current best solution and value
                U_best = U_curr
                b_best = b_to

        # If all neighbours are taboo, choose the one with the smallest tenure
        if tabucount >= neighbourcount:
            b_best = mintenureneighbour
            S_curr[unit] = b_best
            U_best = util(S_curr)
        # set the best solution of the iteration as the current solution
        self._curr_sol[unit] = b_best
        self._curr_U = U_best
        # Check if current best solution is better than current global best
        # and update global best solution if true
        if comp(self._best_U, U_best):
            self._best_sol[:] = self._curr_sol[:]
            self._best_U = self._curr_U
        return b_best, U_best

    def _update_tabu_set(self, unit, b_from, b_to):
        """Add a new taboo move to the set and decrease the tenure counter for
        existing taboos

        unit -- unit indice as int
        b_from -- from branch indice as int
        b_to -- to branch indice as int
        """
        # decrease tenure timers for currently taboo moves
        self._exit_timer[self._exit_map] -= 1
        self._entry_timer[self._entry_map] -= 1
        # set all objects with tenure timer == 0 as non-taboo
        self._exit_map[self._exit_timer <= 0] = False
        self._entry_map[self._entry_timer <= 0] = False
        # set both leaving and entering branches as taboo and update the
        # tenure counters
        self._exit_map[unit,b_from] = True
        self._entry_map[unit,b_to] = True
        self._exit_timer[unit,b_from] = self._exit_tenure
        self._entry_timer[unit,b_to] = self._entry_tenure

    def _is_move_tabu(self, unit, b_from, b_to):
        """
        True if move is tabu, False if not

        unit -- unit indice as int
        b_from -- from branch indice as int
        b_to -- to branch indice as int
        """
        return self._exit_map[unit,b_from] or self._entry_map[unit,b_to]

    def _can_exit(self, unit, b_from):
        """
        True if exit map value at (unit, b_from) is False

        unit -- unit indice as int
        b_from -- from branch indice as int
        """
        return self._exit_map[unit,b_from] == False

    def _get_neighbourhood(self, unit, b_from):
        """
        Get non-taboo neighbours of (unit, from)

        unit -- unit indice as int
        b_from -- from branch indice as int
        num_b -- number of branches as int
        """
        n = self._entry_map[unit,:] == False
        n[b_from] = False
        max_plus_b = self._branches[unit][-1] + 1
        n[max_plus_b:] = False
        return n

    def _run_TS(self):
        """
        Run TS algorithm for a single iteration/repeat

        iteration -- current simulated iteration as int
        """
        nu = self._num_of_units
        shp = (nu, self._data._max_branches)
        self._entry_map = numpy.zeros(shp, dtype=bool)
        self._exit_map = numpy.zeros(shp, dtype=bool)
        self._entry_timer = numpy.zeros(shp, dtype=int)
        self._exit_timer = numpy.zeros(shp, dtype=int)
        i = 0
        # run tabu search for a given number of iterations
        while i < self._max_iterations and not self._abort:
            if not self._get_new_solution():
                break  # optimum solution cannot be found
            self._stat_logger.add_round(self._curr_sol, self._curr_U,
                                        self._best_U)
            i += 1

    def optimize(self):
        """Optimize the current solution space with tabu search algorithm
        """
        runf = self._run_TS
        repeats = self._repeats
        initial_solutions = self._initial_solutions
        self._run_algorithm(runf, repeats, initial_solutions)


def main():
    import doctest
    doctest.testfile('./test/tabusearch-py.rst')

if __name__ == '__main__':
    main()
