#!/usr/bin/env python
# encoding: utf-8

"""
Mock objects for testing optimization algorithms

*mocks4optimizer.py*

Copyright (c) 2007 Simosol Oy. Distributed under the GNU General Public License
2.0.
"""

import numpy
import datetime
from copy import deepcopy as dc
from minimock import Mock


# mock logger
logger = Mock('Logger')
logname = 'optimization-test'

# mock optimization stat logger (OLogger)
ologger = Mock('OLogger')

# mock optimization task definition
taskdef = Mock('TaskDef')
taskdef.discount_rate = 3.0
taskdef.conditional_discounting = None
taskdef.num_of_subobjs = 2
taskdef.num_of_consts = 3
taskdef.period_step = 'year'
so1 = Mock('Subobjective')
so2 = Mock('Subobjective')
so1.weight = 0.3
so2.weight = 0.7
taskdef.subobjectives = [so1, so2]
taskdef.objective_func.type = 'conjunctive'
taskdef.objective_func.target = 'max'
taskdef.subobj_strings = ['1.0 * (sum[1:5](operation:cashflow:discount))',
                          '1.0 * (sum[6:10](operation:cashflow:discount))']
taskdef.constr_strings = ['sum[10:10](comp_unit:V) - '\
                          'sum[1:1](comp_unit:V) gt 0.0',
                          'sum[20:-1](operation:Volume) - '\
                          'sum[10:-1](operation:Volume) gt 0.0',
                          'sum[20:20](comp_unit:V * comp_unit:AREA) - '\
                          'sum[10:10](comp_unit:V * comp_unit:AREA) gt 0.0']
# the below constraints are SIMO postfix form of expressions:
# sum[10:10](comp_unit:V) - sum[1:1](comp_unit:V) gt 0.0
# sum[20:-1](operation:Volume) - sum[10:-1](operation:Volume) gt 0.0
# sum[20:20](comp_unit:V * comp_unit:AREA) - sum[10:10](comp_unit:V *
# comp_unit:AREA) gt 0.0
taskdef.constraints = [Mock('Constraint'),
                       Mock('Constraint'),
                       Mock('Constraint'),]
taskdef.constraints[0].expr = \
    [('data', (u'comp_unit', u'V'),
      (datetime.date(2019, 1, 1), datetime.date(2019, 12, 31)), None, False,
      (2,2)),
     ('aggr', '<function nansum at ...>'),
     ('data', (u'comp_unit', u'V'),
      (datetime.date(2009, 1, 1), datetime.date(2009, 12, 31)), None, False,
      (3,3)),
     ('aggr', '<function nansum at ...>'),
     ('ari', '<function minus at ...>', '-'),
     ('value', 0.0),
     ('eq', '<function gte at ...>', 'gt')]
taskdef.constraints[1].expr = \
    [('op', u'Volume', (datetime.date(2019, 1, 1), None), None, False,
    (4,4)),
     ('aggr', '<function nansum at ...>'),
     ('op', u'Volume', (datetime.date(2009, 1, 1), None), None, False,
     (5,5)),
     ('aggr', '<function nansum at ...>'),
     ('ari', '<function minus at ...>', '-'),
     ('value', 0.0),
     ('eq', '<function gte at ...>', 'gt')]
taskdef.constraints[2].expr = \
    [('data', (u'comp_unit', u'V'),
      (datetime.date(2019, 1, 1), datetime.date(2019, 12, 31)), None, False,
      (6,6)),
     ('data', (u'comp_unit', u'AREA'),
      (datetime.date(2019, 1, 1), datetime.date(2019, 12, 31)), None, False,
      (7,7)),
     ('ari', '<function multi at ...>', '*'),
     ('aggr', '<function nansum at ...>'),
     ('data', (u'comp_unit', u'V'),
      (datetime.date(2009, 1, 1), datetime.date(2009, 12, 31)), None, False,
      (8,8)),
     ('data', (u'comp_unit', u'AREA'),
      (datetime.date(2009, 1, 1), datetime.date(2009, 12, 31)), None, False,
      (9,9)),
     ('ari', '<function multi at ...>', '*'),
     ('aggr', '<function nansum at ...>'),
     ('ari', '<function minus at ...>', '-'),
     ('value', 0.0),
     ('eq', '<function gte at ...>', 'gt')]
    
taskdef.init_date = datetime.date(2000, 1, 1)

taskdef.operand_index = {0: 0, 1: 1, 2: 2, 3: 3, 4: 4, 5: 5, 6: 6, 7: 7, 8: 8,
                         9: 9}

invalid_taskdef = Mock('TaskDef')
invalid_taskdef.discount_rate = 3.0
invalid_taskdef.conditional_discounting = None
invalid_taskdef.num_of_consts = 2
invalid_taskdef.constraints = [Mock('Constraint'),
                               Mock('Constraint'),]
invalid_taskdef.constraints[0].expr = \
    [('data', ('comp_unit', 'V'), (datetime.date(2010, 1, 1), None), None,
    False),
    ('data', ('comp_unit', 'V'), (datetime.date(2000, 1, 1), None), None,
    False),
    ('ari', '<function divide at ...>', '/'),
    ('aggr',' <function nansum at ...>'),
    ('value', 1.05),
    ('eq', '<function gte at ...>', 'gt')]
invalid_taskdef.constraints[1].expr = \
    [('op', 'Volume', (datetime.date(2020, 1, 1), None), None, False),
    ('aggr',' <function nansum at ...>'),
    ('op', 'Volume', (datetime.date(2010, 1, 1), None), None, False),
    ('aggr',' <function nansum at ...>'),
    ('eq', '<function gte at ...>', 'gt')]


# another task definition for testing JLP goal programming
taskdef2 = Mock('TaskDef')
taskdef2.discount_rate = 3.0
taskdef2.conditional_discounting = None
taskdef2.num_of_subobjs = 4
taskdef2.num_of_consts = 0
taskdef2.period_step = 'year'
so3 = Mock('Subobjective')
so4 = Mock('Subobjective')
so5 = Mock('Subobjective')
so6 = Mock('Subobjective')
so3.weight = 0.25
so4.weight = 0.25
so5.weight = 0.25
so6.weight = 0.25
so3.optlevel = 1000.
so4.optlevel = 2000.
so5.optlevel = 3000.
so6.optlevel = 4000.
taskdef2.subobjectives = [so3, so4, so5, so6]
taskdef2.objective_func.type = 'distance'
taskdef2.objective_func.target = 'min'
taskdef2.subobj_strings = ['0.25 * (sum[1:5](operation:cashflow))',
                           '0.25 * (sum[6:10](operation:cashflow))',
                           '0.25 * (sum[11:15](operation:cashflow))',
                           '0.25 * (sum[16:20](operation:cashflow))']
taskdef2.constr_strings = None
taskdef2.constraints = None
taskdef2.init_date = datetime.date(2000, 1, 1)
taskdef2.operand_index = {0: 0, 1: 1, 2: 2, 3: 3, 4: 4}

# mock databases
simdbin = Mock('SimInputDB')
simdbout = Mock('SimOutputDB')

# alternative mock optimization data matrices/handlers
omatrix1 = Mock('OMatrix')
omatrix1.max_iterations = 1
omatrix1.ok = False
omatrix1.solution_feasibility.mock_returns = False
omatrix1.analyze_data.mock_returns = False
omatrix1.construct_data.mock_returns = False
omatrix1.get_status.mock_returns = {
    'utility': 0.7654321,
    'subobjectives': [('1. subobjective', 12345.6789),
                      ('2. subobjective', 234.56789)]}
omatrix1.orig_uids = {'UNIT1': 'UNIT1',
                      'UNIT2': 'UNIT2',
                      'UNIT3': 'UNIT3',
                      'UNIT4': 'UNIT4',
                      'UNIT5': 'UNIT5'}

omatrix2 = Mock('OMatrix')
omatrix2.task_def = taskdef
omatrix2.data_level = 'comp_unit'
omatrix2.max_iterations = 2
omatrix2.num_of_units = 5
omatrix2.unit_ids = ['UNIT1', 'UNIT2', 'UNIT3', 'UNIT4', 'UNIT5']
omatrix2.orig_uids = {'UNIT1': 'UNIT1',
                      'UNIT2': 'UNIT2',
                      'UNIT3': 'UNIT3',
                      'UNIT4': 'UNIT4',
                      'UNIT5': 'UNIT5'}
omatrix2._max_branches = 6
omatrix2.branches = {0: [0, 1], 1: [0, 2, 3, 4, 5], 2:[0, 1, 2],
                    3:[0, 1, 2, 3], 4:[0]}
omatrix2.ok = True
omatrix2.solution_feasibility.mock_returns = True
omatrix2.solution_utility.mock_returns = 1.0
omatrix2.compare_utilities.mock_returns = True
omatrix2.get_random_solution.mock_returns = numpy.zeros(5, dtype=int)
omatrix2.analyze_data.mock_returns = True
omatrix2.construct_data.mock_returns = True
omatrix2._evaluator.evaluate.mock_returns = 1.2345
omatrix2.get_status.mock_returns = {
    'utility': 0.7654321,
    'subobjectives': [('1. subobjective', 12345.6789),
                      ('2. subobjective', 234.56789)]}

omatrix2.data = {}
omatrix2.data['operands'] = {0: numpy.ones((5, 6, 2), dtype=float) * 1.234,
                             1: numpy.ones((5, 6, 2), dtype=float) * 1.234,
                             2: numpy.ones((5, 6, 2), dtype=float) * 2.345,
                             3: numpy.ones((5, 6, 2), dtype=float) * 2.345,
                             4: numpy.ones((5, 6, 2), dtype=float) * 2.345,
                             5: numpy.ones((5, 6, 2), dtype=float) * 2.345,
                             6: numpy.ones((5, 6, 2), dtype=float) * 2.345,
                             7: numpy.ones((5, 6, 2), dtype=float) * 2.345,
                             8: numpy.ones((5, 6, 2), dtype=float) * 2.345,
                             9: numpy.ones((5, 6, 2), dtype=float) * 2.345}
omatrix2.data['date_locs'] = dict([(i, (0, 1)) for i in range(10)])

