#!/usr/bin/env python
# encoding: utf-8

"""
SIMO optimization logger

...

*optlogger.py*

Copyright (c) 2007 Simosol Oy. Distributed under the GNU General Public License
2.0.
"""

__revision__ = "$Revision: Alpha 0.1 $"

import os
import sys
import time
import copy
import numpy

from pygooglechart import Chart
from pygooglechart import SimpleLineChart
from pygooglechart import Axis


CALMER = 1000


class OLogger:
    """class for logging the performance and statistics of SIMO optimization
    algorithms
    """
    def __init__(self, logger, logname, chart_path):
        """Initialize optimization logger

        logger -- SIMO logger instance
        logname -- target log name as string
        chart_path -- path to chart output target folder
        """
        self._logger = logger
        self._log_name = logname
        self._chart_path = chart_path
        self._num_of_rounds = None
        self._start_time = None
        self._branches = None
        self._num_of_units = None
        self._curr_values = None
        self._best_values = None
        self._best_value = None
        self._visit_index = None
        self._visit_list = None

    def start(self, iteration, branches, target):
        """Start logging of the optimization process

        iteration -- current simulated iteration
        branches -- branch ids for each unit
        target -- optimization target, ie. "min" or "max"
        """
        self._repeats = 1
        self._num_of_rounds = 0
        self._start_time = time.time()
        self._branches = branches
        self._num_of_units = len(branches)
        self._target = target
        self._global_curr_values = []
        self._global_best_values = []
        self._curr_values = []  # current utility
        self._best_values = []  # best utility
        self._best_value = 0.0 if self._target == "max" else 1.0
        branchcount = 0
        for unit_ind, branches in branches.iteritems():
            branchcount += len(branches)
        branchmean = branchcount
        if self._num_of_units > 0:
            branchmean = float(branchcount) / self._num_of_units
        solspace = '%3.2f^%s'% (branchmean, self._num_of_units)
        msg = 'Optimization started for simulated iteration %d' % iteration
        self._logger.log_message(self._log_name, 'INFO', msg)
        msg = '  total number of units: %s' % self._num_of_units
        self._logger.log_message(self._log_name, 'INFO', msg)
        msg = '  total number of treatment schedules: %s' % branchcount
        self._logger.log_message(self._log_name, 'INFO', msg)
        msg = '  average number of treatment schedules per unit: %3.2f' \
            % branchmean
        self._logger.log_message(self._log_name, 'INFO', msg)
        msg = '  total solution space size: %s' % solspace
        self._logger.log_message(self._log_name, 'INFO', msg)

    def stop(self, iteration):
        """Stop logging of the optimization process

        iteration -- current simulated iteration
        """
        stoptime = time.time()
        timedelta = stoptime - self._start_time
        msg = 'Optimization process performed successfully in %s min' \
            % (str(timedelta / 60))
        self._logger.log_message(self._log_name, 'INFO', msg)
        # NB: Disable the optimization log drawing as the result is not very
        # usable at the moment.
        #self._save_stats(iteration)

    def add_round(self, currsol, currval, bestval):
        """Add current iterative round parameters to the logger

        currsol -- current solution
        currcal -- current utility value
        bestval -- best utility value so far
        """
        self._num_of_rounds += 1
        if currval is None:
            currval = 0.0
            if self._curr_values != []:
                currval = self._curr_values[-1]
        self._curr_values.append(currval)
        self._best_values.append(bestval)
        if (self._num_of_rounds % CALMER == 0):
            msg = 'Number of iterative rounds complete: %s' % \
                (self._num_of_rounds)
            self._logger.log_message(self._log_name, 'INFO', msg)

    def add_repeat(self):
        """
        Add a single iteration/repeat of the whole optimization process
        """
        msg = 'Optimization process repeated %d time(s)' % self._repeats
        self._logger.log_message(self._log_name, 'INFO', msg)
        if len(self._best_values) > 0:
            bestv = self._best_values[-1]
        else:
            bestv = 0.0 if self._target == "max" else 1.0
        if self._target == "max":
            better = bestv > self._best_value
        else:
            better = bestv < self._best_value
        if better or self._global_best_values == []:
            self._global_best_values = copy.deepcopy(self._best_values)
            self._global_curr_values = copy.deepcopy(self._curr_values)
            self._best_value = bestv
        self._repeats += 1
        self._num_of_rounds = 0
        self._curr_values = []
        self._best_values = []

    def _save_stats(self, iteration):
        """Save the statistics of the current optimization process

        iteration -- current simulated iteration
        """
        C = self._global_curr_values
        B = self._global_best_values
        comb = C + B
        if len(comb) == 0:
            msg = "no utilities found - cannot save optimization log"
            self._logger.log_message(self._log_name, 'INFO', msg)
            return
        N = len(C)
        msg = 'Saving optimization log'
        self._logger.log_message(self._log_name, 'INFO', msg)
        # get the y axis limits
        if max(C + B) <= 1:
            y_max = 1
        else:
            y_max = max(C + B)
        #NB! 400 seems to be the max or the length of the url goes too long
        sets = range(0, N, 400)
        for i in range(len(sets)):
            ind_min = i*400
            ind_max = min((i+1)*400, N)
            # Chart size of 500x250 pixels and specifying the range for the Y axis
            chart = SimpleLineChart(500, 250, y_range=(0, y_max))
            # set the legend
            chart.set_legend(['current value', 'best value'])
            # add current and best values to the chart
            chart.add_data(C[ind_min:ind_max])
            chart.add_data(B[ind_min:ind_max])
            # Set the line colour to red and blue
            chart.set_colours(['FF0000','0000FF'])
            # Set the horizontal dotted lines
            chart.set_grid(0, 25, 5, 5)
            # X axis labels
            chart.set_axis_labels(Axis.LEFT, (0, y_max))
            chart.set_axis_labels(Axis.BOTTOM, (ind_min, ind_max))
            try:
                # try to create the graph
                # NB: pygooglechart needs internet connection!
                chart_name = 'utility-iter-%d-set-%d-%d.png' % (iteration,
                                                                ind_min,
                                                                ind_max)
                chart_loc = os.path.join(self._chart_path, chart_name)
                #msg = 'Saving utility value chart to %s' % chart_loc
                #self._logger.log_message(self._log_name, 'INFO', msg)
                chart.download(chart_loc)
            except Exception:
                pass


def main():
    pass

if __name__ == '__main__':
    main()
