#!/usr/bin/env python
# encoding: utf-8
"""
SIMO optimization data matrix

...

*optdata.py*

Copyright (c) 2007 Simosol Oy. Distributed under the GNU General Public License 2.0.
"""

# these are for running the doctest from src/simo
import sys
sys.path.append('..')

import os
import math
import numpy
import datetime
import time
import pickle
import random
from collections import defaultdict

from simo.utils.utils import secs2str
from simo.utils.utils import calculate_NPV
from simo.optimization.opt_ext import _value2utility
from simo.optimization.tools.postfixeval import PostfixEvaluator
from simo.builder.names import EXPR_DISCOUNT, EXPR_DTYPE_COMB,\
                               EXPR_DTYPE_DATA, EXPR_DTYPE_OP,\
                               EXPR_DTYPE_VALUE, EXPR_OPERATION,\
                               EXPR_DTYPE_FUNC, EXPR_FUNCTION,\
                               EXPR_DTYPE_LEVEL


isnan = numpy.isnan
MAX_STR_SIZE = 16  # maximum length for object ids
DTYPE_FLOAT = numpy.float32


def convert_bytes(bytes):
    bytes = float(bytes)
    if bytes >= 1099511627776:
        terabytes = bytes / 1099511627776
        size = '%.2fT' % terabytes
    elif bytes >= 1073741824:
        gigabytes = bytes / 1073741824
        size = '%.2fG' % gigabytes
    elif bytes >= 1048576:
        megabytes = bytes / 1048576
        size = '%.2fM' % megabytes
    elif bytes >= 1024:
        kilobytes = bytes / 1024
        size = '%.2fK' % kilobytes
    else:
        size = '%.2fb' % bytes
    return size


class OMatrix:
    """Class for in-memory storage of SIMO optimization data structures. Get
    the wanted data from input database, stores it into multidimensional numpy
    arrays for fast access during optimization process.
    """
    def __init__(self, taskdef, run_id, use_existing, keep_data,
                 limit2branches, add_error):
        """Initialize OMatrix
        """
        self._add_error = add_error  # error message logging function
        self.task_def = taskdef  # optimization task definition
        self.limit2branches = limit2branches # only use a subset of branches
        self._data_file = run_id + '.dat'
        self._use_existing = use_existing
        self._keep_data = keep_data
        self._non_feasible_counter = numpy.zeros(taskdef.num_of_consts)
        self._evaluator = PostfixEvaluator()  # RPN evaluator
        self._db = None
        self.data = None  # data file
        self.unit_ids = None  # unit ids
        self.num_of_units = 0  # number of units
        self.unit_index = None  # mapping from unit id to matrix index
        self.orig_uids = None
        self.branches = None # branch inds by unit
        self.max_iterations = 0  # maximum number of iterations
        self._max_branches = 0  # maximum number of branches
        self._max_steps = 0  # maximum number of timesteps
        self._earliest_date = None
        self._latest_date = None
        self._date_index = None  # mapping from date to array index
        self.data_level = self.task_def.main_level
        self._global_max = None  # global maximum subobjective values
        self.ok = False


    ##########################################################################
    # METHODS FOR CONSTRUCTING THE OPTIMIZATION DATA
    ##########################################################################

    def construct_data(self, iteration):
        """
        Construct optimization data tables and fill them with data from
        input databases

        iteration -- current simulated iteration indice as int
        """
        if self._db is None:
            self._add_error('No database to use')
            return False
        msg = "Constructing optimization data set for iteration %d" \
              % iteration
        self._add_info(msg)
        self.ok = False
        self._add_info('  creating optimisation tables')
        if not self._create_tables():
            msg = '  failed to create optimisation tables'
            self._add_error(msg)
            return False
        if not self._use_existing:
            self._db.prepare_db(self)
            if not self._fill_tables(iteration):
                return False
        self._add_info('  data set ready for optimization')
        self.ok = True
        return True

    def close(self):
        """Save the data arrays if keep_data=True
        """
        if self._keep_data and not self._use_existing:
            msg = 'Writing optimization data structure to file...'
            self._add_info(msg)
            try:
                fp = open(self._data_file, 'wb')
                pickle.dump(self.data, fp, pickle.HIGHEST_PROTOCOL)
                fp.close()
            except:
                pass
        elif not self._keep_data:
            try:
                os.remove(self._data_file)
            except:
                pass

    def analyze_data(self, db, add_info):
        """Go through subobjectives, constraints and input database and
        analyze some data properties
        """
        self._db = db
        self._add_info = add_info
        self._add_info('  analysing input database')
        subs = self.task_def.subobjectives
        consts = self.task_def.constraints
        self._latest_date = self.task_def.init_date

        # go first through the subobjectives, and then through the constraints
        # if there are any
        self._add_info('    processing subobjectives')
        self._process_exprs(subs, 'subobjective', True)
        if consts is not None:
            self._add_info('    processing constraints')
            self._process_exprs(consts, 'constraint', False)

        # get simulation unit ids, the number of simulation units and the
        # maximum number of iterations, branches and time steps from data
        # db as operation dtable does not necessarily contain all units,
        # branches or iterations (as operations may not have been done to all
        # units, branches or iterations)
        start = time.time()
        self._add_info('    calculating problem size')
        self.data_filter = None
        if self.task_def.data_filter:
            constraints = []
            for op in self.task_def.data_filter:
                if op[0] == (EXPR_DTYPE_OP, EXPR_DTYPE_COMB,):
                    errormsg = "operation data not allowed in data filter"
                    self._add_error(errormsg)
                    return False
                elif op[0] == EXPR_DTYPE_DATA:
                    constraints.append((op[1][1],))
                elif op[0] == EXPR_DTYPE_VALUE:
                    constraints.append((op[1],))
                else:
                    b = constraints.pop()
                    a = constraints.pop()
                    eq = op[0]
                    if eq in ('and', 'or'):
                        constraints += [a + (eq,), b]
                    else:
                        constraints += [a + (eq,) + b]
            header = ['id', 'iteration', 'branch']
            res = list(self._db.get_data_from_level('data', header,
                                                    self.data_level,
                                                    constraints=constraints))
            self.unit_ids = list(set([id[0] for id in res]))
            self.max_iterations = max([id[1] for id in res])
            self._max_branches = max([id[2] for id in res])
            self.data_filter = constraints
        else:
            ids = set(self._db.get_ids('data', self.data_level))
            ids.update(set(self._db.get_ids('op_res', self.data_level)))
            self.unit_ids = list(ids)
            self.max_iterations = max(self._db.get_max_iter('data'),
                                      self._db.get_max_iter('op_res'))
            self._max_branches = max(self._db.get_max_branch('data'),
                                     self._db.get_max_branch('op_res'))
        end = time.time()
        self._add_info('    size calculated in %0.1f sec' % (end - start))

        if len(self.unit_ids) == 0:
            errorstr = "Empty database, nothing to optimize"
            self._add_error(errorstr)
            return False

        if self.max_iterations is None:
            errorstr = "Maximum number of iterations couldn't be deduced "\
                       "from database contents, database is probably empty"
            self._add_error(errorstr)
            return False

        if self._max_branches is None:
            errorstr = "Maximum number of branches couldn't be deduced "\
                       "from database contents, database is probably empty"
            self._add_error(errorstr)
            return False

        # the actual number of iterations and branches is max + 1
        self.max_iterations += 1
        self._max_branches += 1
        self.num_of_units = len(self.unit_ids)
        self.unit_index = dict([(unicode(v),k) for \
                                k,v in enumerate(self.unit_ids)])
        self.orig_uids = dict([(unicode(uid), unicode(uid)) for \
                               uid in self.unit_ids])
        self._unit_range = numpy.arange(self.num_of_units, dtype=int)

        # add some info messages
        msg = '  number of simulated iterations: %d' % self.max_iterations
        self._add_info(msg)
        return True

    def _process_exprs(self, expressions, exprtype, subobj=True):
        """Go through the expressions and try to deduce data level and
        earliest and latest dates
        """
        for item in expressions:
            for e in item.expr:
                if e[0] == EXPR_DTYPE_DATA:
                    if e[1][0] != self.data_level:
                        errorstr = "Multiple levels in %ss. Check the "\
                            "optimization task definition" % exprtype
                        self._add_error(errorstr)
                elif e[0] != EXPR_DTYPE_OP:
                    continue
                else:
                    pass
                start, end = e[2]
                if not self._earliest_date or \
                    (start is not None and start < self._earliest_date):
                    self._earliest_date = start
                if self._latest_date is not None and \
                    (end is None or end > self._latest_date):
                    self._latest_date = end

    def _create_tables(self):
        """
        Create optimization data tables, using Numpy arrays
        for the unique operand arrays

        The data is stored in a dictionary where keys are unique operand
        indices and values are unique operand data arrays

        Unique operand (subobjective or constraint) array shape:
           (units,
            branches,
            max vector length (number of time steps))
        """
        nunits = self.num_of_units
        maxbr = self._max_branches
        nso = self.task_def.num_of_subobjs

        # construct a container for global maximum subobjective values and
        # one for number of branches of each unit
        self._global_max = numpy.zeros(nso, dtype=DTYPE_FLOAT)

        # construct a table for conditional discount rates if conditional
        # discounting is to be used
        if self.task_def.conditional_discounting:
            self._conditional_discount_rates = \
                numpy.zeros((nunits, maxbr), dtype=DTYPE_FLOAT)

        # open the data file and define compression filter and data atom
        if self._use_existing:
            if not os.path.exists(self._data_file):
                errmsg = 'Reuse optimization data file set, but %s'\
                        ' does not exist' % self._data_file
                self._add_error(errmsg)
                return False
            try:
                fp = open(self._data_file, 'rb')
                self.data = pickle.load(fp)
                fp.close()
            except:
                errmsg = 'Error loading optimization data file!'
                self._add_error(errmsg)
                return False
        else:
            self._total_size = 0.
            self.data = {'operands': {},
                         'date_locs': {}}
            self._date_index = {}
            self._dates = {}
            date_cache = {}

            # go through each unique operand and construct a 3D Numpy array
            # of size num of units x max num of branches x number of time
            # steps for each operand
            opdata = self.data['operands']
            for operand, ui in self.task_def.unique_operands.items():

                # get operands date limits from pre-built task definition
                start, end = self.task_def.date_ranges[ui]

                if self.task_def.one_aggr_level[ui]:
                    if operand[0] == EXPR_DTYPE_OP:
                        dtables = (('op_res', None),)
                    elif operand[0] == EXPR_DTYPE_DATA:
                        dtables = (('data', operand[1][0]),)
                    else:
                        dtables = (('op_res', None),
                                   ('data', self.task_def.main_level),)
                else:
                    dtables = (('op_res', None),
                               ('data', self.task_def.main_level),)

                # check if the dates for the current datatable-level-daterange
                # combination have been retrieved from the database already
                dckey = dtables + (start, end)
                if dckey in date_cache:
                    dd = date_cache[dckey]
                else:
                    dd = []
                    for dtable, level in dtables:
                        constr = (start, end)
                        if constr != (None, None):
                            new_dt = self._db.get_dates(dtable, constr)
                        else:
                            # NB: If start and end time steps are Nones, it 
                            # means that the value should come from the last
                            # records in the database
                            # NB: Still, we have to prevent collecting results 
                            # that are prior to plan start date
                            constr = (self._earliest_date, None)
                            new_dt = self._db.get_dates(dtable, constr)
                            new_dt = sorted(new_dt)[-1:]
                        dt = datetime.date
                        dd += [dt(nd.year, nd.month, nd.day) for nd in new_dt]
                    dd = numpy.unique(dd)
                    dd.sort()
                    # store the dates to date cache
                    date_cache[dckey] = dd

                if start is None:
                    start = dd[0]
                if end is None:
                    end = dd[-1]

                # make sure that there are dates for the current operand
                # if there are no dates, add the start and end dates, which
                # eventually will be just zero value holders in the
                # optimization matrix
                if dd.size == 0:
                    dd = numpy.concatenate((dd, numpy.array([start, end])))


                span = (dd >= start) * (dd <= end)
                dates = dd[span]
                ndates = span.sum()
                shape = (nunits, maxbr, ndates)
                opdata[ui] = numpy.zeros(shape, dtype=DTYPE_FLOAT)
                self._update_total_size(shape)

                # construct date index for finding correct matrix locations
                # with date information
                dtuples = [(d,i) for i,d in enumerate(dates)]
                self._date_index[ui] = dict(dtuples)
                self._dates[ui] = dd

            # go through subobjectives and constraints and find the date
            # locations of each operand
            for sub in self.task_def.subobjectives:
                if not self._update_date_locs(sub.expr):
                    return False
            if self.task_def.num_of_consts > 0:
                for const in self.task_def.constraints:
                    if not self._update_date_locs(const.expr):
                        return False

            self.branches = defaultdict(list)

            nu = len(self.task_def.unique_operands)
            msg = '  number of unique subobjective/constraint operands: %d' \
                    % nu
            self._add_info(msg)
            human_total = convert_bytes(self._total_size)
            msg = '  total size for optimization data arrays: %s' % human_total
            self._add_info(msg)

        return True

    def _update_date_locs(self, expr):
        """Update date location index for current expression

        expr -- postfix epxression in a list
        """
        ok = True
        dtypes = (EXPR_DTYPE_COMB, EXPR_DTYPE_DATA, EXPR_DTYPE_OP)
        for operand in expr:
            if operand[0] not in dtypes:
                continue
            oi, ui = operand[-1]
            start, end = operand[2]
            if start is None:
                start = self._dates[ui][0]
            if end is None:
                end = self._dates[ui][-1]
            i_s = self._get_date_ind(ui, start, 'start')
            i_e = self._get_date_ind(ui, end, 'end')
            # If either the starting or ending date location is missing, set
            # the locations as Nones to indicate that there's no data in the
            # result database for the operands for the particular dates.
            if i_s is None or i_e is None:
                i_s, i_e = None, None
            self.data['date_locs'][oi] = (i_s, i_e)

        if not ok:
            msg = 'Problem placing dates %s and %s for operand %s'\
                % (start, end, operand[1])
            self._add_error(msg)
            return False
        return True

    def _get_date_ind(self, ui, d, date_type='start'):
        """Get correct date indice, find the closest later date in case date
        is not defined in date index

        ui -- unique operand indice, int
        d -- date as datetime.date object
        date_type -- date type as a string, either 'start' or 'end'
        """
        if d not in self._date_index[ui]:
            # get minimum and maximum dates
            mind = min(self._date_index[ui].keys())
            maxd = max(self._date_index[ui].keys())
            dfmt = "%d.%m.%Y"

            # if date is smaller than the smallest date in result database and
            # we are looking for start date, the location indice should be 0
            if d < mind:
                if date_type == 'start':
                    ind = 0
                elif date_type == 'end':
                    msg = "Operand ending date %s outside data date "\
                          "range %s-%s" % (d.strftime(dfmt),
                                           mind.strftime(dfmt),
                                           maxd.strftime(dfmt))
                    self._add_error(msg)
                    ind = None
            elif d > maxd:
                if date_type == 'end':
                    ind = self._date_index[ui][maxd]
                elif date_type == 'start':
                    msg = "Operand starting date %s outside data date "\
                          "range %s-%s" % (d.strftime(dfmt),
                                           mind.strftime(dfmt),
                                           maxd.strftime(dfmt))
                    self._add_error(msg)
                    ind = None
            else:
                candidate = datetime.date(1,1,1)
                delta = abs(d - candidate)
                for cd, ind in self._date_index[ui].items():
                    if abs(d - cd) <= delta:
                        if date_type == 'start' and cd > d:
                            candidate = cd
                            delta = abs(d - cd)
                        elif date_type == 'end' and cd < d:
                            candidate = cd
                            delta = abs(d - cd)
                ind = self._date_index[ui][candidate]
                if candidate == datetime.date(1,1,1):
                    # NB: this should not happen!
                    errormsg = "Could not find date location indice for "\
                               "date %s" % d.strftime(dfmt)
                    self._add_error(errormsg)
                    ind = None
        else:
            ind = self._date_index[ui][d]

        return ind

    def _update_total_size(self, shape):
        """
        Updates the total size of unique operand data arrays
        """
        dim_size = 1
        for dim in shape:
            dim_size *= dim
        size = 8 * dim_size
        self._total_size += size

    def _fill_tables(self, it):
        """
        Fill subobjective and constraint arrays with values from input
        databases. Also fill the conditional discount rate tables, if
        conditional discounting is to be done.

        it -- current simulated iteration indice as int
        """
        self._add_info('  filling tables')
        
        td = self.task_def
        evaluate = self._evaluator.evaluate
        rn = self._unit_range
        n = len(rn)
        ofunc_scope = self.task_def.objective_func.scope
        ofunc_type = self.task_def.objective_func.type

        # get starting time in seconds since epoch
        startt = time.time()

        # if conditional discounting is to be used, get the discount rates
        # before collecting and discounting data values
        if self.task_def.conditional_discounting:
            self._get_conditional_discount_rates(it)

        # collect values for all unique operands
        for operand, ui in td.unique_operands.items():
            array = self.data['operands'][ui]
            if not self._collect_values(it, operand, array, ui):
                return False

        # make sure the zero branch is always there; e.g. if operands only
        # contain operation results (0 branch not having any)
        # also, order branch id lists as the max branch per unit is needed
        # later
        for unit_ind, branches in self.branches.iteritems():
            if 0 not in branches:
                branches.append(0)
            branches.sort()
        self.branches = dict(self.branches)

        # calculate maximum utilities for each sub-objective in order to find
        # the global maximum for each sub-objective...
        for i, so in enumerate(td.subobjectives):
            if ofunc_scope == 'global' and ofunc_type == 'distance':
                # ...except for distance type ofuncs where it's not needed
                continue
            tmp = numpy.zeros((n, self._max_branches), dtype=DTYPE_FLOAT)
            # pass the unique operand data and unique operand index to the
            # expression evaluator
            data = self.data['operands']
            dlocs = self.data['date_locs']
            for br in range(self._max_branches):
                tmp[self._unit_range, br] = evaluate(self._unit_range, br,
                                                     so.expr, dlocs, data, 3,
                                                     False)
            max_of_units = numpy.nanmax(tmp, axis=1)
            max_of_so = numpy.nansum(max_of_units, axis=0)
            self._global_max[i] = max_of_so

        # get ending time in seconds since epoch and log info message
        endt = time.time()
        delta = endt - startt
        msg = '  optimization data arrays filled in %s' % (secs2str(delta))
        self._add_info(msg)

        return True

    def _get_conditional_discount_rates(self, it):
        """
        Get conditional discount rates

        it -- current simulated iteration indice as int
        """
        self._conditional_discount_rates[:,:] = self.task_def.discount_rate
        shp = self._conditional_discount_rates.shape

        # constraint the data fetching to only current iteration and if the
        # task has a data filter, add that to the constraint as well
        if self.data_filter:
            constraints = [('iteration', 'eq', it, 'and')] + self.data_filter
        else:
            constraints = [('iteration', 'eq', it)]

        # get the date range; the condition date should be the same as the
        # earliest date in the optimization task operand dates
        all_dates = set([])
        for item, dates in self._dates.items():
            for date in dates:
                all_dates.add(date)
        dates = tuple(sorted(all_dates)[:2])

        # go through each condition and evaluate the postfix expression
        for cond, expr, drate in self.task_def.conditional_discounting:
            stack = []
            for op in expr:
                if op[0] == EXPR_DTYPE_OP:
                    errormsg = "discount rate can be conditional to "\
                               "data only, not operations"
                    self._add_error(errormsg)
                    return False
                elif op[0] == EXPR_DTYPE_DATA:
                    header = ['id', 'branch', 'data_date', op[1][1]]
                    data = self._db.get_data_from_level('data', header,
                                                        self.data_level,
                                                        constraints=constraints,
                                                        dates=dates)
                    arr = numpy.zeros(shp, dtype=DTYPE_FLOAT)
                    arr.fill(numpy.nan)
                    for uid, branch, date, value in data:
                        i = self.unit_index[uid]
                        arr[i, branch] = value
                    stack.append(arr)
                elif op[0] == EXPR_DTYPE_VALUE:
                    stack.append(op[1])
                else:
                    if len(stack) < 2:
                        errormsg = 'invalid condition "%s"' % cond
                        self._add_error(errormsg)
                        return False
                    b = stack.pop()
                    a = stack.pop()
                    stack.append(op[1](a, b))

            # set the discount rate for objects that passed the condition
            passed = stack[0]
            self._conditional_discount_rates[passed] = drate

    def _collect_values(self, it, operand, array, ui, forced_dates=None):
        """
        Get values for a single expression operand from either data or
        operation result database

        it -- current simulated iteration indice as int
        operand -- subobjective or constraint operand
        array -- target Numpy array
        ui -- unique subobjective/constraint operand indice
        forced_dates -- forced (end, start) dates (only for constraint
                        variables)
        """
        self._add_info('    collecting values for operand %d' % ui)
        
        evalc = self._eval_constraint
        getv = self._get_values_from_db
        setv = self._set_value

        # check if the expression item is normal (length = 4) or operand
        # constraint (length = 2) operand
        norm = len(operand) == 4

        # if forced_dates are defined (only in the case that the values
        # are collected for constraint variables), no need to get the
        # dates otherwise, get the date definition from the operand
        if forced_dates is None:
            dates = self.task_def.date_ranges[ui]
        else:
            dates = forced_dates

        # if operand is normal and it has a constraint, evaluate the
        # operand constraint
        if norm and operand[2] is not None:
            # operand has a constraint: process the constraints and
            # fill the data matrix for objects that passed
            ok = evalc(it, dates, array, operand, ui, norm)
            if not ok:
                return False
        else:
            # get values from database and fill the target matrix with
            # the data values (without any constraints or such)
            constr = dates
            if dates == (None, None):
                # If we have "last date only" -type of aggregation, such as 
                # ("sum[-1:-1]"), we need to still constrain the first date
                constr = (self._earliest_date, None)
            data = getv(it, operand, constr, norm)
            if dates == (None, None):
                # when start and end date are Nones, the value should be from
                # the last recorded date of each iteration-branch-object
                # combination ([-1,-1] syntax in expr definition)
                data = self._filter_data_by_last_date(data)
            for uid, br, date, value in data:
                # we need to ensure we have a date, not a datetime.
                if isinstance(value, unicode) or isinstance(value, str):
                    errorstr = "No values for '%s' on level '%s' in "\
                                "database" % (value, operand[1][0])
                    self._add_error(errorstr)
                    return False
                setv(array, uid, br, ui, date, value)

        # check if variable is of 'normal' type and if it's value should
        # be discounted
        if norm and operand[3]:
            # discount values to init date
            drate = self.task_def.discount_rate
            idate = self.task_def.init_date
            for d, j in self._date_index[ui].items():
                mslice = array[...,j]
                if self.task_def.conditional_discounting:
                    drates = self._conditional_discount_rates
                    array[...,j] = calculate_NPV(mslice, drates, idate, d)
                else:
                    array[...,j] = calculate_NPV(mslice, drate, idate, d)

        return True

    def _get_values_from_db(self, it, operand, dates, norm, condition=None):
        """
        Get value(s) for a single operand from data or operation result
        database

        it -- current simulated iteration indice as int
        operand -- current operand definition
        dates -- start and end dates
        norm -- boolean indicating whether the operand is normal
        condition -- operand condition
        """
        optype = operand[0]

        # NOTE: 'combined' data type and plain 'data' and 'operation' data
        # types will have different constraint structures
        if optype in (EXPR_DTYPE_OP, EXPR_DTYPE_DATA,):
            constraints = [('iteration', 'eq', it)]
        else:
            constraints = [(EXPR_DTYPE_OP, 'iteration', 'eq', it)]

        # add the data filter to the constraint: if optype is EXPR_DTYPE_DATA,
        # this is very simple, if EXPR_DTYPE_COMB, the level has to be added
        # to the operands, and if EXPR_DTYPE_OP, the optype needs to be
        # changed to EXPR_DTYPE_COMB, the level has to be added to the
        # operands and the condition operands need to be updated as well
        change_type = False
        if self.data_filter:
            constraints[0] = constraints[0] + ('and',)
            ml = self.task_def.main_level
            if optype == EXPR_DTYPE_DATA:
                constraints += self.data_filter
            elif optype == EXPR_DTYPE_OP:
                change_type = True
                optype = EXPR_DTYPE_COMB
                constraints[0] = (EXPR_DTYPE_OP,) + constraints[0]
                constraints += [(ml,) + item for item in self.data_filter]
            else:
                constraints += [(ml,) + item for item in self.data_filter]

        if condition is not None:
            # do a little postfix to prefix conversion for the constraint
            # postfix expression
            prefx = []
            for op in condition:
                if op[0] == EXPR_DTYPE_COMB:
                    prefx.append(tuple(op[1:3]))
                elif op[0] in (EXPR_DTYPE_OP, EXPR_DTYPE_DATA,):
                    if change_type:
                        prefx.append(tuple(op[1:3]))
                    else:
                        prefx.append((op[2],))
                elif op[0] == EXPR_DTYPE_VALUE:
                    prefx.append((op[1],))
                else:
                    b = prefx.pop()
                    a = prefx.pop()
                    eq = op[-1]
                    if eq in ('and', 'or'):
                        prefx += [a + (eq,), b]
                    else:
                        prefx += [a + (eq,) + b]
            constraints[-1] = constraints[-1] + ('and',)
            constraints += prefx

        lb = self.limit2branches
        required = None
        if optype == EXPR_DTYPE_OP:
            header = ['id', 'branch', 'op_date']
            header.append(operand[1])
            if operand[1] is not None:
                required = [operand[1]]
            self._add_info('      getting operation results from db')
            data = self._db.get_data_from_level('op_res', header, 0,
                                                constraints=constraints,
                                                dates=dates,
                                                required=required,
                                                limit2branches=lb)
        elif optype == EXPR_DTYPE_DATA:
            header = ['id', 'branch', 'data_date']
            # the varname location is different for normal operands and
            # operation constrain operands (the latter lacks level name)
            header.append(operand[1][1] if norm else operand[1])
            self._add_info('      getting simulated results from db')
            data = self._db.get_data_from_level('data', header,
                                                self.data_level,
                                                constraints=constraints,
                                                dates=dates,
                                                limit2branches=lb)
        elif optype == EXPR_DTYPE_COMB:
            header = [(EXPR_DTYPE_OP, 'id'),
                      (EXPR_DTYPE_OP, 'branch'),
                      (EXPR_DTYPE_OP, 'op_date')]
            header.append((EXPR_DTYPE_OP, operand[1]))
            if operand[1] is not None:
                required = [operand[1]]
            data = self._db.get_data_from_level('combined', header,
                                                self.data_level,
                                                constraints=constraints,
                                                dates=dates,
                                                limit2branches=lb)
        return data

    def _eval_constraint(self, it, dates, array, operand, ui, norm):
        """Evaluate operand constraint and store only those
        objects for which the condition evaluated True

        it -- current simulated iteration indice as int
        dates -- start and end dates
        array -- target Numpy array
        operand -- operand structure
        ui -- unique operand indice
        norm -- boolean indicating whether the operand is normal
        """
        condition = operand[2]

        # get the data values from database
        data = self._get_values_from_db(it, operand, dates, norm, condition)
        if dates == (None, None,):
            data = self._filter_data_by_last_date(data)

        # go through the data objects and evaluate the condition for each
        # iteration-object-branch-date combination
        setv = self._set_value

        for item in data:
            uid, br, date, value = item
            if uid not in self.unit_index:
                continue
            if dates != (None, None):
                if dates[0] is not None and date < dates[0]:
                    continue
                if dates[1] is not None and date > dates[1]:
                    continue
            setv(array, uid, br, ui, date, value)

        return True

    def _set_value(self, array, uid, br, ui, date, value):
        """
        Store a single value into a correct location in a matrix and update
        branch number container

        array -- Numpy array
        uid -- object id, str
        br -- branch, int
        ui -- unique subobjective/constraint indice, int
        date -- datetime object
        value -- value to store, float
        """
        if value is None:
            return

        # ensure datetime.date instead of datetime.datetime
        if type(date) == datetime.datetime:
            date = datetime.date(year=date.year, month=date.month, day=date.day)

        i_u = self.unit_index[uid]
        # NB: In cases where we want a value from last simulation time step, 
        # like when XML syntax is "sum[-1:-1](<level>:<variable> ...), we might
        # encounter a situation in which a stand does not have any values 
        # during the last time step due to simulation errors. In this case we 
        # should just skip this value.
        if date not in self._date_index[ui]:
            return
        i_d = self._date_index[ui][date]

        if isnan(array[i_u,br,i_d]):
            array[i_u,br,i_d] = 0.0
        
        array[i_u,br,i_d] += value

        if br not in self.branches[i_u]:
            self.branches[i_u].append(br)

    def _filter_data_by_last_date(self, data):
        """Filter data so that only the values from the last recorded date of
        each iteration-branch-object combination are selected. This is needed
        when operand dates are defined as [-1:-1]

        data -- data generator object from datadb
        """
        temp = {}
        for uid, br, date, value in data:
            key = (uid, br)
            if not key in temp:
                temp[key] = {}
            if not date in temp[key]:
                temp[key][date] = []
            temp[key][date].append(value)
        filtered = []
        for key in temp.keys():
            dates = temp[key].keys()
            dates.sort()
            for value in temp[key][dates[-1]]:
                uid, br = key
                filtered.append((uid, br, dates[-1], value))
        return filtered


    ##########################################################################
    # METHODS FOR CALCULATING UTILITES AND EVALUATING CONSTRAINTS
    ##########################################################################
    def solution_utility(self, solution):
        """Check objective function utility value in given solution
        """
        # for conjunctive utility function the initial utility must be 1.0
        # instead of 0.0 or else the total utility will be 0.0
        ofunc_type = self.task_def.objective_func.type
        ofunc_scope = self.task_def.objective_func.scope
        if ofunc_type == 'conjunctive':
            u_total = 1.0
        else:
            u_total = 0.0
        evaluate = self._evaluator.evaluate
        urange = self._unit_range
        dlocs = self.data['date_locs']
        data = self.data['operands']
        for i, so in enumerate(self.task_def.subobjectives):

            # evaluate subobjective function
            value = evaluate(urange, solution, so.expr, dlocs, data)

            # subobjective value can be evaluated locally (stand level) or
            # globally
            if ofunc_scope == 'global':
                if ofunc_type == 'distance':
                    # for distance utility functions, the optimal value is the
                    # user-defined optimal value
                    opt_level = so.optlevel
                    if opt_level == 0.0:
                        opt_level = 0.001  # this is to avoid infinite u_subs
                    u_sub = abs((value - opt_level) / opt_level)
                else:
                    # for other utility functions, the optimal value is the
                    # global maximum value
                    opt_level = self._global_max[i]
                    u_sub, err = _value2utility(so, value, opt_level)
                    if err != '':
                        self._add_error(err)
            else:
                # for local utilities, the utility value is calculated already
                # when constructing the optimization data set
                opt_level = self._global_max[i]
                u_sub, err = _value2utility(so, value, opt_level)
                if err != '':
                    self._add_error(err)
                u_sub = u_sub / self.num_of_units
            weight = so.weight
            # each objective function can have a number of subobjective
            # functions.
            # additive: Ui=SUM(aj*uji)
            # conjunctive: Ui=MULTIPLY((uji)**aj)
            # distance: Ui=sqrt(SUM(aj**2*(optj-cij)**2))
            # where aj is weight and uji is the performance of alternative
            # i against the criterion j and SUM(aj)=1
            if ofunc_type == 'additive':
                u_total += u_sub * weight
            elif ofunc_type == 'conjunctive':
                # check that negative values are not raised to fractional power
                if u_sub >= 0:
                    u_total *= (u_sub ** weight)
                else:
                    u_total *= -(abs(u_sub) ** weight)
            elif ofunc_type == 'distance':
                u_total += math.sqrt(weight ** 2 * u_sub ** 2)
        # total utility value is the average of the solution values
        return u_total

    def solution_feasibility(self, solution, first_seek=False):
        """
        Check given solution feasibility.
        For the inital seek for feasible solution goes through all the
        constraints in order to be able to report back to the user the
        constraint feasible/not feasible ratio in case of not finding any
        feasible solutions.
        """
        feasible = True
        nconst = self.task_def.num_of_consts
        if nconst == 0:
            return True
        evaluate = self._evaluator.evaluate
        dlocs = self.data['date_locs']
        data = self.data['operands']
        for i, item in enumerate(self.task_def.constraints):
            evalres = evaluate(self._unit_range, solution, item.expr, dlocs, 
                               data)
            if not evalres:
                if not first_seek:
                    return False
                else:
                    self._non_feasible_counter[i] += 1
                    feasible = False
        return feasible

    def compare_utilities(self, best, candidate):
        """Compare utility values of two solutions: is candidate better than
        the current best solution?
        """
        if best is None:
            # this is for cases when initial utility hasn't been set yet
            return True
        if self.task_def.objective_func.target == 'max':
            return candidate > best
        elif self.task_def.objective_func.target == 'min':
            return candidate < best

    def get_random_solution(self):
        """
        Get a random solution without checkin it's feasibility
        """
        sol_branches = numpy.zeros(self.num_of_units, dtype=int)
        for unit_ind, branches in self.branches.iteritems():
            pick = random.randint(0, len(branches) - 1)
            sol_branches[unit_ind] = branches[pick]
        return sol_branches

    def get_status(self, solution):
        """
        Get a data structure with string representations and current values
        of all subobjectives and constraints

        iteration -- current iteration
        solution -- current solution
        """
        # get total solution utility
        utility = self.solution_utility(solution)
        if not isinstance(utility, float):
            utility = utility[0]
        status = {'utility': utility,
                  'subobjectives': []}
        evaluate = self._evaluator.evaluate
        dlocs = self.data['date_locs']
        data = self.data['operands']
        for i, so in enumerate(self.task_def.subobjectives):
            key = "%d. subobjective" % (i + 1)
            # evaluate subobjective function
            value = evaluate(self._unit_range, solution, so.expr, dlocs, data)
            status['subobjectives'].append((key, value))
        # return the status data structure
        return status


def main():
    import doctest
    doctest.testfile('./test/omatrix-py.rst')

if __name__ == '__main__':
    main()
