#!/usr/bin/env python
# encoding: utf-8
"""
Title: SIMO optimizer postfix evaluator

...

*postfixeval.py*

Copyright (c) 2007 Simosol Oy. Distributed under the GNU General Public License 2.0.
"""

import numpy
from collections import deque

BOOLEANS = (numpy.bool_, numpy.bool8, bool)
FLOATS = (numpy.float, numpy.float_, numpy.float32, numpy.float64, float)
from simo.builder.names import EXPR_DISCOUNT, EXPR_DTYPE_COMB,\
                               EXPR_DTYPE_DATA, EXPR_DTYPE_OP,\
                               EXPR_DTYPE_VALUE, EXPR_OPERATION


class PostfixEvaluator:
    """Class for evaluating expressions in postfix form"""
    
    def evaluate(self, obj, branch, postfx, dlocs, data, dims=3,
                 aggr_over_units=True, violation_metric=None):
        """
        Evaluate postfix expression, where postfx is the expression in
        Reverse Polish Notation (RPN) or postfix form.
        
        obj -- range of object indices [0..n[, numpy integer array, or 
               an integer scalar
        branch -- numpy integer array of size n, or integer scalar
        postfx -- expression in postfix form
        dlocs -- operand date location indices in a list
        data -- unique operand data in a dict
        dims -- data dimensions, int
        aggr_over_units -- aggregate over units, boolean
        violation_metric -- violation metric: None, 'absolute' or 'percentage'
        """
        stack = []  # evaluation stack
        violations = []  # condition evaluation violations
        i = 0  # counter for variable/operation operands
        act = 0  # counter for items inside current aggregation block
        for item in postfx:
            itemtype = item[0]
            if itemtype in (EXPR_DTYPE_COMB, EXPR_DTYPE_DATA, EXPR_DTYPE_OP,):
                # add operand values to the evaluation stack
                if dims == 1:
                    stack.append(data[i])
                else:
                    oi, ui = item[-1]
                    values = data[ui]
                    start, end = dlocs[oi]
                    # If either of date locs are None, the value is considered
                    # to be missing.
                    if None in (start, end):
                        if not aggr_over_units:
                            # When not aggregating over units, we are  
                            # evaluating the expression separately for each
                            # unit and therefore the "missing" value has to be
                            # a numpy array instead of scalar value.
                            value = numpy.zeros([len(obj),1], dtype=float)
                        else:
                            value = 0.
                    else:
                        value = values[obj,branch,start:end+1]
                    stack.append(value)
                i += 1
                act += 1
            elif itemtype == EXPR_DTYPE_VALUE:
                stack.append(item[1])  # add scalar value to stack
            elif itemtype == 'aggr':
                aggregate = self._aggregate(stack, item, act, aggr_over_units)
                stack[-act:], act = aggregate, 0
            else:
                if len(stack) < 2:
                    raise Exception, 'Error: Too few input values'
                b = stack.pop()
                a = stack.pop()
                stack.append(self._nans2zeros(item[1](a, b)))
                if violation_metric:
                    violations.append(self._violation(item, a, b, 
                                                      violation_metric))
                act -= 1
                act = 0 if act < 0 else act
        if len(stack) > 1:
            raise Exception, "Too many input values!"
        # If violation metric was defined in the function call, return the 
        # violations as well
        if violation_metric:
            return stack[0], violations
        else:
            return stack[0]

    def _aggregate(self, stack, item, act, aggr_over_units):
        """Aggregation function; aggregate value matrices over objects and
        time steps into single numerical values, or over time steps into
        value vectors
        """
        if not aggr_over_units:
            aggregate = [item[1](x, axis=1) for x in stack[-act:]]
        else:
            aggregate = [item[1](x) for x in stack[-act:]]
        return aggregate
    
    def _nans2zeros(self, values):
        """Check if value/vector/array is either nan or inf and change the
        nans and infs to zeros
        
        values -- values is either a boolean, a float, or a numpy vector/array
        """
        if isinstance(values, BOOLEANS):
            pass
        elif isinstance(values, FLOATS):
            if values is numpy.nan or values == numpy.inf:
                values = 0.0
        else:
            replace = numpy.isnan(values) | numpy.isinf(values)
            values[replace] = 0.0
        return values

    def _violation(self, item, a, b, violation_metric):
        '''
        Calculate how large the violation of an equality condition is
        '''
        violation = 0.0
        if item[2] in ('lt', 'le'):
            if a > b:
                violation = abs(a - b)
        elif item[2] in ('gt', 'ge'):
            if a < b:
                violation = abs(a - b)
        if violation and violation_metric == 'percentage':
            if b != 0:
                violation = violation / b * 100.
            else:
                violation = float('inf')
        return violation
        


def main():
    pass

if __name__ == '__main__':
    main()
    