#!/usr/bin/env python
# encoding: utf-8
"""
Title: SIMO aggregation output

Module for data aggregation reports; e.g. averages and sums of data variables
over different periods of time.

*aggr.py*

Copyright (c) 2009 Simosol Oy. Distributed under the GNU General Public License 2.0.
"""

from output import Output
import datetime
from datetime import timedelta
from calendar import monthrange
from pysqlite2.dbapi2 import ProgrammingError, OperationalError
from dateutil.relativedelta import relativedelta
from collections import defaultdict
import operator
import numpy
from simo.builder.names import EXPR_DISCOUNT, EXPR_DTYPE_COMB,\
                               EXPR_DTYPE_DATA, EXPR_DTYPE_OP,\
                               EXPR_DTYPE_VALUE, EXPR_OPERATION



class OutputAggr(Output):

    def run(self):
        self.aggregation(self.output_filename)

    def aggregation(self, file):
        self.pads = None
        self.write_started = False
        result = self._run_aggr(self.aggregation_def.data)
        self._write_file(file, result, True)

    def _run_aggr(self, aggr_dict):
        reporting_variables = aggr_dict['reporting_variables']
        discount_rate = aggr_dict['discount_rate']
        init_date = aggr_dict['initial_date']
        result = []
        for reporting_periods in aggr_dict['reporting_periods']:
            time_unit = reporting_periods['unit']
            time_length = int(reporting_periods['length'])
            for var in reporting_variables:
                attr = self._execute_var(discount_rate, init_date, time_unit,
                                         time_length, var)
                if attr is not None:
                    result.append(attr)
        return result

    def _execute_var(self, discount_rate, init_date, time_unit,
                     time_length, var):
        date_dict = {}
        errors = defaultdict(set)
        expression = var['expression']
        proportion = var['proportion']
        grouping = var['grouping']
        if grouping == {}:
            grouping = None
        expression_header = expression
        try:
            expression = self._unpack_weighted_average(expression)
        except Exception, e:
            self.logger.log_message('aggregation', 'error', str(e))
            return None
        all_dates = [datetime.date(d.year, d.month, d.day) \
                     for d in self.datadb.get_all_dates()]
        all_dates.sort()
        all_dates = [d for d in all_dates if d >= init_date]
        if init_date not in all_dates:
            all_dates.insert(0, init_date)
        expr = self._parse_expression(expression, expression_header,
                                      init_date)

        if expr is None:
            return None
        date_ranges = self._get_date_ranges(init_date, all_dates, expr,
                                            time_unit, time_length)
        date_dict, classifiers, errors = self._get_values(expr, expression,
                                                 date_ranges, grouping, errors,
                                                 date_dict, init_date,
                                                 discount_rate)
        if classifiers is None:
            return None
        if grouping is None:
            grouping = {}
        results = self._get_results(date_dict, grouping, classifiers, expr,
                                    proportion)
        if results is None:
            return None
        results, res_grp_header = results

        if self.result_padding:
            results = self._pad_all(results)
        self.pads = None
        data = []
        r = 'Results' if not proportion else 'Proportions'
        data.append('%s for expression (%d keys): %s' % (r, len(results)-1,
                                                         expression_header))
        for row in results:
            drow = []
            for col in row:
                drow.append(col)
            data.append(';'.join(drow))

        ekeys = errors.keys()
        ekeys.sort()
        for key in ekeys:
            errs = errors[key]
            drow = [key]
            errs = sorted(errs)
            for err in errs:
                err = [str(e) for e in err]
                drow.append(', '.join(err))
            data.append(';'.join(drow))
        return data


    def _parse_expression(self, expression, header, init_date):
        """Will parse a string expression into a list of tuples.
        Returns None if there are errors."""
        try:
            expr, err = self.expr_parser.parse(expression, init_date, 'year')
            if err is not None:
                msg = "Parsing '%s' failed:" % header + ' ' + err
                self.logger.log_message('aggregation', 'error', msg)
                return None
            return expr
        except Exception, e:
            msg = "Parsing '%s' failed! check your syntax" % (header,)\
                    + '\n' +str(e)
            self.logger.log_message('aggregation', 'error', msg)
            return None


    def _get_classifiers(self, keys):
        """Build and return classifiers for the given grouping"""
        raw_keys = []
        for key in keys:
            lev, var = key.split(':')
            if lev == EXPR_OPERATION:
                if var in ('id', 'orig_op_id', 'iteration', 'branch', 
                           'op_date', 'op_name', 'op_level', 'op_group', 
                           'op_type'):
                    raw_keys.append(('op_link', var))
                else:
                    raw_keys.append(('op_res', var))
            else:
                raw_keys.append((lev, var))
        classifiers = self.datadb.get_raw_data_from_levels(raw_keys,
                                                           required=raw_keys,
                                                           distinct=True)
        if classifiers is not None and len(classifiers)==0:
            classifiers = ['']
        elif classifiers is not None:
            classifiers.sort()
        return classifiers


    def _get_values(self, expr, expression, date_ranges, grouping, errors,
                    date_dict, init_date, discount_rate):
        """Get the values from the database into the date_dict"""
        dtypes = (EXPR_DTYPE_COMB, EXPR_DTYPE_DATA, EXPR_DTYPE_OP)
        for e in expr:
            if e[0] not in dtypes:
                continue
            # check if the expression item is normal (length = 5) or operand
            # constraint (length = 3) operand
            norm = len(e) == 5
            # get operand values from database
            dates = None if len(e) < 3 else e[2]
            operand = e
            values, res_dates, keys = self._get_values_from_db(operand, dates,
                                                               date_ranges,
                                                       norm, grouping=grouping)
            if (values, res_dates, keys) == (None, None, None):
                return date_dict, None, errors
            classifiers = self._get_classifiers(keys)
            if classifiers is None:
                self.logger.log_message('smt', 'error',
                                        'Bad/missing classifier(s):' + \
                                        ' dropping %s' % expression)
                return date_dict, classifiers, errors
            for range in date_ranges:
                if range not in date_dict:
                    date_dict[range] = defaultdict(list)
                data_pile = date_dict[range]
                ddict = defaultdict(list)
                for val in values:
                    val = list(val)
                    val[1] = datetime.date(val[1].year, val[1].month, 
                                           val[1].day)
                    if not isinstance(val[0], float):
                        val[0] = None #only floats allowed
                    if not self._check_date_range(range, val[1]):
                        continue
                    if norm and e[4]: #do we discount?
                        val[0] = self._calculate_NPV(val[0], discount_rate,
                                                     init_date, val[1])
                        val = tuple(val)
                    if len(val) > 2:
                        ddict[tuple(val[2:])].append(val[0])
                    else:
                        ddict[''].append(val[0])
                for key in classifiers:
                    if key not in ddict:
                        if e[0] == EXPR_DTYPE_OP:
                            unit = '%s:%s' % (e[0], e[1])
                        else:
                            unit = '%s:%s' % (e[1][0], e[1][1])
                    val = ddict[key]
                    data_pile[key].append(numpy.asarray(val, dtype=numpy.float))
        return date_dict, classifiers, errors

    def _get_results(self, date_dict, grouping, classifiers, expr, proportion):
        """Evaluate the expression and return a list of results and the header
        for the result grouping"""
        ddp = self.default_decimal_places
        res_input = [[l+':'+i for i in v] for l, v in grouping.items()]
        res_grouping = []
        for item in res_input:
            res_grouping += item
        ranges = sorted(date_dict.keys(), key=operator.itemgetter(0))
        res_grp_header = list(res_grouping)

        results = [res_grouping]
        flt_str = '%%.%df' % ddp
        for range in ranges:
            results[0].append(self._get_string_range(range))
            total = 0.0
            for key in classifiers:
                val = date_dict[range][key]
                val = self.evaluator.evaluate(None, None, expr, None, val, 1)
                date_dict[range][key] = val
                if proportion and not numpy.isnan(val):
                    total += val
            row = 1
            for key in classifiers:
                val = date_dict[range][key]
                if proportion and not numpy.isnan(val):
                    val /= total
                try:
                    if numpy.isnan(val):
                        flt_text = 'NaN'
                    else:
                        flt_text = flt_str % round(val, ddp)
                    if len(results) > row:
                        results[row].append(flt_text)
                    else:
                        results.append([str(i) for i in key]+\
                                       [flt_text])
                    row += 1
                except Exception, e:
                    self.logger.log_message('aggregation', 'error',
                                    'Something went very, very wrong. '\
                                    'Check your syntax'+'\n'+str(e))
                    return
        return results, res_grp_header

    def _check_date_range(self, range, val):
        """Check if the val is inside the given range"""
        if range[0] > val: #date below range
            return False
        if range[3]: #is last range (is range end ok)
            if val > range[1]:
                return False
        elif val >= range[1]:
            return False
        return True

    def _get_date_ranges(self, init_date, all_dates, expr, time_unit,
                        time_length):
        """
        This gets the date ranges to be used for the data
        In expression parser -1 is mapped to None, i.e. None indicates the
        last date in data
        """
        dtypes = (EXPR_DTYPE_COMB, EXPR_DTYPE_DATA, EXPR_DTYPE_OP)
        all_dates = list(all_dates)
        all_dates.sort()
        last_date = init_date
        for e in expr:
            if e[0] not in dtypes:
                continue
            if e[2][1] is None:
                last_date = all_dates[-1]
                break
            if e[2][1] > last_date:
                last_date = e[2][1]
        first_date = last_date
        for e in expr:
            if e[0] not in dtypes:
                continue
            if e[2][0] is None:
                first_date = all_dates[-1]
                break
            if e[2][0] < first_date:
                first_date = e[2][0]
        date_ranges = self._build_ranges(all_dates, time_unit,
                             time_length, first_date, last_date)
        return date_ranges

    def _build_ranges(self, dates, time_unit, time_length, first=None,
                         last=None):
        """Returns a list of tuples of
        (start_date, end_date, time_unit, last_range)"""
        dates.sort()
        start_date = dates[0]
        last_date = dates[-1] or dates[0]
        if last is not None and last < last_date:
            last_date = last
        if first is not None and first > start_date:
            start_date = first

        if time_unit == 'year':
            inc = lambda d: d + relativedelta(years=+time_length)
        elif time_unit == 'month':
            inc = lambda d: d + relativedelta(months=+time_length)
        elif time_unit == 'day':
            inc = lambda d: d + relativedelta(days=+time_length)
        date_ranges = []
        while (last_date-start_date).days > 0:
            if (inc(start_date)-last_date).days < 0:
                date_ranges.append((start_date, inc(start_date), time_unit,
                                    False))
            else:
                date_ranges.append((start_date, last_date, time_unit, True))
            start_date = inc(start_date)
        if len(date_ranges) == 0:
            date_ranges.append((start_date, last_date, time_unit, True))
        return date_ranges

    def _get_string_range(self, rng):
        """Turns the given range tuple into a string representation"""
        fmt = '%d/%m/%y'
        start = rng[0].strftime(fmt)
        if not rng[3]:
            end_date = rng[1] + relativedelta(days=-1)
        else:
            end_date = rng[1]
        end = end_date.strftime(fmt)
        return  start + '-' + end

    def _unpack_weighted_average(self, data):
        """Unpacks a weighted average into a formula the expression parser
        can handle."""
        data = data.split('wavg[', 1)
        if len(data) == 1:
            return data[0]
        else:
            start, parts = data
        time, parts = parts.split(']', 1)
        parts = parts.split('(', 1)[1]
        count = 1
        i = 0
        while True:
            item = parts[i]
            if item == '(':
                count += 1
            elif item == ')':
                count -= 1
            if count == 0:
                break
            i += 1
            if i == len(parts):
                raise Exception('Bracket mismatch')
        if i == 0:
            raise Exception('No data given for weighted average')
        vars, parts = parts[:i], parts[i+1:]
        vars = vars.split(',')
        if len(vars) != 2:
            raise Exception('too many commas (%d) in weighted average' % len(vars))
        first, second = vars
        first = first.strip()
        second = second.strip()
        return start+'sum['+time+']('+first+'*'+second+\
                ')/sum['+time+']('+second+')'+\
                self._unpack_weighted_average(parts)

    def _build_constraints(self, constraints):
        result = []
        varcount = 0
        i = 0
        while i < len(constraints):
            var = constraints[i][2]
            val = constraints[i+1][1]
            op = constraints[i+2][2]
            varcount += 1
            if varcount > 1 and len(constraints) > i+3:
                andor = constraints[i+3][2]
                result[-1].append(andor)
                i += 1
            result.append([var, op, val])
            i += 3
        return result


    def _get_values_from_db(self, operand, dates, date_ranges, norm, 
                            opresconst=None, grouping=None):
        """Get value(s) for a single operand from data or operation result
        database
        
        TODO: the constraints should be handled in a similar fashion
        to simo.optimization.tools.omatrix or simo.output.expression
        
        NB! valid grouping pairs (in either way):
        level-level
        operation-"comp_unit"
        level-child
        level-parent
        """
        db = self.datadb
        group_header = {}
        grp_levels=0
        keys = []
        constraints = None
        opgroup_by = []
        if len(operand) >= 4 and operand[3] is not None:
            constraints = self._build_constraints(operand[3])
        if operand[0] == EXPR_DTYPE_OP:
            header = ['id', operand[1], 'op_date']
            if grouping is not None:
                for lev in grouping:
                    if lev == 'operation':
                        keys += [':'.join((lev,h)) for h in \
                                 grouping['operation']]
                        opgroup_by = grouping['operation']
                    else:
                        if lev not in group_header:
                            grp_levels+=1
                            group_header[lev] = ['id', 'data_date']
                        group_header[lev] += grouping[lev]
            if operand[1] is not None:
                required = [operand[1]]
            else:
                required = None
            if opresconst is not None:
                constvars = [c[1] for c in opresconst if c[0] == 'operation']
                header += constvars
            data = db.get_data_from_level('op_res', header, 0,
                                          dates=dates, required=required,
                                          constraints=constraints,
                                          group_by_headers=opgroup_by)
            data = [list(l) for l in data]
            for lev, header in group_header.items():
                keys += [':'.join((lev,h)) for h in header[2:]]
                data2 = db.get_data_from_level('data', header,
                                               lev, dates=dates,
                                               constraints=constraints)
                data2 = set([d for d in data2])
                # sort data2 by date and reverse it
                data2 = sorted(data2, key=lambda row: row[1])
                data2.reverse()                
                for range in date_ranges:
                    for row in data:
                        if not self._check_date_range(range, row[2]):
                            continue
                        for row2 in data2:
                            if not self._check_date_range(range, row2[1]):
                                continue
                            if row[0] == row2[0]:
                                # in case single range contains records at
                                # multiple dates, we should take the key
                                # values from the last record (which should be
                                # the first item in reversed data2)
                                if len(row[3:]) >= len(keys):
                                    continue
                                row += row2[2:]
            data = [d[1:] for d in data]
        elif operand[0] == EXPR_DTYPE_DATA:
            # the varname location is different for normal operands and
            # operand constrain operands (the latter lacks level name)
            header = [operand[1][1]] if norm else [operand[1]]
            header.append('data_date')
            level = operand[1][0] if norm else operand[0]
            if db.has_parent(level):
                header = ['id', 'pid'] + header
            else:
                header = ['id']
            if grouping is not None:
                for lev in grouping:
                    if lev == level:
                        header += grouping[level]
                        keys += [':'.join((lev,h)) for h in grouping[level]]
                    else:
                        grp_levels+=1
                        if lev not in group_header:
                            if lev != 'operation':
                                if db.is_parent_level:
                                    group_header[lev] = ['id']
                                elif db.is_child_level:
                                    group_header[lev] = ['pid']
                                else:
                                    msg = "Grouping level '%s' is not "\
                                          "directly related to main level "\
                                          "'%s'" % (lev, level)
                                    self.logging.log_message('aggregation',
                                                             'error', msg)
                            else:
                                group_header[lev] = ['id']
                        group_header[lev] += grouping[lev]
            data = db.get_data_from_level('data', header,
                                          level, dates=dates,
                                          constraints=constraints)
            data = [list(l) for l in data]
            for lev, header in group_header.items():
                keys += [':'.join((lev,h)) for h in header[1:]]
                if lev is not 'operation':
                    # TODO: fix the constraints
                    data2 = db.get_data_from_level('data', header, lev,
                                                   dates=dates,
                                                   constraints=constraints)
                    data2 = set([d for d in data2])
                    if db.is_parent(level, lev):
                        ind = 1
                    else:
                        ind = 0
                else:
                    # TODO: fix the constraints
                    data2 = db.get_data_from_level('op_res', header,
                                                   None, dates=dates,
                                                   constraints=constraints)
                    data2 = set([d for d in data2])
                    ind = 0
                for row in data:
                    for row2 in data2:
                        if row[ind] == row2[0]:
                            row += row2[1:]
            if db.has_parent(level):
                data = [d[2:] for d in data]
            else:
                data = [d[1:] for d in data]
        else:
            # TODO: the constraints should be handled in a similar fashion
            # to simo.optimization.tools.omatrix or simo.output.expression
            self.logger.log_message('aggr', 'error', 'conditional '\
                                    'expressions not supported in '\
                                    'aggregation output!')
            return None, None, None
        if grp_levels > 1:
            msg = "Multiple non-self grouping levels! "\
                  "Results may be corrupted"
            self.logging.log_message('aggregation', 'warning', msg)
        dates = set([])
        result = []
        for i in data:
            dates.add(i[1])
            result.append(i)
        return result, list(dates), keys

    def _calculate_NPV(self, value, discountrate, presentdate, eventdate):
        """Calculate the discounted NPV (Net Present Value) for a single value

        NB! what if discounting period includes months???
        """
        diff = eventdate.year - presentdate.year
        return value / (1 + (discountrate / 100)) ** diff
