#!/usr/bin/env python
# encoding: utf-8
"""
Title: SIMO branching output

Module for data output in branching .dots

*branching.py*

Copyright (c) 2009 Simosol Oy. Distributed under the GNU General Public License 2.0.
"""

import os
from collections import defaultdict
from output import Output
import operator


class OutputBranching(Output):
    
    def run(self):
        self.branching(self.output_filename)
        
    def branching(self, path):
        # TODO: if <save_steps_to_db> in simulation XML is set to 'last', 
        # result database contains only the last date and thus all dates 
        # before the last are missing, and in that case branching graph 
        # output does not really work
        dates_by_id_iter = self.datadb.get_dates_by_id_and_iter(self.main_level)
        for id_iter, dates in dates_by_id_iter.items():
            id, iter = id_iter
            result = self._get_branching_strings(dates, id, iter)
            self._write_file(os.path.join(path, '%s_%d.dot' % (id, iter)), 
                              result, True)
        
    def _get_branching_strings(self, dates, id, iteration):
        dates.sort()
        branching_info = self.datadb.get_branching_info(id, iteration)
        branching_info = sorted(branching_info, key=operator.itemgetter(3))
        groupings = []
        for date in dates:
            groupings.append([date, (0,0)])
        for branch in branching_info:
            try:
                date_ind = dates.index(branch[4])
                for date in groupings[date_ind:]:
                    date.append((branch[2], branch[3]))
            except ValueError:
                errorstr = "cannot find branching date %s from data dates" % \
                           branch[4].strftime('%Y%m%d')
                self.logger.log_message('output', 'error', errorstr)
        result =  [['digraph branches {',
                   'nodesep=0.01;',
                   'ranksep=0.01;',
                   'size="8.27,11.69";',
                   'ratio=fill;']]
        fmt = self._format_date
        result.append(['{', 
                       'node [shape=plaintext fontsize=14 height=.01];', 
                       '',
                       ' -> '.join([fmt(d) for d in dates])+ \
                       ';', '}'])
        branch_roots = ['{', 'node [shape=doublecircle];', '']
        branch_roots.append(self._get_branch_root_str(dates[0], 0))
        for branch in branching_info:                     #date, group
            branch_roots.append(self._get_branch_root_str(branch[4], 
                                                          branch[3]))
        branch_roots.append('}')
        result.append(branch_roots)
        node_points = ['{', 'node [shape=point]', '']
        node_point_groups = []
        for item in groupings:
            node_points.extend(self._get_branch_node_strs(item))
        node_points.append('}')        
        result.append(node_points)
        rankings = []
        for item in groupings:
            rankings.append(self._get_branch_ranking_str(item))
        result.append(rankings)
        links = []
        branch_points = defaultdict(list)
        for branch in branching_info:               #date, from, to, name, bgroup
            links.append(self._get_branch_branch_str(branch[4], branch[2], 
                                                     branch[3], branch[7]))
            branch_points[branch[4]].append(branch[3])
        groups = set([0])
        last_date = None
        for date in dates:
            if last_date is not None:
                if last_date in branch_points.keys():
                    for branch in branch_points[last_date]:
                        groups.add(branch)
                for group in groups:
                    links.append(self._get_branch_link_str(last_date, date,
                                                           group))
            last_date = date
        result.append(links)
        result.append(['}'])
        return result
    
    def _get_branch_root_str(self, date, item):
        branch_str = 'b_%(item)d_%(date)s [label="%(item)d" ' + \
                     'fontcolor=red group="%(item)d"];' 
        return branch_str % {'item':item, 'date':self._format_date(date)}
    
    def _get_branch_node_strs(self, item):
        date = item[0]
        groups = [i[1] for i in item[1:]]
        str_base = 'b_%(item)d_%(date)s [group="%(item)d"];'
        result = []
        for item in groups:
            result.append(str_base % {'item':item, 
                                      'date':self._format_date(date)})
        return result
               
    def _get_branch_ranking_str(self, item):
        date = item[0]
        groups = [i[1] for i in item[1:]]
        rank_str = '{ rank = same; %(date)s; %(nodes)s }'
        nodes = ' '.join(['b_%d_%s;' % (g, self._format_date(date)) \
                          for g in groups])
        return rank_str % {'date':self._format_date(date), 'nodes':nodes}
    
    def _get_branch_branch_str(self, date, group1, group2, name):
        link = 'b_%(G1)d_%(D)s -> b_%(G2)d_%(D)s ' % \
               {'D':self._format_date(date), 'G1':group1, 'G2':group2}
        link += '[ label="%s" fontsize=14 fontcolor=red];' % (name,)
        return link
    
    def _get_branch_link_str(self, date1, date2, group):
        return 'b_%(G)d_%(D1)s -> b_%(G)d_%(D2)s;' % \
               {'D1':self._format_date(date1), 'D2':self._format_date(date2),
                'G':group}
    
    def _format_date(self, date):
        if self.date_res == 0:
            return date.strftime('%Y')
        else:
            return date.strftime('%Y%m%d')

def main():
    pass

if __name__ == "__main__":
    main()