#!/usr/bin/env python
# encoding: utf-8
"""
Title: SIMO by-level output

Module for data output in hierarchical "data in columns" format, with a fixed
width good for spreadsheet programs.

*by_level.py*

Copyright (c) 2009 Simosol Oy.
Distributed under the GNU General Public License 2.0.
"""
from output import Output

class OutputByLevel(Output):

    def run(self):
        constraints = None
        if self.id_limits is not None:
            constraints = [('id', 'in', self.id_limits)]
        self.by_level(self.output_filename, constraints,
                      dates=self.dates, date_res=self.date_res)

    def by_level(self, file, constraints=None, dates=None, date_res=0):
        self.pads = None
        self.write_started = False
        self.rowcount = 0
        strings = self._get_strings_wide(None, constraints, dates, date_res)
        self._write_file(file, strings)

    def _get_wide_header(self, level=1, headers=None):
        if headers == None:
            headers = [['sim unit', 'branch', 'iteration',
                        'data level', 'id', 'date']]
        if not level in self.output_level_constraint:
            level_const = []
        else:
            level_const = self._get_level_constraints(level)
        level_headers = []
        if level_const is not None:
            for item in level_const:
                level_headers.append(item[0])
        elif self.output_constraint is None:
            level_headers = self.datadb.get_level_var_headers(level)
        headers.append(level_headers)
        ch_levels = self.datadb.get_child_levels(level)
        if ch_levels is not None:
            for ch_level in ch_levels:
                self._get_wide_header(ch_level, headers)
        return headers


    def _flatten_header(self, header):
        flat_header = []
        for group in header:
            for item in group:
                flat_header.append(item)
        return flat_header


    def _get_strings_wide(self, wide_header=None, constraints=None, dates=None,
                          date_res=0, level=1, sim_unit=None):
        if wide_header is None:
            wide_header = self._get_wide_header()
        flat_header = self._flatten_header(wide_header)
        items = self._get_strings_wide_items(wide_header, constraints,
                                             dates, date_res, level, sim_unit)
        if not self.write_started:
            items = [flat_header]+items
        if self.result_padding:
            items = self._pad_all(items)
        for i in range(len(items)):
            items[i] = self.sep.join(items[i])
        return items


    def _get_strings_wide_items(self, wide_header=None, constraints=None,
                            dates=None, date_res=0, level=1, sim_unit=None,
                            core=None):
        """
        wide_header - header to be used (method will get one if None)
        constraints is a dictionary that gives sql constraints as
        key op value(s). op is the first param of the value list.
        if op == in, the list can have additional params.
        dates is a tuple with (first_date, last_date)
        date_res is 0 or 1
        date_res = 0: 2009
        date_res = 1: 2009-03-26
        level - where are we in the hierarchy? level 1 == base (for recursion)
        sim_unit - current simulation unit id (for recursion)
        core - current date, iteration, branch and data_id (for recursion)
        """
        if wide_header is None:
            wide_header = self._get_wide_header()
        if constraints is None:
            constraints = []
        results = []
        write_level = level in self.output_level_constraint
        if not write_level:
            level_const = []
        else:
            level_const = self._get_level_constraints(level)
        headers = self._get_level_headers(level, level_const)
        if level==1 or not core:
            headers = ['data_id'] + headers
            rsort = ('id', 'branch', 'iteration', 'data_date')
            # ID, branch and iteration are required values to find children.
            branch_ind = headers.index('branch')
            iter_ind = headers.index('iteration')
            # Date must be set to the proper resolution
            date_ind = headers.index('data_date')
            no_join = False
        else:
            headers.pop(headers.index('branch'))
            headers.pop(headers.index('iteration'))
            headers.pop(headers.index('data_date'))
            rsort = ('id',)
            data_id = core[0]
            iterr = core[1]
            branch = core[2]
            date = core[3]
            no_join = True
        id_ind = headers.index('id')

        levname = self._get_level_name(level)
        iter = self.datadb.get_data_from_level('data', headers,
                                               level, constraints, dates,
                                               rsort=rsort, no_join=no_join)
        flat_header = self._flatten_header(wide_header)
        
        # sim unit and data level are not inherent to the database, so we need
        # to add them here (as well as to the resulting data)
        for item in iter:
            if write_level:
                self.rowcount += 1
            item = list(item)
            id = item[id_ind]
            if not core:
                data_id = item[0]
                branch = item[branch_ind]
                date = item[date_ind]
                iterr = item[iter_ind]
            if date_res == 0:
                ddate = date.year
            else:
                ddate = date
            if sim_unit is None:
                sim_id = id
            else:
                sim_id = sim_unit
            if write_level:
                if not core:
                    pops = sorted((0, id_ind, branch_ind, date_ind, iter_ind), 
                                  reverse=True)
                    for p in pops:
                        item.pop(p)
                else:
                    item.pop(id_ind)
                # sim_id, branch, iteration, level, id, date, <others>
                item = [sim_id, branch, iterr, levname] \
                        + [item[0]] + [ddate] + item[1:]
                item = self._set_wide_vars(item, level, wide_header)
                result = list(item)
                self._listflt2str(result, self.default_decimal_places)
                results.append(result)

            # Activate the following lines if you want to create a single
            # recursive level hierarchy
            # Get the children
            child_levs = self.datadb.get_all_child_levels(level)
            if child_levs is not None:
                if self.output_constraint is not None:
                    child_levs = set(child_levs) & self.output_level_constraint
                if len(child_levs) == 0:
                    continue
                else:
                    child_levs = set(self.datadb.get_child_levels(level))
                children = self.datadb.get_child_ids(level, id, iterr, branch,
                                                     date, child_levs, 
                                                     data_id=data_id)
                if children is not None and self._has_data(children):
                    child_levs = list(child_levs)
                    child_levs.sort()
                    for child_lev in child_levs:
                        new_constraints = self._copy_constraints(constraints)
                        new_constraints += [('id', 'in', children[child_lev],
                                             'and'), ('pid', 'eq', id, 'and'),
                                            ('data_id', 'eq', data_id),]
                        results += self._get_strings_wide_items(wide_header,
                                                          new_constraints,
                                                          dates, date_res,
                                                          child_lev, sim_id,
                                                          core=(data_id,
                                                                iterr,
                                                                branch,
                                                                date))
                    append_header = True
            if level == 1 and self.rowcount > self.flush_limit:
                self._clear_queue(results, flat_header)
                results = []
                self.rowcount = 0
        return results

    def _clear_queue(self, results, header):
        if not self.write_started:
            results = [header]+results
        if self.result_padding:
            results = self._pad_all(results)
        for i in range(len(results)):
            results[i] = self.sep.join(results[i])
        self._flush(results)

    def _set_wide_vars(self, item, level, wide_header):
        base = item[0:len(wide_header[0])]
        vars = item[len(wide_header[0]):]
        result = base
        for lv in range(1, len(wide_header)):
            if lv == level:
                result += vars
            else:
                result += ['']*len(wide_header[lv])
        return result
