#!/usr/bin/env python
# encoding: utf-8
"""
SIMO expression output based on output matrix

...

*expression.py*

Copyright (c) 2009 Simosol Oy. Distributed under the GNU General Public License 2.0.
"""

import sys
from output import Output
import datetime
from datetime import timedelta
from calendar import monthrange
from pysqlite2.dbapi2 import ProgrammingError, OperationalError
from collections import defaultdict
import operator
import numpy
import math

from simo.utils.utils import calculate_NPV
from simo.builder.names import EXPR_DISCOUNT, EXPR_DTYPE_COMB,\
                               EXPR_DTYPE_DATA, EXPR_DTYPE_OP,\
                               EXPR_DTYPE_VALUE, EXPR_OPERATION,\
                               EXPR_DTYPE_AGGR

isnan = numpy.isnan
MAX_STR_SIZE = 16  # maximum length for object ids
DTYPE_FLOAT = numpy.float32


class OutputExpr(Output):
    """Class for expression output, i.e. output variables defined with a
    syntax similar to optimisation task definitions
    """
    def run(self):
        self.expression(self.output_filename)

    def expression(self, file):
        self.pads = None
        self.write_started = False
        result = self._run_expr(self.expression_def.data)
        if result is not None:
            self._write_file(file, result)

    def _run_expr(self, expr_dict):
        self._discount_rate = expr_dict['discount_rate']
        self._init_date = expr_dict['initial_date']
        self._expressions = []
        self._data_levels = []
        self._labels = []
        self._num_of_branches = defaultdict(list)
        for e in expr_dict['expression_variables']:
            expr, err = self.expr_parser.parse(e['expression'],
                                               self._init_date, 'year')
            if err is not None:
                msg = "Could not parse expression '%s': %s" % \
                      (e['expression'], err)
                self.logger.log_message('output', 'error', msg)
                continue
            self._expressions.append(expr)
            self._labels.append(e['label'])
        if not self._analyse_data():
            msg = "Could not analyze expression data - aborting output"
            self.logger.log_message('output', 'error', msg)
            return None
        if not self._create_tables():
            msg = "Could not creating expression data arrays - aborting output"
            self.logger.log_message('output', 'error', msg)
            return None
        if not self._fill_tables():
            msg = "Could not fill expression data arrays - aborting output"
            self.logger.log_message('output', 'error', msg)
            return None
        strings = self._construct_strings()
        del(self.data)
        return strings

    def _analyse_data(self):
        """Go through expressions and input database and
        analyse some data properties
        """
        exprs = self._expressions
        self._latest_date = self._init_date
        self._earliest_date = None
        self._process_exprs(exprs)
        # get simulation unit ids, the number of simulation units and the
        # maximum number of iterations, branches and time steps from data
        self.unit_ids = set([])
        self.unit_levels = {}
        # auxiliary levels are needed when expressions contain only operation
        # results only_ops dummy indicates this type of situation
        aux_levels = []
        only_ops = True
        for level in numpy.unique(self._data_levels):
            if level == 'op_res':
                dtable = 'op_res'
                op_levels = [key[0] for key in self.datadb.content_def.keys()]
                ids = set([])
                for op_lev in op_levels:
                    cur_ids = self.datadb.get_ids(dtable, op_lev)
                    if len(cur_ids) > 0:
                        aux_levels.append(op_lev)
                    ids.update(cur_ids)
                if len(aux_levels) != 1:
                    errorstr = "Operation results should be "\
                               "on a single level only!"
                    self.logger.log_message('output', 'error', errorstr)
                    return False
                # substitute op_res level with the actual level name
                level = aux_levels[0]
            else:
                dtable = 'data'
                ids = self.datadb.get_ids(dtable, level)
                only_ops = False
            self.unit_ids.update(ids)
            for uid in ids:
                self.unit_levels[uid] = level
        # if expressions contain only operation results, use first auxiliary
        # level for deducing iteration and branch counts
        if only_ops:
            dtable = 'op_res'
            level = aux_levels[0]
        else:
            # remove 'op_res' levels from data levels
            while 'op_res' in self._data_levels:
                self._data_levels.remove('op_res')
            dtable = 'data'
            level = self._data_levels[0]
        if len(self.unit_ids) == 0:
            errorstr = "Empty %s database, nothing to output" % dtable
            self.logger.log_message('output', 'error', errorstr)
            return False
        self.max_iterations = self.datadb.get_max_iter(dtable)
        if self.max_iterations is None:
            errorstr = "Maximum number of iterations couldn't be deduced "\
                       "from %s database contents, database is probably "\
                       "empty" % dtable
            self.logger.log_message('output', 'error', errorstr)
            return False
        self._max_branches = self.datadb.get_max_branch(dtable)
        if self._max_branches is None:
            errorstr = "Maximum number of branches couldn't be deduced "\
                       "from %s database contents, database is probably "\
                       "empty" % dtable
            self.logger.log_message('output', 'error', errorstr)
            return False

        # the actual number of iterations and branches is max + 1
        self.max_iterations += 1
        self._max_branches += 1
        self.num_of_units = len(self.unit_ids)
        self.unit_index = dict([(v,k) for k,v in enumerate(self.unit_ids)])
        self._unit_range = numpy.arange(self.num_of_units, dtype=int)

        # find all branches of each object in each iteration
        for level in self._data_levels + aux_levels:
            if level != 'op_res':
                dtable = 'data'
                header = ['iteration', 'id', 'branch']
                data = self.datadb.get_data_from_level(dtable, header, level)
                for item in data:
                    it, uid, br = item
                    if br not in self._num_of_branches[(it, uid)]:
                        self._num_of_branches[(it, uid)].append(br)

        return True

    def _process_exprs(self, exprs):
        """Go through the expressions and try to deduce data levels and
        earliest and latest dates
        """
        self._date_ranges = {}
        self._date_locs = {}
        self._operand_index = {}
        self._unique_operands = {}
        self._one_aggr_level = {}
        modexprs = []
        for expr in exprs:
            modexprs.append([])
            level = None

            # TODO: operands inside a single aggregation block should be
            # possible to link to each other as all objects inside an
            # aggregation should have equal number of time steps
            aggr_stack = []
            for e in expr:
                if e[0] == EXPR_DTYPE_DATA:
                    if level is None:
                        level = e[1][0]
                    elif e[1][0] != level:
                        errorstr = "Multiple levels in single expression "\
                                   "output item. Check the expression "\
                                   "output definition"
                        self.logger.log_message('output', 'error', errorstr)
                elif e[0] == EXPR_DTYPE_AGGR:
                    if len(aggr_stack) > 0:
                        dlevel = aggr_stack[0][0]
                        all_same_type = True
                        for dl, ui in aggr_stack[1:]:
                            if dl != dlevel:
                                all_same_type = False
                            if not dlevel and dl:
                                dlevel = dl
                        for dl, ui in aggr_stack:
                            item = all_same_type, dlevel
                            if ui in self._one_aggr_level:
                                if not all_same_type:
                                    self._one_aggr_level[ui] = item
                            else:
                                self._one_aggr_level[ui] = item
                    aggr_stack = []
                    modexprs[-1].append(e)
                    continue
                elif e[0] not in (EXPR_DTYPE_COMB, EXPR_DTYPE_OP):
                    modexprs[-1].append(e)
                    continue
                start, end = e[2]
                
                if not self._earliest_date or \
                    (start is not None and start < self._earliest_date):
                    self._earliest_date = start
                if self._latest_date and (not end or end > self._latest_date):
                    self._latest_date = end

                # unique operand lacks date info
                uop = e[:2] + e[3:5]

                # update unique operand index
                if uop not in self._unique_operands:
                    ui = len(self._unique_operands)
                    self._unique_operands[uop] = ui
                    smallest, biggest = start, end
                else:
                    ui = self._unique_operands[uop]
                    smallest, biggest = self._date_ranges[ui]

                # get a key (operand indice) for mapping current operand to
                # unique operand and do the mapping
                oi = len(self._operand_index)
                self._operand_index[oi] = ui

                # add the current operand indice to aggregation stack
                dlevel = None if e[0] == EXPR_DTYPE_OP else level
                aggr_stack.append((dlevel, ui))

                # update also date ranges, the aim is to find the maximum and
                # minimum date limits for each unique operand
                if not smallest and start:
                    smallest = start
                elif smallest and start:
                    if start < smallest:
                        smallest = start
                if not end:
                    biggest = end
                elif end and biggest:
                    if end > biggest:
                        biggest = end
                self._date_ranges[ui] = (smallest, biggest)

                modexprs[-1].append(e + ((oi, ui),))

            if level is None:
                self._data_levels.append('op_res')
            else:
                self._data_levels.append(level)

        # store the modified expressions
        self._expressions = modexprs

    def _create_tables(self):
        """Create expression data tables, using Numpy arrays
        """
        nunits = self.num_of_units
        maxit = self.max_iterations
        maxbr = self._max_branches
        num_of_exprs = len(self._expressions)

        # expression output data container
        self.data = {}

        # define expression matrix shape and construct empty matrice
        #   (iterations,
        #    units,
        #    branches,
        #    number of expressions,
        tshape = (maxit, nunits, maxbr, num_of_exprs)
        self.data['expressions'] = numpy.zeros(tshape, dtype=DTYPE_FLOAT)

        return True

    def _update_date_locs(self, expr):
        """Update date location index for current expression

        expr -- postfix epxression in a list
        """
        dtypes = (EXPR_DTYPE_COMB, EXPR_DTYPE_DATA, EXPR_DTYPE_OP)
        for operand in expr:
            if operand[0] not in dtypes:
                continue
            oi, ui = operand[-1]
            start, end = operand[2]
            if start is None:
                start = self._dates[ui][0]
            if end is None:
                end = self._dates[ui][-1]
            i_s = self._get_date_ind(ui, start, 'start')
            i_e = self._get_date_ind(ui, end, 'end')
            self._date_locs[oi] = (i_s, i_e)

    def _get_date_ind(self, ui, d, date_type='start'):
        """Get correct date indice, find the closest date in case date is not
        defined in date index
        """
        if d not in self._date_index[ui]:
            candidate = None
            delta = sys.maxint
            for cd, ind in self._date_index[ui].items():
                if abs(d - cd).days < delta:
                    if date_type == 'start' and cd > d:
                        candidate = cd
                        delta = abs(d - cd).days
                    elif date_type == 'end' and cd < d:
                        candidate = cd
                        delta = abs(d - cd).days
            ind = None
            if candidate is not None:
                ind = self._date_index[ui][candidate]
        else:
            ind = self._date_index[ui][d]
        return ind

    def _fill_tables(self):
        """Fill expression data tables with values from input databases
        """
        expr_array = self.data['expressions']
        for it in range(self.max_iterations):
            # reset operand data container, date index and date arrays
            self.data['operands'] = {}
            self._date_index = {}
            self._dates = {}
            
            # first collect values for all unique operands
            for operand, ui in self._unique_operands.items():
                # get start and end dates for the current unique operand
                start, end = self._date_ranges[ui]
                # get data tables
                only_one_level, level = self._one_aggr_level[ui]
                if only_one_level:
                    if operand[0] == EXPR_DTYPE_OP:
                        dtables = [('op_res', None)]
                    elif operand[0] == EXPR_DTYPE_DATA:
                        dtables = [('data', operand[1][0])]
                    else:
                        dtables = [('op_res', None), ('data', level)]
                else:
                    dtables = [('op_res', None), ('data', level)]

                # get dates for the current operand
                if not start:
                    start = self._earliest_date
                dd = []
                for dtable, level in dtables:
                    constr = (start, end)
                    if constr == (None, None,):
                        new_dates = self.datadb.get_dates(dtable)
                    else:
                        new_dates = self.datadb.get_dates(dtable, constr)
                    dd += [datetime.date(nd.year, nd.month, nd.day) \
                           for nd in new_dates]
                dd = numpy.unique(dd)
                dd.sort()

                # check that operand has dates (which means that the operand
                # has values in the database also)
                if dd.size == 0:
                    return False

                if start is None:
                    start = dd[0]
                if end is None:
                    end = dd[-1]

                span = (dd >= start) * (dd <= end)
                dates = dd[span]
                ndates = span.sum()
                shape = (self.num_of_units,
                         self._max_branches,
                         ndates)
                self.data['operands'][ui] = numpy.zeros(shape, dtype=DTYPE_FLOAT)

                # construct date index for finding correct matrix locations with
                # date information
                dtuples = [(d,i) for i,d in enumerate(dates)]
                self._date_index[ui] = dict(dtuples)
                self._dates[ui] = dd

                array = self.data['operands'][ui]
                if not self._collect_values(it, operand, array, ui):
                    return False

            # go through expressions and find the date
            # locations of each operand
            for expr in self._expressions:
                self._update_date_locs(expr)

            # evaluate the current expression with the values in the temporary
            # matrix and store the results into actual expression value matrix
            evaluate = self.evaluator.evaluate
            rn = self._unit_range
            dlocs = self._date_locs
            data = self.data['operands']
            for i, expr in enumerate(self._expressions):
                for br in range(self._max_branches):
                    values = evaluate(rn, br, expr, dlocs, data, 3, False)
                    expr_array[it,:,br,i] = values

            # delete the temporary expression data table
            del(self.data['operands'])
        return True

    def _collect_values(self, it, operand, array, ui, forced_dates=None,
                        forced_level=None):
        """
        Get values for a single expression operand from either data or
        operation result database

        it -- current simulated iteration indice as int
        operand -- subobjective or constraint operand
        array -- target Numpy array
        ui -- unique subobjective/constraint operand indice
        forced_dates -- forced (end, start) dates (only for constraint
                        variables)
        forced_level -- forced level name, only for condition operands
        """
        evalc = self._eval_constraint
        getv = self._get_values_from_db
        setv = self._set_value

        # check if the expression item is normal (length = 4) or operand
        # constraint (length = 2) operand
        norm = len(operand) == 4

        # if forced_dates are defined (only in the case that the values
        # are collected for constraint variables), no need to get the
        # dates otherwise, get the date definition from the operand
        if forced_dates is None:
            dates = self._date_ranges[ui]
        else:
            dates = forced_dates

        # if operand is normal and it has a constraint, evaluate the
        # operand constraint
        if norm and operand[2] is not None:
            # operand has a constraint: process the constraints and
            # fill the data matrix for objects that passed
            ok = evalc(it, dates, array, operand, ui, norm)
            if not ok:
                return False
        else:
            # get values from database and fill the target matrix with
            # the data values (without any constraints or such)
            data = getv(it, operand, dates, norm, None, forced_level)
            if dates == (None, None):
                # when start and end date are Nones, the value should be from
                # the last recorded date of each iteration-branch-object
                # combination ([-1,-1] syntax in expr definition)
                data = self._filter_data_by_last_date(data)
                    
            for uid, br, date, value in data:
                if isinstance(value, unicode) or isinstance(value, str):
                    errorstr = "No values for '%s' on level '%s' in "\
                                "database" % (value, operand[1][0])
                    self._add_error(errorstr)
                    return False
                setv(array, uid, br, ui, date, value)

        # check if variable is of 'normal' type and if it's value should
        # be discounted
        if norm and operand[3]:
            # discount values to init date
            drate = self._discount_rate
            idate = self._init_date
            for d, j in self._date_index[ui].items():
                mslice = array[...,j]
                array[...,j] = calculate_NPV(mslice, drate, idate, d)

        return True
    
    def _get_values_from_db(self, it, operand, dates, norm, condition=None,
                            forced_level=None):
        """
        Get value(s) for a single operand from data or operation result
        database

        it -- current simulated iteration indice as int
        operand -- current operand definition
        dates -- start and end dates
        norm -- boolean indicating whether the operand is normal
        condition -- operand condition
        forced_level -- forced level, string
        """
        optype = operand[0]
        if optype in (EXPR_DTYPE_OP, EXPR_DTYPE_DATA,):
            constraints = [('iteration', 'eq', it)]
        else:
            constraints = [(EXPR_DTYPE_OP, 'iteration', 'eq', it)]
        const_level = None
        if condition is not None:
            # do a little postfix to prefix conversion for the constraint
            # postfix expression
            prefx, levels = [], set([])
            for op in condition:
                if op[0] == EXPR_DTYPE_COMB:
                    prefx.append(tuple(op[1:3]))
                    if op[1] != EXPR_DTYPE_OP:
                        levels.add(op[1])
                elif op[0] in (EXPR_DTYPE_OP, EXPR_DTYPE_DATA,):
                    prefx.append((op[2],))
                elif op[0] == EXPR_DTYPE_VALUE:
                    prefx.append((op[1],))
                else:
                    b = prefx.pop()
                    a = prefx.pop()
                    eq = op[-1]
                    if eq in ('and', 'or'):
                        prefx += [a + (eq,), b]
                    else:
                        prefx += [a + (eq,) + b]
            constraints[0] = constraints[0] + ('and',)
            constraints += prefx
            # make sure that the constraints do not contain data from
            # multiple levels
            if len(levels) > 1:
                msg = "Multiple data levels in single condition, check the "\
                      "expression ouput syntax!"
                self.logger.log_message('output', 'error', msg)
                return None
            elif len(levels) == 1:
                const_level = list(levels)[0]
            
        required = None
        if optype == EXPR_DTYPE_OP:
            header = ['id', 'branch', 'op_date']
            header.append(operand[1])
            if operand[1] is not None:
                required = [operand[1]]
            data = self.datadb.get_data_from_level('op_res', header, 0,
                                                   constraints=constraints,
                                                   dates=dates,
                                                   required=required)
        elif optype == EXPR_DTYPE_DATA:
            header = ['id', 'branch', 'data_date']
            if forced_level is not None:
                level = forced_level
            else:
                level = operand[1][0]
            # the varname location is different for normal operands and
            # operation constrain operands (the latter lacks level name)
            header.append(operand[1][1] if norm else operand[1])
            data = self.datadb.get_data_from_level('data', header, level,
                                                   constraints=constraints,
                                                   dates=dates)
        elif optype == EXPR_DTYPE_COMB:
            header = [(EXPR_DTYPE_OP, 'id'),
                      (EXPR_DTYPE_OP, 'branch'),
                      (EXPR_DTYPE_OP, 'op_date')]
            header.append((EXPR_DTYPE_OP, operand[1]))
            if operand[1] is not None:
                required = [operand[1]]
            if forced_level is not None:
                level = forced_level
            elif const_level is not None:
                level = const_level
            else:
                level = self._data_levels[self.main_level]
            data = self.datadb.get_data_from_level('combined', header, level,
                                                   constraints=constraints,
                                                   dates=dates)
        return data

    def _eval_constraint(self, it, dates, array, operand, ui, norm):
        """Evaluate operand constraint and store only those
        objects for which the condition evaluated True

        it -- current simulated iteration indice as int
        dates -- start and end dates
        array -- target Numpy array
        operand -- operand structure
        ui -- unique operand indice
        norm -- boolean indicating whether the operand is normal
        """
        condition = operand[2]

        if operand[0] in (EXPR_DTYPE_COMB, EXPR_DTYPE_OP):
            forced_level = None
        else:
            forced_level = operand[1][0]

        # get the data values from database
        data = self._get_values_from_db(it, operand, dates, norm, condition,
                                        forced_level)
        if dates == (None, None,):
            data = self._filter_data_by_last_date(data)
            
        # store the values to the data array
        setv = self._set_value
        for item in data:
            uid, br, date, value = item
            setv(array, uid, br, ui, date, value)

        return True

    def _set_value(self, array, uid, br, ui, date, value):
        """Store a single value into a correct location in a matrix and update
        branch number container

        array -- Numpy array
        uid -- object id, str
        br -- branch, int
        ui -- unique operand indice, int
        date -- datetime object
        value -- value to store, float
        """
        date = datetime.date(date.year, date.month, date.day)
        u = self.unit_index[uid]
        if not date in self._date_index[ui]:
            return  #TODO: error message
        d = self._date_index[ui][date]
        if isnan(array[u,br,d]):
            array[u,br,d] = 0.0
        if value is not None:
            array[u,br,d] += value
    
    def _filter_data_by_last_date(self, data):
        """Filter data so that only the values from the last recorded date of
        each iteration-branch-object combination are selected. This is needed
        when expression dates are defined as [-1:-1]
        
        data -- data generator object from datadb
        """
        temp = {}
        for uid, br, date, value in data:
            key = (uid, br)
            if not key in temp:
                temp[key] = {}
            if not date in temp[key]:
                temp[key][date] = []
            temp[key][date].append(value)
        filtered = []
        for key in temp.keys():
            dates = temp[key].keys()
            dates.sort()
            for value in temp[key][dates[-1]]:
                uid, br = key
                filtered.append((uid, br, dates[-1], value))
        return filtered

    def _construct_strings(self):
        """Construct strings for output
        """
        exprs = [label for label in self._labels]
        hdr = ";".join(['level','unit','iteration','branch'] + exprs)
        strings = [hdr]
        expr_table = self.data['expressions']
        level_items = defaultdict(list)
        for v, k in self.unit_levels.items():
            level_items[k].append(v)
        lord = self.datadb.level_order
        level_names = [self.datadb._level_name[l] for l in lord] + ['op_res']
        for level in level_names:
            units = level_items[level]
            units.sort()
            for unit in units:
                iu = self.unit_index[unit]
                for ii in range(self.max_iterations):
                    branches = self._num_of_branches[(ii,unit)]
                    for ib in sorted(branches):
                        line = "%s;%s;%d;%d" % (level, unit, ii, ib)
                        values = []
                        for ie in range(len(self._expressions)):
                            values.append(expr_table[ii,iu,ib,ie])
                        self._listflt2str(values, self.default_decimal_places)
                        line += ';' + ';'.join(values)
                        strings.append(line)
        return strings


def main():
    pass

if __name__ == '__main__':
    main()
