#!/usr/bin/env python
# encoding: utf-8
"""
Title: SIMO inline output

Module for data output in hierarchical "data in columns" format.

*inline.py*

Copyright (c) 2009 Simosol Oy. Distributed under the GNU General Public License 2.0.
"""

from output import Output

class OutputInline(Output):

    def run(self):
        constraints=None
        if self.id_limits is not None:
            constraints=[('id', 'in', self.id_limits)]
        self.inlined(self.output_filename, constraints,
                     dates=self.dates, date_res=self.date_res)

    def inlined(self, file, constraints=None, dates=None, date_res=0):
        self.write_started = False
        self.pads = None
        self.rowcount = 0
        strings = self._get_strings(constraints, dates, date_res)
        self._write_file(file, strings)

    def _get_strings(self, constraints=None, dates=None, date_res=0, level=1, sim_unit=None):
        self.last_level_header = None
        items = self._get_strings_items(constraints, dates, date_res, level, sim_unit)
        if self.result_padding:
            items = self._pad_all(items)
        for i in range(len(items)):
            items[i] = self.sep.join(items[i])

        return items


    def _get_strings_items(self, constraints=None, dates=None, date_res=0, 
                           level=1, sim_unit=None,
                           core=None):
        """constraints is a dictionary that gives sql constraints as
        key in value, where value is a list
        dates is a tuple with (first_date, last_date)
        date_res is 0 or 1
        date_res = 0: 2009
        date_res = 1: 2009-03-26
        level - where are we in the hierarchy? level 1 == base (for recursion)
        sim_unit - current simulation unit id (for recursion)
        core - current date, iteration, branch and data_id (for recursion)
        """
        if constraints is None:
            constraints = []
        write_level = level in self.output_level_constraint
        if not write_level:
            level_const = []
        else:
            level_const = self._get_level_constraints(level)
        results = []
        headers = self._get_level_headers(level, level_const)
        if level==1 or not core:
            headers = ['data_id']+headers
            rsort = ('id', 'branch', 'iteration', 'data_date')
            # ID, branch and iteration are required values to find children.
            branch_ind = headers.index('branch')
            iter_ind = headers.index('iteration')
            # Date must be set to the proper resolution
            date_ind = headers.index('data_date')
            no_join = False
            new_header = ['sim unit', 'branch', 'iteration', 'data level'] \
                         + headers[4:]
        else:
            headers.pop(headers.index('branch'))
            headers.pop(headers.index('iteration'))
            headers.pop(headers.index('data_date'))
            rsort = ('id',)
            data_id = core[0]
            iterr = core[1]
            branch = core[2]
            date = core[3]
            no_join = True
            if dates:
                no_join = False
            else:
                for c in constraints:
                    if c[0] not in ('id', 'pid', 'data_id'):
                        no_join = False
                        break
            new_header = ['sim unit', 'branch', 'iteration', 'data level'] \
                         + headers
        levname = self._get_level_name(level)
        iter = self.datadb.get_data_from_level('data', headers, level,
                                               constraints, dates,
                                               rsort=rsort, no_join=no_join)
        # ID, branch and iteration are required values to find children.
        id_ind = headers.index('id')
        # sim unit and data level are not inherent to the database, so we need
        # to add them here (as well as to the resulting data)
        append_header = False
        if self.last_level_header != level:
            append_header = True
        for item in iter:
            if write_level:
                self.rowcount += 1
                if append_header:
                    new_header[4] = 'id'
                    new_header[5] = 'date'
                    results.append(new_header)
                    self.last_level_header = level
                append_header = False
            item = list(item)
            id = item[id_ind]
            if not core:
                data_id = item[0]
                branch = item[branch_ind]
                date = item[date_ind]
                iterr = item[iter_ind]
            if sim_unit is None:
                sim_id = id
            else:
                sim_id = sim_unit
            if write_level:
                if date_res == 0:
                    ddate = date.year
                else:
                    ddate = date
                if not core:
                    pops = sorted((0, id_ind, branch_ind, date_ind, iter_ind), 
                                  reverse=True)
                    for p in pops:
                        item.pop(p)
                else:
                    item.pop(id_ind)
                item = [sim_id, branch, iterr, levname] \
                       + [item[0]] + [ddate] + item[1:]
                #[sim_id]+item[:2]+[levname]+item[2:]
                result = list(item)
                self._listflt2str(result, self.default_decimal_places)
                #result = [str(i) for i in item]
                results.append(result)
            # Get the children
            child_levs = self.datadb.get_all_child_levels(level)
            if child_levs is not None:
                if self.output_constraint is not None:
                    child_levs = set(child_levs) & self.output_level_constraint
                if len(child_levs) == 0:
                    continue
                else:
                    child_levs = set(self.datadb.get_child_levels(level))
                children = self.datadb.get_child_ids(level, id, iterr, branch,
                                                     date, child_levs, 
                                                     data_id=data_id)
                if children is not None and self._has_data(children):
                    old_len = len(results)
                    child_levs = list(child_levs)
                    child_levs.sort()
                    for child_lev in child_levs:
                        new_constraints = self._copy_constraints(constraints)
                        new_constraints += [('id', 'in', children[child_lev],
                                             'and'), ('pid', 'eq', id, 'and'),
                                            ('data_id', 'eq', data_id),]
                        results += self._get_strings_items(new_constraints,
                                                           dates, date_res,
                                                           child_lev, sim_id,
                                                           core=(data_id,
                                                                 iterr,
                                                                 branch,
                                                                 date))
                    if len(results) > old_len:
                        append_header = True
            if level == 1 and self.rowcount > self.flush_limit:
                self._clear_queue(results)
                results = []
                self.rowcount = 0
        return results

    def _clear_queue(self, results):
        if self.result_padding:
            results = self._pad_all(results)
        for i in range(len(results)):
            results[i] = self.sep.join(results[i])
        self._flush(results)
