#!/usr/bin/env python
# encoding: utf-8
"""
Title: SIMO operation result output

Module for operation result reports; e.g. volume and income of harvests.

*opres.py*

Copyright (c) 2009 Simosol Oy. Distributed under the GNU General Public License 2.0.
"""

from output import Output
import operator
from collections import defaultdict

class OutputOpres(Output):

    def run(self):
        constraints = None
        if self.id_limits is not None:
            constraints = [('id', 'in', self.id_limits)]
        self.opres(self.output_filename, constraints,
                   dates=self.dates, date_res=self.date_res)

    def opres(self, file, constraints=None, dates=None, date_res=0):
        self.pads = None
        self.write_started = False
        headers = ['op_id','op_level','id','op_date',
                   'iteration','branch','op_group','op_name']
        cfiers = tuple(headers)
        cl_headers = self.datadb.get_opres_def_headers()
        cl_headers = [str(h) for h in cl_headers]
        cl_headers = list(set(cl_headers) & set(self.opres_vars))
        headers += cl_headers
        # Ensure we have data to begin with
        rescount = self.datadb.db.get('SELECT count(*) '\
                                      'FROM %sop_res' % self.datadb._add)[0][0]
        if rescount == 0:
            errormsg = 'No data in op_res database'
            self.logger.log_message('op_res output', 'warning', errormsg)
            return
        # Ensure that we have all the requested headers
        available_headers = self.datadb._get_table_columns('op_res')
        low_h = [h.lower() for h in headers]
        low_ah = [ah.lower() for ah in available_headers]
        for lh, h in zip(low_h, headers):
            if lh in cfiers:
                continue
            if lh not in low_ah:
                errormsg = 'Requested a header (%s) which is not in '\
                           'op_res database' % h
                self.logger.log_message('op_res output', 'error', errormsg)
                return
        
        strings = self._get_opres_strings(headers, constraints, dates, 
                                          date_res)
        self._write_file(file, strings)

    def _get_opres_strings(self, headers, constraints=None, 
                           dates=None, date_res=0):
        date_ind = 3
        if constraints is None:
            constraints = {}
        operation_results = \
            self.datadb.get_data_from_level('op_res', 
                                    headers, 0, constraints, dates, 
                                    rsort=('id', 'iteration', 'branch', 
                                           'op_date', 'op_id'),
                                    required=None, distinct=False,
                                    group_by_headers=self.group_by_vars)
        results = []
        for opres in operation_results:
            opres = list(opres)
            if date_res == 0:
                opres[date_ind] = opres[date_ind].year
            data = opres[1:]
            has_data = False
            for item in data[7:]: # 7 == len(classifiers)
                if item is not None:
                    has_data = True
                    break
            if not has_data:
                continue
            self._listflt2str(data, self.default_decimal_places)
            results.append(data)
        #Clear out the empty headers
        if len(results) > 0:
            empty_heads = [True]*len(results[0])
        else:
            empty_heads = []
        for row in results:
            for i in range(len(row)):
                if row[i] != '':
                    empty_heads[i] = False
        to_remove = []
        for i in range(len(empty_heads)):
            if empty_heads[i]:
                to_remove.append(i)
        results = [headers[1:] + self.group_by_vars]+results
        to_remove.sort(reverse=True)
        for rem in to_remove:
            for i in range(len(results)):
                del results[i][rem]
        
        if self.result_padding:
            results = self._pad_all(results)
        for i in range(len(results)):
            results[i] = self.sep.join(results[i])
        return results
