#!/usr/bin/env python
# encoding: utf-8
"""
Title: SIMO output base class

This class holds the common file writing and text formatting operations for
output

*output.py*

Copyright (c) 2009 Simosol Oy.
Distributed under the GNU General Public License 2.0.
"""

from operator import itemgetter
import operator
import numpy
import os
from simo.builder.optimization.exprparser import ExpressionParser
from simo.optimization.tools.postfixeval import PostfixEvaluator
from simo.output.out_ext import _listflt2str
from simo.output.out_ext import _pad_all

class Output(object):

    sep = ';'
    flush_limit = 300

    def __init__(self, datadb, data_type, id_list, main_level,
                 result_type, output_filename, logger,
                 output_constraint=None, default_decimal_places=1,
                 archiving=False, result_padding=True,
                 aggregation_def=None, expression_def=None,
                 opres_vars=None, group_by_vars=None, full_date=False,
                 dates=None):
        super(Output, self).__init__()
        self.expr_parser = ExpressionParser()
        self.evaluator = PostfixEvaluator()
        self.logger = logger
        self.datadb = datadb
        self.data_type = data_type
        self.id_list = id_list
        self.main_level = main_level
        self.result_type = result_type
        self.output_filename = output_filename
        self.output_constraint = output_constraint
        if self.output_constraint is not None:
            self.output_level_constraint = set(map(itemgetter(1),
                                    self.output_constraint.variables.keys()))
        self.default_decimal_places = default_decimal_places
        self.archiving = archiving
        self.result_padding = result_padding
        if full_date:
            self.date_res = 1
        else:
            self.date_res = 0
        if dates == (None, None):
            self.dates = None
        else:
            self.dates = dates
        self.aggregation_def = aggregation_def
        self.expression_def = expression_def
        self.opres_vars = opres_vars
        self.group_by_vars = group_by_vars
        self.constr_dict = {}
        self.bad_constr = {}
        self.level_hdr_dict = {}
        self.levelname_dict = {}
        self.id_limits = id_list
        self.pads = None
        self.rowcount = 0
        self.write_started = False
        self.run()

    #Override this with the approppriate running sequence in inherited class
    def run(self):
        pass

    def _flush(self, strings, multi_list=False):
        file = self.output_filename
        self._write_file(file, strings, multi_list)
        self.write_started = True

    def _write_file(self, file, strings, multi_list=False):
        mode = 'w'
        if self.write_started:
            mode = 'a+'
        else:
            if not self.bad_constr == {}:
                base = 'ERROR: level "%s" does not have variable(s): "%s"'
                errs = []
                for key, vals in sorted(self.bad_constr.items()):
                    errs.append(base % (key,'", "'.join(vals)))
                if multi_list:
                    strings = [errs] + strings
                else:
                    strings = errs + strings
        try:
            try:
                os.makedirs(os.path.dirname(file))
            except:
                pass
            out = open(file, mode)
            if not multi_list:
                out.writelines([''.join((line.encode('utf-8'), '\n')) \
                                for line in strings])
            else:
                for item in strings:
                    item.append('')
                    out.writelines([''.join((line.encode('utf-8'), '\n')) \
                                    for line in item])
        finally:
            out.close()

    def _get_level_name(self, level):
        try:
            return self.levelname_dict[level]
        except:
            self.levelname_dict[level] = self.datadb._level_name[level]
            return self.levelname_dict[level]

    def _get_level_headers(self, level, level_const=None):
        base_headers = ['branch', 'iteration', 'id', 'oid', 'data_date']
        level_headers = []
        if level_const is not None:
            for item in level_const:
                level_headers.append(item[0])

        return base_headers + level_headers

    def _get_level_constraints(self, level):
        if level in self.constr_dict.keys():
            return self.constr_dict[level]
        if self.output_constraint is None:
            return None
        restricts = []
        for lvl, val in self.output_constraint.variables.items():
            if lvl[1] == level and val is not None:
                restricts.extend(val)
        if restricts == []:
            self.constr_dict[level] = None
            return None
        levname = self.datadb._convert_to_level_name(level)
        bad_headers = self.datadb._check_header(levname,
                                                [r[0] for r in restricts],
                                                'data')
        if len(bad_headers) > 0:
            self.bad_constr[levname] = bad_headers
        self.constr_dict[level] = [r for r in restricts \
                                   if r[0] not in bad_headers]
        return self.constr_dict[level]

    def _copy_constraints(self, constraints):
        result = []
        for constraint in constraints:
            if constraint[0] not in ('id', 'pid', 'branch', 'iteration', 
                                     'data_id', 'data_date', 'op_date'):
                result.append(constraint)
        return result

    def _pad_all(self, rows):
        if self.pads is None:
            self.pads = []
        return _pad_all(self.pads, rows)

    def _listflt2str(self, list, d):
        return _listflt2str(list, d)

    def _has_data(self, level_dict):
        result = False
        for vals in level_dict.values():
            if vals != []:
                result = True
                break
        return result


def main():
    pass

if __name__ == "__main__":
    main()
