#!/usr/bin/env python
# encoding: utf-8
"""
Title: SIMO SMT file format output

Module for data output in SMT file format: values from the last year in
the simulation as whitespace separated fields starting with top-level id and
object id and ending in the year for the values

*smt.py*

Copyright (c) 2009 Simosol Oy. Distributed under the GNU General Public License 2.0.
"""

from output import Output

class OutputSMT(Output):

    sep = ' '

    def run(self):
        constraints=None
        if self.id_limits is not None:
            constraints=[('id', 'in', self.id_limits)]
        self.smt(self.output_filename, constraints, dates=self.dates)

    def smt(self, file, constraints=None, dates=None):
        self.rowcount = 0
        self.pads = None
        strings = self._get_strings(constraints, dates)
        self._write_file(file, strings)

    def _get_strings(self, constraints=None, dates=None, sim_unit=None):
        items = self._get_strings_items(constraints, dates, 1,
                                        sim_unit)
        for i in range(len(items)):
            items[i] = self.sep.join(items[i])

        return items


    def _get_strings_items(self, constraints=None, dates=None,
                           level=1, sim_unit=None):
        """constraints is a dictionary that gives sql constraints as
        key in value, where value is a list
        dates is a tuple with (first_date, last_date)
        
        WARNING: The generated OIDs are not valid for smt output so we
        will replace them. This can result in volatility in the ordering of
        the strata with generated OIDs. The original OIDs are still the same.
        """
        if constraints is None:
            constraints = []
        level_const = self._get_level_constraints(level)
        results = []
        headers = self._get_level_headers(level, level_const)
        levname = self.datadb._level_name[level]
        # TODO: fix the constraints
        iter = self.datadb.get_data_from_level('data', headers, level,
                                               constraints, dates, rsort='oid')
        # create id_dict to get valid oids for the output
        # TODO: fix the constraints
        ids = self.datadb.get_data_from_level('data', ('oid',), level,
                                              constraints, dates, rsort='oid',
                                              distinct=True)
        ids = [i[0] for i in ids]
        count = 0
        for id in ids:
            try:
                int(id)
                count += 1
            except:
                pass
        id_dict = {}
        for id in ids:
            try:
                int(id)
                id_dict[id] = id
            except:
                count += 1
                id_dict[id] = str(count)
        
        # ID, branch and iteration are required values to find children.
        id_ind = headers.index('id')
        oid_ind = headers.index('oid')
        branch_ind = headers.index('branch')
        iter_ind = headers.index('iteration')
        # Date must be set to the proper resolution
        date_ind = headers.index('data_date')
        
        for item in iter:
            self.rowcount += 1
            item = list(item)
            id = item[id_ind] #id for sql queries
            if level >= 1:
                item[id_ind] = id_dict[item[oid_ind]] #oid for output
            branch = item[branch_ind]
            ddate = item[date_ind]
            date = item[date_ind].year
            if level > 1:
                try:
                    int(item[oid_ind])
                except:
                    self.logger.log_message('smt', 'warning',
                                            'Bad stratum id \'' \
                                            + item[oid_ind] + \
                                            '\' on sim unit ' + \
                                            sim_unit + ', year ' + \
                                            str(date) + ' in results. '+\
                                            'Replacing with ' + \
                                            id_dict[item[oid_ind]])
            del item[date_ind]
            iter = item[iter_ind]
            if sim_unit is None:
                sim_id = id
            else:
                sim_id = sim_unit
            del item[oid_ind] #remove the excess oid field from output
            del item[max(branch_ind, iter_ind)]
            del item[min(branch_ind, iter_ind)]
            item.insert(0, sim_id)
            item.append(date)
            self._listflt2str(item, 1)
            result = item
            if level_const is not None and len(level_const) > 0:
                results.append(result)
            # Get the children
            child_levs = self.datadb.get_child_levels(level)
            if child_levs is not None and level == 1:
                children = self.datadb.get_child_ids(level, id, iter, branch,
                                                     ddate,)
                if children is not None and self._has_data(children):
                    for child_lev in child_levs:
                        new_constraints = self._copy_constraints(constraints)
                        new_constraints += [('id', 'in', children[child_lev], 
                                             'and'), ('pid', 'eq', id, 'and'), 
                                            ('branch', 'eq', branch, 'and'), 
                                            ('iteration', 'eq', iter, 'and'), 
                                            ('data_date', 'eq', ddate)]
                        results += self._get_strings_items(new_constraints,
                                                       dates, child_lev, sim_id)
            if level == 1 and self.rowcount > self.flush_limit:
                for i in range(len(results)):
                    results[i] = self.sep.join(results[i])
                self._flush(results)
                self.rowcount = 0
                results = []

        return results
