#!/usr/bin/env python
# encoding: utf-8
"""
cache.py

Bucked tree stem cache and pricearray/pricematrix cache

Created by Antti Mäkinen on 2010-06-10.
Copyright (c) 2010 __MyCompanyName__. All rights reserved.
"""

import os
import cPickle as pickle
import datetime
import numpy
from pysqlite2 import dbapi2 as sqlite3

from simo.db.datadb.db import CoreDB


class StemcacheDB(CoreDB):
    
    def __init__(self, db_path='stemcache.db', make_new_db=False,
                 memory=False):
        """
        If make_new_db == True, the old database will be destroyed
        db_type in (SQLITE, POSTGRESQL)
        """
        super(StemcacheDB, self).__init__(db_path, make_new_db, memory,
                                          'localhost', 5432, 'test', 'test',
                                          'SQLITE', False)
        self.db_path = db_path
        self._create_tables()

    def _create_tables(self):
        """Create tables for storing tree stem and assortment information
        
        """
        if not self.db.table_exists('stem'):
            sql = '''CREATE TABLE stem(
                           id INTEGER PRIMARY KEY,
                           key TEXT NOT NULL,
                           vol REAL NOT NULL)'''
            self.db.execute(sql)
        if not self.db.table_exists('assortment'):
            sql = '''CREATE TABLE assortment(
                           stem_id INTEGER,
                           assortment REAL NOT NULL,
                           vol REAL NOT NULL,
                           val REAL NOT NULL,
                           CONSTRAINT fk_stem_id FOREIGN KEY (stem_id) 
                                   REFERENCES stem(id))'''
            self.db.execute(sql)


class PriceCache(object):
    """Cache class for storing processed cash flow arrays
    """
    def __init__(self, cache_dir, cache_file):
        """Initialize a price cache object

        cache_dir -- cache directory path as string
        cache_file -- cache file name as string
        """
        self.use_cache = False
        self._cache = None
        self._cache_path = None
        writable = True
        # check if cache directory exists
        if not os.path.isdir(cache_dir):
            # try to create cache directory
            try:
                os.makedirs(cache_dir)
            except:
                # it's possible that the user doesn't have the rights to
                # write to the current working dir -> cannot use cache
                writable = False
        if writable:
            # check if cache file exists and if not, create a new cache
            self._cache_path = os.path.join(cache_dir, cache_file + '.dat')
            dt = datetime.datetime.today()
            if os.path.isfile(self._cache_path):
                fp = open(self._cache_path, 'rb')
                try:
                    obj = pickle.load(fp)
                    self._set_cache(obj[0], obj[1], obj[2])
                except EOFError:
                    # the cache file exists but is probably empty, initialize an
                    # empty cache
                    self._set_cache(dt, dt)
                fp.close()
            else:
                self._set_cache(dt, dt)
            self.use_cache = True

    def _set_cache(self, init_date, update_date, price_matrices={}):
        """Set cache values

        init_date -- cache creation date as datetime.date object
        update_date -- last cache update date as datetime.date object
        price_matrices -- cached price matrices in a dictionary
        """
        self._init_date = init_date
        self._update_date = update_date
        self._cache = price_matrices

    def check(self, cashflowtable):
        """Check if given cashflowtable is found in the cache. Cash flow table
        is reduced to a vector of column sums and cast to tuple so that it is
        hashable and can be used as a dictionary key.

        cashflowtable -- two-dimensional numpy array with columns:
                         (price region, species, assortment, min diameter,
                          min length, price)
        """
        psvector = tuple(cashflowtable.sum(axis=0))
        if psvector not in self._cache:
            return True, None
        else:
            return False, self._cache[psvector]

    def update(self, cashflowtable, pricematrix):
        """Update price cache and save it to disk

        cashflowtable -- two-dimensional numpy array with columns:
                         (price region, species, assortment, min diameter,
                          min length, price)
        pricematrix -- price matrices per tree species in a dictionary
        """
        psvector = tuple(cashflowtable.sum(axis=0))
        if psvector not in self._cache:
            self._cache[psvector] = pricematrix

        dt = datetime.datetime.today()
        self._update_date = dt

        fp = open(self._cache_path, 'wb')
        obj = (self._init_date, self._update_date, self._cache)
        pickle.dump(obj, fp)
        fp.close()


class StemCache(object):
    """Cache for storing information on bucked tree stems
    """
    def __init__(self, cache_dir, cache_file):
        """Initialize a stem cache object

        cache_dir -- cache directory path as string
        cache_file -- cache file name as string
        """
        self.use_cache = False
        self.new_data = False
        self._cache = None
        self._cache_db = None
        self._cache_path = None
        self._meta_path = None
        self.initialized = False
        writable = True
        # check if cache directory exists
        if not os.path.isdir(cache_dir):
            # try to create cache directory
            try:
                os.makedirs(cache_dir)
            except:
                # it's possible that the user doesn't have thw rights to
                # write to the current working dir -> cannot use cache
                writable = False
        if writable:
            self._meta_path = os.path.join(cache_dir, 'meta.dat')
            self._cache_path = os.path.join(cache_dir, cache_file + '.db')
            self.use_cache = True

    def initialize(self, divisor, log_reduction):
        """
        Loads the cache from file or creates an empty file if none exists or
        the divisor and log_reduction type don't match to the stored ones

        divisor -- length of pieces the stem is bucked in, float
        log_reduction -- log reduction type for bucking, int
        """
        if not self.use_cache:
            return
        self._divisor = divisor
        self._log_reduction = log_reduction
        dt = datetime.datetime.today()
        
        # check if cache file exists and if not, create a new cache
        if os.path.isfile(self._cache_path) and \
            os.path.isfile(self._meta_path):
            
            # try to load the cache metadata
            fp = open(self._meta_path, 'rb')
            try:
                obj = pickle.load(fp)
                init_date, update_date, divisor, log_reduction,\
                        price_matrices, max_lims_by_sp = obj
                if divisor != self._divisor or \
                   log_reduction != self._log_reduction:
                    self._set_cache(dt, dt)
                else:
                    self._fill_cache(init_date, update_date, price_matrices,
                                     max_lims_by_sp)
            except EOFError:
                # the cache file exists but is probably empty, initialize
                # an empty cache
                self._set_cache(dt, dt)
            fp.close()
        else:
            self._set_cache(dt, dt)
        self.initialized = True

    def _set_cache(self, init_date, update_date):
        """Initialize an empty stem cache

        init_date -- cache creation date as datetime.date object
        update_date -- last cache update date as datetime.date object
        price_matrices -- price matrix column sum vectors
        max_lims_by_sp -- maximum reject diameter and height limits for each
                          tree species in a dictionary
        """
        self._init_date = init_date
        self._update_date = update_date
        self._price_matrices = []
        self._max_lims_by_sp = {}
        self._cache = {}
        self._cache_db = StemcacheDB(self._cache_path, True)
        self._new_stems = set([])
        self.new_data = False
    
    def _fill_cache(self, init_date, update_date, price_matrices,
                    max_lims_by_sp):
        """Fill cache with existing tree stem data from an sqlite database"""
        self._init_date = init_date
        self._update_date = update_date
        self._price_matrices = price_matrices
        self._max_lims_by_sp = max_lims_by_sp
        self._cache = {}
        self._cache_db = StemcacheDB(self._cache_path, False)
        
        # fill the cache with the stems in the cache db
        sql = '''SELECT * from stem, assortment 
                 WHERE stem.id = assortment.stem_id;'''
        data = self._cache_db.db.get(sql)
        for item in data:
            sid, key, vol, stem_id, assortment, as_vol, as_val = item
            if key not in self._cache:
                self._cache[key] = [vol, {}] 
            self._cache[key][1][int(assortment)] = [as_vol, as_val]
        
        self._new_stems = set([])
        self.new_data = False

    def check(self, timber, sp, dbh, height, pricematrix, decimals):
        """Check if the given tree should be bucked (it is not yet in the
        cache)

        timber -- timber assortment collection, dictionary
        sp -- tree species, int
        dbh -- diameter at breast height, float (cm)
        height -- tree height, float (m)
        pricematrix -- price matrix
        decimals -- cache resolution, number of decimals, int
        """
        # first check if tree diameter and height are below the current
        # maximum rejection limits
        #if sp in self._max_lims_by_sp:
        #    if dbh <= self._max_lims_by_sp[sp][0] and \
        #        height <= self._max_lims_by_sp[sp][1]:
        #        return 0.0, False

        psvector = tuple(pricematrix.sum(axis=0))
        # NB: pricematrices are now summarized by a tuple of column sums
        if psvector not in self._price_matrices:
            return 0.0, True
        pmkey = self._price_matrices.index(psvector)

        # construct the cache key
        ckey = "%0.*f;%0.*f;%d;%d" % (decimals, dbh, decimals, height, sp,
                                      pmkey)
        if ckey not in self._cache:
            return 0.0, True
        stemvol, stored = self._cache[ckey]
        for key, val in stored.items():
            timber[key][0] += val[0]
            timber[key][1] += val[1]

        return stemvol, False

    def flush(self):
        """
        cache content is flushed to the disk
        """
        dt = datetime.datetime.today()
        self._update_date = dt
        
        # get the maximum id number from the cache db
        sql = 'SELECT MAX(id) FROM "stem"'
        i = self._cache_db.db.get(sql)[0][0]
        if i is None:
            i = 0
        
        # insert the new items in the cache to the cache db
        for key in self._new_stems:
            
            # update tree stem id
            i += 1
            
            vol, assortments = self._cache[key]
            
            # first add id, key and volume to stem table
            sql = 'INSERT INTO stem'\
                  ' (id, key, vol)'\
                  ' VALUES (?, ?, ?)'
            sql_data = (i, key, vol)
            self._cache_db.db.execute(sql, sql_data)
            
            # then add the various assortments
            for assortment, values in assortments.items():
                # get assortment volume and value and make sure that they are
                # valid numerical values
                vol, val = values
                if numpy.isnan(vol) or numpy.isnan(val):
                    continue
                sql = 'INSERT INTO assortment'\
                      ' (stem_id, assortment, vol, val)'\
                      ' VALUES (?, ?, ?, ?)'
                sql_data = (i, assortment, vol, val)
                self._cache_db.db.execute(sql, sql_data)
        
        # pickle the metadata to disk
        fp = open(self._meta_path, 'wb')
        obj = (self._init_date, self._update_date,
               self._divisor, self._log_reduction, self._price_matrices,
               self._max_lims_by_sp)
        pickle.dump(obj, fp)
        fp.close()
        
        self.new_data = False
        self._new_stems = set([])
        
        self._cache_db.db.commit()

    def update(self, stemvol, timber, sp, dbh, height, pricematrix,
               decimals, too_small=False):
        """Update cache with the current tree information

        stemvol -- tree stem volume, float
        timber -- timber assortment collection, dictionary
        divisor -- log divisor, float
        sp -- tree species, int
        dbh -- diameter at breast height, float (cm)
        height -- tree height, float (m)
        pricematrix -- price matrix
        decimals -- cache resolution, number of decimals, int
        """
        # if tree is tagged as too small, add the tree dbh and height to
        # maximum rejection limits if they are higher than the current maximum
        if too_small:
            #if sp in self._max_lims_by_sp:
            #    dc, hc = self._max_lims_by_sp[sp]
            #    dbh = dbh if dbh > dc else dc
            #    height = height if height > hc else hc
            #self._max_lims_by_sp[sp] = (dbh, height)
            return

        # NB: pricematrices are now summarized by a tuple of column sums
        psvector = tuple(pricematrix.sum(axis=0))
        if psvector not in self._price_matrices:
            pmkey = len(self._price_matrices)
            self._price_matrices.append(psvector)
        else:
            pmkey = self._price_matrices.index(psvector)

        # construct the cache key
        ckey = "%0.*f;%0.*f;%d;%d" % (decimals, dbh, decimals, height, sp,
                                      pmkey)

        # store the values to the cache
        self._cache[ckey] = (stemvol, timber)
        self._new_stems.add(ckey)
        self.new_data = True
    
    def commit(self):
        self._cache_db.db.commit()
    
    def close(self):
        """Close stem cache & databases
        """
        try:
            self._cache_db.db.commit()
        except:
            pass
        self._cache_db.db.close()
        

def main():
    pass

if __name__ == '__main__':
    main()

