#!/usr/bin/env python
# encoding: utf-8
"""
*aggregationcaller.py*

Class for constructing and executing aggregation model calls

Copyright (c) 2008 Simosol Oy. Distributed under the GNU General Public
License 2.0.
"""

import numpy

from simo.simulation.caller.caller import Caller
from simo.simulation.model.aggregationarg import AggregationArg
from simo.simulation.condeval import ConditionEvaluator


class AggregationModelCaller(Caller):
    def __init__(self, deterministic=False):
        """
        Initialize model caller
        """
        super(AggregationModelCaller, self).__init__('aggregation')
        self._evaluator = ConditionEvaluator(self.add_error)
        self.arg = None
        self._error_stack = []
        self._deterministic = deterministic

    def execute(self, model, params, sim, depthind):
        """
        Execute aggregation model
        """
        my_super = super(AggregationModelCaller, self)
        my_super.execute(model, params, sim, depthind)
        if self._construct_call():
            model._v_func(self.arg)
            if not self.arg.success:
                for errorstr in self.arg.errors:
                    self._error_stack.append((errorstr, sim.level, None))
            else:
                # store aggregation results into matrix, but make sure that
                # NaNs are not stored
                # storing of the values depends on the dimensionality (i.e.
                # 'scalar' or 'vector') of the target
                target_level, target_var = self.arg.target_variable
                if self.arg.scalar:
                    mask = numpy.isnan(self.arg.results) == False
                    if len(mask) > 0:
                        self._sim.data.set_value(target_level,\
                            self.arg.target_index[mask], target_var,\
                            self.arg.results[mask], model.name)
                else:
                    # if target is vector, results are stored one variable at
                    # a time as the masks of the various attributes can be of
                    # different lengths
                    n = len(target_var)
                    for i in range(n):
                        mask = numpy.isnan(self.arg.results[i]) == False
                        if len(mask) > 0:
                            self._sim.data.set_value(\
                                target_level, self.arg.target_index[mask],
                                target_var[i], self.arg.results[i, mask],
                                model.name)
        for errorstr, level, objs in self._error_stack:
            self.add_error(errorstr, level, objs)
        self.arg = None
        self._error_stack = []

    def _construct_call(self):
        """
        Construct an aggregation model call
        """
        # get a 'master' tind that can be used to identify target objects
        # and check that there are target objects, abort model call if not
        target_level, target_ind = self._params.target_ind
        through_level = self._params.through_level
        scalar = self._params.scalar
        tind, torem = self._get_targets(target_level, self._depth)
        active = self._sim.data.get_active()
        if tind.size == 0:
            return False
        # construct aggregation argument container
        self.arg = AggregationArg((target_level, target_ind), tind, torem,
                                  scalar, self._deterministic)
        failed = None
        if self._params.condition is not None:
            # evaluate aggregation condition and find out which objects did
            # not pass the condition
            lowest_level = None
            if self._params.lowest_operand_level > target_level:
                # if the lowest operand level in aggregation condition is
                # lower in the hierarchy than target level, pass the lowest
                # level for the evaluator as the condition should be evaluated
                # for the lowest level objects
                lowest_level = self._params.lowest_operand_level
            exprlevel, exprtind, exprtorem = target_level, tind, torem
            expr = self._params.condition
            opmem = self._sim.operation_memory
            passed = self._evaluator.evaluate(self._sim.data, exprtind, expr,
                                              exprlevel, self._sim.main_level,
                                              self._depth, opmem,
                                              self._sim.timespan,
                                              through_level, lowest_level)
            failed = passed == False
        # store the individual operands
        for i, oper in enumerate(self._params.operands):
            if not self._add_operand(i, target_level, through_level, oper,
                                     failed, scalar):
                return False
        # handle the missing target objects, values and weights
        self.arg.handle_missing()
        # log error messages
        if len(self._error_stack) > 0:
            for errorstr, level, errorobjs in self._error_stack:
                self.add_error(errorstr, level, errorobjs)
        return True

    def _add_operand(self, i, target_level, through_level, oper, failed,
                     scalar):
        """Add a single operand to aggregation arguments
        """
        if oper.type == 'variable':
            data_level, data_var = oper.variable
            # get tind for the current operand
            otind, torem = self._get_targets(target_level, self._depth,
                                             data_level, through_level)
            # check that there are target objects left
            if otind.size == 0:
                return False
            # attributes must be in a list or tuple
            if scalar:
                attrs = [data_var]
            else:
                attrs = data_var
            if oper.weight is not None:
                weight_level, weight_ind = oper.weight
                if scalar:
                    attrs.append(weight_ind)
                else:
                    attrs += weight_ind
            # if there is a default value for variable, missing values do not
            # cause error messages
            default = True if oper.default_value is not None else False
            # get values from data handler
            arr, err = self._sim.data.get_value(otind, attrs, default)
            if arr is None:
                return False
            # if data handler returned errors, log the error messages and
            # store removable targets
            lev = target_level if data_level is None else data_level
            if err is not None:
                for e, objs in err[0]:
                    self._error_stack.append((e, lev, err[1][objs]))
                # update removable targets and values
                torem.update(err[2])
            # substitute NaNs with default values if operand has a default
            if default:
                if scalar:
                    self._nan2default(oper, arr, 0, 1)
                else:
                    n = len(attrs)/2 if oper.weight is not None else len(attrs)
                    for i in range(n):
                        self._nan2default(oper, arr, i, n+1)
            # if there was an aggregation condition, store only the values
            # for the objects that passed the condition
            if failed is not None:
                arr[:,failed] = numpy.NaN
            # get the values and weights
            if scalar:
                vind, wind = 0, 1
            else:
                n = len(attrs)/2 if oper.weight is not None else len(attrs)
                vind, wind = range(n), range(n,n+n)
            values = arr[vind]
            weights = None
            if oper.weight is not None:
                weights = arr[wind]
            self.arg.add_operand(data_level,otind, torem, values, weights,
                                 failed)
        else:
            # operand type is 'value': store the value for each target object
            if scalar:
                shape = self.arg.target_index.shape[0]
            else:
                shape = (self.arg.target_index.shape[0], oper.value.size)
            values = numpy.ones(shape, dtype=float)
            #TODO: this transpose should be avoided as it is expensive
            tvalues = numpy.transpose(values * oper.value)
            self.arg.add_operand(None, self.arg.target_index, set([]),
                                 tvalues, None, failed)
        return True

    def _nan2default(self, oper, arr, vind, wind):
        arr[vind,numpy.isnan(arr[vind,:])] = oper.default_value
        if oper.weight is not None:
            arr[wind,numpy.isnan(arr[wind,:])] = oper.default_weight


def main():
    pass

if __name__ == '__main__':
    main()
