#!/usr/bin/env python
# encoding: utf-8
"""
*geotablecaller.py*

Construct geo table model calls and process results.

Copyright (c) 2008 Simosol Oy. Distributed under the GNU General Public
License 2.0.
"""

# these are for running the doctest from src/simo
#import sys
#sys.path.append('..')

import numpy

from simo.simulation.caller.caller import Caller


class GeoTableCaller(Caller):
    def __init__(self):
        """Initialize geotable caller
        """
        super(GeoTableCaller, self).__init__('geo table')

    def __find_geo_indices(self, tind, coord, ulaxis, gridstep, maxsteps, axis):
        """Find indice from the geo table by coordinate variable values
           based on the upper left corner and spacing of the grid.
        """
        vals, err = self._sim.data.get_value(tind, coord[1])
        missing = set([])
        if err is not None:
            # log the error messages and store missing targets
            for e, objs in err[0]:
                self.add_error(e, None, err[1][objs])
            missing.update(err[2])
        if axis == 'X':
            origo = ulaxis + 0.5 * gridstep
            axisind = numpy.array(((vals - origo) / gridstep).round(),
                                  dtype=int)
        elif axis == 'Y':
            origo = ulaxis - 0.5 * gridstep
            axisind = numpy.array(((origo - vals) / gridstep).round(),
                                  dtype=int)
        # sanity check for indices
        outside = (axisind > maxsteps) | (axisind < 0) | numpy.isnan(vals)
        if numpy.any(outside):
            errormsg = "outside bound %s coordinates" % axis
            self.add_error(errormsg, None, tind[outside])
            remove = list(missing.union(list(numpy.where(outside)[0])))
        else:
            remove = missing
        return axisind, remove

    def __which_values_to_extract(self, gtable, param):
        """Which values to return from the geo table
        """
        if param.get_all: # which values to extract
            varind = range(gtable.nvar)
        else:
            varind = []
            for resvar in param.result_variables:
                for i in range(gtable.nvar):
                    if gtable.targets[i] == resvar:
                        varind.append(i)
        return varind

    def execute(self, gtable, params, sim, depth):
        """Execute a geo_table model
        """
        super(GeoTableCaller, self).execute(gtable, params, sim, depth)
        tind, torem = sim.data.get_tind(sim.level, depth)
        xind, xrem = self.__find_geo_indices(tind, gtable.longitude_var,
                                             gtable.ulx, gtable.xdim,
                                             gtable.ncol-1, 'X')
        yind, yrem = self.__find_geo_indices(tind, gtable.latitude_var,
                                             gtable.uly, gtable.ydim,
                                             gtable.nrow-1, 'Y')
        varind = self.__which_values_to_extract(gtable, params)
        remove = list(set(xrem).union(yrem))
        tind = numpy.delete(tind, remove, axis=0)
        xind = numpy.delete(xind, remove, axis=0)
        yind = numpy.delete(yind, remove, axis=0)
        for i in varind:
            levelind, varind = gtable.targets[i]
            if levelind != sim.level:
                elevelname = sim.lexicon.get_level_name(sim.level)
                levelname = sim.lexicon.get_level_name(levelind)
                varname = sim.lexicon.get_variable_name(levelind, varind)
                errormsg = "Level of (%s) target variable '%s' is different "\
                           "from model chain evaluation level (%s)"\
                            % (levelname, varname, elevelname)
                self.add_error(errormsg, None, tind)
                return
            # note that rowind = yind, colind = xind
            resultvalues = gtable.arr[yind, xind, i]
            sim.data.set_value(levelind, tind, varind, resultvalues,
                               gtable.name, sim.value_fixing)


def main():
    import doctest
    doctest.testfile('test/geotablecaller-py.rst')

if __name__ == '__main__':
    main()
