#!/usr/bin/env python
# encoding: utf-8
"""
*operationcaller.py*

Caller class for evaluating operation models

Copyright (c) 2008 Simosol Oy. Distributed under the GNU General Public
License 2.0.
"""

# these are for running the doctest from src/simo
#import sys
#sys.path.append('..')

import copy
import numpy
from collections import defaultdict

from simo.utils.utils import update_dates
from simo.simulation.caller.caller import Caller
from simo.simulation.caller.operationmemory import OperationMemory
from simo.simulation.caller.paramtablecaller import get_indices
from simo.simulation.model.operationarg import OperationArg


class OperationModelCaller(Caller):
    """Class for executing operation models"""
    def __init__(self, warnings):
        """Initialize model caller"""
        super(OperationModelCaller, self).__init__('operation model')
        self.arg = None
        self.memory = OperationMemory()
        self._warnings = warnings
        self._error_stack = []

    def execute(self, model, param, sim, depth):
        """Construct and execute operation model call"""
        super(OperationModelCaller, self).execute(model, param, sim, depth)
        model.dropped = None
        if not self._construct_call(model, param, sim, depth):
            # log error messages
            if len(self._error_stack) > 0:
                for errorstr, level, errorobjs in self._error_stack:
                    self.add_error(errorstr, level, errorobjs)
            self.arg = None
            self._error_stack = []
            return False
        # call the model function and process operation results
        self._call_model(model, sim, depth)
        self._extract_results(model, sim, depth)
        if model.interactive:
            # if operation was interactive, add operation name and group to
            # operation history
            if self.arg.target_index.size > 0:
                ok = self.arg.drop_target == False
                if True in ok:
                    cdates = sim.data.get_date(self.arg.target_index[ok,:])
                    up = update_dates(cdates,
                                      sim.timespan.time_step,
                                      sim.timespan.unit,
                                      deduceaday=True)
                    ok_index = numpy.where(ok)[0]
                    self.memory.update(self.arg.target_index.take(ok_index,
                                                                  axis=0),
                                       model.index,
                                       sim.level, up)
        drop_index = numpy.where(self.arg.drop_target)[0]
        model.dropped = self.arg.target_index.take(drop_index, axis=0)
        self.arg = None
        self._error_stack = []

    def _construct_call(self, model, param, sim, depth):
        """
        Construct operation model call
        """
        self.arg = OperationArg(model.name, model.group,
                                param.name, param.group,
                                model.optype, model.language, sim)
        tind, toremove = self._get_targets(sim.level, depth)
        if len(tind) == 0:
            # we're expecting targets, but got none.
            # can happen at least when errors in forced operation chains
            errmsg = 'No operation targets found: %s' % model.name
            self._error_stack.append((errmsg, None, None))
            return False
        self.arg.set_targets(tind, toremove)
        self.arg.cash_flow_handler = model.cash_flow_handler
        self.arg.sim_effects = param.simulation_effects
        self.arg.notes = param.notes
        self.arg.materials = param.materials

        if not self._assign_cash_flow_table(param, sim):
            return False
        if not self._assign_cash_flow_model(param, sim, depth):
            return False
        if not self._assign_result_weights(model, sim):
            return False
        if model.optype == 'model':
            # If operation type is 'model', construct operation variable,
            # parameter, result and data structures
            self._construct_parameters(model, sim, param)
            if not self._construct_variables(model, sim, depth):
                return False
            if not self._construct_result_structure(model, sim, depth):
                return False
            if not self._assign_param_table(param, sim, depth):
                return False
            if model.data_vars is not None:
                # Get the operation input data from the right level, if the
                # operation has input data structure!
                if not self._constr_data(model, sim, depth):
                    return False
        if numpy.all(self.arg.drop_target):
            return False
        # log error messages
        if len(self._error_stack) > 0:
            for errorstr, level, errorobjs in self._error_stack:
                self.add_error(errorstr, level, errorobjs)
            self._error_stack = []
        return True

    def _assign_cash_flow_table(self, param, sim):
        """Assign cash flow table to operation arguments

        NB! This method takes around 17-18% of the total simulation time!

        param -- operation model parameters
        sim -- Simulator object
        """
        if param.cash_flow_model is not None:
            return True
        if param.cash_flow_table is None:
            return True
        num_of_tables = len(param.cash_flow_table)
        if num_of_tables <= 0:
            return True
        self.arg.init_cash_flow_tables(param.cash_flow_table)
        for k, table in enumerate(param.cash_flow_table):
            if table is None:
                continue
            if sim.level == sim.main_level:
                dates = sim.data.get_date(self.arg.target_index)
            elif sim.level > sim.main_level:
                parents, torem = sim.data.get_parent(sim.level,
                                                     self.arg.target_index,
                                                     sim.main_level)
                dates = sim.data.get_date(parents)
            else:
                sl = sim.lexicon.get_level_name(sim.level)
                ml = sim.lexicon.get_level_name(sim.main_level)
                errorstr = "cannot assign unique cash flow tables when "\
                           "simulation level '%s' is a parent level of "\
                           "simulation main level '%s'" % (sl, ml)
                self.arg.drop_target[:] = True
                self._error_stack.append((errorstr, None, None))
                return False
            un = numpy.unique(dates)
            if len(un) == 1:
                # all simulation main level objects have same dates
                table.set_active_array(dates[0])
                self.arg.set_cash_flow_table(table, k)
            else:
                # targets have unique dates
                for i in range(self.arg.num_of_targets):
                    tablecopy = self._deepcopy(table)
                    tablecopy.set_active_array(dates[i])
                    self.arg.set_cash_flow_table(tablecopy, k, i)
        return True

    def _assign_cash_flow_model(self, param, sim, depth):
        """Assign cash flow model to operation arguments

        TODO: write tests for cash flow model use cases!
        """
        model = param.cash_flow_model
        if model is None:
            return True

        # get values for cash flow model variables
        # TODO: if evaluation level is not the same as main level,
        # using target_index won't work => short circuiting this for now
        rows = self.arg.target_index.shape[0]
        cols = len(model.variables)
        variables = numpy.zeros((rows, cols), dtype=float)
        for i, mvar in enumerate(model.variables):
            level, variable = mvar
            # get target index for variable level objects
            if level == sim.level:
                tind, torem = sim.data.get_tind(sim.level, depth)
            else:
                tind, torem = sim.data.get_tind(sim.level, depth, level)
            values, errors = sim.data.get_value(tind, variable)
            if errors is not None:
                # log the error messages and store the modified target index
                m = " when calling model '%s'" % model.name
                for e, objs in errors[0]:
                    self._error_stack.append((e+m, sim.level, tind[objs]))
                return False
            variables[:,i] = values

        # TODO: not implemented in cash_flow_model.xsd yet!
        parameters = None
        tind = self.arg.target_index
        dates = sim.data.get_date(tind)
        self.arg.init_cash_flow_tables(param.cash_flow_model.table)
        # go through each target unit and set a separate cash flow table for
        # each of the units
        for i in range(self.arg.num_of_targets):
            date = dates[i]
            for k, tab in enumerate(model.table):
                # call the cash flow model function, which processes the cash flow
                # table associated with the model
                tablecopy = self._deepcopy(tab)
                tablecopy.set_active_array(date)
                iteration = tind[i,0]
                errors = []
                if not model.function(tablecopy, date, iteration,
                                      variables[i], parameters, errors):
                    errorstr = "error in cash flow model '%s': " % model.name
                    errorstr += ', '.join(errors)
                    self.arg.drop_target[:] = True
                    self._error_stack.append((errorstr, None, None))
                    return False
                # assign the processed cash flow table to the operation model
                self.arg.set_cash_flow_table(tablecopy, k, i)
        return True

    def _assign_param_table(self, param, sim, depth):
        """Assign parameter table values to arguments
        """
        num_of_tables = len(param.parameter_tables)
        for k, table in enumerate(param.parameter_tables):
            targets = table.targets
            # TODO: should drop_target array be used here?!? It seems to break
            # the indice searching!!!
            #tind = self.arg.target_index[self.arg.drop_target == False,:]
            tind = self.arg.target_index[:,:]
            if tind.size == 0 or numpy.all(self.arg.drop_target):
                return False
            indices, tind, coef = get_indices(tind, table, sim, depth,
                                              self._get_targets,
                                              self.add_error)
            if indices is None or indices.size == 0:
                return False
            # only a single table/array for each parameter table
            arr = table.arrays[0].array
            for i in xrange(table.Ntargets):
                start = table.Nclassifiers + i
                end = start + targets[i].length
                values = arr.take(indices, axis=0)[:,start:end]
                # TODO: the loop here should be discarded, but it requires
                # that operation models are also modified
                for j in xrange(values.shape[0]):
                    self.arg.set_param_table_values(values[j,:], j, k,
                                                    num_of_tables)
        return True

    def _assign_result_weights(self, model, sim):
        """Assign operation result weight variable values to arguments
        """
        weight = model.operation_result_weight
        if weight is None:
            return True
        level, ind = weight
        if level != sim.level:
            msg = "operation result weight variable level should be equal "\
                  "to current simulation level"
            self._error_stack.append((msg, None, None))
            return False
        tind, torem = sim.data.get_tind(level, sim._depth)
        values, err = sim.data.get_value(tind, ind)
        torem = set([])
        if err is not None:
            # log the error messages and store the modified target index
            for e, objs in err[0]:
                self._error_stack.append((e, level, err[1][objs]))
            torem = err[2]
        self.arg.set_weights(values, torem)
        return True

    def _call_model(self, model, sim, depth):
        """Call the operation operation model or the cash_flow function
        """
        evalres = numpy.zeros(self.arg.num_of_targets, dtype=int)
        if self.arg.type == 'cash_flow':
            # evaluate cash_flow models without calling operation library
            evalres = self._evaluate_cash_flow_operation(evalres, sim, depth)
        elif self.arg.type == 'model':
            # evaluate operation models that are implemented in libraries
            if model.function is None:
                errorstr = "function object for '%s' missing!" % model.name
                self.add_error(errorstr, None, None)
                evalres[:] = -1
            else:
                # operation call has always the following parameters:
                #   data - operation data by level
                #   invars - input variable values (from simulation data)
                #   params - parameter values (user-defined)
                #   paramtablevals - parameter table values
                #   cashflowtable - cash flow table instance
                #   cashflow - container for returning operation cash flow
                #   results - container (array) for returning operation results
                #   errors - container (list) for returning errors
                arg = self.arg
                for i in range(self.arg.num_of_targets):
                    if self.arg.drop_target[i]:
                        continue
                    data = arg.get_data(i)
                    vars = arg.get_vars(i)
                    params = arg.get_params()
                    ptvals = arg.get_param_table_vals(i)
                    cftab = arg.get_cash_flow_table(i)
                    cflow = arg.get_cash_flow(i)
                    res = arg.get_results(i)
                    err = arg.get_errors(i)
                    warn = arg.get_warnings(i)
                    restype = model.result_type
                    evalres[i] = model.function(data, vars, params, ptvals,
                                       cftab, cflow, res, restype, warn, err,
                                       sim.price_cache, sim.stem_cache)
                    
                if self.arg.cash_flow_handler == 'simulator':
                    # if the model didn't handle the cash flow, do it now
                    self._evaluate_cash_flow_operation(evalres, sim, depth)

        # check the evaluation result and process the operationarg for result
        # extraction and processing
        self.arg.check_eval(evalres)

    def _classify_results(self, classifiers, sim, depth):
        """Generate operation result variable names by combining the variable
        name and the classifier name from the cash_flow table
        """
        # Get the cash flow table instance and process the classifiers
        # The cash flow table should be the same for all of the operation
        # targets, so get the cash flow table name from the first one
        ret = self._get_classifiers(classifiers, sim, depth)
        nclasses, vals, names, labs, torem, errdata = ret
        if None in (nclasses, vals, names, labs):
            return None, None, torem, errdata
        n = len(classifiers)
        reslabs = [{} for x in range(self.arg.num_of_targets)]
        self._set_labels(n, vals, labs, reslabs)
        return nclasses, reslabs, torem, None

    def _constr_data(self, model, sim, depth):
        """
        Construct operation input data and add to arguments
        """
        levelindices = model.data_vars.keys()
        levelindices.sort()
        hier = sim.lexicon.hierarchy
        min_ord = 9999
        for li, data in hier.items():
            if data['ordinal'] < min_ord:
                min_ord = data['ordinal']
                root_li = li
        still_to_get = set(levelindices)

        # first block all child relatives of simulation level objects
        to_block, clevels = [], copy.copy(levelindices)
        self._children_to_block(to_block, hier, sim.level, clevels)
        blocked = sim.data.block_children(depth, sim.level, to_block)

        objs = {}
        first = True
        to_process_li = hier.keys()

        while len(to_process_li) > 0 and len(still_to_get) > 0:
            # get the level index to collect data for
            if first:
                level = root_li
                parent = level
                to_process_li.remove(root_li)
                objs[level] = None
                first = False
            else:
                next_level = self._pick_a_child(hier, parent, to_process_li)
                if next_level is None:
                    # no more sibling levels for the current level, move down
                    parent = level
                    level = self._pick_a_child(hier, parent, to_process_li)
                    if level is None:
                        # we're at a leave of the hierarchy graph, but there
                        # may be levels left to collect from other lineages
                        if len(to_process_li) != 0:
                            level = to_process_li.pop(0)
                            parent = hier[level]['parent']
                else:
                    level = next_level
            if level in still_to_get:
                still_to_get.remove(level)

            # define map level (for mapping between current level and
            # operation main level objects)
            datalevel, maplevel = level, sim.level
            if level > sim.level:
                # lower in the hierarchy (a child level of maplevel)
                maplevel = parent
            elif level == sim.level:
                # current level is mapping level
                datalevel = None
            # get mapping between parent mapping level and data level, and
            # pass it to the data level construction
            if level in levelindices:
                if not self._construct_level_data(maplevel, depth,
                                                  datalevel, parent,
                                                  objs, sim, model):
                    return False
            else:
                if datalevel == root_li:
                    parent_objs = None
                else:
                    parent_objs = objs[parent]
                objs[datalevel] = None
                if datalevel > sim.level:
                    tind, torem = self._get_targets(maplevel, depth,
                                                    datalevel, None, True,
                                                    parent_objs)
                    objs[datalevel] = sim.data.active_dataobjects

        # release all previously blocked child level objects
        if len(blocked) > 0:
            for blevel, btind in blocked.items():
                sim.data.release(blevel, depth, btind)
        return True

    def _construct_level_data(self, maplevel, depth, datalevel, parent, objs,
                              sim, model):
        """
        Collect data for the model from the specified level

        maplevel -- data mapping level for the data level
        depth -- model chain depth this model call is being processed at
        datalevel -- data level to collect data from
        parent -- parent data level for data level
        objs -- active data object hierarchy as dictionary with level
                indices as keys
        sim -- simulator instance
        model -- model being evaluated
        """
        parent_objs = objs[parent]
        tind, torem = self._get_targets(maplevel, depth, datalevel, None,
                                        True, parent_objs)
        if maplevel == sim.level and len(torem) > 0:
            self.arg.drop_target[list(torem)] = True

        # get data attribute indices
        if datalevel is None:
            level = maplevel
        else:
            level = datalevel
        attrs = model.data_vars[level]
        values, err = sim.data.get_value(tind, attrs)
        if err is not None:
            # log the error messages and store the modified target index
            for e, e_objs in err[0]:
                self._error_stack.append((e, level, err[1][e_objs]))
                for obj in e_objs:
                    torem.update(err[2])
        if values is None:
            levelname = sim.lexicon.get_level_name(level)
            err_str = "no data for model '%s' on level '%s'"\
                       % (model.name, levelname)
            o_ids = sim.data.get_id(maplevel, tind)
            if datalevel is None:
                o_str = " for objects: %s" % o_ids
            else:
                o_str = " for parent objects: %s" % o_ids
            err_str += o_str
            self._error_stack.append((err_str, maplevel, tind))
            return False
        
        maptind, maptorem = copy.copy(tind), copy.copy(torem)
        objs[level] = None
        child = False
        if level > sim.level:
            # when lower in the hierarchy than simulation level, the
            # object indices are retrieved from handler's active
            # dataobjects
            objs[level] = copy.copy(sim.data.active_dataobjects)
            child = True
        elif level == sim.level:
            objs[level] = copy.copy(tind[:, 0:3])
        elif level < sim.level:
            # when higher in the hierarchy than simulation level, the
            # mapping is done with simulation level parent indices
            maptind, maptorem = sim.data.get_parent(sim.level,
                                                    self.arg.target_index,
                                                    level)
        # construct level data array
        data = self._constr_data_level(maptind, values, child, objs[level])
        # if current mapping level differs from simulation main level, get
        # new target index for mapping between current level and sim.level
        if maplevel != sim.level:
            maptind, maptorem = self._get_targets(sim.level, depth,
                                                  datalevel)
        # get index of operation main level object positions in the level
        # data array (start:end -indices similar to tind)
        if child:
            # construct index for simulation level child levels
            index = copy.copy(maptind[:,3:])
        else:
            # construct index for simulation level and higher levels
            n = maptind[:,0].size
            index = numpy.zeros((n, 2), dtype=int)
            index[:,0] = numpy.arange(n)
            index[:,1] = numpy.arange(1,n+1)
        # store current level data to the operation argument container
        self.arg.set_data(data, index, level, attrs)

        return True

    def _constr_data_level(self, tind, values, child, oind):
        """Construct operation input data structure for single level

        tind -- target index (2D numpy array)
        values -- data level values as numpy array
        child -- boolean indicating whether the current level is a child level
        oind -- object index as numpy array
        """
        nattrs = values.shape[0]
        ncols = nattrs + 4  # iteration + branch + object + attrs + tag
        if not child:
            # operation top level, no parent indices and only one data object
            # for each target object
            nrows = tind.shape[0]
            data = numpy.zeros((nrows, ncols), dtype=float)
            data[:,3:-1] = numpy.transpose(values[:,:])
            data[:,0:3] = tind[:,0:3]
        else:
            # level has a parent level to which a linkage should be maintained
            # by adding a parent indice for each object
            ncols += 1  # add a column for parent indices
            nrows = tind[:,-1].max()
            n = tind[:,4] - tind[:,3]
            o = tind[:,2]
            data = numpy.zeros((nrows, ncols), dtype=float)
            data[:,4:-1] = numpy.transpose(values)
            data[:,0:3] = oind
            data[:,3] = numpy.repeat(o, n)
        return data

    def _construct_parameters(self, model, sim, param):
        """Get operation parameter values and add to arguments

        model -- OperationModel instance
        sim -- Simulator instance
        param -- OperationModelParam instance
        """
        for i in range(model.n_params):
            self.arg.set_params(param.parameters[i])

    def _construct_result_structure(self, model, sim, depth):
        """Construct operation result structure, which depends on the number
        of target objects and result variables

        model -- OperationModel instance
        sim -- Simulator instance
        depth -- current model chain hierarchy depth indice as int
        """
        if model.result_type is None:
            return True
        if 'data' in model.result_type:
            # For 'data' type operation results, there's also result
            # target and result level
            self.arg.result_target = model.result_target
            self.arg.result_level = model.result_level
        nres, reslabs, torem, errdata = self._define_result_variables(model,
                                                                      sim,
                                                                      depth)
        if nres is None and reslabs is None:
            level = None
            tind = None
            if errdata:
                level, tind = errdata
            errorstr = "Could not construct result structures"
            self._error_stack.append((errorstr, level, tind))
            return False
        # get the maximum number of result objects and maximum number of
        # results (due to multiple, possibly classified result variables)
        nmaxres = numpy.max(numpy.sum(nres, axis=0))
        nmaxcols = nmaxres
        res_vars = model.result_variables
        # construct result containers for 'operation_result' or 'data' results
        if 'operation_result' in model.result_type:
            results = [{} for i in range(self.arg.num_of_targets)]
            dc = self._deepcopy
            for i in range(self.arg.num_of_targets):
                for res_type, varname, cfier, var_type, do_weight in res_vars:
                    if res_type == 'data':
                        continue
                    # variable type can be either 'scalar' or 'date_array'
                    # the scalar results are classified by the price matrix
                    # classifiers, date_array result structure is an empty
                    # list to which operation model should add an undefined
                    # number of (value, timing) tuples
                    if var_type == 'scalar':
                        if cfier is not None:
                            results[i][varname] = dc(reslabs[cfier][i])
                    else:
                        results[i][varname] = []
            if 'data' in model.result_type:
                ntargets = len([i for i in res_vars if 'data'==i[0]])
                for r in results:
                    r['operation_data_results'] = numpy.zeros(ntargets,
                                                              dtype=float)
        else:
            results = numpy.zeros((self.arg.num_of_targets, nmaxcols),
                                  dtype=float)
        
        self.arg.set_result_structure(model.result_target,
                                      model.result_level,
                                      model.result_type, nres,
                                      model.result_variables, results, torem)
        return True

    def _construct_variables(self, model, sim, depth):
        """Get operation variable values from data and add to arguments

        model -- OperationModel instance
        sim -- Simulator instance
        depth -- current model chain hierarchy depth indice as int
        """
        where = numpy.where
        if model.vars is None:
            return True
        # Create an empty array for operation variables
        nvars = model.n_vars
        torem = set([])
        for datalevel, item in model.vars.iteritems():
            ltind, torem = self._get_targets(sim.level, depth, datalevel)
            vals, err = sim.data.get_value(ltind, item['index'])
            if err is not None:
                # log the error messages and store the modified target index
                for e, objs in err[0]:
                    self._error_stack.append((e, datalevel, err[1][objs]))
                torem.update(err[2])
            # Store the variable values into the prediction argument container
            # in the correct order
            self.arg.set_vars(nvars, vals, item['order'], torem)
            # if warning output is on, check variable limits and add warning
            # messages if values are out of bounds
            if self._warnings and item['limits'] is not None:
                lims = item['limits']
                for i, lim in enumerate(lims):
                    if lims[i] is None:
                        continue
                    ok = lims[i].check(vals[i,:])
                    if False in ok:
                        varind = item['index'][i]
                        varname = sim.lexicon.get_variable_name(datalevel,
                                                                varind)
                        warnstr = "variable '%s' value outside limits %s" % \
                            (varname, lim.limitstr(varname))
                        indices = where(ok == False)[0]
                        # get active data objects
                        if datalevel != sim.level:
                            act = sim.data.get_active()[-1]
                        else:
                            act = ltind
                        sim.add_warning(warnstr, datalevel,
                                        act.take(indices, axis=0))
        return True


    def _define_result_variables(self, model, sim, depth):
        """Define number of operation result variables and construct result
        labels

        model -- OperationModel instance
        sim -- Simulator instance
        depth -- current model chain hierarchy depth indice as int
        """
        nresvars = len(model.result_variables)
        narraycols = numpy.zeros((nresvars, self.arg.num_of_targets), int)
        torem = set([])
        if 'data' in model.result_type:
            labels = None
            if model.result_target == 'new' or model.result_level == sim.level:
                # 'data' type results have only one result object and multiple
                # result variables
                narraycols += 1
            else:
                # result targets are existing objects, but children of 'self'
                # Get the number of result level children for each target (at
                # current level) object
                getc = sim.data.get_children
                plevel = sim.level
                clevel = model.result_level
                nt = self.arg.num_of_targets
                for i in range(nt):
                    it, br, obj = self.arg.target_index[i,:3]
                    children = getc(it, br, obj, plevel, clevel)
                    if children is not None:
                        narraycols[:,i] = len(children)

        if 'operation_result' in model.result_type:
            labels = {}
            for key in model.result_classifiers.keys():
                classifiers = model.result_classifiers[key]
                item = self._classify_results(classifiers, sim, depth)
                nclasses, reslabs, toremove, errdata = item
                torem.update(toremove)
                if nclasses is None or reslabs is None:
                    return None, None, torem, errdata
                # store the class count of each result variable that is
                # associated with the current classifier
                for i in xrange(nresvars):
                    level, var, ckey, vtype, do_weight = \
                        model.result_variables[i]
                    if ckey == key:
                        narraycols[i,:] = nclasses
                labels[key] = reslabs
        return narraycols, labels, torem, None

    def _get_classifiers(self, classifiers, sim, depth):
        """Get values for the operation result classifiers from data or cash
        flow table classifiers

        classifiers -- operation result classifier list
        sim -- Simulator instance
        depth -- current model chain hierarchy depth indice as int
        """
        n = len(classifiers)
        values = [None for x in range(n)]
        names = [None for x in range(n)]
        labels = [None for x in range(n)]
        nclasses = numpy.ones(self.arg.num_of_targets, dtype=int)
        torem = set([])
        for i in xrange(n):
            c = classifiers[i]
            names[i] = c.name
            labels[i] = c.labels
            if c.type == 'variable':
                common = False
                tind, toremove = self._get_targets(sim.level, depth, c.level)
                torem.update(toremove)
                vals, err = sim.data.get_value(tind, c.ind)
                if err is not None:
                    # log the error messages and store objects with missing
                    # values
                    for e, objs in err[0]:
                        self._error_stack.append((e, c.level, err[1][objs]))
                    torem.update(err[2])
                if vals is None:
                    lvname = sim.lexicon.get_level_name(c.level)
                    errorstr = "No objects found on level '%s' when " \
                               "searching values for classifier '%s'" \
                               % (lvname, c.name)
                    self._error_stack.append((errorstr, sim.level, tind))
                    return None, None, None, None, torem, (sim.level, tind)
                if c.level != sim.level:
                    # number of classes is defined by the number of unique
                    # classifier values
                    uq = numpy.unique
                    vals = [uq(vals[s:e]) for s,e in tind[:,3:]]
                    n = numpy.array([x.size for x in vals])
                else:
                    # reshape the value array so that it'll have two dims
                    vals = numpy.reshape(vals, (vals.size, 1))
                    n = 1
            elif c.type == 'cashflow':
                common = True
                try:
                    vals = labels[i].keys()
                except Exception, e:
                    raise Exception(str(e)+'\nName: '+c.name+' '+str(labels)+' '+str((c.name, c.labels, )))
                n = len(vals)
            nclasses *= n
            values[i] = (vals, common)
        return nclasses, values, names, labels, torem, None

    def _evaluate_cash_flow_operation(self, evalres, sim, depth):
        """Evaluate cash flow operation

        evalres -- evaluation result, boolean numpy array
        sim -- Simulator instance
        depth -- current model chain hierarchy depth indice as int
        """
        arg = self.arg
        dtind, torem = self._get_targets(sim.level, depth, sim.main_level)
        dates = sim.data.get_date(dtind)
        for i in range(arg.num_of_targets):
            cash_flow_table = arg.get_cash_flow_table(i)
            cfiers = cash_flow_table[0].classifiers
            if cfiers is not None:
                cvals = numpy.empty((arg.num_of_targets, len(cfiers)), int)
                curlevel = sim.level
                for name, item in cfiers.iteritems():
                    tind = arg.target_index[i]
                    tind.shape = (1, tind.shape[0])
                    if item.level == 'within-operation':
                        raise Exception("'within-operation' classifiers not "\
                                        "allowed in cash flow models!")
                    level, var = item.level, item.ind
                    if level != curlevel:
                        tind, torem = self._get_targets(curlevel, depth, level)
                    vals, err = sim.data.get_value(tind, [var])
                    if err is not None:
                        # log the error messages and store the modified
                        # target index
                        for e, objs in err[0]:
                            self._error_stack.append((e, level, err[1][objs]))
                        arg.drop_target[i] = True
                        break
                    cvals[:, item.order] = vals
                if arg.drop_target[i]:
                    continue
                # access the cash_flow from the cash_flow dictionary
                cash_flow_table[0].set_active_array(dates[i])
                dictn = cash_flow_table[0].active_array.dict
                keystr = ':'.join(numpy.asarray(cvals[i], str))
                arg.set_cash_flow(dictn[keystr], i)
            else:
                if arg.drop_target[i]:
                    continue
                cash_flow_table[0].set_active_array(dates[i])
                dictn = cash_flow_table[0].active_array.dict
                arg.set_cash_flow(dictn[0], i)
            evalres[evalres == 0] = 1
        return evalres

    def _extract_results(self, model, sim, depth):
        """
        Extract operation model results

        model -- OperationModel instance
        sim -- Simulator instance
        depth -- current model chain hierarchy depth indice as int
        """
        if model.optype == 'model' and self.arg.result_type is not None:
            # process the operation model results
            if 'operation_result' in self.arg.result_type:
                # process operation results that do not affect the simulation
                # data and are stored in the operation_result database
                self._process_oper_results(sim, model.interactive)
            elif model.interactive and self.arg.cash_flow_handler != 'none':
                # process cash_flow here, if operation_result wasn't
                # in the result_type
                self._process_cash_flow(sim)
            if 'data' in self.arg.result_type:
                # process results that are stored in the data matrix
                self._process_data_results(sim, model.result_type)
        if model.interactive:
            # if the model is interactive and it wasn't handled yet, handle
            # cash_flow here.
            if self.arg.cash_flow_handler != 'none':
                if model.optype != 'model' or self.arg.result_type is None:
                    self._process_cash_flow(sim)
            # if interactive, process data mods and simulation effects
            # process the operation data
            # (remove cut trees etc. from the data structure.)
            self._process_data(sim)
            # process the simulation effects of the operation, eg. erasing
            # model memory and forcing the time step after operation
            self._process_simulation_effects(sim)

    def _process_cash_flow(self, sim):
        """Process and store operation cash flow

        sim -- Simulator instance
        """
        # Get operation result definitions etc. and process the cash flow for
        # all evaluation level objects
        level = sim.level
        levelname = sim.lexicon.get_level_name(level)
        name = self.arg.db_name
        group = self.arg.db_group
        notes = self.arg.notes
        materials = self.arg.materials
        tind = self.arg.target_index
        cashflows = self.arg.get_cash_flow()
        weights = self.arg.get_weights()
        if weights is not None:
            cashflows *= weights
        oids = sim.data.get_id(level, tind)
        for i in xrange(self.arg.num_of_targets):
            if self.arg.drop_target[i]:
                continue
            iteration, branch, obj = tind[i,:3]
            target = (iteration, branch, levelname, oids[i][0])
            reslist = [{'cash_flow': cashflows[i]}]
            mytind = tind[i]
            mytind.shape = (1, len(mytind))
            sim.queue_opres(mytind, target, [(reslist, None)], name, group,
                            notes, materials)

    def _process_data_results(self, sim, result_type):
        """
        Process operation results that affect the simulation data

        sim -- Simulator instance
        result_type -- result type indicator as str
        """
        modname = self.arg.name
        fix = sim.value_fixing
        level = self.arg.result_level
        plevel = sim.level
        attrs = [x[2][1] for x in self.arg.result_variables if x[0] == 'data']
        if self.arg.result_target == 'new':
            # create a new object to the result level
            for i in range(self.arg.num_of_targets):
                if self.arg.drop_target[i]:
                    continue
                it, br, par = self.arg.target_index[i,:3]
                # add a new object to the result level and store model
                # results to the added objects
                tind = sim.data.add_objects(it, br, level, 1, plevel, par)
                values = self.arg.get_results(i)
                if 'operation_result' in result_type:
                    values = values['operation_data_results']
                sim.data.set_value(level, tind, attrs, values, modname,
                                   fix)
        else:
            if level != plevel:
                # if result target is an existing object and target result
                # level is different from the current evaluation level, it's
                # expected to be a child level
                getc = sim.data.get_children
                getr = self.arg.get_results
                setv = sim.data.set_value
                zeros = numpy.zeros
                na = len(attrs)
                for i in range(self.arg.num_of_targets):
                    it, br, obj = self.arg.target_index[i,:3]
                    results = getr(i)
                    if 'operation_result' in result_type:
                        results = results['operation_data_results']
                    children = getc(it, br, obj, plevel, level)
                    if children is None:
                        continue  # no children - no need process results
                    nc = len(children)
                    # first reshape the result vector into 2d matrix and
                    # then transpose the matrix into (attribute, object) shape
                    resmat = results[0:nc*na].reshape((nc,na))
                    resmat_t = resmat.transpose()
                    # construct a target index for the child level objects
                    # and store the values into data matrix
                    ctind = zeros((nc,5), dtype=int)
                    ctind[:,0] = it
                    ctind[:,1] = br
                    ctind[:,2] = children
                    setv(level, ctind, attrs, resmat_t, modname, fix)
            else:
                # result level should be same as the evaluation level
                # drop_target handling correct???
                use = self.arg.drop_target == False
                tind = self.arg.target_index[use]
                values = self.arg.get_results()
                if 'operation_result' in result_type:
                    for i, val in enumerate(values):
                        values[i] = val['operation_data_results']
                    values = numpy.array(values)
                values = values[use].transpose()
                sim.data.set_value(level, tind, attrs, values, modname, fix)
        return True

    def _process_data(self, sim):
        """Process data objects that were affected by the operation model
        Operation indicator tag values:
            0 - No modifications
            1 - The Object variable values have been modified
            2 - The object has been removed

        sim -- Simulator instance
        """
        if self.arg._data is None:
            return
        if False not in self.arg.drop_target:
            return
        modname = self.arg.name
        opmem = self.memory
        prmem = sim.prediction_memory
        fix = sim.value_fixing
        levels = self.arg._data.keys()
        levels.sort(reverse=True)
        for level in levels:
            attrs = self.arg._data[level][2] #data, iter, attrs
            for i in range(self.arg.num_of_targets):
                if self.arg.drop_target[i]:
                    continue
                data = self.arg.get_data(i)[level]
                # first modify objects with tagged with 1
                mod = data[:,-1] == 1
                if numpy.any(mod):
                    n = len(attrs)
                    values = data[mod,-(n+1):-1].transpose()
                    tind = numpy.array(data[mod,0:3], dtype=int)
                    sim.data.set_value(level, tind, attrs, values, modname,
                                       fix)
                # remove objects with tagged with 2
                rem = data[:,-1] == 2
                if numpy.any(rem):
                    iters = numpy.unique(data[:,0])
                    for it in iters:
                        irows = data[:,0] == it
                        branches = numpy.unique(data[irows,1])
                        for br in branches:
                            brows = irows * (data[:,1] == br)
                            objs = numpy.asarray(data[brows*rem, 2],
                                                     dtype=int)
                            sim.data.del_objects(it, br, level, objs)
                            opmem.del_objects(it, br, objs, level)
                            prmem.del_objects(it, br, objs, level)

    def _process_oper_results(self, sim, interactive):
        """Process operation results that are stored in the operation result
        database

        sim -- Simulator instance
        interactive -- boolean value indicating if model is interactive
        """
        if interactive: #calculate cash flows
            cashflows = self.arg.get_cash_flow()
            weights = self.arg.get_weights()
            if weights is not None:
                cashflows *= weights

        level = sim.level
        levelname = sim.lexicon.get_level_name(level)
        name = self.arg.db_name
        group = self.arg.db_group
        notes = self.arg.notes
        materials = self.arg.materials
        tind = self.arg.target_index
        weights = self.arg.get_weights()
        results = self.arg.get_results()
        resvars = self.arg.result_variables
        nresvars = len(resvars)
        oids = sim.data.get_id(level, tind)
        for i in range(self.arg.num_of_targets):
            op_data = []
            if self.arg.drop_target[i] or results[i] is None:
                continue
            iteration, branch = tind[i,:2]
            target = (iteration, branch, levelname, oids[i][0])
            resdict = defaultdict(dict)
            delay_dict = defaultdict(dict)
            if interactive: #include cash_flow
                resdict[None]['cash_flow'] = cashflows[i]
            for j in range(nresvars):
                res_type, var_name, key, var_type, do_weight = resvars[j]
                if res_type == 'data':
                    continue
                if var_type == 'scalar':
                    # process scalar result values
                    if key is not None:
                        key, popped = self._pop_out_text_classifiers(key)
                        for cfier, value in results[i][var_name].items():
                            en = enumerate(cfier)
                            pairs = tuple([(key[k], c) for k, c in en])
                            if popped:
                                pairs = self._inject_text_classifiers(level, \
                                                    tind[i, 2], pairs, popped)
                            weight = weights[i] if do_weight else 1.0
                            resdict[pairs][var_name] = value * weight

                    elif var_name in results[i]:
                        weight = weights[i] if do_weight else 1.0
                        val = results[i][var_name] * weight
                        resdict[None][var_name] = val
                    else:
                        msg = "variable '%s' not in operation '%s' results" % \
                              (var_name, self.arg.name)
                        self.add_error(msg, level, tind[i:i+1,:])
                else:
                    # process date_array result values, result values are
                    # arrays with corresponding timing values
                    for item in results[i][var_name]:
                        value, delay = item
                        delay_dict[delay][var_name] = value
            if len(delay_dict) > 0:
                res_list = []
                for delay, vals in delay_dict.iteritems():
                    res_list = [vals]
                    op_data.append((res_list, delay))
            # queue operation results only if there are any in the resdict
            # (resdict can be empty if all result variables are of
            # date_array type)
            if len(resdict) > 0:
                res_list = []
                for cfiers, vals in resdict.iteritems():
                    if cfiers is not None:
                        cdict = dict(cfiers)
                        vals.update(cdict)
                    res_list.append(vals)
                op_data.append((res_list, None))
            mytind = tind[i]
            mytind.shape = (1, len(mytind))
            sim.queue_opres(mytind, target, op_data, name, group, notes,
                            materials)

    def _pop_out_text_classifiers(self, key):
        if len(self._model.text_classifiers[key]) == 0:
            return key, []
        pop_these = []
        popped = []
        # text_classifiers defined as a tuple of (level name, var name)
        my_text_cfiers = [var[1] for var in self._model.text_classifiers[key]]
        for ind in range(len(key)):
            if key[ind] in my_text_cfiers:
                pop_these.append(ind)
        key = list(key)
        pop_these.sort(reverse=True)
        for ind in pop_these:
            popped.append(unicode(key.pop(ind)))
        return key, popped

    def _inject_text_classifiers(self, lind, oind, pairs, popped):
        pairs = list(pairs)
        for text_cfier in popped:
            text_val = self._sim.data.text_data[lind][oind].get(text_cfier)
            pairs.append((text_cfier, text_val))
        return tuple(pairs)

    def _children_to_block(self, to_block, hier, simlevel, datalevels):
        """
        Pick the children under simlevel to block

        to_block -- block these levels
        hier -- data level hierarchy
        simlevel -- the data level we're blocking from
        datalevels -- all data levels needed by the model
        """
        lineages2block = set([])
        for lvl in datalevels:
            lineages2block.update(hier[lvl]['lineage'])

        children = hier[simlevel]['children']
        if children is None:
            return
        for child in children:
            if child in datalevels or \
                    hier[child]['lineage'].issubset(lineages2block):
                to_block.append(child)
                if child in datalevels:
                    datalevels.remove(child)
        if len(datalevels) > 0:
            for child in children:
                self._children_to_block(to_block, hier, child, datalevels)

    def _pick_a_child(self, hier, parent, to_process_li):
        """
        Walk the hierarchy tree, and pick each children breadth first for each
        if they are in the to_process_li list

        hier -- data level hierarchy
        parent -- the parent level for which to pick the childer
        to_process_li -- list of level indices to pick altogether
        """
        level = None
        children = hier[parent]['children']
        if children is None:
            return level
        for child in children:
            if child in to_process_li:
                level = child
                to_process_li.remove(child)
                break
        return level

    def _process_simulation_effects(self, sim):
        """Process simulation effects caused by the operation model

        sim -- Simulator instance
        """
        if self.arg.sim_effects is None or self.arg._data is None:
            return
        if False not in self.arg.drop_target:
            return
        # Empty the model memories of each object from all active branch's
        # levels if 'erase_memory' is True
        if self.arg.sim_effects['erase_memory']:
            prmem = sim.prediction_memory
            levels = self.arg._data.keys()
            levels.sort(reverse=True)
            for level in levels:
                for i in range(self.arg.num_of_targets):
                    if self.arg.drop_target[i]:
                        continue
                    data = self.arg.get_data(i)[level]
                    # delete prediction memory for unmodified and modified obj
                    survivors = data[:,-1] != 2
                    if numpy.any(survivors):
                        iters = numpy.unique(data[:,0])
                        for it in iters:
                            irows = data[:,0] == it
                            branches = numpy.unique(data[irows,1])
                            for br in branches:
                                brows = irows * (data[:,1] == br)
                                objs = numpy.asarray(data[brows*survivors, 2],
                                                         dtype=int)
                                prmem.del_objects(it, br, objs, level)

        # Change the simulation time step according to the 'force_step'
        # definitions if 'force_step' is not None
        if self.arg.sim_effects['force_step'] is not None:
            raise Exception('Forcing simulation step not ready!')

    def _set_labels(self, nc, values, labels, rlabs, i=0, row=[],
                    obj=None):
        """
        Set result label keys for each target object. Classifier value can
        be either from data or from cash flow table classifiers. When some or
        all of the classifiers are from data, the number of classes between
        different target objects can be different. If classifiers are only
        cash flow classifiers, the number of classes between target objects
        is always the same. Result label data structure works also as
        operation result structure that is used to pass operation results
        from operation model to operation caller.

        nc -- number of classifiers as int
        values -- data values as numpy array
        names -- classifier names in a list
        labels -- classifier value to label mappings in a dictionary
        rlabs -- list of lists for storing the classifier tuples for each object and result
        i -- classifier indice
        row -- label values from the previous classifiers
        obj -- 'locked' target object indice
        """
        n = self.arg.num_of_targets
        last = i == nc  # last classifier?
        if last:
            # NB: now classifier values are store instead of the labels
            # labels should be transformed into labels with lexicon, if
            # needed
            c = tuple(row)
            if obj is not None:
                rlabs[obj][c] = 0.0
            else:
                for j in range(n):
                    rlabs[j][c] = 0.0
            return
        vals, common = values[i]
        if not common:
            # classifier values are from data and each target object has it's
            # own value array
            no = len(vals)
            for o in range(no):
                if obj is not None and o != obj:
                    continue  # target is 'locked', process only obj
                nv = vals[o].size
                # NB: data classifiers should always be the last ones as
                # it is much more efficient!!!
                for k in range(nv):
                    rowcopy = self._copy(row)
                    rowcopy.append(vals[o][k])
                    self._set_labels(nc, values, labels, rlabs, i+1,
                                     rowcopy, o)
        else:
            # classifier values are from cash flow classifiers and the values
            # are the same for all objects
            for val in vals:
                rowcopy = self._copy(row)
                rowcopy.append(val)
                self._set_labels(nc, values, labels, rlabs, i+1,
                                 rowcopy, obj)

    def _copy(self, data):
        return copy.copy(data)

    def _deepcopy(self, data):
        #NB: this deepcopy takes around 10-15% of the total simulation time!!!
        return copy.deepcopy(data)


def main():
    import doctest
    doctest.testfile('./test/operationcaller-py.rst')

if __name__ == '__main__':
    main()
