#!/usr/bin/env python
# encoding: utf-8
"""
*operationmemory.py*

Operation memory class for storing and accessing operation history.

Copyright (c) 2008 Simosol Oy. Distributed under the GNU General Public 
License 2.0.
"""

import math
import numpy
import datetime
import dateutil


ZERO_DATE = datetime.date(datetime.MINYEAR,1,1)


class OperationMemory(object):
    """
    Class for managing operation history
    """
    def __init__(self):
        """
        Initialize model memory
        """
        super(OperationMemory, self).__init__()
        self._since = None
        self._times = None
        self._shp = None
        self._level_map = {}
        
        # memory dimensions (iterations, branches, objects, operations)
        self._dates = numpy.zeros((1,1,0,0), dtype=datetime.date)
        self._times = numpy.zeros((1,1,0,0), dtype=int)
        
        self._shp = self._dates.shape
        
        self._dates.fill(ZERO_DATE)

    def _size_check(self, obj, opind):
        """
        Check that operation memory is large enough for given objects,
        resize memory if necessary
        
        obj -- object indice row (a row from target index)
        """
        shp = self._shp
        it = obj[0] + 1
        br = obj[1] + 1
        obj = obj[2] + 1
        opind = opind + 1
        if it > shp[0] or br > shp[1] or obj > shp[2] or opind > shp[3]:
            # if target index includes objects or branches that are outside
            # current memory limits, create a larger memory and copy the
            # values from the existing memory to the new one
            d0 = max(it, shp[0])
            d1 = max(br, shp[1])
            d2 = max(obj, shp[2])
            d3 = max(opind, shp[3])
            
            newdates = numpy.zeros((d0, d1, d2, d3), dtype=datetime.date)
            newdates.fill(ZERO_DATE)
            newdates[:shp[0],:shp[1],:shp[2],:shp[3]] = self._dates
            self._dates = newdates
            
            newtimes = numpy.zeros((d0, d1, d2, d3), dtype=int)
            newtimes[:shp[0],:shp[1],:shp[2],:shp[3]] = self._times
            self._times = newtimes
            
            self._shp = self._times.shape
    
    def update(self, tind, opind, level, dates):
        """
        Update operation memory by storing operation and year for objects
        defined in target index
        """
        if not self._level_map.has_key(level):
            self._level_map[level] = set([])
        self._level_map[level].add(opind)
        self._size_check(numpy.max(tind, axis=0), opind)
        
        dsetter = self._dates[:,:,:,opind]
        tsetter = self._times[:,:,:,opind]
        
        dsetter[tind[:,0],tind[:,1],tind[:,2]] = dates
        tsetter[tind[:,0],tind[:,1],tind[:,2]] += 1
    
    def reset(self):
        """
        Reset operation memory
        """
        self._dates[:,:,:,:] = ZERO_DATE
        self._times[:,:,:,:] = 0
        self._dates.fill(ZERO_DATE)
    
    def since(self, tind, opind, dates, timeunit='year'):
        """
        Get number of timesteps operation was done last time relative to
        given date
        """
        self._size_check(numpy.max(tind, axis=0), opind)
        dt = self._dates[tind[:,0],tind[:,1],tind[:,2],opind]
        if type(dates) == datetime.date:
            dd = numpy.zeros(dt.size, dtype=datetime.date)
            dd.fill(dates)
            dates = dd
        if timeunit == 'year':
            diff = [(dates[i].year - dt[i].year) for i in range(dt.size)]
        elif timeunit == 'month':
            diff = [((dates[i].year - dt[i].year) * 12. + \
                     (dates[i].month - dt[i].month)) for i in range(dt.size)]
        else:
            diff = [((dates[i].year - dt[i].year) * 365.25 + \
                     (dates[i].month - dt[i].month) * 30.44 + \
                     (dates[i].day - dt[i].day)) for i in range(dt.size)]
        
        return numpy.array(diff, dtype=int)
    
    def times(self, tind, opind):
        """
        Get number of times a given operation has been done
        """
        self._size_check(numpy.max(tind, axis=0), opind)
        return self._times[tind[:,0],tind[:,1],tind[:,2],opind]
    
    def del_objects(self, it, br, objs, level):
        """
        Delete objects from operation memory (reset given objects)
        """
        if self._level_map.has_key(level):
            opinds = list(self._level_map[level])
            self._dates[it,br,objs,opinds] = ZERO_DATE
            self._times[it,br,objs,opinds] = 0
    
    def add_branch(self, par_obj, new_obj, level):
        """Update operation memory after adding a new branch to simulation
        data matrix
        
        par_obj -- parent object row as a tuple or numpy array
        new_obj -- new object row as a tuple or numpy array 
        """
        if level in self._level_map:
            
            # make room to the operation memory array
            self._size_check(new_obj, 0)
            
            # parent objects might be outside mem limits as it is possible
            # that the objects' operation memory has not been used yet
            self._size_check(par_obj, 0)
            p_it, p_br, p_obj = par_obj
            n_it, n_br, n_obj = new_obj
            inds = tuple(self._level_map[level])
            
            # copy data from parent branches to new branches
            self._dates[n_it,n_br,n_obj,inds] = \
                numpy.copy(self._dates[p_it,p_br,p_obj,inds])
            self._times[n_it,n_br,n_obj,inds] = \
                numpy.copy(self._times[p_it,p_br,p_obj,inds])


def main():
    pass

if __name__ == '__main__':
    main()
