    #!/usr/bin/env python
# encoding: utf-8
"""
*paramtablecaller.py*

Class for evaluating parameter table models

Copyright (c) 2008 Simosol Oy. Distributed under the GNU General Public
License 2.0.
"""

import numpy

from simo.simulation.caller.caller import Caller


def get_indices(tind, table, sim, depthind, get_targets, add_error):
    """Find parameter table indices for target objects
    """
    classifiers = table.classifiers
    level = sim.level
    n_objs = tind.shape[0]
    error_ind = []
    # searchrange: one row for each object, columns:
    #   1. Success indicator, 1 or 0
    #   2. start indice
    #   3. end indice
    searchrange = numpy.zeros((n_objs, 3), dtype=int)
    searchrange[:,0] = 1
    searchrange[:,2] = -1
    arr = table.arrays[0].array
    remove_targets = set([])
    for i, c in enumerate(table.classifier_order):
        cvar = classifiers[c].ind
        clev = classifiers[c].level
        variable_type = classifiers[c].variable_type
        if clev != level:
            _tind, toremove = get_targets(level, depthind, clev)
            values, err = sim.data.get_value(_tind, (cvar,))
            if err is not None:
                # log the error messages and store objects with missing values
                for e, objs in err[0]:
                    add_error(e, level, _tind[objs])
        else:
            #NB! No data_level => toremove==set([])
            _tind, toremove = get_targets(level, depthind)
            values, err = sim.data.get_value(_tind, (cvar,))
            if err is not None:
                # log the error messages and store objects with missing values
                for e, objs in err[0]:
                    add_error(e, level, err[1][objs])
        # remove values for objects that had missing values for previous cfiers
        for e_ind in sorted(error_ind, reverse=True):
            e_ind = numpy.array(e_ind)
            _tind = numpy.delete(_tind, e_ind, axis=0)
            values = numpy.delete(values, e_ind, axis=1)
            if values.size == 0:
                return None, None, None
        # check that values is not None (in case of errors) and if it is,
        # return None values
        if values is None:
            return None, None, None
        # check if there's nans and remove them
        nans = numpy.isnan(values[0])
        if True in nans:
            errorstr = "could not find correct indices for table %s"\
                    % table.name
            e_ind = numpy.where(nans)[0]
            add_error(errorstr, level, _tind[e_ind])
            searchrange = numpy.delete(searchrange, e_ind, axis=0)
            _tind = numpy.delete(_tind, e_ind, axis=0)
            tind = numpy.delete(tind, e_ind, axis=0)
            values = numpy.delete(values, e_ind, axis=1)
            for e_i in e_ind:
                error_ind.append(e_i)
            if values.size == 0:
                return None, None, None
        if variable_type == 'float':
            # float variable should be the last classifier. There can be only
            # one float classifier!
            coef = _interpolate_value(arr, searchrange, values, i)
        else:
            if len(classifiers) > 1:
                coef = _limit_range(arr, searchrange, values, i)
            else:
                coef = []
                where = numpy.where
                for j in range(values.size):
                    searchrange[j,1] = 0
                    searchrange[j,2] = 1
                    start, end = searchrange[j,1:]
                    item = where(values[0][j]==arr[start:,i])[0]
                    if item.size == 0:
                        searchrange[j,1] = -1
                        searchrange[j,2] = -1
                        continue
                    searchrange[j,1] = start + item[0]
                    searchrange[j,2] = start + item[-1] + 1
                    coef.append(None)
    # after all classifiers have been processed, each object (row) should have
    # such start and end indices that end = start + 1
    miss = searchrange[:,2] - searchrange[:,1] != 1
    if numpy.any(miss):
        # construct error message for objects where miss is True and remove
        # those objects from target index
        errorstr = "could not find correct indices for table %s" % table.name
        errorind = numpy.where(miss)[0]
        add_error(errorstr, level, tind[errorind])
        tind = numpy.delete(tind, errorind, axis=0)
        searchrange = numpy.delete(searchrange, errorind, axis=0)
    indices = searchrange[:,1]
    return indices, tind, coef

def _limit_range(table, searchrange, values, i):
    """
    Limit parameter table search range for each object with the i'th
    classifier of the table
    """
    where = numpy.where
    # NB: only a single (first) attribute as a classifier
    V = values[0]
    coef = []
    for j in range(values.size):
        if j>(searchrange.shape[0] + 1):
            continue
        if searchrange[j,0]==0:
            continue
        start, end = searchrange[j,1:]
        if end == -1:
            target = table[start:,i]
        else:
            target = table[start:end,i]
        if end == -1 and V[j] not in target:
            searchrange[j,0] = 0
            continue
        elif end != -1 and V[j] not in target:
            searchrange[j,0] = 0
            continue
        if end == -1:
            item = where(V[j] == target)[0]
        else:
            item = where(V[j] == target)[0]
        searchrange[j,1] = start + item[0]
        searchrange[j,2] = start + item[-1] + 1
        coef.append(None)
    return coef

def _interpolate_value(table, searchrange, values, i):
    """
    Calculate coefficient for interpolation of float value from parameter
    table for object with the i'th classifier of the table
    """
    # NB: only a single (first) attribute as a classifier
    where = numpy.where
    V = values[0]
    coeff = []
    for j in range(values.size):
        start, end = searchrange[j,1:]
        if end==-1:
            le_items = V[j] >= table[start:,i]
            ge_items = V[j] <= table[start:,i]
        else:
            le_items = V[j] >= table[start:end,i]
            ge_items = V[j] <= table[start:end,i]
        le_values = table[le_items,i]
        ge_values = table[ge_items,i]
        le_values.sort()
        ge_values.sort()
        if len(ge_values)>0 and len(le_values)>0:
            start_ind = where(table[le_items,i]==le_values[-1])[0][0]
            le_indices = where(le_items)[0]
            start = le_indices[start_ind]
            searchrange[j,1] = start
            searchrange[j,2] = start + 1
            if le_values[-1]==ge_values[0]:
                coeff.append(0.)
            else:
                diff = ge_values[0] - le_values[-1]
                if diff==0.0:
                    coeff.append(0.)
                else:
                    a = V[j] - le_values[-1]
                    coeff.append(a / diff)
        else:
            searchrange[j,0] = 0
    return coeff


class ParamTableCaller(Caller):
    """Parameter table model caller
    """
    def __init__(self):
        super(ParamTableCaller, self).__init__('parameter_table')

    def execute(self, table, params, sim, depth):
        """Execute parameter table model"""
        super(ParamTableCaller, self).execute(table, params, sim, depth)
        level = sim.level
        fix = sim.value_fixing
        name = table.name
        tind, torem = self._get_targets(level, depth)
        if len(torem) > 0:
            #NB! Since _get_targets doesn't have datalevel set, len(torem)==0
            #That is, this should _never_ happen.
            self.add_error('paramtablecaller needs to remove indices',
                           level, torem)
        
        indices, tind, coef = get_indices(tind, table, sim, depth,
                                    self._get_targets, self.add_error)
        if indices is None or len(indices) == 0 or \
            coef is None or len(coef) == 0:
            errorstr = "cannot find table classifier values for parameter "\
                "table '%s'" % table.name
            self.add_error(errorstr, level, tind)
            return False
        for i in xrange(table.Ntargets):
            trglev = table.targets[i].level
            if coef[0] == None:
                values = table.arrays[0].array[indices, table.Nclassifiers+i]
            else:
                # Calculate value for float classifier using interpolation
                value1 = table.arrays[0].array[indices, table.Nclassifiers+i]
                value2 = table.arrays[0].array[indices+1, table.Nclassifiers+i]
                values = value1 + coef*(value2-value1)
            attr = table.targets[i].variable
            sim.data.set_value(level, tind, attr, values, name, fix)
        return True


def main():
    pass

if __name__ == '__main__':
    main()
