#!/usr/bin/env python
# encoding: utf-8
"""
*predictioncaller.py*

Caller class for evaluating prediction models

Copyright (c) 2008 Simosol Oy. Distributed under the GNU General Public
License 2.0.

TODO:
"""

# these are for running the doctest from src/simo
#import sys
#sys.path.append('..')

import numpy

from simo.simulation.caller.caller import Caller
from simo.simulation.caller.predictionmemory import PredictionModelMemory
from simo.simulation.model.predictionarg import PredictionArg


class PredictionModelCaller(Caller):
    """
    Class for executing prediction models
    """
    def __init__(self, warnings):
        """
        Initialize model caller
        """
        super(PredictionModelCaller, self).__init__('prediction model')
        self.memory = PredictionModelMemory()
        self.arg = None
        self._error_stack = []
        self._warnings = warnings

    def execute(self, model, param, sim, depth):
        """
        Construct and execute prediction model call
        """
        super(PredictionModelCaller, self).execute(model, param, sim, depth)
        
        # get the target index from data handler
        origtind, torem = self._get_targets(sim.level, depth)
        if origtind.size == 0:
            # no target objects, return without calling model
            return
        # check whether this is a model with a 'memory', i.e. one model call
        # produces predictions for several years
        usemem, tind = self.memory.use(model, sim, depth, origtind)
        if not usemem:
            # if tind from prediction memory differs from original tind (part
            # of the objects are blocked due to memory), get tinds always from
            # data handler instead of using stored tinds
            get_tind_from_data = tind.shape != origtind.shape
            # construct model call, call the model function, process results
            # and store them to the data matrix
            if self._construct_call(model, param, sim, depth, tind, torem,
                                    get_tind_from_data):
                self._call_model(model, sim, depth)
                self._process_results(sim)
                if self.arg.error_flag:
                    mname = self._model.name
                    for i in self.arg.error_objects:
                        e = self.arg.errors[i]
                        if e != '':
                            errorstr = "error '%s' returned from model '%s'"\
                                        % (e.strip(), mname)
                        else:
                            errorstr = "empty error string returned from "\
                                        "model '%s'" % mname
                        row = self.arg.target_index[[i]]
                        self.add_error(errorstr, sim.level, row)
                    self.arg.drop_error_objects()
        
        # release possible object blocks caused by result memory
        self.memory.release_blocked_objects(sim, depth)
        self._error_stack = []
        self.arg = None

    def _construct_call(self, model, param, sim, depth, tind, torem,
                        get_tind_from_data):
        """
        Construct prediction model call
        """
        # construct prediction model argument container
        self.arg = PredictionArg(model.name, model.language)
        self.arg.rect_factor = param.rect_factor
        self.arg.allowed_risk_level = param.risk_level
        nresobjs = 1
        if model.result_level != sim.level:
            # model result level differs from evaluation level (presumably a
            # child level of eval level) and number of result objects might
            # be more than one
            nresobjs = sim._max_res_objs
        self.arg.set_targets(tind, torem, nresobjs)
        # construct a warning container for possible warning messages
        self.warn = False
        # store model input variables and parameters into argument container
        if not self._construct_variables(model, sim, depth,
                                         get_tind_from_data):
            return False
        self._construct_parameters(model, param, sim)
        # remove target objects with missing values
        self.arg.remove_missing()
        # find the result level and store it in arg_prediction container
        self.arg.set_result_level(model.result_level, sim.level)
        # store the result variable indices
        self.arg.set_result_variables(model.result_vars)
        # create the containers for passing the results back from the models
        self.arg.set_containers()
        # log error messages
        if len(self._error_stack) > 0:
            for errorstr, level, errorobjs in self._error_stack:
                self.add_error(errorstr, level, errorobjs)
        if self.arg.num_of_targets == 0:
            return False
        if self.arg.model_language == 'c':
            self.arg.python2ctypes()
        return True

    def _construct_variables(self, model, sim, depth, get_tind_from_data):
        """
        Get model input variable values and add to arg
        """
        nvars = model.n_vars
        where = numpy.where
        for datalevel, item in model.vars.items():
            ltind, torem = self._get_targets(sim.level, depth, datalevel,
                                             None, get_tind_from_data)
            values, err = sim.data.get_value(ltind, item['index'])
            remove = set([])
            if err is not None:
                # log the error messages and store the modified target index
                m = " when calling model '%s'" % model.name
                for e, objs in err[0]:
                    self._error_stack.append((e+m, sim.level, ltind[objs]))
                remove = err[2]
            # store the variable values into the prediction argument container
            # in the correct order
            self.arg.set_input_variables(nvars, values, item['order'], remove)
            # if warning output is on, check variable limits and add warning
            # messages if values are out of bounds
            if self._warnings and item['limits'] is not None:
                lims = item['limits']
                for i, lim in enumerate(lims):
                    if lim is None:
                        continue
                    ok = lim.check(values[i,:])
                    if len(remove) > 0:
                        ok[list(remove)] = True
                    if False in ok:
                        varind = item['index'][i]
                        varname = sim.lexicon.get_variable_name(datalevel,
                                                                varind)
                        warnstr = "variable '%s' value outside limits %s" % \
                            (varname, lim.limitstr(varname))
                        indices = where(ok == False)[0]
                        # get active data objects if datalevel differs from
                        # simulation level
                        if datalevel != sim.level:
                            act = sim.data.get_active()[-1]
                        else:
                            act = ltind
                        sim.add_warning(warnstr, datalevel, act[indices])
        return True

    def _construct_parameters(self, model, param, sim):
        """
        Get model input parameter values and add to arg
        """
        if model.n_params is None:
            return
        for i in xrange(model.n_params):
            self.arg.add_parameter(param.parameters[i])

    def _call_model(self, model, sim, depth):
        """
        Call prediction model function
        """
        func = model._v_func
        mlib = model._v_mlib
        for i in range(self.arg.num_of_targets):
            # call the model function once for each target object
            if self.arg.model_language.lower() == 'c':
                ret = func(self.arg, mlib, i)
            elif self.arg.model_language.lower() == 'python':
                ret = func(self.arg, i)
            if ret != 1:
                # if prediction model returned error for the target object,
                # remove the object from the target objects in the arg and
                # raise error flag
                self.arg.add_error(i)
        if self.arg.model_language == 'c':
            self.arg.ctypes2python()
        if model.name in sim.memory_models and self.arg.num_of_targets > 0:
            if not self.memory.update(self.arg, model, sim, depth):
                errorstr = "cannot update prediction model memory for model "\
                           "'%s'" % model.name
                self.add_error(errorstr, sim.level, self.arg.target_index)

    def _process_results(self, sim):
        """
        docstring for _process_results
        """
        attrs = [x.variable for x in self.arg.result_vars]
        reslevel = self.arg.result_level
        if reslevel == 'self':
            reslevel = sim.level
        if reslevel == sim.level:
            self._process_eval_level_results(reslevel, attrs, sim)
        else:
            self._process_child_level_results(reslevel, attrs, sim)

    def _process_eval_level_results(self, level, attrs, sim):
        """
        Store results for existing objects
        """
        modname = self.arg.model_name
        if self.arg.target_index.size == 0:
            return
        if numpy.any(self.arg.mem > 10e300):
            errorstr = "infinite value(s) returned from model '%s'" % modname
            inf = self.arg.mem > 10e300
            indices = numpy.where(inf.sum(axis=1))[0]
            self.add_error(errorstr, level, self.arg.target_index[indices])
        tind = self.arg.target_index
        values = self.arg.mem
        fix = sim.value_fixing
        sim.data.set_value(level, tind, attrs, values, modname, fix)

    def _process_child_level_results(self, level, attrs, sim):
        """
        Create new objects and store model results to the new objects
        """
        modname = self.arg.model_name
        parlevel = sim.level
        nrv = self.arg.num_of_res_vars
        fix = sim.value_fixing
        for i in range(self.arg.num_of_targets):
            iteration, branch, parent = self.arg.target_index[i,:3]
            nro = self.arg.num_of_res_objs[i]  # number of new objects
            if nro == 0:
                self.arg.errors[i] = ', '.join([self.arg.errors[i], 'cannot '\
                                                'construct new objects'])
                continue
            # add a needed number of new objects to the result level and
            # store model results to the added objects
            tind = sim.data.add_objects(iteration, branch, level, nro,
                                        parlevel, parent)
            mem = self.arg.mem[:nrv*nro,i]
            values = numpy.reshape(mem, (nrv, nro,), order='FORTRAN')
            sim.data.set_value(level, tind, attrs, values, modname, fix)


def main():
    import doctest
    doctest.testfile('./test/predictioncaller-py.rst')

if __name__ == '__main__':
    main()
