#!/usr/bin/env python
# encoding: utf-8
"""
*predictionmemory.py*

Class for handling prediction model memory structures

Copyright (c) 2008 Simosol Oy. Distributed under the GNU General Public
License 2.0.
"""

import numpy
import datetime

from simo.utils.utils import update_dates
import simo.builder.names as names


class PredictionModelMemory(object):
    """Class for handling prediction model memory"""
    def __init__(self):
        """docstring for __init__"""
        super(PredictionModelMemory, self).__init__()
        self._map = {}
        self._level_map = {}
        self._mem = None
        self._time_steps = None
        self._shp = None
        self._blocked = None
        # memory dimensions:
        #     (iterations,
        #      branches,
        #      objects,
        #      (model,chain)-pairs)
        #NB! can this fail?
        self._mem = numpy.zeros((1,1,0,0), dtype=datetime.date)
        self._time_steps = numpy.zeros((1,1,0,0), dtype=float)
        self._shp = self._mem.shape

    def _key_check(self, key, level):
        """Check if the given key is mapped to memory indice, create mapping
        if not and return memory indice
        """
        if not key in self._map:
            self._map[key] = len(self._map)
            if not level in self._level_map:
                self._level_map[level] = set([])
            self._level_map[level].add(self._map[key])
        return self._map[key]

    def _size_check(self, obj, ind):
        """Check that model memory structure is large enough in all
        dimensions, resize memory if not

        obj -- object indice row (a row from target index)
        ind -- model indice as int
        """
        shp = self._shp
        it = obj[0] + 1
        br = obj[1] + 1
        obj = obj[2] + 1
        ind = ind + 1
        if it > shp[0] or br > shp[1] or obj > shp[2] or ind > shp[3]:
            # if model memory is not large enough, construct a larger memory
            # structure and copy the values from the existing memory
            # to the new one
            d0 = max(it, shp[0])
            d1 = max(br, shp[1])
            d2 = max(obj, shp[2])
            d3 = max(ind, shp[3])
            newmem = numpy.zeros((d0, d1, d2, d3), dtype=datetime.date)
            newmem[:,:,:,:] = datetime.date(1,1,1)
            newmem[:shp[0],:shp[1],:shp[2],:shp[3]] = self._mem
            self._mem = newmem
            # modify also the time step data structure
            newts = numpy.zeros((d0, d1, d2, d3), dtype=float)
            newts[:,:,:,:] = 1.
            newts[:shp[0],:shp[1],:shp[2],:shp[3]] = self._time_steps
            self._time_steps = newts
            # update memory shape
            self._shp = self._mem.shape

    def _get_time_steps(self, sim):
        """Get time step values for the active evaluation-level objects

        sim -- simulator instance
        """
        tslev, tsind = sim._ctrl.built_ins[names.TIME_STEP]
        if tslev != sim.level:
            tstind, torem = sim.data.get_tind(sim.level, sim._depth, tslev)
        else:
            tstind, torem = sim.data.get_tind(sim.level, sim._depth)
        timesteps, err = sim.data.get_value(tstind, tsind)
        if timesteps is None or err is not None:
            msg = "could not find time_step values: %s" % err
            sim.add_error(msg, sim.level, tstind)
            return None
        return timesteps

    def use(self, model, sim, depth, tind):
        """Check if model memory should be used

        model -- prediction model instance
        sim -- simulator instance
        depth -- model chain evaluation depth
        tind -- target index
        """
        usemem = False
        modname = model.name
        chain = sim.chain_index
        key = (modname, chain)
        if key in self._map:
            ind = self._map[key]
            self._size_check(numpy.max(tind, axis=0), 0)
            enddate = self._mem[tind[:,0],tind[:,1],tind[:,2],ind]
            # check for which objects the existing results still apply,
            # block the evaluation of those objects and get a new
            # target index (tind)
            if sim.level > sim.main_level:
                dtind, torem = sim.data.get_parent(sim.level, tind,
                                                   sim.main_level)
            else:
                dtind = tind
            cdates = sim.data.get_date(dtind)
            # check if the current time_step differs from the original
            # time_step and if it does, make sure that the object is not
            # blocked and new values are calculated
            timesteps = self._get_time_steps(sim)
            if timesteps is None:
                return False
            orig_timesteps = self._time_steps[tind[:,0],tind[:,1],tind[:,2],ind]
            no_block = timesteps != orig_timesteps
            block = enddate >= cdates
            if numpy.any(timesteps < 0):
                block[(timesteps < 0) & (enddate >= cdates)] = False
            if numpy.any(no_block):
                block[no_block] = False
            if True in block:
                self._blocked = tind[block,:]
                sim.data.block(sim.level, depth, self._blocked)
                tind, torem = sim.data.get_tind(sim.level, depth)
                if tind.size == 0:
                    usemem = True
        return usemem, tind

    def update(self, arg, model, sim, depthind):
        """
        Update prediction model memory and divide prediction results depending
        on their time spans and units

        model -- PredictionModel instance
        sim -- Simulator istance
        depthind -- model chain depth indice
        """
        # NB: the one common timespan for all variables is stored
        # separately for each result variable
        resvar = arg.result_vars[0]
        timespan, timeunit = resvar.time_span, resvar.time_unit
        timesteps = self._get_time_steps(sim)
        if timesteps is None:
            return False

        # check if PredictionArg instance contains objects that should be
        # removed due to missing input variable values and remove timesteps
        if arg.remove_targets:
            rem = list(arg.remove_targets)
            timesteps = numpy.delete(timesteps, rem, axis=0)

        # make sure that simulation time step unit and result variable
        # time span unit are the same, and if not, modify the time divider
        # and time steps accordingly
        timesteps, timedivider = self._match_spans(sim, timeunit,\
                                    sim.timespan.unit, timespan, timesteps)
        if timedivider is None:
            return False

        # calculate prediction result divider
        preddiv = timespan / timedivider

        # divide the prediction model result variable values
        if not self._divide_results(arg, model, sim, preddiv, timesteps):
            return False

        # get dates for the current objects
        tind = arg.target_index
        if sim.level > sim.main_level:
            dtind, torem = sim.data.get_parent(sim.level, tind, sim.main_level)
        else:
            dtind = tind

        # NB: there's no need to remove the dates of objects defined in
        # arg.remove_targets as was the case with timesteps; timesteps use
        # a target index retrieved from data handler, dates use a target
        # index from arg, from which the objects with missing values have
        # already been removed by arg.remove_missing() method
        cdates = sim.data.get_date(dtind)
        timespans = numpy.ones(timesteps.shape) * (timespan - 1)
        timespans[timesteps < 0] *= -1
        enddate = update_dates(cdates, timespans, timeunit)

        # check that the (model, chain) pair has a place in the model memory
        key = (model.name, sim.chain_index)
        ind = self._key_check(key, sim.level)
        self._size_check(numpy.max(tind, axis=0), ind)
        self._mem[tind[:,0],tind[:,1],tind[:,2],ind] = enddate
        self._time_steps[tind[:,0],tind[:,1],tind[:,2],ind] = timesteps

        return True

    def release_blocked_objects(self, sim, depthind):
        """Release objects that were blocked from evaluation because the
        model had active results stored in the memory
        """
        if self._blocked is not None:
            sim.data.release(sim.level, depthind, self._blocked)
            self._blocked = None

    def reset(self):
        """Reset prediction memory
        """
        self._mem[:,:,:,:] = datetime.date(1,1,1)
        self._time_steps[:,:,:,:] = 1.

    def del_objects(self, it, br, objs, level):
        """Delete objects from prediction memory (reset given objects)
        """
        if self._level_map.has_key(level):
            inds = list(self._level_map[level])
            if len(objs) > 0:
                tind = numpy.zeros((len(objs), 3), dtype=int)
                tind[:,0] = it
                tind[:,1] = br
                tind[:,2] = objs
                self._size_check(numpy.max(tind, axis=0), max(inds))
                for i in inds:
                    self._mem[it,br,objs,i] = datetime.date(1,1,1)
                    self._time_steps[it,br,objs,i] = 1.

    def add_branch(self, par_obj, new_obj, level):
        """Update prediction memory after adding a new branch to simulation
        data matrix

        par_obj -- parent object row as a tuple or numpy array
        new_obj -- new object row as a tuple or numpy array
        """
        if level in self._level_map:
            # make room to the prediction memory array
            self._size_check(new_obj, 0)
            # parent objects might be outside mem limits as it is possible
            # that the objects' prediction memory has not been used yet
            self._size_check(par_obj, 0)
            p_it, p_br, p_obj = par_obj
            n_it, n_br, n_obj = new_obj
            # copy only the current level indices to the target object
            inds = tuple(self._level_map[level])
            self._mem[n_it,n_br,n_obj,inds] = \
                numpy.copy(self._mem[p_it,p_br,p_obj,inds])
            self._time_steps[n_it,n_br,n_obj,inds] = \
                numpy.copy(self._time_steps[p_it,p_br,p_obj,inds])

    def _match_spans(self, sim, resunit, simunit, timespan, timesteps):
        """
        Match the time span and time steps for prediction model result
        variable when the result variable time unit and simulation time
        unit differ

        resunit -- prediction model result variable time unit, string
        simunit -- simulation time span time step unit, string
        timespan -- result object time span length, float
        timesteps -- time steps of main level objects, numpy array
        """
        factor = 1.
        if simunit != resunit:
            invalid = False
            if resunit == names.YEAR:
                if simunit == names.YEAR:
                    pass  # no need to match the spans
                elif simunit == names.MONTH:
                    factor = 1. / 12.
                else:
                    invalid = True
            elif resunit == names.MONTH:
                if simunit == names.YEAR:
                    factor = 12.
                elif simunit == names.MONTH:
                    pass  # no need to match the spans
                else:
                    invalid = True
            else:
                invalid = True
            if invalid:
                msg = "invalid result variable time span unit '%s'" % resunit
                sim.add_error(msg, sim.level)
                return timesteps, None

        timesteps *= factor

        if 0.0 in timesteps:
            # NB: if timestep is 0.0, there might be problems in dividing
            # prediction result (division by zero!). Use 1.0 instead
            timedivider = timesteps.copy()
            timedivider[timedivider == 0.0] = 1.0
        else:
            timedivider = timesteps

        return timesteps, timedivider

    def _divide_results(self, arg, model, sim, preddiv, timesteps):
        """Divide the result values between timesteps in the timespan.
        Results can be cumulated over multiple years in alternative ways,
        which is controlled with cumulation type attribute. The cumulation
        types are:
        - once: the result is calculated once, do nothing
        - total: the value is divided with result timespan / timestep
        - total%: see _totalpercentage method
        - annual: the result is multiplied with simulation timestep
        - annual%: the same as above, but for percentages
        - totalprob: ...
        """
        for i in range(arg.num_of_res_vars):
            if numpy.any(arg.num_of_res_objs > 1):
                errorstr = "Memory model should return results for a single "\
                    "object only!"
                sim.add_error(errorstr, None, None)
                return False
            values = arg.mem[i,:]
            cumul = arg.result_vars[i].cumulation
            if cumul == 'once':
                pass
            elif cumul == 'total':
                values = self._total(values, preddiv)
            elif cumul == 'total%':
                values = self._totalpercentage(values, preddiv, sim)
            elif cumul == 'annual':
                values *= timesteps
            elif cumul == 'annual%':
                values = ((1 + values / 100) ** timesteps - 1) * 100
            elif cumul == 'totalprob':
                values = self._totalprob(values, preddiv)
            arg.mem[i,:] = values
        return True

    def _total(self, value, preddiv):
        """Divide value with prediction divider
        """
        return value / preddiv

    def _totalpercentage(self, value, preddiv, sim):
        """Calculate total percentage value
        """
        return ((1.0 + 0.01 * value) ** (1.0 / preddiv) - 1.0) * 100.0

    def _totalprob(self, value, preddiv):
        """Calculate total removal probability over a period of time
        """
        return value ** (1.0 / preddiv)
        """
        Vanha toteutus:
        # total removal probability over a period of time
        # i.e. population left at the end = (1-p)InitPop
        # assumption: probability expressed as [0..1]
        # construct the polynomial
        # e.g. xpow2-2x+value for two period case
        # first get the coefficients from the pascal's
        # triangle
        coeff = utils.pascal(preddiv)
        coeff[0] = value
        for i in range(len(coeff)):
            if i % 2 <> 0:
                coeff[i] =- coeff[i]
        # initial guess for the root and the guess based
        #  iteration step
        x = float(value) / float(preddiv)
        denominator = 100.9 - 454.5 * x
        step = x / denominator
        iterate = True
        x -= step
        while iterate:
            x += step
            polyval = 0
            for i in range(len(coeff)):
                polyval = polyval + coeff[i] * math.pow(x, i)
            # stepped over the root, lets go back
            if polyval < 0:
                # using finer steps
                step = step / 10
                while polyval < 0:
                    x -= step
                    polyval = 0
                    for i in range(len(coeff)):
                        polyval = polyval + coeff[i] * math.pow(x, i)
                iterate = False
        return x
        """


def main():
    pass

if __name__ == '__main__':
    main()
