#!/usr/bin/env python
# encoding: utf-8
"""
Title: SIMO simulator condition evaluator

Module for evaluating conditions in SIMO simulator.

*condeval.py*

Copyright (c) 2007 Simosol Oy. Distributed under the GNU General Public License 2.0.
"""

import sys
import numpy
from collections import deque

from simo.utils.utils import update_dates
from simo.builder.names import EXPR_DISCOUNT, EXPR_DTYPE_DATA,\
                               EXPR_DTYPE_DATA, EXPR_DTYPE_OP,\
                               EXPR_DTYPE_VALUE, EXPR_OPERATION,\
                               EXPR_DTYPE_FUNC, EXPR_FUNCTION,\
                               EXPR_DTYPE_LEVEL




class ConditionEvaluator(object):
    """Class for evaluating conditional expressions in postfix form"""
    def __init__(self, add_error):
        """
        Initialize condition evaluator
        """
        self._add_error = add_error

    def cleanup(self):
        try:
            del self._add_error
            del self._data
            del self._tind
            del self._opmem
            del self._timespan
            del self._stack
            del self._failed
            del self._error_stack
            del self._target_indices
        except Exception:
            pass

    def evaluate(self, data, tind, expr, level, main_level, depth, opmem,
                 timespan, throughlevel=None, lowestlevel=None):
        """
        Evaluate conditional expression in post fix. Values are retrieved
        from data matrix using simo.builder.matrix.handler.Handler
        instance
        """
        self._data = data
        self._tind = tind
        self._level = level
        self._main_level = main_level
        self._depth = depth
        self._opmem = opmem
        self._timespan = timespan
        self._through_level = throughlevel
        self._lowest_level = lowestlevel
        self._stack = deque([])
        self._failed = set([])
        self._error_stack = []
        self._target_indices = {
            (level,None,None,True,None): (tind, set([]), data.get_active())
        }
        expected_stack_size = tind.shape[0]
        # go through each item (operand) in the postfix stack
        for i in range(len(expr)):
            if expr[i] is None:
                self._stack.append(None)
                continue
            itemtype = expr[i][0]
            if itemtype == EXPR_DTYPE_DATA:
                if len(expr) >= i+2+1 and expr[i+2][0] == 'ex':
                    self._get_data_exists(expr[i][1])
                else:
                    if not self._get_data_values(expr[i][1]):
                        return numpy.zeros(tind.shape[0], dtype=bool)
            elif itemtype == EXPR_DTYPE_OP:
                if not self._get_operation_values(expr[i][1], expr[i+2][0]):
                    return numpy.zeros(tind.shape[0], dtype=bool)
            elif itemtype == EXPR_DTYPE_VALUE:
                self._stack.append(expr[i][1])
            elif itemtype == EXPR_DTYPE_LEVEL:
                self._get_level_exists(expr[i][1])
            elif itemtype == EXPR_DTYPE_FUNC:
                self._stack.append(expr[i][1](len(tind)))
            else:
                if len(self._stack) < 2:
                    raise Exception, 'Error: Too few input values'
                else:
                    b = self._stack.pop()
                    a = self._stack.pop()
                    # handle arrays of inequal length
                    if hasattr(a, 'shape') and hasattr(b, 'shape'):
                        if len(a) < len(b):
                            a = self._fill_nans(a)
                        elif len(b) < len(a):
                            b = self._fill_nans(b)
                    self._stack.append(expr[i][1](a, b))
        self._data = None
        if len(self._stack) > 1:
            raise Exception, "Too many input values!"
        else:
            # if some of the target objects failed due to missing values,
            # set the evaluation result for those as false
            if len(self._failed) > 0:
                stack_size = self._stack[0].shape[0]
                # in some cases the stack_size may omit the _failed elements
                valid_sizes = (expected_stack_size, expected_stack_size - len(self._failed))
                if stack_size not in valid_sizes:
                    raise Exception, "Stack size deviates from expected"
                self._failed = list(self._failed)
                self._failed.sort(reverse=True)
                stack_max_ind = self._stack[0].shape[0] - 1
                orig_max = stack_max_ind
                resized = False

                for f_ind in self._failed:
                    # resize stack if items missing
                    if f_ind > stack_max_ind:
                        shp = (f_ind + 1, )
                        self._stack[0] = numpy.resize(self._stack[0], shp)
                        stack_max_ind = f_ind
                        resized = True

                    if not resized or f_ind >= orig_max:
                        self._stack[0][f_ind] = False
                    elif resized:
                        # adding failed results between really evaluated ones
                        for ind in range(orig_max, f_ind - 1, -1):
                            self._stack[0][ind + 1] = self._stack[0][ind]
                        self._stack[0][f_ind] = False

            if len(self._error_stack) > 0:
                for errorstr, datalevel, errorobjs in self._error_stack:
                    self._add_error(errorstr, datalevel, errorobjs)
            return self._stack[0]

    def _fill_nans(self, mod_me):
        """
        Fill nan values for those tind objects that failed when collecting
        values

        mod_me -- value array to modify
        """
        to_insert = list(self._failed)
        to_insert.sort(reverse=True)
        if isinstance(mod_me[0], bool):
            fail_value = False
        else:
            fail_value = numpy.nan
        for fail in to_insert:
            pieces = (mod_me[:fail], [fail_value], mod_me[fail:])
            mod_me = numpy.concatenate(pieces)
        return mod_me

    def _get_level_exists(self, level):
        """
        Check if there are any objects on given level
        """
        datalevel = level if level != self._level else None
        tind, torem = self._get_targets(datalevel, None, True)
        if level > self._level:
            start, end = tind[:,3], tind[:,4]
            exists = (end - start) > 0
        else:
            exists = self._data.get_object_map(level, tind)
        self._stack.append(exists)

    def _get_data_exists(self, item):
        """
        Get variable exists from simulation data matrix
        """

        datalevel, variable, vals4all = item
        if datalevel == self._level:
            datalevel = None
        tind, torem = self._get_targets(datalevel, None, vals4all)
        exists = self._data.get_exists(tind, variable)
        self._stack.append(exists)

    def _get_data_values(self, item):
        """
        Get variable values from simulation data matrix
        """
        # get operand datalevel, variable indice and "values for all object"
        # -indicator
        datalevel, variable, vals4all = item
        # when datalevel equals evaluation level, through -level is not
        # defined and condition lowest is not defined, datalevel can be
        # ignored, which makes target index and value retrieval much more
        # efficient
        if datalevel == self._level and self._through_level is None \
           and self._lowest_level is None:
            datalevel = None
        # by default, valuelevel is None as it is used only in aggregation
        # conditions when condition level is lower in the hierarchy than
        # evaluation level, but some of the condition operands are on the
        # evaluation level or some level above the lowest level
        valuelevel = None
        if self._lowest_level is not None:
            if datalevel != self._lowest_level:
                # the value level should be 'self' (the same as the current
                # model chain evaluation level)
                if datalevel != self._level:
                    errorstr = "condition operand level does not equal " \
                               "either lowest operand level or evaluation "\
                               "level ('self')"
                    self._error_stack.append((errorstr, None, None))
                    return False
                valuelevel = datalevel
                datalevel = self._lowest_level
        # get target index for active current level objects
        tind, torem = self._get_targets(datalevel, valuelevel, vals4all)
        if len(torem) > 0 and datalevel <= self._level:
            # if the tind comes from lower than evaluation level, torem
            # doesn't matter, it's in effect already in active_data_objects
            self._failed.update(torem)
        values, err = self._data.get_value(tind, variable)
        if values is None:
            self._failed.update(range(tind.shape[0]))
            return False
        if err is not None:
            # log the error messages and update failed object set
            # NB: failed objects should always be evaluation level objects
            errorstr = " when evaluating condition"
            for e, objs in err[0]:
                self._error_stack.append((e+errorstr, datalevel, err[1][objs]))
            self._failed.update(err[2])
        self._stack.append(values)
        return True

    def _get_targets(self, datalevel, valuelevel, vals4all):
        """
        Get targets for current operand level objects. As handler's
        get_tind method is computationally expensive, it should be avoided
        by storing already fetched target indices and reusing them when
        possible.
        """
        key = (self._level,
              datalevel,
              self._through_level,
              vals4all,
              valuelevel)
        if key not in self._target_indices:
            tind, torem = self._data.get_tind(self._level, self._depth,
                                              datalevel, self._through_level,
                                              vals4all, valuelevel)
            active = self._data.get_active()
            self._target_indices[key] = (tind, torem, active)
        else:
            tind, torem, active = self._target_indices[key]
            self._data.set_active(active)
        return tind, torem

    def _get_operation_values(self, opind, functype):
        """
        Get operation timing related values from operation memory
        """
        if not functype in set(['since', 'times']):
            # when condition atom includes operation (since or times operator),
            # next funtion type should always be 'since' or 'times'
            raise Exception, "invalid operation operator '%s'" % functype
        tind = self._tind
        if self._level is not None and self._level > self._main_level:
            # evaluation level is a child level of main level (where the
            # dates are...)
            tind, torem = self._data.get_parent(self._level, self._tind,
                                                self._main_level)
        if opind is not None:
            if functype == 'since':
                dates = self._data.get_date(tind)
                up = update_dates(dates,
                                  self._timespan.time_step,
                                  self._timespan.unit,
                                  deduceaday=True)
                values = self._opmem.since(tind, opind, up,
                self._timespan.unit)
            else:
                values = self._opmem.times(tind, opind)
        else:
            # this is a special case when there are no operation models
            # in the simulator at all
            values = numpy.zeros(tind.shape[0], dtype=int)
            if functype == 'since':
                values.fill(sys.maxint)
        self._stack.append(values)
        return True


def main():
    pass

if __name__ == '__main__':
    main()

