#!/usr/bin/env python
# encoding: utf-8
"""
Title: SIMO operation model argument container

Operationmodel argument container for SIMO Simulator model library interface.
Operation parameters are passed to the operationmodel library with a operArg
container, where the parameters are stored as 'ctypes' objects so they can be
passed to the C language library.

*operationarg.py*

Copyright (c) 2007 Simosol Oy. Distributed under the GNU General Public License 2.0.
"""

import numpy
from collections import deque

class OperationArg(object):
    """Container class for passing the operation model arguments to the operation
    models"""

    def __init__(self, name, group, db_name, db_group, optype, lang, sim):
        """Initialize a operation_arg container instance"""
        self.name = name
        self.group = group
        self.db_name = db_name
        self.db_group = db_group
        self.type = optype
        self.language = lang
        self.sim = sim
        self.target_index = None
        self.num_of_targets = None
        self.num_of_results = None
        self._cash_flow = None
        self._errors = None
        self._warnings = None
        self._data = None
        self._variables = None
        self._params = None
        self._param_table_vals = None
        self._results = None
        self._weights = None
        self.cash_flow_handler = None
        self.cash_flow_model = None
        self.common_cash_flow_table = None
        self.use_common_cash_flow_table = False
        self.cash_flow_tables = None
        self.cash_flow_table_names = None
        self.result_type = None
        self.result_level = None
        self.result_variables = None
        self.sim_effects = None
        self.aborted = None  # Target objects that were aborted (no errors)
        self.failed = None  # Target objects that failed because of errors

    def set_targets(self, tind, torem):
        """Set the operation model target index
        """
        self.num_of_targets = tind.shape[0]
        self.drop_target = numpy.zeros(self.num_of_targets, dtype=bool)
        self.target_index = tind
        self._cash_flow = numpy.zeros((self.num_of_targets, 1), dtype=float)
        self._errors = [list([]) for x in range(self.num_of_targets)]
        self._warnings = [list([]) for x in range(self.num_of_targets)]
        self.failed = [False for x in range(self.num_of_targets)]
        self.aborted = [False for x in range(self.num_of_targets)]

    def set_cash_flow(self, cashflow, i=None):
        """Set cash flow values either for single target or for all targets
        """
        self._cash_flow[i] = cashflow

    def get_cash_flow(self, i=None):
        """Get cash flow container for target object i or for all targets
        """
        if i is None:
            return self._cash_flow[:,0]
        return self._cash_flow[i]

    def get_errors(self, i):
        """Get error container for target object i"""
        return self._errors[i]

    def get_warnings(self, i):
        """Get warning container for target object i"""
        return self._warnings[i]

    def set_vars(self, nvar, values, loc, torem):
        """Set operation input variable values"""
        self.drop_target[list(torem)] = True
        if self._variables is None:
            self._variables = numpy.zeros((nvar, self.num_of_targets),
                                           dtype=float)
        self._variables[loc,:] = values

    def get_vars(self, i):
        """Get input variable values for target object i"""
        if self._variables is None:
            return None
        return self._variables[:,i]

    def set_params(self, param):
        """Store input parameter value to model arguments"""
        if self._params is None:
            self._params = deque()
        self._params.append(param)

    def get_params(self):
        """Get operation input parameter values"""
        if self._params is None:
            return None
        return self._params
    
    def init_cash_flow_tables(self, cash_flow_table):
        """Initialize containers for cash flow tables
        """
        cft = cash_flow_table
        trange = range(self.num_of_targets)
        self.common_cash_flow_table = [None for x in cft]
        self.cash_flow_table_names = [None for x in cft]
        self.cash_flow_tables = [[None for y in cft] for x in trange]

    def set_cash_flow_table(self, table, index, target=None):
        """Add a cash flow table to operation arguments"""
        if self.cash_flow_table_names[index] is None:
            self.cash_flow_table_names[index] = table.name
        if target is None:
            self.common_cash_flow_table[index] = table
            self.use_common_cash_flow_table = True
        else:
            self.cash_flow_tables[target][index] = table
            self.use_common_cash_flow_table = False
        if self.sim.track_prices and table.persist:
            self.sim._data_db.store_timber_prices(self.sim, target, table)

    def get_cash_flow_table(self, target=None):
        """Get cash flow table for either a specific target object or common
        cash flow table
        """
        if self.cash_flow_tables is None:
            return None
        if target is None and not self.use_common_cash_flow_table:
            target = 0
        if self.use_common_cash_flow_table:
            return self.common_cash_flow_table
        else:
            return self.cash_flow_tables[target]

    def set_weights(self, values, torem):
        """Set operation result weight"""
        self.drop_target[list(torem)] = True
        self._weights = values

    def get_weights(self, i=None):
        """Get operation weights for target object i or all targets"""
        if self._weights is None:
            return None
        if i is None:
            return self._weights
        return self._weights[i]

    def set_result_structure(self, restarget, reslevel, restype, nres,
                             resvars, results, torem):
        """Create operation result structure with given number of rows and
        columns and the result labels
        """
        self.drop_target[list(torem)] = True
        self.result_target = restarget
        self.result_level = reslevel
        self.result_type = restype
        self.num_of_results = nres
        self.result_variables = resvars
        self._results = results

    def get_results(self, i=None):
        """Get result structure for target object i or for all targets
        """
        if i is None:
            return self._results
        if self._results is None:
            return None
        return self._results[i]

    def set_param_table_values(self, values, obj, table, ntables):
        """Store parameter table values for a single object
        """
        if self._param_table_vals is None:
            rng = range(ntables)
            self._param_table_vals = \
                    [[[] for y in rng] for x in range(self.num_of_targets)]
        self._param_table_vals[obj][table].append(values)

    def get_param_table_vals(self, i):
        """Get parameter table values for object i
        """
        if self._param_table_vals is None:
            return None
        return self._param_table_vals[i]

    def set_data(self, data, index, level, attrs):
        """Store input data structure from single level
        """
        if self._data is None:
            self._data = {}
        self._data[level] = [data, index, attrs]

    def get_data(self, i):
        """Get input data structure for single object
        """
        if self._data is None:
            return None
        data = {}
        for level in self._data.keys():
            d, ind, attrs = self._data[level]
            if i >= ind[:,0].size:
                return None  # invalid indice, return None
            s,e = ind[i,:]
            data[level] = d[s:e,:]
        return data

    def check_eval(self, evalres):
        """Check operation model evaluation result

        Check the evaluation result for all result targets.
            1 = model successfully evaluated
            0 = warning(s), model evaluated
            -1 = error(s), model not evaluated
            -2 = model aborted

        evalres -- evaluation result vector
        """
        if numpy.all(evalres == 1):
            # Operation was successfully evaluated for all targets, do nothing
            return
        # Check the objects which were not evaluated successfully
        # (either because of errors or because the operation was aborted)
        level = self.sim.level
        ind = numpy.where(evalres == 0)[0]
        if len(ind) > 0:
            self.sim.add_warning('Group %s Model %s '\
                                 'evaluated with warning(s)!' % \
                                 (self.group, self.name), level,
                                 self.target_index[ind])
            self._add_warnings(ind, level, self.target_index[ind])
        ind = numpy.where(evalres == -1)[0]
        if len(ind) > 0:
            self.sim.add_error('Group %s Model %s failed to evaluate '\
                               'due to error(s)!' % (self.group, self.name),
                               level, self.target_index[ind])
            self._add_errors(ind, level, self.target_index[ind])
        ind = numpy.where(evalres == -2)[0]
        if len(ind) > 0:
            self.sim.add_warning('Group %s Model %s aborted due to '\
                                 'user set limits.' % \
                                 (self.group, self.name), level,
                                 self.target_index[ind])
            self._add_aborts(ind)

    def _add_aborts(self, ind):
        for i in ind:
            self.aborted[i] = True
            self.drop_target[i] = True

    def _add_warnings(self, ind, level, tind):
        for i in ind:
            if self._warnings[i] is not None:
                for j in self._warnings[i]:
                    self.sim.add_warning(j, level, tind)

    def _add_errors(self, ind, level, tind):
        for i in ind:
            self.failed[i] = True
            self.drop_target[i] = True
            if self._errors[i] is not None:
                for j in self._errors[i]:
                    self.sim.add_error(j, level, tind)


def main():
    pass

if __name__ == '__main__':
    main()
