#!/usr/bin/env python
# encoding: utf-8
"""
Title: SIMO prediction model argument container

Prediction model argument container for SIMO Simulator model library interface.
Model parameters are passed to the modellibrary with a PredictionArg container,
where the parameters are stored as 'ctypes' objects so they can be passed to the
C language library.

*predictionarg.py*

Copyright (c) 2007 Simosol Oy. Distributed under the GNU General Public License 2.0.
"""

import numpy
import ctypes
from collections import deque


class PredictionArg:
    """Container class for passing the prediction model arguments to
    prediction library"""
    
    def __init__(self, name, lang):
        """Initialize a prediction model argument container"""
        self.variables = None
        self.parameters = deque()
        self.result_vars = None
        self.error_check_mode = 1
        self.allowed_risk_level = 1
        self.rect_factor = 1.0
        self.model_name = name
        self.model_language = lang
        self.warnings = deque()
        self.errors = None
        self.error_objects = deque()
        self.output_type = None
        self.max_res_vals = None
        self.mem = None
        self.result_level = None
        self.target_index = None
        self.num_of_targets = None
        self.num_of_res_vars = None
        self.max_res_objs = None
        self.num_of_res_objs = None
        self.remove_targets = set([])
        self.warning_flag = False
        self.error_flag = False

    def set_warnings(self, warnings):
        """Store occurred warnings to the argument container"""
        self.warning_flag = True
        # NB: Creating Numpy arrays from lists that include both None values
        # and strings may cause program to crash, maybe due to Numpy version
        self.warnings = numpy.array(warnings, dtype=object)

    def get_warnings(self):
        """Get stored warning messages"""
        return self.warnings

    def add_error(self, i):
        """Add error to target object i"""
        self.error_objects.append(i)
        self.error_flag = True

    def drop_error_objects(self):
        """Trim the prediction model result structures to only contain valid
        results; i.e., no error in model call.
        Empty error strings are also removed.
        """
        tobj = numpy.arange(self.num_of_targets)
        self.target_index = numpy.delete(self.target_index,
                                         self.error_objects, axis=0)
        tobj = numpy.delete(tobj, self.error_objects, axis=0)
        self.num_of_targets = self.target_index[:,0].size
        self.mem = numpy.delete(self.mem, self.error_objects, axis=1)
        self.num_of_res_objs = \
            numpy.array([self.num_of_res_objs[x] for x in tobj])
        if numpy.any(self.errors == ''):
            ind = numpy.where(self.errors == '')[0]  # remove empty error strings
            self.errors = numpy.delete(self.errors, ind, axis=0)
        
    def set_targets(self, tind, torem, maxresobjs):
        """Store target object indices in the argument container"""
        if torem is not None:
            self.remove_targets.update(torem)
        self.max_res_objs = maxresobjs
        self.num_of_targets = tind.shape[0]
        self.target_index = tind
        
    def set_result_level(self, result_level, cur_level):
        """Set model result level"""
        self.result_level = result_level
        if self.result_level == 'self':
            self.result_level = cur_level

    def set_input_variables(self, nvar, values, loc, rem_targets):
        """Store input variable values into the container"""
        self.remove_targets.update(rem_targets)
        if self.variables is None:
            self.variables = numpy.zeros((nvar, self.num_of_targets))
        self.variables[loc,:] = values

    def set_result_variables(self, resvars):
        """Store result variable indices"""
        self.num_of_res_vars = len(resvars)
        self.result_vars = resvars

    def add_parameter(self, param):
        """Add input parameter value to model arguments"""
        self.parameters.append(param)
        
    def set_containers(self):
        """Construct containers for passing results back from the model"""
        if self.num_of_res_vars is None:
            raise Exception("Number of result variables not defined!")
        if self.max_res_objs is None:
            raise Exception("Number of result objects not defined!")
        self.max_res_vals = int(self.num_of_res_vars * self.max_res_objs)
        self.num_of_res_objs = numpy.zeros(self.num_of_targets, dtype=int)
        self.mem = numpy.zeros((self.max_res_vals, self.num_of_targets),
                               dtype=float)
        self.errors = ['' for x in range(self.num_of_targets)]
    
    def remove_missing(self):
        """Remove objects with missing input variable values
        """
        if len(self.remove_targets) == 0:
            return
        torem = list(self.remove_targets)
        self.variables = numpy.delete(self.variables, torem, axis=1)
        self.num_of_res_objs = numpy.delete(self.num_of_res_objs,torem,axis=0)
        self.mem = numpy.delete(self.mem, torem, axis=1)
        self.errors = numpy.delete(self.errors, torem, axis=0)
        self.num_of_targets -= len(torem)
        if self.target_index.size > 0:
            self.target_index = numpy.delete(self.target_index, torem, axis=0)
        self.errors = ['' for x in range(self.num_of_targets)]

    def ctypes2python(self):
        """Change argument types from ctypes to python and remove objects with
        evaluation errors from target objects"""
        # construct a list of errors for all objects (that should have been
        # evaluated) for the error processing later on (proceed is True)
        self.errors = numpy.array([str(e.value) for e in self.errors])
        # process the memory area and other ctypes structures
        rng = range(self.num_of_targets)
        mem = [[float(self.mem[y][x]) for y in rng] for x in range(len(self.mem[0]))]
        self.mem = numpy.array(mem)
        arr = [int(self.num_of_res_objs[x].value) for x in rng]
        self.num_of_res_objs = numpy.array(arr)
        self.allowed_risk_level = float(self.allowed_risk_level.value)
        self.error_check_mode = float(self.error_check_mode.value)
        self.rect_factor = float(self.rect_factor.value)

    def python2ctypes(self):
        """Change argument types from python data types to c data types
        """
        c_int = ctypes.c_int
        c_double = ctypes.c_double
        c_char = ctypes.c_char
        rng = range(self.num_of_targets)
        # make sure that there are variables
        if self.variables is not None and self.variables.shape != ():
            nobj, nvar = self.variables.shape
            vals = self.variables
            self.variables = self._const_vars(vals, nvar)
        self.error_check_mode = c_int(self.error_check_mode)
        self.num_of_res_objs = self._const_nro(rng)
        self.rect_factor = c_double(self.rect_factor)
        self.allowed_risk_level = c_double(self.allowed_risk_level)
        self.parameters = self._const_params()
        self.mem = self._const_mem()
        self.errors = self._const_errors(rng)
    
    def _const_vars(self, vals, nvar):
        """Construct a 2D list of C double type objects
        
        NB: low performance!
        """
        c_double = ctypes.c_double
        new_vals = vals.transpose().tolist()[:nvar]
        return [[c_double(y) for y in v] for v in new_vals]
    
    def _const_nro(self, rng):
        """Construct number of result objects C integer list
        """
        c_int = ctypes.c_int
        return [c_int() for x in range(self.num_of_targets)]
    
    def _const_params(self):
        """Construct parameters C double list
        """
        c_double = ctypes.c_double
        params = self.parameters
        return  [c_double(x) for x in params]
    
    def _const_mem(self):
        """Construct model result memory C double list
        """
        c_double = ctypes.c_double
        mrv = self.max_res_vals
        return [(mrv * c_double)() for x in range(self.mem.shape[1])]
    
    def _const_errors(self, rng):
        """Construct model error C char list
        """
        c_char = ctypes.c_char
        return [(c_char * 200)() for x in range(self.num_of_targets)]
    
    def __construct_ctypes_array(self,valuearray):
        """structure. These arrays are used for passing data to the DLL model
        libraries. Uses ctypes package for the variable types
        """
        c_double = ctypes.c_double
        n = len(valuearray)
        carray = (c_double * len(valuearray))()
        for ival in xrange(n):
            carray[ival] = c_double(valuearray[ival])
        return carray


def main():
    pass

if __name__ == '__main__':
    main()
