#!/usr/bin/env python
# encoding: utf-8
"""
Class Simulator instance forms a simulator object for given simulation task.
Each simulator has input data set, a set of simulation parameters, and a
set of model and operation chains that define the simulation process.

*sim.py*

Copyright (c) 2007 Simosol Oy. Distributed under the GNU General Public
License 2.0.
"""

# these are for running the doctest from src/simo
import sys
sys.path.append('..')

#import math
import numpy
import time
import math
from datetime import date
from collections import defaultdict
import signal
from copy import deepcopy
from dateutil.relativedelta import relativedelta
from operator import attrgetter

import simo.builder.names as names
from simo.utils.utils import update_dates, calculate_date_diff, secs2str, \
        signum_to_msg
from simo.matrix.handler import Handler
from simo.simulation.condeval import ConditionEvaluator
from simo.simulation.caller.aggregationcaller import AggregationModelCaller
from simo.simulation.caller.geotablecaller import GeoTableCaller
from simo.simulation.caller.managementcaller import ManagementModelCaller
from simo.simulation.caller.operationcaller import OperationModelCaller
from simo.simulation.caller.paramtablecaller import ParamTableCaller
from simo.simulation.caller.predictioncaller import PredictionModelCaller
from simo.simulation import cache

class Simulator(object):
    """SIMO simulator class"""

    def __init__(self, logger, log_name, sim_type, lexicon, ctrl,
                 input_db, data_db, operation_db, warnings, max_unit_count,
                 obj_count_multiplier, branch_count, no_op_branch,
                 separate_branching_groups, branch_limit, max_res_objs,
                 max_errors, max_simunit_errors, deterministic, track_prices,
                 do_forced_ops, copy2iters, create_branch_desc, cache_dir):
        """Construct simulator instance

        logger -- logger instance
        logname -- log name as string
        sim_type -- 'simulation' (results to result db) or 'data processing'
                    (results to input db)
        lexicon -- lexicon instance
        ctrl -- simulation control instance
        input_db -- input database instance
        data_db -- data result database instance
        opres_db -- operation result database instance
        operation_db -- operation db instance for forced operations
        max_unit_count -- maximum main level unit count as int
        obj_count_multiplier -- used in data matrix creation to create some
                                extra space to reduce need for matrix resizing
        branch_count -- how many branches are created at once in the matrix
        no_op_branch -- should branching always generate a no-operation
                        branch also?
        separate_branching_groups -- should "no_op_branch" affect all branching
                                     groups simultaneously, or each one
                                     separately
        branch_limit -- maximum number of branches per unit
        max_res_objs -- maximum number of result object
        max_errors -- maximum total error count
        max_simunit_errors -- maximum error count per simulation unit
        deterministic -- if set to True, blocks every random number
                         aggregation model call to return static result
        track_prices -- if set to True, timber prices per main unit are
                        written to the result database 'timber_price' table
        do_forced_ops -- if set to False, will ignore any forced ops
        copy2iters -- copy a single set of forced operations per unit to all
                      iterations (if not, each iteration should have it's own
                      set of forced operations in the db per each unit)
        cache_dir -- where to write the price matrix and stem volume caches
        """
        super(Simulator, self).__init__()
        self._saved_handlers = None
        self.do_init(logger, log_name, sim_type, lexicon, ctrl,
                     input_db, data_db, operation_db, warnings, max_unit_count,
                     obj_count_multiplier, branch_count, no_op_branch,
                     separate_branching_groups, branch_limit, max_res_objs,
                     max_errors, max_simunit_errors, deterministic,
                     track_prices, do_forced_ops, copy2iters,
                     create_branch_desc, cache_dir, first=True)

    def do_init(self, logger, log_name, sim_type, lexicon, ctrl,
                input_db, data_db, operation_db, warnings, max_unit_count,
                obj_count_multiplier, branch_count, no_op_branch,
                separate_branching_groups, branch_limit, max_res_objs,
                max_errors, max_simunit_errors, deterministic,
                track_prices, do_forced_ops, copy2iters,
                create_branch_desc, cache_dir, first):
        """Construct simulator instance

        logger -- logger instance
        logname -- log name as string
        sim_type -- 'simulation' (results to result db) or 'data processing'
                    (results to input db)
        lexicon -- lexicon instance
        ctrl -- simulation control instance
        input_db -- input database instance
        data_db -- data result database instance
        opres_db -- operation result database instance
        operation_db -- operation db instance for forced operations
        max_unit_count -- maximum main level unit count as int
        obj_count_multiplier -- used in data matrix creation to create some
                                extra space to reduce need for matrix resizing
        branch_count -- how many branches are created at once in the matrix
        no_op_branch -- should branching always generate a no-operation
                        branch also?
        separate_branching_groups -- should "no_op_branch" affect all branching
                                     groups simultaneously, or each one
                                     separately
        branch_limit -- maximum number of branches per unit
        max_res_objs -- maximum number of result object
        max_errors -- maximum total error count
        max_simunit_errors -- maximum error count per simulation unit
        deterministic -- if set to True, blocks every random number
                         aggregation model call to return static result
        track_prices -- if set to True, timber prices per main unit are
                        written to the result database 'timber_price' table
        do_forced_ops -- if set to False, will ignore any forced ops
        copy2iters -- copy a single set of forced operations per unit to all
                      iterations (if not, each iteration should have it's own
                      set of forced operations in the db per each unit)
        cache_dir -- where to write the price matrix and stem volume caches
        first -- if this was not the first init, ensure old cache is closed
                 first
        """
        self._logger = logger
        self._log_name = log_name
        self._sim_type = sim_type
        self._warnings = warnings
        self.lexicon = lexicon
        self._ctrl = ctrl
        self._input_db = input_db
        self._data_db = data_db
        self._operation_db = operation_db
        self._evaluator = ConditionEvaluator(self.add_error)
        self._callers = {
            names.AGGREGATION: AggregationModelCaller(deterministic),
            names.GEO_TABLE: GeoTableCaller(),
            names.MANAGEMENT: ManagementModelCaller(),
            names.OPERATION: OperationModelCaller(warnings),
            names.PARAMETER_TABLE: ParamTableCaller(),
            names.PREDICTION: PredictionModelCaller(warnings)
        }
        self.prediction_memory = self._callers[names.PREDICTION].memory
        self.operation_memory = self._callers[names.OPERATION].memory
        self._max_unit_count = max_unit_count
        self._objcount_multiplier = obj_count_multiplier
        self._branch_count = branch_count
        self._no_op_branch = no_op_branch
        self._separate_branching_groups = separate_branching_groups
        self._branch_limit = branch_limit
        self._max_res_objs = max_res_objs
        self._max_errors = max_errors
        self._max_simunit_errors = max_simunit_errors
        self.main_level = self._ctrl.main_level
        self.memory_models = self._ctrl.memory_models
        self.data = None
        self._ids = None
        self.period = 0
        self.timespan = None
        self.chain_index = None
        self._chain_type = None
        self._mccs = None
        self._forced_operations = None
        self._forced_chain_index = None
        self.level = None
        self._depth = None
        self._tasks = None
        self.task = None
        self.value_fixing = None
        self._block_growth = None
        self._block_all = None
        self._branching_operations = None
        self._branching_initialized = False
        self._branching_succeded = False
        self._opres_queue = []
        self._iterations = 1
        self._start_iter = 0
        self._end_iter = 1
        self._orig_time_steps = None
        self.terminate = False
        self._last_step_objs = None
        self._terminated = None
        self.any_terminated = False
        self._error_stack = []
        self._blocked_forced = numpy.array([], dtype=int)
        self._blocked_forced.shape = (0, 5)
        self._blocked_from_db = set([])
        self._last_step_objs = set([])
        self._release_for_updating = numpy.array([])
        self.track_prices = track_prices
        self._data_processing_date = None
        self.do_forced_ops = do_forced_ops
        self.copy2iters = copy2iters
        self.create_branch_desc = create_branch_desc
        if first:
            self.price_cache = cache.PriceCache(cache_dir, 'pricecache')
            self.stem_cache = cache.StemCache(cache_dir, 'stemcache')

        # set the signal handlers for keyboard interruptions etc.
        sigint = signal.signal(signal.SIGINT, self._handle_signal)
        sigterm = signal.signal(signal.SIGTERM, self._handle_signal)
        sigabrt = signal.signal(signal.SIGABRT, self._handle_signal)
        sigill = signal.signal(signal.SIGILL, self._handle_signal)
        if str.lower(sys.platform[:3]) == 'win':
            sigbreak = signal.signal(signal.SIGBREAK, self._handle_signal)
            if not self._saved_handlers:
                self._saved_handlers = (sigint, sigterm, sigabrt, sigill,
                                            sigbreak)
        elif not self._saved_handlers:
            self._saved_handlers = (sigint, sigterm, sigabrt, sigill)

    def cleanup(self):
        try:
            self.cleanup_handlers()
        except Exception:
            pass
        try:
            self._evaluator.cleanup()
        except Exception:
            pass
        try:
            if self.stem_cache.initialized:
                self.stem_cache.close()
        except Exception:
            pass
        try:
            if self.data:
                try:
                    del self.data.matrix.sim
                except Exception:
                    pass
                try:
                    del self.data.matrix
                except Exception:
                    pass
                del self.data
                self.data = None
        except Exception:
            pass

    def cleanup_handlers(self):
        """
        We need to cleanup handlers when done, or the Simulator instance is
        stuck in memory.
        """
        if self._saved_handlers:
            sh = self._saved_handlers
            signal.signal(signal.SIGINT, sh[0])
            signal.signal(signal.SIGTERM, sh[1])
            signal.signal(signal.SIGABRT, sh[2])
            signal.signal(signal.SIGILL, sh[3])
            if str.lower(sys.platform[:3]) == 'win':
                signal.signal(signal.SIGBREAK, sh[4])
            self._saved_handlers = None

    def re_init(self, logger, log_name, sim_type, lexicon, ctrl,
                input_db, data_db, operation_db, warnings, max_unit_count,
                obj_count_multiplier, branch_count, no_op_branch,
                separate_branching_groups,  branch_limit, max_res_objs,
                max_errors, max_simunit_errors, deterministic, track_prices,
                do_forced_ops, copy2iters, create_branch_desc, cache_dir):
        """Construct simulator instance

        logger -- logger instance
        logname -- log name as string
        sim_type -- 'simulation' (results to result db) or 'data processing'
                    (results to input db)
        lexicon -- lexicon instance
        ctrl -- simulation control instance
        input_db -- input database instance
        data_db -- data result database instance
        opres_db -- operation result database instance
        operation_db -- operation db instance for forced operations
        max_unit_count -- maximum main level unit count as int
        obj_count_multiplier -- used in data matrix creation to create some
                                extra space to reduce need for matrix resizing
        branch_count -- how many branches are created at once in the matrix
        no_op_branch -- should branching always generate a no-operation
                        branch also?
        separate_branching_groups -- should "no_op_branch" affect all branching
                                     groups simultaneously, or each one
                                     separately
        branch_limit -- maximum number of branches per unit
        max_res_objs -- maximum number of result object
        max_errors -- maximum total error count
        max_simunit_errors -- maximum error count per simulation unit
        deterministic -- if set to True, blocks every random number
                         aggregation model call to return static result
        track_prices -- if set to True, timber prices per main unit are
                        written to the result database 'timber_price' table
        do_forced_ops -- if set to False, will ignore any forced ops
        copy2iters -- copy a single set of forced operations per unit to all
                      iterations (if not, each iteration should have it's own
                      set of forced operations in the db per each unit)
        cache_dir -- where to write the price matrix and stem volume caches
        """
        self._logger = logger
        self._log_name = log_name
        self._sim_type = sim_type
        self._warnings = warnings
        self.lexicon = lexicon
        self._ctrl = ctrl
        self._input_db = input_db
        self._data_db = data_db
        self._operation_db = operation_db
        self._evaluator = ConditionEvaluator(self.add_error)
        self._callers = {
            names.AGGREGATION: AggregationModelCaller(deterministic),
            names.GEO_TABLE: GeoTableCaller(),
            names.MANAGEMENT: ManagementModelCaller(),
            names.OPERATION: OperationModelCaller(warnings),
            names.PARAMETER_TABLE: ParamTableCaller(),
            names.PREDICTION: PredictionModelCaller(warnings)
        }
        self.prediction_memory = self._callers[names.PREDICTION].memory
        self.operation_memory = self._callers[names.OPERATION].memory
        self._max_unit_count = max_unit_count
        self._objcount_multiplier = obj_count_multiplier
        self._branch_count = branch_count
        self._no_op_branch = no_op_branch
        self._separate_branching_groups = separate_branching_groups
        self._branch_limit = branch_limit
        self._max_res_objs = max_res_objs
        self._max_errors = max_errors
        self._max_simunit_errors = max_simunit_errors
        self.main_level = self._ctrl.main_level
        self.memory_models = self._ctrl.memory_models
        self.data = None
        self._ids = None
        self.period = 0
        self.timespan = None
        self.chain_index = None
        self._chain_type = None
        self._mccs = None
        self._forced_operations = None
        self._forced_chain_index = None
        self.level = None
        self._depth = None
        self._tasks = None
        self.task = None
        self.value_fixing = None
        self._block_growth = None
        self._block_all = None
        self._branching_operations = None
        self._branching_initialized = False
        self._branching_succeded = False
        self._opres_queue = []
        self._iterations = 1
        self._start_iter = 0
        self._end_iter = 1
        self._orig_time_steps = None
        self.terminate = False
        self._terminated = None
        self.any_terminated = False
        self._error_stack = []
        self._blocked_forced = numpy.array([], dtype=int)
        self._blocked_forced.shape = (0, 5)
        self._blocked_from_db = set([])
        self._last_step_objs = set([])
        self._release_for_updating = numpy.array([])
        self.track_prices = track_prices
        self.do_forced_ops = do_forced_ops
        self.copy2iters = copy2iters
        self.create_branch_desc = create_branch_desc
        self.price_cache = cache.PriceCache(cache_dir, 'pricecache')
        try:
            try:
                self.stem_cache.commit()
            except Exception:
                pass
            self.stem_cache.close()
        except Exception:
            pass
        self.stem_cache = cache.StemCache(cache_dir, 'stemcache')
        # set the signal handlers for keyboard interruptions etc.
        signal.signal(signal.SIGINT, self._handle_signal)
        signal.signal(signal.SIGTERM, self._handle_signal)
        signal.signal(signal.SIGABRT, self._handle_signal)
        if str.lower(sys.platform[:3]) == 'win':
            signal.signal(signal.SIGBREAK, self._handle_signal)
        signal.signal(signal.SIGILL, self._handle_signal)

    def _reset(self):
        """Reset simulator, erase all information about past simulation run
        and set all settings to default.
        """
        self.period = 0
        self._start_step = 0
        self.chain_index = None
        self._chain_type = None
        self._mccs = None
        self._forced_operations = None
        self.level = None
        self._depth = None
        self._tasks = None
        self.task = None
        self.value_fixing = None
        self._block_growth = None
        self._block_all = None
        self._blocked_from_db = set([])
        self._last_step_objs = set([])
        self._release_for_updating = numpy.array([])
        self._branching_operations = None
        self._branching_initialized = False
        self._branching_succeded = False
        self._opres_queue = []
        self._iterations = 1
        self._start_iter = 0
        self._end_iter = 1
        self._orig_time_steps = None
        self._curr_op_type = 'simulated'
        self._curr_forced_op_id = None
        self._data_processing_date = None
        self.prediction_memory.reset()
        self.operation_memory.reset()
        self.terminate = False
        if self.data:
            try:
                del self.data.matrix.sim
            except Exception:
                pass
            try:
                del self.data.matrix
            except Exception:
                pass
            del self.data
            self.data = None


    def simulate(self, iterations, id_list=None, ind_seq=None):
        """Run simulation with current simulation setup and data matrix.

        iterations -- number of iterations, int
        id_list -- list of simulation unit ids
        ind_seq -- sequence of simulation unit indices
        """
        # first divide the input data into reasonable sets and simulate one
        # set at a time
        self._iterations = iterations
        idsets = self._divide_data(id_list, ind_seq)
        num_of_sets = len(idsets)
        total_time = 0.0
        self._data_db.static_ids_written = {}
        for i, item in enumerate(idsets):
            self._data_db.data_id_store = defaultdict(list)
            self._data_db.op_id_store = defaultdict(list)
            self._data_db.binfo_store = defaultdict(list)
            self._data_db.price_tracker = set([])
            if self.terminate:
                break

            # a set for blocked main level objects - blocking can be due to
            # passed simulation ending conditions or passed error limits
            # info the ones with passed error limits are kept over time spans
            self._blocked_for_good = set([])

            # reset all settings and simulation information from the last
            # simulation and set current simulation data. Also set the date
            # information for all main level simulation units.
            ids, iterations, start = item
            msg = 'Simulating ids: %s' % ids
            self._logger.log_message(self._log_name, 'DEBUG', msg)
            self._reset()
            self._set_data(iterations, start, ids)
            self._set_dates()
            # check that input unit set contains data
            N = len(ids)
            if N == 0:
                continue

            # log INFO message and start timing
            msg = "simulating iterations %d..%d for set %d/%d with %d units" %\
                  (start + 1, start + iterations, i+1, num_of_sets, N)
            self._logger.log_message(self._log_name, 'INFO', msg)
            start_time = time.time()

            # go through the timespans in the simulation control structure
            # each timespan may have different timesteps and different model
            # chains etc.
            self._terminated = set([])
            self._start_step = 0
            for j, timespan in enumerate(self._ctrl.time_spans):
                self._blocked_from_db = self._copy(self._blocked_for_good)
                self._last_step_objs = set([])
                self._release_for_updating = numpy.array([])
                tind, torem = self.data.get_tind(self.main_level, 0)
                if len(tind) == 0 or self.terminate:
                    self.any_terminated = True
                    break  # stop simulation if stop flag is raised

                # set current timespan and model chains
                self._setup_time_span(timespan, j)
                if not self._get_num_of_steps():
                    continue

                # keep track of units for which the date has been updated, as
                # in the case of multiple end_date spans, objects past all of
                # the end dates will be stored to db during all of the spans
                # with the same date leading to db integrity error
                self._updated_dates = set([])
                self._targets = set([tuple(row) for row in tind[:,:3]])

                # check if forced operations should be done during current
                # timespan
                forceops = self._do_force_operations()

                # first evaluate the initialisation chains
                self._evaluate_chains('init')

                # loop through the number of timesteps or until number of
                # timesteps have been evaluated or stopping condition is met
                self._evaluate_steps(forceops)

                # release all objects from all levels and model chain depths
                # after simulating a full timespan
                self.data.release(None, None, None, False, True)

                # after release of all objects, check which objects were
                # blocked due to exceeding maximum error count and block those
                # objects again
                block = self.data.check_errors(self._max_simunit_errors)
                if block is not None:
                    self.data.block(self.main_level, 0, block, True)

                # block again those objects that should be blocked for good
                # - mostly due to stop condition
                if self._blocked_for_good:
                    btind = numpy.array(list(self._blocked_for_good))
                    self.data.block(self.main_level, 0, btind, True)
                self._start_step = self._end_step

            if self.create_branch_desc:
                msg = 'creating branch descriptions'
                self._logger.log_message(self._log_name, 'INFO', msg)
                self._data_db.fill_branch_desc_table([id[0] for id in ids],
                                                     self._start_iter)
            # stop timing and log INFO message
            end_time = time.time()
            delta = end_time - start_time
            total_time += delta
            msg = "set simulated in %s" % secs2str(delta)
            self._logger.log_message(self._log_name, 'INFO', msg)
            msg = "avg. time per unit %s" % secs2str(delta / N)
            self._logger.log_message(self._log_name, 'INFO', msg)
            msg = "avg. time per set %s" % secs2str(total_time / (i+1))
            self._logger.log_message(self._log_name, 'INFO', msg)
            msg = "total time %s" % secs2str(total_time)
            self._logger.log_message(self._log_name, 'INFO', msg)
        if self.stem_cache.initialized and self.stem_cache.new_data:
            self.stem_cache.flush()
            self.stem_cache.new_data = False


    def _evaluate_steps(self, forceops):
        """Evaluate number of timesteps with possible stopping conditions,
        simulation chains, forced operations and/or operation chains and
        update simulator state
        """
        for step in range(self._start_step, self._end_step):
            if self.terminate:
                break
            self._last_step_objs = set([])  # objects blocked from simulation
            block_all_end_date = False
            block_all_stop = False
            if self.timespan.ending.type == 'end_date':
                # check if current unit has passed the simulation end date
                # and break if all simulation main level objects have passed
                # the end date
                block_all_end_date = self._evaluate_end_date()
            if self._ctrl.stop_logic is not None:
                # evaluate simulation stopping condition and break
                # if all simulation main level objects are blocked
                block_all_stop = self._evaluate_stop_condition()
            simulate = not(block_all_end_date or block_all_stop)
            if simulate:
                if self._do_grow():
                    # evaluate simulation chains for all or some of units and
                    # after that, release blocked units if there are any
                    self._evaluate_chains('simulation')
                    if self._block_growth is not None:
                        self.data.release(self.main_level, 0,
                                          self._block_growth, True)
                        self._block_growth = None
            if not block_all_stop:
                if self._release_for_updating.size > 0:
                    # release objects that were blocked from simulation
                    self.data.release(self.main_level, 0,
                                      self._release_for_updating, True)
                if forceops:
                    # if there are forced operations, evaluate those before
                    # going to the operations defined in the operation chains
                    forceops = self._evaluate_forced_operations()
                    # block the objects that had forced operations on them
                    if len(self._blocked_forced) > 0:
                        self.data.block(self.main_level, 0,
                                        self._blocked_forced, True)

                # evaluate operation chains
                self._evaluate_chains('operation')
                # release the objects that had forced operations on them
                # and clear the list
                if len(self._blocked_forced) > 0:
                    self.data.release(self.main_level, 0,
                                      self._blocked_forced, True)
                    self._blocked_forced = numpy.array([], dtype=int)
                    self._blocked_forced.shape = (0, 5)

            # store data from matrix to db
            if self._release_for_updating.size > 0:
                # release objects that were blocked from simulation, but should
                # still have their dates updated
                self.data.release(self.main_level, 0,
                                  self._release_for_updating, True)
            if self._sim_type == 'simulation':
                self._update(to='current period end')
                self._save_timestep_data(step)
                self._update(to='next period beginning')
            else:
                self._update(to='current period end')
                self._update(to='next period beginning')
                self._save_timestep_data(step)
            if self._release_for_updating.size > 0:
                # block again the objects that were released for date update
                self.data.block(self.main_level, 0, self._release_for_updating,
                                True)
            if self._last_step_objs:
                # update the blocked_from_db set with the units which during
                # the current time step were blocked from simulation, but not
                # yet from db - this way those units will not be added to db
                # during the following time steps
                self._blocked_from_db.update(self._last_step_objs)
            if not simulate:
                # when this condition evaluates true, the simulation of all
                # objects has been stopped due to either stop logic or end
                # date condition and the simulation of the current span can
                # be stopped (after storing current time step data to db)
                break

        steps = self._end_step - self._start_step
        if steps == 0:
            # zero steps span (most likely for only running init chains)
            # using -1 for the step to indicate last step
            self._save_timestep_data(-1)
            self._update(to='next period beginning')

    def _save_timestep_data(self, step):
        """
        Save timestep data to dabase

        step -- current step ordinal
        """
        if self._sim_type == 'simulation':
            if self._check_timestep_saving(step):
                # check that all objects' dates have been updated if we're past
                # the first time span and time span ending type is end_date
                if self._time_span_ind > 0:
                    block = self._targets.difference(self._updated_dates)
                    self._blocked_from_db.update(block)
                # store data to db from simulation data matrix
                self._data_db.add_data_from_matrix(self.data.matrix.dmx,\
                    self.data.matrix.mx, self.data.ind2id,\
                    self.data.linkage.links, self._blocked_from_db,\
                    self._terminated, self.main_level, self.data.text_data, False,
                    self._start_iter)
                if self._time_span_ind > 0:
                    self._blocked_from_db = \
                            self._blocked_from_db.difference(block)
            # store operation results
            self._add_opres_to_db(self._data_db)
            self._revert_blocks()

            # store branching info to db
            if self.data.brancher is not None and \
                self.data.brancher.branching_info is not None:
                self._data_db.add_branching_info(\
                    self.main_level, self.data.brancher.branching_info,
                    self.data.ind2id.ids, self._start_iter)
                self.data.brancher.reset_binfo()
        else:
            if self._check_timestep_saving(step):
                dmx = self.data.matrix.dmx
                if self._ctrl.fixed_dp_date:
                    ddates = numpy.empty(numpy.size(dmx), dtype=date)
                    if self._data_processing_date is None:
                        msg = "fix_data_processing_date_to_init should "\
                              "be FALSE if use_data_date_as_init_date "\
                              "is TRUE"
                        self._logger.log_message(self._log_name, 'ERROR', msg)
                        ddates = numpy.copy(dmx)
                    else:
                        ddates.fill(self._data_processing_date)
                        ddates.shape = dmx.shape
                else:
                    ddates = numpy.copy(dmx)
                self._input_db.add_data_from_matrix(ddates,
                    self.data.matrix.mx, self.data.ind2id,
                    self.data.linkage.links, self._blocked_from_db,
                    self._terminated, self.main_level, self.data.text_data,
                    True, self._start_iter)
                self._revert_blocks()

    def _evaluate_chains(self, chaintype, forcedind=None):
        """
        The model chains are evaluated so that the method goes through the
        top level 'task' elements and calls a method to process each of the
        task elements. chaintype is one of 'init', 'simulation', 'operation'
        or 'force_operation'
        """
        if chaintype not in self._mccs or self._mccs[chaintype]==[]:
            return
        self._chain_type = chaintype
        self.chain_index = -1
        # model chains are in chain collections that map to model chain files
        # each model chain collection consists of chain groups but those are
        # flattened when parsed, so a collection is a list of model chains
        # go through each model chain collection of the current chain type
        for mccind in range(len(self._mccs[chaintype])):
            if forcedind is None:
                chains = self._mccs[chaintype][mccind].chains
            else:
                # only pick those forced operation chains that are set to exec
                chains = [self._mccs[chaintype][mccind].chains[i] \
                          for i in forcedind[mccind]]
            for chain in chains:
                if self.terminate:
                    break

                # update chain index
                self.chain_index += 1
                self.level = self._input_db.get_level_ind(chain.evaluate_at)
                self.task = None
                self._tasks = chain.tasks
                # collect all errors dispatched during the evaluation of the
                # current model chain into error stack
                self._error_stack = []
                # check if modelchain has a condition and evaluate it
                block = None
                if chain.condition is not None:
                    # if condition evaluation fails for all evaluation level
                    # objects, go to the next chain
                    # else block objects for which the condition failed
                    block, all_blocked = self._evaluate_expression(\
                        chain.condition, self.level, 0)
                    if all_blocked:
                        continue
                    self.data.block(self.level, 0, block)
                # go through the top level tasks in the current chain
                # top level task loop can be broken if branching_group
                # condition is not met.
                for task in self._tasks:
                    self._evaluate_task(task, 0)
                    if self.terminate:
                        break
                if block is not None and len(block) > 0:
                    # release blocked objects if there are any
                    self.data.release(self.level, 0, block)
                if len(self._error_stack) > 0:
                    # process the errors dispatched during the evaluation of
                    # the current model chain
                    self._process_errors()

    def _evaluate_task(self, task, depth):
        """Method evaluates a task element of a model chain and moves on
        recursively to any subtasks if there are subtasks. If condition
        element is found, evaluate_expression Method is called and if model
        element is found, evaluate_model Method is called.
        """
        if self.terminate:
            return
        # check if this task record value changes to metadata
        if task.value_fixer:
            self.value_fixing = True
        else:
            self.value_fixing = False
        self.task = task
        subtasksfound = False
        block = None
        if task.condition != None:
            # if the task condition failed for all evaluation level objects,
            # return, else block failed objects and continue evaluation
            block, all_blocked = self._evaluate_expression(\
                task.condition, self.level, depth)
            if all_blocked:
                return
            else:
                # block the objects that should not be evaluated
                self.data.block(self.level, depth, block)
                if task.model != None:
                    self._evaluate_model(task.model, task.model_param, depth)
                    self._debug_last_task = task.name
                elif task.subtasks != []:
                    subtasksfound = True
        elif task.model != None:
            self._evaluate_model(task.model, task.model_param, depth)
        elif task.subtasks != []:
            subtasksfound = True
        if subtasksfound:
            self._evaluate_subtasks(task, depth)
        if block is not None and block.size > 0:
            # release objects blocked by the current task, if there are any
            self.data.release(self.level, depth, block)

    def _evaluate_subtasks(self, task, depth):
        """Evaluate a subtask and all it's subtasks.
        """
        arr = numpy.array
        depth += 1
        if task.type == 'branch_task':

            # check if the current task has a condition from the branching
            # group-level
            tslev, tsind = self._ctrl.built_ins[names.TIME_STEP]
            bgroup = task.branching_group
            partind = task.branching_group_part
            ee = self._evaluate_expression
            start = self.data.brancher.start
            stop = self.data.brancher.stop
            bgcond = self._ctrl.all_branching_groups[bgroup][0][partind][0]
            if bgcond is not None:
                bg_block, all_blocked = ee(bgcond, self.level, depth)
                if all_blocked:
                    return
                self.data.block(self.level, depth, bg_block)

            # container for the target indices of new branches and their parent
            # branches, both of which should not generate more branches and
            # should be blocked (pc_block)
            pc_block = []

            # evaluate the branching conditions
            bconds = task.branch_conditions
            btcond = task.condition
            for bc in bconds:

                # get branch name and current branching condition and evaluate
                # branching condition and block objects (bc_block)
                bname, bexpr = bc
                bc_block, all_blocked = ee(bexpr, self.level, depth)
                if all_blocked:
                    continue
                self.data.block(self.level, depth, bc_block)

                # after evaluating the branching condition, evaluate also the
                # branching task condition, if there is one, and block the
                # objects that failed the evaluation (bt_block)
                if btcond is not None:
                    bt_block, all_blocked = ee(btcond, self.level, depth)
                    if all_blocked:
                        self.data.release(self.level, depth, bc_block)
                        return
                    self.data.block(self.level, depth, bt_block)

                # store the operation results that were generate before the
                # branching operations and empty the queue
                opres_before_branching = deepcopy(self._opres_queue)
                self._opres_queue = []

                # start branching: create necessary amount of new branches,
                # update branching tasks etc. and evaluate branch task's
                # subtasks
                start(task, bname, self.data, self.lexicon, self.main_level,
                      depth, self.prediction_memory, self.operation_memory)
                if self.data.brancher.is_branching:

                    # branch task subtask evaluation is done only if there is
                    # at least a single branching target
                    for subtask in task.subtasks:
                        self._evaluate_task(subtask, depth)

                    # stop branching: inactivate branches that were not
                    # constructed successfully and store branching history
                    # and release currently blocked objects
                    block = stop(bname, self.data, self.timespan,
                                 self.main_level, depth,
                                 self.prediction_memory, self.operation_memory,
                                 self.lexicon, tslev, tsind, self._opres_queue,
                                 opres_before_branching)

                    # block the newly generated branches and their parent
                    # branches for the duration of the currently evaluated
                    # branch task
                    for row in block:
                        row = tuple(row)
                        if row not in pc_block:
                            self.data.block(self.level, depth, arr([row]))
                            pc_block.append(row)
                
                # add the operation results that were generated before
                # branching back to the queue
                self._opres_queue = opres_before_branching + self._opres_queue

                # release first the objects blocked by the branching condition
                # and then the objects blocked by the branch task condition
                self.data.release(self.level, depth, bc_block)
                if btcond is not None:
                    self.data.release(self.level, depth, bt_block)

            # if there was branching condition blocks (only one branching
            # condition can generate a new branch for it, br, obj
            # -combination), release those objects
            if pc_block:
                self.data.release(self.level, depth, arr(pc_block))

            # if there was branching group-level conditions or branch task
            # conditions, release the blocked objects
            if bgcond is not None:
                self.data.release(self.level, depth, bg_block)
        else:
            # evalute normal task subtasks
            for subtask in task.subtasks:
                self._evaluate_task(subtask, depth)

    def _evaluate_model(self, model, param, depth):
        """Evaluate a model stored in a model chain. Models can be of types:
        'prediction', 'aggregation', 'operation', 'management', 'geo_table'
        and 'parameter_table. Model call construction and model calls are
        handled by model caller objects.'
        """
        self._depth = depth
        self._callers[model.type].execute(model, param, self, depth)
        self._depth = None
        try:
            if model.type == names.OPERATION:
                # check if the current operation model is a branching model
                # resulting in a new branch, but only when evaluation level is
                # simulation main level!
                if self.level == self.main_level and \
                   self.data.brancher is not None:
                    self.data.brancher.check_operation(\
                        model, self.data, self.main_level, depth)
        finally:
            self._callers[model.type].cleanup()

    def _evaluate_expression(self, expr, level, depth=0, block_value=False):
        """Evaluate a condition expression. Expressions are in postfix deque
        items and are evaluated with a postfix evaluator.
        Return target index of all blocked objects and boolean value True
        if all evaluation level objects are blocked and False if not.
        """
        tind, torem = self.data.get_tind(level, depth)
        if tind.size == 0:
            return tind, True
        # TODO: a direct reference to operation memory might be a bit more
        # efficient than accessing it through the _callers -dictionary
        opmem = self.operation_memory
        evalres = self._evaluator.evaluate(self.data, tind, expr, level,
                                           self.main_level, depth, opmem,
                                           self.timespan)
        indexer = numpy.where(evalres == block_value)[0]
        all_blocked = indexer.size == evalres.size
        return tind.take(indexer, axis=0), all_blocked

    def _evaluate_stop_condition(self):
        """Evaluate simulation stopping condition and stop simulation if
        condition evaluates True for all simulation level objects. Block the
        objects for which the condition was met.
        """
        # If we have no data to begin with, ignore the stop condition,
        # since the expression would never be evaluated.
        tind, torem = self.data.get_tind(self.main_level, 0)
        if tind.size == 0:
            return False

        self._block_all, all_blocked = self._evaluate_expression(\
            self._ctrl.stop_logic, self.main_level, 0, block_value=True)

        self.data.block(self.main_level, 0, self._block_all, True)
        inds = set([tuple(row) for row in self._block_all[:,:3]])
        self._blocked_from_db.update(inds)
        self._last_step_objs.update(inds)
        self._blocked_for_good.update(inds)
        if all_blocked:
            return True
        return False

    def _evaluate_end_date(self):
        """
        Check if any of the main level objects have passed the simulation
        ending date. Block objects that have passed the ending date.
        """
        tind, torem = self.data.get_tind(self.main_level, 0)
        if len(tind) == 0:
            self._release_for_updating = numpy.array([])
            self._last_step_objs = set([])
            return True
        cdates = self.data.get_date(tind)
        for i, c in enumerate(cdates):
            cdates[i] = date(c.year, c.month, c.day)

        # if time span ending type is 'end_date', check if any of the
        # main level units passes the ending date during the current
        # time step
        tslev, tsind = self._ctrl.built_ins[names.TIME_STEP]
        up = update_dates(cdates,
                          self.timespan.time_step,
                          self.timespan.unit)
        if self.timespan.time_step > 0:
            past_at_start = cdates >= self.timespan.ending.target
            past_at_end = up >= self.timespan.ending.target
        elif self.timespan.time_step < 0:
            past_at_start = cdates <= self.timespan.ending.target
            past_at_end = up <= self.timespan.ending.target

        pas_index = numpy.where(past_at_start)[0]
        pae_index = numpy.where(past_at_end)[0]

        # get growth season and time span ending dates
        ge_month = self._ctrl.growth_season_end_month
        ge_day = self._ctrl.growth_season_end_day
        end_month = self.timespan.ending.target.month
        end_day = self.timespan.ending.target.day
        end_year = self.timespan.ending.target.year
        end_date = date(end_year, end_month, end_day)
        ge_date = date(end_year, ge_month, ge_day)

        # check whether the time step is sub-year
        sub_year = self._sub_year_step()

        # calculate the number of time steps to the last growth season end date
        # if simulation is in yearly or longer steps, and the number of time
        # steps to time span end date if step is less than one year
        if sub_year:
            tgt = self.timespan.ending.target
        elif end_date >= ge_date:
            tgt = date(end_year, ge_month, ge_day)
        else:
            tgt = date(end_year - 1, ge_month, ge_day)
        cdd = calculate_date_diff
        unit = self.timespan.unit
        timeunits = numpy.array([cdd(d, tgt, unit) for d in cdates],
                                dtype=float) + 1

        # make sure that there are no negative time steps if the
        # main time step is positive and vice versa
        if self.timespan.time_step > 0:
            timeunits[timeunits < 0.] = 0.
        elif self.timespan.time_step < 0:
            timeunits[timeunits > 0.] = 0.

        # update the time steps to data matrix for objects that'll be past end
        # date after the current time step
        self.data.set_value(tslev, tind.take(pae_index, axis=0), tsind,
                            timeunits.take(pae_index, axis=0))

        # block the objects that had their time steps adjusted to 0
        block = timeunits == 0.

        # store the blocked objects and add those to last step objects
        self._block_all = tind.take(numpy.where(block)[0], axis=0)
        inds = set([tuple(row) for row in self._block_all[:,:3]])
        self._last_step_objs.update(inds)

        # add objects that'll be past the end date to the last step objects
        inds = set([tuple(row) for row in tind[past_at_end,:3]])
        self._last_step_objs.update(inds)

        # data should be written to db for all objects except those that were
        # already past at the beginning of step
        block_from_db = tind.take(pas_index, axis=0)
        db_inds = set([tuple(row) for row in block_from_db[:,:3]])
        self._blocked_from_db.update(db_inds)

        # objects that will be past end date after current time step should
        # be blocked from simulation, but not from forced and normal
        # operations and date updating
        nosim = numpy.where((past_at_start == False) & past_at_end)[0]
        self._release_for_updating = tind.take(nosim, axis=0)

        if self._block_all.shape[0] == 0:
            self._block_all = None
        else:
            self.data.block(self.main_level, 0, self._block_all, True)
        if numpy.all(block):
            # if all objects have passed ending date, no need to block
            # objects as simulation will be stopped
            return True
        else:
            return False

    def _do_grow(self):
        """
        Check if growth should be calculated for the current time step,
        ie. should the 'simulation' model chains be evaluated, and block
        objects for which the growth simulation should be passed.

        If time step unit is year and length is at least a year, this check
        has to be done only during the first time step.

        If time step is less than one year, the growth season should be
        considered and the sub year steps should be distributed evenly to
        the growth season.
        """
        if not self._sub_year_step() and self.period > 0:
            # if time step length is greater than or equal to one year,
            # no need to check growth after first period
            return True

        # get current dates for each main level object
        tind, torem = self.data.get_tind(self.main_level, 0)
        cdates = numpy.array(self.data.get_date(tind), dtype=date)

        # get the time step values for each main level object
        ts_lev, ts_ind = self._ctrl.built_ins[names.TIME_STEP]
        ts_tind, ts_torem = self.data.get_tind(ts_lev, 0)
        timesteps, err = self.data.get_value(ts_tind, ts_ind)

        # check the growth season end dates and block objects that are past
        # the growth season end date
        m = self._ctrl.growth_season_end_month
        d = self._ctrl.growth_season_end_day
        ged = numpy.array([date(x.year,m,d) for x in cdates], dtype=date)

        end_date = self.timespan.ending.type == "end_date"
        unit = self.timespan.unit

        if self.timespan.time_step == 0:
            # if time step is zero, no need to check this
            return True
        elif not self._sub_year_step():
            # if time span step is greater than or equal to 1 the growth season
            # handling is quite straightforward: units that are past the
            # current growth season end date will be simulated with a decreased
            # time step

            # if timespan ending type is "steps" and if time step is greater
            # than one and growth season end date is passed, decrease time
            # step by one for a duration of one time step and remove the
            # blocks after the current time step
            # if timespan ending type is "end_date" and if timestep is greater
            # than zero, skip the blocking as the possible timestep reductions
            # have already been done in _evaluate_end_date method
            block = cdates >= ged
            if end_date:
                targets = (timesteps > 0) * block
            else:
                targets = (timesteps > 1) * block
            if numpy.any(targets):
                # store the original time steps
                self._orig_time_steps = (ts_tind, numpy.copy(timesteps))
                # when ending type is "end_date", this reduction of the time
                # step is handled by Simulator._evaluate_end_date method and
                # this happens only in the case that the first step is also
                # the last one
                if not end_date:
                    if unit == names.YEAR: decrease = 1.
                    elif unit == names.MONTH: decrease = 12.
                    timesteps[targets] -= decrease
                block[targets] = False
                self.data.set_value(ts_lev, ts_tind, ts_ind, timesteps)
        else:
            # if time span unit is month and/or the step is less than a year,
            # the time steps require more tinkering: the time steps that
            # don't overlap the growth season will have 0 step and the yearly
            # growth is distributed evenly to the overlapping time steps

            # calculate the end date of the current time step
            updated = update_dates(cdates, timesteps, unit, deduceaday=True)

            # get growing season end date for the current year
            m = self._ctrl.growth_season_start_month
            d = self._ctrl.growth_season_start_day
            gsd = numpy.array([date(x.year,m,d) for x in cdates], dtype=date)

            # check if the current time step overlaps the growing season of the
            # current year
            overlap = ((cdates >= gsd) * (cdates <= ged)) + \
                      ((updated >= gsd) * (updated <= ged)) + \
                      ((cdates < gsd) * (updated > ged))

            # check if the time step ending date overlaps with the next year's
            # growth season if time span end date does not come before that
            gsd_next = numpy.array([date(x.year+1,x.month,x.day) for x in gsd],
                               dtype=date)
            ged_next = numpy.array([date(x.year+1,x.month,x.day) for x in ged],
                               dtype=date)
            overlap_next = ((updated >= gsd_next) * (updated <= ged_next))
            if self.timespan.ending.type == 'end_date':
                past = gsd_next >= self.timespan.ending.target
                overlap_next[past] = False
            overlap += overlap_next

            # all objects that do not overlap the growth season should be
            # blocked during the current time step
            block = overlap == False

            # process the time steps of the overlapping objects
            if numpy.any(overlap):
                # store the original time steps
                self._orig_time_steps = (ts_tind, numpy.copy(timesteps))

                # calculate the number of time steps that overlap the growth
                # season and divide the yearly growth evenly to those steps
                start = date(1, self._ctrl.growth_season_start_month,
                             self._ctrl.growth_season_start_day)
                end = date(1, self._ctrl.growth_season_end_month,
                           self._ctrl.growth_season_end_day)
                if unit == names.YEAR:
                    units = ((end - start).days - 1.) / 365.
                    year_as_units = 1.
                elif unit == names.MONTH:
                    units = ((end - start).days - 1.) / 31.
                    year_as_units = 12.
                steps = math.ceil(units / self.timespan.time_step)
                timesteps[:] = year_as_units / steps

                # store the temporary time steps
                self.data.set_value(ts_lev, ts_tind, ts_ind, timesteps)

        if numpy.all(block):
            # if all objects have passed growth ending date, no need to block
            # objects as growth simulation ('loop' chains) will be skipped
            # anyway
            return False
        self._block_growth = tind.take(numpy.where(block)[0], axis=0)
        if self._block_growth.shape == (0, 5):
            self._block_growth = None
        else:
            self.data.block(self.main_level, 0, self._block_growth, True)
        return True


    def _do_force_operations(self):
        """Check if forced operations should be done to any of the current
        simulation units during the current timespan
        """
        if not self.do_forced_ops:
            return False
        self._forced_operations = None
        self._forced_chain_index = None

        # check if current timespans has any forced operation chains
        if not self.timespan.forced_operations:
            return False

        # construct chain index from the time span's forced operation chains
        mccs = self.timespan.chain_collections['forced_operation']
        self._forced_chain_index = []
        for mcc in mccs:
            focs = enumerate(mcc.chains)
            fopd = dict([(c.name,i) for i,c in focs])
            self._forced_chain_index.append(fopd)

        # get the forced operations for all current simulation units and
        # store the forced operations for later use
        levelname = self.lexicon.get_level_name(self.main_level)
        ops = self._operation_db.fill_dictionary(levelname, self._ids)
        if ops == {}:
            return False
        if self.copy2iters:
            # copy all operations from iteration 0 to all current iterations
            # (from start_iter up to end_iter)
            copied_ops = {}
            for key, item in ops.items():
                iterr, branch, oid = key
                if iterr == 0:
                    for i in range(self._start_iter, self._end_iter):
                        copied_ops[(i, branch, oid)] = self._copy(item)
            ops = copied_ops

        # calculate operation timings if operation timings are steps
        tind, torem = self.data.get_tind(self.main_level, 0)
        ids = self.data.get_id(self.main_level, tind)
        dates = self.data.get_init_date(tind)
        cdates = self.data.get_date(tind)
        for i in range(tind.shape[0]):
            key = (self._start_iter + tind[i,0], tind[i,1], ids[i][0])
            if not ops.has_key(key):
                continue
            removables = []
            for j, op in enumerate(ops[key]):
                if op.timing_type == 'step':
                    su = str(op.step_unit)
                    # if operation date is not defined, calculate the target
                    # date for the current forced operation
                    if su == names.YEAR:
                        to_add = relativedelta(years=+op.time_step)
                    elif su == names.MONTH:
                        to_add = relativedelta(months=+op.time_step)
                    else:
                        to_add = relativedelta(days=+op.time_step)
                    op.date = dates[i] + to_add

                # make sure that operations set to be done before current data
                # date will be skipped
                if op.date < cdates[i]:
                    removables.append(j)
            if removables:
                for j in reversed(removables):
                    ops[key].pop(j)
            if ops[key] == []:
                del ops[key]
        if ops == {}:
            return False

        self._forced_operations = ops
        return True

    def _evaluate_forced_operations(self):
        """Evaluate the forced operations for the current units. Forced
        operations can be operation that actually have happened when
        simulating from past to present (updating simulations) or operations
        can be forced to simulation units for some other reason.
        """
        if self._forced_operations is None:
            return False
        if self.timespan.unit not in (names.YEAR, names.MONTH):
            errorstr = "Invalid timestep unit '%s' for forced operations"\
                       " (years or months only)!" % self.timespan.unit
            self.add_error(errorstr, None, None)
            return False
        # check if any of the active units has forced operations during the
        # current time step
        keys = set(self._forced_operations.keys())
        for fkey in keys:
            fo_tind, torem = self.data.get_tind(self.main_level, 0)
            # we want to keep reference to the original tind values
            # unaltered by the processing of the fops model chains
            tind = deepcopy(fo_tind)

            ids = self.data.get_id(self.main_level, tind)
            # get the current and initial dates for all objects
            dates = self.data.get_date(tind)
            # go through each simulation unit and evaluate forced
            # operations for a single unit at a time
            # NB: evaluating forced operation chains for a single
            # (iteration,branch,unit) may not be very efficient
            tslev, tsind = self._ctrl.built_ins[names.TIME_STEP]
            timesteps, err = self.data.get_value(tind, tsind)
            if self.timespan.unit == names.YEAR:
                years, months, days = timesteps, 0, 0
            elif self.timespan.unit == names.MONTH:
                years, months, days = 0, timesteps, 0
            indset = set(range(tind.shape[0]))
            for i in range(tind.shape[0]):
                key = (self._start_iter + tind[i,0], tind[i,1], ids[i][0])
                if key != fkey:
                    continue
                tb = tind[i]
                tb.shape = (1, 5)
                self._blocked_forced = numpy.append(self._blocked_forced, tb,
                                                    axis=0)
                # define current simulation time step end date
                if self.timespan.unit == names.YEAR:
                    end_date = dates[i] + relativedelta(years=+int(years[i]))
                elif self.timespan.unit == names.MONTH:
                    end_date = dates[i] + relativedelta(months=+int(months[i]))
                end_date += relativedelta(days=-1)
                # make sure the end_date is also smaller than the ending date.
                if self.timespan.ending.type == 'end_date':
                    # if time step is 0, set the run end as end_date, otherwise
                    # this is for cases when the current time step is
                    # the last one and the time step has been set to zero
                    # to skip growing
                    run_end = self.timespan.ending.target
                    if timesteps[i] == 0:
                        end_date = run_end
                    else:
                        end_date = min(end_date, run_end)
                # go through the current simulation unit's forced operations
                ops, rem = [], []
                for j, op in enumerate(self._forced_operations[key]):
                    # if forced operation should be evaluated during
                    # the current time step, store the unit id to
                    # operation evaluation list
                    if op.date >= dates[i] and op.date <= end_date:
                        ops.append(op)
                        rem.insert(0, j)
                for j in rem:
                    self._forced_operations[key].pop(j)
                if len(self._forced_operations[key]) == 0:
                    del self._forced_operations[key]
                # sort the operations in the operation queue on operation date
                ops = sorted(ops, key=attrgetter('date'))
                for op in ops:
                    # start from a clean slate
                    self.data.active_data_level = None
                    self._curr_forced_op_id = op.op_id
                    block = list(indset.difference((i,)))
                    self.data.block(self.main_level, 0,
                                    tind.take(block, axis=0), True)
                    ind = []
                    for focind in self._forced_chain_index:
                        mccind = [focind[c] for c in op.chain_names \
                                  if c in focind]
                        ind.append(mccind)
                    orig_values = self._update_data(tind[(i,),:],
                                                    op.data_values)
                    # evaluate the forced operation model chains
                    self._curr_op_type = 'forced'
                    self._evaluate_chains('forced_operation', ind)
                    self._curr_op_type = 'simulated'
                    # forced op chain execution may leave active_data_level set
                    # which will mess up updating the data with tind set here
                    self.data.active_data_level = None
                    self._update_data(tind[(i,),:], orig_values)
                    # release the blocked objects
                    self.data.release(self.main_level, 0,
                                      tind.take(block, axis=0), True, False)
                self._curr_forced_op_id = None
                break
        if self._forced_operations == {}:
            self._forced_operations = None
            return False
        return True

    def _update_data(self, tind, data_values):
        """Update simulator data values by setting the values in the
        'data_values' dictionary
        """
        orig_values = {}
        if data_values is None:
            return orig_values
        for key, value in data_values.items():
            level, variable = key
            lind, vind = self.lexicon.get_variable_ind(level, variable)
            if lind != self.main_level:
                value_tind, to_rem = self.data.get_tind(self.main_level, 0, lind)
                orig, err = self.data.get_value(value_tind, vind)
            else:
                orig, err = self.data.get_value(tind, vind)
            if err or orig is None:
                # NB: in case there's errors, the original values are NaNs and
                # the new data value should be stored for later use instead of
                # the original value
                orig_value = value
            else:
                orig_value = orig[0]
            orig_values[key] = orig_value
            if lind != self.main_level:
                self.data.set_value(lind, self.data.active_dataobjects, vind, value)
            else:
                self.data.set_value(lind, tind, vind, value)
        return orig_values

    def _get_num_of_steps(self):
        """Get the maximum number of timesteps for the current timespan. As
        each simulation unit might have different date stamp, the number of
        time steps will be the maximum over all units.
        """
        if self.timespan.ending.type=='end_date':

            # TODO: check this logic, seems to give one step too many
            # (at least for UMT test stands)

            # get the earliest date from all objects and calculate the maximum
            # number of timesteps
            tind, torem = self.data.get_tind(self.main_level, 0)
            cdates = self.data.get_date(tind)
            if self.timespan.time_step >= 0:
                ext_dates = numpy.min(cdates)
            else:
                ext_dates = numpy.max(cdates)

            # calculate the number of time step units (years, months or days)
            # to the end date
            numberofunits = calculate_date_diff(ext_dates,
                                             self.timespan.ending.target,
                                             self.timespan.unit) + 1
            if (numberofunits <= 0 and self.timespan.time_step >= 0) or \
               (numberofunits > 0 and self.timespan.time_step <= 0):
                span_no = self._ctrl.time_spans.index(self.timespan) + 1
                infostr = "all simulation units are past simulation ending "\
                          "date (%s), time step of timespan n.o %d set to 0"\
                          % (self.timespan.ending.target, span_no)
                self.add_info(infostr, None, None)
                numberofsteps = 0
            elif abs(self.timespan.time_step) != 1:
                # if timestep is something else than a year, the number of
                # steps must be divided by the timestep
                numberofsteps = int(numberofunits / self.timespan.time_step)
                if numberofunits % self.timespan.time_step != 0:
                    numberofsteps += 1
                    warnstr = 'The number of %ss (%d) is not divisible with'\
                              ' the given time step (%d). Forcably rounding '\
                              'up.' % (self.timespan.unit, numberofunits,
                                       self.timespan.time_step)
                    self.add_warning(warnstr, None, None)
            else:
                numberofsteps = numberofunits
            if self.timespan.time_step < 0 and numberofsteps < 0:
                numberofsteps = abs(numberofsteps)
        elif self.timespan.ending.type == 'steps':
            numberofsteps = self.timespan.ending.target
        self._end_step = self._start_step + numberofsteps
        return True

    def _setup_time_span(self, timespan, ind):
        """Setup simulator for current time span

        timespan -- Timespan instance
        ind -- timespan indice, integer
        """
        self._orig_time_steps = None
        self.timespan = timespan
        self._time_span_ind = ind
        self._mccs = timespan.chain_collections
        # set the current time step to simulation level
        tslev, tsind = self._ctrl.built_ins[names.TIME_STEP]
        tind, torem = self.data.get_tind(tslev, 0)
        self.data.set_value(tslev, tind, tsind, self.timespan.time_step)

    def _update(self, to):
        """
        Update simulator state and move to next time step

        to -- 'current period end' or 'next period beginning'
        """
        # if the time step was modified due to growth season date restrictions
        # or such, set the time step back to original
        if self._orig_time_steps is not None:
            ts_lev, ts_ind = self._ctrl.built_ins[names.TIME_STEP]
            ts_tind, time_steps = self._orig_time_steps
            self.data.set_value(ts_lev, ts_tind, ts_ind, time_steps)
            self._orig_time_steps = None
        # get simulation main level objects and update the dates to the
        # current time step end date
        tind, torem = self.data.get_tind(self.main_level, 0)
        if tind.size == 0:
            return
        curdates = self.data.get_date(tind)
        if to == 'current period end':
            tslev, tsind = self._ctrl.built_ins[names.TIME_STEP]
            timesteps, err = self.data.get_value(tind, tsind)
            unit = self.timespan.unit
            # if simulation or data processing time span has an ending date,
            # make sure that current period end is not past that date
            if self.timespan.ending.type == 'end_date':
                # if we're at the last simulation time step, don't deduce day
                # from the updated date as it will cause problems for units
                # for which newdate equals enddate
                newdates = numpy.copy(curdates)
                enddate = self.timespan.ending.target

                # check which of the objects have reached their last step,
                # which is indicated by the unit being in _last_step_objs
                stopped = numpy.zeros(tind.shape[0], dtype=bool)
                for i, row in enumerate(tind[:,:3]):
                    if tuple(row) in self._last_step_objs:
                        stopped[i] = True

                # update the current date with the given timestep
                newdates = update_dates(curdates, timesteps, unit,
                                        deduceaday=True)

                # if some of the updated units did pass the ending date, set
                # the ending date as the new date
                if self.timespan.time_step > 0:
                    past = newdates > enddate
                else:
                    past = newdates < enddate
                newdates[past] = enddate

                # if this is the last time step of some of the units and some
                # of the dates have not yet reached the end date, set the
                # ending date as the new date
                if True in stopped:
                    if self.timespan.time_step > 0:
                        short = newdates < enddate
                    else:
                        short = newdates > enddate
                    newdates[stopped & short] = enddate
            else:
                newdates = update_dates(curdates, timesteps, unit,
                                        deduceaday=True)

            # store units for which the date has been updated
            self._updated_dates.update(set([tuple(row) for row in tind[:,:3]]))

            # store the updated dates to data matrix
            self.data.set_date(tind, newdates)
        elif to == 'next period beginning':
            step = 1
            unit = 'day'
            newdates = update_dates(curdates, step, unit)
            if self.timespan.ending.type == 'end_date':
                update = curdates <= self.timespan.ending.target
                if len(update) == 0:
                    # nothing to proceed to end date with, something is wrong
                    # TODO: next line is commented out because related
                    # issue indicating that the number of steps is one too many
                    # self.terminate = True
                    pass
                else:
                    indexer = numpy.where(update)[0]
                    self.data.set_date(tind.take(indexer, axis=0),
                                       newdates.take(indexer, axis=0))
            else:
                self.data.set_date(tind, newdates)

            # update built-in attributes 'year', 'month' and 'day' for
            # simulation main level objects
            years = numpy.array([d.year for d in newdates], dtype=float)
            months = numpy.array([d.month for d in newdates], dtype=float)
            days = numpy.array([d.day for d in newdates], dtype=float)
            ylev, yind = self._ctrl.built_ins[names.YEAR]
            mlev, mind = self._ctrl.built_ins[names.MONTH]
            dlev, dind = self._ctrl.built_ins[names.DAY]
            self.data.set_value(ylev, tind, yind, years)
            self.data.set_value(mlev, tind, mind, months)
            self.data.set_value(dlev, tind, dind, days)
            self.period += 1

    def _set_dates(self):
        """
        Set initial dates for all main level objects.
        """
        tind, torem = self.data.get_tind(self.main_level, 0)
        nobj = tind.shape[0]
        if self._ctrl.use_data_date:
            # if dates are not set in data, set the default initial date for
            # all objects and log error message
            dates = self.data.get_date(tind)
            missing = dates == 0
            dates[missing] = self._ctrl.initial_date
        else:
            # set data dates either to current date (today) or simulation
            # initial date
            if self._ctrl.use_today:
                ddate = date.today()
            else:
                ddate = self._ctrl.initial_date
            dates = numpy.empty(nobj, dtype=date)
            dates.fill(ddate)
            if self._sim_type=='data processing':
                self._data_processing_date = ddate
        self.data.set_date(tind, dates, set_init_date=True)
        # set built-in attribute 'year' for simulation main level objects
        years = numpy.array([d.year for d in dates], dtype=float)
        months = numpy.array([d.month for d in dates], dtype=float)
        days = numpy.array([d.day for d in dates], dtype=float)
        ylev, yind = self._ctrl.built_ins[names.YEAR]
        mlev, mind = self._ctrl.built_ins[names.MONTH]
        dlev, dind = self._ctrl.built_ins[names.DAY]
        self.data.set_value(ylev, tind, yind, years)
        self.data.set_value(mlev, tind, mind, months)
        self.data.set_value(dlev, tind, dind, days)

    def _set_data(self, iterations, start, ids):
        """Create data handler if first simulation run and reset existing
        handler otherwise. Fill data matrix from input database.

        iterations -- number of iterations for the current data set
        start -- starting iteration round
        ids -- list of simulation unit id strings as tuples of unique id and
               original id
        """
        self._start_iter = start
        self._end_iter = start + iterations
        if self.data is None or self.data.matrix.iterations != iterations:
            # first simulation run, create new data handler
            attrs = self.lexicon.max_num_of_vars
            mcdepth = self._ctrl.max_model_chain_depth
            branchgroups = self._ctrl.all_branching_groups
            objcount = int(len(ids) * self._objcount_multiplier)
            branchcount = self._branch_count
            noopbranch = self._no_op_branch
            separatebgs = self._separate_branching_groups
            branchlimit = self._branch_limit
            self.data = Handler(iterations, self.lexicon, attrs, mcdepth,
                                self.main_level, branchgroups, noopbranch,
                                separatebgs, self, objcount, branchcount,
                                branchlimit)
        else:
            # handler already exists and needs a reset
            self.data.reset()
        # log INFO message
        msg = "filling simulation data matrix..."
        self._logger.log_message(self._log_name, 'INFO', msg)
        # fill data matrix with data from input database
        self._input_db.fill_matrix(self.data, self.main_level, ids)
        # if number of iterations is more than one, copy data objects and
        # values from iteration 0 to all other iterations
        if iterations > 1:
            self.data.add_iteration(0)
        # assign values for the built-in 'iteration' attribute
        iter_lev, iter_ind = self._ctrl.built_ins[names.ITERATION]
        tind, torem = self.data.get_tind(iter_lev, 0)
        arr = numpy.array
        for i in range(tind.shape[0]):
            obj_tind = arr([tind[i,:]], dtype=int)
            obj_value = arr([tind[i,0]], dtype=float)
            self.data.set_value(iter_lev, obj_tind, iter_ind, obj_value)
        # store simulation unit ids
        self._ids = ids
        # set init variable values if there are any
        if self._ctrl.init_variables is not None:
            ivars = defaultdict(list)
            ivals = defaultdict(list)
            for ind, val in self._ctrl.init_variables:
                ivars[ind[0]].append(ind[1])
                ivals[ind[0]].append(val)
            for level in ivars.keys():
                tind, torem = self.data.get_tind(level, 0)
                self.data.set_value(level, tind, ivars[level], ivals[level])
        # log INFO message
        msg = "matrix filling complete"
        self._logger.log_message(self._log_name, 'INFO', msg)

    def _divide_data(self, id_list, ind_seq):
        """Divide the simulation target data into reasonable simulation
        sets

        id_list -- list of object ids
        ind_seq -- index sequence: (from, to)
        """
        #ram = memcheck.get_available_ram() #TODO: Useme
        db_ids = self._input_db.get_ids('data', self.main_level, both=True,
                                        id_list=id_list)
        if id_list is None:
            id_list = db_ids
            if ind_seq is not None:
                start, end = ind_seq
                if end is None or end > len(id_list):
                    end = len(id_list)
                if start < 1:
                    start = 1
                start -= 1
                id_list = id_list[start:end]
        else:
            # check if some of the units in id_list are not in the database
            check_ids = [ids[1] for ids in db_ids]
            missing = set(id_list).difference(check_ids)
            if len(missing) > 0:
                id_list = set(id_list).intersection(check_ids)
                msg = "following simulation units were not found in input "\
                      "database: %s" % ', '.join(missing)
                self._logger.log_message(self._log_name, 'INFO', msg)
            id_list = [db_id for db_id in db_ids if db_id[1] in id_list]
        N = len(id_list)
        if (N * self._iterations) < self._max_unit_count:
            sets = [(id_list, self._iterations, 0)]
        else:
            if self._iterations > self._max_unit_count:
                lo = self._iterations % self._max_unit_count
                mx = self._max_unit_count
                iters = [mx for i in range(int(self._iterations / mx))]
                if lo > 0:
                    iters.append(lo)
                sets = []
                for i in range(N):
                    start = 0
                    for j in range(len(iters)):
                        sets.append(([id_list[i]], iters[j], start))
                        start += iters[j]
            else:
                a = int(self._max_unit_count / self._iterations)
                if a == 0:
                    a = 1
                itr = self._iterations
                sets = [(id_list[i:i+a], itr, 0) for i in range(0, N, a)]
        # log INFO message
        msg = "%d simulation units divided into %d sets" % (N, len(sets))
        self._logger.log_message(self._log_name, 'INFO', msg)
        return sets


    ##########################################################################
    # Public methods that also model caller objects can access
    def add_error(self, errorstr, level, tind):
        """Log error message and update error counters

        errorstr -- error message as str
        obj -- object indice or indices, one of None, int, float, iterable
        tind -- target index as numpy array
        """
        self._add2log(errorstr, level, tind, 'ERROR')

    def add_warning(self, warningstr, level, tind):
        """Log warning message

        warningstr -- warning message as str
        obj -- object indice or indices, one of None, int, float, iterable
        tind -- target index as numpy array
        """
        if self._warnings:
            self._add2log(warningstr, level, tind, 'WARNING')

    def add_info(self, infostr, level, tind):
        """Log info message

        infostr -- info message as str
        obj -- object indice or indices, one of None, int, float, iterable
        tind -- target index as numpy array
        """
        self._add2log(infostr, level, tind, 'INFO')

    def _add2log(self, msg, level, tind, mtype):
        """Add error and warning messages to log
        """
        path = None
        if self.task is not None:
            path = self.task.path
        if tind is None:
            # general log message concerning all simulation units
            # log the message with the task path
            if path is not None:
                msg = "SIMULATION STEP %d, TASK: %s, %s" % (self.period,
                                                            path, msg)
            self._logger.log_message(self._log_name, mtype, msg)
        else:
            # object specific log message(s) for single or multiple units
            if level is None:
                level = self.level
                if level is None:
                    level = self.main_level
            if level == self.main_level:
                main_tind = tind
                simunitids = oids = self.data.get_id(level, tind)
                dates = self.data.get_date(tind)
            elif level > self.main_level:
                if len(tind.shape) == 1:
                    tind.shape = (1, tind.shape[0])
                # check if warning/error level equals current simulation
                # evaluation level and if not, get the object mapping from
                # level to self.level
                pt, tr = self.data.get_parent(level, tind, self.main_level)
                oids = self.data.get_id(level, tind)
                dates = self.data.get_date(pt)
                main_tind = pt
                simunitids = self.data.get_id(self.main_level, pt)
            elif level < self.main_level:
                main_tind, torem = self.data.get_tind(self.main_level,
                                                      self._depth)
                simunitids = self.data.get_id(self.main_level, main_tind)
                oids = self.data.get_id(level, tind)
                dates = self.data.get_date(main_tind)
            # get error object level name
            levelname = self.lexicon.get_level_name(level)
            self._dispatch(mtype, dates, simunitids, tind, levelname, oids,
                           msg, path)
            if mtype == 'ERROR':
                self._error_stack.append((main_tind, simunitids, dates, path))

    def _process_errors(self):
        """Process error objects in the current error stack, update error
        counters and block objects that have passed the maximum error counts

        NB: At the moment _process_errors is called after evaluating a single
        model chain. This means that objects are actually blocked after the
        whole chain has been evaluated and not immediately after the task
        which actually generated the errors.
        """
        for main_tind, simunitids, dates, path in self._error_stack:
            exc = self.data.add_error(main_tind, self._max_simunit_errors)
            if numpy.any(exc):
                n = len(exc)
                inds = [tuple(row) for row in main_tind[exc,:3]]
                self._blocked_from_db.update(set(inds))
                self._blocked_for_good.update(set(inds))
                self._terminated.update(set(inds))
                dates = [dates[i] for i in range(n) if exc[i]]
                simunitids = [simunitids[i] for i in range(n) if exc[i]]
                msg = "****** simulation of unit terminated ******"
                mlevname = self.lexicon.get_level_name(self.main_level)
                self._dispatch('INFO', dates, simunitids, main_tind,
                               mlevname, simunitids, msg, path)
        # empty the error stack and block all objects that have passed the
        # maximum error limit
        self._error_stack = []
        block = self.data.check_errors(self._max_simunit_errors)
        if block is not None:
            self.data.block(self.main_level, 0, block, True)
            inds = [tuple(row) for row in block[:,:3]]
            self._blocked_from_db.update(set(inds))
            self._blocked_for_good.update(set(inds))
            self._terminated.update(set(inds))
        # if total simulation error count has been exceeded, terminate
        # simulation
        if self._max_errors >= 0 and self.data.error_count > self._max_errors:
            msg = "****** maximum error count reached - terminating "\
                  "simulation ******"
            self._terminate(msg)

    def _dispatch(self, mtype, dates, simunitids, tind, level, oids, msg,
                  path):
        """Dispatch log message

        mtype -- message type, string
        dates -- date info, datetime.date numpy array
        simunitids -- simulation unit ids, list of strings
        tind -- target index, numpy array
        level -- level index, int
        oids -- object ids, list of strings
        msg -- message, string
        path -- current task path, string
        """
        msgs = set([])
        start = self._start_iter
        for i in range(len(simunitids)):
            m = "DATE: %s, SIMULATION UNIT: '%s', ITERATION: %d, BRANCH:"\
                " %d" % (dates[i], simunitids[i][1], start + tind[i,0],
                         tind[i,1])
            if path is not None:
                m = "TASK: %s, %s" % (path, m)
            m = "%s, OBJECT: %s '%s': %s" % (m, level, oids[i][1], msg)
            msgs.add((m, simunitids[i][1]))
        for item in msgs:
            self._logger.log_message(self._log_name, mtype, item[0], item[1])

    def queue_opres(self, tind, target, results, name, group, notes='',
                    materials=None):
        """
        Receives the operation results from the operation model caller and
        adds them to the queue to be later added to the db

        tind -- target index as used in the matrix
        target -- target indices for the result, a tuple of (iteration, branch
                  level name, object id)
        results -- a tuple of a list of dictionaries having attribute names
                   as keys and attribute values as values, and the possible
                   delay for the result as years
        name -- operation name
        group -- operation group
        notes -- operation notes
        materials -- materials needed by the operation
        """
        if materials is None:
            materials = list()
        data = (self._curr_op_type, tind, target, self._curr_forced_op_id,
                results, name, group, notes, materials)
        self._opres_queue.append(data)

    def _add_opres_to_db(self, target_db):
        """ Add operation results from the queue to operation database

        target_db -- simo.db.data_db.db instance
        """
        start = self._start_iter
        ops = []
        for opres in self._opres_queue:
            op_type = opres[0]
            ddate = self.data.get_date(opres[1])[0]
            iteration, branch, levname, obj_id = opres[2]
            op_id = opres[3]
            values = opres[4]
            op_name = opres[5]
            op_group = opres[6]
            notes = opres[7]
            materials = opres[8]
            dated_values = []
            for val in values:
                data, delay = val
                if delay is not None:
                    # delay the current operation result if delay has been
                    # defined (add a given timestep to the result timestamp)
                    # NB: the delay is now years!!!
                    year, month, day = ddate.year, ddate.month, ddate.day
                    res_date = date(year+delay, month, day)
                else:
                    res_date = ddate
                dated_values.append((data, res_date))
            data_dict = {'op_type': op_type, 'values':dated_values,
                         'date':ddate, 'id':obj_id, 'op_id':op_id,
                         'branch':branch, 'iteration':start+iteration,
                         'op_name': op_name, 'op_group':op_group,
                         'notes': notes, 'materials': materials}
            ops.append(data_dict)
        if len(ops) > 0:
            errs = target_db.add_opres(levname, ops)
            for err_msg in errs:
                self._logger.log_message(self._log_name, 'ERROR', err_msg)
        self._opres_queue = []

    def _check_timestep_saving(self, step):
        """
        Check if data from current time step should be stored to result db

        step -- current time step, integer
        """
        self._temporarily_activated = None
        if self.timespan.save_steps_to_db == 'all':
            if step == 0 and self._last_step_objs:
                intersect = self._blocked_from_db.intersection(\
                                            self._last_step_objs)
                if intersect:
                    self._temporary_activation(False)
            return True  # store data from all time steps
        elif self.timespan.save_steps_to_db == 'none':
            return False  # don't store any data from the current time span
        elif self.timespan.save_steps_to_db == 'last':
            if step == (self._end_step - 1):
                return True
            if self._last_step_objs:
                self._temporary_activation(True)
                return True
            return False

    def _temporary_activation(self, block_active):
        """
        Activate just blocked objects temporarily. This can be in two cases:
        1. only last time step should be stored, and some of the objects have
           just reached their last time step (due to end date or stop
           condition, but some of the objects are still simulated
        2. all time steps should be stored, but some of the units are blocked
           from db already during first time step because their date is past
           the time span end date

        block_active -- indicates whether currently active objects should be
                        blocked or not
        """
        # get all active objects and block them in order NOT to save
        # them to the result db during the current time step
        if block_active:
            tind, torem = self.data.get_tind(self.main_level, self._depth)
            tmp_block = set([tuple(row) for row in tind[:,:3]])
        else:
            tmp_block = set([])
        # get the objects that are blocked for good, for example due
        # to too many errors ets.
        tmp_block = tmp_block.union(self._blocked_for_good)
        # get the difference to blocked objects that were not blocked just
        # now (ie. objects that were blocked some time before)
        diff = self._blocked_from_db.difference(self._last_step_objs)
        tmp_block = tmp_block.union(diff)
        self._blocked_from_db, self._temporarily_activated = \
            tmp_block, self._blocked_from_db

    def _revert_blocks(self):
        """
        Revert the blocks made in '_check_timestep_storing' method
        """
        if self._temporarily_activated is not None:
            self._blocked_from_db = self._temporarily_activated
            self._temporarily_activated = None

    def _handle_signal(self, frame, signum):
        """
        Handle signals that the simulator receives.

        frame -- frame object
        signum -- signal number as integer
        """
        msg = signum_to_msg(signum)
        self._terminate(msg)

    def _terminate(self, msg):
        """
        Terminate the simulation process. Log message of the termination,
        raise stop flags.
        """
        self._logger.log_message(self._log_name,'INFO', msg)
        if not self.terminate:
            termmsg = 'simulation set to terminate at %s' % (time.asctime())
            self._logger.log_message(self._log_name, 'INFO', termmsg)
        self.terminate = True

    def _sub_year_step(self):
        """
        Check if the current time span has time steps less than one year
        """
        unit = self.timespan.unit
        timestep = self.timespan.time_step
        if unit == names.YEAR:
            sub_year = timestep < 1.
        elif unit == names.MONTH:
            sub_year = timestep < 12.
        else:
            raise Exception("time step unit 'day' not implemented")
        return sub_year

    def _copy(self, data):
        return deepcopy(data)

    def release_all(self):
        """
        Release all blocks from the simulator
        """
        self._block_all = None
        self._block_growth = None
        self._blocked_for_good = None
        self._blocked_forced = numpy.array([], dtype=int)
        self._blocked_forced.shape = (0, 5)
        self._blocked_from_db = set([])
        self._last_step_objs = set([])


def main():
    import doctest
    doctest.testfile('test/sim-py.rst')

if __name__ == '__main__':
    main()
