#!/usr/bin/env python
# encoding: utf-8
"""
Module for logging messages from SIMO to the output stream and db.

*logger.py*

Copyright (c) 2007 Simosol Oy. Distributed under the GNU General Public License 2.0.
"""

import sys
import logging
import re
from datetime import datetime

class Logger:

    def __init__(self, logdb, runid, msgttable, lexttable, lang, console, 
                 consolelevel, output_warnings):
        self.logger = logdb
        self.run_id = runid
        self.console = console
        self.console_loggers = {}
        format='%(name)-12s: %(levelname)-8s %(message)s'
        self.format = format
        loglevel=self._map_logging_level(consolelevel)
        self.loglevel = loglevel
        self.consolelevel = consolelevel
        self.output_warnings = output_warnings
#        self.max_errors=maxerr
#        self.max_simunit_errors=maxsimuniterr
#        self.global_error_count=globalerrs
#        self.error_count=0
#        self.error_count_simunit=0
        self.msg_trans_table = msgttable
        self.level_trans_table = lexttable.level_table
        self.var_trans_table = lexttable.var_table
        self.language = lang
        self.r_patt = re.compile(r'\$(\d)')
        self.logged_warnings = []

    def _implant_orig(self, match):
        i = int(match.groups()[0])
        return self.matches[i-1]

    def _map_logging_level(self,level):
        if level.lower()=='debug':
            return logging.DEBUG
        elif level.lower()=='info':
            return logging.INFO
        elif level.lower()=='warning':
            return logging.WARNING
        elif level.lower()=='error':
            return logging.ERROR
        elif level.lower()=='critical':
            return logging.CRITICAL

    def _translate_message_body(self):
        """Substitute the hardwired message with the translated one"""
        msg = self.message
        mtype = self.message_type
        lname = self.logger_name
        lang = self.language
        if mtype in self.msg_trans_table.keys():
            if lname in self.msg_trans_table[mtype].keys():
                msgpatts = self.msg_trans_table[mtype][lname][0]
                trans = self.msg_trans_table[mtype][lname][1]
                for pattmsg in msgpatts:
                    matches = pattmsg[0].search(msg)
                    if matches is not None:
                        self.matches = matches.groups()
                        msgtempl = pattmsg[1]
                        if lang in trans[msgtempl].keys():
                            replmsg = trans[msgtempl][lang]
                            self.message = self.r_patt.sub(self._implant_orig,
                                                       replmsg)
                        break

    def _translate_data_levels(self):
        """Substitute data level names with their translations"""
        for levelname, trans in self.level_trans_table.iteritems():
            newname = trans.get(self.language)
            if newname is not None:
                self.message = self.message.replace(levelname, newname)

    def _translate_variable_names(self):
        """Substitute variable names with their translations"""
        for varname, trans in self.var_trans_table.iteritems():
            newname = trans.get(self.language)
            if newname is not None:
                self.message = self.message.replace(varname, newname)

    def log_message(self, loggername, mtype, message, sim_id=None):
        mtype = mtype.lower()
        if mtype == 'warning':
            wmes = message
            try:
                wmes = message.split('TASK: ', 1)[1]
                wmes = wmes.split('DATE: ',1)[0]+\
                       wmes.split('SIMULATION UNIT: ', 1)[1].split(' ', 1)[0]+\
                       wmes.rsplit(':', 1)[1]
            except:
                pass
            if wmes in self.logged_warnings:
                return
            self.logged_warnings.append(wmes)
        self.message = message
        self.message_type = mtype
        self.logger_name = loggername
        self._translate_message_body()
        self._translate_data_levels()
        self._translate_variable_names()
#        if mtype in ('critical', 'error'):
#            self.error_count+=1
        timestamp = datetime.now()
        self.logger.add_log_message(self.run_id, mtype, loggername, timestamp,
                                    sim_id, self.message)
        if self.console:
            self._log_to_console(loggername, mtype)

    def _log_to_console(self, lname, mtype):
        if lname in self.console_loggers:
            logger = self.console_loggers[lname]
        elif mtype in ('error', 'critical', 'exception'):
            add_handler = True
            lname += ' error'
            if lname not in self.console_loggers:
                logger = logging.getLogger(lname)
                logger.setLevel(self.loglevel)
                for handler in logger.handlers:
                    if isinstance(handler, logging.StreamHandler):
                        add_handler = False
                        break
                else:
                    handler = logging.StreamHandler(sys.stderr)
                formatter = logging.Formatter(self.format)
                handler.setFormatter(formatter)
                if add_handler:
                    logger.addHandler(handler)
                logger.propagate = False
            else:
                logger = self.console_loggers[lname]
        else:
            add_handler = True
            logger = logging.getLogger(lname)
            logger.setLevel(self.loglevel)
            for handler in logger.handlers:
                if isinstance(handler, logging.StreamHandler):
                    add_handler = False
                    break
            else:
                handler = logging.StreamHandler(sys.stdout)
            formatter = logging.Formatter(self.format)
            handler.setFormatter(formatter)
            if add_handler:
                logger.addHandler(handler)
            logger.propagate = False

        self.console_loggers[lname] = logger
        if mtype=='critical':
            logger.critical(self.message)
        elif mtype=='error':
            logger.error(self.message)
        elif mtype=='warning':
            if self.output_warnings:
                logger.warning(self.message)
        elif mtype=='info':
            logger.info(self.message)
        elif mtype=='debug':
            logger.debug(self.message)
        elif mtype == 'exception':
            logger.exception(self.message)

def main():
    pass

if __name__ == '__main__':
    main()

