#!/usr/bin/env python
# encoding: utf-8
"""
Title: SIMO runner configuration processing

Module for validating and processing the simo runner configuration file.

*runnerconfig.py*

Copyright (c) 2007 Simosol Oy. Distributed under the GNU General Public License 2.0.
"""


import os
import os.path as ph
import ConfigParser
import datetime, time
import re
import codecs

###############################################################################
sep=';'
###############################################################################

class Config:
    def __init__(self, logger):
        """
        Initialize Config object for parsing and checking SIMO
        configuration parameters.

        logger -- SimoLogger object
        """
        self.logger = logger
        self.cfg = {}
        self.ini_err = False
        self.list_separator = ';'
        self.booleans = ('1','TRUE','True','true','Yes','yes','y','Y','on',\
                         '0','FALSE','False','false','No','no','n','N','off')
        self.trues = ('1','TRUE','True','true','Yes','yes','y','Y','on')
        self.falses = ('0','FALSE','False','false','No','no','n','N','on')
        self.output_db_logical_names = ('input','result')
        self.import_formats = ('by_level','inlined')
        self.operation_formats = ('xml', 'text', 'db')
        self.logging_levels = ('debug','info','warning','error','critical')
        self.lp_opt_methods = ('jlp',)
        self.heuristic_opt_methods = ('TS','TA', 'SA','hero', 'GA')
        self.output_formats = ('inlined','by_level','smt',
                               'branching_graph', 'operation_result',
                               'aggregation', 'expression')
        self.output_result_type = ('simulated', 'optimized')
        self.simulation_tasks = ('data processing','simulation')
        self.text_formats = set(['inlined','by_level','smt'])

    def _check_letters_only(self, option):
        if self._option_exists(option):
            if not re.match('^[a-z,A-Z,_]+$', self.cfg.get(option)):
                errmsg = 'Option %s: Value must contain only '\
                         'letters a-z, A-Z or _, got %s'
                self._set_error(errmsg % (option, self.cfg.get(option)))


    def _check_boolean(self, option):
        """
        Check the validity of a boolean parameter and change the parameter
        string to boolean

        option -- config option name as string
        """
        if self._option_exists(option):
            if self.cfg.get(option) not in self.booleans:
                self._set_error('Option %s: value must be one of:'\
                                 ' %s' % (option, self.booleans))
            else:
                if self.cfg.get(option) in self.trues:
                    self.cfg[option] = True
                else:
                    self.cfg[option] = False

    def _check_number(self, option, integer=False, positive=False,
                      non_zero=False):
        """
        Check the validity of a number parameter and change the parameter
        string to a number instance (int or float)

        option -- config option name as string
        """
        if self._option_exists(option):
            try:
                result = float(self.cfg.get(option))
            except ValueError:
                if integer:
                    self._set_error('Option %s: value must '\
                                'be an integer number (f.e. 1 or 2)' % option)
                else:
                    self._set_error('Option %s: value must '\
                                'be a number (f.e. 1.1 or 2.5)' % option)
                return
            if integer and result != int(result):
                self._set_error('Option %s: value must '\
                                'be an integer number (f.e. 1 or 2)' % option)
                return
            if positive and result < 0:
                self._set_error('Option %s: value must be positive!' % option)
                return
            if non_zero and result == 0:
                self._set_error('Option %s: value can\'t be zero!' % option)
                return
            if integer:
                self.cfg[option] = int(result)
            else:
                self.cfg[option] = float(result)

    def _check_date(self, option, allow_empty=False):
        """
        Check the validity of a date parameter and change the parameter
        string to datetime

        option -- config option name as string
        """
        if self._option_exists(option,allow_empty):
            try:
                datestr = self.cfg.get(option)
                self.cfg[option] = datetime.date(\
                    *time.strptime(datestr, "%d.%m.%Y")[0:3])
            except ValueError:
                self._set_error('Option %s: date must be given '\
                                 'in format d.m.yyyy' % (option))

    def _check_delimiter(self, option):
        """
        Check the validity of a delimiter parameter.

        option -- config option name as string
        """
        if self._option_exists(option):
            delim = self.cfg.get(option)
            if delim == '\\t': delim = '\t'
            # Any whitespace converted to None to be used with the split function
            if delim.strip() in [' ', "''", "' '"]:
                self.cfg[option] = None
            if delim.strip() in ("';'", '";"'):
                self.cfg[option] = ';'

    def _check_dir(self, option, setbase=False, create=False, file=False):
        """
        Check that directory exists.

        option -- config option name as string
        setbase -- set the dir as the base dir if True
        create -- if True create a folder if it does not exist
        """
        if self._option_exists(option):
            fpath = ph.expandvars(ph.expanduser(self.cfg[option]))
            folder = ph.normpath(fpath)
            full_path = ph.abspath(folder)
            if (file):
                folder = ph.dirname(folder)
            valid = False
            if ph.isdir(folder):
                valid = True
                self.cfg[option] = full_path
            elif create:
                # create the target directory
                try:
                    os.makedirs(folder)
                    self.cfg[option] = full_path
                    valid = True
                except Exception:
                    pass
            if not valid:
                if create:
                    err_msg = "Option %s: couldn't create path %s"\
                            % (option,folder)
                else:
                    err_msg = 'Option %s: invalid path %s (%s)' % (option,
                                                                   full_path,
                                                                   folder)
                self._set_error(err_msg)
            elif setbase:
                os.chdir(folder)

    def _check_dirs(self, option):
        """
        Check that directories exist.

        option -- config option name as string
        """
        if self._option_exists(option):
            self.cfg[option] = self.cfg[option].split(sep)
            dirs = self.cfg[option]
            for i in range(len(dirs)):
                fpath = ph.expandvars(ph.expanduser(dirs[i]))
                dirs[i] = ph.normpath(fpath)
                valid = False
                if ph.isdir(dirs[i]):
                    valid = True
                    self.cfg[option][i] = dirs[i]
                if not valid:
                    self._set_error('Option %s: invalid path %s'\
                                     % (option,dirs[i]))

    def _check_file(self, option, folder=None):
        """
        Check that file exists.

        option -- config option name as string
        folder -- optional folder name as string
        """
        if self._option_exists(option):
            fpath = ph.expandvars(ph.expanduser(self.cfg.get(option)))
            filename = ph.normpath(fpath)
            valid = False
            if folder is not None:
                filename = ph.join(folder, filename.split(ph.sep)[-1])
            if ph.isfile(filename):
                self.cfg[option] = filename
                valid = True
            if not valid:
                self._set_error("Option %s: file %s "\
                                 "doesn't exist" % (option,filename))

    def _check_filepath(self, option):
        """
        Check that file path exists.

        option -- config option name as string
        """
        filepath = self.cfg.get(option)
        dpath = ph.expandvars(ph.expanduser(ph.dirname(filepath)))
        folder = ph.normpath(dpath)
        valid = False
        if ph.exists(folder):
            fpath = ph.expandvars(ph.expanduser(filepath))
            self.cfg[option] = ph.normpath(fpath)
            valid = True
        if not valid:
            self._set_error("Option %s: directory %s (%s) "\
                             "doesn't exist" % (option, fpath, folder))

    def _check_integer(self, option):
        """
        Check that a parameter value type is integer or can be converted to
        integer value.

        option -- config option name as string
        """
        if self._option_exists(option):
            try:
                val = self.cfg.get(option)
                self.cfg[option] = int(val)
            except Exception:
                self._set_error('Option %s: value %s is '\
                                 'not an integer' % (option,val))

    def _check_logging(self,section):
        """
        Check that logging definitions are valid.

        option -- config option name as string
        """
        self._check_logging_option(section, '.wipe_log', 'boolean')
        self._check_logging_option(section, '.console', 'boolean')
        self._check_logging_option(section, '.console_level', 'value')

    def _check_logging_option(self, section, opt, type):
        """
        Check that a logging option is valid.

        section -- logging section name as string
        opt -- option name as string
        type -- parameter value type as string
        """
        option = section+opt
        value = self.cfg.get(option)
        if value is not None and value != '':
            if type == 'boolean':
                self._check_boolean(option)
            elif type == 'value':
                self._check_value(option, self.logging_levels, True)
            return True
        return False

    def _check_list(self, option, clear_empty=False, force_exist=False):
        """
        Split a string into a list with the separator.

        option -- config option name as string
        """
        if force_exist and not option in self.cfg:
            self.cfg[option] = ''
        self.cfg[option] = self.cfg[option].split(self.list_separator)
        if clear_empty:
            self.cfg[option] = [item for item in self.cfg[option] if item]

    def _check_encodings(self, option):
        """
        Split a string into a list with the separator.
        Check list values against known encodings.

        option -- config option name as string
        """
        self.cfg[option] = self.cfg[option].split(self.list_separator)
        # Clear accidental empty items
        self.cfg[option] = [item for item in self.cfg[option] if item]
        for encoding in self.cfg[option]:
            try:
                codecs.getdecoder(encoding)
            except LookupError:
                self._set_error('Option %s: %s is not a valid encoding. '
                                'A list of valid encodings is available at '
                                'https://docs.python.org/2/library/'
                                'codecs.html#standard-encodings'
                                % (option, encoding,))


    def _check_multi_file(self, option, reqexist, reqlen, dir):
        """
        Check existence of multiple files

        option - config option name as string
        reqexist -- boolean for requirement for the file to already exist
        reqlen -- required number of filenames
        dir -- the directory in which the file should be found
        """
        if not self._option_exists(option):
            return
        self.cfg[option]=self.cfg.get(option).split(sep)
        if len(self.cfg[option])!=reqlen and reqlen is not None:
            self._set_error('option %s: %s filenames required, %s given'\
                             % (option, reqlen, len(self.cfg[option])))
        if reqexist:
            for filename in self.cfg[option]:
                if ph.isfile(ph.join(dir,filename))==0:
                    self._set_error("option %s: file "\
                                     "%s in directory %s doesn't exist" \
                                     % (option,filename,dir))

    def _check_multi_value(self,option,allowedvalues):
        """
        Check for validity of multiple values

        option -- config option name as string
        allowedvalues -- valid values for the option
        """
        if not self._option_exists(option):
            return
        val=self.cfg.get(option)
        if val is not None:
            self.cfg[option]=val.split(sep)
            values=self.cfg[option]
            for val in values:
                if val not in allowedvalues:
                    self._set_error('Option %s: '\
                                     '%s not in the allowed values of: %s' \
                                     % (option,val,allowedvalues))
        else:
            self._set_error('Option %s: no value given'\
                             % (option))

    def _check_range(self,option,allowempty):
        """
        Checks for a valid definition of a range of values

        option - config option name as string
        allowempty -- are empty values allowed
        """
        if self._option_exists(option,True):
            indrange=self.cfg.get(option)
            m=re.compile(r'(\d+)[-]?(\d*)').match(indrange)
            if m:
                if m.group(1)=='':
                    self._set_error('Option %s: start value not given')
                elif m.group(2)=='':
                    self.cfg[option]=(int(m.group(1)),None)
                else:
                    start=int(m.group(1))
                    end=int(m.group(2))
                    if start<=end:
                        self.cfg[option]=(start,end)
                    else:
                        self._set_error('Option %s: start index is greater'\
                                         ' than the stop index' % (option))
        elif allowempty:
            self.cfg[option]=(None,None)
        else:
            self._set_error('Option %s: unable to match the range '\
                             'definition. It should be either '\
                             'from-to (e.g. 1-10), or from- (e.g. 10-) ' \
                             'and not empty!' % option)

    def _check_value(self,option,allowedvalues,lowercase=False):
        """
        Check for a value existense, and optionally whether it has an allowed
        value

        option -- config option name as string
        allowedvalues -- optional list of allowed values
        lowercase -- should the value be converted into lowercase
        """
        if not self._option_exists(option):
            return
        value=self.cfg.get(option)
        if lowercase:
            value=value.lower()
        if value not in allowedvalues:
            self._set_error('Option %s: %s not in the allowed values of: %s' \
                             % (option,value,allowedvalues))

    def _check_value_list(self,option):
        """
       Check a list of values

        option -- config option name as string
        """
        self._option_exists(option)
        if self.cfg.get(option) is not None:
            self.cfg[option]=self.cfg.get(option).split(sep)

    def _option_exists(self, option, allowedempty=False):
        """
        Is the option in the ini file

        option -- config option name as string
        """
        if option not in self.cfg.keys():
            if not allowedempty:
                self._set_error('Missing required option %s' % (option))
            return False
        else:
            val=self.cfg.get(option)
            if val=='':
                if not allowedempty:
                    self._set_error('Option %s: no value given' % (option))
                return False
            else:
                return True

    def _set_error(self, msg):
        """
        Log error message and raise initialization error flag.

        msg -- error message as string
        """
        self.logger.error(msg)
        self.ini_err = True

    def _validate_config_content(self):
        """
        Validates that initialization file includes all necessary components.
        """
        self.ini_err=False
        # program_execution section
        self._check_dir('program_execution.base_folder', setbase=True)
        self._option_exists('program_execution.run_id')
        self._check_boolean('program_execution.import_data')
        self._check_boolean('program_execution.simulate')
        self._check_boolean('program_execution.optimize')
        self._check_boolean('program_execution.output_data')
        self._check_boolean('program_execution.no_wipe_import')
        self._check_boolean('program_execution.no_wipe_simulate')
        self._check_boolean('program_execution.no_wipe_optimize')
        self._check_boolean('program_execution.no_exec_import')
        self._check_boolean('program_execution.no_exec_simulate')
        self._check_boolean('program_execution.no_exec_optimize')
        self._check_boolean('program_execution.no_exec_logger')
        self._check_boolean('program_execution.code_profiling')
        imp = self.cfg.get('program_execution.import_data')
        sim = self.cfg.get('program_execution.simulate')
        opt = self.cfg.get('program_execution.optimize')
        out = self.cfg.get('program_execution.output_data')

        # logging section
        self._check_logging('logging')

        # translation section
        self._option_exists('translation.message_translation')
        self._option_exists('translation.lexicon_translation')
        self._option_exists('translation.lang')

        # error_handling section
        self._check_integer('error_handling.max_error_count_total')
        self._check_integer('error_handling.max_error_count_per_unit')
        self._check_boolean('error_handling.suppress_location')

        # warning_handling section
        self._check_boolean('warning_handling.output_warnings')
        self._check_boolean('warning_handling.suppress_location')

        # simo_db section
        self._check_file('simo_db.db_path')

        # data_db section
        self._option_exists('data_db.simulation_level')
        self._check_value('data_db.db_type', ('sqlite', 'postgresql',
                                              'oracle'), True)
        chk = self._option_exists('data_db.threaded',
                                  allowedempty=True)
        if not chk:
            self.cfg['data_db.threaded'] = False
        else:
            self._check_boolean('data_db.threaded')
        chk = self._option_exists('data_db.unlogged',
                                  allowedempty=True)
        if not chk:
            self.cfg['data_db.unlogged'] = False
        else:
            self._check_boolean('data_db.unlogged')
        dbtype = self.cfg.get('data_db.db_type')
        if dbtype is None:
            return
        self.cfg['data_db.db_type'] = dbtype.upper()

        self._check_boolean('data_db.spatial')
        if self.cfg['data_db.spatial']:
            self._check_integer('data_db.srid')
            if self.cfg['data_db.db_type'] == 'ORACLE':
                self._check_list('data_db.diminfo')
                di = self.cfg['data_db.diminfo']
                if len(di) != 4:
                    self._set_error('diminfo setting must be a list of '\
                            'minx, maxx, miny, maxy values')
                try:
                    di = [float(v) for v in di]
                    self.cfg['data_db.diminfo'] = di
                except Exception:
                    self._set_error('could not convert diminfo setting to '\
                            'a list of four floats')
            else:
                self.cfg['data_db.diminfo'] = None
            self.cfg['data_db.spatial'] = (self.cfg['data_db.srid'],
                                           self.cfg['data_db.diminfo'])
        else:
            self.cfg['data_db.srid'] = None
            self.cfg['data_db.diminfo'] = None
            self.cfg['data_db.spatial'] = None

        if self.cfg['data_db.db_type'] == 'SQLITE':
            try:
                from pysqlite2 import dbapi2 as sqlite3
            except Exception:
                self._set_error('SQLITE3 (pysqlite2) not present '\
                                'or found on system!')
            self._check_boolean('data_db.sqlite.in_memory_db')
            if not self.cfg['data_db.sqlite.in_memory_db']:
                self._check_dir('data_db.sqlite.db_dir', create=True)
            else:
                self.cfg['data_db.sqlite.db_dir'] = ''
            self.cfg['data_db.postgresql.db_host'] = ''
            self.cfg['data_db.postgresql.db_port'] = 0
            self.cfg['data_db.postgresql.db_user'] = ''
            self.cfg['data_db.postgresql.db_pw'] = ''

            self.cfg['data_db.postgresql.db_name'] = ''
            self.cfg['data_db.postgresql.db_schema'] = ''

            self.cfg['data_db.oracle.db_host'] = ''
            self.cfg['data_db.oracle.db_user'] = ''
            self.cfg['data_db.oracle.db_pw'] = ''

            self._option_exists('data_db.sqlite.log_db')
            self._option_exists('data_db.sqlite.input_db')
            self._option_exists('data_db.sqlite.result_db')
            self._option_exists('data_db.sqlite.optimized_db')
        elif self.cfg['data_db.db_type'] == 'POSTGRESQL':
            try:
                import psycopg2
            except Exception:
                self._set_error('POSTGRESQL (psycopg2) not present '\
                                'or found on system!')
            self.cfg['data_db.sqlite.db_dir'] = ''
            self.cfg['data_db.sqlite.in_memory_db'] = False
            self.cfg['data_db.oracle.db_host'] = ''
            self.cfg['data_db.oracle.db_user'] = ''
            self.cfg['data_db.oracle.db_pw'] = ''
            self._option_exists('data_db.postgresql.db_host')
            self._check_number('data_db.postgresql.db_port', True, True, True)
            self._option_exists('data_db.postgresql.db_user')
            self._option_exists('data_db.postgresql.db_pw')

            self._check_letters_only('data_db.postgresql.db_name')
            self._check_letters_only('data_db.postgresql.db_schema')

            self._check_letters_only('data_db.postgresql.log_prefix')
            self._check_letters_only('data_db.postgresql.input_prefix')
            self._check_letters_only('data_db.postgresql.result_prefix')
            self._check_letters_only('data_db.postgresql.optimized_prefix')
        elif self.cfg['data_db.db_type'] == 'ORACLE':
            try:
                import cx_Oracle
            except Exception:
                self._set_error('importing cx_Oracle failed! '\
                                'Oracle db_type can not be used')
            self.cfg['data_db.postgresql.db_host'] = ''
            self.cfg['data_db.postgresql.db_port'] = 0
            self.cfg['data_db.postgresql.db_user'] = ''
            self.cfg['data_db.postgresql.db_pw'] = ''
            self.cfg['data_db.postgresql.db_name'] = ''
            self.cfg['data_db.postgresql.db_schema'] = ''
            self.cfg['data_db.sqlite.db_dir'] = ''
            self.cfg['data_db.sqlite.in_memory_db'] = False
            self._option_exists('data_db.oracle.db_host')
            self._option_exists('data_db.oracle.db_user')
            self._option_exists('data_db.oracle.db_pw')
            self._check_letters_only('data_db.oracle.log_prefix')
            self._check_letters_only('data_db.oracle.input_prefix')
            self._check_letters_only('data_db.oracle.result_prefix')
            self._check_letters_only('data_db.oracle.optimized_prefix')

        # import section
        if imp:
            self._check_value('import.format', self.import_formats)
            self._option_exists('import.mapping_definition')
            self._check_boolean('import.import_date')
            self._check_date('import.data_date')
            if 'import.month_first' in self.cfg.keys():
                self._check_boolean('import.month_first')
            self._check_delimiter('import.data_delimiter')
            self._check_boolean('import.skip_first_line')
            self._check_dir('import.data_dir')
            datadir = self.cfg.get('import.data_dir')
            self._check_multi_file('import.data_file', True, None, datadir)
            if not 'import.encodings' in self.cfg:
                self.cfg['import.encodings'] = 'utf8'
            self._check_encodings('import.encodings')
            self._check_boolean('import.operations.operation_import')
            opimp = self.cfg.get('import.operations.operation_import')
            if opimp:
                self._check_value('import.operations.operation_format', \
                                  self.operation_formats)
                self._check_dir('import.operations.operation_dir')
                opdir = self.cfg.get('import.operations.operation_dir')
                self._check_file('import.operations.operation_file', opdir)
                if self.cfg.get('import.format', None) == 'text':
                    self._option_exists('import.operations.operation_conversion')
                self._option_exists('import.operations.operation_modelchains')
                if self.cfg['data_db.db_type'] == 'SQLITE':
                    self._option_exists('data_db.sqlite.operation_db')
                else:
                    self._option_exists('data_db.postgresql.operation_prefix')


        # simulation section
        if sim:
            self._check_boolean('simulation.use_data_date_as_init_date')
            self._check_boolean('simulation.use_today_as_init_date')
            usedata = self.cfg.get('simulation.use_data_date_as_init_date')
            usetoday =  self.cfg.get('simulation.use_today_as_init_date')
            if usedata and usetoday:
                msg = 'Both use_data_date_as_init_date and ' \
                      'use_today_as_init_date set for simulation'
                self._set_error(msg)
            if not usedata and not usetoday:
                self._check_date('simulation.fixed_init_date')
            self._check_value('simulation.task_type', self.simulation_tasks)
            t_type = self.cfg.get('simulation.task_type')
            if t_type == 'data processing':
                opt = 'simulation.fix_data_processing_date_to_init'
                self._check_boolean(opt)
            self._check_boolean('simulation.deterministic')
            self._check_boolean('simulation.track_prices')
            self._check_boolean('simulation.do_forced_ops')
            self._check_boolean('simulation.copy_ops_to_all_iterations')
            self._check_boolean('simulation.create_branch_desc')
            self._check_dir('simulation.cache_dir', create=True)
            self._option_exists('simulation.simulation_control')
            self._check_boolean('simulation.leave_no_op_branch')
            if 'simulation.separate_branching_groups' in self.cfg.keys():
                self._check_boolean('simulation.separate_branching_groups')
            else:
                self.cfg['simulation.separate_branching_groups'] = False
            self._check_number('simulation.object_count_multiplier',
                               positive=True, non_zero=True)
            self._check_number('simulation.estimated_branch_count',
                               integer=True, positive=True, non_zero=True)
            self._check_number('simulation.max_unit_count',
                               integer=True, positive=True, non_zero=True)
            if self.cfg.get('simulation.max_task_count', None) is not None:
                self._check_number('simulation.max_task_count',
                                   integer=True, positive=True, non_zero=True)
            else:
                self.cfg['simulation.max_task_count'] \
                                        = self.cfg['simulation.max_unit_count']
            if self.cfg.get('simulation.max_branch_limit', None) is not None:
                self._check_number('simulation.max_branch_limit',
                                   integer=True, positive=True, non_zero=True)
            else:
                self.cfg['simulation.max_branch_limit'] = None
            self._check_number('simulation.iterations',
                               integer=True, positive=True, non_zero=True)
            if 'simulation.id_list' in self.cfg.keys():
                self._check_list('simulation.id_list')
            if 'simulation.index_sequence' in self.cfg.keys():
                self._check_range('simulation.index_sequence',False)
            self._check_number('simulation.max_result_unit_count',
                               integer=True, positive=True, non_zero=True)

        # optimization section
        if opt:
            allmethods=self.lp_opt_methods+self.heuristic_opt_methods
            self._check_value('optimization.method', allmethods)
            opmethod=self.cfg.get('optimization.method')

            self._check_list('optimization.limit2branches_include', True, True)
            self._check_list('optimization.limit2branches_exclude', True, True)
            self.cfg['optimization.limit2branches'] = \
                   {'include': self.cfg['optimization.limit2branches_include'],
                    'exclude': self.cfg['optimization.limit2branches_exclude']}
            del self.cfg['optimization.limit2branches_include']
            del self.cfg['optimization.limit2branches_exclude']

            self._check_boolean('optimization.reuse_opt_data')
            self._check_boolean('optimization.keep_opt_data')
            self._check_dir('optimization.chart_path', create=True)
            self._check_boolean('optimization.track_solution')
            self._check_boolean('optimization.create_opt_db')
            if opmethod is not None:
                opmethod=opmethod.lower()
                self._option_exists('optimization.problem_definition')
                if opmethod=='jlp':
                    self._check_integer('optimization.jlp.max_objects')
                    self._check_dir('optimization.jlp.j_folder')
                    self._option_exists('optimization.jlp.file_prefix')
                    self._check_integer('optimization.jlp.timeout')
                    chk = self._option_exists('optimization.jlp.do_split',
                                              allowedempty=True)
                    if not chk:
                        self.cfg['optimization.jlp.do_split'] = False
                    else:
                        self._check_boolean('optimization.jlp.do_split')
                    chk = self._option_exists(\
                                    'optimization.jlp.skip_weight_splits',
                                    allowedempty=True)
                    if not chk:
                        self.cfg['optimization.jlp.skip_weight_splits'] = None
                    else:
                        self._check_value_list(\
                                    'optimization.jlp.skip_weight_splits')
                    fallback = self._option_exists(\
                                            'optimization.jlp.do_fallback',
                                            allowedempty=True)
                    if not fallback:
                        self.cfg['optimization.jlp.do_fallback'] = False
                    else:
                        # Make sure that fallback method is valid
                        self._check_boolean('optimization.jlp.do_fallback')
                        self._check_value('optimization.jlp.fallback_method', 
                                          allmethods)
                if opmethod=='TS':
                    self._check_integer('optimization.TS.repeats')
                    self._check_integer('optimization.TS.initial_solutions')
                    self._check_integer('optimization.TS.maximum_iterations')
                    self._check_integer('optimization.TS.entry_tenure')
                    self._check_integer('optimization.TS.exit_tenure')
                if opmethod=='hero':
                    self._check_integer('optimization.hero.repeats')
                    self._check_integer('optimization.hero.initial_solutions')
                    self._check_integer('optimization.hero.maximum_iterations')
                if opmethod=='SA':
                    sect = 'optimization.SA.'
                    self._check_integer(sect+'repeats')
                    self._check_integer(sect+'initial_solutions')
                    self._check_integer(sect+'maximum_iterations')
                    self._check_number(sect+'starting_temperature',
                                       integer=False, positive=True,
                                       non_zero=True)
                    self._check_number(sect+'freezing_temperature',
                                       integer=False, positive=True,
                                       non_zero=True)
                    self._check_number(sect+'cooling_factor',
                                       integer=False, positive=True,
                                       non_zero=True)
                    self._check_number(sect+'iterations_per_temperature',
                                       integer=True, positive=True,
                                       non_zero=True)
                    self._check_number(sect+'iteration_multiplier',
                                       integer=False, positive=True,
                                       non_zero=True)
                if opmethod=='GA':
                    self._check_integer('optimization.GA.repeats')
                    self._check_integer('optimization.GA.maximum_iterations')
                    self._check_integer('optimization.GA.population_size')
                    self._check_number('optimization.GA.breeding_rate',
                                       integer=False, positive=True,
                                       non_zero=True)
                    self._check_number('optimization.GA.mutation_probability',
                                       integer=False, positive=True,
                                       non_zero=True)
                    self._check_number('optimization.GA.swapping_probability',
                                       integer=False, positive=True,
                                       non_zero=True)

        # output section
        if out:
            self._check_value('output.result_type', self.output_result_type)
            #self._check_range('output.years',True)
            self._check_date('output.start_date', True)
            self._check_date('output.end_date', True)
            self._check_boolean('output.full_date')
            if 'output.id_list' in self.cfg.keys():
                self._check_list('output.id_list')
            if 'output.id_sequence' in self.cfg.keys():
                self._check_range('output.id_sequence',False)
            self._check_value('output.data_type', self.output_db_logical_names)
            self._check_dir('output.output_dir', create=True)
            self._check_multi_value('output.output_format',\
                                     self.output_formats)
            self._check_multi_file('output.output_filename',False,\
                                    len(self.cfg['output.output_format']),None)
            self._check_integer('output.default_decimal_places')
            formats=set(self.cfg.get('output.output_format'))
            if 'aggregation' in formats:
                self._option_exists('output.aggregation.aggregation_definition')
            if 'operation_result' in formats:
                self._check_value_list('output.operation_result.vars')
                self._check_value_list('output.operation_result.group_by')
            if formats.intersection(self.text_formats) is not None:
                self._option_exists('output.output_constraint')
            if 'inlined' in formats:
                self._check_boolean('output.inlined.result_padding')
            if 'smt' in formats:
                pdef = self.cfg.get('output.periods')
                if pdef is not None and len(pdef) == 2:
                    if pdef[0] is not None and pdef[1] is None:
                        errormsg = 'Option output.period: for smt output both'\
                                   ' the start and end period numbers must be'\
                                   ' specified'
                        self._set_error(errormsg)
        # ensure additional options are in the resulting cfg as empty lists,
        # even if they do not exist. Strip any '' values.
        self._check_list('additional_options.input_db', True, True)
        self._check_list('additional_options.result_db', True, True)
        self._check_list('additional_options.operation_db', True, True)
        self._check_list('additional_options.optimized_db', True, True)
        self._check_list('additional_options.controls', True, True)
        self._check_list('additional_options.conversions', True, True)


    def load_config(self, file, replace_cfg=None):
        """
        Loads the configuration file.  Returns a dictionary with key's
        of the form <section>.<option> and the values.

        file -- the configuration file object
        replace_cfg -- a dictionary of <section>.<option>: value pairs that
                       replace the parsed values
        """
        current_wd = os.getcwd()
        cp = ConfigParser.ConfigParser()
        cp.read(file)
        for sec in cp.sections():
            name = str.lower(sec)
            for opt in cp.options(sec):
                cfgname = name + "." + str.lower(opt)
                self.cfg[cfgname] = str.strip(cp.get(sec, opt))
        # Do replacements before validation.
        if replace_cfg:
            for key, val in replace_cfg.items():
                self.cfg[key] = val

        self._validate_config_content()
        os.chdir(current_wd)

def main():
    pass

if __name__ == '__main__':
    main()
