#!/usr/bin/env python
# encoding: utf-8
"""
Title: SIMO utility functions

This module includes general 'helper' functions for SIMO framework; e.g.
tools for creating a simulator instance.

*utils.py*

Copyright (c) 2007 Simosol Oy. Distributed under the GNU General Public License 2.0.
"""


import os
import sys
import datetime
import re
import xml.etree.ElementTree as ElementTree
import cPickle as pickle
import signal
import math
import numpy
from collections import defaultdict
from pygooglechart import PieChart2D
from dateutil.relativedelta import relativedelta


def get_xml_files_from_dir(path):
    """
    Extract xml file names from a given directory

    Parameters:
        path - path string to check

    Calls:
        None

    Returns:
        list of file names with full path
    """
    path = fix_path_ending(path)
    path = os.path.abspath(os.path.dirname(path))
    files = os.listdir(path)
    # check that the file name ends in ".xml"
    test = re.compile("\.xml$", re.IGNORECASE)
    files = filter(test.search, files)
    pathfiles=[]
    for i in range(len(files)):
        pathfiles.append(os.path.join(path, files[i]))
    return pathfiles

def is_path_valid(path, type, logger, logname):
    """
    Check that a file or a path is valid, eg. the directory or file exists.

    Parameters:
        path - path string to check
        type - dir / file
        logger - logging object for error messages and error count
        logname - the name of the active logger

    Calls:
        None

    Returns:
        True/False
    """
    if type=="dir":
        if not os.path.isdir(path):
            logger.log_message(logname,'error','%s is not a valid path.'\
                               % (path))
            return False
        else:
            return True
    elif type=="file":
        if not os.path.isfile(path):
            logger.log_message(logname,'error','%s is not a valid file.'\
                               % (path))
            return False
        else:
            return True

def first_level_higher(firstlevel,secondlevel,alllevels):
    """
    Load the simulation configuration file.  Return a dictionary with key's
    of the form <section>.<option> and the values.

    Parameters:
        firstlevel - name of the first data level
        secondlevel - name of the second data level
        alllevels - a list containing the level names in data hierarchy
                    order

    Calls:
        None

    Returns:
        True/False/None
    """
    try:
        first = alllevels.index(firstlevel)
        second = alllevels.index(secondlevel)
        if first < second:
            return True
        else:
            return False
    except ValueError:
        return None


def inject_list_items(l, item, times):
    """
    A copy of the given entry item in the list is injected to the list after
    the given item as many times as given by the times parameter.

    Parameters:
        l - the list whose items are duplicated
        item - the list item to be injected
        times - the number of injections of the item

    Calls:
        None

    Returns:
        The list with injected items
    """
    newlist = []
    for curitem in l:
        newlist.append(pickle.loads(pickle.dumps(curitem, 1)))
        # NB: here dump must be used in order to compare the list contents
        if pickle.dumps(curitem, 1) == pickle.dumps(item, 1):
            for i in range(times):
                newlist.append(pickle.loads(pickle.dumps(curitem, 1)))
    return newlist

def fix_path_ending(path):
    """
    Add the '\' or '/' character to the end of the path if missing.

    Parameters:
        path - the directory path to be checked and possibly fixed

    Calls:
        None

    Returns:
        The fixed path
    """
    if sys.platform[:3]=='win':
        if path[-1]<>'\\': path=path+'\\'
    else:
        if path[-1]<>'/': path=path+'/'
    return path

def process_constraint_definitions(deffile):
    """
    Read the content of the xml output constraint definition file into a data
    structure.

    Parameters:
        deffile - An xml file conforming to the Output_constraint.xsd
                  schema

    Calls:
        <utils.elem_name>

    Returns:
        output constraint definition data structure
    """
    try:
        # Get the root of definition XML
        defroot = ElementTree.ElementTree(file = deffile).getroot()
        if elem_name(defroot.tag)!='output_constraints':
            errmsg='The specified XML file is not an output'\
                   ' constraint definition file'
            return False,errmsg
        constr={}
        for level in defroot:
            constr[level[0].text]=level[1].text.split()
    except:
        errmsg= 'Error processing the constraint definition XML file. Check'\
                ' that the file exists and has correct contents'
        return False,errmsg
    else:
        return constr,None

def pascal(n):
    """
    Return the n:th line of Pascal`s triangle as a list
    (original author: DR#m <dnpsite.narod.ru>)

    Parameters:
        n - the line number of the Pascal's triangle

    Calls:

    Returns:
        list of numbers
    """
    l=[1]
    for i in xrange(int(n)):
        l2=[1]
        for j in xrange(len(l)-1):
            l2.append(l[j]+l[j+1])
        l2.append(1)
        l=l2
    return l

def print_stats(stats, dataid=None, type=None, heap=None):
    """
    A function for SIMO developers for printing out performance statistics in
    wanted format.
    """
    # Check if stats folder exists.  If not, create one.
    targetdir = os.path.join(os.curdir, 'stats')
    if os.path.isdir(targetdir) is False:
        os.mkdir(targetdir)
    type = '' if type is None else type
    dataid = '' if dataid is None else dataid
    stats.strip_dirs()

    bymodtimes = defaultdict(float)  # container for aggregate module timings

    # output performance statistics by function
    stats.sort_stats('cumulative')
    byfuncfile = open(os.path.join(targetdir,'func %s %s.txt' \
                                   % (type, dataid)), 'w')
    byfuncfile.write('SIMO %s performance profile by cProfile\n' % type)
    byfuncfile.write('%s\n' % stats.files[0])
    byfuncfile.write('Number of primary calls: %s\n\n' % stats.prim_calls)
    byfuncfile.write('Module\tFunction\tcalls\ttottime\tcumtime\tpercall\n')
    for func in stats.fcn_list[1:]:
        byfuncfile.write('%s\t%s\t%s\t%3.3f\t%3.3f\t%3.4f\n' \
                         % (func[0], func[2], stats.stats[func][0], \
                            stats.stats[func][2], stats.stats[func][3], \
                            stats.stats[func][2]/stats.stats[func][0]))
        bymodtimes[func[0]] += stats.stats[func][2]
    byfuncfile.close()

    # output aggregated performance statistics by module
    bymodfile = open(os.path.join(targetdir,'mod %s %s.txt' \
                                   % (type, dataid)), 'w')
    vkpairs = [(val,key) for key,val in bymodtimes.items()]
    vkpairs.sort(reverse=True)
    tottimes = bymodtimes.values()
    tottime = sum(tottimes)
    hdr = "Module\ttotal time, s\ttotal time, %\n"
    bymodfile.write(hdr)
    props = []
    modules = []
    for tt, mod in vkpairs:
        row = "%s\t%0.4f\t%0.1f\n" % (mod, tt, tt/tottime*100.0)
        bymodfile.write(row)
        props.append(tt/tottime*100.0)
        modules.append(mod)
    bymodfile.close()

    # output memory usage information (guppy heap)
    heapfile = open(os.path.join(targetdir,'heap %s %s.txt' \
                                   % (type, dataid)), 'w')
    heapfile.write(str(heap))
    heapfile.close()

    # draw a pie chart using pygooglechart
    n = 10
    chart = PieChart2D(1000, 300)
    chart.add_data(props[:n-1] + [sum(props[n-1:])])
    chart.set_pie_labels(modules[:n-1] + ['other'])
    try:
        chart.get_url()
        chart.download(os.path.join(targetdir, 'mod %s %s.png' % (type,
                                                                  dataid)))
    except:
        pass  # pass silently if offline

def signum_to_msg(signum):
    """
    Generate a message for signals generated by keyboardinterruptions etc.
    """
    if signum == signal.SIGINT:
        msg = 'Terminal interrupt signal received: signum %s' % str(signum)
    elif signum == signal.SIGTERM:
        msg = 'Termination signal received: signum %s' % str(signum)
    elif signum == signal.SIGABRT:
        msg = 'Process abort signal received: signum %s' % str(signum)
    elif str.lower(sys.platform[:3])=='win':
        if signum == signal.SIGBREAK:
            msg = 'Process break signal received: signum %s' % str(signum)
        else:
            msg = 'Unknown signal received: signum %s' % str(signum)
    elif signum == signal.SIGILL:
        msg = 'Illegal instruction signal received: signum %s' % str(signum)
    else:
        msg = 'Unknown signal received: signum %s' % str(signum)
    return msg

def update_dates(dates, timestep, unitofstep, deduceaday=False):
    """
    Add the specified amount of time to the given dates and return the
    result
    """
    newdates = []
    for i, curdate in enumerate(dates):
        if isinstance(timestep, (int, float)):
            curdate = update_date(curdate, timestep, unitofstep, deduceaday)
        else:
            curdate = update_date(curdate, timestep[i], unitofstep, deduceaday)
        newdates.append(curdate)
    return numpy.array(newdates, dtype=datetime.date)

def update_date(curdate, timestep, unitofstep, deduceaday=False):
    """
    Add the specified amount of time to the current date and return the
    result
    """
    if unitofstep == 'year':
        curdate += relativedelta(years=int(timestep))
    elif unitofstep == 'month':
        curdate += relativedelta(months=int(timestep))
    elif unitofstep == 'day':
        curdate += relativedelta(days=int(timestep))
    if deduceaday:
        curdate = curdate - relativedelta(days=1)
    return curdate

def calculate_date_diff(start, end, unit):
    """
    Calculate the difference two dates in given units

    Parameters:
        start - datetime object
        end - datetime object
        unit - 'year', 'month', 'day'
    """
    start = datetime.date(start.year, start.month, start.day)
    end = datetime.date(end.year, end.month, end.day)
    delta = end - start
    if unit == 'year':
        # NB: we need to substract one day from the delta as we want to get
        # the difference in full years
        return int(math.floor((delta.days - 1.) / 365.))
    elif unit == 'month':
        return int(math.floor((delta.days - 1.) / (365. / 12.)))
    elif unit == 'day':
        raise Exception("Processing of other timesteps than 'year' not ready!")

def xcombinations(items, n):
    """
    Take n distinct elements from the sequence, order matters.
    By: Ulrich Hoffmann, taken from
    http://aspn.activestate.com/ASPN/Cookbook/Python/Recipe/190465

    Parameters:
        items - the sequence from which the combinations are generated
        n - the number of elements the new sequence should have

    Calls:

    Returns:
        a generator for the new sequence
    """
    if n==0: yield []
    else:
        for i in xrange(len(items)):
            for cc in xcombinations(items[:i]+items[i+1:],n-1):
                yield [items[i]]+cc

def str2date(text, sep='-', order='year first'):
    """
    Parse a date string from xml (type 'xs:date') or generic string
    to a datetime object.

    text -- date string
    sep -- separator used in the date string
    order -- y.m.d (year first) or d.m.y (day first)
    """
    try:
        if order=='year first':
            year, month, day = [int(x) for x in text.split(sep)]
        else:
            day, month, year = [int(x) for x in text.split(sep)]
        if year>1900:
            ret = datetime.date(year, month, day)
        else:
            ret = None
    except:
        return None
    return ret

def str2bool(text):
    """
    Function: str2bool

    Parse a boolean string from xml (type 'xs:boolean') to boolean value.

    Parameters:
        text - boolean string

    Returns:
        boolean value
    """
    if text.lower() == 'true':
        return True
    return False

def secs2str(seconds):
    """Convert seconds into hh:mm:ss string

    seconds -- seconds as a float or integer
    """
    h = int(seconds // 3600)
    m = int((seconds % 3600) // 60)
    s = seconds % 60
    return "%02.0f:%02.0f:%03.3f" % (h,m,s)

def calculate_NPV(value, discountrate, presentdate, eventdate):
    """Calculate the discounted NPV (Net Present Value) for a single value

    NB: what if discounting period includes months?
    """
    delta = eventdate - presentdate
    years = delta.days / 365.
    return value / (1. + (discountrate / 100.)) ** years


def main():
    pass

if __name__ == '__main__':
    main()
