#!/usr/bin/env python
# encoding: utf-8
"""
Title: ChainEvaluator

Creates dot-files of all model chain xml files in the given directory. 

*chainevaluator.py*

Usage: chainevaluator.py model_chain_directory output_directory

Copyright (c) 2007 Simosol Oy. Distributed under the GNU General Public License 2.0.
"""
__revision__ = "$Revision: Alpha 0.1 $"

import os
import sys
import re
import xml.etree.cElementTree as ET


class ModelChain:
    """ Class: ModelChain """      
    __outformats = ['text', 'dot']
    
    ########################################################
    ## General methods
    ########################################################
    def __init__(self, name, outpath):
        """ initialize modelchain structure """  
        self.__document_name = name
        self.__pdir = outpath
        self.__chain_structure = []      
        self.__chain_groups = 0
        self.__model_chains = 0
        self.__task_levels = 0
        self.__branching_groups = 0
        self.__branch_tasks = 0
        self.__tasks = 0
        self.__conditions = 0
        self.__expressions = 0
        self.__models = {}
        
    def __reset_variables(self):
        """ reset all variables """
        self.__chain_structure = []      
        self.__chain_groups = 0
        self.__model_chains = 0
        self.__task_levels = 0
        self.__branching_groups = 0
        self.__branch_tasks = 0
        self.__tasks = 0
        self.__conditions = 0
        self.__expressions = 0
        self.__models = {} 
    
    def __elem_name(self, elem):
        """ get ElementTree element name without namespace """
        return elem.tag.split('}')[1]
    
    ########################################################
    ##  Methods for constructing & modifying chain structure
    ########################################################
    def __add_chain_group(self, name, igroup):
        """ add a chain_group to chain structure """        
        self.__chain_structure.insert(igroup, [name, []])
        self.__chain_groups += 1
        
    def __add_model_chain(self, name, evaluate_at, igroup, ichain):
        """ add a model_chain to chain group """        
        self.__chain_structure[igroup][1].insert(ichain, [name, evaluate_at, []])
        self.__model_chains += 1
        
    def __add_task(self, igroup, ichain, idepth, itask, name=None, type=None, 
                      branchgroup=None, iparent=None):
        """ add a task to a model_chain to given depth """
        task = [name, type, None, None, None, branchgroup]         
        self.__check_task_depth(igroup, ichain, idepth)
        self.__chain_structure[igroup][1][ichain][2][idepth].insert(itask, task)
        if iparent is not None:
            self.__update_task_children(igroup, ichain, idepth, itask, iparent)
        self.__tasks += 1
        
    def __set_task_name(self, igroup, ichain, idepth, itask, name):
        """ set the name of a task """
        self.__get_task(igroup, ichain, idepth, itask)[0] = name
        
    def __set_task_type(self, igroup, ichain, idepth, itask, type):
        """ set the type of a task """
        self.__get_task(igroup, ichain, idepth, itask)[1] = type
        
    def __set_task_condition(self, igroup, ichain, idepth, itask, condition):
        """ set the condition of a task """
        self.__get_task(igroup, ichain, idepth, itask)[2] = condition
        self.__conditions += 1
        
    def __update_task_children(self, igroup, ichain, idepth, itask, iparent):
        """ update the childlist of a task """
        parent = self.__get_task(igroup, ichain, idepth-1, iparent)
        if parent[3] == None: parent[3] = []
        parent[3].append(itask)
        
    def __set_task_model(self, igroup, ichain, idepth, itask, model):
        """ set the branching_group of a task """
        type = model[0]
        self.__get_task(igroup, ichain, idepth, itask)[4] = model
        if not type in self.__models: self.__models[type] = 0
        self.__models[type] += 1
        
    def __set_task_branchgroup(self, igroup, ichain, idepth, itask, branchgroup):
        """ set the branching_group of a task """
        self.__get_task(igroup, ichain, idepth, itask)[1] = type
        
    def __check_task_depth(self, igroup, ichain, idepth):
        """ check that task depth index is deep enough for new task """
        chain = self.__get_model_chain(igroup, ichain)
        if len(chain[2]) == idepth: 
            chain[2].append([])
            if self.__task_levels <= idepth: self.__task_levels += 1
    
    ########################################################
    ##  Methods for accessing different parts of chains 
    ##  structure
    ########################################################
    def __get_task(self, igroup, ichain, idepth, itask):
        """ get a task structure """
        return self.__chain_structure[igroup][1][ichain][2][idepth][itask]
    
    def __get_model_chains(self, igroup):
        """ get a task structure """
        return self.__chain_structure[igroup][1]
    
    def __get_model_chain(self, igroup, ichain):
        """ get a task structure """
        return self.__chain_structure[igroup][1][ichain]
        
    def __get_model_chain_name(self, igroup, ichain):
        """ get name of a chain group """
        return self.__get_model_chain(igroup, ichain)[0]
    
    def __get_model_chain_level(self, igroup, ichain):
        """ get name of a chain group """
        return self.__get_model_chain(igroup, ichain)[1]
    
    def __get_chain_groups(self):
        """ get a task structure """
        return self.__chain_structure
    
    def __get_chain_group(self, igroup):
        """ get a task structure """
        return self.__chain_structure[igroup]    
       
    def __get_chain_group_name(self, igroup):
        """ get name of a chain group """
        return self.__get_chain_group(igroup)[0]

    
    ########################################################
    ##  Methods for accessing & modifying location indicator
    ########################################################
    def __set_location(self, igroup=0, ichain=0, idepth=0, itask=0):
        """ set the current location """
        self.__igroup = igroup
        self.__ichain = ichain
        self.__idepth = idepth        
        self.__itask = [itask]  # NB! possible errors here! 
        
    def __reset_location(self):
        """ delete location indicators """
        self.__igroup = 0
        self.__ichain = 0
        self.__idepth = 0
        self.__itask = [0]
        
    def __get_location(self):
        """ get current location (task) """        
        return self.__igroup, self.__ichain, self.__idepth, self.__itask[self.__idepth]
    
    ########################################################
    ##  Methods for parsing model_chain xml into object
    ########################################################
    def __parse_xml_file(self, root):
        """ parse model_chain xml into ModelChain object """        
        for chain_group in root:                        
            self.__add_chain_group(chain_group.attrib.get('name'), self.__igroup)           
            self.__parse_chain_group(chain_group)            
        
    def __parse_chain_group(self, chain_group):
        """ parse single chain_group element"""
        self.__ichain = 0
        for model_chain in chain_group:
            name = model_chain.attrib.get('name')
            evaluate_at = model_chain.attrib.get('evaluate_at')
            self.__add_model_chain(name, evaluate_at, self.__igroup, self.__ichain)
            self.__parse_model_chain(model_chain)
        self.__igroup += 1
        
    def __parse_model_chain(self, model_chain):
        """ parse single model_chain element """
        self.__itask = [0]
        for elem in model_chain:
            type = self.__elem_name(elem)
            if type == 'branching_group': self.__parse_branching_group(elem)
            elif type == 'task': self.__parse_task(elem, 'task')
        self.__ichain += 1
        
    def __parse_branching_group(self, elem):
        """ parse a single branching_group element """
        self.__branching_groups += 1
        for branch_task in elem:
            self.__parse_branch_task(elem)
            
    def __parse_branch_task(self, elem, branchgroup=None):
        """ parse a single branch_task element """        
        self.__parse_task(elem, 'branch_task', None, branchgroup)
        self.__branch_tasks += 1
        
    def __parse_task(self, elem, type, iparent=None, branchgroup=None):
        """ parse a single task element """
        itask = self.__itask[self.__idepth]
        name = elem.attrib.get('name')
        self.__add_task(self.__igroup, self.__ichain, self.__idepth, itask, 
                           name, type, branchgroup, iparent)  
        if self.__elem_name(elem[0]) == 'condition':
            self.__parse_condition(elem[0])
            if self.__elem_name(elem[1]) == 'model': 
                self.__parse_model(elem[1])
            else: self.__parse_subtasks(elem, 1, itask)
        elif self.__elem_name(elem[0]) == 'model':
            self.__parse_model(elem[0])
        else: self.__parse_subtasks(elem, 0, itask)           
        self.__itask[self.__idepth] += 1
        
    def __parse_subtasks(self, elem, fromelem, iparent):
        """ parse subtask """
        self.__idepth += 1
        if len(self.__itask) <= self.__idepth: self.__itask.append(0)            
        for subelem in elem[fromelem:]: 
            self.__parse_task(subelem, 'task', iparent)            
        self.__idepth -= 1        
    
    def __parse_condition(self, elem):
        """ parse condition element """        
        condition = self.__parse_expression(elem[0])
        self.__set_task_condition(self.__igroup, self.__ichain, self.__idepth, 
                                  self.__itask[self.__idepth], condition)
    
    def __parse_expression(self, elem):
        """ parse expression element """        
        type = self.__elem_name(elem[0])
        cond = [elem.get('operator'), type, [], None]        
        if type == 'expression':
            for subelem in elem: cond[2].append(self.__parse_expression(subelem))
        elif type == 'variable' or type == 'operation':        
            cond[3] = [(str(elem[0][0].text), str(elem[0][1].text)), None]
            cond[3][1] = self.__parse_condition_value(elem[1])     
        elif type == 'level':
            cond[3] = (str(elem[0].text), str(elem[1].text))       
        return cond
    
    def __parse_condition_value(self, elem):
        """ parse condition value from xml """
        type = self.__elem_name(elem)
        cond = [type, None]
        if type == 'value': cond[1] = float(elem.text)
        elif type == 'value_variable': cond[1] = [str(elem[0].text), str(elem[1].text)]
        elif type == 'enum': cond[1] = [float(val) for val in elem.text.split(' ')]        
        return cond
    
    def __parse_model(self, elem):
        """ parse model element """
        name = elem[0].text
        mtype = self.__elem_name(elem[1])        
        if mtype == 'prediction': param = self.__parse_prediction_model(elem)        
        elif mtype == 'aggregation': param = self.__parse_aggregation_model(elem)        
        elif mtype == 'operation': param = self.__parse_operation_model(elem)         
        elif mtype == 'management': param = self.__parse_management_model(elem)        
        modeltuple = (mtype, name, param)
        self.__set_task_model(self.__igroup, self.__ichain, self.__idepth,
                              self.__itask[self.__idepth], modeltuple)        
        
    def __parse_prediction_model(self, elem):
        """ parse prediction model parameters """
        modelparam = [{}, [], []]
        for subelem in elem[1]:
            if self.__elem_name(subelem) == 'parameters':
                for parameter in subelem:
                    modelparam[0][parameter[0].text] = float(parameter[1].text)
            if self.__elem_name(subelem) == 'rect_factor':
                for factor in subelem: modelparam[1].append(factor.text)
            elif self.__elem_name(subelem) == 'risklevels':
                for level in subelem: modelparam[2].append(level.text)
        return modelparam
    
    def __parse_operation_model(self, elem):
        """ parse operation model parameters """
        modelparam = [{}, [], None, None, []] 
        for subelem in elem[1]:
            type = self.__elem_name(subelem)
            if type == 'parameters':
                for parameter in subelem:                    
                    modelparam[0][parameter[0].text] = parameter[1].text 
            elif type == 'long_term':
                for vareffect in subelem:
                    level = vareffect[0].text
                    variable = vareffect[1].text
                    effects = []
                    for effect in vareffect[2]:
                        effects.append([[int(effect[0][0].text), 
                                        int(effect[0][1].text)],
                                        float(effect[1].text)])
                    modelparam[1].append([level, variable, effects])
            elif type == 'cash_flow_table': modelparam[2]=subelem.text
            elif type == 'cash_flow_model': modelparam[3]=subelem.text
            elif type == 'simulation_effects':
                modelparam[4]=self.__parse_simulation_effects(subelem)
        return modelparam
    
    def __parse_simulation_effects(self, elem):
        """ parse operation model simulation effects """
        effects = [False, None]        
        if elem[0].text == 'true': effects[0] = True       
        if len(elem)>1:            
            effects[1] = [elem[1][0].text, 
                          [elem[1][1][0].text, elem[1][1][1].text]]
        return effects
    
    def __parse_aggregation_model(self, elem):
        """ parse aggregation model parameters """
        modelparam = [elem[1][0].text, None, None, [], None] 
        for subelem in elem[1]:
            tag = self.__elem_name(subelem)
            if tag == 'level': modelparam[1] = subelem.text
            elif tag == 'through_level': modelparam[2] = subelem.text
            elif tag == 'operands':
                for operand in subelem:                    
                    if self.__elem_name(operand[0]) == 'variable':
                        oper = [None, None, None, None, None]    
                        for operelem in operand:
                            ename = self.__elem_name(operelem)
                            if ename == 'variable': oper[0] = operelem.text
                            elif ename == 'weight': oper[1] = operelem.text
                            elif ename == 'level': oper[2] = operelem.text                      
                            elif ename == 'forced': oper[3] = operelem.text
                    elif self.__elem_name(operand[0]) == 'value':
                        oper = [float(operand[0].text)]
                    modelparam[3].append(oper)
            elif self.__elem_name(subelem) == 'condition':
                modelparam[4] = self.__parse_expression(subelem[0])           
        return modelparam
    
    def __parse_management_model(self, elem):
        """ parse management model parameters """
        parameters = []
        if len(elem[1]) == 0: return None  
        for parameter in elem[1][0]:
            if self.__elem_name(parameter) == 'parameter':
                parameters.append(parameter[1].text)
        return parameters
    
    ########################################################
    ##  Methods for outputting model_chain structure
    ########################################################
    def __output_chain_groups(self):
        """ output chain groups into a file """
        for group in self.__get_chain_groups():
            name = self.__get_chain_group_name(self.__igroup)
            if self.__output_type == 'text':
                self.__output('chain_group', '%s' % (name))
            self.__output_model_chains()
            self.__igroup += 1
    
    def __create_dot_chainfile(self, name):
        """
        Method: __create_dot_chainfile        
        """
        name = name.replace(' ', '_')  # replace whitespace, commas etc.
        name = name.replace(',', '_')
        name = name.replace('_', '_')
        path = os.path.join(os.path.abspath(os.curdir), '%s'%name)
        self.__cur_task_id = []
        self.dot_links = []
        self.__outstream = open('%s.dot' % name, 'w')
        self.__write_dot_hdr(name)
        self.__pngfile.write('dot -Tpng %s.dot -o %s.png\n' % (path, path))
        self.__psfile.write('dot -Tps %s.dot -o %s.ps\n' % (path, path))        
    
    def __output_model_chains(self):
        """
        Method: __output_model_chains
        
        output model chain elements
        """
        self.__ichain = 0
        for chain in self.__get_model_chains(self.__igroup):
            name = self.__get_model_chain_name(self.__igroup, self.__ichain)            
            if self.__output_type == 'dot':
                self.__create_dot_chainfile(name)                
            level = self.__get_model_chain_level(self.__igroup, self.__ichain)
            self.__output('model_chain', (name, level), 1)
            self.__output_first_level_tasks(chain)
            self.__ichain += 1
            if self.__output_type == 'dot':
                self.__write_dot_links()
                self.__outstream.close()
    
    def __output_first_level_tasks(self, chain):
        """ print tasks, start from level 0 and move deeper if necessary """       
        for task in chain[2][0]:
            self.__output_task(task)
            if task[3] is not None: self.__output_subtasks(task[3])                
            self.__itask[self.__idepth] += 1
        
    def __output_task(self, task):
        """ output a single task structure """
        self.__output(task[1], '%s' % (task[0]), 2)
        if task[2] is not None: self.__output_condition(task[2])
        if task[4] is not None: self.__output_model(task[4])
        
    def __output_subtasks(self, subtasks):
        """ go through subtasks and print them or more deeper in hierarchy """
        self.__idepth += 1
        if len(self.__itask) == self.__idepth: self.__itask.append(0)
        for itask in subtasks:
            task = self.__get_task(self.__igroup, self.__ichain, self.__idepth, itask)
            self.__output_task(task)
            if task[3] is not None: self.__output_subtasks(task[3])
            self.__itask[self.__idepth] += 1
        self.__idepth -= 1
        
    def __output_condition(self, condition):
        """ Method: __output_condition """
        condstr = self.__linearize_condition(condition)
        self.__output('condition', condstr, 3)
        
    def __linearize_condition(self, condition):
        """
        Method: __linearize_condition
        """
        stack = False
        oper = condition[0]
        type = condition[1]
        if type == 'variable':
            t1 = condition[3][0][0]
            t2 = condition[3][1][1]
            if oper == 'included': oper += ' in'
            elif oper == 'exists': t2 = bool(t2)
            if condition[3][1][0] == 'value_variable':
                t2 = condition[3][1][1][0]            
        elif type == 'expression':
            stack = True
            t1 = self.__linearize_condition(condition[2][0])
            t2 = self.__linearize_condition(condition[2][1])            
        elif type == 'operation':
            t1 = condition[3][0][0]
            t2 = condition[3][1][1]            
        elif type == 'level':
            t1 = condition[3][0]
            t2 = bool(condition[3][1])
        if stack: condstr = '%s\\n%s\\n%s' % (t1, oper, t2)
        else: condstr = '%s %s %s' % (t1, oper, t2)
        return condstr
    
    def __output_model(self, model):
        """
        Method: __output_model
        """        
        self.__output('model', model, 3)        
        
    def __output(self, elem_type, elem_content, indent=0):
        """
        Method: __output
        """
        indentstr = '\n' + indent * '\t' + self.__idepth * '\t'
        if self.__output_type == 'text':
            outstr = elem_type+': '+elem_content            
        elif self.__output_type == 'dot':            
            outstr = self.__construct_dot_string(elem_type, elem_content, indent)
        self.__outstream.write(indentstr+outstr)
        
    def __create_dot_label(self, type, content):
        """
        Method: __create_dot_label
        """
        if type == 'model':
            mtype, mname, param = content
            label = '%s model\\n%s' % (mtype, mname)            
        elif type == 'condition':
            label = content
        elif type == 'model_chain':
            label = '%s\\nlevel: %s\\n%s' % (type, content[1], content[0])
        else:
            c = content.split(' ')
            text = ''
            lrow = 0
            for w in c:
                lrow += len(w)
                text += '%s ' % w
                if lrow > 15:
                    text = text.rstrip()
                    text += '\\n'
                    lrow = 0
            text = text.rstrip('\\n')
            text = text.rstrip()
            label = type.upper() + '\\n' + text
        return label
        
    def __construct_dot_string(self, elem_type, elem_content, indent):
        """
        Method: __construct_dot_string
        """        
        pid = None
        shape = 'record'
        fontsize = 12
        nodeid = 'node'+str(self.__igroup)  # current node id
        instr = (indent+2) * '\t'
        if elem_type == 'chain_group':            
            self.__cur_group_id = nodeid
        elif elem_type == 'model_chain':            
            nodeid += '_%s' % self.__ichain
            self.__cur_chain_id = nodeid            
        else:  # type is task, model or condition
            nodeid += '_%s_%s_%s' % \
                 (self.__ichain, self.__idepth, self.__itask[self.__idepth])
            if self.__idepth >= len(self.__cur_task_id):
                self.__cur_task_id.append(None)            
            if elem_type in ['condition', 'model']:                
                if elem_type == 'condition':                    
                    shape = 'plaintext'
                    fontsize = 10
                    i = 0
                elif elem_type == 'model':
                    fontsize = 10
                    i = 1                
                nodeid += '_%s' % i
                pid = self.__cur_task_id[self.__idepth]                               
            else:
                if self.__idepth == 0:
                    pid = self.__cur_chain_id
                else:
                    pid = self.__cur_task_id[self.__idepth-1]
            self.__cur_task_id[self.__idepth] = nodeid 
        label = self.__create_dot_label(elem_type, elem_content)
        if pid is not None: self.dot_links.append((pid, nodeid))
        nodeid += ' [shape=%s,fontsize=%s,nodesep=0.01,label="%s"]' \
               % (shape, fontsize, label)                     
        return nodeid
    
    def __write_dot_hdr(self, name):
        """
        Method: __write_dot_hdr
        """
        self.__outstream.write('digraph %s {\n\tcompound=true;\n' \
                               '\tlabeljust = l;\n\trankdir=LR;' % name[0])
        
    def __write_dot_links(self):
        """
        Method: __write_dot_links
        """
        for link in self.dot_links:
            self.__outstream.write('\n'+'%s -> %s;' % (link[0], link[1]))
        self.__outstream.write('\n}')
        
    ########################################################
    ##  Public functions for manipulating ModelChain instance
    ########################################################
    def parse_model_chain_xml(self, xmlfile):
        """ parse a model_chain xml file to ModelChain object """
        self.__set_location()
        self.__parse_xml_file(ET.ElementTree(None, file=xmlfile).getroot())
        self.__reset_location()
        
    def get_chain_structure(self):
        """ return the model chain structure """
        return self.__chain_structure
    
    def get_modelchain_description(self):
        """ get the general description of ModelChain object """
        desc = {'name': self.__document_name,
                'chain_group': self.__chain_groups,
                'model_chain': self.__model_chains,
                'task level': self.__task_levels,
                'branching_group': self.__branching_groups,
                'branch_task': self.__branch_tasks,
                'task': self.__tasks,
                'condition': self.__conditions,
                'model': self.__models}
        return desc
 
    def output_chain_structure_as_dot(self):
        """ output chain structure into file """
        outdir = os.path.join(self.__pdir, self.__document_name)
        if not os.path.isdir(outdir):
            os.mkdir(outdir)
        os.chdir(outdir)
        self.__output_type = 'dot'
        pngpath = os.path.join(self.__pdir, 'createGraphsPNG.bat')
        pspath = os.path.join(self.__pdir, 'createGraphsPS.bat')
        self.__pngfile = open(pngpath, 'a')
        self.__psfile = open(pspath, 'a')
        self.__set_location()
        self.__output_chain_groups()
        self.__pngfile.close()
        self.__psfile.close()
        self.__reset_location()
        os.chdir(self.__pdir)        


def main():
    chainpath = sys.argv[1]
    chainpath = os.path.abspath(chainpath)
    if len(sys.argv) == 3:
        outpath = os.path.abspath(sys.argv[2])
    else:
        outpath = os.path.abspath(os.curdir)  
    files = os.listdir(chainpath)
    test = re.compile("\.xml$", re.IGNORECASE)  # check that the file name ends in ".xml"    
    xmlfiles = filter(test.search, files)
    for xmlfile in xmlfiles:
        docname = xmlfile.split('.')[0]    
        modelchain = ModelChain(docname, outpath)
        modelchain.parse_model_chain_xml(os.path.join(chainpath, xmlfile))    
        modelchain.output_chain_structure_as_dot()
    
if __name__ == '__main__':
    main()
    



